/// Careful Triangles and Artful Graphics
/// Animations
/// Copyright (c) 2018 by Howard I. Cannon - MIT License

"use strict";

/// Class to run an animation
function Animator($scope, drawer, params) {
  this.$scope=$scope;
  this.drawer=drawer;
  // Animation options
  this.isAnimationQueued=false;
  this.shouldStop=false;
  this.direction="forward";
  this.stepsPerCycle=parseInt(defarg(params.stepsPerCycle_, "1"));
  this.lastTime=new Date(0);
  this.delayMS=parseInt(defarg(params.delayMS_, "0"));
  this.clearValue=0;
  this.stepFn=function() { };
  this.stepCount=0;
  return;
}

jQuery.extend(Animator.prototype, {
  fillQueryParams: function(params) {
    params.stepsPerCycle_=this.stepsPerCycle;
    params.delayMS_=this.delayMS;
  },

  setRunState: function(runstate) {
    this.shouldStop=!runstate;
    if (!this.isAnimationQueued && runstate) {
      // Not already running and wanting to, start up
      requestAnimationFrame(this.animationStep.bind(this));
      this.isAnimationQueued=true;
    };
    this.lastTime=new Date(0);
    return runstate;
  },

  setDirection: function(dir) {
    if (dir===false) dir="backward";
    else if (dir===true) dir="forward";
    this.lastTime=new Date(0);
    this.direction=dir;
    return dir;
  },

  isRunning: function() {return this.isAnimationQueued && !this.shouldStop;},

  run: function(dir) {
    if (supplied(dir)) this.setDirection(dir);
    this.setRunState(true);
    return dir;
  },

  pause: function() {
    this.setRunState(false);
    return;
  },

  step: function(dir) {
    this.pause();
    if (supplied(dir)) this.setDirection(dir);
    var steps=this.stepsPerCycle;
    for (; steps>0; steps--) this.stepFn(this.direction);
    this.stepCount+=(this.direction=="forward" ? 1 : -1)*this.stepsPerCycle;
    this.update();
    return;
  },

  clear: function(reset) {
    this.drawer.fill(this.clearValue);
    if (reset) this.reset();
    if (this.clearFn) this.clearFn(reset);
    this.update();
    return;
  },

  reset: function() {
    this.stepCount=0;
    return;
  },

  update: function() {
    this.drawer.update();
    return;
  },

  animationStep: function(timestamp) {
    this.isAnimationQueued=false;
    if (this.shouldStop) {
      this.shouldStop=false;
      return;
    };
    // Requeue the animation frame regardless
    requestAnimationFrame(this.animationStep.bind(this));
    this.isAnimationQueued=true;
    // Check to see if enough time has elapsed
    var now=new Date();
    if ((this.lastTime.getTime()+this.delayMS)>now.getTime()) return;
    this.lastTime=now;
    // Do next animation step
    var steps=this.stepsPerCycle;
    for (; steps>0; steps--) this.stepFn(this.direction);
    this.$scope.$apply((function() {
      this.stepCount+=this.stepsPerCycle*(this.direction=="forward" ? 1 : -1);
    }).bind(this));
    this.update();
    return;
  },

  setupMidi: function(midi) {
    // F1 does a reset
    var animator=this;
    midi.onCommand(54, "noteon", function(data) {
      if (data==127) animator.clear(true);
      return;
    });
    // Rew, FF, pause, play
    midi.onCommand(91, "noteon", function(data) {
      if (data==127) animator.run(false);
      return;
    });
    midi.onCommand(92, "noteon", function(data) {
      if (data==127) animator.run(true);
      return;
    });
    midi.onCommand(93, "noteon", function(data) {
      if (data==127) animator.pause();
      return;
    });
    midi.onCommand(94, "noteon", function(data) {
      if (data==127) animator.step();
      return;
    });
    return;
  },

  end: ""
});

/// The animations
function AnimBase($scope, ics, animator) {
  this.$scope=$scope;
  this.animator=animator;
  this.drawer=animator.drawer;
  ics.setupControls(this.controlsKey, animator, this.calcParams.bind(this), this.clear.bind(this),
                    this.onchange.bind(this), this.init.bind(this), this.uispec());
  // Default pixel value increment to 1
  $scope.pixelInc=1;
  return;
}

jQuery.extend(AnimBase.prototype, {
  init: function() { return; },

  calcParams: function(reset) {
    this.calcCenter();
    return reset;
  },

  onchange: function(param, reset) { return reset; },

  clear: function(reset) {
    // When reset, recalculate all parameters
    reset=this.calcParams(reset);
    if (reset) this.init();
    return;
  },

  uispec: function() { return []; },

  findUiSpec: function(specs, name) {
    for (var i=0; i<specs.length; i++) {
      if (specs[i].name==name) return specs[i];
    };
    return null;
  },

  calcCenter: function() {
    // Default is to always center in drawer viewport
    this.$scope.centerX=zround(this.drawer.width, 2);
    this.$scope.centerY=zround(this.drawer.height, 2);
    return;
  },

  end: ""
});


/// Radius support
function AnimRadiusBase($scope, ics, animator) {
  AnimBase.call(this, $scope, ics, animator);
  return;
}

jQuery.extend(AnimRadiusBase.prototype, AnimBase.prototype, {
  radiusBase: function() {
    var dw=this.drawer.width, dh=this.drawer.height;
    return Math.min(dw, dh)/2;
  },

  calcParams: function(reset) {
    reset=AnimBase.prototype.calcParams.call(this, reset);
    var $scope=this.$scope;
    // Recalculate positions based on percentages
    var rbase=this.radiusBase();
    $scope.radius=Math.floor(rbase*$scope.radiusPct/100);
    return reset;
  },

  controlsKey: "radius",
  uispec: function() {
    var uis=[{name: "radiusPct", label: "Radius %:", default: 90, type: "slider", reset: true,
              extras: {from: 1, to: 100, step: 1}, midi: ["bend", 224, 1, 100]}
            ];
    return AnimBase.prototype.uispec.call(this).concat(uis);
  },

  end: ""
});


function AnimInnerRadiusBase($scope, ics, animator) {
  AnimRadiusBase.call(this, $scope, ics, animator);
  this.innerRadiusOld=-1;
  return;
}

jQuery.extend(AnimInnerRadiusBase.prototype, AnimRadiusBase.prototype, {
  drawInnerCircle: function() {
    var $scope=this.$scope;
    if ($scope.innerRadius>0) {
      this.drawer.drawFilledCircle($scope.centerX, $scope.centerY, $scope.innerRadius,
                                   ["noop", 0, "set", 0], undefined, Circle);
    };
    return;
  },

  init: function() {
    AnimRadiusBase.prototype.init.call(this);
    this.drawInnerCircle();
    return;
  },
  
  calcParams: function(reset) {
    reset=AnimRadiusBase.prototype.calcParams.call(this, reset);
    var $scope=this.$scope;
    $scope.innerRadius=Math.floor(this.radiusBase()*$scope.innerRadiusPct/100);
    return reset;
  },

  onchange: function(param, reset) {
    reset=AnimRadiusBase.prototype.onchange.call(this, param, reset);
    if (param=="innerRadiusPct") {
      // This requires some extra work
      var $scope=this.$scope;
      if (this.innerRadiusOld>$scope.innerRadius) {
        // Making the blanking circle smaller, erase the old one first
        this.drawer.drawFilledCircle($scope.centerX, $scope.centerY,
                                     this.innerRadiusOld, ["noop", 0, "set", 255], undefined, Circle);
      };
      this.drawInnerCircle();
      this.innerRadiusOld=$scope.innerRadius;
    };
    return reset;
  },

  clear: function(reset) {
    AnimRadiusBase.prototype.clear.call(this, reset);
    // Then, redraw inner circle
    this.drawInnerCircle();
    return;
  },

  controlsKey: "innerRadius",
  uispec: function() {
    var uis=[{name: "innerRadiusPct", label: "Inner Radius %:", default: 0, type: "slider", reset: false,
              extras: {from: 0, to: 100, step: 1}, midi: ["bend", 225, 0, 100]}
            ];
    return AnimRadiusBase.prototype.uispec.call(this).concat(uis);
  },

  end: ""
});


/// Basic circle animation
function CircleAnim($scope, ics, animator) {
  AnimInnerRadiusBase.call(this, $scope, ics, animator);
  return;
}

jQuery.extend(CircleAnim.prototype, AnimInnerRadiusBase.prototype, {
  init: function() {
    AnimInnerRadiusBase.prototype.init.call(this);
    var $scope=this.$scope;
    this.circle=CircleUtils.makeCircle({class: Circle,
                                        centerOf: $scope.drawer, radius: $scope.radius,
                                        period: $scope.period, increment: $scope.increment,
                                        cx: $scope.centerX, cy: $scope.centerY
                                       });
    this.animator.stepFn=function(command) {
      if (command=="setcenter") {
        this.circle.setCenter(arguments[1], arguments[2]);
      } else {
        var p0=this.circle.point();
        var p1=(command=="backward" ? this.circle.back() : this.circle.step());
        this.drawer.drawTriangleAddDet(p0[0], p0[1], p1[0], p1[1], $scope.fixedX, $scope.fixedY,
                                       $scope.pixelInc);
      };
      return;
    }.bind(this);
    return;
  },

  calcParams: function(reset) {
    reset=AnimInnerRadiusBase.prototype.calcParams.call(this, reset);
    var $scope=this.$scope;
    // Recalculate positions based on percentages
    var dw=this.drawer.width, dh=this.drawer.height;
    $scope.fixedX=Math.floor((dw/2)*(1+$scope.fixedXPct/100));
    $scope.fixedY=Math.floor((dh/2)*(1+$scope.fixedYPct/100));
    if (this.circle) {
      this.circle.setCenter($scope.centerX, $scope.centerY);
      this.circle.increment=$scope.increment;
    };
    return reset;
  },

  controlsKey: "circle",
  uispec: function() {
    var uis=[{name: "period", label: "Period:", default: 256, type: "slider", reset: true,
              extras: {from: 4, to: 1000, step: 1}, midi: ["bend", 226, 4, 1000]},
             {name: "increment", label: "Increment:", default: 1, type: "slider", reset: false,
              extras: {from: 1, to: 1000, step: 1}, midi: ["bend", 227, 1, 1000]},
             {name: "fixedXPct", label: "Fixed X %:", default: 0, type: "slider", reset: false,
              extras: {from: -100, to: 100, step: 1}, midi: ["bend", 228, -100, 100]},
             {name: "fixedYPct", label: "Fixed Y %:", default: 0, type: "slider", reset: false,
              extras: {from: -100, to: 100, step: 1}, midi: ["bend", 229, -100, 100]}
            ];
    return AnimInnerRadiusBase.prototype.uispec.call(this).concat(uis);
  },

  end: ""
});


/// Rotor-based animations
function RotorAnimBase() {
  AnimBase.apply(this, arguments);
  this.dtBase=Math.PI/256;
  return;
}

jQuery.extend(RotorAnimBase.prototype, AnimRadiusBase.prototype, {
  init: function() {
    AnimRadiusBase.prototype.init.call(this);
    CircleUtils.defaultCircleClassName="STCircle";
    this.rotorFn=this.makeRotorStepFn();
    this.rotorFn("sort", this.sortFn);
    this.animator.stepFn=this.rotorFn;
    return;
  },

  calcParams: function(reset) {
    reset=AnimRadiusBase.prototype.calcParams.call(this, reset);
    var $scope=this.$scope;
    $scope.dt=this.dtBase*$scope.dtPct/100;
    if (this.rotorFn) {
      this.rotorFn("setcenter", $scope.centerX, $scope.centerY);
    };
    var sfn=RotorUtils.rotorSortOriginalFn;
    switch ($scope.rotorSort) {
    case 2: sfn=RotorUtils.rotorSortBackwardsFn; break;
    case 3: sfn=RotorUtils.rotorSortRadiusUpFn; break;
    case 4: sfn=RotorUtils.rotorSortRadiusDownFn; break;
    case 5: sfn=RotorUtils.rotorSortSpeedUpFn; break;
    case 6: sfn=RotorUtils.rotorSortSpeedDownFn; break;
    };
    this.sortFn=sfn;
    return reset;
  },

  onchange: function(param, reset) {
    reset=AnimRadiusBase.prototype.onchange.call(this, param, reset);
    if (param=="rotorSort") {
      this.rotorFn("sort", this.sortFn);
    };
    return reset;
  },

  controlsKey: "rotor",
  uispec: function() {
    var uis=[{name: "nRotors", label: "Rotors:", default: 19, type: "slider", reset: true,
              extras: {from: 1, to: 100, step: 1}, midi: ["bend", 225, 4, 1000]},
             {name: "dtPct", label: "DTheta %:", default: 20, type: "slider", reset: true,
              extras: {from: 1, to: 100, step: 1}, midi: ["bend", 226, 1, 100]},
             {name: "rotorSort", label: "Sort by:", default: 1, type: "slider", reset: false,
              // extras: {from: 1, to: 7, modelLabels: {1: "Def", 2: "Bwds", 3: "Rad+", 4: "Rad-", 5: "Freq+", 6: "Freq-"}}
              extras: {from: 1, to: 7, step: 1,  scale: [{val: 1, label: "Default"},
                                                         {val: 2, label: "Backwards"},
                                                         {val: 3, label: "Radius Up"},
                                                         {val: 4, label: "Radius Down"},
                                                         {val: 5, label: "Freq Up"},
                                                         {val: 6, label: "Freq Down"}
                                                        ]}
             }
            ];
    return AnimRadiusBase.prototype.uispec.call(this).concat(uis);
  },

  end: ""
});

function BesserAnim() {
  RotorAnimBase.apply(this, arguments);
  return;
}

jQuery.extend(BesserAnim.prototype, RotorAnimBase.prototype, {
  makeRotorStepFn: function() {
    var $scope=this.$scope;
    return RotorAnimations.getBesserFn(this.drawer, $scope.radius, $scope.nRotors, $scope.dt, $scope.amp, 1);
  },

  controlsKey: "besser",
  uispec: function() {
    var uis=[{name: "amp", label: "Amplitude:", default: 4, type: "slider", reset: true,
              extras: {from: 0.5, to: 45.5, step: 0.1, round: 1}, midi: ["bend", 227, 1, 50]}
            ];
    var specs=RotorAnimBase.prototype.uispec.call(this).concat(uis);
    var srotors=this.findUiSpec(specs, "nRotors");
    var maxrs=150;
    srotors.extras.to=maxrs;
    srotors.default=maxrs>>1;
    return specs;
  },

  end: ""
});

function GibbsSquareAnim() {
  RotorAnimBase.apply(this, arguments);
  return;
}

jQuery.extend(GibbsSquareAnim.prototype, RotorAnimBase.prototype, {
  makeRotorStepFn: function() {
    var $scope=this.$scope;
    return RotorAnimations.getGibbsSquareFn(this.drawer, $scope.radius, $scope.nRotors, $scope.dt, 1);
  },

  controlsKey: "gibbsquare",

  end: ""
});

function GibbsTriAnim() {
  RotorAnimBase.apply(this, arguments);
  return;
}

jQuery.extend(GibbsTriAnim.prototype, RotorAnimBase.prototype, {
  makeRotorStepFn: function() {
    var $scope=this.$scope;
    return RotorAnimations.getGibbsTriFn(this.drawer, $scope.radius, $scope.nRotors, $scope.dt, 1);
  },

  controlsKey: "gibbstri",

  end: ""
});

function DragonAnim() {
  RotorAnimBase.apply(this, arguments);
  return;
}

jQuery.extend(DragonAnim.prototype, RotorAnimBase.prototype, {
  makeRotorStepFn: function() {
    var $scope=this.$scope;
    return RotorAnimations.getDragonFn(this.drawer, $scope.radius, $scope.nRotors, $scope.dt, 1);
  },

  controlsKey: "dragon",

  end: ""
});

function HilbertAnim() {
  RotorAnimBase.apply(this, arguments);
  return;
}

jQuery.extend(HilbertAnim.prototype, RotorAnimBase.prototype, {
  makeRotorStepFn: function() {
    var $scope=this.$scope;
    return RotorAnimations.getHilbertFn(this.drawer, $scope.radius, $scope.nRotors, $scope.dt, 1);
  },

  controlsKey: "hilbert",

  uispec: function() {
    var specs=RotorAnimBase.prototype.uispec.call(this);
    var spct=this.findUiSpec(specs, "dtPct");
    spct.default=30;
    var srotors=this.findUiSpec(specs, "nRotors");
    var maxrs=RotorAnimations.hilbertCoeffs.length;
    srotors.extras.to=maxrs;
    srotors.default=maxrs>>1;
    return specs;
  },

  end: ""
});

function Base2Anim() {
  RotorAnimBase.apply(this, arguments);
  return;
}

jQuery.extend(Base2Anim.prototype, RotorAnimBase.prototype, {
  makeRotorStepFn: function() {
    var $scope=this.$scope;
    return RotorAnimations.getBase2Fn(this.drawer, $scope.radius, $scope.nRotors, $scope.dt, 1);
  },

  controlsKey: "base2",

  uispec: function() {
    var specs=RotorAnimBase.prototype.uispec.call(this);
    var radpct=this.findUiSpec(specs, "radiusPct");
    radpct.default=50;
    var srotors=this.findUiSpec(specs, "nRotors");
    var maxrs=(RotorAnimations.base2Coeffs.length-1)/2;
    srotors.extras.to=maxrs;
    srotors.default=maxrs>>1;
    return specs;
  },

  end: ""
});

function CoeffsAnim($scope, ics, animator, specfnname) {
  this.spec=(eval(specfnname))();
  this.controlsKey=specfnname.replace(".", "_");
  RotorAnimBase.apply(this, arguments);
  return;
}

jQuery.extend(CoeffsAnim.prototype, RotorAnimBase.prototype, {
  makeRotorStepFn: function(specfn) {
    var $scope=this.$scope;
    var rotorfn=RotorUtils.getRotorFnFromCoeffs(this.drawer, $scope.radius*defarg(this.spec.size, 1.0), $scope.nRotors, this.spec.coeffs,
                                                $scope.dt, this.spec.freqfn, 1);
    // Reset center as specified
    var cx=zround(this.drawer.width, 2), cy=zround(this.drawer.height, 2);
    rotorfn("setcenter", cx*defarg(this.spec.xofffactor, 1.0), cy*defarg(this.spec.yofffactor, 1.0));
    return rotorfn;
  },

  uispec: function() {
    var specs=RotorAnimBase.prototype.uispec.call(this);
    var srotors=this.findUiSpec(specs, "nRotors");
    var maxrs=this.spec.coeffs.length;
    srotors.extras.to=maxrs;
    srotors.default=maxrs>>1;
    return specs;
  },

  end: ""
});


/// Spirograph-ish animation
function SpiroAnim() {
  AnimInnerRadiusBase.apply(this, arguments);
  this.dtBase=Math.PI/32;
  return;
}

jQuery.extend(SpiroAnim.prototype, AnimInnerRadiusBase.prototype, {
  init: function() {
    AnimInnerRadiusBase.prototype.init.call(this);
    var $scope=this.$scope;
    var randds=[];
    if ($scope.radius1!=0) randds.push([$scope.radius1, $scope.dt1, 0]);
    if ($scope.radius2!=0) randds.push([$scope.radius2, $scope.dt2, 0]);
    if ($scope.radius3!=0) randds.push([$scope.radius2, $scope.dt3, 0]);
    // Only color final circle
    randds[randds.length-1][2]=1;
    this.animator.stepFn=RotorAnimations.getSpiroFnFromList(this.drawer, randds, $scope.linewidth, Circle);
    return;
  },

  calcParams: function(reset) {
    reset=AnimInnerRadiusBase.prototype.calcParams.call(this, reset);
    var $scope=this.$scope;
    // Calculate radii based on overall radius
    $scope.radius1=Math.floor($scope.radius*$scope.radiusPct1/100);
    $scope.dt1=this.dtBase*$scope.dtPct1/100;
    $scope.radius2=Math.floor($scope.radius*$scope.radiusPct2/100);
    $scope.dt2=this.dtBase*$scope.dtPct2/100;
    $scope.radius3=Math.floor($scope.radius*$scope.radiusPct3/100);
    $scope.dt3=this.dtBase*$scope.dtPct3/100;
    return reset;
  },

  controlsKey: "spiro",
  uispec: function() {
    var uis=[{name: "radiusPct1", label: "Radius % 1:", default: 70, type: "slider", reset: true,
              extras: {from: 1, to: 100, step: 1}, midi: ["bend", 226, 1, 100]},
             {name: "dtPct1", label: "DTheta % 1:", default: 10, type: "slider", reset: true,
              extras: {from: 1, to: 100, step: 1}, midi: ["bend", 227, 1, 100]},
             {name: "radiusPct2", label: "Radius % 2:", default: 37, type: "slider", reset: true,
              extras: {from: 0, to: 100, step: 1}, midi: ["bend", 228, 0, 100]},
             {name: "dtPct2", label: "DTheta % 2:", default: 123, type: "slider", reset: true,
              extras: {from: 1, to: 200, step: 1}, midi: ["bend", 229, 1, 200]},
             {name: "radiusPct3", label: "Radius % 3:", default: 0, type: "slider", reset: true,
              extras: {from: 0, to: 100, step: 1}, midi: ["bend", 230, 0, 100]},
             {name: "dtPct3", label: "DTheta % 3:", default: 70, type: "slider", reset: true,
              extras: {from: 1, to: 200, step: 1}, midi: ["bend", 231, 1, 200]},
             {name: "linewidth", label: "Line width:", default: 0, type: "slider", reset: true,
              extras: {from: 0, to: 50, step: 1}, midi: ["bend", 232, 0, 100]}
            ];
    return AnimInnerRadiusBase.prototype.uispec.call(this).concat(uis);
  },

  end: ""
});
