/// Careful Triangles and Artful Graphics
/// Circles
/// Copyright (c) 2018 by Howard I. Cannon - MIT License

"use strict";

// Basic Minsky circle
var SimpleCircle=function(x, y, epsilon) {
  this.x=x;
  this.y=y;
  this.epsilon=epsilon;
}

jQuery.extend(SimpleCircle.prototype, {
  next: function() {
    this.x=this.x-this.y;
    this.y=this.y+this.epsilon*this.x;
    return [this.x, this.y];
  },

  back: function() {
    this.y=this.y-this.epsilon*this.x;
    this.x=this.x+this.y;
    return [this.x, this.y];
  },

  end: ""
})

var CircleUtils={
  makeSinWave: function(args) {
    var amplitude=defarg(args.amplitude, 1.0);
    var period=args.period;
    var phase=defarg(args.phase, 0.0);
    var dtheta=(period ? (Math.PI*2)/period : Math.PI/2);
    var sin=Math.sin(dtheta), sin2=2*Math.pow(Math.sin(dtheta/2), 2);
    var x=amplitude*Math.cos(phase),
        y=amplitude*Math.sin(phase);
    return new SimpleCircle(x, x*sin+y*sin2, sin2*-2);
  },

  makeCosWave: function(args) {
    var amplitude=defarg(args.amplitude, 1.0);
    var dtheta=defarg(args.dtheta, Math.PI/2.0);
    var phase=defarg(args.phase, 0.0);
    var sin=Math.sin(dtheta), sin2=2*Math.pow(Math.sin(dtheta/2), 2);
    var x=amplitude*Math.cos(phase),
        y=amplitude*Math.sin(phase);
    return new SimpleCircle(x, x*sin2-y*sin, sin2*2);
  },

  defaultCircleClassName: "STCircle",
  getCircleClass: function() {return eval(CircleUtils.defaultCircleClassName);},
  makeCircle: function(spec) {
    var circleclass=spec.class;
    if (!circleclass) circleclass=CircleUtils.getCircleClass();
    return new (circleclass)(spec);
  },

  end: ""
}

// Circle using two internal circle algorithms (per Lisp code)
// Args keys include: cx, cy, radius, phi, dt, period, centerOf (object with width and height), r
var ZCircle=function(args) {
  // Initialize center of circle
  if (supplied(args.centerOf)) this.setCenterFrom(args.centerOf);
  else this.setCenter(defarg(args.cx, 0), defarg(args.cy, 0));
  var phi=-defarg(args.phi, 0);
  var dt=defarg(args.dt, (Math.PI*2)/args.period);
  var sin=Math.sin(dt), sin2=2*Math.pow(Math.sin(dt/2), 2);
  var r=defarg(args.radius, 1);
  var x=args.x, y=args.y;
  if (supplied(x)) {
    if (!supplied(y)) y=Math.sqrt(r*r-x*x);
  } else if (supplied(y)) {
    x=Math.sqrt(r*r-y*y);
  } else {
    x=r*Math.cos(phi);
    y=r*Math.sin(phi);
  };
  // Create inner simple circles
  if (!supplied(x)) x=0;
  if (!supplied(y)) y=0;
  this.c1=new SimpleCircle(x, x*sin2-y*sin, 2*sin2);
  this.c2=new SimpleCircle(y, x*sin-y*sin2, 2*sin2);
  this.dt=dt;
  this.period=defarg(args.period, (dt!=0 ? Math.abs(zround(2*Math.PI, dt)) : 1.0));
  this.radius=Math.abs(r);
  this.forward=true;
  this.increment=defarg(args.increment, 1);
  return;
}

jQuery.extend(ZCircle.prototype, {
  x: function() { return this.cx+this.c1.x; },
  y: function() { return this.cy+this.c2.x; },

  reverse: function() { this.forward=!this.forward; },
  setForward: function(fwd) { return (this.forward=fwd); },
  setCenter: function(x, y) { this.cx=x; this.cy=y; },
  setCenterFrom: function(drawer) { return this.setCenter(zround(drawer.width, 2), zround(drawer.height, 2)); },

  next: function() {
    return [this.cx+this.c1.next()[0], this.cy+this.c2.next()[0]];
  },

  point: function() {
    return [this.cx+this.c1.x, this.cy+this.c2.x];
  },

  back: function() {
    return [this.cx+this.c1.back()[0], this.cy+this.c2.back()[0]];
  },

  step: function() {
    var point;
    for (var ni=0; ni<this.increment; ni++) point=(this.forward ? this.next() : this.back());
    return point;
  },

  end: ""
});

// Circle using sin/tan
// Args keys include: cx, cy, radius, phi, dt, period, centerOf (object with width and height), r
var STCircle=function(args) {
  // Initialize center of circle
  if (supplied(args.centerOf)) this.setCenterFrom(args.centerOf);
  else this.setCenter(defarg(args.cx, 0), defarg(args.cy, 0));
  var phi=-defarg(args.phi, 0);
  var dt=defarg(args.dt, (Math.PI*2)/args.period);
  var r=defarg(args.radius, 1);
  var x=args.x, y=args.y, rsign=sign(r);
  if (supplied(x)) {
    if (!supplied(y)) y=Math.sqrt(r*r-x*x)*rsign;
  } else if (supplied(y)) {
    x=Math.sqrt(r*r-y*y)*rsign;
  } else {
    x=r*Math.cos(phi);
    y=r*Math.sin(phi);
  };
  // If dt is near an odd multiple of pi, then cut in half and double increment
  // Otherwise, tangent misbehaves!
  var addtlincr=1;
  var pimultiple=zround(dt, Math.PI);
  if (Math.abs((dt/Math.PI)-pimultiple)<0.01 && isOdd(pimultiple)) {
    // dt near enough a an odd multiple of PI to cause trouble
    // Use half its value and do two steps per step
    addtlincr=2;
    dt/=2;
  };
  this.sin=Math.sin(-dt);
  this.tan=Math.tan(dt/2);
  this.x=x;
  this.y=y;
  this.dt=dt;
  this.period=defarg(args.period, (dt!=0 ? Math.abs(zround(2*Math.PI, dt*addtlincr)) : 1.0));
  this.radius=Math.abs(r);
  this.forward=true;
  this.increment=defarg(args.increment, 1)*addtlincr;
  return;
}

jQuery.extend(STCircle.prototype, {
  x: function() { return this.cx+this.x; },
  y: function() { return this.cy+this.x; },

  reverse: function() { this.forward=!this.forward; },
  setForward: function(fwd) { return (this.forward=fwd); },
  setCenter: function(x, y) { this.cx=x; this.cy=y; },
  setCenterFrom: function(drawer) { return this.setCenter(zround(drawer.width, 2), zround(drawer.height, 2)); },

  next: function() {
    // Three skews make up a step
    var x=this.x, y=this.y;
    x+=this.tan*y;
    y+=this.sin*x;
    x+=this.tan*y;
    this.x=x;
    this.y=y;
    return [this.cx+this.x, this.cy+this.y];
  },

  point: function() {
    return [this.cx+this.x, this.cy+this.y];
  },

  back: function() {
    // Reverse three skews to go backwards
    var x=this.x, y=this.y;
    x-=this.tan*y;
    y-=this.sin*x;
    x-=this.tan*y;
    this.x=x;
    this.y=y;
    return [this.cx+this.x, this.cy+this.y];
  },

  step: function() {
    var point;
    for (var ni=0; ni<this.increment; ni++) point=(this.forward ? this.next() : this.back());
    return point;
  },

  end: ""
});

/// Circle using trig and an integer index
// Args keys include: cx, cy, radius, phi, dt, period, centerOf (object with width and height), r
var Circle=function(args) {
  // Initialize center of circle
  if (supplied(args.centerOf)) this.setCenterFrom(args.centerOf);
  else this.setCenter(defarg(args.cx, 0), defarg(args.cy, 0));
  var phi=args.phi;
  var r=args.radius;
  var x=args.x, y=args.y;
  if (!supplied(r)) {
    // No r, must compute it from x and y, which must both be supplied
    r=Math.sqrt((x-this.cx)*(x-this.cx)+(y-this.cy)*(y-this.cy));
  } else if (supplied(x) && !supplied(y)) {
    y=Math.sqrt(r*r-x*x);
  } else if (!supplied(x) && supplied(y)) {
    x=Math.sqrt(r*r-y*y);
  } else if (!supplied(x) && !supplied(y)) {
    // Neither x nor y supplied
    x=this.cx+r;
    y=this.cy;
  };
  this.radius=Math.abs(r);
  // At this point we have consistent x, y and r
  if (!supplied(phi)) {
    phi=Math.atan2(y-this.cy, x-this.cx);
  } else {
    // phi goes counterclockwise
    phi=Math.PI*2-phi;
    if (r<0) {
      // Negative radius, rotate phi by half a circle
      if (phi>=Math.PI) phi-=Math.PI;
      else phi+=Math.PI;
    };
  };
  this.period=args.period;
  var dt=defarg(args.dt, 0);
  // Use large dt argument to indicate increment greater than 1
  this.increment=defarg(args.increment, Math.max(1, Math.ceil(dt/Math.PI)));
  // If both supplied, dt wins over period
  if (supplied(args.dt) || !supplied(args.period)) {
    // Cut down dt
    // Would like to find a way to emulate behavior of ZCircle, but
    // that benefits from being able to use an arbitrary epsilon
    // For now, just use excess over pi
    var maxdt=Math.PI/2;
    if (dt>maxdt) {
      var extra=Math.floor(dt/maxdt);
      dt-=extra*maxdt;
    } else if ((-dt)>maxdt) {
      var extra=Math.floor(-dt/maxdt);
      dt+=extra*maxdt;
    };
    this.period=(dt!=0 ? Math.abs(zround(2*Math.PI, dt)) : 1.0);
  } else {
    dt=(Math.PI*2)/args.period;
  };
  // Positive dt should go counterclockwise
  dt=-dt;
  // Having rounded period, need to adjust dt to be as exact as possible
  dt=(Math.PI*2)/this.period*sign(dt);
  this.dt=dt;
  // Index 0 is defined to be the point directly to the right of center (theta 0)
  // Compute starting index based on phi
  this.index=(dt!=0 ? zround(phi, dt)%this.period : 0);
  if (this.index<0) this.index+=this.period;
  this.forward=true;
  return;
}

jQuery.extend(Circle.prototype, {
  x: function() { return Math.cos(this.index*this.dt)*this.radius+this.cx; },
  y: function() { return Math.sin(this.index*this.dt)*this.radius+this.cy; },

  reverse: function() { return (this.forward=!this.forward); },
  setForward: function(fwd) { return (this.forward=fwd); },
  setCenter: function(x, y) { this.cx=x; this.cy=y; },
  setCenterFrom: function(drawer) { return this.setCenter(zround(drawer.width, 2), zround(drawer.height, 2)); },

  stepInternal: function(dir) {
    var idx=(this.index+this.increment*dir)%this.period;
    if (idx<0) idx+=this.period;
    this.index=idx;
    return this.point();
  },

  point: function() {
    var angle=this.index*this.dt;
    var cos=Math.cos(angle);
    var sin=Math.sin(angle);
    return [cos*this.radius+this.cx, sin*this.radius+this.cy];
  },

  next: function() {
    return this.stepInternal(1);
  },

  back: function() {
    return this.stepInternal(-1);
  },

  step: function() {
    return this.stepInternal(this.forward ? 1 : -1);
  },

  end: ""
});
