/// Careful Triangles and Artful Graphics
/// Core graphics
/// Copyright (c) 2018 by Howard I. Cannon - MIT License

"use strict";

var Graphics={

  // yslope solves for the equation x=my+b
  // We never have a horizontal line (y0=y1), so this will always be finite
  yslopeIntercept: function(x0, y0, x1, y1) {
    var m=(x0-x1)/(y0-y1);
    var b=x0-m*y0;
    return [m, b];
  },

  slopeIntercept: function(x0, y0, x1, y1) {
    var m=(y0-y1)/(x0-x1);
    var b=y0-m*x0;
    return [m, b];
  },

  triDeterminant: function(x, y, ax, ay, bx, by) {
    return (bx-ax)*(y-ay)-(by-ay)*(x-ax);
  },

  // Add delta to end of vector from p0 to p1
  extendPointAlongVector: function(p0, p1, delta) {
    // Relocate vector to the origin
    var x=p1[0]-p0[0], y=p1[1]-p0[1];
    var vlen=Math.sqrt(x*x+y*y);
    // Calculate offset vector
    x=(x/vlen)*delta;
    y=(y/vlen)*delta;
    // Then add it to the end of the original
    return [p1[0]+x, p1[1]+y];
  },

  drawRectangle: function(x0, y0, x1, y1, pixelfn) {
    var x0=Math.round(x0);
    var y0=Math.round(y0);
    var x1=Math.round(x1);
    var y1=Math.round(y1);
    var startx=Math.min(x0, x1);
    var endx=Math.max(x0, x1);
    var starty=Math.min(y0, y1);
    var endy=Math.max(y0, y1);
    for (var y=starty; y<endy; y++) {
      for (var x=startx; x<endx; x++) {
        pixelfn(x, y);
      };
    };
    return;
  },

  // This function draws a triangle whose coordinates comprise an
  // abstract grid between the pixels. The function chooses which pixels
  // belong to the interior of the triangle and calls the pixelfn to
  // perform whatever operation it wishes.
  // The guarantee is that triangles with shared edges will always hit
  // each pixel once and exactly once, so it is possible to reliably draw using xor.
  // The specfied "direction" of the triangle is ignored.
  drawTriangle: function(x0, y0, x1, y1, x2, y2, pixelfn) {
    var t, tx, ty;
    // sort points in increasing Y
    if (y1<y0) {tx=x0; ty=y0; x0=x1; y0=y1; x1=tx; y1=ty};
    if (y2<y0) {tx=x0; ty=y0; x0=x2; y0=y2; x2=tx; y2=ty};
    // Point 0 now has minimum y
    // Sort points 1 and 2
    if (y2<y1) {tx=x1; ty=y1; x1=x2; y1=y2; x2=tx; y2=ty};
    // Degenerate triangle hits no pixels
    // +++ Should have a more general check for colinearity?
    // +++ Consider fuzzy equal
    if ((x0===x1 && x1===x2) || (y0===y1 && y1===y2)) return;
    if ((x0===x1 && y0===y1) || (x2===x1 && y2==y1) || (x0===x2 && y0===y2)) {
      // Degenerate triangle with same two points, don't draw anything
      return;
    };
    // Now, mark all of the horizontal lines inside the triangle
    var lx0=x0, ly0=y0, lx1, ly1, rx0=x0, ry0=y0, rx1, ry1;
    if (y0!=y1) {
      // March from left to right
      // The tallest line, from point 0 to point 2 defines a half plane
      // Figure out whether the point 1 is in the "left" half or right half
      // You can't just compare the x coords -- draw a triangle
      // with the tallest line at 45 degrees and the third point near
      // the midpoint of that line and move the third point left and right to see why.
      var det=Graphics.triDeterminant(x1, y1, x2, y2, x0, y0);
      if (det<0) {lx1=x1; ly1=y1; rx1=x2; ry1=y2;}
      else {lx1=x2; ly1=y2; rx1=x1; ry1=y1;};
    } else {
      lx1=rx1=x2;
      ly1=ry1=y2;
      if (x0<x1) {rx0=x1; ry0=y1;}
      else {lx0=x1; ly0=y1;};
    };
    var line=Graphics.yslopeIntercept(lx0, ly0, lx1, ly1);
    var lm=line[0], lb=line[1];
    line=Graphics.yslopeIntercept(rx0, ry0, rx1, ry1);
    var rm=line[0], rb=line[1];
    // Draw horizontal lines until we run out of Y extent
    // on one or both sides
    var replaced=false;
    // Need to handle the first Y specially
    // The lines will all be scanned at pixel+0.5
    // Figure out which pixel to start with
    var testy=Math.floor(y0)+0.5;
      // If the starting point is beyond the halfway point, then actually start at the next pixel
    if (y0>testy) testy++;
    for (; testy<y2; testy++) {
      if (!replaced && y0!=y1 && testy>=y1) {
        // We now need to move to the next point along one of the lines
        if (testy>=ly1) {
          // Replace left hand line
          lx0=x1; ly0=y1;
          lx1=x2; ly1=y2;
          line=Graphics.yslopeIntercept(lx0, ly0, lx1, ly1);
          lm=line[0]; lb=line[1];
        } else {
          // Replace right hand line
          rx0=x1; ry0=y1;
          rx1=x2; ry1=y2;
          line=Graphics.yslopeIntercept(rx0, ry0, rx1, ry1);
          rm=line[0]; rb=line[1];
        };
        replaced=true;
      };
      // Calculate x position on left and right line corresponding to middle or y'th pixel
      var lx=testy*lm+lb;
      var rx=testy*rm+rb;
      // +++ The determinate above should prevent left and right from switching, but double check
      if (lx>rx) {
        // debugger;
        var t=rx;
        rx=lx;
        lx=t;
      };
      // Now, what we do depends upon how lx to rx spans pixels
      var lpx=Math.floor(lx);
      var rpx=Math.floor(rx);
      // Offset into leftmost and rightmost pixel
      var lpoff=lx-lpx;
      var rpoff=rx-rpx;
      // Iterate over pixels spanned
      for (; lpx<rpx; lpx++) {
        if (lpoff<=0.5) {
          // The left offset halfway or less across, we include
          // the center point of the pixel, so we include it in the triangle
          pixelfn(lpx, Math.floor(testy));
        };
        // From now on we consider the entire pixel from the left
        lpoff=0.0;
      };
      // We've now run into the last pixel on the right, decide whether to include it
      if (lpoff<=0.5 && rpoff>0.5) {
        // This triangle covers the center point of the pixel, so draw it
        pixelfn(rpx, Math.floor(testy));
      };
    };
    return;
  },
  
  // Draws a filled circle with triangles from the center
  // If not supplied, period is estimated based on the radius
  drawFilledCircle: function(cx, cy, radius, pixelfn, period, circleclass) {
    if (!period) {
      //  Decide how good the circle needs to be based on the circumference
      var circum=Math.PI*2*radius;
      // Hitting every other point should do it
      period=zround(circum, 2);
    };
    var circle=CircleUtils.makeCircle({cx: cx, cy: cy, radius: radius, period: period, class: circleclass});
    // Run around circle fully once
    var p0=circle.point(), p1=p0;
    for (var n=0; n<circle.period; n++, p0=p1) {
      p1=circle.step();
      Graphics.drawTriangle(p0[0], p0[1], p1[0], p1[1], cx, cy, pixelfn);
    };
    return;
  },

  // poly as an array of [x, y] pairs
  // Returns a function (closure) to call to do the next step of the animation
  // This function returns true when there's more to do, otherwise false
  // If the function is called again after it returns false, it starts over
  createAnimatePolyWithTriangles: function(poly, drawfn, centerp) {
    var polylen=poly.length;
    if (!centerp) {
      // No center point provided, use average value
      // This will work with all drawing modes for convex poly's
      var x=0, y=0;
      for (var np=0; np<polylen; np++) {
        x+=poly[np][0];
        y+=poly[np][1];
      };
      centerp=[Math.round(x/polylen), Math.round(y/polylen)];
    };
    var index=0;
    return function() {
      var p0=poly[index++];
      if (index>=polylen) index=0;
      var p1=poly[index];
      drawfn(p0, p1, centerp);
      return (index==0);
    };
  },

  // Colorspace conversion support
  // Adapted from http://axonflux.com/handy-rgb-to-hsl-and-rgb-to-hsv-color-model-c
  /**
   * Converts an RGB color value to HSV. Conversion formula
   * adapted from http://en.wikipedia.org/wiki/HSV_color_space.
   * Assumes r, g, and b are contained in the set [0, 255] and
   * returns h, s, and v in the set [0, 1].
   *
   * @param   Number  r       The red color value
   * @param   Number  g       The green color value
   * @param   Number  b       The blue color value
   * @return  Array           The HSV representation
   */
  rgbToHsv: function(r, g, b) {
    r=r/255, g=g/255, b=b/255;
    var max=Math.max(r, g, b), min=Math.min(r, g, b);
    var h, s, v=max;
    var d=max-min;
    s=(max==0) ? 0 : d/max;
    if (max==min){
      h = 0; // achromatic
    } else {
      switch (max) {
      case r: h=(g-b)/d+((g<b) ? 6 : 0); break;
      case g: h=(b-r)/d+2; break;
      case b: h=(r-g)/d+4; break;
      };
      h/=6;
    };
    function clip(n) {return Math.min(Math.max(n, 0.0), 1.0);};
    return [clip(h), clip(s), clip(v)];
  },

  /**
   * Converts an HSV color value to RGB. Conversion formula
   * adapted from http://en.wikipedia.org/wiki/HSV_color_space.
   * Assumes h, s, and v are contained in the set [0, 1] and
   * returns r, g, and b in the set [0, 255].
   *
   * @param   Number  h       The hue
   * @param   Number  s       The saturation
   * @param   Number  v       The value
   * @return  Array           The RGB representation
   */
  hsvToRgb: function(h, s, v) {
    var r, g, b;
    var i=Math.floor(h*6);
    var f=h*6-i;
    var p=v*(1-s);
    var q=v*(1-f*s);
    var t=v*(1-(1-f)*s);
    switch (i%6) {
    case 0: r=v, g=t, b=p; break;
    case 1: r=q, g=v, b=p; break;
    case 2: r=p, g=v, b=t; break;
    case 3: r=p, g=q, b=v; break;
    case 4: r=t, g=p, b=v; break;
    case 5: r=v, g=p, b=q; break;
    };
    function tobyte(n) {return Math.min(Math.max(n, 0.0), 1.0)*255;};
    return [tobyte(r), tobyte(g), tobyte(b)];
  },

  /**
   * Converts an RGB color value to HSL. Conversion formula
   * adapted from http://en.wikipedia.org/wiki/HSL_color_space.
   * Assumes r, g, and b are contained in the set [0, 255] and
   * returns h, s, and l in the set [0, 1].
   *
   * @param   Number  r       The red color value
   * @param   Number  g       The green color value
   * @param   Number  b       The blue color value
   * @return  Array           The HSL representation
   */
  rgbToHsl: function(r, g, b) {
    r/=255, g/=255, b/=255;
    var max=Math.max(r, g, b), min=Math.min(r, g, b);
    var h, s, l=(max+min)/2;
    if (max==min) {
      // achromatic
      h=s=0; 
    } else {
      var d=max-min;
      s=(l>0.5) ? d/(2-max-min) : d/(max+min);
      switch (max) {
        case r: h=(g - b) / d + (g < b ? 6 : 0); break;
        case g: h=(b - r) / d + 2; break;
        case b: h=(r - g) / d + 4; break;
      };
      h/=6;
    };
    function clip(n) {return Math.min(Math.max(n, 0.0), 1.0);};
    return [clip(h), clip(s), clip(l)];
  },

  /**
   * Converts an HSL color value to RGB. Conversion formula
   * adapted from http://en.wikipedia.org/wiki/HSL_color_space.
   * Assumes h, s, and l are contained in the set [0, 1] and
   * returns r, g, and b in the set [0, 255].
   *
   * @param   Number  h       The hue
   * @param   Number  s       The saturation
   * @param   Number  l       The lightness
   * @return  Array           The RGB representation
   */
  hslToRgb: function(h, s, l) {
    var r, g, b;
    if (s==0) {
      r=g=b=l; // achromatic
    } else {
      function hue2rgb(p, q, t) {
        if (t<0) t+=1;
        if (t>1) t-=1;
        if (t<1/6) return p+(q-p)*6*t;
        if (t<1/2) return q;
        if (t<2/3) return p+(q-p)*(2/3-t)*6;
        return p;
      };
      var q=(l<0.5) ? l*(1+s) : l+s-l*s;
      var p=2*l-q;
      r=hue2rgb(p, q, h+1/3);
      g=hue2rgb(p, q, h);
      b=hue2rgb(p, q, h-1/3);
    };
    function tobyte(n) { return Math.min(Math.max(n, 0.0), 1.0)*255; };
    return [tobyte(r), tobyte(g), tobyte(b)];
  },

  end: ""
}

var ArrayDrawer=function(w, h) {
  this.width=w;
  this.height=h;
  // Represent array as a vector w x h in length
  this.data=new Array(w*h);
  this.fill(0);
  return;
};

jQuery.extend(ArrayDrawer.prototype, {
  // Changes size
  setSize: function(neww, newh, preserve) {
    var newdata=new Array(neww*newh);
    if (preserve) {
      // Want to preserve old data
      // Copy appropriate section of old data and zero out an extra
      // +++ nyi
    };
    this.width=neww;
    this.height=newh;
    this.data=newdata;
    if (!preserve) this.fill(0);
    return;
  },

  aluFn: function(op, m) {
    switch (op) {
    case "or":
    case "ior":
      return function(p) {return p|m;};
    case "and": return function(p) {return p&m;};
    case "xor": return function(p) {return p^m;};
    case "set": return function(p) {return m;};
    case "remap": return function(p) {return p;};
    case "add": return function(p) {return p+m;};
    case "noop": return null;
    };
    return null;
  },

  // An interpreted alu function is a list of [pixel, alpha]
  // Each entry is either a number or a function that returns a number
  interpretAlu: function(alu) {
    var pixelval=0, pixelmask=1;
    var alphaval=255, alphamask=255;
    if (Array.isArray(alu)) {
      pixelval=alu[0];
      if (alu.length>1) pixelmask=alu[1];
      // Optimize special case of set
      if (pixelval==="set") pixelval=pixelmask;
      else if (typeof(pixelval)!=="function") pixelval=this.aluFn(pixelval, pixelmask);
      if (alu.length>2) {
        // Have an alpha spec, too
        alphaval=alu[2];
        if (alu.length>3) alphamask=alu[3];
        if (alphaval==="set") alphaval=alphamask;
        else if (typeof(alphaval)!=="function") alphaval=this.aluFn(alphaval, alphamask);
      };
    } else if (typeof(alu)==="number") {
      pixelval=alu;
    };
    return [pixelval, alphaval];
  },

  getPixel: function(x, y) { return this.data[this.width*y+x]; },

  setPixel: function(x, y, val) {
    this.processPixel(x, y, val);
    return val;
  },

  // Keep this as a separate function so could also do color mapping for only modified pixels
  processPixel: function(x, y, ialu) {
    var idx=this.width*y+x;
    ialu=ialu[0];
    if (ialu!==null) {
      this.data[idx]=(typeof(ialu)==="function" ? ialu(this.data[idx]) : ialu);
    };
    return;
  },

  update: function(force) {
    // Nothing to update, there is no display surface
    return;
  },
  
  fill: function(alu) {
    return this.drawRectangle(0, 0, this.width, this.height, alu);
  },

  drawRectangle: function(x0, y0, x1, y1, alu) {
    var ialu=this.interpretAlu(alu);
    var ad=this;
    Graphics.drawRectangle(x0, y0, x1, y1,
                           function(x, y) {
                             ad.processPixel(x, y, ialu);
                             return;
                           });
    return;
  },

  drawFilledCircle: function(cx, cy, radius, alu, period, circleclass) {
    var ialu=this.interpretAlu(alu);
    var ad=this;
    if (!period) {
      // If no period supplied, estimate the number of points in the circumference
      period=radius*Math.PI*2;
    };
    Graphics.drawFilledCircle(cx, cy, radius,
                              function(x, y) {
                                ad.processPixel(x, y, ialu);
                                return;
                              },
                              period, circleclass);
    return;
  },

  drawTriangle: function(x0, y0, x1, y1, x2, y2, alu) {
    var ialu=this.interpretAlu(alu);
    var ad=this;
    Graphics.drawTriangle(x0, y0, x1, y1, x2, y2,
                          function(x, y) {
                            ad.processPixel(x, y, ialu);
                            return;
                          }
                         );
    return;
  },

  // Adds a value to the triangular area, using the determinant for sign
  // If drawing triangles from a common point, pass the common point last
  drawTriangleAddDet: function(x0f, y0f, x1f, y1f, xcf, ycf, val, linewidth) {
    var det=Graphics.triDeterminant(x0f, y0f, x1f, y1f, xcf, ycf);
    if (det==0) {
      // alert("determinant is zero!");
    } else {
      this.drawTriangle(x0f, y0f, x1f, y1f, xcf, ycf, ["add", val*sign(det), "noop", 0]);
      if (typeof(linewidth)==="number" && linewidth!=0) {
        // Draw an opposing triangle of opposite sense to leave just a line around the edge
        var cp=[xcf, ycf];
        var np0=Graphics.extendPointAlongVector(cp, [x0f, y0f], linewidth);
        var np1=Graphics.extendPointAlongVector(cp, [x1f, y1f], linewidth);
        // this.drawTriangle(np1[0], np1[1], np0[0], np0[1], xcf, ycf, ["add", val*sign(-det), "noop", 0], 0);
        this.drawTriangleAddDet(np1[0], np1[1], np0[0], np0[1], xcf, ycf, val, 0);
      };
    };
    return;
  },

  createAnimatePolyWithTriangles: function(poly, alu, centerp, linewidth) {
    var drawfn;
    var drawer=this;
    if (typeof(alu)==="number") {
      drawfn=function(p0, p1, cp) {
        drawer.drawTriangleAddDet(p0[0], p0[1], p1[0], p1[1], cp[0], cp[1], alu, linewidth);
        return;
      };
    } else {
      drawfn=function(p0, p1, cp) {
        drawer.drawTriangle(p0[0], p0[1], p1[0], p1[1], cp[0], cp[1], alu);
        return;
      };
    };
    return Graphics.createAnimatePolyWithTriangles(poly, drawfn, centerp);
  },

  end: ""
});

var ColorMappedCanvasDrawer=function(canvas, mapspec) {
  // Get width and height from canvas
  this.canvas=canvas;
  this.width=canvas.width;
  this.height=canvas.height;
  this.context=canvas.getContext("2d");
  // this.context.fillStyle="black";
  // this.context.fillRect(0, 0, this.width, this.height);
  this.imageData=this.context.createImageData(canvas.width, canvas.height);
  this.colorMap=this.makeColorMap(mapspec);
  this.modX0=0;
  this.modY0=0;
  this.modX1=this.width;
  this.modY1=this.height;
  ArrayDrawer.call(this, this.width, this.height);
  return;
}

jQuery.extend(ColorMappedCanvasDrawer.prototype, new ArrayDrawer(0, 0), {
  setColorMap: function(newcm, noremap) {
    this.colorMap=newcm;
    if (!noremap) {
      // Remap all of the pixels
      var ialu=[null, null];
      for (var y=0; y<this.height; y++) {
        for (var x=0; x<this.width; x++) {
          ialu[0]=this.data[this.width*y+x];
          this.processPixel(x, y, ialu);
        };
      };
    };
    return;
  },

  colorMapToString: function(cm) {
    var scm="";
    for (var nc=0; nc<cm.length; nc++) {
      var c=cm[nc];
      if (nc>0) scm+=",";
      scm+=c[0]+(c[1]<<8)+(c[2]<<16);
    };
    return scm;
  },
  
  makeColorMap: function(mapspec) {
    if (!mapspec) mapspec=256;
    if (typeof(mapspec)==="number") mapspec={length: mapspec, type: "default"};
    var ncolors=mapspec.length;
    var cm=new Array(ncolors);
    var expmap=mapspec.map;
    if (expmap) {
      // If the map is explicitly supplied, use it
      var cs=expmap.split(",");
      for (var nc=0; nc<ncolors; nc++) {
        var c=parseInt(cs[nc]);
        cm[nc]=[c&255, (c>>8)&255, (c>>16)&255];
      };
    } else {
      switch (mapspec.type) {
      case "gray": {
        for (var nc=0; nc<ncolors; nc++) {
          var n=Math.round(255*(nc/ncolors));
          cm[nc]=[n, n, n];
        };
        break;
      };

      case "random": {
        cm[0]=[0, 0, 0];
        for (var nc=1; nc<ncolors; nc++) {
          cm[nc]=[Math.round(256*Math.random()), Math.round(256*Math.random()), Math.round(256*Math.random())];
        };
        break;
      };

      case "highcontrast":
      default: {
        // Default to a high contrast color map
        var mask=0;
        for (var nc=0; nc<ncolors; nc++, mask++) {
          // Always skip black once we get past color 0
          if (nc>0 && ((mask&7)==0)) mask++;
          cm[nc]=[255*(mask&1), 255*((mask>>1)&1), 255*((mask>>2)&1)];
        };
        break;
      };
      };
    };
    return cm;
  },

  resizeCanvas: function(neww, newh, preserve) {
    // Optimize out no change
    if (this.width!=neww || this.height!=newh) {
      this.canvas.width=(this.width=neww);
      this.canvas.height=(this.height=newh);
      var olddata=this.imageData;
      this.imageData=this.context.createImageData(neww, newh);
      // +++ preserve is nyi by setSize
      this.setSize(neww, newh, preserve);
      if (preserve) {
        // Copy from old image data into new
        // +++nyi
      };
    };
    return;
  },

  absMode: false,
  processPixel: function(x, y, ialu) {
    var w=this.width, h=this.height;
    // Clip
    if (x<0 || y<0 || x>=w || y>=h) return;
    var idx=this.width*y+x;
    var ididx=idx*4;
    var val;
    var pixel=ialu[0];
    var cm=this.colorMap;
    var id=this.imageData.data;
    if (pixel!==null) {
      if (typeof(pixel)!=="number") pixel=pixel(this.data[idx]);
      this.data[idx]=pixel;
      // Map through color map and also modify drawing surface
      // There are 4 bytes per index
      var cmidx, cmlen=cm.length;
      // Just for convention purposes, negate pixel here, since animations seem
      // to mostly start with negative winding numbers
      if (this.absMode) cmidx=Math.abs(-pixel)%cmlen;
      else {
        cmidx=(-pixel)%cmlen;
        // Wrap around negative numbers
        if (cmidx<0) cmidx+=cmlen;
      };
      var cme=cm[cmidx];
      id[ididx++]=cme[0];
      id[ididx++]=cme[1];
      id[ididx++]=cme[2];
    } else {
      // No modification to pixel value, skip
      ididx+=3;
    }
    // Calculate alpha, maybe based on old alpha
    var alpha=ialu[1];
    if (alpha!==null) {
      if (typeof(alpha)!=="number") alpha=alpha(id[ididx]);
      id[ididx]=alpha;
    };
    // Maintain modified rectangle
    this.modX0=Math.min(this.modX0, x);
    this.modY0=Math.min(this.modY0, y);
    this.modX1=Math.max(this.modX1, x+1);
    this.modY1=Math.max(this.modY1, y+1);
    return;
  },

  update: function(force) {
    if (force) this.context.putImageData(this.imageData, 0, 0);
    else if (this.modX1>this.modX0 && this.modY1>this.modY0) {
      // Have a non-empty region to update
      this.context.putImageData(this.imageData, 0, 0,
                                this.modX0, this.modY0,
                                this.modX1-this.modX0, this.modY1-this.modY0);
    };
    this.modX0=this.width;
    this.modY0=this.height;
    this.modX1=0;
    this.modY1=0;
  },

  end: ""
});
