/// Careful Triangles and Artful Graphics
/// Rotors
/// Copyright (c) 2018 by Howard I. Cannon - MIT License

"use strict";

function bount(n) {
  return (n>>1)*Math.pow(-1, n);
}

function parseRotorCoeffs(scoeffs, scale) {
  scale=defarg(scale, 1.0);
  // First, remove all whitespace and brackets
  scoeffs=scoeffs.replace(/(\s|{|}|\[|\]|\(|\))+/gm, "");
  // Also fix exponent format
  scoeffs=scoeffs.replace(/\*10\^/g, "e").toLowerCase();
  // Then split into individual coefficients
  var ascoeffs=scoeffs.split(",");
  var acoeffs=[];
  var reparts=/([-+]?[0-9]*\.?[0-9]+(?:e[-+]?[0-9]+)?i?)/g;
  for (var nc=0; nc<ascoeffs.length; nc++) {
    var matches=ascoeffs[nc].match(reparts);
    var real=0, imag=0;
    var n=parseFloat(matches[0]);
    if (matches[0].indexOf("i")>=0) imag=n;
    else real=n;
    if (matches.length>1) {
      n=parseFloat(matches[1]);
      if (matches[1].indexOf("i")>=0) imag=n;
      else real=n;
    };
    if (imag==0) acoeffs.push(real*scale);
    else acoeffs.push({r: real*scale, i: imag*scale});
  };
  return acoeffs;
}

function parseXYRotorCoeffs(sxs, sys, scale) {
  scale=defarg(scale, 1.0);
  var xs=parseRotorCoeffs(sxs, scale), ys=parseRotorCoeffs(sys, scale);
  // Now, the X's are the real parts and the Y's are the imaginary parts
  var acoeffs=[];
  for (var nc=0; nc<xs.length; nc++) {
    acoeffs.push({r: xs[nc], i: ys[nc]});
  };
  return acoeffs;
}


var RotorUtils={
  // With no default prefix, getRotorFn is used
  defaultRotorFnPrefix: "",
  getRotorMaker: function() {return RotorUtils["get"+RotorUtils.defaultRotorFnPrefix+"RotorFn"];},
  getDefaultRotorFn: function() {
    return RotorUtils.getRotorMaker().apply(this, arguments);
  },

  defaultRotorSortFn: function(circles) {return circles;},
  rotorSortNoneFn: function(circles) {return circles;},

  rotorSortOriginalFn: function(circles) {
    circles.sort(function(a, b) {return a.index-b.index;});
    return circles;
  },

  rotorSortBackwardsFn: function(circles) {
    circles.sort(function(a, b) {return -(a.index-b.index);});
    return circles;
  },

  rotorSortRadiusUpFn: function(circles) {
    circles.sort(function(a, b) {return -(b.radius-a.radius);});
    return circles;
  },
  rotorSortRadiusDownFn: function(circles) {
    circles.sort(function(a, b) {return b.radius-a.radius;});
    return circles;
  },
  rotorSortSpeedUpFn: function(circles) {
    circles.sort(function(a, b) {return Math.abs(a.dt)-Math.abs(b.dt);});
    return circles;
  },
  rotorSortSpeedDownFn: function(circles) {
    circles.sort(function(a, b) {return -(Math.abs(a.dt)-Math.abs(b.dt));});
    return circles;
  },

  getGRotorFn: function(drawer, ncircles, radiusfn, dtfn, phifn, colorfn, startrotor, colorfnignore, sortfn) {
    radiusfn=funkify(radiusfn);
    dtfn=funkify(dtfn);
    phifn=funkify(phifn);
    colorfn=(supplied(colorfn) ? funkify(colorfn) : funkify(1));
    var circles=[];
    var cx=zround(drawer.width, 2), cy=zround(drawer.height, 2);
    var x=cx, y=cy;
    var period, firstmover;
    // First, create all of the circles and the colors to go with them
    for (var nc=0; nc<ncircles; nc++) {
      var r=radiusfn(nc);
      var dt=dtfn(nc);
      if (r!=0 && dt!=0) {
        var phi=supplied(phifn) ? Math.PI*phifn(nc) : 0;
        var circle=CircleUtils.makeCircle({cx: x, cy: y, radius: r, dt: dt, phi: phi});
        circle.index=nc;
        circle.color=colorfn(nc);
        circles.push(circle);
        var pc=circle.point();
        x=pc[0];
        y=pc[1];
        // alert(nc+", r="+r+", x="+x+", y="+y+", dt="+dt+", phi="+phi);
        // alert(nc+", period="+circle.period);
        if (typeof(period)==="undefined") {
          period=circle.period;
          firstmover=nc;
        };
      };
    };
    // Fixup for case where r was zero for some circles
    ncircles=circles.length;
    sortfn=defarg(sortfn, RotorUtils.defaultRotorSortFn);
    circles=sortfn(circles);
    startrotor=defarg(startrotor, 0);
    // Set up other variables for "loop"
    var subperiod, count=0, connected=0, inc=1;
    function sendcircles(msg, arg) {
      for (var nc=0; nc<ncircles; nc++) circles[nc][msg](arg);
      return;
    };
    return function(command) {
      if (command=="setcenter") {
        cx=arguments[1];
        cy=arguments[2];
        return;
      };
      if ((count%period)==0) {
        if (connected!=ncircles) {
          circles[connected].setCenter(0, 0);
          circles[connected].color=colorfn(connected);
          connected++;
        };
      };
      if (connected<ncircles) {
        var c=circles[connected];
        subperiod=c.period;
        if (((subperiod+count)%period)==0) {
          c.color=-(zround(count, subperiod)*c.color);
        };
      };
      switch (command) {
      case "backward": sendcircles("setForward", false); inc=-1; break;
      case "forward": sendcircles("setForward", true); inc=1; break;
      case "reverse": sendcircles("reverse"); inc=-inc; break;
      case "sort": circles=arguments[1](circles); return;
      case "circles": return circles;
      };

      var x=cx, y=cy, ox=cx, oy=cy;
      for (var i=0; i<ncircles; i++) {
        var circle=circles[i];
        var alu=["add", circle.color];
        var point=circle.point();
        var step=circle.step();
        if (i<connected) {
          var ox0=ox, oy0=oy, x0=x, y0=y;
          ox=ox+point[0];
          oy=oy+point[1];
          x=x+step[0];
          y=y+step[1];
          if (circle.color!=0 && i>=startrotor) {
            drawer.drawTriangleAddDet(x0, y0, ox0, oy0,   ox,   oy, alu[1]);
            drawer.drawTriangleAddDet(x0, y0, ox, oy, x, y, alu[1]);
          };
        } else {
          drawer.drawTriangle(circle.cx, circle.cy, point[0], point[1], step[0], step[1], alu);
        };
      };
      count+=inc;
      return;
    };
  },

  getRotorFn: function(drawer, ncircles, radiusfn, dtfn, phifn, color1fn, startrotor, color2fn, sortfn) {
    radiusfn=funkify(radiusfn);
    dtfn=funkify(dtfn);
    color1fn=funkify(color1fn, 1);
    color2fn=funkify(color2fn, color1fn);
    phifn=funkify(phifn);
    startrotor=defarg(startrotor, 0);
    var circles=[];
    var non0dt;
    for (var n=0; n<ncircles; n++) {
      var r=radiusfn(n);
      if (r!=0) {
        var dt=dtfn(n);
        if (Math.abs(dt)==Math.PI) {
          // Place to put a breakpoint trying to find Chrome issue
          dt=dt;
        };
        if (!non0dt && dt!=0) non0dt=dt;
        var c=CircleUtils.makeCircle({cx: 0, cy: 0, radius: r, dt: dt, phi: Math.PI*phifn(n)});
        c.index=n;
        c.colors1=color1fn(n);
        c.colors2=color2fn(n);
        circles.push(c);
      };
    };
    sortfn=defarg(sortfn, RotorUtils.defaultRotorSortFn);
    circles=sortfn(circles);
    return RotorUtils.makeRotorStepFn(drawer, circles, zround(2, non0dt), startrotor);
  },

  // In original code period was used to stop after one cycle
  makeRotorStepFn: function(drawer, circles, period, startrotor) {
    var cx=zround(drawer.width, 2), cy=zround(drawer.height, 2);
    var count=0;
    var ncircles=circles.length, inc=1;
    function sendcircles(msg, arg) {
      for (var nc=0; nc<ncircles; nc++) circles[nc][msg](arg);
      return;
    };
    return function(command) {
      switch (command) {
      case "backward": sendcircles("setForward", false); inc=-1; break;
      case "forward": sendcircles("setForward", true); inc=1; break;
      case "reverse": sendcircles("reverse"); inc=-inc; break;
      case "setcenter": cx=arguments[1]; cy=arguments[2]; return;
      case "sort": circles=arguments[1](circles); return;
      case "circles": return circles;
      default: return;
      };
      var x=cx, y=cy, ox=cx, oy=cy;
      for (var i=startrotor; i<ncircles; i++) {
        var circle=circles[i];
        var point=circle.point();
        var step=circle.step();
        var ox0=ox, oy0=oy, x0=x, y0=y;
        ox=ox+point[0];
        oy=oy+point[1];
        x=x+step[0];
        y=y+step[1];
        var c1=circle.colors1;
        // At this point:
        // ox0 and oy0 hold the previous center of the circle
        // x0 and y0 hold the new center of the circle
        // ox and oy hold the previous point on the circle
        // x and y hold the new point on the circle
        // So we're going to fill in the quadrilateral consisting of these four points
        // (x0, y0) and (ox, oy) are the common edge as required by the careful triangle
        if (c1!=0) {
          drawer.drawTriangleAddDet(x0, y0, ox0, oy0, ox, oy, c1);
        };
        var c2=circle.colors2;
        if (c2!=0) {
          drawer.drawTriangleAddDet(x0, y0, ox, oy, x, y, c2);
        };
      };
      count+=inc;
      return;
    }
  },

  coeffMag: function(c) {
    var mag=c;
    if (typeof(c)=="object") {
      // Assume a complex number
      mag=Math.sqrt(c.r*c.r+c.i*c.i);
    } else mag=Math.abs(mag);
    return mag;
  },

  coeffPhi: function(c) {
    // Use arctangent to determine initial rotor angle
    var real=c, imag=0;
    if (typeof(c)=="object") {
      real=c.r;
      imag=c.i;
    };
    var phi=Math.atan2(imag, real);
    if (phi<0.0) phi+=2*Math.PI;
    return phi;
  },

  getRotorFnFromCoeffs: function(drawer, r, maxn, coeffs, dt, freqfn, colorf) {
    if (typeof(coeffs)=="string") coeffs=RotorUtils.parseCoeffs(coeffs);
    colorf=defarg(colorf, function(n) {return 1<<(hn%4);});
    var phis=[];
    var dts=[];
    var res=RotorUtils.getDefaultRotorFn(drawer, Math.min(coeffs.length, maxn),
                                          function(n) {return r*RotorUtils.coeffMag(coeffs[n]);},
                                          function(n) {
                                            var dt0=dt*freqfn(n);
                                            dts.push(dt0);
                                            return dt0;
                                          },
                                          function(n) {
                                            var phi=RotorUtils.coeffPhi(coeffs[n])/Math.PI;
                                            phis.push(phi);
                                            return phi;
                                          },
                                          colorf);
    return res;
  },

  end: ""
}

var RotorAnimations={
  getSpiroFn: function(drawer, ncircles, radiusfn, dtfn, phifn, colorfn, linewidth, circleclass) {
    radiusfn=funkify(radiusfn);
    dtfn=funkify(dtfn);
    colorfn=funkify(colorfn, 1);
    phifn=funkify(phifn, 0);
    var colors=[], circles=[];
    for (var n=0; n<ncircles; n++) {
      var dt=dtfn(n);
      colors.push(colorfn(n));
      circles.push(CircleUtils.makeCircle({cx: 0, cy: 0, radius: radiusfn(n), dt: dtfn(n), phi: Math.PI*phifn(n),
                                           class: circleclass}));
    };
    return RotorAnimations.makeSpiroStepFn(drawer, circles, colors, linewidth);
  },

  makeSpiroStepFn: function(drawer, circles, colors, linewidth) {
    var cx=zround(drawer.width, 2), cy=zround(drawer.height, 2);    
    var count=0;
    var ncircles=circles.length, inc=1;
    function sendcircles(msg, arg) {
      for (var nc=0; nc<ncircles; nc++) circles[nc][msg](arg);
      return;
    };
    var prevpoints=[];
    return function(command) {
      switch (command) {
      case "backward": sendcircles("setForward", false); inc=-1; break;
      case "forward": sendcircles("setForward", true); inc=1; break;
      case "reverse": sendcircles("reverse"); inc=-inc; break;
      case "setcenter": cx=arguments[1]; cy=arguments[2]; return;
      };
      var x=cx, y=cy, xoff=cx, yoff=cy;
      for (var i=0; i<ncircles; i++) {
        var circle=circles[i];
        var point=circle.point();
        xoff+=point[0];
        yoff+=point[1];
        point=[xoff, yoff];
        circle.step();
        var color=colors[i];
        if (prevpoints.length<=i) {
          prevpoints.push(point);
        } else {
          var prev=prevpoints[i];
          if (color!=0) drawer.drawTriangleAddDet(prev[0], prev[1], point[0], point[1], cx, cy, color, linewidth);
          prevpoints[i]=point;
        };
      };
      count+=inc;
      return;
    }
  },

  getBogorFn: function(drawer, r, ncircs, dt, phimpy) {
    if (!supplied(r)) r=222;
    circs=defarg(ncircs, 3);
    phimpy=defarg(phimpy, 1);
    var basedt=defarg(dt, pi/256);
    return RotorUtils.getDefaultRotorFn(drawer, circs, r, basedt,
                                         function(n) {return phimpy*n/2;});
  },

  // params object includes keys with defaults: radius=200, n=19, dt=pi/256, amp=4, color=1
  getBesserFn: function(drawer, r, n, dt, amp, color) {
    r=defarg(r, 200);
    // The expected effect is obtained only with an odd number of rotors, so force it
    // (Not to say that the even cases aren't pretty...)
    n=defarg(n, 19)*2+1;
    dt=defarg(dt, pi/256);
    amp=defarg(amp, 4);
    var colorfn=funkify(color, 1); // function(n) {return 1<<(n%8);}
    var rfn=function(n) {
      var jn=bount(n+1);
      var jna=Math.abs(jn);
      // amp is the argument, jna is the order
      var res=besselj(amp, jna);
      res*=r;
      // For special case of negative order, J(arg, -order)=(-1^order)*J(arg, order)
      if (jn<0 && isOdd(jna)) res*=-1;
      return res;
    };
    return RotorUtils.getDefaultRotorFn(drawer, n, rfn,
                                         function(n) {return dt*bount(n+1);},
                                         0, colorfn);
  },

  getBessemerFn: function(drawer) {
    return RotorUtils.getBesserFn(drawer, 444, 19, pi/1024, 6.5, 1);
  },

  getSpiroFnFromList: function(drawer, rsanddts, linewidth) {
    return RotorAnimations.getSpiroFn(drawer, rsanddts.length,
                                      function(n) {return rsanddts[n][0];},
                                      function(n) {return rsanddts[n][1];},
                                      0.0,
                                      function(n) {return rsanddts[n][2];},
                                      linewidth
                                     );
  },

  getGibbsSquareFn: function(drawer, r, n, dt, colorf, startrotor) {
    r=defarg(r, 239);
    n=defarg(n, 7);
    dt=defarg(dt, pi/1024);
    colorf=defarg(colorf, function(n) {return 1<<(n%8);});
    return RotorUtils.getDefaultRotorFn(drawer, n,
                                         function(n) {return r/Math.pow(-1, n)/(n+n+1);},
                                         function(n) {return Math.pow(-1, n)*dt*(n+n+1);},
                                         function(n) {return 0;},
                                         colorf, startrotor);
  },

  getSmoothTriFn: function(drawer, r, n, dt, colorf) {
    r=defarg(r, 190);
    n=defarg(n, 14);
    dt=defarg(dt, pi/1024);
    colorf=defarg(colorf, function(n) {return 1<<(n%8);});
    return RotorUtils.getDefaultRotorFn(drawer, n,
                                         function(n) {return r*Math.pow((n%3==2 ? 0 : 1)/(n+1), 2);},
                                         function(n) {return dt*Math.pow(-1, n%3)*(n+1);},
                                         0, colorf);
  },

  getGibbsTriFn: function(drawer, r, n, dt, colorf, startrotor) {
    r=defarg(r, 180);
    n=defarg(n, 14);
    dt=defarg(dt, pi/1024);
    colorf=defarg(colorf, function(n) {return 1<<(n%8);});
    return RotorUtils.getDefaultRotorFn(drawer, n*3,
                                         function(n) {
                                           var x=0;
                                           switch (n%6) {
                                           case 0: x=1; break;
                                           case 3: x=-1; break;
                                           };
                                           return (r*x)/((n>>1)+1);
                                         },
                                         function(n) {return dt*Math.pow(-1, n)*((n>>1)+1);},
                                         0, colorf, startrotor);
  },

  getDragonFn: function(drawer, r, n, dt, colorf, pow) {
    pow=defarg(pow, 0.5);
    colorf=defarg(colorf, function(n) {return 1<<(n%4);});
    return RotorUtils.getDefaultRotorFn(drawer, n,
                                         function(n) {
                                           var n2=(n>>1)+1;
                                           return r/n2/n2;
                                         },
                                         function(n) {
                                           var n2=(n>>1)+1;
                                           return dt*Math.pow(-1, n)*pow*n2*n2*n2;
                                         },
                                         function(n) {return 0.5+n/3.0;},
                                         colorf);
  },

  hilbertCoeffs: [0.06223887138466256, -(1/2) /* + I/ 2 */, -0.03757699960423023, 0.3534622421100067, 0.001991248730794406,
                  0.02263634713979416, -0.06127348004948964, -0.05035135619339813, -0.04758128681867736, 0.08836556052750166,
                  -0.05751588565102905, 0.01990485205073149, 0.01555971784616565, 0.01207443558318487, -0.03331928114438672,
                  0.01328808465702301, -0.02024432653812967, 0.005659086784948563, -0.02312387553544684, 0.04295107772777291,
                  -0.009394249901057534, 0.03252934642836088, 0.004622962864207519, 0.02909644098020461, 0.01588426613603546,
                  -0.01258783904834953, 0.0009063918613197954, 0.03731563786373518, 0.000497812182698582, 0.01368955264400237,
                  -0.02349203825649674, -0.006686675767394851, -0.01110056796080487, 0.02209139013187543, -0.001140488640825559,
                  0.01561532049755264, -0.01531837001237242, 0.02375744009931453, -0.003754433926688407, 0.01836628031913467,
                  -0.005110042822623129, 0.00497621301268281, -0.003810068008611362, 0.008587662820422375, -0.01189532170466933,
                  0.009727021320018096, -0.00050411957818439, 0.01097491651483933, -0.001946950453038532, 0.003018608895796204,
                  -0.001632739044164659, 0.007702135104158005, -0.01437897141275724, 0.006614835571164981, 0.005698185743091395,
                  0.00845891302783349, -0.001031362005851175, 0.003322021164255779, -0.01199842836214862, 0.0008555117640799113,
                  0.003889929461541422, -0.00655083422020749, -0.003956361177097896, -0.003362472537382089, 0.005557912565736062,
                  0.001414771696237165, 0.005247349047891251, 0.01519943166862378, -0.008329820286096738, 0.009808179927835343,
                  -0.003205587901945435, 0.0008143322778257947, -0.002127297739457049, 0.01073776943194319, -0.001602784224736215,
                  0.0005558826227172634, -0.005061081634532488, 0.001824506653626429, -0.004034635053493906, 0.0006148939397054995,
                  -0.003751054629672773, 0.00813233660709018, -0.002353487597718222, 0.0006311522734217512, -0.005780968883861763,
                  0.001817619296684152, -0.003567027147756754, 0.001902570923195148, -0.00835128440376175, 0.007274110245051123,
                  -0.0100706148355437, -0.002660970094028548, -0.002348562475264354, -0.001681730037139576, 0.003820232797154878,
                  0.004535106229689979, 0.004707599633100043, -0.003146959762087377, -0.002005547666291996, 0.006649106837315133,
                  0.001155740716051917, -0.00211434016181398, -0.00658963522290451, -0.006423241345659125, -0.005296909011015884,
                  0.00932890946593377, -0.007639979813265072, 0.0004045531436072537, 0.003971066534008885, -0.000284490064479035,
                  -0.007580545870469911, -0.0009483515635467404, -0.004991027339946337, 0.003422388161000601, -0.005512248408155111,
                  0.00554059593925073, 0.0002265979653300432, 0.005126280854934592, -0.001237333189778612, 0.004437415023292161,
                  0.002898552881221581, -0.001671668941848663, 0.0006391056113377033, 0.008628533282493614, 0.000124453045674628,
                  0.004593947714447126, -0.006114253036090815, -0.000983464276133556, -0.001413310362544937, 0.005522847532968873,
                  0.002205601784596595, 0.004862888303959315, -0.005873009564124123, 0.007338958973388599, -0.0007728328190390376,
                  0.00423586998609516, -0.0004950210379072408, 0.00390383012438826, -0.00003877568799838723, 0.0006140030567896404,
                  -0.002775141990201283, 0.001160930694110041, -0.003458207482514088, -0.0001916264477744515, -0.003307034526061018,
                  0.005939360024828611, -0.002716371994552343, -0.0006705928223811072, -0.000285122160206434, 0.0006481349082489176,
                  -0.005570254013874692, 0.0002656926872077723, -0.004895961282965411, 0.004591570079783698, -0.00209819742128886,
                  0.001081598476448882, -0.003829592503093119, 0.003775621220268773, -0.0006409521892965854, 0.004575337721336643,
                  -0.002235066124275481, 0.001244053253170635, -0.003066084744969638, 0.0005743223144445868, -0.000938608481672101,
                  0.00008804772336298813, -0.002353593843365625, 0.0008999431208116946, -0.002096309328876612, 0.00214691570510559,
                  -0.002134009909879341, 0.001761890250422856, -0.001277510705655801, 0.00152120626257285, -0.002441686096612816,
                  0.00125872429131159, -0.002238181281288213, 0.002431755330004541, -0.002058949446100431, 0.001353421564823904,
                  -0.0009525170021528194, 0.001461102146603759, -0.003384811359924373, 0.001140503434731367, -0.00347087906312315,
                  0.00274372912870986, -0.002069237475718652, 0.001103543067748216, -0.002973830426167325, 0.002441355957479614,
                  -0.0001355378475247463, 0.003375134318286645, -0.001593466704405271, 0.0007546522239490392, -0.002987604383122198,
                  0.0006119717693622004, -0.0001260298945461244, -0.000478847201473909, -0.002404601245414132, 0.0001318295615911588,
                  -0.002176772672579778, 0.001925533776039497, -0.002445744196167361, 0.001569003094086856, -0.0004867376132596315,
                  0.001231182748337168, -0.002341321779096503, 0.001120775733455021, -0.002029878524413234, 0.001653708892791233,
                  -0.002000838666315617, 0.001635562122219605, -0.0004081847610411777, 0.001522806638705557, -0.003351182219612117,
                  0.0011224434833845, -0.003181579269862162, 0.002114728256958395, -0.00125857147650526, 0.001398256670325407,
                  -0.003594742853189293, 0.002892230223599603, 0.0001509210410761142, 0.003674072970188372, -0.002871385960764938,
                  0.0008305052910639733, -0.004933181179907147, -0.0009300354479400973, 0.001424546435772842, -0.001818965827432853,
                  -0.002734483324740853, 0.0002966140036928695, -0.002934655087996448, 0.0002138779410200188, -0.003514046123168347,
                  0.002582950987704468, -0.0002578405014626508, 0.002011675597722261, -0.0006704146346037957, 0.002967667739462253,
                  -0.0004892154326121379, -0.001637708555051845, -0.0008628991200493141, 0.003865905077299429, -0.002999607090537095,
                  0.002696302054318873, 0.002064465135180341, 0.002921900497932673, 0.0007738729459218142, -0.0008406181343456021,
                  -0.002593603739143034, 0.00156660364464094, 0.0009724823653853659, -0.001288967835347924, -0.0005640211247844295,
                  -0.00112721967293219, 0.001935476936549804, 0.0003536929240593176, 0.00091638914090145, 0.003848754073418842,
                  -0.0009890902942744219, 0.001649401731897833, -0.001962149391466704, -0.001215433770825039, -0.001532845867268418,
                  0.003799857917155961, -0.002367628765244806, -0.0001926858127673764, 0.001389478141434096, -0.0002732869463351612,
                  -0.003586339801281728, -0.0009282205198204929, -0.002578535589536818, 0.002452044981958773, -0.002756051208602553,
                  0.001446489009010723, 0.00131183726197284, 0.00169452687153101, -0.002700605481863785, 0.001194542134977341,
                  -0.0001691137410212641, 0.0002035830694564686, 0.001171293026873436, 0.003887129980002576, -0.002082455071524243,
                  0.003953992056724249, -0.002236520120578821],
  getHilbertFn: function(drawer, r, n, dt, colorf) {
    var coeffs=RotorAnimations.hilbertCoeffs;
    n=Math.min(n, coeffs.length);
    colorf=defarg(colorf, function(n) {return 1<<(n%4);});
    dt/=10;
    return RotorUtils.getDefaultRotorFn(drawer, n,
                                         function(n) {return r*Math.abs(coeffs[n]);},
                                         function(n) {return dt*(bount(n+2)+1);},
                                         function(n) {return (coeffs[n]>=0.0) ? 0.0 : 1.0;},
                                         colorf);
  },

  base2Coeffs: [-0.0004747476453367828,0.0007610695792685598,0.00009932781069767989,0.0003953768008454699,
                -0.0007177264297245438,-0.000815233345241471,0.00004234948561983365,0.0001015171954136594,
                -0.0003843433455610063,0.0005121101693467125,-0.0008522597265172447,-0.00007068340605832614,
                -0.0005677903024088744,0.0009479326098987811,0.0009363339500065274,-0.0002885653504065064,
                0.0007210449104085295,0.000887229099075375,0.0002041916563497175,-0.0005838569915054722,
                0.0001637598450835509,-0.0005241942072383007,0.0007733236550685385,0.0007947327381749779,
                -0.0002992367130301747,0.0006923497055742907,0.001057644723847063,-0.000124141779293968,
                0.00008114690766602539,0.0003215158930378972,6.830091531663808e-6,-0.0003908437305064103,
                -0.000685701762647188,-0.0001070268996247652,0.0004341129775080726,-0.000252138887388691,
                0.000640481361212125,-0.001009558298472226,-0.0000695934891994536,-0.0005912443087004612,
                0.000940577845618317,0.0008914128715965654,-0.0002766094842587323,0.0007075953648505779,
                0.0008240722782692167,0.0002746658771666924,-0.0007280950454353839,0.0001613703145024841,
                -0.0006630180386569547,0.001009190318531357,0.001023477761766492,-0.0004108588410688351,
                0.001011644068922774,0.001364211313720035,0.00004631580928357041,-0.0002960520326440468,
                0.000365843378190334,-0.0002740554357903702,0.00004169478416816259,-0.0002321017192411498,
                -0.0002010224050416118,0.0005477664591509723,0.00008321072028947642,0.0003525056103459081,
                -0.0004539670520291939,-8.169626716052314e-6,-0.00008930332830024942,-0.0000158572666713221,
                -0.0001841812263168028,0.00008377133627423899,-0.0002578896582457987,-0.0006148216730432609,
                0.0001407331855673084,-0.000133160329287015,-0.0002665590286597764,-0.00002842932227926146,
                0.0003928876400642146,0.0006975732422763131,0.00008125594393085188,-0.0003397622121393119,
                0.0004442843777206494,-0.0006299934343248351,0.0009085960242401795,0.0001969775186454348,
                0.0004096058357662074,-0.0007719753250495193,-0.0008502873185571122,0.00002355681444970001,
                0.0001203666268400091,-0.0003028140451402802,0.0003795076482704691,-0.0005611555540784244,
                -0.00004095382692814759,-0.0002874949312587736,0.000411074097339744,0.0003375207315168393,
                -0.00008240646849926409,0.0001501937853640695,0.0001042637642806617,0.00002260925037393982,
                -9.86057809882082e-6,-0.00004271187433215533,0.00007377774122464992,-0.00009628411367240978,
                -0.00006607657719092878,0.0001364036594880643,-0.0004379165440872559,-0.0003577985455130275,
                -0.0002667522072057463,0.0005346091984301758,-0.00009545342144275162,0.000371114286641536,
                -0.0004735123182824507,-0.0003902524391457295,0.0001581328771651466,-0.000317720777026111,
                -0.0003282899117369678,0.00006816829647654599,-0.0002042450607388355,-0.00004318199341821984,
                -0.0003241915528494617,0.0007168633034561196,0.0009087623585307196,-0.0003189215681119338,
                0.0008797461163945812,0.001512850769401443,-0.0000384810355735199,-0.000267491681584881,
                0.0005520853185302254,-0.0003856462040211279,-0.00003644121037046246,-0.0006051445076235705,
                -0.0004217139530552292,0.001326140458081157,7.37075389746208e-6,0.001347966863372073,
                -0.002174331543663823,-0.0001935116949216575,-0.001159982350182567,0.001939117884349437,
                0.002038054088672318,-0.000217454759640557,0.00008122121517243899,0.00100575475446905,
                -0.0009609997072716015,0.001613656972392924,0.0001536286195376402,0.001167761829678075,
                -0.002051332153857964,-0.002156769793764382,0.0006818409718877397,-0.001731459751286826,
                -0.002504354486013214,-0.0001561795062451927,0.0007872756081180259,-0.0007046086415094981,
                0.0007603304771741543,-0.0004136767585431952,0.0002443747680587526,0.0005515853296249657,
                -0.00155899384667971,-0.0001905675710937464,-0.001356459934482549,0.00217985177937034,
                0.0002430955772816849,0.001129321154823705,-0.002006123499652483,-0.002260973951497567,
                0.0001292786871169421,0.0002628978436041237,-0.00109114356544289,0.001456540725326342,
                -0.002457098101440236,-0.0002067733715407055,-0.001685398189788375,0.002859633797006753,
                0.002875708417156905,-0.0009041665004348743,0.002310146870031414,0.002924633560531263,
                0.0006724611370092001,-0.001985202300299638,0.0006152452379823443,-0.001884616137400813,
                0.002758811835799657,0.002781528182488231,-0.001212347275230469,0.002999123096594312,
                0.004095806607613339,0.00002914342024834363,-0.0006405554076753516,0.001217247567925346,
                -0.0006430978124299998,-0.0003863781806535195,-0.001450973622444424,-0.0006444523908391108,
                0.001989214300732704,-0.0001139407585442628,0.001820038577145276,-0.002620816969003167,
                -0.0001139233428220632,-0.001021087515037414,0.001101107118462061,0.0004362095069060559,
                -0.00002044748909584576,-0.0001790210721590425,-0.001515604995750142,0.0008300414295856165,
                -0.001209297391771349,-0.001047800409263347,-0.0006138573491205261,0.002558552533111127,
                0.004132374173379407,0.000266327377686873,-0.001636072484814069,0.002930305214080358,
                -0.004040864006123965,0.006227342786268781,0.001326742340789062,0.003302064029473028,
                -0.006440521003084582,-0.007395229674260587,0.0004517166770024701,0.0005074129356344071,
                -0.003613144672629123,0.004091800983764622,-0.006813043190922041,-0.0006252610233629199,
                -0.0049052708969728,0.008728288853313504,0.009375365872706544,-0.003062030453416887,
                0.00807390459877046,0.01216303745391401,0.0007673403200997245,-0.004088571923131428,
                0.004024905919076513,-0.004329621179761712,0.002025316184508175,-0.002577987394167176,
                -0.003821245718325299,0.01175500259200331,0.0002256794612919266,0.01292621754462911,
                -0.02233348589767415,-0.003180260437538581,-0.01380802324965276,0.02777290342099757,
                0.0354093494062212,-0.002098095938228558,-0.005881592814292376,0.02682498310156699,
                -0.04514907846640619,0.1025231614301487,0.01349863431004029,0.2689264108436459,
                0.890650731381337,0.2181060735280177,-0.03808142274531038,0.0660224779686737,
                0.06192412971336819,0.01147222101297167,-0.02735627140231787,0.006556117342093166,
                -0.01828771508863248,0.02384082679595441,0.02190150980456198,-0.007415844548122702,
                0.0155583611222426,0.0217904342007247,-0.002428290761582019,0.001569493088288757,
                0.005216984893640134,0.0002155650097884898,-0.005695919990646718,-0.009244289253580281,
                -0.001280529656184413,0.004984176951373795,-0.00291972097417001,0.006734689542274032,
                -0.01009268237249961,-0.0006725811493921987,-0.005419807893175207,0.008307896700676949,
                0.007627536967621221,-0.002283702619776386,0.005634582546959464,0.006552391276056456,
                0.001822206642341578,-0.004816222022303914,0.001293460455544124,-0.004122102034146647,
                0.005684720029267055,0.00520491730508656,-0.002391257248069029,0.005867578865344953,
                0.006586572697642368,0.001048174347977065,-0.00268789172729956,0.001455739552674133,
                -0.001887023871116981,0.001429844781484832,0.0004407785003404197,-0.0008987762344855174,
                0.0020417714578303,0.000974229069267761,0.0005406285422372806,-0.0003817188139160472,
                0.00008070885173867048,0.0005410340561768817,-0.00149325289555985,-0.002066644035758704,
                0.0007236999112003396,-0.001951952836610288,-0.003452208264606565,0.0002541923220184363,
                0.0002335899978288216,-0.001173855400879715,0.000487624096451365,0.0006219723424877674,
                0.001784031329900936,0.0005893149746816244,-0.001898101235244957,0.0008130687107427519,
                -0.002329730694652849,0.003488127589324961,0.0005778922366245982,0.00164973431136644,
                -0.002990320856089006,-0.003302377713666393,0.0001403052223826906,0.0004171263347483645,
                -0.001359972641943646,0.001759535250624438,-0.002811461937899756,-0.000224460648520493,
                -0.001741658379529855,0.002813126483110163,0.002691969789247427,-0.0008045867241607198,
                0.001951654134448302,0.002326451707653119,0.0005317248262160727,-0.001476682046802583,
                0.0003850036936829506,-0.001267892610886837,0.001872352641101342,0.001933575013964939,
                -0.0006763043411831676,0.001518070366685083,0.002438973109432962,-0.0003855882265151045,
                0.0003486388132540345,0.0007390529311127869,0.00009898790960772237,-0.001003709851746023,
                -0.001650150241652217,-0.0002240916869889953,0.0009513516879481676,-0.0005929336643058369,
                0.001427493527939655,-0.002225986488578112,-0.0001516399855611742,-0.001265783343097786,
                0.001971614992716058,0.001819096438649202,-0.0005490237441212694,0.001364732325042439,
                0.00149639178791404,0.000550439122034463,-0.001388616166337202,0.0002263980806989827,
                -0.00119608722773995,0.001896577891009512,0.002003695442105651,-0.0006596794348027782,
                0.001522181433609968,0.002394652187433033,-0.0002346721543253904,0.00002337359582381079,
                0.0006589139724811142,-0.0001511391130293895,-0.000444379132290744,-0.0009172166010867669,
                -0.0002522521672248166,0.0008120743283728542,-0.0001622619186840758,0.0007950229042948752,
                -0.001125574771300707,-0.00005698061984026417,-0.0004722079048044581,0.0005957757045671076,
                0.0004085044092995715,-0.00009853866932397938,0.0002004388039635317,1.138742010991045e-7,
                0.0002199423598694207,-0.0004012136224796264,-0.00009560049776024254,-0.0002659307309001768,
                0.0005670721988190914,0.0006900931467879611,-0.00008081295232173788,0.0001164452215543805,
                0.0005128598927177369,-0.0002242487921429254,0.0002398193507995029,0.0001165024694403042,
                0.00005107024957107143,-0.0001202636725608546,-0.0001051331781950655,-0.00001649318340921004,
                0.00002785332068580491,0.00005471403313378693,-0.00009667362770881605,0.0002304111815274309,
                0.00002446746190204686,0.0002350183941142219,-0.0004544521592317944,-0.0005089309047530737,
                0.0001752986094566337,-0.0004844481528336571,-0.0006630689231402082,-0.0001512544934500453,
                0.0004762138063090883,-0.000173586575165717,0.0004881172028172609,-0.0006953631890503609,
                -0.0006710636142090584,0.0003474092417477284,-0.0008975513804983433,-0.001073782367065868,
                -0.0001441911996594866,0.000402749545251214,-0.0002949797682127101,0.0003063249708723016,
                -0.0001456227692892117,0.0001003241217011435,0.0001911148218169522,-0.0005026386831432279,
                -0.0001181420739635343,-0.0002911660141212351,0.0003670131059414427,2.142270048511971e-6,
                0.00004275265408440108,0.000092098420104886,0.0002631264763375878,-0.000110248959318329,
                0.000331659871182222,0.0007295684222404806,-0.0001279284693875024,0.00007611264391258438,
                0.0003120351047406284,-0.00004222785272974589,-0.0003241748311060841,-0.0006694469359488976,
                -0.0001541809999816025,0.0005626251376693132,-0.0003075182493166317,0.0007877625573809462,
                -0.0012072017678431,-0.0001760576786048058,-0.00060761036859196,0.001065327680970034,
                0.001136937580931714,-0.00009286121266248248,-0.00001212528786758352,0.0005107706835390731,
                -0.000510637047350766,0.0008265658987992449,0.0000752651811423576,0.0005616753027214653,
                -0.0009644283983845579,-0.0009958848023021733,0.0003102326905892532,-0.000777469289767817,
                -0.001118628800358639,-0.00005951258799679879,0.0003226291287208634,-0.0003118070995232912,
                0.000305767560865907,-0.0001247291374126672,0.0001744925605012326,0.0002216798557843056,
                -0.0006331282536773764,-5.130486629061289e-6,-0.0005916295871193322,0.0009296902993312991,
                0.000120298033247458,0.0004654831308631012,-0.0008307923716463408,-0.0009273370866120558,
                0.00004724786496753686,0.0001116535855581989,-0.0004151700760861554,0.0005440246410318271,
                -0.0008904547347752099,-0.00007264317713950222,-0.0005740296314171785,0.0009427675417814091,
                0.0009160786959889447,-0.000277713313112038,0.0006825291629338884,0.0008258966836584315,
                0.0001869296791353846,-0.0005254187965885775,0.000144660618462373,-0.000455482697605114,
                0.0006605477113831464,0.0006675178595866958,-0.0002457099470104133,0.0005564999333285721,
                0.0008391618540039548],
  // n goes from 1 to floor(number of coeffs/2)
  getBase2Fn: function(drawer, r, hn, dt, colorf) {
    var coeffs=RotorAnimations.base2Coeffs;
    var halfmax=(coeffs.length-1)/2;
    hn=Math.max(Math.min(halfmax, hn), 1);
    var offset=halfmax-hn;
    colorf=defarg(colorf, function(n) {return 1<<(hn%4);});
    return RotorUtils.getDefaultRotorFn(drawer, hn*2+1,
                                        function(n) {return r*Math.abs(coeffs[n+offset]);},
                                        function(n) {
                                          return dt*(1/3+((n+offset)-halfmax));
                                        },
                                        function(n) {return (coeffs[n+offset]>=0.0) ? 0.0 : 1.0;},
                                        colorf);
  },

  TestPIFreqCoeffs: [0, 0.5, 0.25, 0.1, 0.05, 0.025, 0.00125],
  
  // There is a problem on Chrome when dt is exactly equal to pi, let's see if we can find it
  // Yes, STCircle called tan() on dt/2, which was bad when dt was near an odd multiple of pi
  getTestPIFreqSpec: function() {
    return {coeffs: RotorAnimations.TestPIFreqCoeffs, freqfn: function(n) { return n*Math.PI/3; }};
  },

  TriFlowSnakeCoeffs: parseRotorCoeffs("{-0.5542462962455832+0.2118413127223185 I,0.1295974670774961 -0.1717329693698436 I,0.03980882668573128 +0.2086626824754133 I,-0.06555011120411201-0.03573922190513552 I,-0.005576667898833564-0.03154887293607099 I,0.01297399218643278 +0.0620491739644784 I,-0.03850381822673776+0.02727417010976463 I,0.03465776571571852 +0.0145847673572052 I,-0.01572130177487293-0.04933056573671502 I,0.007783196099711141 +0.008623889407155078 I,-0.003598857158930216-0.01322118516134766 I,-0.03802074832821237-0.03226204139094693 I,-0.01553715033881818+0.02032803384561738 I,0.01576681372655875 -0.0274680542463637 I,-0.003363297221585933+0.02976569897440502 I,-0.02306153544705971-0.02000425574154976 I,0.0391385117995438 +0.007580477219011236 I,-0.0278132800126223+0.01109650150218689 I,-0.01146991765060083-0.00001846889028025581 I,0.021549440201455 -0.01019026543790942 I,0.01979729879530864 +0.01887074731242762 I,0.007940393919230848 -0.01658704173607436 I,-0.02339264322526065+0.01227103312612604 I,0.002563842405016903 -0.003110909665683201 I,-0.0003961361222134825+0.01126481564723887 I,-0.007004010473459091+0.0104742449453136 I,0.007493514137013241 +0.01148301861477468 I,-0.003040977390178347-0.004151232661249615 I,-0.0002927822841578881-0.00335205950162669 I,0.004485940091405546 +0.01081653662658929 I,-0.004665919577935661+0.003016802048600516 I,0.001290128683891628 +0.009269268122336237 I,0.001729687963665733 +0.00002624139309155811 I,-0.0128821056238452+0.002242594575750187 I,0.005008435860748981 -0.01048610197119811 I,0.002197320385085153 +0.005445155708521709 I,0.01015438121748925 +0.001476027205945193 I,-0.003422356712334888+0.00005475170374071079 I,-0.00007547612434538664+0.00142269530714859 I,0.009384132729079717 -0.001284412767313683 I,-0.001076521295299934-0.004309233617346235 I,0.002021015870644393 +0.003407993481711579 I,-0.004048875566085926-0.0001346588556043221 I,0.004362565810976707 -0.00401165939539098 I,0.002125355854713809 +0.007107570660761463 I,-0.0005132600395102364-0.002054645917982942 I,-0.001053233947612524-0.002319855902771462 I,0.002464742786816698 +0.007872437136329204 I,-0.002750968514676302+0.0005879069978803011 I,0.008545950802800965 -0.001059293534710536 I,-0.001903070078771856-0.006424094812278786 I,-0.001126963672997775+0.001688798631962248 I,-0.00366535345871405+0.006468656663965455 I,0.005930122334554215 -0.003713775174468292 I,-0.006081221725328237+0.0007070975370741523 I,0.003183719053105165 +0.005261500275928303 I,-0.001383359368573848-0.007129879800913327 I,0.001702520548506056 +0.005279694047369376 I,-0.01125024322056856+0.0006242674271956933 I,0.0002293266404984124 -0.003757134962085137 I,0.002776024902459632 -0.0003302011690660794 I,-0.006064226482084095-0.001105959461902557 I,-0.002405899727067798-0.001449061999180867 I,-0.002426408253657655+0.006111616024553039 I,0.005144207707998546 -0.002585466104358235 I,-0.0005366759177123354+0.0003246814679853571 I,-0.001018563076965694-0.002831332277264155 I,0.00197094547677172 -0.003447133327552795 I,-0.00430002629755194-0.002540684271838635 I,0.002750467911309908 +0.002065905655235713 I,-0.001168602271981175-0.0008329060758765526 I,-0.001581519956901992-0.002044001994866661 I,-0.003363693634824912+0.0002281770155841933 I,-0.002874907385232173-0.005489046352581683 I,0.0007377362292441878 -0.00005590842024700292 I,0.00160367955979767 -0.004943539335322604 I,-0.0006353659661476428+0.005968474184552829 I,-0.006128615635791194+0.00229524300533421 I,-0.001940837945809238+0.003114909644834965 I,0.002098563600644485 -0.001686730225992824 I,0.001831616613295744 -0.006579840724181506 I,-0.007484471438165344+0.003526514174121272 I,0.004659143446805186 -0.00006089094149064878 I,-0.002300648507867669-0.00426059081640773 I,0.0001942518807306743 +0.004786432616543564 I,0.0001251815020141971 -0.004236585464111005 I,0.005546754174631742 +0.002038285722979036 I,0.0004067365690061125 -0.0003935951763945266 I,-0.0009192899181308414-0.001031239134905831 I,0.003619103856843104 +0.004189386166030283 I,-0.004311604320791346+0.002191667809534922 I,0.002418030194611765 -0.0009151344612602872 I,-0.002791230391421744-0.005051977303744343 I,0.001295524315229761 +0.0011162736896946 I,0.0001647390657612405 -0.0017110017110912 I,-0.005732985110278822-0.004120822798545917 I,-0.001847596717925736+0.003008115158604652 I,0.002122703352181438 -0.003997806978456996 I,-0.00058281574939481+0.004219967402741345 I,-0.00328517494723679-0.00304278766859997 I,0.005174396439140065 +0.001806079515975476 I,-0.004730759850709263+0.000507781718820277 I,-0.001543916445078511-0.0004820602447015572 I,0.00355985122169378 -0.001366202345636236 I,0.001915191422377007 +0.005461947358491269 I,0.001284439833541289 -0.0007062619371633841 I,-0.005208159377524544+0.0004209308898845236 I,0.001295986707988768 -0.001018642832070855 I,-0.002430644436481422-0.001278939723325667 I,-0.006077754995499663+0.0039857664507853 I,0.00168648137296502 +0.002687840778660831 I,-0.003002386052435907-0.003652969535729224 I,0.004315834936616101 +0.002389282501460247 I,-0.005093027906326593+0.002008548279279046 I,0.003570204475491155 -0.005778813110325074 I,-0.001078597214765185+0.004821810481853347 I,-0.000801304355782429+0.001997848935604315 I,-0.0002792243689839049-0.005996979453562795 I,0.006144404015734993 -0.0009467738467346941 I,-0.003583188786414298-0.001901330149325329 I,0.0004350698412564297 +0.007288921753529854 I,-0.0003045493178732599-0.001587944324728911 I,0.001823006959656107 -0.0002451247977462624 I,0.001266282676125868 +0.005088636542665804 I,0.004751864854803924 -0.001808083873174228 I,-0.002842747048870531+0.0003999730096329412 I,0.00002373293524506297 +5.300255887199416*10^-6 I,0.003200046650152743 -0.0005907490448360767 I,0.0008350471995679906 +0.0009029452196044699 I,0.003552966249395728 -0.0001185539089887299 I,0.0001180192328537011 -0.0006079690969198313 I,0.0006243618752757333 +0.00472945130647442 I,-0.004243009598520542-0.001506582253574565 I,0.002022532386515806 -0.0004856500901333102 I,0.00007416768416141929 -0.004312051984880055 I,0.0003204156351192414 +0.00124657593212671 I,0.0002359559358836917 -0.0006980259327920158 I,-0.002168680299874932-0.003268874689754701 I,-0.00132009935209076+0.00102117984946562 I,0.0007454403194598404 -0.001917283203993203 I,0.0008949352053852143 +0.002560290351666168 I,-0.002768913266433332-0.001665681406025779 I,0.004204031267389635 -0.002148820607832324 I,-0.0007769803489065799+0.001362289108563727 I,-0.001202577328174233+0.0008571330080223396 I,0.003554929718344123 -0.001727140446494624 I,0.0007702181234406758 +0.0006139170543205653 I,-0.0009515276490700917-0.004498805962217104 I,-0.002761398812856888+0.001399481631507009 I,0.0007984530937028548 +0.0005902708044599827 I,0.003791974170317947 +0.00165003273311475 I,-0.002047833194101942-0.003280622996065994 I,0.0006249487301439224 +0.00316339142890207 I,0.002771214348871879 -0.002090171141314328 I,-0.003776457564433+0.001399793066287759 I,0.00258512278014556 -0.001405046346296686 I,0.001396517388343733 +0.005893715075380825 I,-0.002109850861833258+0.0003587816996040792 I,-0.0004119338294100663-0.001365538498922415 I,-0.0002576688588770014+0.003058485816308756 I,-0.0003077816653919482+0.001107011938277738 I,0.00308809441943096 +0.0008777052284741403 I,-0.001278399401234834-0.002092766225059314 I,0.0001387848815432167 +0.0001676839345896358 I,-0.001183349398173695+0.0002566970818851336 I,-0.001856957829571722-0.0004733680133232539 I,-0.0005284457722232461+0.001807776770751334 I,0.000428710883472085 -0.001728270925294673 I,0.00002810349925767231 +0.001423281019370364 I,-0.001242634098082566-0.00001054524305456441 I,0.002086227448260291 +0.0009207373346648163 I,-0.002031667982146216+0.001955621103841544 I,-0.0005932404387211694-0.0001359510134583323 I,-0.0005179361445145473-0.0002698695437593146 I,0.002434033858310361 +0.0001816224686397333 I,0.001495244554955922 +0.0004798386645664402 I,0.0000157688873377037 +0.0007757927151715154 I,-0.0004838955385258634-0.0004782464381132862 I,-0.001554387726781957+0.0004248460039468584 I,0.0008930083129180693 +0.001869877775678577 I,0.0001889284219856046 -0.0005698532707014345 I,-0.001007813953342592+0.0002177691369624957 I,0.001097150357503419 +0.0003058274762908908 I,-0.001065318069782189-0.00005010460865141062 I,0.001660165960622572 -0.001190040391253887 I,-0.0002935606612348181+0.0008042264964264653 I,-0.0006853166237943568+0.0002801471737004242 I,0.001700918088810669 -0.0006047396491333619 I,0.001175459677111592 +0.0008875298922480374 I,0.0006006613241576093 -0.002032088814670165 I,-0.002219634925116207+0.0009919609590565393 I,0.0002511368881608543 -0.0001188790752271285 I,0.0002111963088026634 +0.001547185036735498 I,-0.0007128622533903582+0.0008529632667806354 I,0.0008976550170237703 +0.001570248101547943 I,-0.0003515092213211854-0.0005514650720242145 I,9.068394861061762*10^-6-0.0004632020496402137 I,0.0006241428125538337 +0.001626928134577089 I,-0.0004860058030944566+0.0003354364205362316 I,0.0002944385863336852 +0.001673504670631126 I,0.0001807203508660949 +0.00001868079721095826 I,-0.001943460484613228+0.0005949990772929346 I,0.0008643768086579357 -0.001839959340700851 I,0.0008851502982025741 +0.0005996645845537441 I,0.001400329076904528 -0.00008182554758262835 I,-0.0006192889090724095+0.0001274805290541232 I,-0.0002990556998936697+0.0008145878462897029 I,0.001618989567800225 -0.0004392211513738836 I,-0.0005742484479810639-0.0002063290237800867 I,0.0005202888732131337 +0.0005146804304601727 I,-0.0005612496324451502-0.0002491297554772352 I,0.0003743567660132978 +0.00001624160256346222 I,0.00002523900756780705 +0.001045250490076786 I,-0.0005991283573369593+0.0002888837752676198 I,-0.0001123453438316482-0.0002781799119172121 I,-0.0005542687432589875+0.0006285136758206164 I,0.0007224733600294126 -0.0002440981089113811 I,0.001387165432743708 +0.0005418582943699803 I,0.0004156964624013339 -0.0003717584797484783 I,-0.0005046393744315846-0.00002617921261311682 I,-0.00128402451997797+0.0009006178141706812 I,0.001681944854397375 +0.0003059441747430417 I,-0.001016245656387929-0.000510781722685736 I,0.0001083155333594903 +0.001086642664177778 I,0.0002660887315987447 -0.0009100216002499032 I,-0.0004863163373418626+0.0005722300839884556 I,-0.0008491851165696327-0.0004143086114357989 I,-0.0007343555576941183-0.0003370632923295883 I,-0.00001904124736707083+0.00009144368275331569 I,0.00003084198211409951 -0.0012887926920443 I,0.001379976970966026 +0.001107948713258998 I,-0.0003136874482219703+0.0002404635430426539 I,-0.000792159666681617+0.001527796095997526 I,0.00001254328965587862 -0.000705918152270556 I,-0.001014012993534126+0.0001121488769305769 I,0.00008882693179871666 +0.002965497050728051 I,0.001291881031242859 -0.0003710076817753817 I,-0.001863556338499571+0.0004815234648212695 I,0.001463065526859104 -0.001055720529945751 I,0.00007638207721144305 +0.001760054236820245 I,-0.00133644996387664-0.001815990903961791 I,0.001632822100016766 +0.0006801284685203059 I,0.0005048872275626729 +0.0003725581439112564 I,-0.001500775388403287-0.00007639387938555159 I,-0.000234077669550726-0.001674187155894608 I,-0.0005449154352513477+0.0006997531272082798 I,0.001590734318961231 -0.00009981933927271363 I,-0.0003111344837958508+0.0000609434278959956 I,-0.0001864103778836033-0.0004998245745423096 I,0.000697109572077828 +0.00004347754202173386 I,-0.0001678378526990984-0.0008453558826568333 I,-0.0004199713737003141+0.000278593066596338 I,0.0008233507681800658 +0.0001929899617694899 I,-0.0008809722804228224-0.0001675262905663961 I,0.0006244976061122822 -0.001665442380132798 I,0.0003346753085024186 +0.0002794809390242748 I,-0.0001818779955104993+0.0005233437281425929 I,0.0006850313329407698 -0.001250742283778559 I,0.0007401857007009058 -0.0002010452644401007 I,-0.001080044906789352-0.001078424967442954 I,-0.0001167627065555713+0.001526044585363661 I,0.00003201896443395332 -0.0001820187635288688 I,0.0008956629526946937 +0.0001352996997748185 I,0.0002955775928668974 +0.0006901002211309502 I,0.001109581045528319 -0.0003204530589646195 I,-0.0005207183169106454+0.0001111948378024142 I,-0.0000709785957986285-0.00001349023791483551 I,0.0008676437335388975 -0.0001594425163080192 I,0.0003965813814683503 +6.243917935336428*10^-6 I,0.001121519445854375 +0.0001643781152770168 I,-0.00007466393410805035-0.00001851919136641865 I,0.0004534810025364563 +0.001182817380924019 I,-0.0009180353442392803-0.0007955990269315293 I,0.0006208394104715362 -0.001002370877863169 I,-0.0001689012801564919-0.0008348105971456414 I,-0.00002765436956621824+0.0005106713957369926 I,0.0006125929700434406 +0.000988669121233958 I,0.0003938523461093226 -0.001458282484085374 I,-0.0005612754189592932+0.0005993304694333845 I,0.000832370263170621 +0.0001844279458195204 I,-0.0006499002118158102-0.000493181624985331 I,0.0005423364041712828 +0.0004366885458579346 I,-0.0008639524072388664+0.0004930640887239757 I,0.00004905940427712945 +0.00002498685871955791 I,0.0002394689387355363 -0.00001310750121253631 I,-0.0008497230367330938+0.00005235594234941636 I,0.00001375340034119539 -0.0007337465113857202 I,-0.0002286162969654429+0.0003830846346870092 I,0.000886694029970615 +0.00009432578668131214 I,-0.0002013211169098256+0.00007238278602174582 I,0.0003518126522177497 -0.00001124375481630733 I,0.0007629137056804216 -0.0004402202031972543 I,-0.00009770369051314814-0.0004488127376942452 I,0.0003265867182772528 +0.0003815017347615698 I,-0.000532614083813703-0.0001121516179579216 I,0.0006264661763044324 -0.0004670089654276823 I,-0.0001543167057129776+0.0006555504266429009 I,0.0002467484751778038 -0.0005982733695103064 I,0.00004845493372507026 -0.0002133052625068771 I,0.0003018391470191454 +0.0007843329861792623 I,-0.0009785053263059078+0.0001016404340300408 I,0.0002488430844414409 -0.00004838662463547349 I,-0.0001633031964030828-0.0009187936378619574 I,0.0001681900803555046 +0.0002722471051337946 I,0.000260088427577591 -0.0002289030104571482 I,-0.0007391763804448923-0.001001496677067214 I,-0.0003337795379672672+0.0004533382971264705 I,0.0004718989796912787 -0.0006354744219365707 I,-0.0001799109272711948+0.0007848592053052287 I,-0.0004840321333865526-0.0006243385516312051 I,0.001226882806790386 +0.0001785806927823578 I,-0.0006271398957214885+0.0003537922287015805 I,-0.0003985446567584256+0.00004016430931256409 I,0.000885887269425852 -0.0002163153885579254 I,0.0005658783508347344 +0.0004643067262172353 I,0.0002913325860192454 -0.000998204075028179 I,-0.000991112974279362+0.000468988849700584 I,0.0001141787081997337 -6.248280400479488*10^-6 I,0.0003426811306565802 +0.0007795114842520439 I,-0.000230486946088912+0.00008700775141541095 I,0.0003761622175445447 +0.0007570719417285059 I,0.00007402981550904047 -0.0002229813460368698 I,-0.0002960962232924581-0.0002411125278533888 I,0.0005640258563372272 +0.0006559126783412018 I,-0.0003417869867335518+0.0004753481886085793 I,0.0001644870181759268 +0.0007111228681764536 I,0.0001337506001726707 -0.00003217253291507817 I,-0.001114650517315264+0.0004097977401625255 I,0.0004172430871065293 -0.001147455816614978 I,0.0003739017473458834 +0.0004138600296403779 I,0.001116443029576713 +0.0001046339350301689 I,-0.0004425625110734827+0.00005035451088039542 I,-0.00006970914597465648+0.0004099694681527007 I,0.001516768916406162 -0.00008600608283170988 I,-0.0001764667037054936-0.0007350640560107661 I,0.0001048321254897461 +0.0008040329958578147 I,-0.0003027750309611643-0.0004376390105303097 I,0.0006468104941930495 -0.0002114003742776307 I,-0.0002230086665200256+0.0003349080672878414 I,0.0005273091570970512 -0.0004457938404477172 I,-3.100876081599667*10^-6-0.0001518608031646048 I,0.0003883873975553166 +0.0009820590084752745 I,-0.0009017525354322592-0.0002801629417620721 I,0.0007130639141983851 -0.0005333463682790016 I,-0.000216414618707048-0.00106232103240541 I,-0.00003103775649097028+0.0004639322092122256 I,0.0002182120878142908 +0.0009303628566410465 I,0.0004721891881768995 -0.00150412607801559 I,-0.0009922862649963558+0.0005554937266233468 I,0.001023164627081994 +0.0003825062557661284 I,-0.0006873881501346501-0.0007315636765805011 I,0.0003644438745803643 +0.0005466740662998328 I,-0.001401972390873472+0.0005413278511937322 I,-0.0002960054895986433-0.0003563365959823343 I,0.0003600372030735553 -0.00003535252436286579 I,-0.001018076899418722-0.0003981142791033959 I,0.00006760484300604114 -0.0003128789143775043 I,-0.0006728661569572773+0.0008348874611165796 I,0.0009072124331463592 +0.0002787775143994681 I,-0.0000780437901018339-0.00007698646583509452 I,0.0003779985108251144 -0.0006355838808731997 I,0.0001431784814217508 -0.0002192986820821633 I,0.000166325735222038 -0.0004439877649942969 I,0.0002100615993918563 -0.0001090469919648903 I,-0.0005019067397221302+0.0005870259343626034 I,0.0003562433295384278 -0.001057713686612208 I,0.001000961578091328 +0.0008559891667638081 I,-0.0000963140349753139-0.0002991692852538763 I,-0.0003055793116056913-0.0002603509379436262 I,0.001050699947826493 +0.0009960436832497717 I,-0.0006251709455150379+0.0003915364121054107 I,0.0009128305891410083 -0.0008421701276473038 I,-0.0009093571582042434-0.0008326591525885195 I,0.0001364001993445328 +0.0003523223950044901 I,0.0003337468228890929 +0.0008517083300941321 I,-0.0002447198110352233-0.001198303932845221 I,-0.0004977653097245759+0.0009690445872397431 I,0.0009218350014702452 -0.0002259269413837127 I,-0.0008997636482612078-0.0001205781917168748 I,0.0005536159650972044 +0.0001683089189775264 I,-0.0003006126907979584+0.001414569067407612 I,-0.0006046418854492285+0.0001690268541466559 I,0.0000245615831143823 -0.0003296535973289815 I,-0.0006780367130948998+0.0007399567897269894 I,0.000195172462486475 -0.0002308980293588923 I,0.001079991264686871 +0.0005170041261667845 I,0.0002902509997259997 -0.0006154574036824359 I,-0.0003114620046841263+0.0001151182570924868 I,-0.0006750260680612509+0.0008049765932322664 I,0.001113691614858749 -0.000323378813015905 I,-0.0008604945753042206-0.00003642423358347376 I,0.0004303701491392118 +0.0007956757951543652 I,-0.0002085192376150781-0.001002930929824266 I,0.0002276085344385472 +0.0007993347806403632 I,-0.001563744511712266+0.000121395715167312 I,-0.00005449474254680961-0.0002959512674188894 I,0.0003921561384655689 -5.94554422872658*10^-7 I,-0.001185985392423158-0.0002884053903242303 I,0.0000974891892055666 -0.0006383762350907553 I,-0.000379504865551868+0.0009021348318372548 I,0.001252576977143454 +0.0000445896632601236 I,-0.0002973124037120692+0.00003639655331882279 I,0.0001547132643846071 -0.0000894626075310692 I,0.001227175406709207 -0.0007068264111933322 I,-0.0006275179817369563-0.0006966322279956976 I,0.0008335936799084974 +0.0007524074525786926 I,-0.001106606604279925-0.0003322289360868321 I,0.0008654976623248147 -0.0007984548667883302 I,-0.0006502006724863647+0.001753459557549954 I,-0.0005425167835147746-0.001419765157938746 I,0.00008169851547662829 -0.0006028800611587063 I,0.0003273286053963096 +0.00101968680716756 I,-0.001001861843721647+0.001223936883944841 I,0.0008432509521980538 +0.0007989966025356568 I,-0.0006415987293486088-0.001353022708073941 I,0.0001968706050520243 +0.00005049746061955732 I,-0.0009586711602181622-0.000314694241782405 I,-0.0008020935852372712-0.0001046788874130197 I,-0.00070573501608676+0.0003397245767608746 I,0.00006098743101480541 -0.000228188703617265 I,0.000465535253550718 +0.00004910014713002068 I,-0.0009836995048081608+0.0002752976335370068 I,-0.0001553213765858307-0.0005662000288649312 I,-0.0007003095025877694+6.662141540510257*10^-6 I,-0.00001081887038755669+0.0001505596907284729 I,-0.00001039058948436585-0.001122089172365145 I,0.000734527333383699 +0.0006559398757191266 I,-0.0005617253910827024+0.0001889562774524246 I,-0.0003440674426796903+0.0007809853870484449 I,0.0001205200042843012 -0.0003102518945362227 I,-0.000107574151883533-0.0002526352259308389 I,-0.0004594241214058689+0.0008156918735389357 I,0.0004665211175467315 +0.0001597334720156006 I,-0.000320249276010862-0.0002301484552325822 I,0.00006453733346829014 +9.17833085337766*10^-6 I,0.0001158661320721925 +0.0002811049609041664 I,-0.0001757265595764549+0.0002938003397173077 I,-0.0001241186519442961+0.0002829587407292073 I,0.0001013821096403144 -0.000112779180572582 I,-0.0007878455984988261+0.0001844523215328554 I,0.0003766142774510492 -0.000252627174473793 I,0.0003805494393201063 +0.0008991721633466934 I,0.0006929528890706137 -0.00003999756254418375 I,-0.0002992906138629968-0.0001910163586794691 I,-0.000984743227918863-0.0001677939929001556 I,0.0008573134019965962 +0.0009239973948703564 I,-0.0001791338239932075-0.0008718284565184451 I,-0.0006557406800738995+0.0006842898783864647 I,0.0008528919563707175 -0.0004533795553298859 I,-0.0005440954001396007+0.0003344974416853208 I,-0.0003963041090062848-0.001333103606434889 I,0.0004689143777425834 -0.0001822431389225007 I,0.00008459457401466998 +0.0002991383005003819 I,0.0001530470806309897 -0.0007065341010026806 I,0.00005652495023876094 -0.0001474719954647782 I,-0.0006862953182022854-0.0002614522949600419 I,0.000181421078828407 +0.0004754578862976207 I,-0.00001173896406835365-0.00003176868712228497 I,0.0003004386798585762 +0.00001141343262061083 I,0.0004241718349038474 +0.00004623883040354554 I,0.0000814382625584091 -0.0003565308796454228 I,-4.101693252603626*10^-6+0.0004242670044374109 I,-0.0001323458770027848-0.0003896982740655918 I,0.0003953547063664202 -3.521756022157201*10^-7 I,-0.0006081120546030229-0.00001666764831090272 I,0.0004147935656834442 -0.0005725216755898644 I,0.0001605162974374208 -0.00002511506859628714 I,0.0001221242109620192 +0.0001496703604318001 I,-0.0006879057187124051+0.00006563745940956148 I,-0.0003180778604186171-0.00007209915046181659 I,-0.0001002074673424715-0.0002982496511437266 I,0.0001626620889221073 +0.0001396782289970741 I,0.0004459018954332537 -0.0001074978150646575 I,-0.0003882343526653015-0.000617490752673004 I,-0.00006843415305752876+0.0003133709816295077 I,0.0004048263654533846 -0.0001778964571736375 I,-0.0004903892050676899-8.24268139169343*10^-6 I,0.0004232373058703646 -0.00004836396196069679 I,-0.0004055902397343812+0.0007393734928916049 I,-0.00005867504674437469-0.0001988064391168825 I,0.0001785898880814478 -0.0002020704876679232 I,-0.0005811663294941656+0.0005199183349170269 I,-0.00043041426669609-0.000275737082205312 I,0.0002436636101239941 +0.0007227072764116089 I,0.0006711864079313284 -0.0006838713249386034 I,-0.000137604373334535+0.0001247545392405929 I,-0.0002706131603462117-0.0002280533223187683 I,0.0004219764047174586 -0.0005884503579071834 I,-0.0006966605401911544-0.0003648971704306789 I,0.0004094996366330988 +0.0003227123258607815 I,-0.0001704502712633583-0.00008924821798760125 I,-0.0001970736988021279-0.000330825959177731 I,-0.0003130414558291139+0.0001148224380679297 I,-0.0003517894721959734-0.0006725816805755196 I,0.0000560868285302939 -0.00005346254541853135 I,0.0002289321690156965 -0.0003361819332079073 I,-0.0001904964000106453+0.0005721922618064245 I,-0.0003827128593307034+0.0002542338799359593 I,-0.00008123260423333123-5.246201162771661*10^-7 I,0.0001245149579044335 -0.00006299813020759352 I,0.00004797679857659826 -0.0004222406597134738 I,-0.0002485757925714099-0.00009322409902083687 I,-0.0001120175742944273-0.000153353670129774 I,0.0001807924666370421 +0.00001075870411614992 I,-0.0002408372130023213+0.00008500549429790414 I,0.00005107223117846501 -0.0004319223276131669 I,-0.0001324474415336803+0.0004280736072477567 I,-0.0002893021356006668-0.0007201511821258086 I,0.00002222501104273345 -0.0001806779780121069 I,0.0004433548176589993 +0.00003193056142596152 I,-0.0004963927173379826+0.0009587995416955836 I,-0.0001611447402040828+0.0002514336153614467 I,-0.0006282054951119681-0.0002885720650039595 I,0.0003203758049837167 -0.0000964225573671667 I,-0.0001496228375088082-0.000648019709102356 I,-0.001193821124463268+0.0002899265986819238 I,0.0002436224100350035 +0.0002765314005406129 I,-0.0002859894649139465-0.0005721187196148776 I,0.000338690485163417 +0.0004207309466440888 I,-0.0004012077824015309-0.00008974977653691527 I,0.0003220421196662461 -0.000305710117350315 I,0.00002570760930112091 +0.00005685167701070094 I,-4.563546488152329*10^-6+0.00003968671750098863 I,0.0001959432012739185 -0.00004639471896538444 I,-0.0004346500944561744+0.0001740921295788437 I,-0.0002353248631898831-0.00003690574078075424 I,-0.0002075153774986828-0.0004308061285271199 I,0.0002457520850363069 +0.0001354745180130972 I,0.000366317846633437 -0.0004977122710264425 I,-0.001168218855496415-0.0008460519843500065 I,-0.000202801348830751+0.0006606637132345252 I,0.0005409477161391646 -0.0009501555224192312 I,-0.0003424017580314831+0.000987224631467227 I,-0.0005262229076532351-0.0006548238349012495 I,0.001141851497573845 +0.0006740548285236028 I,-0.001157787236619953+0.0003967776292856301 I,-0.000309976894158117-0.00009459114886853985 I,0.0002201198403956381 -0.0004065909460326051 I,0.001037527506569782 +0.000656405495280996 I,0.0002344453586727303 +0.0001500057894963891 I,-0.0002950581842275678+0.0007141557214683255 I,-0.00005916139660033574-0.0003483720815160575 I,-0.0004721612376109495-0.00009141328671175151 I,0.0001126123912545354 +0.001177927988174639 I,0.0004843983375885223 -0.0003399723802314007 I,-0.0005839492576202399+0.00009239962270955468 I,0.0002738356035260709 +0.00005483040893698454 I,0.00005569547660408402 -0.0001270708510884298 I,0.0003853717212050543 +0.0000305751748492613 I,0.0001486510536708792 -0.00004998095502698484 I,-0.0001453541566456867-0.0001346918310109986 I,0.0005225451691091678 +0.0005999672349281727 I,-0.0003434589613062357+0.0003815108083829439 I,0.0007864294850873793 -0.0002769008561958566 I,-0.0007157871990470527-0.0007116320726780089 I,0.00006801983417770407 +0.0001284814955471205 I,-0.0005443820182294239+0.0005287781093572887 I,-0.0001773321417682376-0.0001457485088531724 I,-0.0005575668620489464+0.0005231713374512296 I,0.0001225228433813442 +0.0001107908537622513 I,0.0002778762851650042 -0.0005925219931586601 I,-0.0004204358601200314+0.000971383490118746 I,-0.001179980466079644-0.0004857435967705477 I,-0.0003210567433829961+0.00008871278641957459 I,0.0003301861092471644 +0.0001642571410902618 I,-0.001147790006894182-0.001108398356339121 I,0.0007294078027433105 -0.0000720543992585849 I,-0.0009096488771033234+0.001147445780871785 I,0.0008581811794986832 +0.0008485041310947353 I,-0.0001106105482745374-0.0004328262098973773 I,-0.0004722337236120434-0.001000332251287825 I,0.0001005181130527396 +0.001338160276906801 I,0.0005145224765785614 -0.0009456388208808172 I,-0.0009231197640345675+0.0001472313154271345 I,0.0008417453557543251 +0.000197724335385601 I,-0.0005348331039110338-0.0002351643525941114 I,0.00036982627345368 -0.001174530024817928 I}"),

  getTriFlowSnakeSpec: function() {
    return {coeffs: RotorAnimations.TriFlowSnakeCoeffs, freqfn: function(n) {n++; return 1/2-(3/2)*Math.pow(-1, n)*(2*n-1);}};
  },

  InCircCoeffs: parseXYRotorCoeffs("{53.59450767170912,-10.47703416042127,15.0935368706331,4.913533599094313,32.29865291080515,-27.51751769713111,19.53546696641377,2.801546438744995,4.126951900065939,1.21416498984536,-1.022788215516768,-1.063805732233603,-0.2881560125930435,0.8936587678669995,0.4987773799102299,0.1587485066169816,-0.002051397826234903,-0.001230991194299674,-0.5257033866105648,-0.08526027501043203,0.08884709012210042,-0.2121410999870546,0.1573406656409539,0.1292111465947617,-0.1027458508303088,-0.06778952891168737,0.05966780160317663,-0.05599622021604458,0.04264745346023471,0.1286252542416956,0.01430125953376567,0.005539551820275202,-0.03009225101366275,-0.02324534184128078,-0.05778795039667228,0.0003152387480854192,0.0443760621497174,-0.002892543754023242,-0.01266313243520827,-0.01363453015891604,-0.02743948477689173,-0.02193582097112863,0.04296310299393636,0.01478312765351368,0.006427737896465495,0.01645179454106109,-0.01023116856371544,-0.0007485185915721254,-0.003073442455403715,-0.002372650670122334,-0.001359916031192226,0.006794364422176977,-0.0004210810585223296,-0.0000871171261097683,-0.01220365201000205,-0.01917053065040674,0.004164942165125964,0.003742269704478856,0.01184376052480431,0.007518468190235528,-0.001525283286081935,-0.006140899130058541,-0.002837225437258026,0.006138409539653876,0.003703986748369275,0.002657405290564095,0.0006845729929533342,0.001001413626764354,0.009456249836535643,-0.00233799815097523}",
                                   "{-63.93200362933003,-12.49788112751415,85.13505472658557,-27.71477324666746,18.58740382274922,15.83593021816443,-7.151646759933464,1.025604893187348,-11.42154406247881,3.360261826753148,-1.760123075161363,1.830710393790804,-0.04983982049061835,-0.1545683262672748,-0.4217065560861401,0.1342187691456935,-0.4885066244128282,0.2931402896692298,-0.4369538761453293,0.07086659244716417,-0.01613675264040497,-0.03852988826814691,-0.2760804162930853,0.2267224877705333,-0.277051487380628,0.1827926837146286,0.02127694597596352,0.01996769648593024,-0.02502211648507057,0.0754670169828193,-0.08480377228179417,0.03284849771453099,-0.03529024542408227,0.02726063324983666,0.0004853492682348262,2.647626272471996*10^-6,-0.05384766591919665,-0.003509926797871463,-0.06809415856162915,0.07331770897168792,-0.01548571631321587,0.01237967488888355,-0.01613861869926154,0.005553119857665771,-0.01826768433448454,0.04675613633486178,-0.01727132495583601,0.001263580767926669,-0.0005050395366311057,0.0003898828146006905,0.001169508870340225,0.005843058893197813,-0.03342293577549538,0.006914844664659282,-0.00997132046540582,0.01566379510415573,-0.0007926425466049641,0.0007122024919044767,-0.02119365758193086,0.01345382153166457,-0.004009277120500067,0.01614163520055772,-0.000984947310856578,-0.00213095860823151,-0.002215228438882199,0.00158930368092848,-0.004278207108359104,0.006258287926241596,0.01090286847085993,0.002695665487465151}",
                                   0.01),

  getInCircSpec: function() {
    return {coeffs: RotorAnimations.InCircCoeffs, freqfn: function(n) {return bount(n+2)}};
  },

  TDragCoeffs: parseXYRotorCoeffs("{97.74594998565652,-26.11678360023883,-7.971388635005575,-4.90812276231423,18.60463881388537,-7.434074392000997,5.203176043892377,4.284465787928508,4.332462093221351,-0.8040957355557405,-4.908003209416126,-1.794277909066934,3.044027539949237,0.4405471051515736,0.19075116555794,-1.552647137616237,0.2413882669209286,-2.301455629386627,2.068641388716631,1.352427815708823,1.450858927950233,0.7508134886113039,-1.990570912741372,1.01827819974038,-0.07485337699835727,2.783972481442621,-2.058459791731896,1.5910577359665,-2.432961500077542,0.5668533828208707,-1.230574532300103,1.226071234162307,-0.5493323549247683,-0.04970256415490745,-1.291029407688358,-0.6860614282373466,0.8216094148883828,0.7081672706013455,0.002063257900978015,0.1864766308743928,-0.5498939288068719,0.2477251516675496,-0.1752796647580086,0.9869841240921932,-0.3815561685134268,0.474177228774921,-1.147813822527757,-0.2085856477484398,0.05412409315456698,0.4187640758604395,-0.1263854514984652,-0.5043314105507922,-0.175445207213045,-0.7439723938267389,0.7182850106994194,0.3411408696612713,0.4926580560228137,0.2074968445244877,-0.5689669843114847,0.2650988099221792,0.02980722128468609,0.5431393373848505,-0.5751678478160226,0.3460868422856721,-0.3178306998431683,0.3762551419986083,-0.5206773773105473,0.2721118745340496,-0.1536132672773644,0.409931355251095,-0.5149055856920526,0.2503032927767729,-0.1461976919832858,0.3379528538058996,-0.5392311644168144,-0.02018942297349486,0.09920785854938412,0.528332912846187,-0.5590728611698179,-2.960900259794264,0.6209643125493375,-0.6155454335965652,2.900864452407047,0.3775213160952455,-0.3663734227957044,-0.2487811247715731,0.07689122094184697,0.3648696241834608,-0.02115807995782525,0.2652076064754383,-0.3388182450310824,0.1064402297019938,-0.2600316280037905,0.02936533196671987,0.009060128215104483,0.2449916739419881,-0.1776716478871462,-0.3180030904947269,-0.05760504094197275,0.1672108345438797,0.2704748769102952,0.1709137217328154,-0.2938501231205368,0.2962101309832851,-0.1173505878807185,-0.006624222805955307,-0.3505301047074689,-0.2418432928801832,0.2568635870823255,0.2866102024593222,0.001555852889100108,0.0801031700307189,-0.2245069664189211,0.1795131363313825,-0.1207797363455368,-0.1748453612626669,-0.08118741096926191,0.01406464528953411,0.1849229154415792,0.06166536717976476,-0.09403025034833626,-0.009995790456877842,-0.006641387766755783,0.07353122928157704,-0.02597459210522792,0.07829089369467095,-0.07028913845298673,0.03995943980445633,-0.09336017127730178,0.001872304400367648,0.02527727365734181,0.1641947532811366,-0.1250371830625146,-0.7221882965677806,0.08817969627335791,-0.06939079340582983,0.7382955168938018,0.2003541995930067,-0.2418515061023783,0.08860877304566121,-0.04147669030954136,0.1264704369698219,-0.2116672874803562,-0.04703272762726057,0.02991953654731367,0.2014701828914039,-0.09135334574066642,0.05688258415289765,-0.1300727977757292,0.133868911134762,-0.0983392068248917,0.2180211826721976,-0.1222927342419565,0.174568290050334,-0.2852467320016296,-0.04403098409241843,-0.03042343839952931,0.1593442829556482,-0.05348152165975351,-0.1667360835367348,-0.07952417770910625,-0.2452379003278912,0.2423859726693944,0.09779784911666584,0.1654346791676002,0.0622352961427847,-0.1757782283131124,0.06436714568673213,0.02455099217649394,0.2193993875926528,-0.1841325830562891,0.1162428452684021,-0.1463775062977876,0.1141525081238781,-0.157516993306474,0.0975861657036011,-0.05145040389903744,0.1189527518259859,-0.1713075069195309,0.05968354507445318,-0.02723897283918431,0.103669286076763,-0.1551459245752732,-0.008173132313972343,0.0252908557163329,0.1156442231021197,-0.1458728751520951,-0.7346054911404205,0.1849034768634812,-0.1942388467555273,0.7029567172701604,0.03384668990797916,-0.009551388404163325,-0.1483433372820903,0.05275286653678629,0.06925025556341463,0.1162025973698821,0.1393922492112578,-0.1554457541915819,-0.08902636421648141,-0.04933942355394971,-0.02670660007796563,0.09207206008807849,0.0157895216482712,-0.0102387643122672,-0.3220213943572216,0.06363479052781333,-0.05370170527170547,0.356094876971075,0.1310488520871732,-0.1299956392296005,0.03406126178659522,-0.02029730854869943,0.1512205589652293,-0.1409342485287116,0.07244141917587912,-0.06088345277529353,0.1199138580953532,-0.1654028758256159,-0.0001341511117870128,0.007712358074288527,0.1003931224515171,-0.1381681594114657,-0.4952605534386178,0.1457388166178404,-0.1536753584522798,0.4288039883170605,-0.08965799381934816,0.1056098471631976,-0.2229088536408855,0.08069055472265457,-0.08482973021871809,0.3154726402256729,0.1709855190053979,-0.1753504458763607,-0.4800574097828186,0.149271192694176,-0.1390437827307708,0.2999784320825553,-0.6041249265276858,0.412075949470872,1.279681780665942,0.06328452737199128,-0.05108793032969632,-1.276450645296636,-0.3858234079744314,0.574281781217258,-0.2464261009968063,0.1209367105006828,-0.1478812104169545,0.4093101390115257,0.1165004251271147,-0.1143539422613657,-0.2515228497430009,0.08979440644481838,-0.06298391615273644,0.16714531989101,-0.1085994878177173,0.08741686459993837,-0.2503172597368378,0.1169989448429764,-0.1272835543054145,0.2994339585747241,0.06031763046966401,-0.0165660008586589,-0.06518351063029405,0.02255224646087571,0.1083225296134029,-0.00494741391001186,0.09665696532741092,-0.1048338010597007,0.01573552070224395,-0.09680143992288343,-0.009305038038717272,0.03758568150489935,0.02898255083116164,-0.04865633577717897,-0.1702959028181516,0.05856814639559084,-0.04535307738748907,0.1445348765025555,-0.01570357146851398,0.02245206033443967,-0.05624376154947498,0.01972576812115097,0.002474210770548663,0.0686647544710208,0.07097523793738468,-0.06357826414146654,-0.09960887114777436,-0.005347540495604538,-0.03873724645514363,0.09198255309037269,-0.08164791918101122,0.03993150570378712,-0.1225652393331693,0.09707780896047076,-0.1058284432824931,0.1349870452267844,-0.01231651087353569,0.06099906154784865,-0.08880484805997298,0.03618441469476157,0.04707047273471812,0.06647296148665668,0.07133436706411136,-0.08470644433624067,-0.007401122895168704,-0.04281736175275092,0.0003188573390262622,0.02147753089454963,0.05129687294835845,-0.03376044046431628,-0.09976218860402584,-0.005257837410642422,0.03978701709863923,0.09013538795250221,0.05160852616220873,-0.08310788678166041,0.08943023058427553,-0.03404931826061565,-0.001477944589526216,-0.1091668073628015,-0.0822123947804605,0.08388081565379835,0.1020760837318912,0.0009175325301709087,0.02995141789339673,-0.08150620686059891,0.07049251115093673,-0.04599819819018063,-0.07823537465754694,-0.03073559084164217,0.001530646648859983,0.08377804478062556,0.02248693294965499,-0.0360749113269105,-0.018337471573507,0.001151516918458669,0.02579814318501189,0.0104151827275024,0.05402817349980812,-0.04771390171740905,-0.04596304560269562,-0.02525002018879572,-0.02166126439671225,0.05457014444192823,-0.04009310504882605,0.01307358043920927,-0.04496741011724439,0.05317098037475874,-0.04357229559842411,0.03818214912137536,-0.0217084847591313,0.0431881251664832,-0.03851831723046698,0.01719827469656233,-0.0003031497715932129,0.04281443844765024,0.0226208364777582,-0.02612916867977683,-0.02547499362146578,0.0007847724800663323,-0.006925836030651025,0.01985479447592957,-0.008813247521054298,0.006407314190937635,-0.04318930662799849,0.01486593658345645,-0.01825085698877308,0.05083972100136783,0.01363120654947218,-0.007436585129880118,-0.01038373741224629,0.002846478164908238,0.0262945123501226,-0.004945408780064406,0.02533992322574841,-0.02445149465360427,0.006303936512917809,-0.02847626379915829,-0.003573809983018263,0.01220226722386268,0.0111935186063463,-0.0189932129351311,-0.08411580000133209,0.02828784306783648,-0.02811324576343338,0.0734538868964071,-0.01602147022479741,0.01952314338486127,-0.04421576141542326,0.01523279507984875,-0.01220710717066731,0.06179094106362408,0.04769847706709733,-0.04361344688083078,-0.1129830991992437,0.02234123302840446,-0.03880121932983664,0.08248347497757948,-0.1473135308431405,0.09004447161829945,0.2106376618874173,0.05117812969434351,-0.05803051726967376,-0.2095841085684578,-0.09579031792322819,0.1559079301154066,-0.104486372480192,0.04595520861718741,-0.0150387539849918,0.1365284862161661,0.06988275859643567,-0.0741822217302817,-0.07761177365446661,0.002668142008311375,-0.01913864160364272,0.06144583008023625,-0.005613912390728094,0.00829417842936126,-0.1408341000366434,0.03492753252567696,-0.02038626795179484,0.152987649850533,0.05175937711141108,-0.05458241888211,0.02969356154126961,-0.01215919263838022,0.03851152820960616,-0.06668826892150245,-0.01358278033358368,0.01168206147734299,0.06597295347684537,-0.03465073490572063,0.01432633509937078,-0.03460202899097416,0.05091077284030329,-0.04381496622225332,-0.09821557041605933,0.001146824004677252,-0.01124996827942128,0.09203485594513404,0.01063110220665698,-0.01604296888413594}",
                                  "{19.05036354381815,13.22798183605533,-28.51794769687614,0.7253028123449318,-1.849423072501305,4.382294551957541,-6.322047937803731,-11.89662252909096,3.210739426511752,-2.47329525674975,9.697430899404559,-0.8567268532407323,1.055684891126017,-2.534524639054801,0.981148532842961,-0.4970415959563827,3.197468513491583,1.611471527476057,-1.826643452872546,-3.153584910224594,0.8653198980422286,-0.7721688220204268,1.947077839119054,-2.721608552730876,2.080222686151696,5.440229768542677,0.1509799641573017,0.2107521669173631,-5.276203836143671,-1.066512042828849,1.634914211598657,-0.08922507771343507,0.1483933500044196,-0.5806674293748573,0.5175498370497393,-0.3387936364386799,0.3636014902937891,-0.1805732854212968,0.4225505840204199,-0.03341197757760778,-0.03511077201470842,-0.4419641416332638,0.3081055226234669,0.9819997682301036,0.03400058741198679,-0.2610856923097749,-0.9682510230054104,-0.5616167813269001,0.7858882692484105,-0.750601108132596,0.2919350226690172,-0.2155790021988924,1.064741268791522,0.5586685079147313,-0.5825692595358538,-1.051158405747607,0.2557084213658927,-0.2585905543968412,0.6784585760432481,-0.7914764604159276,0.6272326052524778,1.51982691245942,0.03902205960122693,0.2301683363916023,-1.4681757865575,-0.08082420613289795,0.1788145046471557,0.5230719856963728,-0.1359032334968241,-0.1471762785576155,-0.5787218185431703,-0.8279881225591241,0.7599124386939039,1.158535193297155,0.07315032689017112,0.429054651774814,-1.017213692740783,1.192585067801938,-0.6591204917918759,-0.768954164482877,-0.7345987710173586,0.6348115350305571,0.7957617594233188,0.6112517358710651,-1.100035808891207,0.7274271755931838,-0.3366431720796813,0.06633820115474004,-0.8780122614843588,-0.3983393860704532,0.4591114145519727,0.4223219948890527,-0.02090913342245106,0.1009003018964546,-0.3211850096864071,0.0799866902589884,-0.07083890125241205,0.5604212628264582,-0.1799256890468499,0.1539409920775361,-0.6161237214450218,-0.1513097104649032,0.1271548921008064,0.03015741658624511,-0.003065604246662584,-0.1730673337013047,0.0917283201916865,-0.1168136807086651,0.1196020446003703,-0.05824585273527017,0.1556026627202752,0.009496383945102008,-0.03848634899400106,-0.05330708496044088,0.086649984126139,0.2859849856529674,-0.09633818237611701,0.07914547324984234,-0.2418144879157492,0.03307177633439677,-0.04483081462247799,0.1018436103994723,-0.03603783406136992,0.006282823612697882,-0.1304244102759457,-0.1249110598966435,0.1104268440364847,0.2102977315989947,-0.007006878856567215,0.08193738665248698,-0.1833769486460245,0.2206531791512076,-0.1152664024696456,0.075294552625273,-0.1912990616656946,0.2284389784711654,-0.09745567556257359,0.0947540816327098,-0.2203639098680703,0.2561628918581802,-0.1033787022559931,-0.07896749351313836,-0.244797318796225,-0.2134535579020745,0.2376068377378026,0.1000612281847547,0.08414768681283402,0.02222411490808851,-0.1136204610826971,-0.1009032020303868,0.06053734460447855,0.4126277702668858,-0.03586502600922367,-0.05268926864973826,-0.416158682267376,-0.1970005179974789,0.2637455532377862,-0.2409200617466811,0.09281421891774029,-0.07663623925394321,0.3533283996798731,0.1887896601375241,-0.1914452061214087,-0.35172082264361,0.0810853913434964,-0.08902695045644639,0.2328376440553233,-0.263066066213621,0.2036257273523153,0.4689358239220067,0.0288355954997943,0.06157986177820836,-0.4569552867571254,-0.03665387024578411,0.07243150936095186,0.151659436291113,-0.03726197251957271,-0.05348482206538656,-0.1613605478449248,-0.2457138166799398,0.2283265605011505,0.3274789302402154,0.02830791977997711,0.1185416609116572,-0.287082055560641,0.3310340645354787,-0.1865344338554043,-0.3204846978338876,-0.1773967657966623,0.1184996042404076,0.3436537548368505,0.1811043548603295,-0.2972573463507221,0.1355836068741968,-0.07151192276941086,0.07455258328056687,-0.2013916373063852,-0.02833516761281997,0.03888540759167951,0.1135280410095224,-0.06318217094619111,0.02926552357136876,-0.06184523688280427,0.110898337370685,-0.07483876287392097,-0.04248878619319539,-0.05820636157810268,0.1217104859603538,0.01748898226764707,0.08253040395623673,-0.1479138641775152,0.2308990808983538,-0.08110646303922431,-0.03935816220782471,-0.2623280093584163,-0.2681734689838972,0.2606478854842041,0.323969356088061,0.03108933941874299,0.1084009105762474,-0.2808814302790854,0.2669457615338309,-0.1559665666335127,-0.2937532567690135,-0.1367536919144185,0.01791963157036196,0.3341506442047316,0.1194352835190208,-0.1789742892479599,-0.08439259586860992,0.0003701118747333392,0.1624410749544061,0.03467649736684009,0.3376468592812332,-0.2794437703884768,-0.2655191131901294,-0.1894294333245864,-0.1474167625478486,0.3623944356741488,-0.2048993214037632,0.02331755963021376,-0.8725989970595973,0.471742690713401,-0.4599660461600755,0.8619179446134383,-0.04126727765673815,0.2153923093171949,-0.3495068135439277,0.1444435302569561,0.1518929045396483,0.2750325610668117,0.2550735974615412,-0.3084592474652019,-0.06927926758113971,-0.1251607381530616,-0.01219420888107284,0.09800805882632523,0.1172450298354082,-0.0784007135570139,-0.3176624337681364,0.01257758485817712,0.0670298665008425,0.2997610838197662,0.1320047514301317,-0.2023697764474076,0.1778515308170449,-0.0711983152542909,0.0181024196093296,-0.2303075277558224,-0.1387649978256555,0.1454886725476319,0.1815162029901717,-0.01300117815324026,0.04866648723121825,-0.1361954770935628,0.1092851833544703,-0.07668037839555368,-0.1017368380609413,-0.04777653012661418,0.01032071376730101,0.105349707315135,0.03624492567821412,-0.05541902052144384,0.001066772265420666,-0.005608970927334047,0.0393667996065929,-0.02306183489418426,0.02857903092212443,-0.02755095878444309,0.0411603709621787,-0.04780923232343328,0.009971682658451083,-0.006971959931882031,0.1012157012307893,-0.07176888516237213,-0.3090353470970894,0.01689694949730866,-0.0001465025265033158,0.3091268583354753,0.09429819530300776,-0.1298572139518655,0.06920023393677063,-0.03043162803012686,0.0371261167203337,-0.1151584899171048,-0.04638109546181744,0.04357422410921689,0.08447445824023228,-0.02132367658488,0.02229555031093113,-0.06072824867560501,0.02975121347923636,-0.02451044664707446,0.1250721180231738,-0.04430041462448146,0.04570732183834574,-0.1455189230579534,-0.03960141531786839,0.0268174059588037,0.01245255808029084,-0.002878468532549858,-0.05564143810912309,0.0253765637978339,-0.03928957095085023,0.03959978153753215,-0.02170298604285419,0.05370221027318812,0.003527408397063065,-0.01331632305534087,-0.01917296180307812,0.0326316333909385,0.110901687355234,-0.03917831552106423,0.03348917355635512,-0.09223390442264556,0.01942082873356292,-0.02479123569140958,0.03941979161115278,-0.01555604570091285,0.01644886386883818,-0.0552450072099212,-0.02037517108823837,0.02410805251831464,0.06101224003618825,-0.0252307341080498,0.01391234310333969,-0.03253244543073518,0.06226237510468938,-0.04915774682627025,-0.1588436581037901,0.007249136210644705,-0.01391797344160221,0.1552298885805604,0.02894018368789206,-0.04066000759984644,-0.002198910341836739,-0.002020002760641869,0.01911645476062,-0.0114228634698957,0.01320396847249604,-0.01453715767494441,0.005558748295716396,-0.01464616892693684,0.001264577006529848,0.001716757405330684,0.01751879254766285,-0.01171522440496934,-0.03588655042460733,-0.002219555958322468,0.01361445344763078,0.03444859318727689,0.02251494214685175,-0.03278051926946477,0.03812038274161655,-0.01389648053068862,0.002746814788395642,-0.05013770811148419,-0.03975640512097481,0.03815173662736407,0.05869696977818558,-0.002834674751418879,0.01838086246196247,-0.04659889868461865,0.04872136621119041,-0.03076208190181001,-0.05911209499814151,-0.02100220591861566,0.00267633927692251,0.06447759352061577,0.02292035603236076,-0.03288270940482135,-0.0124421761794848,-0.000722038687760222,0.03434695822426367,0.0002650019227252098,0.06159500013738418,-0.05001048935998817,-0.03085317389368247,-0.04499818261175294,-0.02270916486222832,0.05951025834728785,0.0003709313742544107,-0.02237998578144901,-0.3075809881883864,0.09897730172260884,-0.1006944139515466,0.3121098978355653,0.03431746842608265,-0.01036362518766858,-0.05302761041115883,0.02001815725517762,0.05882269106244105,0.01271327969540323,0.04655544679043941,-0.06034349503636646,0.0254845401356734,-0.0467417184229864,0.00865376399294405,-0.004050321001176088,0.05441588603356215,-0.03812872989051575,-0.0332666754673993,-0.02124284614407532,0.04977501304351956,0.01623482572398347,0.02311944570227082,-0.05436366804588033,0.0713245569523406,-0.02698366406791848,-0.02651150594195564,-0.06809730536414568,-0.07387109903399305,0.07654556019209882,0.05261050176587376,0.02710400344096111,0.01915042089111615,-0.05660914609302055,0.02529182883540384,-0.007415442270190704,0.03369866735649388,-0.03850121431543685,0.02269733584639357,-0.02303135539002527,0.01968426064863748,-0.02883290097848781}",
                                  0.005),

  getTDragSpec: function() {
    return {coeffs: RotorAnimations.TDragCoeffs, freqfn: function(n) {return 1/3+bount(n+1);}};
  },

  DDragCoeffs: parseXYRotorCoeffs("{-6.999850399052599,-21.78352842630434,27.33241815885687,8.356678844135278,-14.12879374189589,0.2383607471571052,-11.17704229834825,7.679632204481579,-2.640525392794413,0.4153715794512828,-4.051176878657165,-1.215079919814256,4.851390437874287,5.364649491135819,-4.695137789629341,4.104144088538686,-1.87573421757868,-2.020047143672763,2.975365026921032,0.3149119053827648,3.160594522835157,1.899154344138188,-2.703154106111588,-0.1151968635988156,1.908253412778093,1.764496951967527,-0.631632210152171,-2.872109914214904,2.290628162937402,0.8667915222227581,0.4043595345706991,1.03264629203144,-2.091164621626234,0.6342034533286516,0.2538630368709342,1.651017205368268,-0.3353668014276506,1.403153559190313,-0.04698708203741406,0.2366081527926226,1.505767829776348,-1.148076071174912,1.703831841619806,-0.7245544816722286,1.463500852247758,0.8445534319326552,-0.8945953469481998,-0.3541373750567987,0.5386571847996922,0.7004434522126142,0.2105747852582902,-0.4615590449572662,0.6469083565227257,-0.4742817230012762,0.9275785600151896,-1.510292382380664,0.9102808592811701,0.2480451850355002,-0.1470156707472919,-0.7779777427225452,-0.5036563600355454,-0.8739514114665362,0.5193148990812286,0.5039183594673852,-1.017002087395583,0.8069245148229782,-1.166372382080629,-0.2357946665230748,0.1618243796097154,1.123452721528319,-0.3214234398225777,0.5971513525438576,-0.4508581612898027,0.3820249215060021,0.6355588235269347,-0.1847371116561455,0.5322147702374821,0.3156031258801931,0.4673424921384202,-0.4527228485333991,0.2297978933851966,0.047881431230698,0.5908816260293122,0.8521346243793172,-0.6042721175486917,0.001520381398724638,0.4850010646407988,-0.3144266498496448,0.868871243138184,-0.7206713521683022,0.6858958286058844,-0.3514832366234428,0.34703639807444,-0.02059559681026207,-0.3137685970186602,-0.3376233994169268,0.1591911029049336,0.7721510140020457,-0.3537336648725032,-0.1491427044623374,-0.003714951295864209,-0.09878146603067253,0.7806710552639611,-0.4932568444497507,0.1910782202584499,-0.4608160293757621,0.5803530203384241,0.3871207849656772,-0.4241337613784396,-0.4204351559051109,0.1686496385432365,-0.667874209275238,0.6252410913264148,0.08900744473605793,-0.1677065112373107,-0.2540740814367427,-0.2005457447925676,-0.7293232468585231,0.4634550364926738,-0.3295733861004574,-0.1218398590955202,-0.1409026523809953,-0.4101964529396694,0.1552838119206434,-0.458613509858069,-0.04422233773534627,-0.2723286744030711,0.5142337623258928,-0.2391438970531499,-0.06349696102384086,-0.2267896115958318,-0.4164686699736659,0.608768871413332,0.3403664667404811,-0.4414765225748105,-0.08634955565074574,-0.04054609843020702,0.4814486429830843,-0.1742469178197008,-0.5519902950833238,0.2268469863769199,-0.08553492722809886,0.302012072605164,0.0993569951636265,-0.4459105028811994,0.4889975570921796,-0.2553142269617006,0.0313322742325954,-0.05299651556513791,0.3047042617862159,0.241039881459,-0.2429830336392377,0.1160839490695443,-0.01363495217187604,0.4438624560286928,-0.02545440970279985,-0.1552335532552185,-0.4814481566697196,0.4218185229724148,-0.3114174763217085,0.04947337249344828,0.3453043221349313,-0.3889130193961318,-0.03825399659096502,-0.3189312287043588,0.1643506887907784,0.1686017143739124,0.2771752298277371,-0.3489033004302169,0.01640714742441847,0.2609656342940231,0.2604301027503853,-0.03230722657263192,0.06166746645597473,0.192635951418382,-0.336082844085174,0.3227595143298895,-0.1416933592503322,0.3326278498577088,0.4017620311935877,-0.3774144234643375,-0.1906648255225713,0.2299371026263984,-0.1159273533276022,0.3179828995401137,-0.4461101089018226,0.2191966240562541,-0.01772679218097213,0.05349293342098922,-0.03508347667419547,-0.2703530376677827,-0.1150590208373813,0.1389119387644268,0.3416765894949355,-0.24288488113937,-0.3336133370689259,0.1589623139067365,-0.05607636427216478,0.2832732911906197,-0.2375003863586429,-0.1230970445243275,0.1089724869498195,0.07941433837992759,0.1588746860864208,-0.3256031168172087,-0.3554555540421742,0.3846485553979142,-0.2992838004547463,0.1153329139639361,0.04267354682653847,-0.1095046815715287,0.1753889621345464,-0.3903420538341946,-0.08829846361686791,0.1564848782362773,0.1661868513012223,-0.119898422165255,-0.3101749473169811,0.1927348428970881,0.3009911291016484,-0.05286983872943531,0.04154846040662363,-0.1528201610131019,-0.2798412466337792,0.3871711166702236,-0.04125706723325216,0.1016587897568934,-0.09429438399702622,-0.06480162661170494,-0.372683249717059,0.2403973302878586,-0.1521746488037142,-0.05383436878910163,-0.1185675245172327,-0.1368575112932234,0.2886208809382069,-0.3170888372945286,-0.1122390592574126,-0.04999012766308484,-0.001442480214505396,0.1296126091864086,-0.249928541319306,0.005168064997290181,-0.08054060929750377,0.04427593548006005,-0.1357695940577421,-0.1955112121081932,0.1739324058693641,-0.146700225624407,0.075878251275143,-0.2921533894822319,-0.05074257664243328,0.1104555246571884,0.3022480193495089,-0.2223711530741984,0.2551797543288782,-0.124227407006982,-0.1290964402630339,0.2100833804298718,0.09659713374541841,0.1969911584658695,0.07504578797126484,-0.09868202082593672,-0.1558313110756179,0.3064754901167419,-0.1076423166695248,0.07336944678610968,-0.2195842576201258,0.1561932304209421,0.003331794113521896,-0.1060356746439098,-0.03849931178014144,-0.02917681422207645,0.2346965617217199,-0.1473035760372088,-0.1858897942506171,0.1280257452849588,-0.1327951821437149,0.2163560353235993,-0.3041947836278372,0.06240513847665753,0.008188196788790882,-0.04449903874972505,-0.04780341749404999,-0.213362938661572,-0.203380485657504,0.06938060289079243,0.118086794849308,-0.1498480216772474,0.1763547616451686,-0.2803859753794709,-0.2212988227268437,0.1608209939825811,0.1798616604922446,-0.05240326177524525,-0.05157430124149817,-0.1160448775550454,0.1700662392120396,0.06246097756357485,-0.06792780350572912,-0.04193188854077024,-0.1944578317314374,0.2708508121867709,-0.1941925929850982,-0.004646303492008791,0.1845477898236285,-0.1277419708843164,0.0306136985046696,-0.2201446574430097,-0.1000646181436706,0.228476748640703,-0.06306101338035969,-0.07099596755569768,-0.1552332966852222,0.1065921194390543,0.2290438543338698,-0.2250333900844191,-0.1419391052981773,-0.002598040094318913,-0.03869964984488329,0.1398774567167234,-0.04053565213076359,-0.1038350187314427,-0.2340750658008136,0.1062811749009383,-0.03859679019900701,-0.08680511816576021,-0.01396865253308546,-0.1789834220639323,0.254052451250315,-0.2070394801081694,-0.09713208280894835,-0.06795005946340578,-0.09502924167385035,0.1707433032968445,0.09926768784006818,-0.2139893906041193,0.1064704153477969,-0.1076495945104788,0.04059534005676596,-0.05991036172354007,0.2037456421891849,-0.02934034146760337,0.06599121778494313,-0.06677242945841129,0.01788742174511727,0.1936098825522224,0.07452480023491044,-0.02107465722211807,-0.1546284658312152,0.1868819950935028,-0.2101828594551161,0.120774259232321,0.04893040197030457,-0.05294716178523795,0.03680914924594698,-0.2190855022032096,-0.02876327133366883,0.08181126950968168,0.1557283762763741,-0.1720893759957955,-0.01280713857768053,0.04264704426062841,0.1835619179715969,-0.05033769298478563,0.05467311363566687,0.03370167071589935,-0.1007605718207264,0.1470231391865645,-0.1639606888701926,0.1926492555790999,0.2035562674555487,-0.1562852897355668,-0.07626403730289946,0.00934761485876591,0.04335291311944454,0.1138066758299179,-0.2054670748729709,0.05734388761543109,0.003466716753698085,0.09290246830756277,0.09698669032543195,-0.2023467208590053,-0.0306627108143414,0.09293061206119042,0.08882715472383682,-0.006085881424823955,-0.1358678955843881,0.1164113340871777,-0.09426830672858796,0.1347781099810414,-0.2097070933602221,0.0399873136401915,0.02030390473389116,-0.01176081481003837,0.01503452162031912,-0.1701540976241568,-0.2208884662017676,0.1374193778015908,-0.08809548786887292,-0.01304878755338848,0.1168078967536341,-0.187426024775616,-0.08145575976201017,-0.08523665183369877,-0.03779584312140694,0.08058137049433914,0.03032929849706785,-0.1721395349347214,-0.02077407252840061,0.02469291823909636,0.2024472785843551,-0.1673592076398819,-0.1085593509971227,0.06767245821738699,-0.08166508180280749,0.1254425284842355,-0.08578392643778773,-0.01887425295630092,-0.09040457782709835,-0.03602471929242526,-0.06772593093937991,-0.05888070050492419,0.09108925500595656,-0.1807097923874902,0.1417522435658772,-0.1539520045700323,0.05068182296346242,-0.05970988257196525,0.003498868433832431,0.07375382441443219,0.06229786575378705,-0.06995203066394069,-0.09253789096332789,0.06491228241901961,0.1471655371327134,-0.09148272785167898,0.1145745918445098,-0.09920420435191679,0.1540415433809416,-0.001370088133382914,-0.08546586355346515,0.1220337549124755,0.05105206158886946,0.1148021159476684,0.06041339801597587,-0.04307692304933958,-0.1646510900780409,0.1649192792064427,0.0009131715063451784,0.08479381890385077,0.01520093379960914,-0.09533103759940706,-0.1797270099294667,0.1533150218240243,-0.06378974418376807,-0.009112374132216685,-0.07141295031235982,-0.03762883905545914,0.157325508831079,-0.1611504975252433,-0.07239584242520977,-0.01211948984507817,-0.00222556923688981,0.06032163915080957,-0.1172607268194863,-0.008800023355249827,-0.01871736846906153,-0.0144086355727711,-0.01519682383940899,-0.1113521241985384,0.1630110896780718,-0.1267146697603598,0.03129644202999173,-0.08522407157710302,0.07422066919716508,0.04832106950507576,0.132025799457274,-0.06131720329059906,-0.001405382806399014,0.0633009619058636,-0.1605540984376521,0.1800287780381706,0.1010720251447108,-0.03047047768005808,-0.1008247796319891,-0.0327043494388212,0.01583909091732826,0.07608851036301804,-0.0705244786076948,-0.08067015689150142,-0.001498865350563195,0.03887241208700599,0.1109522650328351,-0.1615231294532292,0.1366851588114518,-0.03449301336988464,-0.08432991212046782,0.04878549763379363,-0.04588966226112781,0.1554226493357666,0.0574723835273776,-0.1295204654115671,0.004978032029129352,0.04238508040734921,0.1110389767348991,-0.07698881882137049,-0.1589307851493874,0.1128610030744582,0.05607915898174735,0.01813232391303291,0.06103510649177225,-0.01379791750463859}",
                                  "{83.66809687360099,0.4937129801786174,-12.16478296388238,2.495384179248328,14.43334597675149,11.65404699812837,-10.80686618016573,1.652563606597628,4.391404608031663,1.019546806076275,4.467566943388138,-4.825695234701511,6.395946261558191,1.676728138701923,0.4473928560383838,-1.285501954648718,-0.4462575566058523,2.759560707151659,0.3740754819683195,0.9981013301864023,-0.5670103954124578,-0.2229643074383472,3.100334659499914,-2.840967815722808,2.778456705463451,-1.468492280008085,2.056136788005111,-0.9063523491714549,-0.6170573274907907,-1.764394872419558,0.4608324278265512,-0.7564799862033362,-0.1714938603863095,2.443296326672252,-2.364753358552746,-1.110067853988604,-0.02890658797364122,0.5406895076356992,1.357477582870637,-0.3381089521298837,-0.8334357695221052,-0.04344627753341569,1.285553292976849,0.9417413276015981,-0.9136668403081522,-0.9422788650845194,1.308051784458782,-1.765732646268794,1.305780617863222,0.04049169959131541,0.1456626607149747,0.6134168410660673,-1.464107017245489,-1.242061941405736,1.17408392204683,-0.3437945699416706,0.1731265958022498,-1.277859132010315,0.1561700422985823,0.6110851066835793,-0.578009662036086,-0.05347584732632396,-0.94162243397498,0.1269483310728766,0.4479579669255615,0.3277455318834232,-0.4998551712135474,-1.222811027585308,0.9924791757465368,-0.1513799041231904,0.1202508499639568,-0.7598615134242482,-0.2960075409919614,0.7541102483754196,-0.390986404174963,-0.2417377546201776,-0.6437844157882776,-0.9663555217542872,1.014370253802063,-0.2762051122328177,-0.4320650955125213,0.6240726544178852,-0.7618807983403125,0.05340816591160827,-0.62951997599582,0.1980219108075835,0.3679101200378441,0.218251306952587,-0.5054143588369724,-0.3367177607761791,0.6212907883976528,0.7728578125343786,-0.4540609902058382,-0.4743090636833451,0.3150119678906321,-0.7722946737695735,0.7934292014235816,-0.05072741499677327,-0.09372021593910568,-0.2370192307308105,-0.4601795402002601,-0.5366497190174258,0.3015237779123452,0.2041494551698318,-0.6438738192394999,0.08588453893311593,-0.3143099950234608,0.6858176414196923,-0.5424729437988477,-0.06785986677506044,0.1414654983204439,-0.07781972133454032,0.2692350665725642,-0.6661623000100914,0.3767293892121805,0.4161186289487887,-0.3298438893522142,-0.1167841682805842,-0.3575606686431587,0.5047932708724516,-0.02467949428812585,-0.3250751363553242,-0.1044826621432158,0.01739068853594426,0.4350360798243403,0.6243296880864871,-0.6402051564110606,0.05920273106983486,0.2088546666549506,-0.04797553051253058,0.3975221029992687,-0.03143838967807666,0.2932572898379243,-0.2929034005915814,0.2824512983413343,-0.6194104565992558,0.482307503560218,-0.2495920156624906,0.04224227088494835,-0.05931509235685879,-0.2561846471065604,-0.4244152907413304,0.01320278285941604,-0.2051042743992204,0.04708559790491466,0.3502718934239355,-0.510616875595362,-0.4577247103054635,0.05500064846281368,-0.332436464907621,0.2987398564087805,-0.2022127575664752,-0.3627516541131904,0.02451683019491086,-0.205982384006983,0.3692882785570733,-0.5529804826289905,-0.280552020484663,0.08572124204782673,0.1195969045797122,0.03597469603434349,-0.1306509289269415,-0.2096357685631139,-0.4245841838588013,0.2469573463118404,0.020910837206027,-0.2857330456245807,0.02616234873022418,-0.3683416819128226,0.5001742463668761,-0.3640626123338206,-0.1020434364092071,-0.142124915017454,-0.2421784283245803,0.3874221798371843,0.05829749357595183,-0.3046009291574139,0.0893259146877951,-0.1630580034403295,0.1614184261480846,-0.2782926106348549,0.211570529978739,-0.04642927556087706,0.3076086960982995,-0.2345114413536907,-0.1526004302247385,0.2757046316399568,0.3852541377546473,-0.04495793540056374,-0.007668467987930467,0.02398733138314887,-0.3262107370747636,0.444051499461712,-0.2187853168755279,0.1860991356834535,-0.07912925076978622,-0.1380092598558634,-0.4023254320398295,0.08307717635965324,-0.06214766650506647,-0.1429953964781047,0.08120912917877737,-0.3606425935285192,0.1240156871807608,-0.2613342785616606,-0.1089929814414058,-0.02154261097220091,0.2522035538116743,-0.1840504191224731,-0.1112320699209744,-0.1292448332527639,-0.1265273465868231,0.1089996165494701,0.2203337106042414,-0.3222548452357915,0.3129065300856702,-0.3180319622464622,0.2544627323793389,-0.07768426736152029,0.1312439793707395,0.04344810978885957,0.04743257085902371,0.1774361911774267,-0.08204804625385428,0.07665569404158248,-0.2932663595169091,0.2480901740875546,0.2501168308239787,-0.249532876649718,-0.05972506911175061,-0.1491685360235733,0.2975162953199357,-0.05934710450817878,-0.02102661324763784,-0.1262226059424192,0.03511959329949278,0.1697329510622694,0.3279195498739414,-0.2332442547958138,0.2266539838449144,-0.04638729317787937,-0.239776833680702,0.2921992895789655,0.1938666330747733,0.1123450096202955,0.1848698586383122,-0.2237312699450006,0.06938352580278641,0.2368237915542119,-0.01763989015745441,0.08782436742401072,-0.2618642832222534,0.3283753203623984,0.1201545515921993,-0.05930446859039197,-0.1038176961983798,-0.03083805062073875,0.2206639783981855,0.01582464485328791,0.1182311214851873,-0.1194396515091829,-0.03149646844650589,0.2890917174459575,-0.1077948190164305,0.1195979593803013,-0.2411775395191244,0.2680326180070783,0.2233937023014979,-0.1314892857259997,-0.1123872708595818,0.02056078588104145,-0.2864286186923097,0.3077019739393231,-0.0756395671056715,0.03705048437453497,0.03975834814469879,-0.1668993455609618,-0.2827111809387268,0.1435013556217551,0.008777523426241413,-0.02458946156900139,-0.22026257673551,-0.07758807443891463,0.1553620896928984,-0.07133123644679439,0.1112071324017349,-0.2580135759041816,0.01024224680639094,0.1423507080997993,-0.04842089183323978,0.04952574723227,-0.1898608299837327,0.1857593473254179,-0.2125323397589643,0.069084873474614,-0.2334263675077639,0.01737771488133957,0.03720288935413832,-0.1703353252679315,0.06764495623194608,-0.2383224289701879,-0.2146636587418598,0.07166088224336375,0.02466579086877385,-0.05114493743115633,-0.04164690324457011,-0.1753950694840856,-0.1246279398050995,0.005957973164473704,-0.03581925859119619,-0.1205526906659652,0.1998542811338938,-0.256516700632103,0.07226377146424785,-0.1606566655996817,0.1325904989962502,-0.01024485974254734,0.1898814852729088,-0.09491704754435354,0.09389835591345481,0.04797630215418672,-0.2346398497155071,0.2366496312989304,0.05922276228773549,0.0902928710934276,-0.1005916682132376,-0.09684819132630247,-0.124598128199266,0.1750785142979551,-0.04781703475406054,-0.1216973118978374,-0.1337996876243255,0.03263123700082936,-0.03430002063359594,-0.123095339444249,0.1081576031151467,-0.121619679892036,0.1375621571726769,-0.1918633250067003,-0.2563797029709422,0.1878313203682486,-0.02308964003986794,0.009013449214225503,-0.09043696178235124,-0.1449904718014415,0.164840335926795,-0.1293332812470313,0.07008255775627192,-0.1950434507504349,-0.1337733891486905,0.1408330988272416,0.105521107987729,-0.08995259293546914,-0.09527437726971424,-0.0750973674907769,-0.09460682100006111,0.0597358274033117,0.09248181500617043,-0.2129212004139129,0.1112493493220404,-0.1504869658271567,0.2339600938013628,-0.1500455357198668,-0.1080155911563649,0.05343275650603294,-0.004845929549477837,0.1146003380550186,0.1035534957385174,-0.1853504450639955,0.0312407860469951,-0.01886310037356755,0.1032264602306914,-0.07371579457503485,0.1878583473562735,-0.07016172594671233,0.1517259565373819,-0.0493409383825331,-0.02720616263541636,0.1664544299034741,0.1035003469697393,0.06098775144755222,0.0148383853318569,0.05531368260463837,-0.1942825387296897,0.1894607176762241,-0.08168704099207236,0.1219585738071961,0.07600368880659102,-0.1578912997764083,-0.2035910176950489,0.1004911546292423,-0.01117573974589453,-0.005453627900462256,-0.1012730838888237,-0.08483341982120611,0.1312374520419535,-0.1062851016657947,0.005010584818182284,-0.1186267470337612,0.08065978222062362,0.03024154514425658,-0.124244019565714,-0.003127537295951083,-0.1139569889557722,0.112116043614958,-0.03061321721621163,-0.1493969013655792,0.09290540088175438,-0.1263387220326803,0.1360537468696821,-0.193210442824018,0.01065652073808847,0.008884197880580452,0.1095110914479941,-0.02996328196936986,0.1521930114353839,-0.08132859445099896,-0.1722567789000296,0.124492693713117,0.04480887470799554,0.08320095274110641,0.03852640765642396,-0.1765638061778703,0.01190472668034948,0.07531647131547696,0.07539505429624227,-0.1088236875373876,-0.179709014815562,0.1192826679634367,0.04305304342502252,-0.05350530471630369,0.1607624404726457,-0.1897567950699585,0.1077025823244947,-0.09183344921866538,-0.08553308817595694,0.1159812010677957,0.1564347165437054,-0.052776370909562,0.01083548684961146,-0.06017645296810475,0.1372669571560392,0.0459428780043156,0.02484138695292667,-0.03398564706277289,-0.09412041276284491,0.1871873246380356,-0.06770469133504507,0.00728890619427254,-0.08541541887649018,0.02286065011586409,0.09929136938661823,-0.1505760435937608,-0.005587682853209083,-0.07154242959005581,0.1690378170116697,-0.07718816001665248,0.009632959031565927,-0.06308787556863857,0.01222566985748375,0.0809077666442795,0.1606366513997771,-0.1110585979705095,0.149947638477047,-0.04873839035074357,-0.1056409993663097,0.1190572534630302,0.02967150477061198,0.1063465230350632,0.12445907879402,-0.1499574566876519,0.03738332781397598,0.0635104871304574,0.1177825111304172,-0.01660486361907085,-0.1474835312849552,0.1342515768495654,0.03560232129705204,0.06241835974359975,0.03290888425211586,-0.1091770901705899,0.02129191718173712,0.02902261319040013,0.1483928655134433,-0.08303280998659622,0.117300859328188,-0.02163360871181208,0.1012851826310569,0.06149929766318049,-0.06777527977605984,0.1273030708119951,-0.04169025423193475,0.1479018218954296,0.1270615512269899,-0.06968738956747346,-0.03191148229161774,0.05270643108848488,0.02522549892023993,0.1004049589078833,0.0676456972852123,0.0009991394465641706,-0.01358477214937351,0.09051786066080911,-0.1337112679410141,0.161602980988559,-0.07637477420105071,0.08556085370880628,-0.06950166167152763,-0.002401989240924607,-0.1020499275528518,0.01566777937700769,-0.0509578191239069,0.1291306860967203}",
                                  0.01),
    getDDragSpec: function() {
      return {coeffs: RotorAnimations.DDragCoeffs, freqfn: function(n) {return (1/2)+bount(n+1);}};
  },

  BigDoilyCoeffs: parseRotorCoeffs("{-0.5792220189563448-1.042885132666832 I,0.007280590820737548 -0.1945864246136738 I,-0.2466112704025641-0.224317279071344 I,-0.1715822209086627+0.05552675314651254 I,0.01658473773071799 -0.03813431203127174 I,0.09945513410617911 +0.02641773377967594 I,0.07650772922747362 -0.02058236167582868 I,0.01204249993790051 -0.04776907957664141 I,-0.04791485124731084-0.07507493120707275 I,0.01649923060787758 -0.06485773426624157 I,-0.03806933250205745-0.0005678090147974458 I,-0.01412339763139972+0.05276395007686572 I,-0.02246166814577345+0.04455434577389427 I,-0.03210254375761815+0.0006322201181151774 I,-0.03088970204381454-0.03042414142182714 I,-0.03682908755159709+0.01050059451626549 I,-0.001503453487345009-0.008074200916322445 I,0.01313596450481051 -0.01009349220170478 I,0.02483335629354723 -0.02893038923677306 I,-0.008610259256857433-0.03027320802705819 I,0.0007799263943088454 -0.00253409653046466 I,-0.002089946338273161+0.01341192454353165 I,0.004263773091019926 +0.02756810219250418 I,0.02408222675101194 +0.01455123209487466 I,0.02005215433137302 +0.001161089854958496 I,0.004589919981303149 +0.01443355859270739 I,-0.0162371509726803-0.01063589572857112 I,-0.01888378743356438-0.000389179382049533 I,0.003457231821968744 -0.02252620283858072 I,0.02805253553108434 -0.000640828889011526 I,0.01949774180353306 -0.003731656136360195 I,0.003297957135147789 -0.01194474528611636 I,-0.01118909979759407-0.0197903860501996 I,0.01088270087435668 -0.01582953159547408 I,-0.005362388080580116-0.007336311359195222 I,-0.01474578525391872+0.001484722664598335 I,-0.01656025744402426-0.003603046898936182 I,0.004831184612590424 -0.01343119746855703 I,0.01568377612911995 -0.005963582454521879 I,0.005908826336991566 -0.01505424489565314 I,-0.006216864731626921-0.002192385963286976 I,-0.002130543294612319-0.003492225469924687 I,-0.004436961258952852+0.01630249773560583 I,0.01184498802864234 +0.01246029088566395 I,-0.003061328134471203+0.006272200981793405 I,-0.006461809155236887-0.005588162309125565 I,-0.01462167468935784-0.0009613883803203084 I,-0.003895334607867272+0.0134141398163331 I,-0.006863349412750457+0.01196182679668394 I,-0.01485006594030189-0.001840200639512408 I,-0.01247051116574128-0.003897401225893844 I,-0.003769318532197313+0.01006501695833491 I,0.01004011249017919 +0.005958418988525442 I,0.003768804591208329 +0.00897606175068561 I,-0.0005149471494829421-0.005529516407758174 I,-0.006208885228983909-0.002337571420868889 I,-0.006021038088374296-0.009811681364016734 I,0.002988839818665672 -0.0105311482913564 I,-0.001790537419005552-0.004693191133922504 I,-0.00360109504599572+0.005614423941709166 I,-0.00312190587821175+0.009300174563195703 I,-0.007464197428093639+0.001885493327002516 I,-0.006126724409651101-0.005880753102511406 I,-0.008828823849275127+0.002434918283419139 I,-0.0007665788134342131-0.001926901179605617 I,0.002087361298566989 -0.004323683651043255 I,0.005180014573314732 -0.00952949892402344 I,-0.005269596080747161-0.006347630018239857 I,0.0002671333876539323 +0.00363580953791485 I,-0.003180953196156296+0.00629603434411358 I,0.003386808909717423 +0.004256786139362074 I,0.004640693300367505 -0.001223109487026449 I,0.00784776059304665 -0.005393614241920768 I,0.00798000377849691 +0.002191827478371539 I,-0.0000596675836747111-0.003183704509339352 I,-0.00676711456416385-0.001239097786822533 I,-0.00685236580030152-0.004062388376351503 I,0.001211513155628499 -0.006012897876208725 I,0.005656528473923789 -0.005471675387195018 I,-0.001681352614613965-0.007218823509478941 I,-0.00002602631794558506-0.00139014097175435 I,0.002685894449215554 +0.0008009880555286601 I,0.005227607590934547 +0.004658732893619083 I,0.004993122299583929 -0.001625321598025014 I,-0.00239516044276428-0.0001688724280950478 I,-0.003257649086673668-0.003527277714036152 I,-0.004375544589356331+0.0002026877131481894 I,-0.002260598035403681+0.003714213770076721 I,-0.0004709743955926068+0.006665992884245971 I,-0.005064336940986161+0.002758531667606327 I,0.00292540926048842 -0.00008850545931036746 I,0.003948465053085282 -0.004483117406485518 I,0.007004252273879826 +0.00006864423366875526 I,0.0006647185563800785 +0.006876226492057362 I,-0.0007175614717856438+0.00799880686782232 I,0.005960858058294554 +0.004005924546122151 I,0.005078974314127831 +0.0006031322643122906 I,-0.00001994294329770686+0.005062155929588427 I,-0.003130813781210228+0.0008822886672027127 I,0.0009021196065223137 +0.0005857153301471841 I,0.004915240893986115 -0.005060017325769297 I,0.006779907378002738 +0.002717905219954938 I,-0.0005680940417076827+0.00210528042891134 I,0.0004575547937758431 +0.003001695460832766 I,-0.003958595582488671-0.002925955001934403 I,-0.0009080026567827701-0.004831562355728425 I,-0.005020604912029209-0.00415401968361752 I,-0.004308976371526559+0.003733051073353997 I,0.0006891968943535344 +0.003097892188972064 I,-0.000431739356092937+0.003251681654473344 I,-0.0007399551642043255-0.003041772909557123 I,-0.004245970437664741-0.0007192887582982892 I,0.0002945684108722984 -0.00554819402977069 I,0.005943230809440319 -0.002938564410483912 I,0.005723662087594271 -0.006020870786903813 I,-0.00296257174477162-0.007491627123029732 I,-0.002672512260206332-0.001857400472854383 I,-0.0003638781399709779+0.0007721518022211609 I,0.002664361447419927 +0.006375193089388675 I,0.006913303878352002 +0.002174414283925304 I,0.005157372411631989 +0.0007841523254337216 I,0.001147686936098071 +0.004256773623947554 I,-0.004007350964668977-0.001209262430163564 I,-0.004067507796804137+0.0001364237739694802 I,0.0004921113508473778 -0.006007864374685314 I,0.007101488139966988 -0.0007039607487515341 I,0.004859375147722954 -0.0009824097834112287 I,0.0008374910272851697 -0.002987242424834579 I,-0.002749748878947294-0.005010175009176667 I,0.003095204991787447 -0.003946768084838886 I,-0.0009349050832327239-0.002289652880133509 I,-0.004042335678446194-0.0005960527998039295 I,-0.004133868074540539-0.001466600326096661 I,0.0017297178090573 -0.002710771855944729 I,0.004656945657065994 +0.0001000353943240051 I,0.003464746100174626 -0.003219819965200746 I,-0.0003118050964348424-0.001532443690839351 I,0.0007783038756619282 -0.000558955943735433 I,-0.003386808223612082+0.002859977413106301 I,-0.000341908061081547+0.004460597745750988 I,-0.002504392037307303-0.000159100413707561 I,0.0006333881048213627 -0.003235564745995459 I,-0.001349828297645444-0.00404517527125146 I,-0.0039583206638697-3.329089864589257*10^-6 I,-0.004427839630452914-0.0007283912696910756 I,-0.0004380042173440115-0.004863734078137399 I,0.0004396062221015389 -0.003831204890307512 I,-0.003540050241335397-0.0009916179553591366 I,-0.002694322813384302+0.00314340601449849 I,-0.003092278814269159+0.0002123484535750255 I,0.00260290748735957 +0.0003035901751694549 I,0.002383338178170604 -0.002285044964024597 I,0.004313279540280703 -0.0006985187286778629 I,0.003022088314654139 +0.003218998923606013 I,0.0007855177548788556 +0.001030854922614613 I,-0.001280306084198139-0.001406803883793498 I,-0.002477791350557822-0.004308118267251314 I,0.003096745240650591 -0.003756505744213857 I,0.004039162208866205 -0.001082830326049638 I,0.001512663808349126 -0.003722477057189164 I,-0.001342585239374965-0.0006235057809015424 I,0.0005073502017148629 -0.0003811760884318207 I,-0.0003046183247328104+0.003423568954364209 I,0.001377864744918554 +0.001582891391146731 I,-0.003025677293643913-0.0005203381512617955 I,-0.0006627619153748244-0.004500124509753366 I,-0.0002073922662073513-0.001943710836802869 I,-0.0008258234629799198-0.0008777615831357526 I,-0.001733012188718883+0.003537687658061905 I,-0.003250650840474218+0.001243426176083734 I,0.0008374859418491737 +0.003568882387709646 I,0.004428760189575365 -0.0001618125392196578 I,0.002764966005156028 -0.0001779791327235626 I,0.0007903185146372009 +0.001099418744002191 I,-0.001650779494079518+0.003439260706459635 I,0.002724171054614802 +0.003042219984712806 I,-0.0005634447163405371+0.001964325446437597 I,-0.002825983389919433+0.0004379089744174215 I,-0.004782782114197975+0.0006116053343517654 I,-0.001056177758082359+0.003637278557698716 I,0.001011364898826968 -0.0004717286938825391 I,0.002521708461246542 -0.0006654387356250545 I,0.001196054905525518 -0.002507976535338407 I,-0.001162833260409072-0.002242248534146089 I,-0.003305715742956346-0.001300149416165563 I,-0.0009393526398768789-0.002543109441588514 I,-0.0008041019291098622+0.002454200417419549 I,0.002565223424291625 +0.002873884538920878 I,-0.00030349339760509+0.002524236822652368 I,-0.002039299031995304-0.0006893493257337402 I,-0.003767816440712999-0.0001387347462531577 I,-0.000988416035913646+0.00336453200058861 I,-0.001793006196592512+0.003039161281524924 I,-0.004107236580715683-0.0006127480974795758 I,-0.00367146692728116-0.0006399962494371705 I,-0.0008979168498002885+0.003608729085272705 I,0.002400645599871062 +0.002166658723655058 I,0.0005287061613206629 +0.001699892867513968 I,-0.0004551171108572601-0.003103583598460721 I,-0.003379303990193779-0.001547736080178921 I,-0.002707402676256695-0.002199988432640208 I,0.00008400302034983457 -0.003392834817892045 I,0.00143835780766548 -0.0002658481647185942 I,0.001418627324394865 +0.0006246740371758339 I,3.99257483104841*10^-6+0.004235225779463075 I,-0.004265549219867886+0.001768715015341158 I,-0.003379086572147026+0.000768921706891335 I,-0.0006841683010999968+0.002639590806596509 I,0.002263682989161219 +0.001715142429122555 I,-0.0006115984804603317+0.002184943242408055 I,0.0005573129919353591 -0.0007176506286076975 I,0.0007974273578991066 -0.00111250373991203 I,0.002993260777789191 -0.0009690881462134128 I,0.001595741616899676 +0.002712360403997832 I,-0.000890774338339024+0.002053844181678119 I,0.0007645366430514628 +0.002187089669037784 I,-0.00006075801047651052-0.001144953658312249 I,0.0002228781599732785 -0.000937163946125445 I,-0.002261743764185205-0.002354601244404861 I,-0.002607047393389919+0.001122608202443124 I,0.0002241803135579293 -0.0002197029713463587 I,0.001446849520457882 +0.0006959053716696674 I,0.0007129898254984806 -0.001465611824247634 I,-0.001451897529730684-0.0007006314973224541 I,-0.001428447276197512-0.002561136852462541 I,0.001533058007537993 -0.002716960043370624 I,0.0003112870215851977 -0.002716856703657434 I,-0.002763094423155343-0.0008495339395450044 I,-0.001926378244158578+0.0004882081289586484 I,-0.0001607125807265937-0.0005211539034104486 I,0.002517597825202488 -0.0003154938261894014 I,0.001220907425936523 -0.002399717266461406 I,0.0003756605631145859 -0.001260915013016844 I,0.001030881712937046 -0.0002877808583844799 I,-0.0002640295016358624+0.00193451592177471 I,0.001046657588164195 +0.001575609808696667 I,-0.001551071184795682+0.0007321130176907852 I,-0.001393777982126158-0.001707140506420847 I,-0.001982791471869316-0.000795080013935826 I,-0.0009289195521194114+0.001638597855771241 I,-0.0009883539392235654+0.002325057759845789 I,-0.002667186881076809+0.00003377620439563815 I,-0.001718364977476808-0.0001485144321260694 I,-0.00005060386744774288+0.001811173732145943 I,0.002107781713843573 +0.001312895933143062 I,0.0003980426615186505 +0.001881825027476635 I,-0.0006558972466184333-0.001093887590428121 I,-0.001614292851253797-0.0001188108239252333 I,-0.001213374665376327-0.001852559732682234 I,0.0005791232152818382 -0.002239230761595574 I,-0.0001686635302570547-0.00130518090808817 I,-0.000874561692499849+0.001028888907943712 I,-0.0006098789584962486+0.002179236714888841 I,-0.001807827986845475+0.0005709567812822142 I,-0.001436103523177905-0.001363574560166782 I,-0.002183586954802728+0.0005971420892282779 I,-0.0002166701858890536-0.0004759780711825327 I,0.0004374968687279491 -0.001206069496327741 I,0.001196662225146579 -0.002541392622759208 I,-0.001576233840171493-0.001449179078495011 I,0.00005946146292291943 +0.001214326678852762 I,-0.0009148091933649862+0.001748186432130073 I,0.001124869519682062 +0.0007344862123573846 I,0.000991408774485813 -0.0008989983318555567 I,0.001941880462747807 -0.0019505561838768 I,0.00252031132616055 +0.00012390477350003 I,0.0002303446582069498 -0.0005531638152975243 I,-0.001546015998536669+0.00007567861984744322 I,-0.002550488125203428-0.0003254817602693203 I,-0.0008402329054184315-0.002155408744979908 I,0.0005644888688035153 -0.002008744245050642 I,-0.001245672486205842-0.001393921296001864 I,-0.00008270876482698023+0.0006487868620731603 I,-0.0001872515459496037+0.0004104332340965224 I,0.001532811164120982 +0.001088860062806645 I,0.001659501579835361 -0.0003203478092582275 I,0.0004317449954826431 +0.000846502547829459 I,-0.001424284313764847+0.000421411656887814 I,-0.001940106048370735+0.0001430718821261298 I,-0.001207074762981786+0.001583152188388595 I,0.0004119668216491637 +0.001129907709029869 I,-0.0005571827027145107+0.0004980512060738334 I,0.001185074975903876 -0.001582215287786184 I,-0.0003466677529943613-0.001930784637431238 I,0.001539068466039986 +0.00001338058304783658 I,0.000154429375205069 +0.002201953432626909 I,0.0008552593702451547 +0.001592663247168362 I,0.00123157928281953 -0.0002439053911812723 I,0.00203429311066318 -0.001255800800044635 I,0.001891511704206038 +0.001215462573382732 I,0.001052633485740522 -0.0001638008913801309 I,0.0001207244008851211 -0.001679069460139503 I,0.00009966139690064128 -0.001984733101383868 I,0.001461253791324988 +0.0001354795945439346 I,0.0005950855168738282 +0.001825057532130458 I,0.001830047103878298 +0.0008063166775287361 I,0.0001665462046405792 -0.0003579097907344543 I,0.0001039260409719025 -0.0001102586568433835 I,-0.002128695754401765-0.0003734639316351343 I,-0.001404833147380231+0.001561574604740713 I,-0.0005273592190000688-0.000480051315435719 I,0.0009644822173656652 -0.0006897113028755651 I,0.00009049091850112394 -0.001817550252986041 I,-0.001791017726452243-0.0004071988923711961 I,-0.001797807405335518-0.001126741690034399 I,0.0003571634323263884 -0.002347515285716731 I,0.0004581358415478003 -0.002090170135945348 I,-0.001820230897038045-0.000669495949243003 I,-0.001097388794410434+0.001426283156371886 I,-0.001292293974380149+0.000502743885909969 I,0.001202623132337212 +0.00008052391317340845 I,0.0007807387279675703 -0.0011238764245877 I,0.00178034126447338 -0.00101400618734205 I,0.001828749023670471 +0.0007022254802724566 I,0.0004948569478790565 -0.0001348947782055911 I,-0.00108439740176148-0.0004532859032029427 I,-0.001680073100601488-0.0009433460627823023 I,0.0001643197030577843 -0.001555251882477639 I,0.001317772709954243 -0.001218273444741586 I,-0.0003915024097642367-0.001780261014039993 I,-0.00002053498732682607-0.0003958277661142958 I,0.0008843615922882477 +2.823727920107593*10^-6 I,0.001552366154471308 +0.000856191961559341 I,0.0008556859900510949 -0.0008811805245394364 I,-0.00110895650160011+6.208272466839953*10^-7 I,-0.001106767589582223-0.000974062822202849 I,-0.0003919678829390379+0.000542642445379827 I,0.0005279207843760657 +0.0005285814627581651 I,0.001005853149190116 +0.001493610468066801 I,-0.0004255441702832599+0.001613512763464183 I,0.0003220842991166842 +0.0004245872946177668 I,0.0003337246407208093 -0.0009075915541846046 I,0.0005910148985402905 -0.001529459684810072 I,0.00145358005521869 -0.0002045452851003318 I,0.001034090044333663 +0.0007474534056380111 I,0.001121775425753665 -0.000370215007042921 I,-0.0004225782524339132-0.00006149153532624245 I,-0.0003560579142424479-0.000182092543984711 I,-0.001164911501838393+0.0006867191959369142 I,-0.0002257534676803061+0.00120900590682359 I,-0.0004846488596535856+0.00005988122984747026 I,0.0002405837299623218 -0.0008717337986531095 I,0.0004345686418717502 -0.001265875338901007 I,-0.000679114902193624-0.001191631957832635 I,-0.00095594032810842-0.0002434379814515345 I,-0.0001137408320920911-0.000719009222432782 I,0.0004534906602301861 +0.001135413003946056 I,0.001191593516301569 +0.0006459748002936905 I,-0.0005963784273917604+0.001150080104144431 I,-0.001544893738440671-0.0006622448490551662 I,-0.001483910756755041-0.0002821733937303094 I,-0.000660671748753125+0.0009450234815439397 I,-0.0001809092655359509+0.001655041514960282 I,-0.001502240883216415+0.0003698008977992039 I,-0.0002323300980385988+0.0002760965978789239 I,0.0006504830289305434 +0.0006439625707743579 I,0.001131927589029946 +0.001220222518891846 I,-0.0009667930795396539+0.001098793145394501 I,-0.001172814597801346-0.000461265856364112 I,-0.001270060907083158+0.000474988899918536 I,0.0004876392254911522 -0.00007551030979146946 I,0.0005008510084331987 +0.0001528436741099414 I,0.001503416915176676 -0.001138862469500365 I,-0.0003807917780494301-0.001750292511240547 I,0.00006188681955812436 -0.0000864966295397758 I,-0.0004024094827412825+0.0005116529134658786 I,0.001106349059767054 +0.0007796488982278635 I,0.001136969880557411 -0.001058658220880592 I,0.001822585403169297 -0.00006046537768295703 I,0.0009531404972490582 +0.001938221335837039 I,0.001177269666193731 +0.002017432519325964 I,0.002120813049232586 -0.0005156917105593802 I,0.0002011843003499881 -0.001321545040802409 I,0.0002471236947507539 -0.000876998687231204 I,-0.001537040779448864+0.0004960299900932935 I,-0.0005368440099222557+0.001647271723783425 I,-0.0007430027335793414+0.00191400162576639 I,-0.001812765991647687+0.000416674063763624 I,0.0002851978002992968 -0.0002025372855021961 I,0.0009762664967119909 -0.0009189480052525322 I,0.002077870887106769 +0.0001440319497110502 I,0.0002323545133929156 +0.001944664590286824 I,-0.0002479651589404572+0.002050550558585861 I,0.0014867860634223 +0.001022594432016549 I,0.0012744271713933 +0.0001689930733550999 I,-0.00004306894330128086+0.001352324097624671 I,-0.0006997658530373314+0.0004782960099587856 I,0.0005930658615829784 +0.0002286517474323169 I,0.001222922975043756 -0.001192001074846469 I,0.001378179081191292 +0.0004589668775435614 I,-0.0006789792330853892+0.0003780846047117123 I,-0.0003609521020970773+0.001119513178009524 I,-0.00102655494604038+0.00002652405798936166 I,-0.0007819796497583636-0.0007871934165598789 I,-0.000542993154017742-0.001077738074540547 I,-0.0004643545382669824-0.00007326672256827912 I,0.001127264730205876 +0.0005746187606004147 I,0.00009895440960188927 +0.001708858178132645 I,0.0001474838642256053 +0.0007046792714876889 I,0.0004224234978189844 +0.00003470572826101128 I,0.001148496597137171 -0.001198555530065961 I,0.001549139281974635 +0.0005686069040186404 I,0.000840250725417885 -0.0003559042252694202 I,0.00008130238360589927 -0.001378842963870865 I,-0.0001538874858709558-0.001713410479635777 I,0.001179603622130243 -0.0003541942927871804 I,0.0002569432988791052 +0.00146802028566976 I,0.001370498665621192 +0.001004169524200153 I,-0.0002165519958835985+0.0006606047997717271 I,-0.0009803540372688771+0.0002898910714824403 I,-0.001630499197767222+0.00006379250839165045 I,-0.0001572320294637162+0.0008417702348682138 I,0.001071402182079336 -0.0005715186924823094 I,0.001410764916326585 +0.00066890649261549 I,-0.0001038104639006274+0.0002832824391374038 I,0.00008657817123476553 +0.0007092703191596236 I,-0.0009922631004335497-0.0007316055404788426 I,-0.0001884736449597608-0.00122785938765465 I,-0.001414002121155881-0.001329976462437842 I,-0.001675143848672164+0.0006708564735506533 I,-0.0001832484893077266+0.00008325789126413895 I,0.0008554708473956906 +0.0001181478077787348 I,0.001278867266613025 -0.0006883122340159121 I,0.00005248986222639525 -0.001325261463946559 I,-0.0007987385115829144-0.001347026957725325 I,0.000433766866710146 -0.001254179442517384 I,-0.0004233182421610651+0.0005507077307035862 I,0.0002898852922714918 +0.001279053395892947 I,-0.0007582417445513189+0.001337285970976323 I,-0.001396090710621185-0.0002940092283128208 I,-0.001476666499939781-0.0006379783012424028 I,-0.0009871347042661952+0.0009765425662551586 I,-0.00003767314619923+0.0008008410138904567 I,-0.0006853879313219595+0.0001531446464803844 I,0.00002207336186492828 -0.0005998033827386458 I,0.00004426448965332159 -0.0001300340917064274 I,0.001189872932012436 +0.0002568683802771022 I,0.0002807298275372105 +0.00133083078735938 I,-0.0004631977737064289+0.0004728720486328015 I,-0.0001392924397176978+0.0007707714457703688 I,-0.00008421088250771135-0.0007293664311386658 I,0.0002551240066344729 -0.0007307841935447906 I,-0.0006147318691967733-0.001371660761967018 I,-0.001258038633832856+0.0003165030120221313 I,0.00009151642324975311 +0.0005155373717799996 I,-0.00006511691502914025+0.0007991392964632911 I,0.00002283065820791956 -0.0006830684683943294 I,-0.001028777650815375-0.0002160539030183558 I,0.00003638686181586124 -0.001380751512519381 I,0.001354431917705232 -0.0009432375578132092 I,0.001254481201295875 -0.001866799090568191 I,-0.001256740118881289-0.001747995811256788 I,-0.0006082939457970855+0.00001677869843283023 I,-0.000286444111747144+0.0007048086312577517 I,0.001187296877721484 +0.001324442984591533 I,0.001751837643076919 -0.0002724965867856166 I,0.001775104314181826 -0.0009395046489952796 I,0.001602655657707067 +0.0007590799485234345 I,-0.0006540690645380535-0.0003764975245353881 I,-0.001656096299112589+0.0001721090412456873 I,-0.00161736720199241-0.00137180121581267 I,0.0008831175790410482 -0.001743491663186887 I,0.00148185225273748 -0.001669182702206149 I,-0.0006978931093377448-0.001873542133101836 I,-0.0007235654740652946-0.0006462922090534875 I,0.0006678160949733792 -0.0005995330200822006 I,0.001162041814282156 +0.0006799421130584957 I,0.0006455595235436227 -0.0003748786285334348 I,-0.001160452355066655+0.0003793942637132078 I,-0.001232331259996631-0.001069595605551551 I,-0.000639895494510154-0.0001299245575801635 I,-0.000364996787093996+0.0004073811822855537 I,0.0003426355391348246 +0.001590356065910545 I,-0.0008941547913336069+0.001057684985847937 I,0.001016606142746913 +0.0002806199101534208 I,0.001040617618726068 -0.00134487133803896 I,0.001403256772725187 -0.0003630008897632409 I,0.0002877835985594849 +0.001256242628100745 I,0.0001640967606528124 +0.001961990539966145 I,0.001713869192604901 +0.0006547325344575885 I,0.001309275516249381 +0.0002821456345934946 I,-0.00001079231706097834+0.001482232140861297 I,-0.0007187545016022983+0.000749667092107799 I,0.0006000654497191361 +0.0002444909678433703 I,0.001102840455653701 -0.001388195531289873 I,0.001643445513385914 +0.0003543072720643208 I,-0.0003481050651710206+0.0004945351388299672 I,0.00005422543591338023 +0.0009154655700740975 I,-0.001081192182047912-0.0003534987671344668 I,-0.0005177136786718232-0.001047422950584962 I,-0.001255800088901138-0.0009416660991528159 I,-0.0009262033676373812+0.000857021708397842 I,0.0004253639315561835 +0.0006459339264186517 I,0.0001057682327476345 +0.0009347055593073899 I,-0.000125329350561582-0.0005938348978301449 I,-0.0009303387296862676-0.0001452466211532555 I,7.316210091224934*10^-7-0.001468756496135024 I,0.001498959530811151 -0.0008342232990648187 I,0.001303390345248714 -0.001557698554262407 I,-0.0008708743064008511-0.001818610585747589 I,-0.0007796463117503236-0.000471503951497967 I,-0.00007306972743122072+0.0001059955703103931 I,0.0007881558574607429 +0.001541802148432635 I,0.001775748111225722 +0.0004266232338640793 I,0.001297523557196417 +0.0001782642179267807 I,0.0003096676953816606 +0.001084697147391726 I,-0.001001205696454593-0.0002173435879848673 I,-0.0009826102776185008+0.00005593421688087011 I,0.00009086812877809096 -0.001520639716885678 I,0.001778337396417154 -0.0002124707467778137 I,0.00121324756509461 -0.0002506682463759987 I,0.0002101092728548815 -0.0007468901126866482 I,-0.0006844973704058857-0.001256440283805381 I,0.0007968352071813578 -0.0009868452366267612 I,-0.0002065262766462287-0.0006020382204606758 I,-0.001033089637645417-0.0002108241364109827 I,-0.001033759939543835-0.0003912675516544653 I,0.0004453298671888974 -0.0006300174878938945 I,0.001203917791761097 +0.0001198844054447218 I,0.0009765635787232624 -0.0007750210402721288 I,0.00002845844684398748 -0.0004443124706578405 I,0.0003036039967380388 -0.00009736498825319046 I,-0.0009245224454661431+0.000624705230390087 I,-0.0002870033874902138+0.001118761286791351 I,-0.0007506555511307118-0.00016298379337518 I,0.0002609656807495935 -0.0009276985287735654 I,-0.0001397213629562304-0.001187570272160432 I,-0.00106833320173728-0.0002622030269577916 I,-0.001121620782930191-0.0003421163754467145 I,0.0001005985825566998 -0.001249097262913758 I,0.0003864422493844189 -0.0008660163095105924 I,-0.0007351407094567193-0.0004395169353483788 I,-0.0008566432658210324+0.0007251474311224211 I,-0.0009250007349077601-0.0001325423825977132 I,0.0004958873439560961 +0.0002142675466811282 I,0.0006350965018226812 -0.0003210559162505753 I,0.001164909877395493 +0.0002322287363805409 I,0.0004672292163988493 +0.001085614140028225 I,0.0002650641193428841 +0.0002158346731132039 I,-0.0001876585070887461-0.0006550245409853165 I,-0.0001112504433050157-0.001271154811696057 I,0.00108719840552578 -0.0004483155489241059 I,0.001165181328687391 +0.0002960708864544549 I,0.0008950004722557147 -0.0007720110613302415 I,-0.00002213516198207351-0.0003610138342098084 I,0.0003279235282107879 +0.0001549218262626101 I,-0.0003496325975932411+0.0008415575692754845 I,0.0002270405326445073 +0.0003808017878274443 I,-0.0005979597637211891-0.0007215564636137513 I,0.0005873339405040409 -0.00111972465673941 I,0.0002034316847487604 -0.0002713120962606476 I,0.0001011521833112425 -0.000157150304784222 I,-0.0008982272687296899+0.0005683021174868116 I,-0.000950026330486011-0.0002081592427437752 I,-0.0005959214978192486+0.0008329897100785883 I,0.0008187149344020002 +0.0008074910121766137 I,0.0006359492577995554 +0.0003391325489922198 I,0.0001420753793577598 +0.000360763004926484 I,-0.0008854017281748012+0.0002696734104680565 I,-0.0001592298535585912+0.001042111440231299 I,-0.0006564495350700737+0.00008223646526716117 I,-0.0005123021479036973-0.0007383834031270716 I,-0.0009648668792833177-0.001078959903315822 I,-0.001178348576300737+0.0003334295606584058 I,0.00008331060577111446 +0.0002200963225718737 I,0.0004704505486128099 +0.0004090302928254711 I,0.0009575182444394133 -0.0001206408096372312 I,0.0004970151297494272 -0.0007817571307972762 I,-0.00001798416411827917-0.001037450322527709 I,0.0006283697965778531 -0.0004726658009741318 I,-0.0007170857227894714+0.0002511653350827115 I,-0.0004197338112703221+0.001035792274288454 I,-0.0008354761836042735+0.00008387283692189459 I,0.00003944155210486227 -0.000826868352381889 I,-0.0002836568803745242-0.001104011479060466 I,-0.001003676476714448-0.000024728764563743 I,-0.001123867635838802-0.0001698026493304567 I}",
                                   0.5),

  getBigDoilySpec: function() {
    return {coeffs: RotorAnimations.BigDoilyCoeffs, freqfn: function(n) {n++; return -(1/2)-(3/2)*Math.pow(-1, n)*(2*n-1);}};
  },

  TriFillCoeffs: parseRotorCoeffs("{0.1230801364394257 -0.343672354289428 I,-0.09553570895507285-0.0679546559890317 I,-0.1722391378515112-0.1747026242582879 I,0.0555886660047933 +0.1450432966058455 I,0.0453452506021847 -0.01345621965091374 I,0.06907993209569388 +0.1117784436597022 I,0.03077003410985643 -0.085918088572357 I,-0.02388392723876821-0.01698866399725793 I,0.05981416374676882 -0.1062440999113016 I,-0.04082467538556211+0.05501945522166546 I,-0.02134041186857817-0.01116526374481603 I,-0.03594785442324427+0.07765995961818636 I,-0.02249519318798059-0.0498905781080931 I,0.0009479656852295321 -0.0003824880532620355 I,-0.0430597844628778-0.04367565606457197 I,0.01389716650119833 +0.03626082415146137 I,0.00727253421242536 -0.03173200631670426 I,0.03769178325143584 +0.02654349944960931 I,0.006048200871054792 -0.06308964526105419 I,-0.002877016756771081+0.03546145801081046 I,-0.01126419820867469-0.02592008045309105 I,-0.00897327961387038+0.05377825498883629 I,0.01133631265054617 -0.003364054912728434 I,0.01726998302392347 +0.02794461091492556 I,-0.002277337123177328-0.003746125587175382 I,-0.008549053417322797-0.00004104417314859972 I,-0.02180333558460944-0.02776931685378976 I,0.01582953217982573 +0.004889651395069468 I,0.01829459449852087 -0.002985565089523786 I,0.01423341973190128 +0.0260651388793558 I,0.007692508527464108 -0.02147952214308925 I,-0.005970981809692053-0.004247165999314482 I,0.02013618577004373 -0.02267494600064367 I,-0.005267567274297973+0.01263259921358051 I,-0.01189004192497292-0.02056443007168031 I,-0.01295347400261692+0.01032009894707577 I,0.0089965544863969 -0.02382633426124732 I,0.01450850130093723 +0.02287143752629302 I,0.0149535409366922 -0.02656102497782541 I,-0.01020616884639053+0.01375486380541637 I,0.003779246052157576 -0.02311498563913246 I,-0.01442946310415408+0.02810300386540074 I,0.008263775276092868 +0.001578335248964743 I,-0.006525863611852304+0.01566194714413348 I,0.002878574147951514 -0.02095966495875367 I,-0.01189356596509291+0.004989206070448803 I,-0.005335102967144543-0.002791315936204008 I,-0.008986963605811069+0.01941498990454659 I,-0.009886031984778717-0.02109515897851941 I,-0.00372195158101795+0.002711877853205529 I,-0.008839499190881435-0.006652576963478368 I,0.009470951387522263 +0.02143478929049313 I,-0.00217579584839795+0.0004282853345597515 I,0.003224479316968352 +0.003450726432229091 I,-0.005623798296995147-0.01247264452702328 I,0.0002369914213073831 -0.0000956220133155091 I,0.00819029803301628 -0.01751645483510298 I,0.0009119290057304105 +0.008470057421414235 I,-0.003554922177298364-0.006993297147217141 I,-0.006101686318977244+0.01722021725543049 I,-0.006638372509924239-0.01072092226177267 I,0.0006787408340117564 +0.001642725943122991 I,-0.01076494611571945-0.01091891401614299 I,0.003474291625299581 +0.009065206037865342 I,0.001845037364924793 -0.01100080395180837 I,0.01024823014969623 +0.005914489294923078 I,-0.001798158382295812-0.0180970192706085 I,-0.001532459800001966+0.0145619690888135 I,-0.00797204396904016-0.006700105143988218 I,0.002436116956648838 +0.01395002504251939 I,0.00181813355310634 -0.007933001579176065 I,0.00942294581285896 +0.006635874862402327 I,0.005539779846321308 -0.001310574855195037 I,-0.0001470164087164327+0.005027027299654169 I,-0.00583673886995769-0.01040607103708867 I,-0.003588926302382339+0.002277986142029019 I,0.002435119682303502 -0.01153075897976546 I,0.007643872267607885 +0.008262550831467774 I,0.001512050217763698 -0.01577241131526355 I,-0.0007192541891927703+0.008865364502702614 I,0.001355618971666919 -0.007162861811924629 I,0.0008752197972425532 +0.01501822620395134 I,0.005749980593636919 -0.006471656662720584 I,-0.003220986960508178+0.005855866832716251 I,0.0006448104361220647 -0.01276199016278224 I,-0.003866020849889294+0.006570774932660848 I,-0.002816049552168674-0.006480020113272761 I,-0.002243319903467595+0.01344456374720907 I,-0.006934825263799706-0.008021060859865026 I,0.004102646943152792 +0.008219824131994501 I,0.004665850223095175 -0.01089026281443943 I,0.005414459707029587 +0.01031823534517996 I,-0.00312397685364899-0.002806818299195212 I,-0.004904430033967008+0.01184056912884435 I,0.002834078162636543 -0.0008410137281821084 I,0.004317495755980868 +0.006986152728731389 I,-0.00270928130604494-0.001581680637221316 I,-0.002878369622255278+0.003319889707147283 I,-0.0006386161048394706-0.005067188000731583 I,0.006523024913699753 +0.00302589940550788 I,0.004383526770720597 +0.0002180518083265993 I,-0.002873345798529883+0.006067168891757155 I,-0.0005693342807943318-0.0009365313967938452 I,-0.002137263354330699-0.00001026104328715002 I,0.002533754276745412 -0.009079849563412913 I,-0.001651355005707538+0.0002584090222815953 I,-0.005067945570488159-0.004484097336501677 I,0.0002391736011641754 +0.008889145540037984 I,-0.002812769949098829-0.001830139674020479 I,0.002164669000509653 +0.001765082597333787 I,-0.005450833896152361-0.006942329213447441 I,0.003957383044956433 +0.001222412848767367 I,0.004881852767451587 -0.003599739631657343 I,0.006986999133484537 +0.004150676920568785 I,0.001029064045257045 -0.01316048165976907 I,-0.002917800476756308+0.004150570173037538 I,-0.001542645985156575-0.00665983523909014 I,-0.0008731667642349505+0.01263359616679335 I,0.004573648624630218 -0.0007463912723809465 I,0.003558354932975321 +0.00651628471983895 I,-0.0006561551198983475+0.0001735079992093254 I,-0.003135030703360934+0.000340695401099987 I,-0.004983464248147783-0.005863915334891494 I,0.003701824771246676 +0.00004584801176685735 I,0.004980174383643457 -0.0008506461828659583 I,0.00346370709366243 +0.006288428334514313 I,0.001923127131866027 -0.005369880535772312 I,-0.001492745452423013-0.001061791499828621 I,0.005269656200039512 -0.005542266704746532 I,-0.000880870846148605+0.00320436353756969 I,-0.002910015001195238-0.006603146155814144 I,-0.003184712020475079+0.002662123538609718 I,0.002092789154523937 -0.005922102613334565 I,0.003386004790479779 +0.007794629445657496 I,0.005034046442510934 -0.005668736500160917 I,-0.001316891818574493+0.003158149803395128 I,0.002382545890552197 -0.004904665949029979 I,-0.005215408660396183+0.005252847231589427 I,-0.001815445039560193+0.00001767824838478807 I,-0.001760725157663557+0.002610795358470078 I,0.002830746317251384 -0.005330617175519084 I,0.0007766513274169573 +0.0004969399247969246 I,-0.00297251048124323-0.005141107517920076 I,-0.003238368500654231+0.002580024736768942 I,0.002304051682140021 -0.00813016431513887 I,0.002462776363536745 +0.002870918876157482 I,-0.002035634584066031-0.007167783679786396 I,-0.003516117441878752+0.007141781883589099 I,-0.003636235705203151-0.006177981343534547 I,0.003064610965258805 +0.006366263854397655 I,0.002249138621599226 -0.00595658356531183 I,0.003627125325234306 +0.005717859381573256 I,0.000838647218346613 +0.0001676952891760267 I,-0.0007772846335723107+0.00474685855796663 I,-0.001160396162532106-0.005017358315866194 I,-0.0003551613140400036-0.0006224089416117836 I,0.003938916122472855 -0.005327668161833978 I,0.002926118109874714 +0.006047211651995836 I,0.003738385234173052 -0.006640256244456352 I,-0.002551542211597632+0.003438715951354091 I,0.001493409936772648 -0.004615598909717712 I,-0.002933573219293447+0.006949334627663959 I,0.0009863221544027694 -0.001489215435369778 I,-0.002604101390826089+0.001903119361695552 I,0.002494376493969408 -0.007804448675424216 I,0.0003374451423506531 +0.003115190155386163 I,0.000944811513039394 -0.005778746409783115 I,-0.003607365776038519+0.007025750966350185 I,-0.003326074126900646-0.005465497110827465 I,-0.0006975978485041424+0.00722265144554094 I,0.003928934903482857 -0.002687436649452506 I,0.001905057238945591 +0.00444309313669659 I,0.001074861761395239 -0.003096496986581047 I,-0.003936269411840584+0.00517071097398314 I,0.002065943819023217 +0.000394583812241186 I,-0.001631465902963076+0.003915486786033371 I,-0.001531327876214-0.004366602112371399 I,-0.00383335740749948+0.000984493216165844 I,-0.002818622181876312-0.0007396676177021017 I,0.001958456079451765 +0.003275317160202711 I,0.002438034737273626 -0.001805655009952913 I,0.001758225671949016 +0.001616537382177722 I,0.0007196435369878788 -0.005239916239688417 I,-0.002973391491273227+0.0012473015176122 I,0.001118410089916543 -0.006689612485924361 I,-0.00205814080126873+0.005736364687835853 I,0.00192291029752715 +0.00006833454948172951 I,-0.00160456468629204+0.004949060315495291 I,0.00003199609538837118 -0.004851896466147211 I,-0.002983561346634398+0.001081840942900878 I,-0.001333775741786136-0.0006978289840510019 I,-0.002246740901452768+0.004853747476136648 I,-0.002603159430716628-0.005836119433077303 I,-0.001426837798540947+0.0008301043668389943 I,-0.002616674409112125-0.0008296407552085661 I,0.002230665465740596 +0.005969296319154526 I,-0.0004730277353286159-0.0005169888902976291 I,0.001479638139453732 -0.0003396390171190769 I,-0.00247150799619468-0.005273789744629853 I,-0.0009304878952544875+0.0006779694633013819 I,0.001542879772116231 -0.006401227921406402 I,0.001489544667384681 +0.004772527252513103 I,0.001908920641711077 -0.002475050052869024 I,-0.00212691153483797+0.006768133156560617 I,-0.003864873800876501-0.004213052613569568 I,-0.002217476786681774+0.001766189672801329 I,-0.002209874797720359-0.001663144240869592 I,0.002367737846880566 +0.005358697322623283 I,-0.002335954400167196-0.001411651751622516 I,0.001532284015029207 +0.002044027523707007 I,0.000671458108302996 -0.003801844684688506 I,0.002956169008997569 +0.003219222895800415 I,-0.00005397012137689501-0.00003493314058266483 I,-0.001968544485890911+0.003404988269540751 I,-0.0005439489620994876+0.000107071333639938 I,0.0008061198292420882 +0.0008626816080572725 I,0.0008639003075551371 -0.002698549468667266 I,-0.000481504922736617-0.0004917693189292414 I,-0.002946547713651419-0.00275270500302512 I,0.0007878295660437213 +0.002850860225984861 I,0.0004222932186950364 -0.0007599598046092343 I,0.001491204068567413 +0.001599593374963293 I,-0.001405949574248787-0.003118161131755819 I,0.0000592478553268458 -0.0000239055033288775 I,0.002608881162590991 -0.003670642090523254 I,0.001395125936582705 +0.001355745081976515 I,-0.001841050030799572-0.004646069670932534 I,-0.001867482316428704+0.003099132846177265 I,-0.0004589699199403484-0.003263407558321435 I,0.002414230712044668 +0.004119121196916063 I,0.00204757450825407 -0.004379113708775746 I,0.0002279822514326025 +0.002117514355353559 I,0.001469625437955039 -0.002892414466622795 I,-0.001799549015937491+0.004369069319794052 I,0.0007720249964185639 -0.0005824505833059135 I,-0.001806293190103533+0.002259622930815717 I,0.0004353254419316079 -0.004637372433277226 I,-0.001139735820929405+0.001208234570750308 I,-0.0008887305443245909-0.001748324286804285 I,-0.001525421579744312+0.004305054313857621 I,-0.002085591197190158-0.004074217056293811 I,-0.0004849081222991797+0.001625111979830245 I,-0.00106493366034691-0.001367220939130917 I,0.001816178601845995 +0.004481492659930117 I,-0.0006568780722141996-0.0002776461171409098 I,0.0002121572412816791 +0.0006507867490975525 I,-0.00165959312748106-0.002680230565443168 I,0.0001696852085029391 +0.0004106814857807474 I,0.001608068873884268 -0.003946911476442011 I,0.0006201412655600155 +0.001890422577400822 I,-0.0007908039672910448-0.001960711177822453 I,-0.001310413054116858+0.004165726369117267 I,-0.001701330724280409-0.002570823850538782 I,0.000188252985351963 +0.0005193174175172147 I,-0.002691236528929863-0.002729728504035748 I,0.0008685729063248951 +0.002266301509466336 I,0.0004666724191756783 -0.002965805938989 I,0.002601258572991835 +0.001419223576974702 I,-0.0007534285957248331-0.004660913569470704 I,-0.0004354644457261158+0.004049923926430574 I,-0.002292393900630462-0.001662373560600274 I,0.001082295698904106 +0.003485015909712385 I,0.0004612593412311981 -0.002750200987952094 I,0.002562057537424056 +0.00147862232373077 I,0.002104491536520612 -0.0007719528964159066 I,-0.0001155872971150433+0.001991836774515506 I,-0.001247518501503009-0.002371535019328794 I,-0.00194149413908042+0.001013245304667862 I,0.0002028895346880961 -0.004293029708928238 I,0.001760315842207482 +0.001616770127543044 I,-0.000449539595573953-0.004524254817652125 I,-0.0003831149500004916+0.003640492272203375 I,-0.0006618088222150953-0.002617295249558275 I,0.0008199914186528976 +0.0042683020340195 I,0.00169989434690699 -0.001998457219328236 I,-0.0001938582489094094+0.00302121343712845 I,-0.001031875825577762-0.002856491002076508 I,-0.001421630868016393+0.001449697225493429 I,-0.00199301099226004-0.001675026285997054 I,0.0006090292391622096 +0.003487506260629846 I,-0.001340654211912745-0.001965930872605176 I,0.002288561589244267 +0.001510243079989241 I,0.0001903657064577725 -0.004461191284457901 I,0.001294330377569523 +0.003240872848739155 I,-0.001544784127564579-0.0007680546327363298 I,0.00007501738498188389 +0.00374301924993157 I,0.0004545333882765849 -0.001983250394794016 I,0.00235573645321474 +0.001658968715600582 I,0.0005903368151847445 +0.00001453462791043633 I,0.0005387606677790721 +0.002157362707490981 I,0.0006369141828298681 -0.003376799653470203 I,0.0005281145675801405 +0.0004675530563710895 I,0.00105980295019664 -0.001546735872183553 I,-0.0008503913206898041+0.003893521829441997 I,0.001384944961580327 -0.0003276437137987593 I,-0.00003675410217910805+0.001256756824913543 I,0.0006595833666821025 -0.001750524347494706 I,-0.001717907244778969+0.0004200236844814196 I,-0.001962335744936287-0.0009819172980045157 I,0.0001272827628514349 +0.00130891850823234 I,0.0010747612571961 -0.001461774556730098 I,0.001015301299139346 +0.0005483164311598532 I,-0.001459184717489423-0.002601517759272168 I,-0.0008972315755955849+0.0005694965355072548 I,0.001433298840135616 -0.003543617207551041 I,0.001458000990829959 +0.00127042379508761 I,-0.001054198119133362-0.003618747319355824 I,-0.001515651264418075+0.003406456055940274 I,-0.001852856973737319-0.002445925736134477 I,0.001485344668829587 +0.002888552494254616 I,0.0006087799205758755 -0.002882689744941366 I,0.001910968066901971 +0.002065637707866943 I,0.001136658222283911 -0.0004149757085948756 I,0.00002309911893664196 +0.002081897635418863 I,-0.0008597611514956518-0.002462546257245355 I,-0.001001935085515567+0.0005516478881730863 I,0.0005963800991803439 -0.003017784986758129 I,0.001699159396192501 +0.002138646518779493 I,0.0003780125544409245 -0.003943102828815887 I,-0.0001798135472981926+0.002216341125675654 I,0.0006347779493932812 -0.00184282491085238 I,0.0004751677939365717 +0.003749892829897793 I,0.001351909860750154 -0.002320848403388738 I,-0.001313657330934872+0.001530386788590444 I,-0.0002370655398168613-0.003707185384461791 I,-0.0005611750390142805+0.00262107547262535 I,0.0003389047429167294 -0.001790715452981157 I,0.0002188049493106383 +0.003754556550987835 I,-0.001310789373481442-0.0008513222616931475 I,0.0001415624784004127 +0.002094564504510507 I,0.0003078109554246948 -0.002574271547498901 I,0.001113512395397182 +0.0008900159677007927 I,0.001214138359089857 -0.00150425163927116 I,0.00002610959936854387 +0.003100216926853128 I,0.00143749514840923 -0.001617914165680146 I,-0.0008052467401270442+0.001463966708179063 I,0.0002140831512677802 -0.002270988557979149 I,-0.00145510245607167+0.001852360990151182 I,-0.0006969616673147798-0.0008259358999296184 I,-0.0002398494321818494+0.001629938094174359 I,0.0006655021040688935 -0.002141451203975588 I,0.0009003556107097028 +0.001125697929441282 I,0.0001612026090305161 -0.00319049754069556 I,-0.0009665052124723237+0.001642693733165212 I,0.0004058943719694684 -0.002906333573106326 I,-0.0003241450517576225+0.00336594673180136 I,0.001152629055378004 -0.0008046744281370949 I,-0.001161946166391519+0.002424793874479403 I,-0.0004441186533274454-0.003261755766643515 I,-0.0009372406596640027+0.001006840558935812 I,-0.0007040123880421687-0.00162000502831819 I,-0.0005608299758668988+0.003361140936802268 I,-0.001487297403612742-0.002532253776192846 I,0.000162445411223639 +0.001929951033028331 I,0.000249394920522021 -0.001203563649376445 I,0.0006903194914998432 +0.003185018349865127 I,-0.001304427631462981-0.001326333709164965 I,-0.0003848323237495606+0.0005830120634930726 I,-0.001733706315949927-0.002005265214966257 I,0.001025661735788198 +0.002054956032998625 I,-0.0001913038440742571-0.001267831973145372 I,0.001971857261047491 +0.001659997405537046 I,0.0002741648231856022 -0.003586978535749076 I,-0.00001014362105763634+0.002026762132478075 I,-0.001010198430003061-0.001978057234310091 I,0.0006307435520980053 +0.003163102474223541 I,0.001166462555773794 -0.002722565703609858 I,0.001353614926757397 +0.002579558836294989 I,-0.0004057704372642841-0.00001309110380252873 I,-0.0003901698740709412+0.003578303772768849 I,0.002024205309304244 -0.001014262448858297 I,0.0003572039426261939 +0.0001904065390249017 I,0.001216435315501599 -0.002490956194625048 I,-0.001901252564260515+0.001495197092833872 I,-0.0007809942134122478-0.0007017045747988028 I,-0.001226107508491752+0.002960142282211086 I,-0.001977706587189227-0.00232668750776034 I,0.0008667873832590156 +0.001548046896494955 I,0.001074197729554559 -0.002569331051932187 I,0.001637015577141854 +0.002783006805244005 I,-0.0008275674690358648-0.0004689181374391903 I,-0.001317084187214194+0.002875291257448145 I,0.0007085195406591356 -0.0002102534320455271 I,0.001079373938995217 +0.001746538182182847 I,-0.0007625881455628234-0.0004238530797968384 I,-0.0007811687834645498+0.001077783677279585 I,0.0001742242909725233 -0.001066782332426844 I,0.001572412354820992 +0.0007908090955246479 I,0.001087488066868432 -0.0004156585541964264 I,-0.001062284183707565+0.00128049831388775 I,-0.000677320326511235-0.0003954201593053289 I,-0.0007195924055638192+0.0008299724267868208 I,-0.0003129239314401047-0.002408589625649894 I,0.0003974921902750067 +0.0004807460635224974 I,-0.0004264855771891398-0.00215048031041354 I,0.0008189229113941721 +0.002707391574840171 I,-0.001071136206841938-0.0001928587366691434 I,-0.00005295606573896431+0.001872035059992103 I,-0.0001596540262098675-0.001266797000182896 I,0.001630756228424938 +0.0007564748513769699 I,0.0007130244914779884 -0.00001980753628834722 I,0.0005586354415724078 +0.001386591608903924 I,0.0005929401231416194 -0.002511007184443896 I,0.000223176620329819 +0.00003342103408158381 I,0.001247240276516437 -0.001516576497564997 I,-0.0009370830883195996+0.002970949949517526 I,0.001095881692680149 +0.00005451295208164994 I,-0.0007183364496324708+0.001516792222939289 I,-0.0006674545812960888-0.001608569412229904 I,-0.001247805450271832+0.0004332834264822152 I,-0.000656651667266146-0.0009033136340770932 I,0.001327858429072073 +0.001359513265101195 I,0.000874188724023607 -0.00001333055653305311 I,-0.0004488738133346281+0.001286419137389987 I,-0.0001423335701985829-0.0002341328491984611 I,-0.0005343158385826746-2.565260821787594*10^-6 I,0.0007151281586750009 -0.002173020039548777 I,-0.000423011697781893-0.0002374752954986787 I,-0.001790754116096443-0.001618362205207032 I,0.0000762824691300173 +0.001725142790828894 I,0.0003419753326175989 -0.000635041727045852 I,0.001387301039982274 +0.001457319415928877 I,0.0006334385691863531 -0.002269962390853228 I,-0.0004128387514268845+0.00006460225557039857 I,0.001293223914203226 -0.002482305671722179 I,-0.000884002825669367+0.002058672226364755 I,0.000146841282150459 -0.0001868458382906504 I,-0.001256103107989806+0.002128247469349586 I,-0.0006666201551988361-0.002265071809615079 I,-0.0005802284339780428+0.0001682147652028578 I,-0.00126698639262204-0.001121024334125419 I,0.0000597934002910438 +0.002222286385009496 I,-0.0008677697218985852-0.001629902569995671 I,0.0007285284521931227 +0.0009304272821538131 I,-0.00003881358363204284-0.001702866995204767 I,0.0009857401174869036 +0.002176843454516638 I,-0.0004750466638972361-0.0001674758454125899 I,-0.000539126360025185+0.001240732581355134 I,-0.0007031924872747072-0.0004575349185051198 I,0.0005411672501274134 +0.0004412706493334467 I,0.0005633125480406532 -0.001559686145053719 I,0.0003096738092448111 +0.00001141923181165667 I,-0.001251898726938933-0.001643106358425514 I,0.00007434532417346796 +0.001985610477268888 I,-0.0007409329864568767-0.0004849202639862051 I,0.0006919210005542325 +0.0006722764127242384 I,-0.00136270847403809-0.001735582303361861 I,0.0009893457612391082 +0.0003056032121918416 I,0.001229436014689192 -0.001153168254757932 I,0.001848530346235991 +0.0007737162738002593 I,-0.0002883957330539635-0.003528913148381075 I,-0.0007506168930647841+0.001601568565219252 I,-0.001070549874056978-0.001299185292490942 I,0.000521764704806523 +0.003159726090636905 I,0.001220463191862897 -0.0008999349079143356 I,0.001746749783371134 +0.001037669230142197 I,0.001153947817261196 +0.0004101913232802899 I,-0.0008035879877340689+0.0004122296760260052 I,-0.001699119058527816-0.00140487575335376 I,-0.0005942446551520585-0.0004901686023001218 I,0.001260952140051303 -0.001949036094141223 I,0.001978427230410195 +0.001133453364421351 I,0.0002572660113142615 -0.003290120414942267 I,-0.0007294501191890772+0.001037642543259384 I,0.0008494360185786002 -0.001786874091260605 I,0.0003017798207134169 +0.002906028389043823 I,0.001007004188356873 -0.001349778066251822 I,-0.001426382888545227+0.001227935540724593 I,-0.0002268205065706598-0.003091502414821444 I,-0.0003607604462706315+0.001348977797553487 I,-0.0003856614962891439-0.001664958809772535 I,-0.0002182916910587377+0.003158399041698339 I,-0.001626484868692903-0.001152639360270358 I,0.0009733985658099105 +0.002002193710479056 I,0.001272520801944373 -0.002431203891726757 I,0.001167253749287292 +0.001712710828493945 I,-0.0004754163353124022-0.000696076116509699 I,-0.001009654590309199+0.002876227730482625 I,0.001143412156157555 -0.0001865978180952366 I,0.0008895887332438305 +0.001629071179959738 I,-0.0007330844485310137-0.00002850527871990146 I,-0.001017477852622708+0.0009830006880307042 I,0.0001493924670967801 -0.0007148995402146084 I,0.0015278829053889 +0.0002482422115236547 I,0.001324204142927044 -7.686538190905177*10^-6 I,-0.0009304130871250752+0.001195705100363365 I,-0.0001640387799745869+0.00004337699980233162 I,-0.0007837576758402333+0.00008517385027499663 I,0.0002979731568784753 -0.002107035106674145 I,-0.0003964324276350592-0.0000875478457396898 I,-0.001156436479351255-0.000987939089994398 I,0.0003284805094610641 +0.002042713379331585 I,-0.0005835877494066432-0.0001147257189361157 I,0.0004411022726196765 +0.0004144827507996018 I,-0.001245866062036946-0.001465978833722874 I,0.0009254561928116689 +0.00001146200294171417 I,0.001327801496206872 -0.0008264828523110136 I,0.001639546852428776 +0.0008147786567577361 I,0.0001322888960443151 -0.003191640892154094 I,-0.0008082616984105681+0.000905254394126001 I,-0.0003502518253793529-0.001637167067317061 I,-0.00009039006872315106+0.003117457519257856 I,0.001245043595910864 -0.0002126615457164895 I,0.0008659267734156078 +0.001572107083628578 I,-0.0001412389274287264+0.0000999151621011797 I,-0.0008496347181948371+0.0001142208207941359 I,-0.001222024972861491-0.001397527284816017 I,0.0008985343993646438 -0.00005788935790847664 I,0.001270012932897625 -0.0002219697882328817 I,0.0008612258759494388 +0.001556368468309618 I,0.0004807817829665069 -0.001342470133943078 I,-0.0003731863631057532-0.0002654478749571552 I,0.001331967045710299 -0.00137998492740288 I,-0.00019102150338038+0.0008054498543667309 I,-0.0007223114021913656-0.001747934321421833 I,-0.0007969565518923332+0.0006823307499429438 I,0.0004896163239080749 -0.001490943353860003 I,0.0008352112815575916 +0.002079555755312804 I,0.001317414050009878 -0.001385566676186633 I,-0.0002202177115371512+0.0008010908843924228 I,0.0006808536807494237 -0.001154399948003576 I,-0.001348712090447317+0.001237917792752837 I,-0.0006594717679091055-0.0001142726511841359 I,-0.0004645823191738671+0.0005829203762192213 I,0.000792747974742562 -0.001422930016843913 I,0.0004442713565319577 +0.0001995655321443185 I,-0.0007275037502988092-0.001650786538953536 I,-0.0007961780051187699+0.0006655308846524292 I,0.0007388131559600202 -0.002088800745531408 I,0.0007267500768234644 +0.001077745125252095 I,-0.0002498822403756165-0.001993510507932583 I,-0.001089687976968413+0.001779229591015073 I,-0.0008813925114552192-0.001982195521598988 I,0.0005634481781366815 +0.001835642946960379 I,0.0005231972886309843 -0.001480525653333641 I,0.0008465011976199447 +0.001948657361414375 I,-0.0002452488716236162-0.00002441277774346963 I,-0.00002679912788795621+0.001271646371623766 I,-0.0001369803184917805-0.001719022451229613 I,0.0004068827485999028 +0.00007656872074077939 I,0.0009285521487086875 -0.001035316799433237 I,0.0005144867755870068 +0.00211456617138763 I,0.001258511610627734 -0.001417184125040229 I,-0.0003292229546436233+0.0007895374508487822 I,0.0004370860905372199 -0.0009131289421406944 I,-0.001017275921346258+0.001606206213983657 I,0.00006678001627467168 -0.0005544899366610225 I,-0.0002805693728547998+0.0001820326642094891 I,0.001235365730963002 -0.001686695519142816 I,0.00001135208902597908 +0.001187150791302972 I,0.0005956364726380491 -0.001226166487257494 I,-0.001303852165099046+0.001313211807897357 I,-0.0004665688300468238-0.001997422118118405 I,-0.0008274419514108132+0.001584998175745666 I,0.0004559128139439954 -0.0002165241215287982 I,0.0003791548381717753 +0.001544335641034127 I,0.0002407346413288083 -0.0006826529093862482 I,-0.0009617656972232963+0.0006502409259018005 I,-0.0004538612598900484+4.419562096196996*10^-6 I,-0.000440181289415889+0.0006526988396175193 I,0.00009701144372706578 -0.001703034330889757 I,-0.00007162014414659612-0.0002157086953837256 I,-0.001216979062116287-0.001416086159767834 I,0.0001918410687055608 +0.00155795130200443 I,-0.00009385328664375878-0.0004048952590506273 I,0.0009005519415116703 +0.001361276683509303 I,0.0007076865793128464 -0.001332654293879771 I,0.0001941628318542395 +0.0001242349811992312 I,0.001037550831500987 -0.001139611146700312 I,-0.0009687020774115169+0.00100666319444106 I,-0.0006726702124100548-0.0001359133349713228 I,-0.0006702816030042511+0.0006935789134591612 I,0.0005886894390648505 -0.001373978148748471 I,0.000349037175459151 +0.00007914861124165833 I,-0.0007431276203108075-0.001285276879480019 I,-0.0008095921251635578+0.0006450061841922352 I,0.0005242121484340263 -0.002149631071154399 I,0.0004514912437225289 +0.0005507573263614226 I,-0.0009235264102940817-0.001582363213973486 I,-0.0007043258806197093+0.001931229443453769 I,-0.0007671611261268385-0.001159044363284039 I,0.001215757498765951 +0.00127661043453874 I,0.0005760129205350054 -0.002032541078784718 I,0.0006156940908841862 +0.0007177297190393705 I,0.001056076702093207 -0.0007823842920988482 I,-0.000539751894808189+0.001701472325491338 I,-0.0001285686908684601-0.0003468109735131231 I,-0.001224157603494096+0.0004767636533998847 I,0.0004071738135952287 -0.002353757888696935 I,0.0003743270185858782 +0.0002672506288251946 I,-0.000508908646016508-0.001791945919946599 I,-0.0008790293604696882+0.001785445470897275 I,-0.0004159427680464325-0.002012124920954144 I,0.0003423218068885104 +0.001542272355869245 I,0.000164569381656912 -0.0008947612009133435 I,-0.0003160049222735727+0.002141260976196736 I,-0.001109784183824884-0.001346245980412872 I,0.00007814427334166398 +0.0004506149374272071 I,-0.0009090589263007878-0.001544495335883637 I,0.0007661527413147014 +0.001591565963599414 I,-0.0001891679732773606-0.000609041019011714 I,0.001193811858911304 +0.001248863555998963 I,0.0002793188406749379 -0.002157392166577439 I,-5.87990187901096*10^-6+0.001144418981130434 I,-0.0006071794963195137-0.001154505529277889 I,0.0003018075679805347 +0.001905751781195205 I,0.0005622846553998065 -0.001489145891327958 I,0.0009067813313085766 +0.001429464845393314 I,-0.0000891825530007906+0.00005337190754654683 I,0.00001725061313221025 +0.001815534760376546 I,0.0009559342411568452 -0.001154787399078617 I,0.0002527072940362764 +0.0001249170180622514 I,0.0005916286517929575 -0.001122587969496873 I,-0.0009599200668783071+0.001550619115642054 I,0.0002096618045866532 +0.00004192382229400676 I,-0.0001943211583930777+0.001186714639491658 I,-0.0003617663028733568-0.0007933890714653459 I,-0.0003348445124230495+0.00008020818869638588 I,-0.0008373522794034657-0.001045825396376842 I,0.0006929705043689161 +0.0007405312433868503 I,0.0004846981982764334 -0.0001512660471737107 I,0.0005423772252506624 +0.0008389017265135796 I,-0.0002900990406330263-0.001254339578966549 I,-0.00008879032851000091-0.000155602235402946 I,0.001212155357342218 -0.001313939397285358 I,0.0006188964960453231 +0.000739394204594404 I,-0.0003614229083464799-0.002033076534580669 I,-0.001140411237922085+0.001044295093764946 I,-0.0005225520485676544-0.001768134918825185 I,0.000729546790093853 +0.002017409640152446 I,0.000984729030618214 -0.001331917040458495 I,0.0007315295274686784 +0.001511802912998959 I,0.0001277801970250825 +0.0001265325842508469 I,-0.0004788516089870891+0.001178517932735481 I,-0.0002924759087893577-0.001118437576145904 I,-0.00002895967890455991-0.0002950386940908119 I,0.001078905310136029 -0.001321638022017778 I,0.000713836262648764 +0.0014948973660673 I,0.0009345963085432631 -0.001660064061114088 I,-0.0006378855528994082+0.0008596789878385229 I,0.0003787451532354936 -0.001088831246288386 I,-0.0006649114141839938+0.001769207359891414 I,0.0003492069910561385 -0.0005253388991743122 I,-0.0006247925222547288+0.0004257386842134496 I,0.0006095248921389111 -0.001928849616668629 I,-0.00006261155264655661+0.001018072718645461 I,0.0003733524841931618 -0.001153899727429428 I,-0.000733393304823362+0.00173733365691599 I,-0.0007345820330520158-0.001135631943004143 I,-0.0003681237541384348+0.001416161033995505 I,0.0004116130324728866 -0.0006773418619431534 I,0.0005675005080828675 +0.0008257683088507382 I,0.0006396889538368932 -0.0006199427246500497 I,-0.000465954734037551+0.001101927609117272 I,0.0002465805386006924 -0.0003723038588424443 I,-0.0006510253477065222+0.0004757798404238878 I,0.0003291489023814236 -0.001601228483742871 I,-0.0004678107521097996+0.0002955163911027364 I,-0.0009530063732212964-0.000900946887065041 I,-0.0003925545317891348+0.00124559329391809 I,-0.000130847611793927-0.001102656427980114 I,0.001097455269508819 +0.0009749453761891167 I,0.0006235941234923523 -0.001951112168856054 I,0.00008436128558766315 +0.0007787975388465405 I,0.0006300481811095914 -0.0008319667545185764 I,-0.0005353336595547306+0.001848871475412697 I,0.000298747839033982 -0.0003562906386305628 I,-0.0008881227287780722+0.0003664942745358206 I,0.000553129129927952 -0.00212039174881083 I,0.00009741099028863464 +0.0006955444762197153 I,0.0002362028782598484 -0.001444686602445779 I,-0.00090184144400963+0.001756437741587546 I,-0.0005671256791500908-0.001559705218032169 I,-0.0004421803252515072+0.001674811090478654 I,0.0004057759062188535 -0.0001171394631366451 I,-0.0001454118344572278+0.001364181771306356 I,-0.0002343822462864426-0.00151430813607061 I,-0.0002381781985662424+0.0001042905471445735 I,-0.0008315185317251617-0.001366374277706867 I,-0.0001743994621260355+0.001805662861385235 I,-0.0008798873945618698-0.000996863100080069 I,0.0009831741023875735 +0.001448410643892368 I,0.0008265488133418858 -0.001801517875300558 I,0.0005495506184753268 +0.001244223756043002 I,-0.0002967894749977817-0.000690497835295721 I,-0.0003616110692407876+0.00201732903246191 I,0.0009822337258707143 -0.0006718591623631264 I,0.0004762643097363979 +0.001110773284174148 I,-0.0001340610844452878-0.0001573037349624774 I,-0.0007802946161225893+0.001282341812018664 I,0.000397656139441987 -0.0003916706356968959 I,0.000227340623386227 -0.00002928456696925294 I,0.0010269770515258 -0.001248743654648594 I,-0.0004105637823563025+0.0009186555815080571 I,0.0002687154403488098 -0.0007741242466452616 I,-0.000984067352960146+0.001292677743495785 I,-0.0007367017659146712-0.001393976132416047 I,-0.000348997073202824+0.001288250332353471 I,0.0006882692442237206 -0.0005979625023546622 I,0.0005737606596930363 +0.001106444537832513 I,0.0002919919389764923 -0.0006483438238298691 I,-0.000910680049906046+0.001209666665090147 I,0.0005164859547558043 +0.0000986459530602967 I,-0.0004078664757407689+0.0009788716965083424 I,-0.0005423037425089682-0.001063659307699405 I,-0.001059343550788374+0.0003165197590769412 I,-0.0005661408877537157+0.00007436450736499925 I,0.0007539213791246518 +0.0008248205878609949 I,0.0007448719694922978 -0.0002606692419862689 I,0.0001026040769244631 +0.00005802136779607192 I,-0.0003828319690535-0.00109165052809285 I,-0.0009583393518748702+0.0002461233040414609 I,0.00005609128737594126 -0.002092845384140331 I,-9.370807494419597*10^-6+0.001150377416445688 I,0.000419997913157946 -0.0004427442720902127 I,-0.00008683977468688651+0.001968555686332147 I,-0.0007751896803067886-0.0006024662004497868 I,-0.0007573097929081821+0.0004149943758741456 I,-0.0007046555454690782-0.0001849169044255255 I,0.0004896140198629411 +0.0008188292900506775 I,-0.00003606039744154657-0.0006832977579568481 I,0.0004961664274793741 +0.0001142740004973629 I,-0.0003451885184668699-0.001524447524538813 I,0.000293613640906443 +0.00093726545322802 I,-0.0001262039795441642-0.0003779480558670086 I,0.0001280884970897262 +0.001497786374138513 I,0.0006095086843184066 -0.0004514137524882283 I,0.0004395564179872539 +0.0004041343455444305 I,0.0004835487026572558 -0.0003404157302511842 I,-0.0006809140464892093+0.0003279304764795922 I,-0.0008000207863162822-0.0005468107570659284 I,-0.00006003265513619049+0.00006150588537944095 I,0.0009053395035888512 -0.0008379631443663478 I,0.0006625010854747647 +0.0006281169059630331 I,0.00017991088424697 -0.001309979059922104 I,-0.0007433478728183068+0.0003118253794030501 I,0.0006077804673020435 -0.001472742776011187 I,-0.0002525717693817623+0.001032802298870024 I,-0.0002618619789034859-0.000652843234585129 I,-0.0008069679342344116+0.0008789611885938493 I,0.0001418227441149763 -0.001664801953102359 I,0.0005653610768014379 +0.000731409397582434 I,0.0002796025224791358 -0.00167240312148109 I,-0.0005145352003171828+0.001434091171958963 I,-0.0004468753689554338-0.001170454402976037 I,-0.0001392583831010484+0.00188546594086278 I,0.001010201892577064 -0.0005650992705359488 I,0.00009083518344226509 +0.0009771145070206978 I,0.0001612058536437154 -0.001063866141157753 I,-0.0009252255828317524+0.001152651926965822 I,0.0004807275743817874 +0.00001708363737043264 I,-0.0004011411715730101+0.001237265078873823 I,-0.0006200260174417051-0.001013131437407046 I,-0.0009487383249835924+0.0003618960446002704 I,-0.0005537503159247128-0.00002414988907039289 I,0.0006388800477185645 +0.000836725377383233 I,0.0005840584015026343 -0.0003368585690004844 I,0.000283311090968806 +0.0002311406743271899 I,7.999023847093026*10^-6-0.001212974116536803 I,-0.0007458903366585997+0.0002704602357252193 I,0.0002526360076368497 -0.001697712926228422 I,-0.0004081764356137032+0.001354288269622339 I,0.0004743107412319938 -0.00001723778783185941 I,-0.0003832140688096125+0.001291805555747454 I,-0.00003237856353589265-0.001186662417393474 I,-0.0007461577885519087+0.0002650549230194019 I,-0.0003334439354465339-0.0001744572460127503 I,-0.0005616852253631922+0.001213436869034162 I,-0.000657776486693206-0.001492846829889102 I,-0.0003916000329519018+0.0002176193134927145 I,-0.0006813321014047136-0.0001484961786134735 I,0.0005490289749472816 +0.001517516851181115 I,-0.00007610828185457307-0.0001760757265179804 I,0.0004145707276271786 -0.0001564550821516691 I,-0.0006507898576791569-0.001459029858269326 I,-0.000356709449635237+0.0002075260917097483 I,0.0003454399490407983 -0.001797981028174639 I,0.0003761636499014582 +0.001392887937139672 I,0.0006817209233034866 -0.0005481327574271189 I,-0.0005658883478091142+0.001886604024397339 I,-0.001061619064770745-0.001140780884635254 I,-0.0008144673008229428+0.0004170077591431798 I,-0.0006541686022780315-0.0002074101888021414 I,0.0005576663664351491 +0.001492324079788632 I,-0.0007633597553482223-0.0005015859897349645 I,0.0003510723226330003 +0.0002967828288123843 I,-0.00008548113016855422-0.000988318551469956 I,0.0009447387724873707 +0.001060869794106988 I,-0.0001351263966840105+0.0002543346350080661 I,-0.000307341182200612+0.0008315239543429486 I,-0.0001182569338321539-0.0001292472225744071 I,0.0003699095348634333 -0.00008490975427976942 I,0.0006649927826829617 -0.0007256789256048108 I,-0.0002470115613815838-0.00009509481506204219 I,-0.0009967823814864285-0.0005558738832692169 I,-0.0002238222071327311+0.0006521963793146723 I,0.0001989384155598981 -0.0006158458573581107 I,0.00069699863359022 +0.0001158631684763847 I,-0.0002326219738136221+0.0001694923658253452 I,0.0006390561969683119 +0.0007222611766431311 I}",
                                  0.75),

  getTriFillSpec: function() {
    return {coeffs: RotorAnimations.TriFillCoeffs, freqfn: function(n) {return bount(n+2);}};
  },

  MandelflakeCoeffs: parseRotorCoeffs("{ 0.0 +0.0 I /* 0.499999999999985 +0.2886751345948042 I */,0.1847024349957967 -0.1946881485999848 I,-0.06102631605366464-0.03859579794740142 I,-0.1135919541781429-0.0889791466969014 I,-0.07389479694390785+0.08081734198636385 I,-0.01067388588958987-0.02403477355885762 I,0.05240581188638025 +0.07341410081365642 I,-0.03921151221545357-0.03774827290458913 I,0.0372907619232205 -0.0040881069793774 I,0.02581400607921392 -0.05005704125026614 I,0.02311650634776449 +0.01335280086604162 I,0.02292541248042066 -0.00851532465003751 I,-0.01208804543742597+0.05894503656766743 I,0.02638606214226492 -0.02781259265713708 I,-0.008718045150514787-0.005513685421057616 I,0.00707964191855068 -0.03655405072046718 I,-0.02796370820517976+0.02221694442499495 I,-0.008401899733503335+0.01210502447605563 I,0.0009315908212541345 +0.03309735849989883 I,0.01393270517545893 -0.02476570410261353 I,0.003470556264675217 -0.003335846160305298 I,0.009112766171202057 +0.0009300895843521409 I,-0.02660038007947489+0.001417238407593239 I,-0.01133393228300277-0.01842427135790679 I,-0.008098972721755545+0.01318155162651956 I,0.001629574704565962 -0.01619392466056085 I,0.00934620757032276 +0.005800929581712876 I,-0.01622742202544891-0.01271130667099541 I,-0.01055639956341805+0.01154533456946822 I,0.006693105525031762 -0.00919587216046952 I,0.008376358452925616 +0.007875160311991074 I,-0.0007852041143282456-0.01646733192310504 I,-0.004576642259704855+0.00196995055543629 I,-0.01444436976457949-0.019050886483271 I,-0.001179204196714202+0.02010444301096911 I,0.004839827082702132 -0.02046712079856853 I,0.006438947853549475 +0.01499530714196241 I,0.00006481093352586125 +0.001178252821730869 I,-0.001567874678092492+0.0204234710106454 I,-0.009247686277146006-0.006394908731518608 I,-0.00331311018728534+0.00007916366808238728 I,-0.001524840841382371-0.003433539079854501 I,0.007486544555191512 +0.01048772868768062 I,0.0004895339562044818 -0.005349692946265501 I,0.001654061421775758 +0.003093412890877906 I,-0.003792883818045526-0.008756531931875393 I,-0.003140454383185596+0.00786702227771468 I,-0.00322049865571231+0.0001248206332781523 I,0.005651419465794443 +0.007207257569201708 I,-0.005896232201179126-0.008536538466172711 I,0.002443963568269638 +0.0008801810735519581 I,0.004749282721747288 -0.004990863976529933 I,0.00631375291160914 +0.01046171951043018 I,-0.003367315725757035+0.001471841559546327 I,-0.004365004005826396+0.00275385428350711 I,-0.005601644602199899-0.005392610414954063 I,0.005327251703330153 -0.0005840152827585355 I,0.005286940337247996 +0.002376172473832744 I,0.00645872245135866 +0.0004437039785343105 I,-0.003620266350396636-0.01455626392867415 I,-0.001386764753481808-0.001058188238437248 I,0.001532364919340355 -0.007388132232437302 I,-0.0009359807420999121+0.01126025680863471 I,-0.001361615367936582-0.006001074342389186 I,0.008214691626629418 +0.005240378183334973 I,0.001994176863199053 -0.005417172909957696 I,-0.0005243680241126735+0.003068897353337346 I,-0.00004309907781510446-0.003947907890077612 I,-0.0008024947906555824+0.007642371619108047 I,0.003687715154173717 -0.00715100589288576 I,0.003302358049675709 +0.001907542980872539 I,-0.0008518732907958256-0.001633895449913673 I,-0.0001512499180733655+0.005852140272494083 I,0.001327106922479728 -0.005675382108775474 I,0.003168976636293488 +0.003141847937372116 I,0.007461374945739293 -0.007394585853719815 I,-0.004542230117183734+0.003250794421549343 I,0.00336814231566771 -0.0002332030472608403 I,-0.0007994680199254844+0.007893378160744037 I,-0.003252489712916463-0.005741281563897921 I,-0.004437205644056849-0.001194071716583195 I,0.002367527037972992 -0.008637920196885595 I,0.001883368972356898 +0.005964533830081264 I,0.003275058925767094 -0.001216474950004814 I,-0.001726863633922098+0.008420719509664606 I,-0.001697649853048016-0.007202619817540496 I,-0.002779407052604063+0.003225925554784684 I,-0.001605897973039312-0.002151659571314277 I,-0.001736542464091353+0.006334610900045985 I,-0.0007141490396197921+0.00284142563509976 I,0.003859894249726428 +0.002697896212807129 I,-0.005412730650833779-0.002774043794939752 I,0.002419862516482974 -0.0002755194038872746 I,0.003750585142266521 -0.006190832442112236 I,0.004329395422963537 +0.0001080257042233404 I,0.003769278779235055 -0.0006355647300920426 I,-0.001956945481148731+0.008076560204881693 I,0.003769437448903243 -0.003973227522444555 I,-0.00124543502149338-0.0007876693458656439 I,0.000820506650944967 -0.005032326652824255 I,-0.003696399779751913+0.003612410519906952 I,-0.001100070351635539+0.0003039344550526608 I,0.0005623424073369112 +0.004970004715868914 I,0.003937928231672255 -0.003934601367213369 I,0.00002724866336822615 +0.0003498980920935781 I,0.0006842645332302404 +0.000835274389547449 I,-0.002757814094197698+0.0008600888228936462 I,-0.001707123980494688-0.005954891446153891 I,-0.001990334473661613-0.0006889746770818928 I,0.003166644376964822 -0.005322522513298767 I,0.002934069905669142 +0.00560038033219168 I,0.0010113774169374 -0.005222007245783118 I,-0.003994815457889848+0.003173849203566158 I,-0.001526874765849453-0.00510096392427965 I,-0.001569124469632426+0.00597550433951432 I,0.003131124380467496 +0.0001763981725594723 I,0.00174650152862838 +0.003619941775910691 I,-0.0007699892707944814-0.005270981772320891 I,-0.002435847336223658+0.001244462803765937 I,0.003122613343497892 -0.003447837366523431 I,-0.003488466079552865+0.00632142085221968 I,-0.001566680358765866-0.0005690479954635095 I,-0.0004976807020096048+0.004357718138265109 I,-0.001048439095267808-0.002865425249523126 I,-0.001189489595211154+0.00001654740812414857 I,-0.001200271390510977+0.001729289210870885 I,0.000133084403032078 +0.004728194071419925 I,0.001419910715479528 -0.003701494466990478 I,-0.0001956662094600399-0.0008255390400548149 I,-0.003325400186198578-0.003194617879613431 I,-0.003123294016905578+0.003131399225522278 I,-0.0025511476084248+0.001837196828483321 I,0.004009840674586669 +0.003595580992011766 I,-0.004202217554000953-0.003296100324187604 I,0.003196051897929582 -0.001155040803081985 I,0.003040155284254917 -0.003352780358689755 I,0.003081357739532031 +0.001869465732962144 I,0.00109042671465793 +0.00002526024851871019 I,-0.001404765670698281+0.0045393568127997 I,0.001990386453643171 -0.003537957728939915 I,0.0004957937521060616 -0.000476549451476519 I,-0.0002035080906153011-0.002294123829807752 I,-0.001467632111147364+0.003117744926267145 I,0.0006128244952660099 -0.0006131394990811776 I,0.001359628640443414 +0.003214909339153806 I,0.003203401120040811 -0.001740244271861315 I,-0.0008932282956680539-0.0001881597897221683 I,-0.0002369000734997426+0.0004605816826196332 I,-0.002427512903357893+0.002168891300656192 I,0.0001153048467692489 -0.004385128506566993 I,0.001310968492406465 -0.001356896189421363 I,0.004009061304895244 -0.001451787287129223 I,-0.0009324071967600291+0.004871631630589855 I,0.001301823738737477 +0.0001328699406292979 I,-0.00380005429706836+0.0002024626296392473 I,0.00273295963411327 -0.005814444521452891 I,-0.003553797776935477+0.000237509247369536 I,-0.003267573708820127+0.001488226983392196 I,-0.002376966456510717+0.003641578525527624 I,-0.001369403234462006-0.003287066130107892 I,0.00235400896724888 +0.001387626846749986 I,0.0007045559894268863 -0.001531763003652698 I,-0.001634436125244867-0.0009434840571818804 I,-0.001116532815381645-0.002006200839297202 I,-0.001319755517452428+0.003201021717826308 I,-0.0004225716804224695-0.002461623935727084 I,0.001479725993823448 +0.001018207822740093 I,-0.001619133183284899-0.002632038765422449 I,-0.001156996103108555+0.001883078803784895 I,0.0006665127221895061 -0.002174811144953998 I,0.000496779676975802 +0.001794131475960247 I,-0.0004874066114041303-0.001770328675058273 I,-0.0009011482607279777+0.001416700034527597 I,-0.002480264715653422-0.002770441677702219 I,0.0008019352960861347 +0.002984735336773572 I,0.00006211397514495578 -0.003512504218645514 I,0.0009195951121511644 +0.001221484732410114 I,0.0005277761810390608 -0.0005931761901438325 I,0.0002129545858383519 +0.003970744399347068 I,-0.001380255800397821-0.0008449945446843297 I,-0.0004875085164281037+0.001002241002829328 I,0.0002327963863702721 -0.002313417808648841 I,0.001335172510049996 +0.0008287042259646232 I,0.00003383946125176801 -0.001278597108616932 I,-0.0003500092498619222+0.001848794004881793 I,-0.0002276438529438683-0.0007035520407177636 I,0.0004347866644577161 +0.002314152786661163 I,0.001720304985180238 -0.002005403941067992 I,0.00001994157680763104 +0.0001753786364017598 I,0.0006156761953186796 +0.00007323047327551493 I,-0.002356057444448069+0.0005139919748722348 I,-0.001176062016725236-0.002503191356984766 I,-0.0007989868079954775+0.001198240612190392 I,0.0004071921808123666 -0.00217203770842034 I,0.001204499793238499 +0.001022289459190195 I,-0.002318203146492623-0.001815900953008835 I,-0.001508057080490934+0.001649333509911697 I,0.001096193556237673 -0.001225408292171731 I,0.001471265209207576 +0.00106101837975276 I,-0.0003362542040845966-0.003124839048730033 I,-0.0008995758914556377+0.00005226925483318888 I,-0.002494166643384776-0.002954789650188554 I,7.995073027972242*10^-6+0.003850068246550312 I,0.0002448807378946548 -0.003652241295557913 I,0.002137565932408307 +0.002291681867453803 I,0.0005826940472361118 -0.0002791371181115969 I,0.00003232258162183539 +0.003076166112060034 I,-0.0010044267589764-0.0009845363094133593 I,-0.0007631547618140837+0.001265554330147457 I,0.0009561579321536756 -0.00131369602292197 I,0.001196622636130092 +0.001125022901714225 I,0.0001429813937266331 -0.001632374452171514 I,-0.0007988439704467707+0.001141935261524336 I,-0.0009559610169669987-0.0006610461165963302 I,-0.0002711520645306487+0.002187831790375587 I,0.0008013838001098649 -0.00109477311839484 I,0.0008726196945592045 +0.0007595314608681544 I,-0.0004033738570070605-0.0005379715772387659 I,-0.000500958570772384-0.0002454407544544844 I,-0.00004366724179736581-0.00221114204427628 I,-0.0001210989317702094+0.001079394168388667 I,0.0004921938868420552 -0.001260711851519801 I,0.0005174547860347814 +0.001951920918425377 I,-0.0001121720163379889-0.002352475989010657 I,-0.0006538060370997428+0.0002814215079169364 I,0.0006022259260965755 -0.001552774294306021 I,-0.0003911698947289976+0.001669124559155345 I,-0.0003735595338203392-0.0004696516370382874 I,0.00004858194463558459 +0.001363714792484225 I,0.0005063082763098348 -0.003620395772518888 I,-0.0002878507875998781+0.001063922029507465 I,0.001895696596799577 -0.0006751755612425158 I,-0.001497192478798604+0.0007621052340143725 I,-0.002030622700129806-0.001107254969358437 I,-0.002644178745022761+0.002312486420777294 I,-0.001305957882387202-0.003185911219323083 I,0.002311974528597307 -0.0002774379190154139 I,-0.002063481394928874-0.002721555211916845 I,-0.0001684577423914479+0.002872063287280029 I,-0.0006772310062019918-0.0001456187047604001 I,0.002105373137560213 +0.002133487668030137 I,0.001089950844796013 -0.003061571893528644 I,0.0002292025661694534 -0.00006177422837353278 I,-0.001441570107778174-0.002208854622969248 I,-0.0006495177627851696+0.002930478126190121 I,0.0004099172344153716 -0.002687239092694913 I,0.001652736402165751 +0.002231946575570883 I,-0.0000594287859433852-0.0004654725324258712 I,-0.0008542796558394631+0.002204038305603868 I,-0.001081071239744103-0.002081776490450923 I,0.0007295578869468153 +0.001035136109867545 I,0.0006914038689739283 -0.002923874399784769 I,0.0009198496933550569 +0.002142186734559101 I,-0.00203139908703984-0.000667902879732235 I,0.0003417474475426763 +0.003367704671896889 I,0.00257745154974986 -0.0008953017512545033 I,0.001924848052041597 +0.00143034580340771 I,0.001033024618177544 -0.001743893242232045 I,-0.001940525449275983+0.0001762324392916675 I,0.001339134137461096 -0.001009128221642661 I,-0.001618580410502442+0.003516966008995931 I,-0.0008846293012865469-0.0006783235799079562 I,-0.0004636454432670025+0.001152324983883146 I,-0.0001456096829611214-0.002033858087907552 I,-0.00006055623765260491+0.0004679435146255259 I,9.258704769763935*10^-6+0.0001683218316729489 I,-0.0002239820968987353+0.002917638715806822 I,-0.0005738211404476003-0.002155332870350149 I,-0.0004193586536143934+0.000927393177883301 I,-0.0000276938391081517-0.0004856811727064317 I,-0.0004752332812626336+0.002203250128072213 I,-0.0009465897371109298+0.001068159650916949 I,0.001224124190695567 +0.0004004359429909756 I,-0.001429702666627337-0.001281059478838944 I,-0.0006541018967913094+0.0001661543901462321 I,0.0009334019125996414 -0.001376243082935308 I,0.001841166940730019 +0.001432080152106216 I,0.0003205631409936506 +0.0002235083910639898 I,-0.0007468377040260308+0.000922125968221138 I,-0.001321098039598894-0.0009135583902390539 I,-0.0004733014553006666+0.00001130909543299626 I,0.001071229183408693 -0.0006665577289089286 I,0.0008068267467183197 +0.0005238607282535794 I,-0.0001327433940427554-0.001974720909112031 I,-0.0003451266976712665+0.0001481121642164307 I,-0.001240046378333089-0.002298457301687096 I,-0.0004029107563213564+0.002250282126084457 I,0.0006515028267174523 -0.002379025975077774 I,0.0006794267491738547 +0.002112589476849331 I,0.00001074737504564601 +0.00007991303614906367 I,-0.000124270791420261+0.002466001891849712 I,-0.001159692138301262-0.0008574944067692737 I,-0.000526861661329472+0.00007273302802404395 I,-0.0002178344059241578-0.0004905055828540567 I,0.001069506365021693 +0.001498246955398365 I,-0.00002835258308395603-0.0006495505006123609 I,0.0003729121273529737 +0.0005818311326563704 I,-0.0002523100623134885-0.001562396630223819 I,-0.0002102133254123562+0.001057242176698199 I,-0.0001432641423802305-0.0001663780749459842 I,0.0003974577703579613 +0.001159306506180633 I,-0.0007165758574570241-0.001232336483618123 I,0.000414246114996751 +0.0008125999834350997 I,0.0006113889723879236 -0.001067941690723963 I,0.000817085970088035 +0.001071876585582881 I,0.00007894472156293264 -0.0003340206329242145 I,-0.0006228473328796345+0.001163206691735511 I,0.00006993342230424489 -0.0007642418494594032 I,0.0002362944888402996 +0.000441916127281906 I,0.0003944499974150369 -0.000955055705527085 I,-0.0001468574571286044+0.0006492531700386012 I,-0.0004027580173454583-0.0009163229351926809 I,-0.0001350843973276336+0.00108736484422341 I,0.0003235291103036603 -0.001036012198291467 I,0.0001629183883667159 +0.0008151298450604061 I,0.00003639827177380187 -0.0006363072841774717 I,-0.0003597374121248378+0.0006097784756433341 I,-0.0002399489748280408-0.001229631603301767 I,-0.0002005888281009507+0.0008790473299824282 I,0.00009852841880891085 -0.001203346820150132 I,0.0003985001911368611 +0.000890938674629957 I,-0.000541840545420565-0.00125093313313639 I,-0.0004486363404497445+0.001123860325377382 I,-0.00004905400256228516-0.0009548221560436009 I,0.0003950392756349187 +0.001226509829053683 I,0.0005024466851378017 -0.001162782792565059 I,0.0000643471950853773 +0.0006761819462623196 I,-0.0009904114628612889-0.001368006257883824 I,-0.0003175078616098212+0.001356001429502376 I,0.000337052767390847 -0.001933904310628309 I,0.0000394646330098906 +0.001827778930324563 I,0.0001756164122459118 +5.374418534806819*10^-6 I,0.0004058601406659408 +0.002120598620491489 I,-0.0008514370740642945-0.0003595004364008062 I,-0.0008078083972955632-0.00001689678681248717 I,-0.0004600712365431525+0.00001783151902885292 I,0.0008073456379618107 +0.001029608224184601 I,0.0004962890935537042 -0.0006537536952982496 I,0.0002879368724794751 -0.00002686509527328275 I,-0.0008356015449140913-0.001388585681934949 I,-0.0006356004152110302+0.0009086143154185998 I,-0.0003855870972164419-0.00008495469939425296 I,0.00084166706021961 +0.001158470302607452 I,-0.0008221685544981825-0.001087878258001109 I,0.00043258035594215 +0.00003017889960060511 I,0.0005484152542952426 -0.0009656783246742906 I,0.000683995362700076 +0.001278458399199652 I,-0.000260548086859447-0.0001467243449952778 I,-0.0002798421839911537+0.0006665719689497375 I,-0.00084231888587592-0.001219505495181399 I,0.0003491376526231151 +0.0001257401533768557 I,0.000545383044688441 +0.00009859531708892918 I,0.000990619722393898 +0.000622714454305518 I,0.00004890197779996196 -0.002139485524312444 I,0.00005352282576760962 -0.0001563918094344806 I,-0.0002371204549073272-0.001889166021543833 I,-0.0006727193509194274+0.001945834143228849 I,0.0005168589757488971 -0.001596886522265739 I,0.00103305554377481 +0.001892752881334161 I,-0.00005196448820509869-0.0002402679713177477 I,-0.0004587776843552247+0.001314009890488165 I,-0.0006533925573301791-0.001142556296993621 I,-0.0001852755342626243+0.0006365152928943898 I,0.0006784689602481216 -0.0007129805680655255 I,0.0009019647016495375 +0.001494531358643736 I,-0.0007286509214794603-0.0008146397928210058 I,-0.0002177986616259041+0.00110022999072132 I,0.0003769411634499695 -0.0005933141669581285 I,0.0003875776509889566 +0.001438110781953454 I,0.0007158582242326803 +0.0003534781926734077 I,0.00005404635470847665 -0.00002282743453072753 I,-0.0005856576615317137-0.0002842156721734334 I,-0.0007619035039463858+0.0008128865348417941 I,0.0003399611632735184 -0.001281606718132588 I,0.001004496661588977 +0.0002320048832979094 I,0.0007191685974131093 -0.0001930880034336912 I,-0.0006691154514531597+0.001010134330552707 I,-0.00048104510369244+0.0002102630799277534 I,-0.0006235720008272449+0.0003934077547821694 I,0.001048832038015136 -0.00123930521222905 I,-0.000100689778376874-0.0002476032244975897 I,-0.001071176495876353-0.0009977673895794038 I,-0.00100025560493556+0.0007159106034582692 I,-0.001221997901073369-0.000523586218804949 I,0.0009367514651431605 +0.001456736715281166 I,-0.0006310083247878926-0.001590138878312051 I,0.0006776088018826659 +0.0001125625249887631 I,0.0006865563217132478 -0.0003982738367997984 I,0.0008003696234027466 +0.001796596527313858 I,-0.0006276624193927001+0.0002626834534288001 I,-0.0006642731418859832+0.0002810742835176668 I,-0.00080023494316366-0.0007703729164347681 I,0.0007610359576315317 -0.00008343075467012627 I,0.0007531721039594872 +0.0002602240572883422 I,0.0009345283417689613 +0.0001317480947992275 I,-0.0003420649834242528-0.002126846100318149 I,-0.0001770322261647511-0.0001636913693246942 I,-0.0002349089134204409-0.001239068200505554 I,-0.0003047339109159563+0.00182717160666876 I,-0.00009675078242658376-0.001515310120115496 I,0.001248348924646165 +0.001475980867942885 I,0.0006230065062027955 -0.0003022114936293173 I,0.0005325466396479594 +0.001066240190600766 I,-0.0002092813416698175+0.0000783570856902112 I,-0.001226036078427618+0.0008940725741099484 I,0.0007552771910162153 +0.0003394532105543122 I,0.0009226746359111715 +0.00006338628266502563 I,0.0009715223892699075 -0.0008954351145193819 I,-0.0007670515756813293-0.0005852731094704232 I,-0.001936401801296486-0.0005332309557580413 I,-0.0009148045369298189+0.001210079685952681 I,0.0008558192965729767 -0.0001227764423087794 I,0.001345148704096941 +0.0005455813852970174 I,-0.0008093398943488975+0.0002527243278013886 I,0.0002290503672542512 -0.001435214354859753 I,0.00005224955508016668 -0.001725431993565712 I,-0.0002530613918866853+0.0002026208103889368 I,0.0002792610103483876 -0.000718344227363607 I,0.0007590894547061687 +0.001229863278920677 I,-0.0005171809071785836-0.002079466275522669 I,-0.0001981092504802518-0.0001511697483449238 I,0.0002589123522690246 +0.0001188637269230867 I,0.0007275843101171761 +0.001165666060709919 I,0.0003135739721815811 -0.001508482834421948 I,0.0004364469768385851 -0.0001980275036498357 I,0.0005422744451444103 -0.00241152549939624 I,-0.001004219898134217+0.001380295945985033 I,0.001084824738994543 -0.0006418906401164458 I,0.0003107155862693228 +0.001496737168735767 I,-0.0008140159652812075-0.0006480340695284042 I,-0.001382632622942312+0.0005481663088583165 I,-0.0001770681782901515-0.00183205436138862 I,0.0008138033773709508 +0.0007171841634302547 I,0.0002189092742064815 -0.001055447461775121 I,-0.0001337115346005705+0.001608608115507271 I,-0.0005413039637412765-0.0009250070709792731 I,0.0000235007652886254 +0.001186011507364977 I,0.0003334118150835932 -0.0006837004556285442 I,0.00004513601122398076 +0.0007649178353330243 I,-0.0006614794201772045-0.0007557961715595849 I,0.00004423329939401985 +0.0009828305435479567 I,0.0000460378237803494 -0.001394488190777835 I,0.0003753681263806379 +0.0008978229655686073 I,0.0001693032720304847 -0.00006306785481655955 I,0.00003938797862803995 +0.001658262910072764 I,-0.0009728226851469908-0.0004347626448394379 I,-0.0002471186031761511+0.00002832502326162446 I,-0.0001945164811269554-0.0008572963346198185 I,0.001173527375218762 +0.0007486254547772806 I,-0.0002252278957122813+0.0002439569642967938 I,0.0006450497324485615 +0.0006463751856392253 I,0.0001445273813638731 -0.001505058515108219 I,0.0003279642934928476 +0.0003901063899170191 I,0.0008546535853382089 -0.0003210207933896792 I,-0.000267734011672028+0.0007180623918891301 I,-0.0003128731831011435-0.0002716975321239106 I,0.0003372881860784297 +0.0007671871363425638 I,0.0001610004005797002 -0.001134945659078436 I,0.0001248754153089498 -0.00005950208111399784 I,0.0004374016273336874 -0.0006736609696647658 I,-0.00005427152936658232+0.00121318838456021 I,0.000284882409022702 -0.0007738818442807575 I,-0.0000749097177318036+0.0004384139076122407 I,0.00006757454236147291 -0.0006050865257960298 I,-0.0000754611625297517+0.0007116435040242307 I,-9.137336159071744*10^-6-0.0007961315919168007 I,0.0001375746284688717 +0.0005554977885270704 I,0.0004882577303489209 -0.001014262072428934 I,-0.0002998335813528107+0.0007020296609104057 I,0.0002979690456998421 -0.000399482855806054 I,-0.00001172801527734306+0.0009050192637924951 I,-0.0003228238572751723-0.0007960362242467134 I,-0.0004641815236261195+0.0003405101621472123 I,0.00007820511158345985 -0.00111413492310906 I,0.0002765549850606559 +0.0007132106820558667 I,-6.157011112266791*10^-6-0.0005639868414451887 I,-0.000114642112951073+0.001091767374157209 I,-0.00007597397418339709-0.0007114553820645118 I,0.0001208204612938214 +0.0005865536512976073 I,-0.0001207022975090851-0.0009940837480416796 I,-0.0002153228494720598+0.0005459444457729082 I,-0.000424900829642616-0.0003218084414310413 I,0.0002651630054005189 +0.001091382113433297 I,-0.0004693574231625145-0.001046723921024381 I,0.0008414218219543957 +0.0005180422984208052 I,0.0004868770666893353 -0.0006020733888110512 I,0.0003239616350151129 +0.0007228817384678636 I,-0.00008387595134773757-0.0003490090486987065 I,-0.0002513580442445273+0.0009598719769511868 I,0.0005268164505965453 -0.001021572270402849 I,0.0004717654356630262 +0.0002725061401340993 I,-0.0002519479468071444-0.0003962408474027396 I,-0.0001450955133626839+0.0009707586837288619 I,0.0003251566174050325 -0.0005155925046874137 I,0.0004941582907143327 +0.0007208731735874055 I,0.0008642467532896991 -0.000847661425063523 I,-0.0004923171604595614+0.0002333298833148815 I,0.0003364556938524594 -0.0001845540994379703 I,-0.0004663533572696497+0.0009969503392312014 I,-0.0002883563260665537-0.0007483656963186749 I,-0.0002210531849132864+0.0003030753749095082 I,0.0001005989908189275 -0.0008095549368070251 I,1.552790183457786*10^-6+0.0005658888263347406 I,-0.0001216961844011889-0.0002334136357087459 I,-0.00002160713115651665+0.0008360200389252924 I,0.0002130002741370038 -0.0005896839568253184 I,0.0001101889460783307 +0.0002305497546512491 I,-0.0004109590955442776-0.001008120438646434 I,-0.0003133356159483393+0.0004858137019537025 I,-0.00008736728706966301-0.0003582391676703917 I,0.0002868559001817645 +0.0009696168997656805 I,-0.0003512498094429115-0.0007028954717709833 I,0.0006410412114198834 +0.0003618211976312934 I,0.0002424130914636712 -0.0006942042047528474 I,0.0000699133917038517 +0.0005247790950708866 I,-0.00009209138433350333-0.0005082287556165546 I,-0.00007171553174967194+0.0007779205900418463 I,0.0001895867032101845 -0.0008107688726787486 I,0.0004527109480386145 +0.0004488354196296741 I,-0.00007323748364624156-0.0001400911752207377 I,0.0002112856528066869 +0.0005936715697869409 I,-0.00004115202252919519-0.001148243146224233 I,0.0001865752477718876 +0.0003756755039276749 I,0.0008449293052788533 -0.0003916826172682321 I,-0.0001900144972010845+0.0007157325668847737 I,-0.0003322258276336496-0.00005166490164798444 I,0.0005739506483726359 +0.0005483156339781416 I,0.0001226060968539131 -0.001367810676928024 I,-0.0001112026358480849-0.0006737197566879649 I,0.0009839456108637153 -0.0009873435172383007 I,0.0002695041995855015 +0.00183348465156456 I,0.001065910706530111 -0.00105636940766357 I,-0.000648890016734457+0.0004643992030750318 I,-0.0004127997102758242-0.001257856197149438 I,-0.0008304302255411444+0.001306367623962946 I,0.0004863817566099344 +0.000272446157601378 I,0.000449840584202001 +0.001013593135731738 I,0.0004279992210231004 -0.001060359422055627 I,-0.0004855923158958832+0.0000839781817660643 I,0.0006016295257741184 -0.0003569667977861841 I,-0.0006326646104059691+0.001149615436194617 I,-0.0004027039178611716-0.0005057498987469265 I,-0.0003817113318958795+0.0004147903870179247 I,0.0001125479160053912 -0.000989062489366975 I,0.00002010116886832375 +0.0005928404398800174 I,0.000481163187946209 -0.00003331472103655688 I,-0.0001142097171354073+0.001127625451530041 I,-0.000177410425922162-0.001251519698623312 I,-0.0006762131045333756+0.0004170053677017156 I,-0.0004094253577778833+0.0001539562106223495 I,-0.0002881049643718988+0.00134329736571066 I,0.0001782151641497008 +0.0005049424734028383 I,0.0008140115348275299 -0.0001092608471652485 I,-0.0007552738760172717+0.00001173827531398273 I,-0.0004956840369712383-0.0005398648253507641 I,0.0002685194530558742 -0.001227109592646975 I,0.0005662693617825073 +0.0001075335321553105 I,0.0006053701473466395 -0.0001787725777552917 I,0.0000587979194934447 +0.0009436251982048489 I,-0.0004646413875567178-0.0008201830805591018 I,-0.0006338865205730445-0.0001705816737983002 I,0.0006064756675874959 -0.0004787079104410708 I,0.0001955179175964098 +0.0004083425192004812 I,-0.0002525568273083313-0.0008468308313810738 I,-0.0001983966063415687+0.00005755420843505116 I,-0.0003149648704987128-0.001539444899029227 I,-0.0001737140763451406+0.001195366445297037 I,0.0004620514175515436 -0.0008568367826637339 I,0.0005232326294317845 +0.0006756038770773714 I,-0.0002687855400549691-0.0003502830948130836 I,-0.0007240308982997661+0.0008124218459159127 I,-0.0003465133869473858-0.001046632638475516 I,0.0004736079540070278 +0.0005530374151008684 I,0.0003382181482842739 -0.001233988599551653 I,0.0002690527103357471 +0.0008520762614395095 I,-0.0006657325345564381-0.0003637226251387287 I,0.00004468250010448516 +0.001287134850260514 I,0.0007665283214600034 -0.0004426308219686933 I,0.0006511771927261369 +0.0005839825010361374 I,0.0005999049826876005 -0.0008733751393342083 I,-0.0007033299545903915+0.0002354829247194488 I,0.0005234637465070532 -0.0002833693847395158 I,-0.0004047001040149132+0.001285801399035176 I,-0.0004080736295731445-0.0005604400042449429 I,-0.0004316421214314707+0.0001675619971227027 I,0.0002007070477935215 -0.001102955689716357 I,0.0001640484978324521 +0.000655162944534605 I,0.0004678655608165845 -0.0001737821357144282 I,-0.0002466948048501168+0.001202959929949916 I,-0.0003327536132613988-0.001199266707046351 I,-0.0005729932543700328+0.0006103269093329519 I,-0.0001124506427993976+0.0001357820806060127 I,-0.0001486853275152555+0.001183158243926682 I,-0.00001757212706841224+0.0002173378268171081 I,0.0005484514038843159 +4.448785918442811*10^-6 I,-0.0004185646117048916-0.0002038729854659899 I,-0.0004217392511469147-0.0001704488062909295 I,0.00004533221089662377 -0.000888965135118698 I,0.0001800010771235176 +0.000342175045804054 I,0.0002419085047604061 -0.0005715121686048045 I,0.0003219664781727621 +0.0006705559259933779 I,-0.0002425214075745945-0.001028945688217831 I,-0.0003970581503661588+0.0004608465078243475 I,-0.0002456025619471875-0.0005004618575238547 I,0.0001773244202669367 +0.001028964201060377 I,0.0006552766637670047 -0.0004437937148887106 I,0.0003671060261095285 +0.0004281695938294511 I,-0.0001224155476379623-0.0008618519599456219 I,-0.0004860672461741435+0.0004220309706314387 I,0.0003852188387384509 -0.0006517391812477128 I,-0.0004127865393563684+0.001129845691574315 I,-0.0002057806071074622-0.0003245201397333856 I,-0.00005789915760160523+0.0006477730121763862 I,-0.00002759219350988601-0.0006865116469454303 I,0.0000308168369799527 +0.000279033125398862 I,-0.000229413996147615-0.000307379938766717 I,-0.0002480774948723317+0.0009049444142857676 I,-0.000142245163578201-0.0007065253936501192 I,0.0001105661157834065 +0.0006024244713547441 I,0.0003556767949290192 -0.0005162228931874374 I,-0.00001113447002984902+0.0005089023621045396 I,-0.0007899863731479368-0.0003813106133472043 I,-0.00008409560052520804+0.0006601080377853509 I,-0.00001368441316758629-0.001241852519138565 I,0.0000366311453837388 +0.001129902819070202 I,0.0003590615194290335 +0.000142915334129534 I,0.0006567368730255709 +0.001240749324401349 I,-0.0004305717500495767+0.0001557168708417835 I,-0.000870672541731293-0.00001770956820275959 I,-0.0001020212913844506+0.0004059179478675562 I,0.0005514134642446307 +0.0003854137446964298 I,0.0004634120571828536 -0.000495370882672211 I,-0.00008821529993223842-0.0003147504688648454 I,-0.0007627153187214811-0.0005565960423521974 I,-0.0005132250048158+0.0006818850948702345 I,-0.0001359411803496968+0.0001829151122519557 I,0.0007019442065906632 +0.0004517089119977644 I,-0.0005744591191794585-0.0004756844682351734 I,0.00006361753171608923 -0.000242964638114152 I,0.0003429221819933801 -0.000532709976601193 I,0.000525654224194176 +0.0007723061033608081 I,-0.0002319247472597748+0.0001297529319978562 I,-0.0003017552924136987+0.0001667180276029682 I,-0.0007732472358333292-0.0003962919707164105 I,0.0003456946452228362 -0.0000393599148334017 I,0.0006362518884587536 +0.0002925213089355025 I,0.0007958615488793041 -0.0001328571537943867 I,-0.0004657884010390717-0.001651336123089964 I,-0.0003163312518697557-0.0002976292118124571 I,-0.000210878583701882-0.0006469029875796059 I,0.00003930624453636422 +0.001421836519885343 I,-0.0003320182563216393-0.0008876290626232607 I,0.001273756733634994 +0.0004841715678525186 I,0.000447612065949941 -0.0004410622857530861 I,0.0001229162225750802 +0.0004607342552820058 I,-0.0001216731741396064-0.0002128988925929008 I,-0.0002459174453402303+0.0008439426841960243 I,0.0005357978774713786 -0.0008844046345805363 I,0.0006184850604264071 +0.00001543224346668288 I,-0.0001404558588020375+8.176901462200911*10^-6 I,0.00001033532907387077 +0.0007022947941833318 I,0.0002373852591307634 -0.0007098368239601187 I,0.000522675219373864 +0.0003014891020354216 I,0.001185551488291622 -0.0009363326080914665 I,-0.000719213995652298+0.0003082152046243447 I,0.0004930444470768321 +0.0001285311889944967 I,-0.0001327841043588585+0.00107810296860578 I,-0.0004738577494307555-0.0007672020028550168 I,-0.0006404512223612853-0.0003522545012364166 I,0.0003991703942127953 -0.001184021945880806 I,0.0002682019612073252 +0.0008336844311116946 I,0.0005384683970288486 -0.00009079496143799371 I,-0.000279563640171143+0.00115379431498255 I,-0.0002709025393458447-0.00102605596931683 I,-0.0004593181931880513+0.0004280637957746684 I,-0.0002154129470580834-0.0002015345174150828 I,-0.0002328831635648547+0.0009055843027400469 I,-0.00007095385036993504+0.0004661478123421863 I,0.0005516977037041852 +0.0003012523956282734 I,-0.0007745074598852641-0.0003423081540396174 I,0.0003042884412237108 -0.00006891111122356076 I,0.0005310995161045193 -0.0008803747268428436 I,0.0006244233256515932 -0.00002010959610043232 I,0.0005490792971406472 -0.00008336067979434544 I,-0.000279798324153714+0.00114968836126165 I,0.0005384910641372887 -0.0005676039317741937 I,-0.000177919288776036-0.0001125241922667908 I,0.0001103826411823922 -0.0007163477566563162 I,-0.0005216877556226102+0.0005303404769702487 I,-0.0001382273903616414+0.00001696484204518533 I,0.0000987507060300703 +0.0007190101595380035 I,0.0005892758714991432 -0.000592946817816639 I,-0.00002656675174674005+0.00006625424429902867 I,0.0001172025022887863 +0.00009403668196993234 I,-0.000407426801195349+0.0001966415764386426 I,-0.0002523869033942926-0.0008704095261215889 I,-0.0002828782288361933-0.00005988976627599285 I,0.0004321129912016557 -0.0007917092627844222 I,0.0003855727675636873 +0.0008087846624040682 I,0.0001172152358469991 -0.0007189038075471253 I,-0.000528057111398588+0.0005160586456997101 I,-0.0001658733100342499-0.0007484182103788442 I,-0.0001908526527614264+0.0008130977822194244 I,0.0003436058108910437 -0.0001265492774972551 I,0.0001810257667349576 +0.0005451656852254596 I,-0.0001060773576892547-0.0007460879555887098 I,-0.000270342176855437+0.0003120725215988598 I,0.0003485166838606897 -0.000554479527603569 I,-0.0003570333577868086+0.0008716764975503862 I,-0.0001708728570181395-0.0002073199040063543 I,-0.0000616855091601308+0.0006519854565884383 I,-0.0001488946531572047-0.0004788291050072864 I,-0.0001433826168198627+0.0001408320266941597 I,-0.0001571529073789517+0.00004341920786937743 I,0.00008033462962055638 +0.0007100006737016494 I,0.0001589477231379817 -0.0004887256283591113 I,0.00006312205835804372 +0.00003604041821323694 I,-0.0003810748313842829-0.0006991531200844193 I,-0.0003489138577009585+0.0004383681020939631 I,-0.0002321133648401709+0.0001175282326205519 I,0.00042466892095623 +0.0006774574010086825 I,-0.0005932393337407348-0.0005467955766171967 I,0.0005942472192938968 +0.000100649635160672 I,0.000466073839216658 -0.0006469299703611152 I,0.0004353432097679798 +0.0001721738763088772 I,0.0003133159076107429 -0.0001536660050460651 I,-0.0002485467931862307+0.0009659825198079472 I,0.000562561175962349 -0.0005620859095931872 I,3.892666198917042*10^-6+0.0000499854417197667 I,-0.0000776946901189081-0.0006411435118972232 I,-0.0004599761122107507+0.0006032396956597574 I,0.00006734638930438395 +0.00003482197351117925 I,0.0002030883288580975 +0.0007486655349425647 I,0.0005042148896464618 -0.0003277780879264214 I,-0.00008622694651604836-0.00005544956758053416 I,-0.00001241810908662249-0.00002849072373527093 I,-0.0005014232997381965+0.0004149623054515619 I,0.00001654177806331334 -0.0006970627125189009 I,0.0002488306553813057 +0.00002096495192307914 I,0.0004888304411227649 -0.0002238003078962721 I,-0.0002687589474291731+0.0006954386898127409 I,0.00009775207617041634 +0.0001193249128009168 I,-0.0003939734420332653+0.0001228698318349428 I,0.0005881852805923116 -0.0008858937032851403 I,-0.0004681990534792123-0.0001488968669153602 I,-0.0008487485471060963-0.00009244563218741303 I,-0.0005661675843867343+0.0006446746511381677 I,-0.00009711169927390432-0.0002940095927347868 I,0.0006650630529151695 +0.0004114010089017844 I,-0.0002997506375291868-0.0001880423697240014 I,0.00003759020185780418 -0.0004994944628944041 I,0.00006350751724528569 -0.0007540396382969824 I,-1.702962583293386*10^-6+0.0003223251150628706 I,0.0002004555916966369 -0.000297943927387735 I,0.000232273617428995 +0.0006373862534703252 I,-0.000243874854355019-0.000850698778029446 I,-0.0002843334962338227-0.00009842495387734836 I,0.0003553889311676905 -0.0002484253859539135 I,0.0001673935830540031 +0.000432933181034331 I,-0.0001347686544390296-0.0006385933267846575 I,-0.0000526445828504607-3.404068985808607*10^-6 I,-0.0001242041099490685-0.001382098100605088 I,-0.000260234672779179+0.0009261922727900673 I,0.0005473160494655702 -0.0006578573170104465 I,0.0004276981869405802 +0.000638161174061507 I,-0.0003204594576197967-0.0001391822555828589 I,-0.0007634967241938426+0.0006902008005233977 I,-0.0004153330851923612-0.0008979955818606455 I,0.0003494024529835712 +0.0003268645976365983 I,0.000452377768145875 -0.0007603603590415576 I,0.0004191528436678922 +0.000800054333176765 I,-0.0005368208776476652-0.000367670947679851 I,-0.00017146301328683+0.0008121605580300743 I,0.0002293289655255787 -0.0001369832956829763 I,0.0003283097882296207 +0.0007216571479241111 I,0.0007396623302612225 -0.0001259581685673119 I,-0.00005372434247429597-0.0001060910718166033 I,-0.00006277021691595304+0.0002456491973995089 I,-0.0004029898530940542+0.0001298806472061771 I,-0.0002066032605757416-0.0009375079027298291 I,-0.0002125643798112582-0.0002668533898178903 I,0.0005132567047087376 -0.000770678187129339 I,0.0004352441173137397 +0.0008456512247237539 I,0.0001444824881355023 -0.0007460010351139664 I,-0.000570687922562718+0.0004534070290763301 I,-0.000170626464562307-0.0007087564783590245 I,-0.0002218673092953793+0.0007967430466805706 I,0.0002593644538319295 -0.00002238453002715268 I,0.0001619859866660891 +0.00052043157888612 I,0.00004080980660186711 -0.0007755266961769858 I,-0.0001822345552951228+0.0002258785527702063 I,0.0003874640388848331 -0.0002672459326453634 I,-0.0003928274742033247+0.0004753361092427047 I,-0.0004041575356294916-0.0004067425345317568 I,-0.0004855847563708057+0.0004269678513581749 I,-0.00007948156974620172-0.0008887968154087835 I,0.0005056640103624395 +0.0002443189970208229 I,-0.0002181249665488945-0.0007287091320439996 I,-0.0002241606385230636+0.0008536434770760876 I,-0.0003939449100735696-0.0003604699676922578 I,0.0002762476689882274 +0.0008367047328234955 I,0.000599476038543742 -0.0005516275299961088 I,0.0002534839040035114 +0.000252315504835178 I,-0.0003494792707764067-0.0004943064485080354 I,-0.0003621379009728411+0.0005644501901526633 I,0.0001315315329035793 -0.0008155004844158114 I,0.0001672688884812412 +0.001073475459276495 I,0.0001318626135355086 -0.00007129865306979374 I,0.0001769587037247333 +0.0005693167207545442 I,-0.00009219498085023758-0.0002907462063681139 I,-0.0003802858971975783+0.0003735017757978589 I,0.0004473034829261145 +0.00002519973894372403 I,0.0002495002183774132 +0.0005171345394179638 I,-0.00006266324800640567-0.0006746401893794326 I,-0.0004479238929414594+0.000220455124929134 I,-0.0002950532940469616+0.00003008493376056642 I,-0.0001080125042669903+0.0008468855595918501 I,0.0003713785202283836 +0.0001708003706885227 I,0.0004011031498819223 -0.00007442476342427182 I,-0.0003414924332009183+0.0001042606852822215 I,-0.0003179089141623214-0.000253374539404461 I,0.00001548998809928212 -0.0008339559609698293 I,0.00009120467738607479 -8.678388517530536*10^-6 I,0.0003455448378225467 -0.0004138870868414827 I,0.0002499473404249788 +0.0006613943704313597 I,-0.0001099984672546937-0.000752997396051554 I,-0.0003479781908922661+0.0001777804005332054 I,-0.00002543814304243079-0.0003305402166987327 I,0.00008356633546930883 +0.0006506426662945936 I,0.0002843632211544981 -0.0004160281451414618 I,0.0002272567021141091 +0.000217158678905097 I,0.0002255981623671792 -0.0008472092713850297 I,-0.0004144851102361887+0.0004078265936465224 I,0.0003888868522722939 -0.0002199451740409447 I,-0.000107242346134237+0.0007681455825460322 I,-0.0003443121462954066-0.0005136993277299891 I,-0.000477993948081749-6.098379120782788*10^-6 I,0.000311368883326708 -0.000991419286717836 I,0.0003975102452858488 +0.0006825556987802029 I,0.0004460876205066299 -0.0004925481952150446 I,-0.0004983522970900515+0.0009030601217485346 I,-0.0005068652974782085-0.001033138533486568 I,-0.0004962255564692509+0.0009009068230174183 I,0.0004528663906602937 +0.0003809253743301144 I,0.0001772894263661715 +0.0008772018525948282 I,-0.0002715193394387764-0.0002168516172553275 I,-6.775234764925967*10^-6-0.00008919548127047318 I,0.0001378175260545661 -0.0004627391538241065 I,-0.0006379181338121985+0.0005402618876452056 I,-0.0000907179660637737-0.0001734693508539132 I,0.0001456728413143887 +0.0006715523254072583 I,-0.0001202706233518468-0.0003425841739338402 I,-0.00008459195793144588+0.00004278422347010196 I,-0.0002238114798299459-0.00008129257077896114 I,-0.0000710972431479709+0.0006225311626116405 I,-0.00003519195301461175-0.0005354428661123909 I,1.551388184559357*10^-6+0.0002675262674823963 I,-8.621988974762992*10^-6-0.0002562981461909521 I,-0.0001392110482171587+0.0005482410175560649 I,-0.0003251858622853036+0.0001334583413147114 I,0.0002962411646565858 +0.0002450330526561532 I,-0.0003808655844665629-0.0004598211159315943 I,-0.00009769010845475403+0.0001322417341054648 I,0.0002983439783672218 -0.0004345037545659902 I,0.0005188074468816232 +0.0004210613623034584 I,0.0001573118544304662 +9.148860824575475*10^-6 I,-0.0002142785938118729+0.0004369225435876137 I,-0.0001497770136112345-0.0004093464642206021 I,-0.0001699270850277146+2.363915439181235*10^-6 I,0.0001548209681872551 -0.000361946692033307 I,0.00002088862371370384 +0.0004034866105838463 I,0.0001780767122139918 -0.0002901959754921834 I,0.0001105311766896559 +0.0002715522039290409 I,-0.00008240602511274543-0.000778268243836639 I,-0.0003164071849387846+0.0003446081701309428 I,0.0004125751242184349 -0.0005039059193028798 I,-0.0003207366667338233+0.0008613450958592434 I,-0.0002012339433248277-0.00006749441890291381 I,-0.00008456715218763624+0.0006115608979767118 I,-0.0001632756452284621-0.0003988899537987149 I,-0.0001748639521989615+2.247440864932477*10^-6 I,-0.0001714673415120692+0.0002470413158444932 I,0.00001901205757178416 +0.0006754562959229388 I,0.0002045925170832777 -0.0005610974182174616 I,-0.00004113202127488678-0.0001112127446852341 I,-0.0004738443559974314-0.0004133428177525597 I,-0.0004652946110216988+0.0004854871841770133 I,-0.0004520578935747259+0.0002610458304108571 I,0.0006148235773414674 +0.0005017998037042968 I,-0.0006056914366438345-0.0005264327381323173 I,0.0003800931021343988 -0.0001889806686816278 I,0.0004456900430011281 -0.0004737763704893165 I,0.0004968575722589895 +0.0003771504572378499 I,0.0001564792389892996 +0.00004081191946638187 I,-0.000217094402048612+0.0006302931425807948 I,0.0002028443879295461 -0.0005287849238555774 I,-0.00002795231563043105-0.0001179341485849852 I,0.00007744054561229222 -0.0004591170913364292 I,-0.0002699282947130916+0.0004544277222195787 I,6.583149703272635*10^-6+0.00006598997247759884 I,0.0001118727265911563 +0.0005398775046355066 I,0.000264353139336271 -0.0005577459975191795 I,-0.00004874181712725794-0.00001670815666787661 I,0.0002356029947676329 -0.00004106623986178064 I,-0.0006244674458119143+0.0001923170003642644 I,-0.0002935404945987332-0.0004720082609997951 I,-0.0001874595239245873+0.0003295589646266792 I,0.0000908268201468675 -0.0005183832641726698 I,0.0003001421703711584 +0.0001950145459669617 I,-0.000475057169459865-0.0004563739828103461 I,-0.0004461848595648117+0.0004473427464980377 I,0.00006480752844803154 -0.0004196888371569685 I,0.0002384417557558187 +0.0004612439253781776 I,0.0002696206901586586 -0.0004041715749714926 I,-0.00003491082679131083+0.0001404446744403731 I,-0.0007911298567898915-0.0007760576451019371 I,-0.0001889209060333883+0.000651850449347797 I,0.0003557611611984389 -0.001057517856798705 I,-0.00001473842319165424+0.0009186730123459175 I,0.0000125692413283566 +0.0003889335092816454 I,0.0001197678797705788 +0.001269592236500172 I,-0.0006949616753182892-0.00004512951184903222 I,-0.0004439778746961423-0.0004263407854670273 I,-0.0003644496583627903+0.0002624566897866673 I,0.0005728343820883844 +0.0005136544274414412 I,0.0002713436842922543 -0.000137616332835436 I,0.0002402888803390651 -0.0002646905345931644 I,-0.0005198652008840316-0.0006542549688569399 I,-0.0004242195098729756+0.0003247426390407052 I,-0.0003477383090700712+0.0002309952777261484 I,0.0005765033732055205 +0.0005197002098553276 I,-0.0005405622698584891-0.0005409798485967014 I,0.0003170145563554794 -0.0002328984395373857 I,0.0003867919234816677 -0.00026228995148096 I,0.0004727236095029279 +0.0007336206634234057 I,-0.000339383834534411+0.0002342605544168485 I,-0.0002603923761297767+0.00003641714528521318 I,-0.0006003167934348454-0.0004708714748974493 I,0.0004565788425754135 -0.0001650058290092086 I,0.0003877391477065969 +0.0004722576528411923 I,0.0007408365781551218 +0.00001632967539170055 I,-0.0001267246937283639-0.001313505199451237 I,-0.00002526604341011063-0.0003686469863954546 I,0.00003309126667282023 -0.0006925997131857598 I,-0.0002710863450133784+0.001018782558584964 I,7.9233510068797*10^-6-0.0006019074572314464 I,0.0008401981820683052 +0.0006553028530080108 I,0.0001569964612681443 -0.0002240296484983609 I,-0.0001033330104809981+0.0002962017818123196 I,-0.0001360363384025953-0.000325086274455174 I,-0.0001275095944154963+0.0005106323087810127 I,0.0004343078977600521 -0.0004789686226627804 I,0.0004401939627994585 +0.0002670665332889587 I,-0.0002106820677463641-0.0001154893553953889 I,-0.00006248669079221953+0.0004845588445876563 I,0.0001695162687019899 -0.0003871723092775946 I,0.0003156768584650253 +0.0003537026010062655 I,0.0007092758906805099 -0.0003689371544864848 I,-0.0003418148368495011+0.0001049877963475579 I,0.00017229879521614 +0.00006486340881968303 I,-0.000186427074442474+0.0006178328120965964 I,-0.0002142599335257994-0.0005301236029169264 I,-0.0002120735109148353-0.0001586710899633182 I,0.0002648348286744034 -0.0006218133842978215 I,0.00006878377994070452 +0.0004930286876857282 I,0.0001557752449460716 +3.608606940390028*10^-6 I,-0.0002006808101103495+0.0006484795446847281 I,5.231014814448357*10^-6-0.0006108162592532981 I,-0.0001761969276769911+0.0001419123095665477 I,-0.00026470182867161-0.0003108980125167958 I,-0.0002707017052622701+0.0004766995899044379 I,-0.0002328254887638097+0.0001925295642062119 I,0.0004133329566430179 +0.0004122553961381765 I,-0.0005091225324945651-0.0004082713539835666 I,0.0003843963463664224 -0.00004589646013349126 I,0.0003873785717860836 -0.0004784663951427987 I,0.0003894677595259644 +0.0002021128721753307 I,0.0001716228306639593 -0.00002989746406381125 I,-0.0001889582928107812+0.0006563568631831602 I,0.0002843409219552052 -0.0005054225327008274 I,0.0000708276788818965 -0.00006807849307240785 I,6.945742740024028*10^-6-0.0002945779060918401 I,-0.0001796764459452666+0.0004334198924042683 I,0.00005203359677217547 -0.0001585808306276773 I,0.0001838408974018113 +0.0003883873885397095 I,0.0004495711162670928 -0.0004881438963195713 I,-0.000185360545083401+0.00009680211464540153 I,0.0001726775156651261 -8.124527160732019*10^-6 I,-0.0002759947256492631+0.0003648004393051137 I,-0.0001964066962240995-0.0004878130038022246 I,-0.0001947874056096693+0.00005407225183023181 I,0.0001677038496095459 -0.0005233501628816091 I,0.0001692043613727896 +0.0004048635979293824 I,-0.00002907258437269729-0.0003277319756868638 I,-0.0002096617301685454+0.0004453921323196375 I,-0.0000426720315536031-0.0003927848769255823 I,0.00001837577022604597 +0.0003828662522218894 I,0.000106238140217765 -0.0003611137910808893 I,0.00001720694591988083 +0.000272170503573524 I,-0.0001311026560498592-0.0003700699283715375 I,-0.00009648746705203214+0.0004102183978451849 I,0.00005932639321797137 -0.0004729417985390173 I,0.00009043134403761098 +0.0005637975964376907 I,0.00002832834791028849 -0.0001721012242668822 I,0.0000339806274259184 +0.0004011319740663004 I,-0.0001033667066107153-0.0002978053788433945 I,-0.0001500327147061422+0.0002325566983643017 I,0.00008754635646179781 -0.00008759135700775467 I,0.0001942326629191975 +0.0004592727627418023 I,-0.00004184691564039869-0.0002901593364332324 I,-4.515656451366141*10^-6+0.0001676301136433441 I,-0.0001064154137696183-0.0004523865979701763 I,-0.00006840884946381246+0.0003737130081235814 I,0.0001010655027555192 +0.0001776341403077266 I,0.0001666666053351434 +0.0003086505495042836 I,-0.000440102736592701-0.0002318743423334398 I,0.0002798364846630511 +0.0001810057811737341 I,0.0003481669796473087 -0.0005706906334266969 I,0.0004119560129167192 -0.00009548554113129831 I,0.0004161478483127154 -0.00006853296161833708 I,-0.0002670977537655731+0.0008109094658706208 I,0.0004576287314387075 -0.0002486063245402421 I,-0.0001276040422318986-0.00002687996996694363 I,0.00006083866251250171 -0.0005743858885466521 I,-0.000444991370346149+0.0003246006021010631 I,-0.00007637002132207747+0.00009702783378296948 I,0.00006708863436955604 +0.0005682217294508353 I,0.0003852542963206324 -0.0002959458023121027 I,-0.00002598350652074731-0.0000514521175467351 I,0.00006923791914421408 +0.00004079537458464242 I,-0.000420197195840803+0.0002618137616318966 I,-0.0001159521544762832-0.0005115669447866606 I,0.00001641254454475868 -6.741632712529569*10^-7 I,0.0003178368007779414 -0.0003349358879859953 I,-0.0000372795759762502+0.0004282526862784092 I,-0.0000338428676523182+0.00006579738323538613 I,-0.000346787557626405+0.0003098416143709551 I,0.0003122152565922327 -0.000635094822711665 I,-0.0002073799326275183-0.00005099250567629815 I,-0.0004938644343281905-0.0002108536887095454 I,-0.0004342637484074934+0.0004024464116021828 I,-0.0003980485757618759-0.00004413221679282121 I,0.0005191204644557213 +0.0004752228709793733 I,-0.0003978126540324576-0.0004555778594800521 I,0.0003430785530908808 -0.0002392364594776487 I,0.0003182147863337106 -0.0003624877381803528 I,0.0002774544803108783 +0.0004355454809789523 I,-0.00002733078726445635+0.0000372989731864713 I,-0.0001115538580148723+0.0004372201265829701 I,0.00001647212097503702 -0.0006264469295136169 I,0.0001872812132150388 -0.0001938423127766155 I,0.0001102715100205915 +0.00003414592402783727 I,0.0001261649170465647 +0.0003762464469219752 I,0.00008419289087893342 -0.0005168895907168101 I,0.0002329675803284535 +0.00004905536603849 I,0.000521403914695619 -0.0007995151233022526 I,-0.0004418526696796773+0.0003798663027905005 I,0.0003621708039612107 -0.00007294323542488572 I,0.00007238302799210756 +0.0006826305164744253 I,-0.0002725861974639778-0.0004162672135505149 I,-0.0004585374657579979-0.0001235911397240015 I,0.0002201558847381972 -0.000766144155478689 I,0.0001900408603388129 +0.0006000164127179147 I,0.0005727230435616417 -0.0002073981838607193 I,-0.0001332010281194042+0.0006959473758006368 I,-0.0003523937701190668-0.0007473054025404842 I,-0.0004987521668199016+0.0004961770309761127 I,0.00006421155095998097 +0.0002177773698557577 I,0.00007317013520918955 +0.000792623988402202 I,0.00031103747793759 +0.0002853626999362669 I,0.0001921614690870685 -0.0002228484502910872 I,-0.0003219538532607263+0.0001136320491181048 I,-0.0003846728847652267+0.00004240849327905872 I,0.000133846205981886 -0.0007131095711716744 I,0.0003682101553290916 -0.0002587568802270811 I,0.0006332887939752694 -0.0001481024949138182 I,-0.0001502755642272024+0.0007456626132910385 I,0.0001859748198139661 +0.00001898142009746322 I,-0.0005428648995817128+0.00002892323278867711 I,0.0003674850096915006 -0.0008101482052090986 I,-0.0004844851229216928+7.967122394565245*10^-6 I,-0.0005287838664523638+0.0001032358188804915 I,-0.0003792365310091253+0.000530992097973834 I,-0.0001041075412562458-0.0001806867001840845 I,0.0004717543348089747 +0.0002436072207172184 I,-0.0002549121537752435-0.00009526560550251548 I,0.00006564513028064676 -0.0003597771999191715 I,0.00007098530724196349 -0.0006668467986747431 I,-0.0000250728952138424-0.00002085770286169168 I,0.0004383376284310077 -0.0003952417630870611 I,0.0002936476804032568 +0.0008352409844210151 I,0.0003904228048913969 -0.0008306349316325507 I,-0.0005076853967043092+0.0000339298924662162 I,-0.0002384660258912344-0.0007309462912179942 I,-0.0004952921350344866+0.000890391300281316 I,0.0004578366856582227 +0.0005528274029649453 I,0.0003563788951363557 +0.0006354134382056481 I,0.0001274268899681551 -0.0007902346990577089 I,-0.0003394219962314632-0.0002732301016307643 I,0.000604317256462431 -0.0001218703431554692 I,-0.001021474017687386+0.0005889542806170536 I,-0.0004300401237520486-0.0001268079802703321 I,-0.0001983925475292439+0.0005147818972209334 I,-0.00003920285035602898-0.0004023372145187872 I,0.00004459544755752633 -0.0001348437904377994 I,-0.0004667962441345688+0.0002126038547640032 I,-0.0003395666366538003+0.0005202255036402499 I,0.0003721751187923939 -0.0005670458267205442 I,0.00001562768732706647 -0.0001550281079390559 I,-0.0004755108305270604-0.0002670277560253509 I,-0.0005317955579294656+0.0002901663193077329 I,-0.0007571234031468837-0.0002067847197002593 I,0.0005191380007170289 +0.0005892378289345721 I,-0.0001724251006372978-0.0006870728544912065 I,0.0002666023050041925 -0.0001161543470766456 I,0.0001441957823901435 -0.00003271129837046318 I,0.00008388718898327844 +0.000848785213120093 I,-0.0004045266305012247-0.00009378832245250415 I,-0.00002505774442241845-0.00001317573060337999 I,-0.0001956290334923973-0.0004695808757384455 I,0.000336286995326245 +0.0001982324066868934 I,-0.00005537135642532724+0.00001135438686746074 I,0.0002379805330963205 +0.0003137956474660626 I,0.00008149975607875379 -0.0004559657535432635 I,0.00008809332337337776 +0.0002096307200101177 I,0.0001151373455270906 -0.00009930771789991039 I,-8.576659485934475*10^-6+0.0002438051239966618 I,-0.0002133711569521343-0.0001974358970984322 I,0.00002311565492700355 +0.0002081691567091033 I,0.0002011512450047543 -0.0004613875253058227 I,0.0003101646579920971 +0.0000594546699360247 I,0.0003180448863962106 -0.00002019211304997055 I,-0.0001915087576892986+0.0005618260914835094 I,0.0001006508556331395 -0.000218823286233554 I,-0.0002334908750264053-0.0001347834367495244 I,0.0003096008558728009 -0.0004596298569963249 I,-0.0002525584540051774+0.0001539180581936734 I,-0.0002378793981046719-0.00002024426942619856 I,-0.0001013730484163692+0.0003165079450352778 I,0.0000654605072157398 -0.0006935550063669448 I,2.386465302533339*10^-6+0.0002272872502102991 I,0.000311698518976955 -0.0001572427231655635 I,-0.0002548758196990173+0.0001821747898287704 I,-0.0002956574192131975-0.0002655924477099102 I,-0.0003610797099185744+0.0003540556583808158 I,-0.00004797155650229304-0.0005196698579729638 I,0.000256735371624706 +0.0001449558041946302 I,-0.000159504687913639-0.0002866001199029772 I,-0.0001885365025009234+0.0004572888168300598 I,0.0000312023404003469 -0.0004127980069258018 I,0.00001062487333284227 +0.0002633445705470901 I,-0.0000810596913228914-0.0002064151660769695 I,-0.0001617321854982222+0.0002895482329811571 I,-0.0004037609709694352-0.0003052208762014906 I,0.0001872196689046253 +0.0004025196668231682 I,-0.00003591628939201567-0.0005264493815853388 I,0.0001001694052132895 +0.0001362579021640256 I,0.00008507984692000913 -0.0000930595702122223 I,0.00006183368555096959 +0.0006322004467273676 I,-0.0002373221379770781-0.0001191625961830358 I,-0.00006836107020206282+0.0001051139626709459 I,-0.00006036738291695919-0.0003516605622497331 I,0.0002113894276947323 +0.0001454582603951982 I,0.00003620585935036661 -0.0001434580334385806 I,0.00004080940088182025 +0.0002517767957808654 I,-0.00005484733069518556-0.0002412338612348353 I,0.00003039113964070218 +0.0002856313952941295 I,0.0001843985530487752 -0.000324128866402227 I,0.00001415448818285979 +0.0001580849901270279 I,0.00005529215521372683 -0.00008426092608350571 I,-0.0001955818159571536+0.0001068955585160082 I,-0.0001297575045250852-0.0003790828314020029 I,-0.0001229630876845313+0.0002158003395280807 I,0.00003921431669125614 -0.0003567822156426559 I,0.000179576903299401 +0.0002279528639734779 I,-0.0002313047404680609-0.0003760055379246857 I,-0.0001652851575875564+0.0002690112576799428 I,0.00007329708278800147 -0.0002194061797840324 I,0.0001584980137560849 +0.0002827378584618352 I,0.00002218210326153098 -0.0004037952985653744 I,-0.0000474368249552154+0.0001093705952227115 I,-0.000232990044253477-0.0004642920387088701 I,-0.00004091018906520768+0.0004806449036714068 I,0.00007231046451615905 -0.0004627613398059992 I,0.0001718142943241221 +0.0003579334297768133 I,0.00001664002865403544 -0.0001483019958221127 I,-0.00004904573492868173+0.0004179932984557961 I,-0.00009283056164835196-0.0002651606218011196 I,-0.00001582112052251066+0.0002455552751315139 I,0.00009521610317116182 -0.0003106873064196624 I,0.0000709685252830398 +0.0002563044965679387 I,-0.00004104041846458436-0.0003066735650284643 I,-0.000088707169507111+0.00031864935172993 I,5.446809043117513*10^-6-0.000120149064758997 I,0.00003228285022038226 +0.0003544662037794746 I,0.00006600465576728203 -0.0002941115365767458 I,0.00001427898799127755 +0.000144252024381873 I,0.00005058225824823057 -0.0001841376963021241 I,-0.0001746475609819137+0.0001914685881680561 I,-0.00006003756387474004-0.0002960391383543583 I,-0.00002833258032294614+0.0002724485102880239 I,0.00004258899751312999 -0.0002637740438658687 I,0.00003987384336621486 +0.000248926292355912 I,-0.00006962951591531379-0.0002529040964385104 I,-0.0001287354658171085+0.0002023857192146014 I,0.0001188437668275429 -0.0003621743767248271 I,-0.0000657511576949238+0.0002168145481558556 I,-0.0001192141609432037-0.0001191795309284121 I,-0.00009772486846915572+0.0002894357821828399 I,-0.0001360807066854852-0.0004810181808848674 I,0.0000809138505858205 +0.0002666742552521098 I,0.000186562410707919 -0.0003037939023798486 I,-0.0001532654434060248+0.0001287430852602708 I,-0.0001940204903385886-0.0001239650460020725 I,-0.0002492633365320612+0.0005730965311392565 I,-0.0003052528043113174-0.0003848063739599147 I,0.0002120663781706881 -0.0000779090825253808 I,-0.0003543235308054199-0.0003957773825392044 I,0.00011456218515104 +0.0004263907623998853 I,-0.00004530318074396609-0.00001247586600530992 I,0.0003533245562626363 +0.0002792618636966885 I,0.00007823085932664679 -0.0005732884692053617 I,-0.00001058374765144318+0.00005826808100776641 I,-0.0001966683744753225-0.000253694680308643 I,-4.405452076085466*10^-6+0.0004873086507719656 I,-0.00008285643076176234-0.0004396768797193089 I,0.0002696006581614854 +0.0002710333352595476 I,0.0001024003279333274 -0.0001960660047031488 I,0.00003769955920738695 +0.0003723401347340954 I,-0.0001081335475012165-0.0002077382695379892 I,0.00001327064351151404 +0.0002580195968561975 I,8.87342502768817*10^-6-0.0005017863169499803 I,0.0001313707303098835 +0.0001744978189194286 I,-0.0001232196886958617-0.0000702519955126451 I,0.0001254074438356678 +0.0004628733009111608 I,0.0002813946954109268 -0.0003397331166902458 I,0.0002422783664638385 +0.0001412420609791031 I,0.0001657268482895225 -0.0005432993082438528 I,-0.0003491329303012277+0.000247912142293424 I,0.0003010562890911584 -0.0002611819472537367 I,-0.0001123950347579291+0.000654961523445975 I,-0.0001862187374557225-0.0001361831714137786 I,-0.0002104002779172315+0.0002663938057538515 I,-0.00007779911934148197-0.0004590994190275023 I,0.00003682341774041076 +0.0001451700687163287 I,0.00007539659728669259 -0.00008473945573210169 I,0.00003042208368263381 +0.0005672491999082872 I,-0.0001447812841909659-0.0003901775247862515 I,-0.00008305513135717883+0.0002545769491295789 I,0.00001994782588309728 -0.0001355765840377475 I,-0.0000336164249638174+0.0004308496253735554 I,-0.00006408115371037998+0.00008524125588435057 I,0.0001534476924342364 +0.0001004485473760151 I,-0.0001861271407599643-0.0002289294390446404 I,-0.0001338405507653482+0.000137015267292118 I,0.000105576583028209 -0.0002919925225973787 I,0.0002452167864923 +0.0002612728370756932 I,0.00005397705490786156 -0.0000790572429056023 I,-0.0001187147980149773+0.0002045619927498506 I,-0.000197179400060586-0.0001207135063887343 I,-0.00006964407377319561+0.0001431772861165547 I,0.0001986808228505587 -0.0002211978714617792 I,0.0001011909102221947 +0.0000631457849065507 I,-0.0001573172138030601-0.0004059977848897875 I,-0.0001645675223706413+0.0001344840330676549 I,-0.0002551717094193768-0.0002550344497045866 I,0.0001171070552548848 +0.0004673134563449849 I,-0.0001023729501615009-0.0004525208942169299 I,0.0002582354011718596 +0.0001891020714204044 I,0.0001416471403518517 -0.0001528414989698372 I,0.0001019113290683692 +0.0004194006510622215 I,-0.0001243559349823405-0.0001044223701656635 I,-0.0001015128243227093+0.0002221437053348684 I,0.00003325662662803063 -0.0003304882583756958 I,0.0001907389300110293 +0.0001183863180005871 I,-0.00001140005886984057-0.0001053357120297238 I,0.00004837327620007168 +0.000274080566022968 I,0.0000521847907357223 -0.0003035080803248455 I,0.0001003036001322447 +0.0002097272002970078 I,0.0001911401001615458 -0.0003033469404034856 I,-0.0001172824367447927+0.0001997360653220159 I,0.00005860709557256157 -0.0001556282464318271 I,-0.00006296125702089062+0.0003321239525602647 I,-0.00004420692965197915-0.0002877749848151037 I,-0.00003039251302374599+0.0001499639797804029 I,0.00006066322067013683 -0.0002769415333482897 I,-4.631727962857621*10^-6+0.0002753944106184845 I,4.834208748863404*10^-6-0.0001826567298043295 I,-0.00005000132141174012+0.0002641134292659803 I,0.00003268876379060847 -0.0002824713928928383 I,-0.00001968214382429174+0.0002084150107596911 I,-0.0000240730238859847-0.0002488167861992911 I,-0.00002585585020959447+0.0002420257309857505 I,-0.00002680678719481945-0.0002674140343509126 I,-9.761856974431522*10^-6+0.0002658235380120542 I,0.0000364497976815317 -0.0002770607879561584 I,-2.722403988934921*10^-6+0.0003094717855959287 I,-0.00001966988304399716-0.0001900402894250589 I,-9.718499222249406*10^-6+0.0002866250303801451 I,-0.00005332748060302773-0.0002489282436143848 I,-0.00004739661062661477+0.000168667917852077 I,-0.00003252055042734968-0.0001005074343866823 I,0.0000621123806325884 +0.000330593255241724 I,0.00003400044609278192 -0.0002351717000650422 I,0.00003400738685595646 +0.0001099338027229197 I,-0.0001269067310274445-0.0003580946898582203 I,-0.0001142845623583203+0.0002429868015105942 I,-0.00006975659388807374+0.00002294178067307429 I,0.0001670960998366259 +0.0003240109389449038 I,-0.0002814060873760104-0.0002772925293088564 I,0.000248823056246482 +0.0000985378579295435 I,0.0002309066860791883 -0.0003470512919913326 I,0.0002367401494874393 +0.0001082052031228881 I,0.0001586156560914271 -0.00007556043783609712 I,-0.0001424041798068183+0.000489646921082255 I,0.0002457578550272618 -0.0002864862772837292 I,2.848796697093548*10^-6+0.0000250540909124304 I,-7.76783621782282*10^-6-0.0003201500559988807 I,-0.0002059270223113066+0.0002976196463602477 I,0.00001810057952576964 -0.00003727493295270706 I,0.00009590871028002356 +0.0003751254144788926 I,0.0002822562124444945 -0.000193843864649373 I,-0.00005199743886520498+2.584928897331505*10^-6 I,-0.00001206290808863491+1.130738689244071*10^-6 I,-0.0002529651680676629+0.0002245965702062664 I,-1.405314488830345*10^-6-0.0004322403954869673 I,0.0001164067861673286 -0.00005157006109050828 I,0.0003467234662279968 -0.0001712944237701615 I,-0.00009790942883726358+0.000458305719858151 I,0.00008795374217681866 +0.00001046149618672986 I,-0.0003365796349172694+0.00007342742496654191 I,0.0002952512613237623 -0.0005849449233446935 I,-0.0003184888189588691+8.823270912717676*10^-6 I,-0.0003803105727443956+0.00006087200167159256 I,-0.0002788979987764049+0.0003854683480443417 I,-0.0001569773579924785-0.000313585284046561 I,0.0002927448670621766 +0.0002116151198212027 I,3.784495513898238*10^-6-0.0001985542731333021 I,-0.00009890285183047536-0.0001259725441036273 I,-0.00006945075730599711-0.0002701835663728764 I,-0.00009842416212964222+0.0003538648322557354 I,-0.00002994191760814526-0.000255731865873538 I,0.0001451692697322246 +0.0001781202212016891 I,-0.0001680088595258783-0.000357598765291436 I,-0.0001141409725644971+0.0001711772303074653 I,0.00009227644520538185 -0.0002358842924005157 I,0.00006670334747371336 +0.0002349707218250981 I,-0.00005249089687999599-0.00025080835899456 I,-0.00008273195041555361+0.0001554236008983139 I,-0.000246065107410608-0.0004078278902709164 I,0.00005160830515903182 +0.0003877269473302959 I,0.00004701539368858984 -0.0004251853144725953 I,0.0001236065716062038 +0.0001968475902519439 I,0.00003719034506595473 -0.0000912287130830244 I,-0.000023452280961496+0.0004803081429269546 I,-0.0001709223221335664-0.0001580732366311536 I,-0.00003930260745899375+0.0001504376775584269 I,0.00005817031155043114 -0.0003102911011971961 I,0.0001720713990347881 +0.0001460413513207907 I,-0.00001794832030161404-0.000186897257673756 I,-0.0000598330177776922+0.000263254734174879 I,-0.00002384314541048286-0.00008916866929042186 I,0.00006369201184445413 +0.0003272510284462059 I,0.0002435208138935038 -0.0002598544868773319 I,5.379816269295204*10^-6+0.00001758660636980211 I,0.00007764098302277658 +0.00001514041552777253 I,-0.0003251768540986906+0.00006961600936127893 I,-0.0001620971984044518-0.0003566902924799679 I,-0.0001089980869618607+0.0001594107511968106 I,0.00006158014296495794 -0.0003079582727155644 I,0.0001703759252395913 +0.0001494481019648812 I,-0.0003311718780702956-0.0002594144218678963 I,-0.0002154367257870605+0.0002356190728321938 I,0.0001594696211137157 -0.0001730967564941722 I,0.0002167270496967942 +0.0001507258333875505 I,-0.00004998092666162422-0.0004649060395583686 I,-0.0001333711010547973+6.767684207266512*10^-7 I,-0.000376880354927823-0.0004332662494843479 I,1.088299940947676*10^-6+0.0005776283669688376 I,0.00003086223648188195 -0.0005570784686674655 I,0.0003365524350490791 +0.0003472785631533028 I,0.0001020771065746043 -0.00002805550142718356 I,0.00001978873699158285 +0.0004725617710143523 I,-0.00016270441752788-0.000124457094523544 I,-0.0001429881601830417+0.0001858088462485687 I,0.0001565990794658041 -0.0001750583274488258 I,0.0002101807441737768 +0.0001515740542538222 I,0.00004673173588078657 -0.000244461116806774 I,-0.0001348962135471589+0.0001387305292500161 I,-0.0002005932604075609-0.0001001547733134605 I,-0.00006590561016838297+0.0003462006915610372 I,0.0001539460886072239 -0.0001480877350751533 I,0.0001765887947009097 +0.0001102345450536003 I,-0.00009215712013570902-0.00004561158659874898 I,-0.00007304681844375602-0.0001168434572013512 I,-6.864638829628144*10^-6-0.0003892373816572651 I,-0.00002969264779613288+0.0001450765055568445 I,0.00009489029150523879 -0.0002032155654956711 I,0.0001219633800023231 +0.000331531860878787 I,-0.0000480363148669346-0.0004464055783917118 I,-0.0001285108416322193+7.467036399487861*10^-6 I,0.0001016473136677415 -0.0002139855946977118 I,-0.00002388303687748738+0.0003054926717245666 I,-0.00001785084910274936-0.0001080778438197704 I,0.0000397194680486247 +0.0001695223339225079 I,0.0000781546020427981 -0.0007031316050323142 I,-0.00012047306004109+0.0002124463006560036 I,0.0003982859472999836 -0.0001431917702665092 I,-0.0002436072874631361+0.0002124604145383877 I,-0.0003846030057092531-0.0001316041805138603 I,-0.0005069401785480087+0.0004270604747809473 I,-0.0002708772465191832-0.0005757561407041921 I,0.0003983781423404967 -0.00009907704698616034 I,-0.0003563095204771367-0.0004221128071834085 I,1.142153281113759*10^-6+0.0005500097495093214 I,-0.0001237149035834052-0.00002998229791443557 I,0.0003676497988243345 +0.0003372031927710706 I,0.0001577263536863861 -0.0005359835003620066 I,2.510996076548783*10^-6-2.683304313829572*10^-6 I,-0.0003217029853534631-0.0003299561433980777 I,-0.00005327261032675097+0.0005305331346273376 I,0.00002630515173369947 -0.0005271522728923636 I,0.0003319952676418714 +0.0003419965795952007 I,0.00003657835274820851 -0.00002419588444519691 I,-0.00009943693634899225+0.0004936470938283336 I,-0.0002998007034246936-0.0002815632278225543 I,0.00007709121966124478 +0.00009095618805639439 I,0.0000349829625648963 -0.0005217487565062421 I,0.0003053665617724819 +0.0003273831239286079 I,-0.0003173778885521843+0.00004654914466258317 I,0.0001945009516869352 +0.0005299610211511088 I,0.000420053493620584 -0.0003482560586137269 I,0.0003416551751668886 +0.0001492807912366119 I,0.0002568850979145921 -0.0002940217293223747 I,-0.0003650725074822163+0.0001250030243758638 I,0.0001735443330249331 -0.0001696302948361393 I,-0.0001291656579500369+0.0006245201424269766 I,-0.00009924070894201419-0.0001769603216703357 I,-0.00006354845041004007+0.0001201794793523785 I,0.0000166464345152329 -0.0003413191155367863 I,-0.00003249380042953665+0.0001898330806785268 I,0.00008324200674342559 -0.00003987673115892651 I,4.617511652951346*10^-6+0.0004394523017194607 I,-0.0000979249152288167-0.0003286704344682663 I,-0.0000781730767590865+0.0002005777459975579 I,0.0000134760026589028 -0.0001418255093257047 I,-0.00001548965824813544+0.0003532025564405856 I,0.00001393018213004516 +0.00004790238501613198 I,0.000093093733701707 +0.00008829356560405538 I,-0.0001367272454463097-0.0001498308217055439 I,-0.0001128581555104266+0.0001473270131854589 I,0.00005386473006037462 -0.0002935139015687377 I,0.0001554998327270692 +0.0001457565916613401 I,0.00009105470823973281 -0.0001440916479422027 I,-0.00004497702759904603+0.0002218800610484996 I,-0.0001434895369942942-0.0001406480442063968 I,-0.000109022108829098+0.0001807934757384392 I,0.0001099925351503854 -0.0002090599282049076 I,0.00009127055842471474 +0.0001228668538855705 I,-0.00004734028017173248-0.0003598711722665904 I,-0.00009516077666457615+0.0001069048026293438 I,-0.0002330599365206689-0.0002383809351450774 I,0.00002912760382020075 +0.0004166046258888833 I,-0.00005623576501480967-0.000429679450322427 I,0.0002423746717287655 +0.0002915996312338599 I,0.0001264339826420953 -0.0001008289702239367 I,0.00009618467250317167 +0.0003422658103032836 I,-0.00008724819617366817-0.000108938064276724 I,-0.0001395147600222738+0.0002181843589936453 I,0.0001365939903139489 -0.0001876708604151769 I,0.0001709460908735885 +0.0001607175573889608 I,-0.00003027046782969178-0.0001941275478314397 I,-0.00007424173067719688+0.0002042101067334828 I,-0.00002375687405239721-0.0001560404499810401 I,0.00004144435464998211 +0.0002945075244626276 I,0.0001856574164923882 -0.0001132733833747818 I,0.00001638226771306448 +0.00008772694300567414 I,-0.00005443525813737659-0.00006769527423448718 I,-0.00009253154878027932+0.0001374122725060934 I,1.436500273212311*10^-6-0.0003319006168334612 I,0.00003378907760970092 +0.00006834383351134598 I,0.0001277115212865945 -0.0002148825405525488 I,0.00001946009051158565 +0.000312096190057743 I,0.00002042591339041697 -0.0002331963503161487 I,-0.0001141205672140561+0.0001631336087882019 I,0.00001498551388432614 -0.0002781735101188193 I,-0.00005715205446887534+0.0002488190184883104 I,0.00003577648399383901 -0.0001267186800264478 I,0.00002515345848436745 +0.0002393622346450997 I,9.09059009533566*10^-6-0.000298186010867086 I,-0.00005301756107127021+0.0001568967935088021 I,0.0001048277701913937 -0.0001948048534830146 I,-0.0001465180023902544+0.0002818261129270732 I,-0.00008557627523166884-0.0001726856122709031 I,-0.00004842090807988393+0.0002575998494543557 I,-0.00003417955897389839-0.0002521539778496351 I,0.0000139236295773062 +0.0001002206458201651 I,-0.0001365658595715406-0.0000944351595063417 I,-0.00003873600922085271+0.0003125473986298599 I,0.00005624248680164605 -0.0002165977820158596 I,0.00008482647459881393 +0.0001053903014591824 I,-0.00006942715701906291-0.0003337386943391955 I,-0.0001168614541272668+0.0001562933705001049 I,-0.0002274185782005486-0.0001005378669079681 I,0.0001038944163421607 +0.0003630025753955629 I,-0.0001428043154223333-0.0003891534113545328 I,0.0002216212257854415 +0.0002029982566715984 I,0.0001684416740184836 -0.0001406499925872651 I,0.0001723389103330678 +0.0002998538086100562 I,-0.00003621055281616404-0.00005409443638883085 I,-0.0001654777113307629+0.0002410174573077708 I,0.0001144834000181177 -0.0001563961597715742 I,0.0001246599563691975 +0.0001085044944057662 I,0.00001590524784493632 -0.0002179992415062848 I,-0.00008919774822761586+0.0001617937071683871 I,-0.00005994144871135999-0.0001324278889313368 I,-4.182104334584351*10^-7+0.0002989441693214995 I,0.0001069682214019546 -0.00005734038731241076 I,0.00007159788309329002 +0.00008658736053526244 I,-0.0001247457969224537-0.0001133143686773429 I,-0.00009006185549691775+0.00009650181804211291 I,0.0001011136556507463 -0.0002990644149828135 I,0.0001981388876221129 +0.0001248854106540071 I,0.0001310694338698661 -0.00001137709447268322 I,-0.0001492660559407656+0.0002785391188208211 I,-0.00005762483671374767-0.00007685308245901586 I,-0.00007156551011556391-0.00003506296492764186 I,0.0002767841571776689 -0.0002223786719779624 I,-0.00004236667594419913-0.00001575329264951827 I,-0.0002760087599798117-0.0002388422219917924 I,-0.0001622841429183621+0.00014546026655028 I,-5.704215381553179*10^-6-0.0003415406594035207 I,0.0001035439567489709 +0.0003141345745264266 I,0.00001753644189242818 -0.0001780755344536682 I,0.00009581316436965787 +0.00004335032325245052 I,-0.00003721276004760502-0.0002464838708811667 I,-0.0001166459892597335+0.0002054220258963057 I,-5.342335277021818*10^-6-0.0002489825607510877 I,0.00009215619713195878 +0.0002407218740371014 I,-6.238177402192771*10^-6-0.0003158774348972707 I,-0.0000172998473979479+0.0001541991669139579 I,0.00002527380961915496 -0.000173969796778919 I,0.00001959312581836326 +0.0002359997298612707 I,-0.00001034446301194559-0.0002294466899573773 I,7.183951325340381*10^-6+0.0001487800656301348 I,2.460016797910175*10^-6-0.0003644624018048106 I,-0.00005084421797651661+0.0002721304517877444 I,0.0001012091904422934 -0.0002354285762604545 I,0.00004989104110771692 +0.0002538284345384093 I,-0.00005983629177170904-0.0001521486870458904 I,-0.0001227876322762957+0.0002516934158941616 I,-0.00006941598351348764-0.0002744550945772156 I,0.00002918623514731685 +0.0001737120210917426 I,0.00007031341240471768 -0.0001801016930731731 I,0.0000739221122908834 +0.0002788458454861301 I,-0.00003473349062965528-0.0002316559492137923 I,-0.00005666988600919027+0.0001801094444347285 I,-0.00008586034632445919-0.0001522917882777864 I,-0.00004023873960857272+0.0002882561732165173 I,0.00007091155112215857 -0.00003517250351866587 I,0.0001374084670799268 +0.0001300041980103117 I,-0.000151398526040401-0.00009088462825573031 I,-2.846702502149613*10^-6-0.00002480157825937414 I,0.00006147062204103333 -0.0003465729768911756 I,0.00007127702932089675 +0.00008874027277546422 I,0.0001171495385076913 -0.0001483613027712329 I,0.00003978742363781597 +0.0003269736001243467 I,-0.0000160245737679522-0.0003360679984257454 I,-0.00009340086244186962+0.00004020307255702877 I,0.00005575967175051107 -0.0001730672923097754 I,9.18688197271322*10^-6+0.0002533082864846224 I,0.00002535385169084988 -0.0002084816614633776 I,0.00004310431234924614 +0.000131309106287323 I,0.00005307512355863324 -0.000410056705614955 I,-0.0001058915845246325+0.0002417535985402863 I,0.0001277028272051435 -0.0001827832626458583 I,-3.372153458277839*10^-6+0.0002765934387993015 I,-0.00007830177857440646-0.000204544230137885 I,-0.0001274043259035767+0.0001650031037423299 I,0.00003076481496387097 -0.0002938729649878341 I,0.00006170702359370517 +0.0002471281848102231 I,0.0000860322751659534 -0.000221824899183701 I,-0.00005588141352999009+0.0002384463655831228 I,-0.00003288446245077592-0.0002930792403342559 I,-0.00009791918591632996+0.0002519154532474468 I,0.00001090010800734855 -0.00007267007831851216 I,0.00001242976582034089 +0.0002701052132701825 I,0.00001015624339262967 -0.0002461719970343614 I,-2.93207676856544*10^-6+0.0001300736138430184 I,0.00006654890320270174 -0.0001798629700306998 I,-0.0001145405314008751+0.0002056086201342513 I,-0.00004714716140686415-0.0001691155873913439 I,-0.00002281002529055228+0.0002463213494737883 I,-0.0000248463521779117-0.0001832243133870661 I,-0.00003124814539913988+0.000131126108155195 I,-0.00005336564770363471-0.00006709309101311718 I,6.940277795675778*10^-6+0.0001948163989218957 I,0.0001391191143098651 -0.0002231818985391819 I,-0.00001976501807429585+0.00001845037022947123 I,-0.0002582398047169751-0.0002320939209598677 I,-0.0001736553764626262+0.0002007536241461292 I,-0.0000205180597855496-0.0001304804517885184 I,0.0002114344885013607 +0.0002754275347758169 I,-0.0001290800399747317-0.0001341252184674239 I,0.0001602435248822205 -0.0001113564802153828 I,0.00001682267575717512 -0.0003280179213169812 I,-0.00008787352589356617+0.0001189806544702397 I,0.00004438775473583414 -0.0002810312990093186 I,0.000185284333737235 +0.0003373208880347955 I,0.00007232975375948207 -0.000517199396073698 I,-0.00004112154107617958+0.0001519888613670369 I,-0.0001660899482676699-0.0001476457004944576 I,0.00001539817278586224 +0.0004614419087262614 I,0.0002497300827240411 -0.0001521452420122781 I,0.0002033509141560222 +0.0001852333204454765 I,0.0001498888871743749 -0.0002936855587063472 I,-0.0002037688874173073+0.0001120138681717428 I,0.00007391334022089388 -0.000119747001099375 I,-0.00009046524691360001+0.0003718433335791419 I,-3.966310428773927*10^-6-0.0002806766926356931 I,0.00002860090162258553 -4.675758690866377*10^-6 I,0.0002444056199380329 -0.0002082512861547647 I,-0.00005991071498153461+0.000422145382942109 I,0.0002708137995360718 -0.00009645365160217385 I,-0.0002138846398130167+0.0001088721762720064 I,0.00006021143608502869 -0.0005382590823630106 I,-0.0003985056134636022+0.0002009714618965347 I,-0.0001346934512384886+0.0002833715869100178 I,-0.00004627882834370309+0.0004679233979413662 I,0.0001361519320936244 -0.0003006803115374553 I,0.0001022500984397815 -0.0001119901911956549 I,0.0001803936349172839 +0.0000695440935096926 I,-0.0004901420310878633-0.00003043098321162364 I,-0.0002747319753379509-0.0003036195363544924 I,-0.0002236453108341354+0.0001615538843596224 I,0.00009075842838661107 -0.0003928850564986562 I,0.0002641972445263758 +0.00009295719664960507 I,-0.0002900889571607039-0.000158179281346109 I,-0.0003777398207258732+0.0003303552029592288 I,0.0001419889452212903 -0.0004028188406754902 I,0.0000268247875887087 +0.0001161413396235522 I,-0.000161235108689408-0.0001239162836793326 I,-0.0002798458283909946+0.0001704800683181229 I,-0.0006557724994246595-0.0002706003153627041 I,0.0002965700796673815 +0.0004368590568022598 I,-0.00006592586493750556-0.0005974095308706935 I,0.0001956286066914215 -0.00005302506013261993 I,0.0001354130214083766 +0.00007981360809450216 I,0.00008048476612286991 +0.0007653784479363553 I,-0.0003954480134591014+0.00004239758358619461 I,-0.00004754785053940964-0.00008267008535199086 I,-0.0001865654117643645-0.0004551301741973395 I,0.0003302820755265547 -0.00003963398841749206 I,0.00003052298099751964 +0.0001452316171145257 I,0.0002208294475759238 +0.0002541022055879197 I,0.0000727548597594525 -0.0003053769082963556 I,0.0001455429672129971 +0.00004066023145797701 I,0.0002249545160235828 -0.0005411548186951347 I,-0.0001989394110842633+0.0001776135027268786 I,0.0002979417737239493 -0.00006875912971745256 I,-0.0001830062546018395+0.0002360791513800697 I,-0.0002909556216635929-0.0001599395048251948 I,-0.0003744686375127489+0.000275075685265772 I,-0.0001728401641009725-0.0004545268164101416 I,0.0003109266320122839 -0.00002994391208517763 I,-0.0002947830564073568-0.000388793601723395 I,-0.00002406539177391168+0.0004102947553244463 I,-0.0001148870518876914-8.432272056123936*10^-6 I,0.0003197960345384197 +0.0003390224825655707 I,0.000246117310887847 -0.0004380580369082581 I,0.00008643533219365553 -0.00002443108402222485 I,-0.0002478823708718423-0.0003932843893324676 I,-0.0001963013481822978+0.0004088437406322181 I,0.0001656574124720927 -0.000484255318896183 I,0.0001302011831708924 +0.0005244160713591846 I,-0.0000150191769335334+0.00009975755290059239 I,-0.00003847220261880456+0.0005046040657648346 I,-0.0003050779710481658-0.000148796345008517 I,-0.0001238881591381339-0.00008985345948678007 I,-0.00009674728661366786-0.00002080267210938006 I,0.0003007675910761933 +0.000304783952595235 I,0.00004768925597833428 -0.00001761419918826243 I,0.0001529985521204321 -8.079083203448482*10^-6 I,-0.0001838417285741564-0.0004336755956356738 I,-0.0001193040579827242+0.0001150523377275485 I,5.901428354922914*10^-6+0.00001741522749275818 I,0.000168025793570172 +0.0003269051336672795 I,-0.000237996990948578-0.0001831764692510512 I,0.0003059073393407509 -0.00003370323951410234 I,0.0001474178053872043 -0.000304290702193953 I,0.00006715395419114023 +0.00008333875623410726 I,0.00006768343110791747 -0.0001283051739125684 I,0.00002442369428358074 +0.0003853994179226309 I,0.0001557072635594352 -0.0004373674133534193 I,0.00003274322375252558 -8.824889764939189*10^-6 I,-0.00004898150589753182-0.000137118582563658 I,-0.00007038270938483031+0.00035614638140444 I,0.00009769966909204323 -0.00006350792606424462 I,0.0001634974006988233 +0.0001953499503370342 I,0.0002726006142961339 -0.0004856730364480587 I,-0.0001833158008354912+0.00005604088397046786 I,0.0002908945154822681 +2.878668587253072*10^-6 I,-0.0002624661532205296+0.0002295320534335939 I,-0.0002974567096145469-0.0002107427184538621 I,-0.0003617511833413929+0.0001687952847810232 I,-0.0000747324828690681-0.0004752235714227914 I,0.0003000918923040543 +0.00004838288129792546 I,-0.0002059385868157741-0.000315550660429521 I,-0.00009278825183918128+0.0004186397323046468 I,-0.000087956347908679-0.0001093052211146117 I,0.0002083257355349212 +0.0002883835573994678 I,0.0001268213526931062 -0.0003621637256554482 I,0.000010185255904709 +0.00003741488364301884 I,-0.0002279886561523073-0.000320436567211595 I,-0.00006958000508774359+0.0003985233804764696 I,0.0000795473338452411 -0.0004057228355957779 I,0.0002142114827873705 +0.0003161923872617349 I,-0.00001699779896367039-0.000018127078550947 I,-0.0001306333500255578+0.0003913967567080718 I,-0.0002383900965991053-0.0002768795134248103 I,0.00007677574547139065 +0.00007119334989276307 I,0.00005855960492163729 -0.0003838912989520805 I,0.0002361052003011859 +0.0003188495108075301 I,-0.0002654132331533882-0.00001820160551396966 I,0.0001061227987255285 +0.0004155514031807292 I,0.0002948797477379233 -0.0002265436457225241 I,0.0002486636602000533 +0.0001849670547952275 I,0.0002410424145596331 -0.0001711218582063019 I,-0.0002272123007437184+0.00004500315572493816 I,0.00009626404048320511 -0.00006340517354442869 I,-0.0001683024275354899+0.0003979785773399489 I,-0.0000939852896508211-0.0002262139225698372 I,-0.00004671961219081405+0.00003962689740615885 I,0.00008878922950145924 -0.0002893956236057452 I,0.00002867493762570854 +0.0001944728602672377 I,-8.489826562912368*10^-6-0.0000664960760641027 I,-0.0001220399508305626+0.0003148626150762855 I,5.868555264141364*10^-6-0.0003036925631667957 I,-0.00004483308218476801+0.0001433386707482467 I,-0.00005448877128651554-0.0001336120016691568 I,-0.0000925841874726947+0.0002276549378990714 I,-0.0001765063498074289-0.0001236235669714165 I,0.0001199361521543349 +0.000215697662293229 I,-0.00004657396145774495-0.0002724410928677379 I,0.00002991802669869554 +0.00007802007740462626 I,0.00002458952966717377 -0.0001176023804891265 I,0.00001534098746003964 +0.0003455652386545395 I,-0.0001409182753501518-0.0001366850717656661 I,0.00002981664032901197 +0.00006011499794704778 I,-0.0001544387485218615-0.0002973966414906727 I,0.000104222555270685 +0.0001478765871354681 I,5.902601280492706*10^-6+0.00001421870093868285 I,0.0002004471547273712 +0.0002039959119035201 I,0.00005698685511913618 -0.0003895420546596474 I,0.00004376404754679966 +0.00001287406974205014 I,-0.00001250117231144313-0.0003465531077114983 I,-0.0001095157289964517+0.0003432186689662474 I,0.00009338517673602436 -0.0002606752725673828 I,0.0001753984545457436 +0.0002783097865643632 I,-0.00005837893090527425-0.0001155689403793159 I,-0.0001836871317714799+0.0002392368005129725 I,-0.0001329637808718475-0.0003305986354234794 I,0.0001362869605426439 +0.0001610833123262239 I,0.0000987719812984707 -0.00041769634281566 I,0.0001314070990415685 +0.0003060266763586282 I,-0.0002746702075652288-0.00008251574233209363 I,0.00004868281275969128 +0.0004658128690896434 I,0.0003325437042219982 -0.0001304075559932035 I,0.0002694367637773907 +0.000205206194655928 I,0.0002056806105059264 -0.0002875579245061684 I,-0.0002677437766873116+9.492893824882896*10^-6 I,0.0002373839266378206 -0.00009138055546163296 I,-0.0002913247430607825+0.0004404487693278854 I,-0.000215046463500244-0.0001004981008487439 I,-0.0001758944098002777+0.0002266601085350712 I,-0.0001485865566989311-0.0003559128103681604 I,0.0001145969084856979 -0.00008276169278797266 I,-0.0002901998695698499-0.00009541469712433428 I,0.00004882106390953112 +0.0004811006674082883 I,-0.00006471460272442306+0.00001341692891737947 I,0.0003444842541252567 +0.0001282086131733178 I,0.0001047056109435451 -0.0006149358568450871 I,-0.00002852219795945078-0.00005093122165288833 I,-0.0002347636986291987-0.00004351523906851948 I,-0.00005798353216101638+0.0005051356960844929 I,-0.0001570674210054042-0.0004649380172864123 I,0.0004903279843359085 +0.0004307801971977671 I,0.0002206665176135382 -0.00006290976202450679 I,0.0001547416600119498 +0.0001867676282559379 I,-0.00005414758255969111-0.00009894473917899702 I,-0.0002080001240912149+0.0002975100485818011 I,0.0003682073642343414 -0.0001279002501510985 I,0.0002749782931424757 +0.0002043351147902408 I,-0.0001863642872149194-0.0001839726901322773 I,-0.0001754814950041413+0.0002374740041721748 I,0.00004209453378542105 -0.00006258966954320135 I,0.0001312412582970124 +0.0003730984539640924 I,0.0004170901744805141 +0.0001700808773276069 I,0.00002175951151532894 -0.0001093030559603837 I,-0.0002085477302337966+0.000159864103796478 I,-0.0001317155855083637+0.00003767257048071082 I,0.00006763837121346145 -0.0005014945989985523 I,0.0001413308982866438 -0.000273648747030785 I,0.0003823035296533541 -0.0001725777021954077 I,0.0000228221128517948 +0.0005205879067080347 I,0.0001475749454672513 -0.0002491276060340629 I,-0.0002772179213131445+0.00002517606273596879 I,0.00001646686449559415 -0.0003155712114239506 I,-0.0001606355395507921+0.0002865086579793079 I,0.00008517584706536465 -0.00002633160647905969 I,0.00009420308153856139 +0.0001989929641863303 I,0.0001626113239289382 -0.0004231669563469093 I,-0.0001879430887911162+0.0001004336188225503 I,0.0002472200055669203 -0.00004695936100649605 I,-0.0001597384375694026+0.0003633014317620194 I,-0.0002271124075051612-0.0002278262294370306 I,-0.000280464025684673-5.511695980261999*10^-6 I,0.0001297258367573999 -0.0005094002684794064 I,0.0002003279788621954 +0.000268621339691578 I,0.0001913048767850648 -0.0001441611744980008 I,-0.0002312257729574665+0.0005024237155468525 I,-0.0002308258730069556-0.0005122864242016526 I,-0.0002405436776386219+0.0003551105823320907 I,0.0001348227012160975 +0.000170735608155207 I,0.00001353576207583903 +0.0004448652581654799 I,-0.00015837833740862+0.00005504956490601982 I,0.0001167038339186523 -0.00004554087675130323 I,-0.00007626696361739335-0.0001952237300824003 I,-0.0002440894386042877+0.0001115503528524122 I,0.00004069424779532372 -0.0001499534518523454 I,0.0001673623127805561 +0.0002655495431463987 I,4.799152585910358*10^-6-0.00008112425917505626 I,-0.00005834019855921806+0.00008882895736118975 I,-0.0001263756144843783-0.00009690336857232658 I,-0.00006623506333581808+0.0001646178548314909 I,0.00005095799535078585 -0.0002085842845281725 I,0.00002641162363566238 +0.0001190768555794501 I,-0.00002518380925406055-0.0001670105389332048 I,-0.00006448988895808946+0.000180591400761225 I,-0.000148341226698599-0.00009609932466931729 I,0.00008516886944053626 +0.0001959880419785185 I,-0.0000885806123813434-0.0002428183064740878 I,0.00002653565360364757 +0.00009822636960282842 I,0.0000815587293853413 -0.0001671102769945082 I,0.0001070565347771128 +0.000233044769356141 I,4.21310776185579*10^-7-0.00008695048812112075 I,-0.00002103790964102138+0.0001882748787043002 I,-0.00002080138326838979-0.0002905511554256678 I,-8.650891074785624*10^-6+0.00006684907351788759 I,-0.00003242900921754129-0.0001194144477337501 I,0.00003172159797359588 +0.0002883421168889177 I,0.0001752093394961735 -0.0001465050370232308 I,0.0001412211123981485 +0.000127081207078613 I,0.00007872705278812456 -0.0003959048363485122 I,-0.0002301246701045091+0.0001255525803735817 I,0.0002481291520288331 -0.0001690684000559678 I,-0.0001745287302889301+0.0004541196254367889 I,-0.0001640704661325023-0.00007378472727347297 I,-0.0001483904578294282+0.0002026271423094302 I,-0.00004600728246180024-0.0003186131239494051 I,0.00002271759795381559 +0.00005309859602415256 I,1.32267209032146*10^-6+0.00002404597595038891 I,-0.00003199744244248433+0.000416805530829882 I,-0.00004514966240777487-0.0003158301598091488 I,-0.00006920462792183601+0.0001002272835560929 I,-0.00007813852002884667-0.0000824977282910408 I,-0.0001131603518139811+0.0003013545046644289 I,-0.0001348263337023219+0.00009430729207597708 I,0.0002088234540807208 +0.0001116459925193833 I,-0.0001775764365963794-0.0001739505597970955 I,-0.00001781273872853507-0.00003167189374587786 I,0.0000749893759841606 -0.0002223577294014065 I,0.0001160280019582045 +0.000188412431905952 I,0.00002121223943664816 -0.0001247611117625532 I,0.00004091864116663356 +0.0001800942203506604 I,-0.00008197444862111237-0.0003079046957751512 I,-0.00005990837908454689+0.0001324847396925964 I,-0.00007273637859403582-0.0001269294467572833 I,0.00008064023354212333 +0.0003074317188856665 I,0.0002093122720414884 -0.0001793346442979224 I,0.00011704081430551 +0.0001232775228412878 I,-0.00005056488720624375-0.0002599829294282479 I,-0.0001619061315589513+0.0001567171417696415 I,0.0001105783711321064 -0.0002390560770922512 I,-0.00009577326302090068+0.0003974566713867396 I,-0.00004278060798779687-0.0000879862289084657 I,5.160177153936881*10^-6+0.0002124652420273377 I,-6.289515575277696*10^-6-0.0002158320193442655 I,-0.00002362467749765772+0.0001077500229228763 I,-3.956262741997209*10^-6-0.00006938302467147597 I,-0.00006789046876969193+0.0003147500182883626 I,-0.00008371437035111127-0.0002955452834341555 I,-0.00004327615268040877+0.0002109905599408371 I,0.0001196658698714905 -0.00005818394028074389 I,3.999049789597782*10^-7+0.000261754271950378 I,-0.0002484248768436403-0.00002681423866272459 I,0.00002827898508358853 +0.0001087371337989325 I,-0.00005185802114103692-0.0003696866128559078 I,-0.0001063450248012435+0.0002977165686281177 I,0.0001416411693244818 +0.00003045131870056214 I,0.0003012131585815778 +0.0004322025999284921 I,-0.0001382927847245805+0.0001299986476546745 I,-0.000322136365789127-0.00004709251058885542 I,-0.0001352271053422214+0.0001525942358261595 I,0.0001748748843990411 +0.00005720513473530693 I,0.0002251232622842521 -0.0001062584492446312 I,0.00003311208398111083 -0.0001575223641643822 I,-0.0002809170764271898-0.000253108209406622 I,-0.0001804206825610781+0.0001697233508514546 I,8.843753065702485*10^-6+0.0000238676173222781 I,0.0002062512389582081 +0.0001729220497260482 I,-0.000177672296606586-0.00009040243354334693 I,0.00001876537572322492 -0.00008098571380346914 I,0.00007111294927325505 -0.0002692574339799429 I,0.0001202135959553924 +0.0001189269425639521 I,0.00006642436527211453 -0.00005620487457407593 I,-8.418184249087*10^-6+0.0001763504669682817 I,-0.0002042432380851819-0.0001830084970042444 I,-0.00009344312809490637+0.00002373634143764773 I,0.0001730744873523644 -0.00007191796018999049 I,0.0001620775944876263 +0.00006656622126561701 I,-0.00006164271624989403-0.000383248060071172 I,-0.000092088551165781-0.00004266656682477163 I,-0.0002495523873641149-0.0003611613324264539 I,-0.00003276227466397407+0.0004586720430981024 I,0.00005062948822650142 -0.0004299954359989149 I,0.0003026908085787165 +0.0002965250425862438 I,0.00006212816266964297 +0.0000239051341136545 I,-0.00003339982989632719+0.0003761687728028508 I,-0.0002021211340158299-0.000134845460275285 I,-0.00007781684588226772+0.00009045264201571084 I,0.0001333431303739418 -0.0001966061546849148 I,0.0002630238486680478 +0.0002045828789104465 I,-0.0001397984756606625-0.00007530470981898993 I,-9.408878441128283*10^-6+0.0002310441650807452 I,0.000101725570284242 -0.0001494302447274536 I,0.0001270873148548236 +0.0002349452808283206 I,0.0002241131772880801 -0.00001159614621576434 I,-0.00006856188143986645+3.887984118217511*10^-6 I,-0.0000517844691111473-0.00002968018002914991 I,-0.0001272436303669773+0.0002257006034897093 I,0.00002951458913931663 -0.0002511834505212809 I,0.0001202540511686008 -6.099595367664453*10^-6 I,0.000156619592127506 -0.0001072385321258019 I,-0.000106590575485946+0.000246875103505932 I,0.00004579473439732236 +0.00003192977015786666 I,-0.0001066911005713476+0.000131732281155951 I,0.0001108949032949272 -0.0003159593669562711 I,-0.0001387151945832897+0.00002125224639169835 I,-0.0001717827798806335-0.00004382686106804886 I,-0.0001335832760008997+0.0002688332781132428 I,-0.00006136297443349251-0.0000132689990495414 I,0.0001859838865685981 +0.0000969358016136293 I,-0.0001297630978629886-0.0001234911108500255 I,-0.00007254643688187306-0.00004421912012009226 I,0.00006865454808284235 -0.0002191785284957788 I,0.0001428330170801771 +0.0001786513853686686 I,0.00004455436842717539 -0.00001855184458199811 I,-0.00005235680995273347+0.0001414880260764765 I,-0.0001887282913778779-0.0001305083414848318 I,-0.00006761449358857039+1.615585054511833*10^-6 I,0.0001985887459432621 -0.00008434245426123038 I,0.000121703270015352 +0.00002168139847599347 I,-0.0001157666958646097-0.0003425785848597202 I,-0.0001091635047675842-6.279043386752071*10^-6 I,-0.0001716651217031263-0.0002979448714922876 I,6.916631491373846*10^-6+0.0003953905237664409 I,-1.082584320353505*10^-6-0.0003341699982691781 I,0.0002506110186318452 +0.000219363055227599 I,0.00008503116947689478 -0.00006907823045959869 I,0.00002845006250853846 +0.0002521116291452075 I,-0.0000501423446217216-0.00007678874278447155 I,-0.000119965726587875+0.0001936474222426034 I,0.0001530327404806999 -0.00009522253267835808 I,0.0001152609638306666 +0.00007483724690343466 I,0.00006125700272106322 -0.000208671608479803 I,-0.0001414323144876791+0.00007407825559209888 I,-0.0001758600440705691-0.00004502981670842118 I,-0.00006540073192947505+0.0002581117364732908 I,0.0001111732541665403 -0.0001087873903714044 I,0.0001366564775701499 +0.00007643121833279396 I,-0.00006209567788294014-0.00003431104989518318 I,-0.00004499146004793105-0.00008277025235030173 I,-3.001391177652809*10^-6-0.000262188902241439 I,-0.00002203529798555584+0.0001102665002644765 I,0.00004899479343914282 -0.0001382845181906807 I,0.00006660941348571635 +0.0002214791816415403 I,-0.00001896334199703603-0.0002821029870226008 I,-0.0000493038139392681+0.00002115888059482692 I,0.00006645719555206722 -0.0001374211770624604 I,-0.00001717409586614744+0.0001862380923910267 I,-0.00004102203813486009-0.000101295275891966 I,0.00001772490551446391 +0.0001330378786178944 I,0.0001066506298928547 -0.0003971516001660137 I,-0.00004883847275536049+0.0001349789034674495 I,0.0001917613532971179 -0.00005026671386529799 I,-0.0001324386398471076+0.00009595638316981401 I,-0.0002179618494307271-0.0001751034727905744 I,-0.0002961327154651277+0.0001890338646973847 I,-0.00008689675974762732-0.0003821027010834033 I,0.0002672038622687084 +0.00004624283834818714 I,-0.0001771494825972757-0.0003283510431171215 I,-0.00005755867948897643+0.000321468875150503 I,-0.00008827928495330149-0.00006262484408644401 I,0.0001932608412750348 +0.0002765200211390149 I,0.0001377477729912529 -0.0003157293679858023 I,0.00004096531656393044 +0.00002144810686444877 I,-0.0001455406018865711-0.0002882316385507555 I,-0.00009376479801038598+0.0003240939004377975 I,0.00007882891741884545 -0.0003055007077590476 I,0.0001619976198460392 +0.0002882572550511134 I,-0.00003378858195694717-0.00006133127201600839 I,-0.0001294726466892516+0.0002560203996046705 I,-0.0001241372941047895-0.0002854218581495973 I,0.0001017321526117832 +0.0001267311762336655 I,0.00009307183240605018 -0.0003398608535692695 I,0.00009706096415567432 +0.0003017984966947531 I,-0.0002618139512337715-0.0001132590497341725 I,0.0000263524014884293 +0.0004147899307663308 I,0.0003124141152456199 -0.0000991899188154284 I,0.0002237715652010884 +0.0002019515918114194 I,0.0001042034654702302 -0.0001757907076990648 I,-0.0002112264288946654+0.00001912434591846892 I,0.000132889597060621 -0.0001394843552400725 I,-0.0001942919042296178+0.000420409088756987 I,-0.00008486719227151315-0.00009243552458005419 I,-0.00002296899218972874+0.0001545211182682713 I,-0.00001665968422740589-0.0002345753754693322 I,-0.00002546251328257101+0.00006917517104035843 I,1.53533927248129*10^-6+0.00001141614801811091 I,-0.00001775297022946619+0.0003522859845472927 I,-0.00006976688443512049-0.0002666463193319661 I,-0.00004910245773139073+0.0001233815802946114 I,1.684360558179077*10^-6-0.00007505309335531218 I,-0.00005075779297652793+0.0002826405986174278 I,-0.00009912489239000737+0.0001300626075607484 I,0.0001445658810279353 +0.00005206724432267525 I,-0.0001832986280008562-0.0001584001505458045 I,-0.00008873225429782797+0.00003684055450530027 I,0.0001191056572825436 -0.0001959520285537554 I,0.0002390186027040809 +0.0001707012878745231 I,0.00005829719056014794 +0.00001726939226759205 I,-0.0000966559633369455+0.0001401450873654705 I,-0.0001656703054773091-0.0001224992009886846 I,-0.0000752659515942552+0.00001039043256746003 I,0.0001398236676247743 -0.0001047688163018379 I,0.00009887436171886724 +0.00007863919601274796 I,-0.0000135754339224425-0.0002630109132324369 I,-0.00004505370631107076+0.00002928300456280735 I,-0.0001705972561608794-0.0003145583678155699 I,-0.00005665718089185979+0.0003064762986626345 I,0.00009015338448018283 -0.0003314243090088154 I,0.0000898778678763467 +0.0002988145379907165 I,2.301218513652861*10^-6+9.897662938445081*10^-6 I,-0.00001426471495074538+0.0003449717892108004 I,-0.0001625194320155804-0.0001210850246274315 I,-0.0000764151207214695+0.00001146469259656071 I,-0.00003111920085869878-0.00007007222613970751 I,0.0001527866235688614 +0.0002140352793580425 I,-5.952213183588057*10^-6-0.00009032157325025497 I,0.00005631744952504106 +0.00008612441258621918 I,-0.00002912269498896284-0.0002299069028913424 I,-0.00002474264142664261+0.0001489708610639948 I,-0.00001744599325336135-0.00003270381158190604 I,0.0000458457601012483 +0.0001698663594401224 I,-0.00009243538364198658-0.0001838443099569193 I,0.00005711980202218182 +0.0001378496637469299 I,0.00008214265171103343 -0.0001453134990284121 I,0.0001129030323576009 +0.0001660821169646965 I,-4.185384884951927*10^-6-0.00005327776412755825 I,-0.0000969266749650563+0.0001523150789343819 I,-4.050369024660202*10^-6-0.00009279292865143066 I,0.000053273161065685 +0.00008311873325322298 I,0.00006591503712543064 -0.0001252878915356677 I,1.517182379384288*10^-7+0.0000725900030546939 I,-0.00008152824400336223-0.000180637555657795 I,-0.00003844109117101407+0.0001463359131197856 I,0.00003726738524055531 -0.0001272398257512284 I,0.00004366599961104695 +0.0001576059268568325 I,-0.00003302220250150052-0.0001125011243079366 I,0.00001488047899223509 +0.00007814256060678709 I,-3.268349833860624*10^-6-0.0001944545037137319 I,-0.00001392170846788153+0.0001219166802019181 I,0.00001481517649874236 -0.0001710386313818726 I,0.00004867234376281532 +0.0001686083374373234 I,-0.00003604429459540133-0.0002231995186116659 I,-0.00003003047504884396+0.0001510345966616516 I,-0.00003290477448657303-0.000123445600282821 I,0.00005126226298159794 +0.0002147026152854816 I,0.00009772110816497526 -0.0001833100795529026 I,0.00005004363255702256 +0.0001114265374969169 I,-0.00005299035023483-0.0002056802936405797 I,-0.00008246522103023997+0.0001775959749662526 I,0.00005120333352701282 -0.0002312442305800874 I,-9.656093765705453*10^-6+0.0002901751974659194 I,0.00001143292675338651 -0.00006612813001956472 I,0.00004185897627427087 +0.0002282866570640381 I,-0.00005272433142092707-0.0001007546372392208 I,-0.00009333173792502117+0.00007868069637841206 I,-0.00002046630606490262-0.00002376829643628405 I,0.00005677968146408134 +0.0001656152151760951 I,0.00005966160796119756 -0.0001563761499837281 I,-3.214677512143906*10^-6+0.00004224880479334567 I,-0.000101612795240871-0.0001717673580324045 I,-0.00007751575119017139+0.0001638034168207806 I,-0.00004670783657999478-0.0000763101455810082 I,0.00009332285253565443 +0.0001703757686990082 I,-0.00006437823287281882-0.0001707686665252509 I,0.00003305677284729283 +0.00006100282581122771 I,0.00003438042693195858 -0.0001384882980409582 I,0.0000409852235846449 +0.0002097140364644855 I,-0.00005873604992365146-0.00009682841733127399 I,-0.00001206842644267205+0.00009225684705219852 I,-0.0001023679796212557-0.0001760480691018988 I,0.00005917801642741619 +0.0001160857119411523 I,0.00003163720847061138 -0.00003183090301984075 I,0.0001280546644822941 +0.0001201726955686965 I,0.00001212969509271158 -0.0003097666899908788 I,1.057334960230455*10^-6+0.00004045822924746758 I,-0.00004129566695510609-0.0001833984534558317 I,-0.0000455543903780308+0.0002690958700032059 I,-0.00001030141568779639-0.0002367411812331077 I,0.0001524498125844014 +0.0002338251070141332 I,0.00004968058072533976 -0.00009759314471173325 I,0.0000162142368405169 +0.0001799512929435751 I,-0.00005179372267932929-0.0001257136700284152 I,-0.00005623872132736465+0.0001568264211710681 I,0.00008734128176258768 -0.0001525630986581267 I,0.0001167265671468754 +0.0001531252265300526 I,-0.00006261763582323431-0.00010639231969803 I,-0.00001471986989006205+0.0001766101846003674 I,0.00004884196586967087 -0.0001488369534040168 I,0.00007122353198034681 +0.0001813290009889066 I,0.0001544605696743481 -0.00007610000490359152 I,-0.00005175960094155964+0.00006131936322854127 I,-0.00001694274986699445-0.00004857930369647794 I,-0.00007314756327837007+0.0001918900748275919 I,-6.514511226983392*10^-6-0.0002183760226250367 I,0.00003677973237827749 +0.00001650580404495887 I,0.0001063131359508663 -0.0001564604030287193 I,-0.00002694287699933473+0.000206207404133372 I,0.00001127781734682586 -0.00004771723327901789 I,-0.00008897819041218922+0.0001661723845136968 I,0.00004744965678539429 -0.0002304145148145531 I,-0.00004767434572199202+0.00008241563060873777 I,-0.00006818299639926066-0.0001317568694253742 I,-0.0000758713713528057+0.0001730282353898362 I,-0.00009648678733789322-0.00006849178621815211 I,0.00008397544892178418 +0.0001764261900159231 I,-0.0000849557506858603-0.0002020412907231455 I,0.00005890970491521678 +0.0001002060179923469 I,0.00008390076165926094 -0.0001370564476881483 I,0.0001066444031811153 +0.0001771898554580687 I,5.07850767532816*10^-6-0.0000482802873481651 I,-0.00008596719628398735+0.0001625399776480042 I,9.990488889925321*10^-6-0.0001091774070585321 I,0.00003375635556380552 +0.00006313087533962031 I,0.00006145861843692664 -0.0001321205333966805 I,-0.00001304976953005996+0.00008601187569747437 I,-0.00006455568418130897-0.0001648278241094163 I,-0.00002817193725423664+0.0001389941003791777 I,0.0000285123261359839 -0.0001343682084441701 I,0.0000243236394955197 +0.000159292485398482 I,-0.00003465443916078318-0.0001282294454778165 I,0.00003690951225906758 +0.0001136786818858055 I,0.00003072176936877312 -0.0001792823522290471 I,0.00002852759923612327 +0.0001013043458454503 I,0.00004971657385643137 -0.000118666022425724 I,-0.00002734570296462802+0.0002078918049398736 I,0.00005634999962924946 -0.0001364365293490184 I,-0.0000209796367180567+0.00009275045285507365 I,0.0000276276154568784 -0.0001897870303220683 I,-0.00006806457392169404+0.0001345769175791124 I,-0.00003818998180504857-0.00007732441930138315 I,-9.915833471870448*10^-6+0.0001839514547372004 I,0.00003482667918873328 -0.0001667982630981657 I,0.00001714440946617147 +0.0001018586827084403 I,0.00002405027527829826 -0.0001049277401726973 I,-0.00006668283319649233+0.0001076186963736774 I,-0.00003380077906785388-0.0001565531282571186 I,-0.00002565379099650972+0.0001496957898093107 I,-3.944134296716673*10^-6-0.0001569370316012605 I,0.00002242535696070717 +0.000116428135209037 I,-0.00005753685962489371-0.0001309032764753212 I,-0.00001929777105152886+0.0001553378348814466 I,0.00003026764979627439 -0.0001325931975444035 I,0.00003710299841769417 +0.000112324526303443 I,-0.00002828696719205229-0.0001943258529346121 I,-0.00003496458979634834+0.0001135481653170797 I,-0.0000538405149736994-0.0001350324548137327 I,0.00002422850553386377 +0.0002003519690057499 I,-0.00003386006077548924-0.0001864366185246435 I,0.00007396155497365279 +0.0001370893845853988 I,0.00003991686629246997 -0.0001367987608850985 I,0.00002998175814240568 +0.0001540536503175465 I,1.070938731148954*10^-6-0.000116185448089125 I,-0.00002983615048141964+0.0001706327306107455 I,0.00004621844432453715 -0.0001480017426010772 I,0.00002327405548689961 +0.000116447120734938 I,-6.365157356884532*10^-6-0.0001542319877733522 I,-0.00003473613354753251+0.000149944285337063 I,-2.110496225953747*10^-6-0.0001092716615020552 I,0.00001287630280728461 +0.0001740954390049149 I,0.0000449173306333993 -0.0001148168503175166 I,4.460345844877906*10^-6+0.0000987744865338826 I,-0.00002034755268895756-0.0001050434616210352 I,-0.00004219911514430874+0.0001274687383254267 I,0.00001636758704246969 -0.0001779409835868191 I,0.00004529058762110092 +0.00009983087878835546 I,0.00006046373977398036 -0.0001047483275814843 I,-0.0000366515026909818+0.0001803083815610379 I,5.199753096791674*10^-6-0.00009090104060208809 I,-0.00005139105885234593+0.00008711121080096945 I,0.00007838533839080217 -0.0001936248617726302 I,-0.0000515253218537367+0.00008051198498650221 I,-0.0000926104537326498-0.0001021372026463876 I,-0.00006150174168853017+0.0001589683673784152 I,-0.00002055634296835129-0.000169514876388485 I,0.00005864714420070192 +0.0001492231940723774 I,2.10167743665923*10^-6-0.0001291418413879295 I,-8.466972058542998*10^-6+0.00006279359291886778 I,-0.00001743259012070249-0.0001565253924963654 I,-0.00003018613696466921+0.0001588476425929239 I,-2.149115155297653*10^-6-0.0001522997801347688 I,0.00003760457502978927 +0.000137741551175424 I,-0.00003427842496436084-0.0001756616576179028 I,-0.00002865554686522895+0.000125578189992842 I,0.00002328616246860236 -0.0001424704838058277 I,0.00001466804374112802 +0.0001438896810992369 I,-0.00001464359923794667-0.0001476637986711008 I,-0.00002013910292440714+0.000119786934460869 I,-0.00006038613049030067-0.000196888361607807 I,9.80801436637148*10^-6+0.0001856251096450136 I,0.00002024390493966649 -0.0001926796869452311 I,0.00003615306231452274 +0.0001345753920625334 I,2.687611274627745*10^-6-0.0001002691620252021 I,-0.00002021275779362383+0.0002120622972386895 I,-0.00005661767781892424-0.0001313679970655296 I,-1.517965966799516*10^-6+0.0001131119442564947 I,0.00001407548840416802 -0.0001719066885881572 I,0.00005692859873109155 +0.0001272769535306548 I,-0.00001428386961076021-0.0001172560745208908 I,-6.018716655773008*10^-6+0.0001569704376408077 I,9.939866520452385*10^-7-0.000113784416785391 I,0.00002500043562042621 +0.0001610570533282256 I,0.00007904082246934942 -0.0001536522130640114 I,-7.041173112775849*10^-6+0.00008621294053097872 I,0.00002681497510831912 -0.00006642855524002866 I,-0.00008169812575803527+0.00009965539914703535 I,-0.00005506955930808142-0.0001822699509302955 I,-0.0000518150960171125+0.0001094747110734777 I,0.00001814680874840666 -0.0001878418164248387 I,0.00007150068988402708 +0.0001277239618467072 I,-0.00007740579218842768-0.0001787047333165328 I,-0.00006409090577319429+0.0001605514750434827 I,9.51111234819957*10^-6-0.0001277194464322196 I,0.00005702560789037998 +0.0001588295968634256 I,0.00002855056949861951 -0.000186238809010146 I,-6.415747325249504*10^-6+0.00008628479918317946 I,-0.00008846349643338642-0.0002145777691181212 I,-0.00002900373360898215+0.0002124122095139695 I,0.00004414055865639689 -0.0002203701333220547 I,0.00004744432886619781 +0.0001997494410491229 I,-8.998992912428878*10^-6-0.00007389016788074819 I,-0.00002948654053301235+0.0002201198103623811 I,-0.00007021017482338005-0.0001497899102609518 I,-4.143976165981201*10^-6+0.0000870876805861763 I,-7.007714656103276*10^-6-0.0001364031651480725 I,0.00005643418222537792 +0.0001752156898739076 I,-0.00001855704815650734-0.0001129536545313826 I,0.00002578988019505617 +0.0001361209348314027 I,-5.275350570482822*10^-7-0.0001715399415627779 I,-2.486202279483559*10^-6+0.0001324685901472157 I,-8.99057024974104*10^-6-0.0001077697654925262 I,0.00001368328996684026 +0.0001588250899312922 I,-0.0000354561398422771-0.0001549963959380704 I,0.00005124367612407058 +0.000127476509189097 I,0.00003530312825997106 -0.0001469702797191867 I,0.00002527441294618489 +0.0001288307998598282 I,0.00001634660578161657 -0.0001183172272200004 I,-0.00001862941241672562+0.0001848082408105263 I,0.00007177809788083858 -0.0001661118274985984 I,9.192456450417642*10^-6+0.00009659742089907575 I,-0.00001212243316859886-0.0001716878384443291 I,-0.00007249446486332577+0.0001706955719723775 I,-1.844169440350468*10^-6-0.00003442046624310826 I,0.00002870665590635462 +0.0001975434522396717 I,0.00008772218837201721 -0.0001841419538596818 I,-5.363600211685492*10^-6+0.00004266879398598153 I,0.00007087797146357951 -0.00004685310752961293 I,-0.0001504362487914025+0.00008305915805836343 I,-0.000101025255574501-0.0001631181299444839 I,-0.0000987417014965165+0.0001465845816373718 I,-0.00002810628571618781-0.0002145563992842078 I,0.0001118584935441642 +0.00006468784002202094 I,-0.0001414873518165742-0.0001954294654247358 I,-0.0000453582659525498+0.0001937144899167988 I,-1.039909845195302*10^-6-0.00007262409782060447 I,0.0001245663512104032 +0.0001596317252612334 I,0.00004369160761542615 -0.0002528098499758654 I,1.339117735306831*10^-6+0.00004007829677858114 I,-0.00007271420696745534-0.0002273941625813779 I,-0.00004982135405621149+0.0002486149220038574 I,0.00005319541091316524 -0.0002021015833802563 I,0.00009736611968788466 +0.0002014901355917538 I,-0.00004458971370085135-0.0001056120093683647 I,-0.0001115999868719048+0.0001624898139278342 I,-0.00004896174137520268-0.0002607880825299031 I,0.0001086594140470073 +0.0001370339121205406 I,0.00004815039535615707 -0.0002762720443593703 I,5.637804697407635*10^-6+0.0002611112757584799 I,-0.0001901537324903573-0.0001352315715163158 I,0.00002372534148922849 +0.0003378529153825319 I,0.0002467937978572347 -0.00009607549911544798 I,0.000149628693950655 +0.0001574545851705131 I,-0.00001575766909111054-0.0001717307401214693 I,-0.0001561877901417137+0.00007797879508317755 I,0.0001070054229059203 -0.0002039361419178534 I,-0.0001033179055066617+0.0003737064840926149 I,-0.00004124788397540886-0.00002647382388498147 I,-1.741466551756859*10^-7+0.0002109636253358195 I,-0.00007236639574995934-0.0001739593596870798 I,-0.00005495755879571798+0.00003300528939716798 I,0.00002508805887434053 +7.67774070812646*10^-7 I,0.00005798002007383787 +0.0003029426600647879 I,-0.00009277398646362429-0.0002014330344206584 I,-0.00003919252638221259+0.0001225190796924467 I,0.00003148866628737303 -0.00006625591529571992 I,-7.140350165551534*10^-6+0.0002627075890379685 I,-0.00001878446637007918+0.0001594884906710621 I,0.0001034659826027713 -8.799592982962238*10^-7 I,-0.0001795171416661058-0.00009489181680231963 I,-0.000114832826382162+0.00005255169869794303 I,0.0001161182940524479 -0.0002062015525752562 I,0.0002504998195309538 +0.00008400124830451204 I,0.0001150709796977903 +0.0000304685969904887 I,-0.0001158323564124538+0.000162907315468312 I,-0.0001216338677246075-0.00005135720521839483 I,-0.0001154011995932019-2.413826692975433*10^-6 I,0.0001495382311991769 -0.0001511466325738767 I,0.00003429274849582784 +0.00003833262160808654 I,-0.00004558108521552196-0.000181074359900727 I,-0.00007041137409379433+0.00006216539152449245 I,-0.0001746988154175218-0.0002367535187918532 I,-9.653612832739291*10^-6+0.0002421684778961525 I,0.00005126232520758664 -0.000289169579554193 I,0.00004463344179896447 +0.0002334628559043201 I,0.00002565507677010383 +5.980570729463616*10^-6 I,0.00004503422267544662 +0.000310280060276128 I,-0.0001240872908350107-0.00005214635909941443 I,-0.0001132286752732816+1.585123064668466*10^-7 I,-0.00006572446237613006+2.547359850381572*10^-6 I,0.0001153350911285774 +0.0001470868891821569 I,0.00007262911950793439 -0.00008785517733567374 I,0.00005172487070307427 -6.53301848881868*10^-6 I,-0.0001118825777419292-0.0002317447075457453 I,-0.00009105031238754766+0.0001119680532825299 I,-0.00006955409216136299-5.21025437884034*10^-6 I,0.0001028334395514869 +0.0002026187357373281 I,-0.0001403117227897442-0.0001906677696772729 I,0.0001342132194057672 +0.00006477328559974424 I,0.0001298451709563178 -0.0001412656569367171 I,0.0001467354874526772 +0.0001370853065784137 I,0.00003000942699972134 +2.711358604004216*10^-6 I,-0.0001224082323218013+0.0001966816163388678 I,0.00007089844193024308 -0.00009339338503790075 I,0.00004113383893874534 -3.837870744567229*10^-6 I,0.00007824167501150908 -0.0001566766338365093 I,-0.00008647490526891026+0.00006840245253668654 I,-0.0001039451227689336-0.00005256224128779154 I,-0.00003053202477739249+0.0001850045015627738 I,0.00009048166299187004 -0.0001425105391884456 I,0.00005531529032342785 +0.00005921153628551554 I,8.77171169563001*10^-6-0.00002930784272714545 I,-0.0001049476949053588+6.674652386817893*10^-6 I,-0.0000563765141690538-0.0001963740355970721 I,-0.00004936390892126862+0.0001061619681608232 I,0.00002442183469729616 -0.0001684531417666965 I,0.0000896900293230372 +0.0001162513835456058 I,-0.0001193716492631099-0.000198369383144444 I,-0.00009080005931339729+0.0001298020450494972 I,0.00001871891406241954 -0.0001119258769633301 I,0.00008291217028606418 +0.0001655548688154722 I,0.00006435512950813232 -0.0001848466934177962 I,2.844461555111248*10^-6+0.00005712869892540401 I,-0.0001540057546966288-0.0002458457217436889 I,-0.00006027522616252115+0.0002228889335323458 I,0.00007773381756845148 -0.0002842722418746682 I,0.00002791509001472926 +0.0002679277271520549 I,5.679639194549054*10^-6+0.00001538948885603286 I,0.00001726550910751929 +0.0003102253333715217 I,-0.00013654602347013-0.00007578313441905404 I,-0.00009225217802742281-9.614009690926592*10^-6 I,-0.00005508387104395999-0.00001213638563974465 I,0.0001202381514524561 +0.0001654957575289487 I,0.00005040979621211909 -0.00009506160177789416 I,0.00003854854395867344 +0.0000132744334887968 I,-0.0000824408440942079+0.0001359010467574947 I,0.0001184181694606602 +0.0001588518632330132 I}",
                                      1.75),

  getMandelflakeSpec: function() {
    return {coeffs: RotorAnimations.MandelflakeCoeffs, freqfn: function(n) {return bount(n+1);}};
  },

  RingNoRingCoeffs: parseRotorCoeffs("{-1.006780986458608-1.862583056370508 I,0.248697469011681 -0.02826440919586293 I,-0.3531246336938139-0.3321349642778486 I,-0.1828794967514688-0.262291950125772 I,-0.005734991318812721+0.07071158102917159 I,-0.09937878284864737+0.05611199261053353 I,0.01107140590322386 -0.09915503778643599 I,0.01044310249298683 +0.03968017709253217 I,0.068649480609332 +0.03253114141664123 I,0.08763715190226488 -0.00562640980282296 I,-0.005909525909923372+0.009220993381337132 I,0.04395587071586252 +0.01086440569217704 I,-0.04768388830592255-0.05285720319541753 I,0.04056107862207196 -0.02891974073568606 I,-0.0954228376274939-0.04780535685798449 I,-0.09271376802702393-0.0496499520618246 I,0.0398729661737845 +0.04054157255767937 I,-0.05771584114531111+0.0324569120094127 I,-0.04892822814640926-0.007858934098467205 I,0.04498975677435392 +0.0386781244535175 I,-0.03841624075336945-0.02657787439101563 I,-0.0388123154710041-0.04601589915332698 I,0.02471910664438113 -0.01748149210419191 I,-0.01455730267016137-0.03368388222635627 I,0.001690900649746849 +0.006211247571557329 I,-0.006414945501679835-0.006282830016512381 I,0.009940488237116074 +0.02204327372664761 I,0.0001017419977382972 +0.02783751405888676 I,0.004798640687474331 -0.01058086311258722 I,0.01190194978813247 +0.01457879469696983 I,0.003146614567540479 +0.009564440071918492 I,0.01251839245744254 -0.01328879844315739 I,-0.004034251986269815+0.02173657624283324 I,-0.01052872452276893+0.01684285410768062 I,-0.004981047688706958-0.01312750164481828 I,-0.01424161410335701+0.002772293486172297 I,0.001260842640458247 -0.01575803227814625 I,0.006970820062244375 -0.0008104738935410423 I,0.0083572777339498 -0.02154413691873248 I,0.01190826069677977 -0.006025180716082374 I,0.002343396241536577 +0.01019050039326839 I,0.001658332342107338 -0.01711610308202245 I,-0.006802729418647031+0.0066035105835553 I,-0.007399994825878609-0.001105020062550383 I,-0.005062212418512251-0.01031180750551663 I,-0.005000756767032814+0.004010679346032994 I,-0.003372969828962791+0.003924348092862774 I,-0.0009880857488513486+0.008910775187121132 I,0.007447846556824421 +0.00780854493795579 I,-0.004301887514880986+0.009175630413494382 I,0.009323992111422777 -0.001104062704854379 I,0.01109628862088674 -0.01106807804499179 I,-0.01352594375118933-0.002825180072122969 I,-0.0006382935019947304-0.004810788567165459 I,0.01106045629998253 -0.0140065757875106 I,-0.01059688859227177+0.003804350490838094 I,0.0156333782979103 +0.0002334119213875281 I,0.019000579173033 -0.002568424126227745 I,0.00002314949552344181 +0.007772370346612316 I,0.01175359760354256 +0.001490205916564536 I,-0.002634737416887453+0.0002143789907605928 I,0.001458248919142641 +0.0005966928618327053 I,-0.006410041276132279+0.004383515652952952 I,-0.007697850468864752-0.002273691147326051 I,-0.005148285972735316-0.001860697171469875 I,0.00117137491649188 +0.008083008633503364 I,0.005508616239913344 -0.01076158095517829 I,-0.0008388392846586691-0.0008630473489952628 I,0.01459977089152207 +0.008052938273610491 I,0.0156562838851472 +0.001773919029359475 I,-0.00241478753462677+0.001783671815680101 I,0.007339431914260891 +0.002077824223939867 I,-0.00771376519153587-0.00877547843540174 I,0.006664071878013904 -0.004116659537357448 I,-0.01304996948242993-0.00713067962568838 I,-0.01232989588787461-0.007363715896601764 I,0.002389942751141766 +0.0061395003900993 I,-0.005096861532009429+0.003839198995002623 I,-0.003881789414879077-0.007379670065706299 I,0.00117009934166757 +0.004696521699822391 I,0.00825668549150496 -0.001782401973571001 I,0.01180482500207545 -0.007049590432023654 I,0.001215319742828517 +0.002487578341731854 I,0.00911799351729308 -0.003346975069102148 I,-0.01102566318638307-0.006634745590214705 I,0.006954959076907292 -0.00935942982984727 I,-0.02205995244217575-0.001868000909192106 I,-0.02552851443721518+0.0003807394888377435 I,0.01491329557637275 +0.008208867551623688 I,-0.00865817052505651+0.0113020400402221 I,-0.01816503384856399+0.006983873893398599 I,0.01623490412941451 +0.00441343299793981 I,-0.01809034038178543-0.007658023979562786 I,-0.02187788030225289-0.008711580755510498 I,0.005153827731775954 -0.007399775508713163 I,-0.009209209534469739-0.007360507622692592 I,0.005503320282398357 +0.005720101391956193 I,-0.004287203135264067+0.002461650913940736 I,0.01074578399675613 +0.00879504991846766 I,0.003820560876945881 +0.009427408654385305 I,-0.004834728450000953-0.003686305677046819 I,0.01057895079875749 +0.006848867432626888 I,-0.0008833409057998924-0.003533074419402854 I,0.005156270880082256 -0.008373660698261877 I,0.008450157633848189 +0.005334359854284563 I,-0.003336731530268722+0.002432082431875482 I,-0.005975594446588509-0.007162575915654689 I,-0.0112521830965475+0.002180832934456418 I,-0.009593541117246742-0.002851384355231334 I,-0.002629017769567084-0.005748195072737593 I,-0.002616021866862988-0.006963243406738592 I,0.009881252966903552 -0.001381891549384703 I,0.005306463516246866 -0.003126471944026967 I,-0.003382790120926509-0.009169250737932979 I,0.006890230577431 +0.007414272151568616 I,-0.006959579455967954-0.001192002682083904 I,-0.00005440296148416166+0.003239272365511739 I,0.007516816920879207 +0.007673846598313057 I,-0.00343240891241292+0.004464177152532623 I,0.005947591878919066 +0.007318598123388045 I,-0.002971846458845538-0.002630181530176817 I,0.003021139466479361 -0.001364611872932703 I,-0.01327032121551204-0.006296173002739063 I,-0.0116829032110489-0.01039255579089259 I,0.006170533806081181 +0.003147988760919762 I,-0.01364175254510709+0.005060949018139022 I,-0.003629047045861919+0.001077512051271628 I,0.008046419921212628 +0.00816854517057429 I,-0.0119258308835895-0.0005973223988491125 I,-0.008931812051932955-0.004415314718058751 I,0.005714543632500569 -0.004188729584407221 I,-0.003468473995944207-0.005650464566945965 I,0.005021569217379523 -0.002326913252853052 I,0.001615391987792023 +0.001452601561811957 I,0.006049608869109856 -0.002559528349951668 I,0.002202506291511312 -0.0007073122884306226 I,-0.0009989841929985161+0.003419144721140011 I,-0.0006563945170249587-0.002181308235905803 I,-0.00361520072602497-0.00169365179217487 I,0.001211018124414911 +0.003393212180467827 I,-0.001823596269346232-0.004777865251099851 I,-0.003246844648071045-0.006359604088418178 I,0.003251733525089085 -0.001423724607032866 I,-0.002111604590576761-0.00536469072869324 I,0.000362514853580893 +0.0008853350901658463 I,-0.0008538376974095247-0.00118929548380814 I,0.003532062871628498 +0.002345909978326091 I,0.002938767624161272 +0.00495123637065885 I,0.001165841525178958 -0.0001178966128933087 I,0.002491211286838485 -0.002511410226170571 I,-0.001507878632361502+0.003032239137013998 I,-0.0006370961068378131-0.0006736033563363303 I,-0.002243473117592529-0.002606411592629465 I,-0.002928515544091367-0.001476886907734268 I,0.0008762134488938126 +0.00002889132906326933 I,-0.001239266383229617-0.001824971056252252 I,0.001346381237793192 +0.002936434430016207 I,-0.002250467689561518+0.0003708779430701505 I,0.003852020112496491 +0.004637046610582986 I,0.003580332800870782 +0.005689499317143767 I,-0.0007889353226208243-0.003161274704906395 I,0.004734056968616495 -0.0002422033834328487 I,0.002570282118866852 +0.002648567243207889 I,-0.001402346396130209-0.004413518760472994 I,0.001103614673904771 +0.00365929775069868 I,0.0005195847182528678 +0.005458380413397829 I,-0.002552759323681229-0.00007876565774584451 I,-0.0007387141669117393+0.003179556788034709 I,0.0004581807202529101 -0.001282327456887018 I,0.0006591860818142602 +0.00075284969239643 I,0.0008423530252808734 -0.003106876629199888 I,0.00200753436853397 -0.001847747324232867 I,-0.0004818596831829107+0.001178876502096842 I,0.0006503824604081531 -0.003076698829796292 I,-0.0004099784815896649+0.0006006132042074672 I,-0.002502596289927303-0.0007433590033013734 I,0.0005916065863900977 -0.002020345969401891 I,0.0008525807047980766 +0.00128153110020828 I,-0.001562554085523514+0.001916235870621381 I,0.001151814799910869 +0.003127642092020384 I,0.0001764614466689043 +0.002235528060119888 I,-0.001057927273909407+0.001304291900048365 I,-0.001296120730086843+0.0001219260470770715 I,-0.0003929912550314427-0.003363128788626679 I,-0.001375736769386746-0.0006882757059631925 I,-0.002437678541284011+0.0004143878411623013 I,0.001378514922681456 -0.00239399722409613 I,-0.0006947142640679194+0.002232048260494119 I,0.0009353813985323745 -0.001244830145559052 I,0.001896782170392048 -0.002752675646684142 I,0.001259200464722606 +0.001432318600703536 I,0.002212795400983585 -0.001398573454485143 I,-0.0004503184593968113+0.002333438117321926 I,-0.001434311395211747-0.0002850562524929537 I,-0.001471209862630143+0.005392645960610964 I,-0.002491550967466171+0.003491669158959908 I,-0.0000896179496959566-0.003456324980499255 I,0.001481472569691682 +0.005134302580586294 I,0.002594867654025112 +0.0003322835925017183 I,0.002300913537726165 -0.002823813536292387 I,0.0008699837176061216 +0.006046205642133769 I,0.0001420163687943198 +0.004316014175987148 I,-0.001674489727216507-0.001770824757907365 I,-0.001227610827426044+0.001877131656059894 I,-0.0006648159137497495-0.00272730447085105 I,0.001108568080540632 -0.0002247682425392601 I,-0.001447343143595277-0.003026167196121664 I,-0.000931988702864605-0.002267856725084245 I,0.0006195639700240976 +0.0003931167385917266 I,-0.001799742426573476+0.00129979973447911 I,0.0003899838504450451 -0.0001158323782405849 I,0.001891858881436222 +0.0001748159851675105 I,-0.001027658190716612+0.002558125828845321 I,-0.001387669111306988+0.001620589524092542 I,-0.0002253096230929261-0.002040793137249093 I,-0.002295636132507159+0.0003152581514895793 I,0.0001766751234289314 -0.002607406692605078 I,0.0009584201941815317 -0.0000382479639600042 I,0.001099444718241626 -0.004093800942817433 I,0.002538267288761277 -0.002487336561485421 I,0.0001694086315353017 +0.001022730267297025 I,-0.001766434572890016-0.00277849890599102 I,0.001116052272295207 +0.0003463758371429816 I,-0.00132912570675042+0.001289434228833381 I,-0.0007854807882833456+0.0005725576081825759 I,0.00120418959070221 +0.001092121958178316 I,0.0005718948343112376 +0.0003278412666399698 I,0.000865937280304797 +0.0001572736761856701 I,0.001513695954445053 -0.00250785204172755 I,0.002225878386344659 +0.0007510495140959814 I,0.001202764350913679 -0.003856212196974002 I,0.0004019513658757201 -0.002616258011915704 I,-0.000619003122257533+0.003520151143465385 I,-0.0002767008109597537-0.003238253746546001 I,-0.002721105899271645-0.001205434683544019 I,-0.001826101759749789+0.001640680953218995 I,0.0004545998234425632 -0.004667856773135026 I,0.000718477562916162 -0.002617110395620429 I,0.0004074907913290331 +0.001850365400983805 I,0.001710688333509258 +0.0002232019956127525 I,0.0001470365919514043 +0.002294418961212776 I,-0.000746952460401793+0.0006835090255276734 I,-0.0009144876903197183+0.001722341744894816 I,-0.001596198726469845-0.001242398071452287 I,-0.001112480210219658-0.0002693785812167427 I,-0.001293729988294697+0.001698005753574834 I,0.0000790460661222736 -0.002690288614621272 I,0.0007755194693767338 +0.001874291760122028 I,0.001153033696905574 -0.0008460584370819809 I,0.000913718779993179 -0.003280935112388646 I,0.001464178409518553 -0.00006208825109578356 I,0.0006264418381603854 -0.002493229497189642 I,-0.001002381905668051+0.0003704391490719516 I,-0.000317511952188028-0.001100191696349558 I,-0.0007126574734281125+0.002519287562329452 I,-0.001291193046509508+0.002651621214190283 I,0.0007239422498495646 -0.001029127196224088 I,0.0008287446854812518 +0.001997970164699582 I,0.0002615798814721965 +0.0004526088892252906 I,0.001380526533220033 -0.0007978071030946745 I,0.00009204986485217628 +0.001940735032983802 I,-0.0002986363870047205+0.000479145789150928 I,0.0001059451753350291 -0.001201447946532286 I,-0.0008400887521228849-0.0009803301864941788 I,-0.0005847122658175644-0.001823760933618404 I,0.0009610747772028708 -0.001002185353236745 I,0.00009272288054691701 -0.001196049425398271 I,0.00001780968985208705 +0.001239562846313519 I,0.00130426151865458 +0.001184774918689308 I,0.001053814167529038 -0.001096151520722204 I,-0.001887989918124616+0.001648337409191112 I,0.00005488091629145915 -0.0008428360095788442 I,-0.0008408271850232534-0.001310083242155239 I,-0.001477426238884078+0.00007205511848503267 I,-0.0005310942353112034-0.0001733508208456196 I,-0.0009298582434498106+0.0003674827362818642 I,-0.0004082159127431775+0.000811942340461643 I,-0.000684856788841262-0.0003526473513503238 I,-0.0000672616172077904+0.0002956535145161325 I,0.000783856334233161 +0.0008123529055784513 I,0.0009474459620827627 -0.0005421994310316315 I,0.0002015892553054854 -0.0008484124437642035 I,0.00002949487769300391 +0.002135445692405498 I,-0.0004056321512872479-0.0008485104422849869 I,-0.00168283991203745-0.0001299824820887117 I,-0.001354585387529558+0.001589683767829541 I,-0.0006573556700930345+0.0003791459939177274 I,-0.0002278380184225068+0.001532976962053542 I,0.001154315346716921 +0.001360471003216898 I,-0.0008541264828494646+0.001182479934444382 I,0.0006914119870910051 +0.0006410703055482127 I,0.0005705288717218803 -0.0006325475573983525 I,-0.0009867146268537662-0.001084244768367354 I,-0.00004546213513002394+0.0008350565619784917 I,0.001268538329440996 -0.0002703542450341811 I,0.0002617493958157551 -0.0006762113351676492 I,-0.00009711417297936622+0.001616634246765863 I,-0.000580055956832061+0.001957082956236162 I,-0.0008453847359305635-0.0004681854709879081 I,-0.0009697353300750443+0.001743040831825225 I,0.0008964060890258043 -0.0001360030160084231 I,-0.0002450125382828989+0.001353654479698983 I,0.001156589573863327 -0.001849488192702208 I,0.001936113243763356 -0.003407144338733369 I,-0.002106983792546092-0.000154924944231941 I,-0.001732670908299856-0.0007569605632186224 I,0.00215208278091257 -0.00309092192693609 I,-0.001139996241539128+0.001987757811934554 I,0.002480899803546876 +0.0002018942826536672 I,0.003255655678738634 -0.001879456321349043 I,0.0009052168233579009 +0.0005369534541293092 I,0.00133594741172085 -0.00143026397738441 I,-0.001664890675620161-0.001000714980110581 I,0.0004684912292558896 -0.001552409552010554 I,-0.002240033185822892+0.0002173839694331221 I,-0.001333433579759874+0.0006837793185925339 I,0.001997966036306872 -0.0004257665320536034 I,-0.001692366697193267+0.000379799506035275 I,0.0008017094051830783 +0.001941945171426036 I,0.0006566016222987012 +0.0007863188319541414 I,-0.002746007669563453+0.001776363591035156 I,-0.0005727687761117651+0.001156919945742985 I,0.001251224124950786 -0.0003435027668916727 I,0.0005949693926073523 -0.001025578628154989 I,0.002301683595298417 -0.001997085653911759 I,0.00181107115272047 +0.0007689944084919909 I,0.001997617937197957 -0.001305826728111444 I,-0.0004147356280653942+0.0005622988476239953 I,-0.0003234499787205412+0.00219821005331817 I,0.001689705206479934 -0.000603649655094389 I,-0.002682137488187484+0.0002664123564627744 I,0.0005138632064348317 -0.0007143618682347863 I,-0.0009700076873209245-0.001522011141608726 I,-0.002607663193323287+0.0002288881867729513 I,-0.0003703545194052868-0.0007230517293306432 I,-0.001896389102436663+0.0005989841135304555 I,0.001552646375478007 +0.0006141309778354956 I,-0.0005198573974083348+0.00149193554875398 I,0.00357264771600258 -0.001298288183363737 I,0.003999445325880923 -0.001454065194383283 I,-0.002464636673085954+0.0004583664948962208 I,0.0008597351112567759 -0.003250698293644007 I,0.0009915752875457708 -0.00194731957360304 I,-0.003161268952835381+0.0007247357137119835 I,0.002753827052558145 -0.002157663376786025 I,0.003181236767799496 -0.001438874650093541 I,0.0001280663561250991 +0.002080742595296314 I,0.002240535646749732 +0.00001721004243379295 I,-0.0003577206512637254+0.001303678267510476 I,-0.000411216756662015+0.0001404071983729847 I,-0.0005603973965244958+0.001860067422327402 I,-0.0005391401113804721+0.0005501315836521777 I,-0.0008334161376573801-0.0006925199546116926 I,0.0007250250346597571 +0.0004582873338238481 I,0.000947998796689341 -0.001014768649993178 I,-0.001006431551385463-0.0002449719536321117 I,0.001903752243998648 +0.0003146602473248633 I,0.002429055410570256 +0.00004835022182683095 I,-0.0001964902589494323+0.001122761065375352 I,0.001891059058860635 +0.0003052546511937131 I,-0.0004669784555954681+2.357693816448878*10^-6 I,0.0004287664173173379 -0.00009344423260384768 I,-0.001409809499726473+0.0008073227359672815 I,-0.002092619429678277+0.0001432153575591053 I,-0.0001874005338869814+0.000238267967031612 I,0.0004765924232105617 +0.001428804067898987 I,-0.0006681564341844955-0.0009805260391098699 I,0.000451651984653946 -0.0003058928451110318 I,0.001211652617635049 -0.0001769800057830587 I,0.0006433210762010818 -0.0005180586900359891 I,-0.0003542290619104314+0.0001794896816050698 I,0.0004616173644188103 +0.0001920742860306399 I,-0.001062003697077669+0.0004600410755506241 I,-0.0002930592011811946-0.0004727940836787229 I,-0.001695715827946587+0.001015751756503163 I,-0.00170730666541678+0.0001875573499897265 I,0.0003868084813480029 -0.0006842084241227362 I,-0.0004932475582615477+0.00207845827413139 I,0.0002046089872625181 -0.0003731644208730181 I,0.001484021302789443 -0.00008279411742857399 I,0.00003429418649559396 +0.001323314851970366 I,-0.0005319964288628058-0.00009020903443719155 I,0.00009752933438763808 -0.001091852821778388 I,-0.001047303799381633-0.0007005441429526614 I,-0.0007291736001162014-0.0007994443771302774 I,-5.858116476618359*10^-7-0.0008010035197110283 I,-0.00004262441097987015-0.0002862409375689545 I,0.0007844927988413573 +0.0009664202484484235 I,0.0009445579723795228 -0.000521859983081958 I,0.0001455786215371216 -0.0005729947013995722 I,0.0008673071474383518 +0.001738523859174352 I,-0.0003512126066176555-0.0007790430954587497 I,-0.0009645932670320821+0.001560055194579304 I,0.00004110295660131498 +0.002327954213313755 I,-0.0004353383733806361+0.00002346010670950657 I,0.0002228509071360786 +0.001079213403994861 I,0.001067907135712702 -0.001272356027661168 I,0.001098041227770866 +0.0006979595985880124 I,0.0004994206450529013 -0.002125441183564283 I,-0.0000474197843426602-0.001563208403468246 I,-0.0002664577940067157+0.001618143070211468 I,-0.0002716678178651217-0.001138493849367112 I,-0.001407255402236209-0.0001677672574866157 I,-0.0003068492059710177+0.0005505308193931095 I,-0.0009116040987060751-0.001860341264213633 I,-0.001218931427636768-0.0005338123966088049 I,0.0001937594488884225 +0.0003429915630588835 I,-0.0002561432832375366+0.0003498530843674062 I,0.001590505182511064 +0.001489907441921187 I,-0.000850320089584113+0.001593577226641906 I,0.002530397809221317 +0.000808047771242213 I,0.002226176091323541 -0.0006091257616459332 I,-0.002712008739197568-0.0007149238572189693 I,0.001125377169653114 -0.0006705794172838506 I,0.001904440877932364 -0.002496052753846127 I,-0.002002152667471937-0.00005960739428393248 I,0.003448018437407971 +0.0002277444816982236 I,0.003592962478739518 -0.000166085245932792 I,-0.000354716761534864+0.001387485124672908 I,0.002039731417823949 +0.0004945446893735491 I,-0.0007686920941120433+0.00004969436892481808 I,0.0002203559481645365 -0.0000236339095139892 I,-0.001359045805960782+0.0007817275154693845 I,-0.001430949545138111-0.0003310247046704258 I,-0.0007928943936245424-0.0003606553696439246 I,0.0002447005047416118 +0.001437547099978383 I,0.0009187349146561651 -0.001772902133273621 I,-0.0001267024183892289-0.0002190003538561829 I,0.00248801925899005 +0.00138524400724292 I,0.002630284889596265 +0.0003338953218550222 I,-0.0004317095623390418+0.0002981403566626277 I,0.001223491113355247 +0.0003531052052096776 I,-0.001279344019492728-0.001461844981788154 I,0.001110393292430051 -0.0006751195877674886 I,-0.002102455325749808-0.001163976261543185 I,-0.001985087518344043-0.001151596940544701 I,0.0003643841580925253 +0.001024021373995644 I,-0.0007296894969919983+0.0005671581981184267 I,-0.0006542973367771856-0.001171943181784775 I,0.0001242245200418633 +0.0006987876709502923 I,0.001326342930239755 -0.0003241510694015972 I,0.001860762781726167 -0.001046456837189214 I,0.00016218374905479 +0.0004417045665624847 I,0.001457385422735239 -0.0004344195183671839 I,-0.001556597799395584-0.0009694625478975532 I,0.001063167109057731 -0.001262110578949159 I,-0.003150329777164075-0.0002852088970842904 I,-0.003743061670577816-0.00008670387394321016 I,0.001897683479729673 +0.001269435123564438 I,-0.001129634853404095+0.00169735003005719 I,-0.002616407879684388+0.0005572413428956797 I,0.002249097751670235 +0.0006517793916845748 I,-0.002020182119877007-0.00117077827934469 I,-0.002718598305812027-0.001483944680079591 I,0.0006506999277480635 -0.001018930407298637 I,-0.001269674405948106-0.001066871664541154 I,0.0002837184353881236 +0.000783678140044734 I,-0.0006914786013724906+0.00003063128584408293 I,0.0009945671051589295 +0.001231524025133382 I,0.0005326452185930885 +0.001266630706313776 I,-0.0003080714550763866-0.0008118026749266811 I,0.0009208954992070115 +0.0008625303146843501 I,0.0005084786876062538 -0.00008300395991234233 I,0.0005563461045106899 -0.0008967932940808204 I,0.0006860871894295137 +0.001411448053226772 I,-0.00002299716127847503+0.0008638406486104644 I,-0.000505961291352501-0.0008193739471999405 I,-0.0009053386140386998+0.0001767377702821532 I,-0.0006470827588691379-0.001209899000118211 I,0.0004777376283084608 -0.0005406310143720127 I,-0.0001463792136823408-0.001617794978412412 I,0.0006497044585288972 -0.0003773985623947397 I,0.0007346145105672562 +0.0006375498297801988 I,-0.0002306970386150504-0.001355293430310052 I,-0.000373249291738051+0.001051206579358078 I,-0.0005644339309107611-0.00008061239311640692 I,-0.0008834217365540994-0.0004738547298509356 I,-0.0004496563545473107+0.0008673956800170648 I,-0.0003081429336165746+0.000327183848071758 I,0.00002305738028368924 +0.001051149271749989 I,0.001017750972122548 +0.0001954993289875896 I,0.00008196041043944096 +0.001091796209052815 I,0.000968361442267472 -0.001093570875105065 I,0.000957409942696867 -0.002082579012828267 I,-0.001516000223798274+0.0005914965641900924 I,-0.0006035838626039791-0.001044684247673403 I,0.0004673054467713358 -0.002075253168703581 I,-0.00129759657871524+0.001329683058782091 I,0.001640209332854798 -0.001235901888099387 I,0.00218960342642207 -0.001626966183961191 I,0.0006359869976546274 +0.00115093193977812 I,0.001750484575889088 -0.0006119909139258062 I,-0.00009944650493151565+0.0003746972246794601 I,0.0001685513854217033 +0.000137880368764331 I,-0.0003583204335978248+0.001152561936245714 I,-0.001050040451876202+0.0001412709043687951 I,-0.001045223676196443+0.00001457086475298991 I,0.0006435587938950756 +0.0009938656754554776 I,0.0002983537353663064 -0.002135863427131 I,-0.000339932088222914+0.0001510355909387257 I,0.002855906118828865 +0.000228373742426187 I,0.002848548476873036 -0.00102233311237937 I,1.502942344527511*10^-6+0.0005428689067251434 I,0.001573275314026771 -0.0005336510605035687 I,-0.001733366475203294-0.001065821026752464 I,0.001014736826486673 -0.001185911409547056 I,-0.002470764701717868-0.0002073017578777283 I,-0.00265196524714771-0.00004302223618055415 I,0.0008471957357499354 +0.001137504976251289 I,-0.0002711162576471754+0.0008027632491351056 I,-0.001529459338620728-0.000978388444632388 I,0.0003558111500778839 +0.0007377158216608616 I,0.001456442362013365 -0.001324206880749035 I,0.001802819930461379 -0.002436118639903021 I,0.0005293009438695332 +0.0005827249885027968 I,0.001788212329084302 -0.00155588463037886 I,-0.002546377420891928-0.0003333855173082387 I,0.0007344233378270294 -0.002400803397817774 I,-0.00440084237380456+0.001732168454204934 I,-0.005394852741703708+0.002731195797291353 I,0.003725635177275769 +0.000686884913926976 I,-0.0001294671510556982+0.003074820068598342 I,-0.003853114139723102+0.003180118536430862 I,0.003811071708517911 -0.0009507487146115349 I,-0.00447184327904109-0.00055301733538617 I,-0.005924322697127042+0.0001823242681974808 I,0.00009589969821697901 -0.001984207885213535 I,-0.002882217592694541-0.0002786215564370532 I,0.001835435703897012 +0.00135323386606428 I,-0.001076307788039967+0.001296666449428383 I,0.003445250005953319 +0.00114486552679505 I,0.001994929295851242 +0.001307273067084333 I,-0.002086633017172404-0.0005205936907517567 I,0.003119637562017777 +0.0006857031570843154 I,-0.0003710614367051146-0.001684439653924398 I,0.0002199490129461944 -0.002254913633766066 I,0.003539909478337835 +0.00005885627181487338 I,0.0001193296206205394 +0.00007799386404024099 I,-0.002145404260242474-0.0009636814184901516 I,-0.002329467643454466+0.001387393678192928 I,-0.003501679357686924+0.0006602627536200009 I,-0.001617805830472215-0.001683196817517537 I,-0.002246756499899778-0.0006669438804311184 I,0.001928125064789649 -0.0008961612467762046 I,0.001432143896625142 -0.002051252511043099 I,-0.002128625580997385-0.001312197628584257 I,0.003206125336406823 +0.001365839218231148 I,-0.001628476196105577+0.0004067403815664843 I,0.000408059792478196 +0.001808447207975241 I,0.003264038098218504 +0.001421495292091303 I,-0.0003033151500219974+0.001350521979313506 I,0.00257212128841096 +0.0009406612872958375 I,-0.001517824613935905-0.001215573912843571 I,0.001213809612446935 -0.001297739200972902 I,-0.005065294231174236-0.00063418409244605 I,-0.005262382329062238-0.001118502964592548 I,0.003039090929827707 +0.000708601520840631 I,-0.003285669763239307+0.003077421041269069 I,-0.001943552497259393+0.00162717229485017 I,0.003854255760909632 +0.001243121252219642 I,-0.004523401487955306+0.0008065595299586139 I,-0.00436491131088925-0.000271789586086677 I,0.001163429507191895 -0.002286042005817336 I,-0.00239600916040967-0.001339097966953084 I,0.001290966928450735 -0.001070276452986136 I,0.000421806915829977 +0.0001608469899599295 I,0.001840127311960261 -0.001531661235817724 I,0.001127517311861098 -0.0001519940528909229 I,0.000486203862084406 +0.001086684476305302 I,-0.000430526583857987-0.001164596706272484 I,-0.001461178920161621+0.0008193814328910518 I,0.000601297307929755 +0.0007689895289480147 I,-0.002054674690202023-0.001635683108046362 I,-0.002519124620595047-0.001309179817542654 I,0.000841680751833474 -0.0006715017154635156 I,-0.001389392609028513-0.00118992730563537 I,0.000888502309977865 +0.0009197734587284894 I,-0.0007949959575195113+0.0002276466627285352 I,0.002073523312675489 +0.0009347002436990796 I,0.001891747559450985 +0.00151274003934942 I,-0.00008814569842436346-0.0005883314144773689 I,0.0009716223214554485 -0.0008827127196403861 I,0.0001415781168398768 +0.001199090665380831 I,-0.0002884064356610411-0.000626777050338692 I,-0.001062557275552016-0.000257231134897058 I,-0.001505511695047642+0.0005135820880815154 I,-0.00005471692371096992-0.0001464984062997183 I,-0.0009182361798470923+0.0001652406910217459 I,0.001259264454796532 +0.001196198580564582 I,-0.001026463188441937+0.001018551559148648 I,0.002499283915631399 +0.001139551573491777 I,0.002482150743697246 +0.000880742501642411 I,-0.001665197449839241-0.001329198978248155 I,0.001723391397982621 -0.000345904796892569 I,0.001899208885279024 -0.0006182997059871294 I,-0.001125181941482904-0.001356369239738933 I,0.00217078484154659 +0.001558679177566404 I,0.001664469484909753 +0.001565785595448891 I,-0.001095764564143346+0.0001116230572530462 I,-0.000108082889995546+0.001115896406730909 I,-0.0009375203944029162-0.0007292911295639338 I,0.0001208983217536644 -0.000396708866545632 I,-0.0006694027062829141-0.001304304456657238 I,0.0009559911901797673 -0.0008197889027282051 I,0.0004126926544059414 -0.000142800752647762 I,-0.0005614105866806952-0.00147361218811341 I,0.001093719784020553 +0.0008051498610590016 I,-0.001471050188754157-0.00005246986067666305 I,0.0000620288305726044 +0.0003472814967577662 I,0.00153477414221573 +0.001401889534519265 I,-0.0005010335612557537+0.00113582861145429 I,0.00144548923476796 +0.001446876580755573 I,0.0001877926049013016 -0.0001950621943520755 I,0.000507210979704594 +0.0003981432041493133 I,-0.00144053594348099-0.0009441945037857274 I,-0.001359745526850345-0.002072719210344509 I,0.0000949373440565014 +0.0005691745124108819 I,-0.001915199942066621+0.0001599095861269358 I,-0.0001015617693740703-0.0005398823777048855 I,0.0000623941946725627 +0.0016012772770148 I,-0.001059377500237933-0.0007544668788280943 I,0.0001317706626131534 -0.001071748134836275 I,0.001136196201917283 +0.0006759686810417618 I,0.001374354148033255 -0.000794715995856044 I,0.001362658861523115 +0.0006989852283953122 I,0.00003856988928966818 +0.0008998402222053749 I,0.0007200289141170481 +0.001966580284123686 I,-0.001349409608455128+0.0009202375096627328 I,-0.001422224451923424-0.0003805408130133116 I,0.001427510616675689 +0.002442716330815145 I,0.00009099666314676198 -0.001847882651241556 I,0.001042273201474888 -0.0009077584937139665 I,0.002049099862954986 +0.001549908764258162 I,0.0008169980787376904 +0.0003375935384088664 I,-0.000543017573388294-0.0007383166034075043 I,-0.0002472552600440107+0.00006336008758311984 I,-0.001115965642640159-0.001320780353803403 I,0.0007525184113974178 -0.0006725464550134459 I,-0.001194922424710917-0.0009744207931326299 I,-0.001454375052288711-0.000269216810277602 I,0.0005151222024571522 +0.0008211040586719387 I,0.0001249119569164351 +0.001118198427543936 I,-0.001402699989883321-0.0004649865337630142 I,0.001445238868348758 -0.0004565647527551893 I,0.0004782990435573508 +0.00005041316522431526 I,-0.001121688520667049-0.0002445088726745758 I,-0.000707669509855788-0.001294357852255852 I,-0.001906044908645474+0.0002853375764686196 I,-0.001633058361885669-0.0005095558742203736 I,-0.0003075805810756694-0.0009826631862872464 I,-0.00107082562942359-0.001617208755763716 I,0.0008135037096704057 -0.001269247719870888 I,0.001172623596720351 -0.0001472930400817986 I,-0.002074998961546478-0.0008497130642494207 I,0.000648922294808636 +0.0007049592638956691 I,0.0001912296297651667 +0.001176218974807024 I,-0.001222911000509393+0.0002961852335570789 I,-0.0003404229561727527-0.0003600617234608286 I,0.0007049166767801483 -0.0005279696966837604 I,-0.0005919829880497204-0.0008846051923533738 I,-0.00003813815918370512-0.001061801342071409 I,0.0004529504054316301 -0.0005532169220291588 I,0.0004056443462229994 -0.001423830756502625 I,0.001319330954491618 -0.0001423220169303704 I,0.0009883585152212213 +0.0006130977610049268 I,-0.0007759701351706266-0.002426841586843312 I,0.00009032879118495672 +0.001573090725988496 I,-0.00143032788260467+0.0007891107328588866 I,-0.001460985891779401-0.000974519065310572 I,0.0005488504444042142 +0.0002294653701455844 I,0.0005871613654453228 +0.001237470151349602 I,0.001704467933290253 +0.0001343270801746434 I,0.001323289795973653 +0.0001047235713702765 I,0.000718561464518122 +0.0007298230831281944 I,-0.0000718516215388145+0.00002039680459487868 I,-0.001639262118770379-0.0006038761400122638 I,-0.0002364376302795828+0.001217174009589175 I,-0.0003236592705989623+0.0004852681701822678 I,-0.001774509358412609-0.0006917378209278641 I,0.0007468973484576606 +0.0009630185680409922 I,-0.0009119954898736396-0.001837846487745259 I,-0.001357095962962938-0.001906951943675207 I,0.0009485151208990204 -0.00005273088846783844 I,-0.00008732303687786127-0.001215762074834615 I,0.0008772613447272958 +0.001185003645562282 I,-0.0005443498264469217+0.0005032558516496431 I,0.001426283852147192 +0.00191545368623535 I,0.00005902069997502837 +0.001485870460045465 I,-0.0009624918980559868-0.0003950051330982288 I,0.001548718744914684 +0.001073717117051672 I,-0.0003864794701128253-0.0008637691259047468 I,0.0006700511031626966 -0.0007049871767694062 I,0.0009772445787319937 +0.00003916225181906216 I,-0.0006569316587302046-0.0006203101439371894 I,-0.0001340701565674504-0.0007845553468192837 I,-0.001261157393752383-0.0006476931303887908 I,-0.0005116401478869594+0.0002087641735854979 I,-0.0005824898258444706-0.0003636837180383183 I,0.001191796905801042 +0.0004534683460380185 I,0.001973466004895408 +0.001361255469109133 I,-0.0002326979720315915-0.0006244421236693666 I,0.001512454896382753 -0.00128611260812795 I,0.0007640322936085775 +0.0005560385299383981 I,-0.001486031634955529-0.00111540273838452 I,0.001314535384824074 -5.308866594702882*10^-6 I,0.001361031404945715 +0.0008468996315587509 I,-0.0008706860244342715+0.0008817751728717854 I,0.0007438136396818309 +0.001044440360945158 I,-0.0006555044778443324+0.0006224052619848027 I,-0.0003956665058866291-0.0001113214714425792 I,-0.00100338316346827+0.0006708898409553666 I,-0.0003308351262226756+0.000108472332477671 I,-2.797470897371573*10^-6-0.0008188713404488243 I,0.0001552387414079088 +0.0004748974852333005 I,0.001025809350488818 +0.0001026616745476863 I,-0.000454821629767692-0.000679921320954211 I,0.0008390295341337412 +0.001158138071415041 I,0.001369441245329571 +0.001349045402316397 I,-0.0007432991509574067+0.0005327039236300929 I,0.0009515949271044644 +0.001231571601228245 I,-0.0004109101310733952-0.0003858994829555713 I,0.0004226325939054791 +0.00006306205561510195 I,-0.001569702665433945-0.0005497396268644256 I,-0.001565265862034781-0.001330083869996125 I,-0.00009005428723591106+0.0002476413813935032 I,-0.0008266273929386145+0.001101893392541385 I,0.00002022847816027045 -0.001158253113440496 I,0.000444197775234607 +0.0004113632226334741 I,0.001083041213806714 +0.0004409344892624189 I,0.001324223832934296 -0.0004046444689929424 I,-0.0001115655999243256+0.00004934195461632387 I,0.0007766624991428011 -0.00001744939516063382 I,-0.00154905406207345-0.0009488794826689804 I,0.0007712721374868471 -0.001188075204075542 I,-0.003101730645712978-0.0007110836591678233 I,-0.0030988689172941-0.0007010755770755116 I,0.002046504823471946 +0.0008529228887893299 I,-0.002093112130939944+0.001430536783994209 I,-0.001811235578540326+0.0008953103991639859 I,0.002221060013295739 +0.001229237179822984 I,-0.002890404225924884-0.0006782687274575654 I,-0.002804611143785747-0.001350858620786386 I,0.001219690672149103 -0.001035122637873502 I,-0.000978568281553176-0.001573528630221516 I,0.001083987680480124 +0.0003497090938308528 I,-0.0002086500986210463+0.0002864713103112419 I,0.001650097212291585 +0.001238842342234537 I,0.00008427292680663581 +0.001816020127701061 I,-0.0004184439708799566-0.000121740384313337 I,0.001853323583515002 +0.001161603275645871 I,-0.0009472970582005082-0.0001156251327537948 I,0.00125214564610404 -0.001445563865903718 I,0.0005666180616943355 +0.0004409895055195035 I,-0.001675112935937551+0.0003277768364099268 I,-0.0009045581572321856-0.001448007149429789 I,-0.002429804652247224+0.0002902158860313191 I,-0.001035265815663342-0.00004849774737273043 I,-0.0007538335814440097-0.0005526842571571602 I,0.0006150671071509634 -0.0008938040828694103 I,0.002497215922511976 +6.014690854947509*10^-6 I,0.0002708789531187621 -0.000697210468762657 I,0.0001129417623571002 -0.001763936176540277 I,0.001292989217990401 +0.0008591080546256841 I,-0.001488188440209615-0.0005615973485521562 I,0.000845190464461133 +0.0003474734149381801 I,0.001587809298540796 +0.001130040137064594 I,-0.0009419069024236625+0.0008198945214678174 I,0.0008172456445012266 +0.001425880486939641 I,-0.001336432259588494-0.00001783863147009795 I,-1.506561066805506*10^-6-0.0007244085015110364 I,-0.003067454516590425-0.000587285215261058 I,-0.001973354235447141-0.001395893004493749 I,0.001749878771709749 -0.0003172703877761583 I,-0.002790990155335087+0.001155385712191813 I,0.0001517127880224097 +0.001216444026949627 I,0.001623909681076974 +0.001103326755587998 I,-0.002953164988183199+0.0009837553201070127 I,-0.002026853604585722+0.0003722678465857604 I,0.0007978192842532318 -0.0009812217863961294 I,-0.0009551267556682185-0.0004855566073079025 I,0.001257119232014597 -0.00067083896852861 I,0.0002938504315449144 +0.0005011766782056042 I,0.001234500681687668 -0.001028285775210826 I,0.0008145359865962061 -0.0007193915329249426 I,-0.0001513527832259708+0.0004023057655349106 I,-0.0007696615374311411-0.0005345023886352308 I,0.00001115675305040026 -3.954935303655512*10^-6 I,0.0002524742900317957 +0.0008397846605832238 I,-0.001057502733991969-0.0001092711537483688 I,-0.0008109538328719833-0.0004503253615706058 I,0.0007698170657242788 -0.0004406909996089432 I,-0.00051502901703847-0.0008877274591393018 I,0.0008603042468789317 -0.0003849099906988991 I,0.0001767822482338122 +0.0002416465907397785 I,0.001520254368828331 -0.0003180700460883433 I,0.001231763313207015 +0.0004406761743672829 I,-0.0002323108491030054+0.0002525345290117955 I,0.0005181612125842802 -0.001056436412665264 I,0.00005800597182814798 +0.0004190328312751444 I,-0.0006057560391930479-0.0001183049952855206 I,-0.0003070604153019949-0.0001537563189695363 I,-0.00004577665320877633+0.0004727383348741027 I,0.00003868270044029665 +0.0002179199847031218 I,0.0001597951323653282 +0.0002197510040988934 I,0.000849474529076752 -0.0001381712311654091 I,0.0002946880482830737 +0.0006157320079819026 I,0.0009736671920810567 -0.0005049262715556362 I,0.00045828756016599 -0.0004827868272708841 I,-0.0007761429786630771+0.0006400747107433506 I,0.0004555732098600531 -0.0005832416545273238 I,-0.0004713764396271033-0.0008032036784178348 I,-0.0004732114967762274+0.0001284184603004517 I,0.0007815971148603146 -0.0009761108145915435 I,0.0003610642278279628 -0.0006724980646025029 I,-0.00008613107597024835+0.0003209275912423976 I,0.000174176200166941 +0.00006760705866529649 I,-0.0003258750084974961+0.0005695535730619365 I,-0.0003483387964722741-0.00002614570107189291 I,-0.0003076808444769999+0.0004910592587276518 I,-0.000128715111923215-0.0001181322676320953 I,-0.0001795427852938974-0.000382518706770264 I,-0.0002535241985639426+0.000467050842680169 I,0.000420667190428276 -0.0004879790034780993 I,0.0002128269151931907 +0.0001963928633260545 I,0.0004280160993332159 +0.0004057697660142663 I,0.0003410669220615938 -0.000303941468799501 I,0.0001412215843816932 -0.0002523973007088026 I,-0.0001544139512829718-0.0004358155132182242 I,-0.000503283529099043-0.0004445526522609481 I,0.0001198163183856497 -0.0005042255530464725 I,-0.0002877424877608515-0.0001901726457461473 I,0.0001234889243887483 +0.0003809845567402084 I,0.000561415980189265 -0.0000496648832427638 I,-0.0000219647137498244-0.000324816015957938 I,0.0001575386072638983 +0.0007210989854131856 I,-0.0001781531605358443-0.0002467838799260978 I,-0.0002789878247930946+0.0004107902312553342 I,0.0001869891752758247 +0.000789558292837935 I,-0.0002726436552590153+0.00009297828791625926 I,0.0001991579173321233 +0.0005815064166094868 I,-0.00003729580695260967-0.0005150618750681252 I,0.0004535600127043271 -0.00001520718887420148 I,-0.0007746236760753266-0.001072973900847001 I,-0.0007448255265666648-0.00120766856454885 I,0.0004556066355335567 +0.0006856482537095062 I,-0.001296744032667764-9.653099232289361*10^-6 I,-0.0006370643706917851-0.00003187372353073378 I,0.0005872348316079474 +0.001044189273171765 I,-0.001181419619353044-0.0006100587046756328 I,-0.0009126327602125486-0.000858994366905384 I,0.000747096092646553 -0.0003145198370893494 I,-0.0002913270850951948-0.0007922722254288609 I,0.0006019027697592818 -0.00009508841695224276 I,0.0001392683393273691 +0.0001778362310688143 I,0.0008189220732660839 -0.000137766656604444 I,0.0003077148386701721 +0.00002947279095675789 I,-0.0001645014718274879+0.0004473628107766627 I,-9.78174166207566*10^-6-0.0003400623279989992 I,-0.0005395172986363522-0.0002529710948419461 I,0.0001314147301715692 +0.0004943809621202691 I,-0.0002348474560079968-0.0007830017627603997 I,-0.0004817014706015233-0.001025584198944371 I,0.0005197609531707729 -0.0002052070224169914 I,-0.0003436426108276559-0.000887329462019995 I,0.0000627004602604058 +0.0001433436354577947 I,-0.0001386554678817731-0.0002027479061933002 I,0.0006518604101327684 +0.0003496330273733649 I,0.0006130756849140116 +0.0008383691524679351 I,0.0002112714216478644 +4.463351612320546*10^-6 I,0.0004088565468386971 -0.0005964759031032182 I,-0.0002260304975689861+0.0006048688036119545 I,-0.0002078958893487007-0.00007097574110084934 I,-0.000508712737420803-0.0004687286918596229 I,-0.0005288781256084367-0.0001835021711167011 I,0.0002155246331674134 +0.00002660094465832671 I,-0.0002057908788618124-0.0003381875287954649 I,0.0004785394027627948 +0.0003830555854668822 I,-0.0002509785845379038+0.0002109097916622085 I,0.001057933046011323 +0.0005198075701665693 I,0.0009429921033346472 +0.0008439262901507831 I,-0.0003420927435689833-0.0002296388968835907 I,0.001007733880213356 -0.0006007803198090587 I,0.0002349531346909349 +0.000318334178171526 I,-0.00070865899185997-0.0006776228008783574 I,0.0004371347577233422 +7.101317207274037*10^-6 I,0.0003993963647058231 +0.0007041169540821915 I,-0.0004829119166522229+0.0003916301315024828 I,0.000266857635437984 +0.0007416466303198129 I,0.00008264776133773053 +0.0001766727196258902 I,-5.520576103285263*10^-7+0.0002793533221390671 I,-0.0001202527047859254-0.000281183049058514 I,-0.00002540002526824504-0.0007048742581971139 I,-0.0003200667804712877+0.0001829445017594393 I,-0.0002718403577938798-0.0001747690391622159 I,-0.00005138832756675451-0.0005471593810149689 I,-0.0002462169262979631+0.0004071704387859007 I,0.0003511647946136739 -0.0005830601997658552 I,0.0003875741934105954 -0.0006911178012425867 I,0.000183598258930593 +0.0002989330796567809 I,0.0003644653244912301 -0.000248113202315148 I,-0.0002329076167238557+0.0004140571269169635 I,-0.0002208400741657319-0.0001384790074463085 I,-0.0003604566426260995+0.0007177939532056068 I,-0.000391099451295425+0.0003651640768719507 I,0.0000757559442458183 -0.0003710254573305847 I,-0.00002506470828815067+0.0005572336964400843 I,0.0002405645917805702 -0.00004316956308973377 I,0.0002729212160056074 -0.0000466187598252457 I,0.00008564997649987047 +0.0004786966842184092 I,0.0001339647093311169 +0.00004684155521662695 I,0.00006521756216120388 -0.0001789717809610736 I,-5.499848122689655*10^-6-0.0002180723920392127 I,-0.0002176477171273439-0.0004354591464265385 I,0.0002760630253101838 -0.0002958015759632598 I,-0.0003247708933559898-0.0002848865752831503 I,-0.0003858557965135473+0.000119669692882127 I,0.00049502150209783 +0.0003381231192106196 I,-0.0001556720941049949-0.00003136231615545474 I,-0.0005897914418264568+0.0004217704918846568 I,0.0003417707983381853 +0.0001142599223466288 I,-0.0006757944536240124-0.0003600455647484869 I,-0.0007537126423439336-0.0002430846085966922 I,0.0001751357827703126 -0.0002290043686763178 I,-0.0003766127474602198-0.0002632836942998163 I,0.0001892000630322461 +0.00009649936492085395 I,-0.00007032735566618023+6.819662501135037*10^-6 I,0.0004031017022457467 -0.00002170333955133093 I,0.0003297884639227652 +0.0002008615695904083 I,0.0001122286571268427 +0.0001386916724845024 I,0.00006592710087831375 -0.0003825661033259947 I,-0.0003394287646477074+0.0002913913816314523 I,-0.00002901078831604286+0.000119325738623562 I,-0.0004489241077300845-0.0005904200110429404 I,-0.0005495587588645771-0.0004140753079949547 I,0.0002104538262228358 -4.433356755465098*10^-6 I,-0.0002098632645600206-0.0003220061620028826 I,0.0002193194638067702 +0.000476378663998859 I,-0.0003190001974623345+0.00005370828283336135 I,0.0005129725783137012 +0.0005822458232476358 I,0.0004157941328921933 +0.0006563778533331835 I,3.969481227126225*10^-6-0.0002650225709992554 I,0.0004004852102045245 -0.0001357444480924903 I,-0.00003684626108543365+0.0004332019684704287 I,-7.288453229661093*10^-6-0.0002556091814146429 I,-0.0004116048619979604-0.0001713719031900231 I,-0.0006916978434676436+0.00004357468602586177 I,0.00001616608123101591 -0.0001023051250438535 I,-0.0004624167422553896-0.00007616025052072518 I,0.0004223283237393067 +0.000596799046307055 I,-0.0005460619654467988+0.00031268111448514 I,0.001101805142628718 +0.0006751464368431199 I,0.001273229652875148 +0.0006970141292277587 I,-0.0005725368303845144-0.0007462269078876155 I,0.0008604633388741411 -0.0003727968327778816 I,0.001029097271458925 +0.0001218962009918751 I,-0.0007550587287038434-0.0007384053352867421 I,0.0007812258029150258 +0.0008166796559897093 I,0.0008963434847306401 +0.001091703937925178 I,-0.0005400505777211162+0.0002777738907341637 I,0.0002813895739528843 +0.0007407598457909541 I,-0.0001299939987189414-0.0004016125429943808 I,0.0002753970598144957 +0.00003145608374335824 I,-0.0003274128145803002-0.0008042905556845542 I,0.00008809974191880559 -0.0007190051417816177 I,0.00009753733168551269 +0.0003551506230970588 I,-0.000405020839150443-0.0007252964572969588 I,-0.00005540838746977226+0.0001409729594345056 I,-0.0005947779781040049+0.00035668209944001 I,-0.0002487167533329653-0.0006503173041019861 I,0.0003246548347342892 -0.0001173878910689786 I,0.0001250940496701528 +0.0006543388588429628 I,0.0007776294068511986 +0.000232944711780747 I,0.0004436726355672868 +0.0005625574149544807 I,-0.00008071906668437768+0.0004310015572336567 I,-0.0001190351680646483+0.0006052413614935597 I,-0.0006844517029324023-0.0002412015558026673 I,-0.000456896789381978-0.00004993319125001073 I,-0.0001242849368431974+0.0007581559251175491 I,-0.0001216374881978801-0.0007603055137164073 I,0.0004137085456747277 +0.0003395150739987206 I,0.0001753323397177695 -0.0001800077130417342 I,-0.0001571424474158736-0.0008236759697293965 I,0.0004030584864772694 -0.0001623847400775867 I,-0.00009994804963407808-0.000749896553286498 I,0.00003566567509975989 +0.0003087980327728092 I,-0.0002866206455834035-0.00007610421705391007 I,0.0005835743533589087 +0.001040080121956425 I,0.0003441845020381481 +0.001273751750347835 I,-0.0002406544620660631-0.0005490994207522697 I,0.001164195799395716 +0.0002866333570540385 I,0.0003628373417450505 +0.0001585415241267166 I,-0.0001796882419790023-0.0009799507927542076 I,0.0007721961265197717 +0.0006904116850788215 I,0.0004178105339540917 +0.00083367964041795 I,-0.0006297308772751749+0.00001544649550568489 I,-0.00003377173975388948+0.0006231275018574323 I,-0.0004503865472790335-0.0001057142053309837 I,-0.00003686994528335672-0.0001620058212423821 I,-0.0005684315383535728-0.0001647443972777075 I,-0.0001815011737782902-0.0001287950359143721 I,0.0002316401712694102 -0.0002171635490882974 I,-0.0001295331713082774+0.0002092859923118123 I,0.000254110493833862 +0.0002916357269342644 I,0.0001002122516040947 -0.0002801361819661381 I,-0.0001353156029165559+0.0005492091557099731 I,-0.00006851559782172594+0.0006942084451041809 I,-0.0002911561796551931-0.00009712742728885298 I,-0.0001360606959510861+0.0004928410286262932 I,0.00006535527468638877 -0.0001996670240267505 I,0.00007055015289905361 +0.0001473908570336632 I,-0.00007261400125893906-0.0004958315496957082 I,0.0001122478068064577 -0.0005864468632639577 I,-0.0001084517731298182+0.00001966267086679711 I,-0.0003342452171782855-0.00006032213677838096 I,0.0002818783323543054 -0.0001499565664543449 I,-0.00006448169013538506+0.000123597150338753 I,0.00002926777220931388 +0.0003015895623849514 I,0.0002207947224050805 +0.000293726602810064 I,-0.00003700934264020772-7.278271295451159*10^-6 I,0.00007412714496561146 +0.0001974875905271577 I,0.0001263460285808846 -0.0003884482030135811 I,0.0002972312232396037 +0.0001246156803785321 I,0.00001537250454447005 -0.0006801833895435819 I,0.00001618646813251828 -0.0006679303448474919 I,-0.000136394590232577+0.0004568907198454926 I,-0.0002927061312043719-0.0003752945424925447 I,-0.0002384103751433412-0.0003618134946115348 I,-0.0002238889865440349+0.0004097792793957953 I,0.0001430468904998532 -0.0005859237527433016 I,0.0002509425684861018 -0.0005060771598231424 I,0.0001478911262458653 +0.000248481794811195 I,0.0002982089353303275 -0.00009608390753747655 I,-0.00007372184096013617+0.000235950806554851 I,-0.00006779106250998925-3.153467705209619*10^-6 I,-0.0002552756538682898+0.0002465807731873796 I,-0.0003037502874947731-0.0001402927973408357 I,-0.00003859029941530204-0.00005283568060466024 I,-0.0003003301647845514+0.0002457118379176988 I,0.00005529258772021589 -0.0002850187467093931 I,0.000121756180720206 +0.0003534105489669735 I,0.00001537263135454605 -0.00006962393451307559 I,0.000151788464430209 -0.0004730814377151324 I,0.000320296479860922 +0.0000264805039162997 I,0.0002211667672078623 -0.0004583954848574743 I,-9.071427161741019*10^-6-3.912381565917547*10^-6 I,0.0000470634883610768 -0.0001145949436700082 I,-0.00005355596283507268+0.0004389516637203457 I,-0.0003731686239669327+0.0004891683326721639 I,0.00008099421554486291 -0.00006345752312682804 I,0.0002205822131770011 +0.0004597414356767148 I,-0.0001355328429689601+0.00001817724518730256 I,0.0003849143496202256 -0.0001410493041437325 I,-0.00006012750853272623+0.0002838852216124029 I,-0.0002753142295107401+1.142068639753858*10^-6 I,0.00008055093558924516 -0.0003322237288225368 I,-0.0003165700174187599-0.000301858893219487 I,0.00004248353507281988 -0.0003387810675179656 I,0.000170906023984545 -0.00009270186731999977 I,0.0004431288443808595 -0.0002352017460753616 I,0.0003640226643615565 +0.0003554834528498401 I,0.00001027105151535628 +0.0002974836168772811 I,0.0005619640128983518 -0.0005005343568553742 I,-0.0003878899767013863+0.0001558211741404882 I,-0.0003120744255114204-0.0002795012456947733 I,0.0002723205032893635 -0.0005102461937365166 I,0.00005551028986686645 -0.0001047468184375324 I,-0.0002080527946498929+0.000182009519531829 I,0.0000224228389399323 +0.0001379781665093131 I,-0.0002763390260353732+0.000345446797229952 I,-0.000220353297598644-0.0001446016438568601 I,-0.0002137982277772109+0.0003218050831738253 I,0.00002757007075348335 +0.0003365952646394807 I,0.0001932016129302632 -0.0002173581695556234 I,0.0002251116629843697 -0.00008470171089275212 I,0.00004026180708182356 +0.0004238950301494731 I,-0.0001953893333319411-0.0003546933510655733 I,-0.00006507259119515598-0.00001533140369983099 I,-3.169697308846972*10^-6+0.0004222013264792349 I,-0.0003102042002505354+0.0002400710848037485 I,0.0001707249242436111 +0.0005352742228402523 I,0.00002201587648269362 +0.0003870691269041944 I,-0.0002007671045199583+0.0001837270682520214 I,-0.0003127009112508475+0.0002601812432183544 I,-0.0003240953612835368-0.0002441148930255938 I,-0.0001502628773670412-0.000271632103309071 I,-0.0001415972271520389+0.0005443429128950098 I,0.0002045313284616405 -0.000243674323562465 I,0.0002549344703763898 -0.00009918046921951571 I,0.0001635228893634368 +0.0003674758159344983 I,-4.677381531199777*10^-6+0.0002018135001771544 I,-0.0001933054298665367-0.0001408537067394338 I,-0.0001590572951632638+0.0002748860309172711 I,-0.0002208107333610795-0.00003816493983625687 I,-0.00007452609435168434-0.00003313825072907752 I,-0.0003859563966835509-0.0002321451524438857 I,-0.0001282576001873164-0.0005622793209828303 I,0.00004367364433511341 -0.0001356655938598243 I,-0.0006893127338300563+0.0002433425535204512 I,0.0003090713731253033 -0.0002652072951668617 I,0.0002385376770767466 +0.0004821320506076954 I,-0.00011710300054052+0.0002277748938484793 I,0.00009708560903679952 -0.0003728271289030485 I,0.0003590648391938897 -0.0001578506453570057 I,0.00002258171011578587 -0.0005272515248940189 I,-0.00008654619466587723-0.0002666368884193324 I,0.0001160096396699135 -0.0002491694307482711 I,-0.00007429563799561873+0.0001405285769864211 I,-0.0001752595723642044+0.0004709034419355648 I,0.0003356968214958097 -0.00005115324788861232 I,0.00006253839663624089 +0.0001851101770470676 I,-0.00005323046991873547+0.0004294265396294214 I,0.0003004564908748741 -0.0001595603068501619 I,-0.0004416719527511861+0.0003869540576983671 I,-0.000418178475933369+0.0003380306325164953 I,0.00003633359152711271 -0.0002971376418660481 I,-0.000311492186602896-0.00008367404629557946 I,0.0002518445829228159 -0.0003632875042719636 I,0.0001997869441899799 +0.00005325073664032157 I,0.0004541986713569821 -0.0004091835662479767 I,0.0003436528284624268 +0.0000905240038828877 I,7.121715345865349*10^-6+0.000279901557753958 I,0.0003097965630363736 -0.000406713061454807 I,-0.0002621763093192239+0.0002306184308114555 I,-0.0001425191265357336-0.0002343589529389258 I,-0.00007588256720852211-0.000165452404853993 I,-0.0002195517481095227+0.0003008364708418183 I,-0.0002304528645677594-0.00001898136274451064 I,-0.000217804773968878+0.0003664122011446623 I,0.00008952145142872606 -0.00001490754102025592 I,0.00001567154064673916 +0.000153705972630906 I,0.00014415596067692 -0.0005567064163951582 I,0.0003649200829568694 -0.0006464884983744497 I,-0.0001187204355181785+0.0002609222706178076 I,-0.0003922673486534234-0.000566700918885357 I,-0.00006382257214168879-0.0001890680989618735 I,-0.0003148424604455415+0.0004983644564585025 I,-0.0000973449473471794-0.0006314930056262311 I,0.0001075059196390514 -0.0005626495158212539 I,0.0003028206770541744 +0.0002523952540756623 I,0.0002505297965804075 -0.0002976756113543106 I,0.0001804923333324722 +0.0002910244601436673 I,-0.00008900221626505309+0.0000894697026727388 I,0.000258413622803091 +0.0003792737103337691 I,0.0001036584326505411 +0.0002319873321630166 I,-0.0001838745501332979-3.455088275079331*10^-6 I,0.0002245875280319161 -0.0001520279964297646 I,-0.0000569670728495852-0.0001532739819021998 I,-0.0002582330952764926+0.00007580179843269628 I,0.0002397341614659238 -0.0004110872577250315 I,0.0002760334486754919 -0.0004114189291921121 I,0.0001237367266529079 +0.0002886000974579062 I,0.0003679260548335788 -0.0002116976277821975 I,-0.00006614315488636345+0.0003573461953374309 I,-0.000123773162283188-0.0000618175438639081 I,-0.0001455973532811604+0.000700806729588227 I,-0.00040800633693764+0.0005537892318241068 I,0.0000271973787613057 -0.0001324187638824356 I,0.000381279772055976 +0.0004101240082299206 I,-0.000284896158782865+0.00001532599193410599 I,0.0002199923094586583 -0.0002371744656539526 I,0.00009084772250043157 -0.0002294835571273008 I,-0.0003049258444527366-0.0003086835225561815 I,-0.00006087960561989279-0.00005087809046612463 I,-0.0001960564171416709-0.0001378633864891673 I,-0.0002728713009864671+0.0006162981941364989 I,-0.0005461970121334393-0.0001654587538969479 I,-0.00009653566304500597+0.001029803484815718 I,0.00007309158267654889 +0.0008771850861153011 I,0.0001233201242807243 -0.0009083849345201388 I,0.0003385168005056118 +0.0006734447921358503 I,0.0007335760680750454 +0.0004590683059805314 I,0.0002985302240616628 -0.000653622614872826 I,-0.0000474289970192976+0.001277308470959821 I,-0.0001098081711106949+0.001003870683397508 I,-0.0002852370660858531-0.000400479186825933 I,-0.000411650765336857+0.0002239565441120095 I,0.00001951732476301166 -0.0006379986210857629 I,0.0002525793441951404 -0.000079988123281876 I,0.0002595674760192314 -0.0007062771079668651 I,0.0005189033079778906 +0.00005832302928333475 I,0.0002180022747142143 +0.0002205951843608063 I,0.0003104896098866474 -0.000738180034212271 I,-0.0001000044383420709+0.0006788937312038817 I,-0.0004779867315741679-0.0004371425537843133 I,-0.0002631002734732345-0.0001319676128517368 I,-0.0001076665873557767+0.0008178515221107484 I,-0.0004504222696961314+0.000313260196670429 I,0.0001144888333438278 +0.0009034983137224986 I,0.0003520454890241388 +0.000289758520627144 I,-0.00007539795574516069+0.0004902677978870283 I,0.00003166445339927802 -0.0003636756355222465 I,0.00008295796149602921 -0.0009714367884127484 I,-0.0004526784539638244+0.0001224339316332977 I,-0.0004424700539210967-0.0001510005180327716 I,0.00005985828313193032 -0.0006056526867504222 I,-0.0002292833200633743+0.0005035271419076715 I,0.0001763905968135378 -0.0005998830794118068 I,0.00021572978128026 -0.0007066709226271175 I,0.000256340028599586 +0.0003321651596616314 I,0.0003524381968000413 -0.0002587579681232643 I,0.00005244296545435588 +0.000749188581021371 I,-0.0004810880841357036+0.000111626628039496 I,0.00007203419411200577 +0.001288202697970153 I,-0.00009281526918597136+0.0007624556518051545 I,-0.0003503262131892888-0.0008735563617688097 I,0.0005595411986155918 +0.0008761271286452563 I,0.0008136190960052131 -0.0001503823433872437 I,0.0001195780507983983 -0.0006906911771428648 I,0.0006589463929844037 +0.001236354832141762 I,0.0005095430772969397 +0.000921758055103388 I,-0.0004611223676737937-0.0001296416170965124 I,0.00001627896714146804 +0.0005203984360923602 I,-0.0002880742710515308-0.0004446685206689255 I,0.0001761251907428527 -0.00007407854144861364 I,-0.0005132806418496856-0.0004150754974440964 I,-0.0003279037625007884-0.0003977254331203064 I,0.0000797046893203508 -0.00005454759408786235 I,-0.000256262233052696+0.0003861738389600377 I,0.0002265061746950743 -0.0000853142568769678 I,0.000259760869090029 -0.0001115438996009593 I,0.00007419117927205741 +0.0006422840191860105 I,0.00006351235388455658 +0.0004912635036352083 I,-0.0002013588195189445-0.0002370006395300482 I,-0.0001893227885513638+0.0002955716287769134 I,-0.0001104973502977273-0.0004686358999025889 I,0.0002074981973799433 -0.00001794371115355296 I,-0.0001922952982867097-0.0007010675028914869 I,0.0000434799396705339 -0.0006215873917408861 I,0.00002209148419877419 +0.0001637395511333345 I,-0.0004230171437933584-0.0001305422590181298 I,0.0001709980241735872 -0.0001708734171835498 I,-0.00005553768859545879+0.0002353749190628681 I,0.00009098842403566364 +0.0001864946246354324 I,0.0003501565411742867 +0.0000768189972484284 I,0.00002900013608153563 +0.0000193376968756727 I,0.0001840467886090695 +0.00004275707776509974 I,-0.0001202641135072416-0.0004212308178972621 I,0.0003227247014261664 -0.000119155577649269 I,-0.0004116859409959804-0.0005153502712299455 I,-0.0004256761555441852-0.0004243242405962651 I,0.0002602328738020575 +0.0003925295647969365 I,-0.0004462863700064466-0.00001890838665449715 I,-0.0003836972116998669+4.769756969068111*10^-6 I,0.0002171563670874228 +0.0003652579383227898 I,-0.0003947381073015177-0.0003264372620595356 I,-0.0003061870764977066-0.0003710776524574751 I,0.0002478820958282574 -0.00006832474082115402 I,-0.00004154939605625885-0.0002852328070795611 I,0.0001386136090482601 +0.00006428545540713217 I,-0.00001092229675165723+0.0000139518106039736 I,0.0001733148923290047 +0.0001969680968663603 I,-0.00004455340088607978+0.0002764728660946741 I,5.514169817487195*10^-7-3.067579142064807*10^-6 I,0.0002356114844074894 +0.0001413230990864351 I,-0.0001472303029513929+0.0000827313515412011 I,0.0001600248734987066 -0.0002216359770439945 I,-0.00003909150029497938+0.00007855158331693539 I,-0.0002883009118652454+0.0001820384787435053 I,-0.0001594371015050863-0.0001698422420773686 I,-0.0003140796123344541+0.0001737802046361742 I,-0.00001127812081694064+2.872687737582225*10^-6 I,-0.00007932346470175143+0.00005292301454314498 I,0.0001517038857815856 -0.0002447448073140967 I,0.000369776830192419 -0.0002401970849936865 I,-0.00009241490323188216-0.00001140612035250649 I,-0.00008257623003560526-0.0003157387857896642 I,0.0001433537774830548 -0.00005550704308284767 I,-0.0002827116142819241+0.00007947311285634074 I,0.0001512536143337623 -0.0001918881598592875 I,0.0002611461915959583 -0.00009074190844368245 I,-0.00003936416153103406+0.0002403309126894343 I,0.0002471119715662734 +0.0001160849997742595 I,-0.0001285230884367815+0.0003178019556788757 I,-0.0002077373807120221-0.00001999350262463267 I,-0.0003694148588606051+0.0004754319881689476 I,-0.0003217020831649518+0.0001170159973057756 I,0.00004484947090468267 -0.0004476704255249313 I,-0.0001320576945160869+0.0006213985988939401 I,0.000429860815620303 +0.00004574775608481944 I,0.0003220751157174758 -0.0001848106514917581 I,-0.00005715639387685674+0.0008167120306813289 I,-0.00001612757499691044+0.0005467901023074793 I,-0.0001316289247189901-0.0002447483905878331 I,-0.0001518908433176466+0.0002083811579765197 I,9.294056061398548*10^-6-0.0004089253087608528 I,0.0001815668377169826 +0.00001737735829005486 I,-0.0001348321282564006-0.0004633454348645134 I,-0.0001126167273490165-0.0003811111001189879 I,0.00004894068429504622 +0.00007535240882267097 I,-0.0003022020979195988+0.0001918954666733677 I,0.00007657627887768003 -0.00006277427194114932 I,0.0002971431849081403 +0.00006310165301252657 I,-0.0001595632251590563+0.000407030007654667 I,-0.0002154686591471723+0.0002437943174771798 I,-0.00002363773748727933-0.0003351172827166976 I,-0.0003804014341539997+0.00004858119620493155 I,0.0000284341378077017 -0.000434050396330309 I,0.0001543089050024516 -4.377482060186272*10^-6 I,0.0001686866400791092 -0.000697178015319156 I,0.000426273391298649 -0.0004590920915516553 I,0.0000296532454649854 +0.0001520833972874235 I,-0.0003560898474218386-0.0004422967568446104 I,0.000234403500558467 +0.00004523601176501261 I,-0.0001950407730957203+0.0002372554492837994 I,-0.0001443321999268768+0.0001688095000193524 I,0.000201278629970689 +0.0002269770300380125 I,0.00008359389833375772 +0.00001485506289131854 I,0.00009700165915608235 +0.00006309054527671876 I,0.0002241937530773906 -0.0004939405325743748 I,0.0003951171911891472 +0.0001098580682975414 I,0.0001374356457118585 -0.0008135238764953851 I,0.0001046704128978354 -0.0006431614281276917 I,-0.00005776663677617021+0.0006113056976438682 I,-0.0003161359797421407-0.0006216920946294657 I,-0.0003602909512196904-0.0002236122070625571 I,-0.0003190909896105003+0.0005093709507184151 I,-0.00004870068938183985-0.0007778602244596998 I,0.0002029885617770163 -0.0005800806742358631 I,0.0002575868765745444 +0.0003321807629028091 I,0.0004181165972387682 -0.000184302203666096 I,0.00008624708009515709 +0.0002622901324266584 I,-5.619353025876437*10^-6+0.0000516211267235923 I,-0.0001807255160627014+0.0003401814140863028 I,-0.0004747041946642188-1.701658074274547*10^-6 I,-0.00001910443001690711+0.00004750496476506463 I,-0.0001702789030289084+0.0003909798721498865 I,-0.0002063872974024127-0.0002162278362665835 I,0.0003558305217250034 +0.0002488443899248085 I,-0.0001910142590767038-0.00009939810670883979 I,-0.0003094285132099897-0.0004224706732021408 I,0.0002880880989375513 -0.0002000348467530854 I,-0.000166469135832068-0.0004567629725249243 I,0.0001407608862287395 -5.680514192994105*10^-6 I,-0.00001437349953819047-6.1467199817377*10^-6 I,0.0004236336156308262 +0.0002393133945591352 I,0.0001884610025846516 +0.0004686564545555673 I,-0.00009946314328796986-3.151180308126188*10^-6 I,0.0004800674310245004 +2.233209972543994*10^-6 I,-0.0001358024537378019+0.00002145001960713786 I,0.00002401378636338459 -0.0003059746741737431 I,0.0002066678721168682 -3.81837063382983*10^-6 I,-0.0001391302570403979+0.0000865863126245148 I,-0.0001961498447216732-0.0001293596793658401 I,-0.0003095920220011678+0.0001208765073203809 I,-0.0001950866769404174+0.00006130130194467097 I,-0.0001608059789255057-0.00008665062929178314 I,0.00003022534484236868 -0.000102179899039224 I,0.0004085811807245619 -0.00003410398143302942 I,0.00007433641752914119 -0.0001874257987638624 I,-0.0001016150725151783-0.0002968644815667943 I,0.0003299302764614825 +0.0002304541130822695 I,-0.0002487633660262211-0.00002376561696134928 I,-0.00008402704900886476+0.0001437643267966276 I,0.0001438234037781548 +0.0002790541706082037 I,-0.00005960293863837109+0.0001301945673317497 I,0.0001130456284336716 +0.0002055689120644521 I,0.000114651024671497 +0.00003014872433313397 I,-0.00003088608716884087+0.0001206463453091104 I,-7.741231083762605*10^-6-0.0000451742820661366 I,0.0000891646530676094 -0.0002224147321707302 I,-0.0001022435575118429-0.0001581143939437134 I,-0.0002225484177235254+0.00003693627973178559 I,0.0004131232226871956 -0.00005656934791272269 I,-0.00006441946813567705+0.00009676739936151005 I,-0.00004053468779416499+0.0004359738861787832 I,0.0002381712624850439 +0.0003057743439678445 I,0.00007226677773972806 -4.564264289815055*10^-7 I,0.0001388420635994189 +0.00002918930432022711 I,0.0002550156934335708 -0.0003912847975396176 I,0.0002868462303029359 +0.0001432828996575319 I,0.00021675991787546 -0.0004144382959662523 I,0.0001019883744377393 -0.0002636317349803389 I,-0.0001789327125954481+0.000262569104669084 I,0.00002516233541450001 -0.000250108306019677 I,0.00004218551079506405 -0.0001738091086642062 I,-0.0002533615561389533+0.00005880494501351712 I,0.0001909188706371009 +0.00005978637392319556 I,0.0004296608794862349 +0.0002252393848078712 I,-0.00002631726572651055+0.0001799419177762842 I,0.0003700782927860657 +0.000197110757685936 I,0.0001861101730788347 -0.0003978541825423286 I,0.0004574855413612864 +0.0001706751956222016 I,-0.00008441663169975265-0.000633076320569727 I,-0.0004242776863857214-0.0006577554148119997 I,-0.0001842781042615912+0.0007534391869175364 I,-0.0001294700494270634-0.0001627370272571594 I,-0.0007763565981254407-0.0005849156343121027 I,8.748537581766474*10^-6+0.0004064031405215988 I,0.0002169686966922858 -0.001007441838131592 I,-0.00006481132982767896-0.0008948540690938834 I,0.0001340248500142687 +0.0001509101074833235 I,0.00007563460661044405 -0.0001796082444890402 I,-0.0002062230236502653+0.0005663282724880739 I,-0.0003387365164389928-0.00002149564317453366 I,-0.0002088467916175021+0.0005946134544858481 I,-0.0001910992260339058-2.879086491809168*10^-6 I,-0.0001671056303423946-0.000363686318501221 I,-0.000249899613787364+0.0005311791501875141 I,0.0003565616436519707 -0.0004930048738379918 I,0.0002588949056549246 +0.0002753035178632841 I,0.0003412769833761538 +0.000338687254424191 I,0.0003320619039393583 -0.0004448845856948145 I,0.0002813302945079282 -0.0001965420799167709 I,0.00001940634622024119 -0.0006000523435388809 I,-0.0003518085141525973-0.0003799874146759689 I,0.0001634545992719596 -0.0004296936786356891 I,-0.0002218040508440987+0.0001140654461049211 I,-0.0002121388806303969+0.0006101530736109316 I,0.000448666587262947 -9.605147838742754*10^-6 I,0.0001846721273672582 +0.000115729408244279 I,-0.0000984260380358699+0.0005073756850138972 I,0.000218705855998645 -0.0002872395780907067 I,-0.0002898404172442295+0.0004311693280198723 I,-0.0002372798263240171+0.0005098579391831203 I,-0.0001118121972645758-0.0002552138146411074 I,-0.0002440377549953953+0.0001278448796255822 I,0.0001144028892859134 -0.0005776364787254967 I,0.0003971900220426733 +6.666509679514963*10^-6 I,0.0001603255682497978 -0.0009357926627156468 I,0.0001568074646254959 -0.0004999911470901627 I,0.0001000276385592163 +0.0007061139668883978 I,-0.0001831128943587715-0.0007075928001136647 I,-0.0006413038334105959-0.00002722740790433726 I,-0.0002146729488922072+0.0004050030013275031 I,-0.0001961966244095903-0.0009975723985040992 I,-0.0001780357723050139-0.0007034988986700777 I,0.000240661553008057 +0.0001873502607173283 I,0.00007945125489323577 -0.0003054582345688294 I,0.00005363769389771494 +0.0004006348510908206 I,-0.0001741227910723863-0.0000117699172933286 I,0.0001985761045532992 +0.0004209417245549922 I,0.0001469937693690323 +0.000376285747113483 I,-0.00001503955136795607-0.0000508607328357906 I,0.0002725824809244798 -0.0002122886887568967 I,-0.00009951265835757706+0.0001307271588879611 I,-0.000252768898726192-0.00008237367335079727 I,0.00005521653203327613 -0.0003574780242697822 I,0.00009968030743984913 -0.0001233954533789314 I,-0.00002448424845798375+0.0002772007732706498 I,0.0002769211502225973 +0.00006559172466132773 I,0.00001932076555693195 +0.0003959624077048512 I,-0.000157649510859336+0.00006650833905177608 I,-0.0001388680228462624+0.0005169362358446308 I,-0.0003179862749613874+0.0002215375003894962 I,-0.00008176053122654979-0.0001709990622896233 I,0.000182870364382924 +0.0004092476657429695 I,-0.00005838950207414869-0.000106967124729772 I,0.0001750683920680784 -0.0001765102226516442 I,0.0001482214309566752 +0.00004844985271984027 I,-0.0000830260914064108-9.098738486371907*10^-6 I,-0.0001301975429206168-0.00004595089852319525 I,-0.00008234890853695161+0.0001147968751486682 I,-0.0001785157219684698+0.0002056162777489075 I,-0.0001858354425158432-0.0000383354115873726 I,-0.0002494068773552994+0.000232246492359762 I,-0.0001739490392906321-0.0000433192482167172 I,0.00001177662311357605 -0.0002724712650239574 I,-0.0002156736064036838+0.0004312309361912743 I,0.0002220975124100409 -0.0001343026531557003 I,0.0003102309726322603 +0.00007576046315674908 I,-1.297529145055719*10^-6+0.0003560948912198023 I,-0.00007469285106093701-0.00009137228057329685 I,0.0001100909577762238 -0.0002534333884595256 I,-0.0002097442940549896-0.0002798943208798002 I,-0.0001415156149165054-0.0001983649179861153 I,6.704174523883689*10^-6-0.0002000838305752698 I,0.00002451602287340037 +0.00004053671632608469 I,0.0001144619903684482 +0.0003590760980953056 I,0.0002150218604223538 -0.0001456925091392263 I,0.0001259268455630902 +0.00003263759607298792 I,0.0001457450997800737 +0.0003600094925095435 I,0.00009299968814550593 -0.0002557326377258974 I,-0.0001613213526157271+0.0004570831526548953 I,-0.0001281160677583977+0.0004879087345102212 I,-0.0001286963886368328-0.0001950698310554957 I,-0.0002271278048271581+0.0001418563944736031 I,0.00005248490722362087 -0.0003924451472261798 I,0.0002021730730050159 -0.00001940318713213008 I,0.0001326432053198727 -0.000584084292720706 I,0.0002730626820637393 -0.0001914221124026732 I,0.0001272070393711856 +0.0002677191317883011 I,-0.00001743444529819283-0.000499117091445978 I,-0.0001369455172793501+0.0002841503927429191 I,-0.0002446496340908381-0.00004668068572327309 I,-0.0002053982869042929-0.0001894816756444585 I,-0.00006075252745995447+0.0002990703762100849 I,-0.0001536799313746099+0.0001479175301008803 I,0.00005963788887372752 +0.0004116166096191104 I,0.0001869370683018518 +0.00004541751689991254 I,0.00004842845099470739 +0.0002367961051676764 I,-0.00004089846872745685-0.0004134001258139892 I,-1.002455080017671*10^-6-0.0007558306537673437 I,-0.0001733240186976692+0.0002130640622214915 I,-0.0005475571487262242-0.0001569611630516801 I,-0.00001290678864413809-0.0003803231224418281 I,-0.0000417413358815862+0.0005722295621412318 I,-0.0001189923143436371-0.0003466134648490137 I,0.00006908669215769042 -0.0005262749846856147 I,0.0003641041098523511 +0.00008463128107313579 I,0.0001827601300127067 -0.0003698217179266758 I,0.0002177926052733093 +0.00002898503977783211 I,0.00006909721407213128 +0.00009627358072524672 I,0.0002653461279492341 +0.0001356092055302156 I,-0.00001442568178469932+0.0000464359083973923 I,-0.0002157358963772872+0.0001593036916505403 I,0.0001039726252753714 -0.00002230972940501705 I,-0.0001648683050762736-0.0003638103619132252 I,-0.00003527509355426566+0.0002033656950949534 I,0.0002710075783546874 -0.0003300484586837452 I,0.0001761640252433874 -0.0005502201019307396 I,0.0002046447374854219 +0.00004328405214723584 I,0.0001109173171610384 -0.0004042632097973161 I,-0.0001623701783501175+0.00004265007145716146 I,4.492589328425074*10^-6-0.0001935468521214664 I,-0.00003745085465089147+0.0002888429239897466 I,-0.0001255431960516604+0.0004313338511745072 I,0.0001515659929618618 +0.00004954267749672156 I,0.0002206999571950068 -0.00001101012211790763 I,-0.0002724713930450272+0.000133783355584055 I,0.00003114689042588268 -0.00004274146364393385 I,-0.00003867482334554923-0.0003149649193906319 I,-0.0001536353755322304-0.0003234745332347594 I,0.00008332704880383306 +0.00002618425775867118 I,-0.0000243290200856492-0.0002505169069382337 I,-0.0001013619626191154+0.0003113365078640332 I,-0.0002278148068393745-0.000133347230918372 I,0.00003732574362069773 +0.0006203721634332564 I,0.000017180863306328 +0.0007127304185267977 I,0.0001386337034791937 -0.000363223187204266 I,0.0004079812485817884 +0.0002324865258910133 I,0.0001517176556554357 +0.0004107344663385928 I,0.00009539008973405313 -0.0004779538474047294 I,-0.00009309902487491525+0.0004596969869016653 I,-0.0001901090364138906+0.0005727299952780819 I,-0.0002341046103530004-0.0001381978017079065 I,-0.0002140004509872505+0.0002619770708155037 I,0.00008781333329729987 -0.000148063712504244 I,0.00005098707599843753 +0.00007686682552361331 I,0.0001925104585116278 -0.0002971168615291414 I,0.0002781113583850154 -0.00008718966984874759 I,-0.00005118627527790675+0.00009379491003570751 I,0.0001610644958225543 -0.0003064571208916958 I,-0.00003446661042301317+0.0001099755117167651 I,-0.000236566733612865-0.0001676117980979854 I,0.00007366685697673546 -0.0001263921528418076 I,0.00005214563739172509 +0.0002320949085112165 I,-0.0002129012135342021+0.0001458339512320923 I,0.00004466287556736124 +0.0003647079098536434 I,4.20479049084157*10^-6+0.0001572747632243502 I,-0.0000775206381286907+0.0001210472471535954 I,-0.0001077537300123798-0.0001165043890561939 I,0.00003447891795694928 -0.0003841206487514494 I,-0.0001088236895846436-0.00001469625414422956 I,-0.0002460925233587188-0.00008046385676214373 I,0.0001056466530004392 -0.0001707625908832358 I,-0.00013055556781385+0.0002065961491942315 I,0.00004258535053345593 -0.0001769258370992062 I,0.0001598189878601295 -0.0001951591295705444 I,0.00008013141615171703 +0.0001761918691490293 I,0.0002177562094043959 -0.00002109096970958805 I,0.00001720734481791184 +0.0002520438543194964 I,-0.0001376845265174156+0.00005396564323553542 I,-0.0001170504056938646+0.000401803864505629 I,-0.0001890210062990522+0.0001000307556665646 I,-0.00009706664052257816-0.0002777014687601061 I,-0.00001567155535338988+0.0004086518196783816 I,0.0002813035132347353 -0.0001212363722732271 I,0.0001547342100571646 -0.00007638232714680986 I,0.00009988384910276764 +0.0004547283061803345 I,0.0001120895588925072 +0.0002244860529861828 I,-0.0000341334963616871-0.0001012672779889164 I,1.180102505861803*10^-6+0.00005505588156815803 I,0.00001138540100884802 -0.0002406979993356329 I,0.0001348893761086425 +0.00002319110525368277 I,-0.00005659562267708784-0.0002134662028162088 I,-0.0001219417538202435-0.000213404808295428 I,-0.00005994062968460608+0.00009085178523678632 I,-0.0001307540910800415+0.0001500444064609069 I,4.982246041196911*10^-6-0.0002280207020854142 I,0.0001595841582187753 +0.0001170987890863142 I,0.00009537897974072477 +0.0001443233460054959 I,0.00003672329808139356 -0.0001007559080186539 I,0.00008010490222298502 -0.00017235531678243 I,-0.0001288556877134417-0.0001819600620949069 I,-0.00009272095353774558-0.0003053807164408751 I,0.0001511036808285585 -0.0001278680549564003 I,0.00004722470664661254 -0.0003128073577039793 I,0.0001406616475788803 -0.00004467531084671587 I,0.0001137587191170715 +0.0001889165621514966 I,-0.00006716140278487332-0.0003116594446169943 I,-0.00007781460315487272+0.00008137208087327712 I,-0.0001422516291728463+0.0001419015029921705 I,-0.00006234281157461985-0.0001658835475708209 I,0.000118183662872689 -0.0001302056650320025 I,0.000134472726078969 +0.0000860470523934282 I,0.0001625684924829236 -0.0001778647124521019 I,0.00006574033610697765 -0.0001239651323124981 I,0.0001658359627315154 -0.00004751597510098833 I,0.0000660689455213696 -0.00003242958767150602 I,-0.0001025995997729812+0.000221190062031641 I,0.00009849623099245899 +0.0002569564745832128 I,0.0002802610394074891 -0.0001836664161548273 I,-0.0004057991719973341+0.0001920642223001395 I,-0.00008122320558287403-0.0001622658317328635 I,-0.00009479631187412832-0.0003689803670844569 I,-0.0002066977069764742+0.00004416759836524533 I,-0.0001364343759224248+0.0001138722823501502 I,-1.83307572203641*10^-6+0.0002500369191186647 I,0.0001192742433058086 +0.0002998590242638171 I,-0.0001306337496788507+0.0002039066216230566 I,0.00009400493990643053 +0.00007438646555176729 I,0.00004529874502052925 -0.0002588197793540577 I,-0.0002973778567237987-9.058904913505233*10^-6 I,-0.00001682371259310736+0.00004696703722094234 I,0.00002482833239072124 -0.0004380375576431051 I,-0.00008346775775375107+0.0001350676023330512 I,0.0004069178113408431 -0.0002275950445933262 I,0.0002852169643380852 -0.0003881343137024105 I,0.00002510257869958233 +0.0001056113674736672 I,0.0001321286375427688 -0.0001106531061595627 I,-0.0003141195557798998+0.0001907463703369346 I,-0.000145859380141837-0.0001800286294537777 I,-0.0003647179707593209+0.0003440779054220991 I,-0.0002326318931582261+0.0001470558065422633 I,0.0001383074418973845 -0.0002535499875732074 I,-0.0001562191934266105+0.0003043161538547626 I,0.0002153575034949743 +0.00002235069248211329 I,0.0001470559362464146 +0.00005217217811984913 I,-0.00001656978597672427+0.0003109362168584011 I,0.0001300108301657221 +3.903565058983524*10^-6 I,0.0001075680974911333 -0.00006570295450872383 I,0.00009749329366085451 -0.0001931738129355027 I,-0.00005554488285408702-0.0002524682653307008 I,0.0001842568776874205 -0.0001358963751453725 I,-0.0001320448199302069-0.0000285977644673308 I,-0.0003000852606866689+0.0002080069057926642 I,0.0001881945411105599 +0.000169413872706096 I,0.0001248033326992852 +0.0001294826876656687 I,-0.0003343364053775814+0.0001304759451030495 I,0.0002071703041386251 -0.0001035722116089898 I,-0.0001150245807994355-0.00008362910027767329 I,-0.000268323199678221-0.0000417800461404169 I,-0.00003467845515455815-0.000133193311303484 I,-0.0001835308729309226-0.00004083235820513768 I,-0.00007602514647383628-5.601608823720106*10^-6 I,-0.0000151704373530288-0.00008090252487012083 I,-6.375758909464655*10^-6-0.0000345012586008709 I,0.0000102966134584941 +0.000104958167676708 I,0.0001265055914940193 +0.0000522515663346387 I,0.00006443998283854613 -0.00005006445392199244 I,-0.0001847058562978797+0.0001558739485542878 I,0.00006142715608934362 -0.0000779027209945255 I,-0.0001118746444480526-0.0001445435839395249 I,-0.000228370172386098-0.000025197014206278 I,-0.00006357123611314673-0.00005086685884144658 I,-0.0001573826427371458+0.0000511706723699102 I,-0.00006511426566244372+0.0001421374514508758 I,-0.0001267623318049026-0.0000293452113167864 I,-0.00003083974183691439+0.00005185413992939429 I,0.00007957571828774278 +3.048736090478679*10^-6 I,0.00006881771515059704 -0.0001046313553066325 I,-0.00009821832782593173+1.101624402193079*10^-6 I,0.00005821098192205869 +0.00009536092571298533 I,0.00005466742566222947 +0.00003253939982473782 I,-0.0001354264466391621+0.00002307395884870776 I,-0.0001383977886417869-0.00002931727318372668 I,0.00004248849314953415 -0.00007062178353568139 I,-0.0001250961194083611-0.0000811867386200543 I,6.497341617154543*10^-6+0.00002062960493614407 I,-0.000061191023528426-0.00003790937027296429 I,0.0001134867448726437 +0.00002708051873155447 I,0.0002177080085480853 +0.0001360315846787304 I,0.0000689444100938748 -0.00008376786041473401 I,0.00004315777167613287 -0.0001855186441782261 I,0.000117040625996687 +0.0002287310078807038 I,-0.0001493175348993547-0.00006730132743226324 I,-0.0001236429324697257+0.00005376005644397647 I,0.00003698683064964285 +0.0002115975440403553 I,-0.00002981932730307895+0.0001023299468329639 I,0.0001227342852843072 +0.0001299124671386415 I,0.0001674984244154682 -0.00001334338171041612 I,0.00007976429842227839 +0.000127638466602903 I,0.00005788166407039641 -0.00009415697017017062 I,-0.00008178745078872302-0.0001552230241720995 I,-0.000104548915078141+0.0001696031634566131 I,0.00001089573096760206 -0.00004146878931956291 I,-0.0002155515098992174-0.0001615901788287436 I,-0.00001841113093062287+0.00009764961555730408 I,0.00005158006176724059 -0.0003497833957119844 I,-0.00005518770893952747-0.0003053033539551943 I,0.000049134703925575 +0.00008032203385870411 I,0.00004147991701049041 -0.00007501038549392793 I,-0.00003768248023724725+0.0003118482390200064 I,-0.0001937034502421136+0.000015699326240101 I,-0.00002228611987734139+0.0004085571887137173 I,-0.00006736335915516717+0.000168387728412207 I,-0.00009754552953096052-0.0002393819276278559 I,0.00005561296730955252 +0.0003015373636612433 I,0.0001634689936514382 -0.0001646410414273138 I,0.0001279307800362861 -0.00002585368749542777 I,0.0002049981916254579 +0.0001994977296152384 I,0.0001129011788679251 -0.0001296637624802303 I,0.00005158496970323955 -0.0001055772291478614 I,-0.00004176969251360905-0.0002356692909533569 I,-0.0002284121745111512-0.0001196783046115964 I,0.0000229897636512714 -0.0002472040534614639 I,-0.0001364353796626946+0.0001414723869050751 I,-0.00009969440476013318+0.000448499405683488 I,0.0002916663174324626 -0.00004855696340274999 I,0.0002296305961862578 +0.00004875547936906711 I,-0.0001510096362183451+0.0004064382814618682 I,0.0001067504190670969 -0.0002983641417242105 I,-0.000179060120491615+0.00007584410213480863 I,-0.0002878151747999654+0.0002480935134420479 I,-0.000169236661198205-0.00009943953475111827 I,-0.0002209463143771627+0.0001574622163884848 I,-0.00007168885428159157+0.00006246755115468035 I,-0.00009943194569735629-0.00005570495190286568 I,-0.00003754590757633797-0.00003734670856046134 I,0.0001417519493392028 +0.00007905109567333011 I,0.0001796060488714056 -0.0001007388650235968 I,-9.627882811506685*10^-6-0.0001374238110807827 I,0.00003135131399827919 +0.0003833273661145587 I,-0.00005434329105448655-0.0001411192026334428 I,-0.0003123945220787599+0.00001111938541642856 I,-0.000239821470882511+0.0002877741489689583 I,-0.0001105901369786059+0.00005660666704415896 I,-0.00003967099202114823+0.0002567906492564533 I,0.0001900672685309405 +0.0002283853190180259 I,-0.000145289023352357+0.0001890523728671185 I,0.00009604600361013289 +0.0001288508204459895 I,0.00006134772721930361 -0.00007792677245473735 I,-0.0001459258028936792-0.0001843629515230756 I,4.129524734837789*10^-6+0.0001725756498150026 I,0.0001726609990352324 -0.00004149654215313274 I,0.0000753920403222029 -0.0001140130288080099 I,-7.809106915023173*10^-6+0.0002199556807157415 I,-0.0001208197542565376+0.0002289394120139215 I,-0.0001219702317409612-0.00008368543671445617 I,-0.0001690746758857155+0.0002056870210469953 I,0.00005170413280213725 +0.00005740152802712647 I,-0.0001143236354383978+0.0001208697765143858 I,0.00008937319582412261 -0.00008555854075742005 I,0.0002320887125423846 -0.0002745099677312744 I,-0.0001809356446416715-0.0001769613115062812 I,-0.0001952904146115875+0.00002823741709093744 I,0.0003784160572492752 -0.0002169211670023192 I,-0.00004892983459720696+0.000125494238709208 I,0.000173043245626007 +0.0002714296916487071 I,0.0002743348086408562 +2.939305514899871*10^-6 I,0.00006948297274063999 -0.00003059444353432648 I,0.00005462882383961998 -0.0001227928647790064 I,-0.0001020290484338562-0.0002368772937109379 I,0.0001156512306433958 -0.0001386001009837243 I,-0.0001033142381302277-0.000120427787803168 I,-0.0000261214451640535+0.00009893427194747096 I,0.0002020220317193918 +0.00004564734185982382 I,-0.00002130211081979091-0.0001123797590404603 I,-0.00003106983023510614+0.0002908948835623442 I,-0.00002515059822669191-0.00005944239613649735 I,-0.0002599704151555505+0.00004679075139628168 I,-0.0001336010431735806+0.0002280570788854474 I,-0.00003048470885297426+0.00002652409920975244 I,0.00001031730860298624 +0.0001406885734302391 I,0.0002089258311168003 +5.632825326655966*10^-6 I,0.00003167198340777404 +0.0001615370958758597 I,0.0001453613772944428 -0.00009975469385654776 I,0.00004635900058876986 -0.00017260440238072 I,-0.0001511912326893317+0.00007779808396205342 I,-9.12202231652772*10^-6-0.00003751719921013113 I,-0.0000448123216629726-0.0001651912404321057 I,-0.0000368733754333123+0.00007930598074010534 I,0.0000784391577423736 -0.0001620018941148529 I,0.00002030951868574652 -0.0001908670280175269 I,0.00005016879329402868 +0.00004853883818182981 I,0.00003619442766228327 -0.00009284438605328219 I,-0.00003164547631483795+0.0001851986209699684 I,-0.0001382864523758534-0.00002018013648573332 I,0.00001347937177545628 +0.0003599757468242686 I,-7.609172033289742*10^-6+0.0003151698550246838 I,-0.00001808170937981469-0.0002587268814522385 I,0.0002212504024572426 +0.0002361269705718015 I,0.0001918119135061898 +0.00009925164472670981 I,0.00006718396936018469 -0.0003025510651187408 I,0.0001114315394111423 +0.0003917607861698888 I,0.00002227961329782859 +0.0003833773659534541 I,-0.000208421144128245-0.00008869330453727573 I,-0.0001097624170451554+0.000232102446596874 I,-0.0001148235933556684-0.0001198681471853621 I,0.00001120483094306122 -0.0000497518108227489 I,-0.0001748983306266688-0.0001780811560532099 I,-0.000020323809396375-0.0001015529329719845 I,0.0001391065010699435 -0.00006460882121414024 I,-0.0001135247659761696+0.00004227197865712431 I,0.00008538197254350752 +0.0002242265981356585 I,0.00008242089183008302 -0.00012126130601147 I,-0.0002329287933421563+0.0002432573926589918 I,-0.0002114711851418609+0.0003918315179393003 I,-0.0001460074824322765-0.00008619376277232328 I,-0.0001673550281099861+0.0002987284723912654 I,0.0001344911403263024 -0.00002561338700339096 I,-0.0000375273476467343+0.0001600155725732079 I,0.00007969289550326965 -0.0002404537622480857 I,0.0001845536584700755 -0.0003565963811063418 I,-0.0001289782767119015-0.00009049910191518063 I,-0.0002040938226992591+5.275072554972724*10^-6 I,0.0002909181222617564 -0.00007604815796415277 I,-5.228584447160506*10^-6+0.00001179316937519336 I,-0.0000237047947278606+0.0003426268384636711 I,0.00004009417020639467 +0.0003707711072352351 I,-0.0001090251106536992-0.0000744718308967442 I,-0.00009498415298325074+0.0003011057700532714 I,0.0001372547049219581 -0.0002420456065916106 I,0.0001259649973566631 +0.000192785844150856 I,0.00006834449229316139 -0.0006171912066827608 I,0.000201795915582483 -0.00082460375318627 I,-0.0002556695596066221+0.0001998496981161337 I,-0.0005353099712748733-0.0002380095660162891 I,0.0001842246049145302 -0.0005447226359641653 I,-0.0001566545185447533+0.0005817285580091282 I,0.000238186036689407 -0.0002033328106628121 I,0.0004686402440062622 -0.0005780949664178781 I,0.0003228455073481685 +0.0001126508143165182 I,0.0002871486310282492 -0.000431723333408306 I,-0.0002108298633202965-0.0001160890338034036 I,0.00008205884118217674 -0.000268146400365591 I,-0.0003433223794454246+0.0002046225835659017 I,-0.0003526733185115375+0.0002896280458190342 I,0.0003668578594489756 -0.0000802453621422081 I,-0.000206114644857351+0.0002415896681871742 I,0.00003982812171970094 +0.0003639508478304621 I,0.0003055742764468581 +0.0000578423490074689 I,-0.0005568033518681347+0.0004045752404678332 I,-0.0003323587236759814+0.0002452180041412649 I,0.000201483264805077 -0.0002444486383962107 I,-0.0001472850889621982-0.0002203548424516105 I,0.0004104016120014491 -0.0004523946405108366 I,0.0003247052656346027 +0.0001142868944305543 I,0.0005420112184404188 -0.0004125285936402404 I,0.0002182973914878922 +0.0001135339961298058 I,-0.0000427746781773606+0.0004357741307002494 I,0.0003870598050234331 -0.0003935955105706843 I,-0.0004667520665205488+0.0001757411121697819 I,-0.00009995155829574642-0.0002144142550979996 I,-0.0001449755293724389-0.0003510051065293063 I,-0.0003949111136397607+0.0002030309282182842 I,-0.0002043812317012018-0.00001704105082461633 I,-0.0002797588304199163+0.0003701891355518558 I,0.0002437155179225068 +0.0001904793329140467 I,-0.0001298832065871972+0.0003201674521352678 I,0.0004214049060150578 -0.0003550274705001566 I,0.0006053190140361106 -0.000634934207441563 I,-0.0004328540998874461+0.00005565708341847951 I,-0.0002510194223906474-0.0004913261072291672 I,0.0002983686599173174 -0.0004623626967164041 I,-0.000460155678462674+0.0004091939394000405 I,0.0003146722041480392 -0.0003116859442705291 I,0.0005438221345177984 -0.0003673538744387684 I,0.0002054654346516342 +0.0003207484335226086 I,0.000433399924277726 -0.0001537239731545617 I,0.00009376980274129033 +0.0001002277444034597 I,0.00003859784302942754 +0.0000964416889777481 I,0.00003462926038248857 +0.0002313137247596511 I,-0.0001446412573125716-0.00002101110297254345 I,-0.0002676562547856601+0.0000224811522144493 I,0.00007939204754358795 +0.0001472733123760022 I,0.00006258602055659023 -0.0004615745798178768 I,-0.00005832961234864535+0.0001226441407520627 I,0.0004775454733415828 -0.00001839382892745542 I,0.0004508612438144568 -0.000308192350691421 I,0.00009149078589099264 +0.00005879175344601797 I,0.0002079832962318456 -0.000229748483143892 I,-0.0002612364793312268-0.0001783349568259833 I,0.000145536575942703 -0.0002191290666808384 I,-0.0002736427823853721-4.230469986281967*10^-6 I,-0.0002777833404941782+0.0001091549159806306 I,0.0001727338765043503 +0.0001623377713407132 I,-0.0000123193742208056+5.707308618294147*10^-6 I,-0.0002299978138677306-0.0000189356819950402 I,0.00003129636006517125 +0.0001116995841959649 I,0.00006656036702880092 -0.0002342616418217847 I,0.0001035449289250329 -0.0003531233993621661 I,0.000120064288458815 +0.00004161815680360536 I,0.0001211414947479996 -0.0002790435229821592 I,-0.0002479543016688012+9.70246966764635*10^-6 I,0.00001652981274384975 -0.0002831350866652734 I,-0.0003170907617273124+0.0002573763143669933 I,-0.0003246972759372821+0.0004598521804608735 I,0.0004282917796585813 -0.00002782046331480746 I,0.00005325361557109596 +0.0001153374225567971 I,-0.000219804895685443+0.0005389486695352456 I,0.0001948845334724863 -0.0001539071605324019 I,-0.0005360660513257761+0.00005406700569390952 I,-0.0004507722758373499+0.0002464486861325933 I,0.00001069075366400717 -0.00007473295448017043 I,-0.00008897202827241169+0.00005165514566471389 I,0.0003379789810764166 +0.00001462909530271694 I,0.00006180575801085895 +0.0001981500563427565 I,0.0003256695239932699 -0.00003098723896174718 I,1.138689145966233*10^-6+0.00004644777986804831 I,-0.000156554634774998+0.0001939165856000703 I,0.0002833390638098076 +0.00001396658118972723 I,-0.0003746989610366302-0.00009362637244146368 I,0.0001023577642923783 -0.0001341994809830696 I,0.00002865610082733146 -0.0003223108095503918 I,-0.000369132676303359-0.0001458038256171981 I,-0.0001003586659693608-0.00009626643321730214 I,-0.0003063437547067797+0.00005643551982561961 I,-0.00001152970658921989+0.0003161820740431403 I,-0.0002683847372026206+0.00007812930276122519 I,0.0002682072965592771 +0.0001801459895471211 I,0.0004020662112757691 +0.00006262804527729356 I,-0.0002152278790581306-0.000212842317634648 I,0.0001890384702171157 -0.00009419928040458263 I,0.0002078491823985228 -0.0001503871055529742 I,-0.0001826909320786495-0.000132240220307846 I,0.0004079667453798867 -0.00002302363251530459 I,0.0002427065108212929 -0.00005221926402965447 I,-0.0001679637709418003+0.00006640115037402693 I,-0.00003078503836643982+0.0001100292461090183 I,-0.0004149142301802365+0.0001865918868773904 I,-0.0002314057506187701-0.0002397303017242281 I,-0.0004508570415313515+0.0002025062616270862 I,-0.00001983992747985866+0.0001013340792747299 I,0.0003376883773125282 -0.0003998962947908801 I,-0.0003125198113024832+0.0001128517439280381 I,0.0004015179697940039 +0.000463146499393115 I,0.00007913881544810647 -0.00004808181866514338 I,-0.0004596967320921811+0.0004761743034604695 I,-0.0001807992796873683+0.0004722052969893005 I,-0.00002003603444566325-0.00006949760708826515 I,-0.0000836511883507604+0.0001842878219185255 I,0.0002621759053351827 -0.0000958127684557202 I,9.136539691413279*10^-6+0.0001759426287589858 I,0.0001634866342760888 -0.0002143031191974878 I,0.0002258023562684005 -0.0003061297012793601 I,-0.0001039241299697564-0.00009599671198130022 I,-0.0003135278092015454-0.00007929542653928887 I,0.0003812123287409154 -0.00002485326067036656 I,-0.00005780222602530294+0.0002267434837753901 I,-0.0002133771744929285+0.0003185098864493723 I,0.0001051729900345867 +0.000145975904173898 I,0.0002232477265253231 -0.00001508223710453935 I,0.0001166147636888443 -0.0001758586258261262 I,0.0003926430667269226 -0.0003237890799109138 I,0.000253784136018198 +0.0001684591920520929 I,0.0004672333189601873 -0.0002207765692252128 I,0.0002192356173105248 +0.00004332887214725513 I,-0.0002022280972600163+0.000245542108248697 I,0.0002866176287953505 -0.0003154520624003795 I,-0.00006783078901370301-0.00004867184069367006 I,-0.000286692734265279-0.00006900422299520813 I,0.0001302943624353899 -0.0001037003476739234 I,0.0002367343912567067 +0.0001553751423798456 I,-0.00002481742935178893+0.0002025558001794535 I,0.0002935896518113396 +0.000130275745564442 I,0.0002640485228305473 -0.00007918929409145168 I,0.000212431203119211 +0.0002229778275043105 I,0.0001753844385092168 -0.000115103981754018 I,-0.0002033660975711152-0.0001318531611714659 I,-0.0002999206663607394+0.0003665185209007776 I,0.0003418275386312203 +0.00002835549572846781 I,-0.0004734602230002697-0.0004046743964799007 I,-4.176681343495724*10^-6-0.0001047514832808873 I,0.0003975592475104697 -0.0004736400850235433 I,-0.00005842083953841817-0.0002714100459171551 I,-0.0001822546359534788+0.00003107426419972663 I,-0.0001390185510090917+0.0001684117800964105 I,-0.0002241126419766427+0.0003641669935152583 I,-0.0002826022602002368+0.00002103911601097319 I,-0.00007862746256728+0.0001769014262116053 I,0.0001802227530314936 -0.000192200433433169 I,-0.0002083572389002122-0.0002642410390886954 I,-0.0001241313842058797+0.00006582267744888743 I,0.000381262917638106 -0.0003569943723216354 I,-0.0001445908678442418+0.0001080180845126986 I,0.0004784158976809171 +0.0001230052091564066 I,0.0005250462845954387 -0.0001718643028868434 I,-4.014465796581745*10^-6+0.0001010190939258734 I,0.0002041384011654236 -0.00006615690044518144 I,-0.000386257772479681-0.00005752041977221165 I,-5.04384530444123*10^-6-0.0002598666276294705 I,-0.0005202066009781118+0.0001072536871080565 I,-0.0003005031830135485+0.00003990764357711835 I,0.0002750541961970977 -0.0001569694855289073 I,-0.0002856502722335044+0.0002050070050229861 I,0.0001955049476622217 +0.0001415888774151737 I,0.0001104191523505193 +0.0001054392698155077 I,-0.0001428387130216723+0.0003080165404996849 I,0.0001542512475090757 +0.00007187748668584362 I,0.0001199116796250819 -0.00001184570964561679 I,0.0001906883685005117 -0.0001436706340890345 I,-0.0000145752646420739-0.0002762723277538896 I,0.0002275548500148982 -0.0001212346189184094 I,-0.0002442747956277221-0.00004434623034383167 I,-0.0004714934441363828+0.0001560222978325587 I,0.0002531708374196633 +0.0001779154176665121 I,0.00002339320804287397 +0.0002765129721523304 I,-0.0003649395054473507+0.0001870218859241156 I,0.0003724182193134616 -0.0001011363617470501 I,-0.0002944573572782824+0.0000471698141552757 I,-0.0004627561038901714+0.00006430012556147446 I,-0.00003460202077868688-0.0002611771596773514 I,-0.0003376582963142524-0.00001872018984488161 I,-0.0000158465860658068-0.00009872537503703279 I,5.556221240001185*10^-6-0.00006439713368228746 I,0.00005756186304888472 -0.0002052354821677636 I,0.000116534418261624 +0.00001125549462899748 I,0.0001967615557378927 +0.00006236090216971135 I,-0.00007486147045521218-0.0001318573531750198 I,-0.0001455850548990475+0.0003266574593552249 I,0.0001144977405665364 -0.00003901525167844362 I,-0.0004416420832476447-0.00005326184792569321 I,-0.0004998499637861148+0.0001526271810072354 I,-0.00002391446203265941-0.0001394390117622968 I,-0.0003176008213292835+0.00009901370713039174 I,0.0002322027258655378 +0.0001147319451794735 I,-0.0001382538534604967+0.0001671093142756853 I,0.0003719449049730175 -0.0000898308292608081 I,0.000458981034224825 -0.00009578238675230326 I,-0.0001223503984961175-0.0001384084603620149 I,-0.00003766163046335994-0.0001982917151555579 I,0.0002764155345198845 +0.000130516582718641 I,-0.0001046284137902115-0.00005687317387520333 I,-0.0002098078722622771+0.0002417369636360672 I,-0.0001593713111926766+0.0004304365025043185 I,-0.00007143686036316339-0.00005704184911432903 I,-0.0001745774659760974+0.0002806487606968821 I,0.000432710921792571 -0.0001184249425205611 I,0.00006206516261977196 +0.0003997999012319145 I,0.0006161259716118497 -0.0005384346614388579 I,0.0006649218766996989 -0.0006848536318537947 I,-0.0005925449133833337+0.0001989437697452964 I,-0.00008869259081455784-0.0005599745751830997 I,0.0002642130834821158 -0.0006840287596091308 I,-0.0005658773990843823+0.0003543252055552374 I,0.0007077373750968677 -0.0003593301497378287 I,0.0008419643726818529 -0.0004427378309575423 I,0.00007795538545748727 +0.0003636496244806591 I,0.0004727668879581219 -0.00007504134586696362 I,-0.0003487836091154531+0.00005608474290845386 I,5.419223692548747*10^-6-0.000204468362391726 I,-0.0005970388705719906+0.0001362512363180129 I,-0.0004004552231517747-0.0001572328578306271 I,0.0002334520398907192 -0.0001063543861332786 I,-0.0005631185015483355+0.0001912191055667159 I,0.0002291523067580106 +0.00001210972617700961 I,0.0001290918488875358 +0.0004473003682004053 I,-0.0003456300115299664+0.0002055184177656246 I,0.0001334701608584644 -0.0002163462055980088 I,0.0004487811574754433 +1.314023294496087*10^-6 I,0.0003047745468966695 -0.0005220666763834749 I,0.0003003756273843436 -0.0003451396338907411 I,0.0003381393558048697 -0.00001313115323726548 I,0.0002073001481098693 +0.00009427694431571283 I,-0.0003500021179035152+0.0004618107542847232 I,0.00008147196068042796 +0.000306771887002526 I,0.0004016648675532329 +0.00018177872126324 I,-0.0005101887882225907+0.0001571538586997476 I,0.0003167275671989645 -0.0002381229044285232 I,-0.0002575446872321026-3.635393560320118*10^-6 I,-0.0005860921550900433+0.0001722233362970821 I,-0.00005447901198363047-0.0003147975118962887 I,-0.000480348748480699+0.00002309507866912213 I,0.0002942807647746701 -0.0002014395177673272 I,0.0001058974517699607 +0.0002058632734446877 I,0.0007344785794241087 -0.0005918648123580757 I,0.0007613747084059879 -0.0003161076125023549 I,-0.0003365656151164865+0.0004082953099012801 I,0.0001746343162960273 -0.0008750780516237214 I,-0.0001884373514432263-0.0002760419604683119 I,-0.0006112742907209193+0.0002348299628632589 I,0.0003647286121288968 -0.0008311116692057737 I,0.0003902050257145223 -0.0005457250591196394 I,0.0001078556282233479 +0.0004066324634478263 I,0.0003588835207935036 -0.0001363222224826937 I,-0.00004724767262920542+0.0004032599984072942 I,-0.0001536271702409554+3.514779650219961*10^-6 I,0.0000157143172450126 +0.0004841614110396787 I,-6.522321568326237*10^-7+0.0002831262685355626 I,-0.0001164721162589323-0.0001131425918389945 I,0.0002487931749662171 -0.00005337970758869312 I,0.00004619309379394937 -0.00007338056715029074 I,-0.0002540886869398135-0.0000633520827668711 I,0.0002779705023730643 -0.0001875770216087792 I,0.0003376950796368805 -0.0001219555416144177 I,-0.0000139116555903063+0.0002718985433506661 I,0.0003512539531793952 +0.0000112442579614318 I,-0.00006595131399095294+0.0002185206090162961 I,-0.00004475922494611231-8.635332687138744*10^-6 I,-0.0002079375331803002+0.0003964886787191583 I,-0.0003629963400804271+0.0002144905162034013 I,-0.00002606107108913536-0.0000613276742537436 I,0.0001824175752993294 +0.0003120670838440349 I,-0.0001390815576109116-0.00008961671319383395 I,0.0001230957147481069 -0.0001254654053579938 I,0.0001344151637798915 -0.00006553000187586706 I,-0.00005889673996083746-0.0001074427825833122 I,-0.00006468804062532545-6.668898116808582*10^-6 I,-0.00003762662875667262+6.808584319404514*10^-6 I,-0.0001342031158713984+0.0002252125420953463 I,-0.0001834354601927075-0.00004309891370276556 I,-0.0001012490306885175+0.0003365812398313664 I,-0.00004720291218872564+0.000188717541594854 I,-0.00001574436537910126-0.0002814959238550973 I,0.00006112194758000781 +0.0002793090055494728 I,0.0002318920445083297 -9.377167607841074*10^-6 I,0.000107231422607641 -0.0001381218976386162 I,0.0001146931580456666 +0.0003546153063872269 I,0.00006423955135490775 +0.0001686874352075441 I,-0.00006505263941977425-0.0001130660706144646 I,-0.00009463291022560933+9.541846335112651*10^-6 I,-0.0001180775810737158-0.0001720132126576048 I,0.00004251251191252894 -0.0001048850201788772 I,-0.000067055268225931-0.0001444310083218539 I,0.00008932386854272216 +0.0000194836719630629 I,0.0001390870982177766 -0.00001991691163884263 I,-0.00006654551232249687-0.0001450416684117309 I,0.0001049507472374915 +0.0002176702162119244 I,-0.00009696262332169582-7.762842003176327*10^-6 I,-0.0001651368389475951+0.0001100350605044908 I,0.00005122956430144663 +0.000206277857196767 I,0.00001904533200356389 +0.00007367239895889085 I,0.0001411267334971629 +0.00006513016107769732 I,0.0001921524196475769 -0.000108351331582173 I,0.0001433612439361494 +0.0001093042897739779 I,0.00007263069745453278 -0.0001177750085490782 I,-0.0001266433135042643-0.00008029317843396915 I,-0.0000578211992719093+0.0001916753361367284 I,0.00005257496075738164 +9.681032877652279*10^-6 I,-0.0002348258483170431-0.00005693167287371999 I,0.00008119128018597093 +0.00001717586673876935 I,-0.000106020985280516-0.0001697491949663158 I,-0.0002542250182725554-0.00007464293648150024 I,0.0000299369749137676 -0.00007018552373772624 I,-0.000168763447791501-0.00002876820505815357 I,0.0001905756140504442 +0.00008887980565706202 I,-0.00006674964514670045+0.0001631284993125237 I,0.0004507489455201493 -0.00003173739130936815 I,0.0004513446181276454 -0.0000351748145902925 I,-0.0003110551100268636+0.00002195045451107021 I,0.0002239040986427912 -0.0003426818732658172 I,0.0001171695414615539 -0.0002317076357051121 I,-0.0003925763764200791+2.644946903255022*10^-6 I,0.0004142742418487776 -0.0002344028972255564 I,0.0004284739357622099 -0.0001389783235558557 I,-0.00003133032749926587+0.0002678170224290416 I,0.0002892052530795851 +0.00003812820765769284 I,-0.00008820696747608879+0.0001781056016696882 I,-0.00006489604974171189+2.116046557100655*10^-6 I,-0.0001125674127639468+0.0002617947338682941 I,-0.00008478427343340983+0.00008872759663421974 I,-0.0001028361688515671-0.0001121868629461162 I,0.0001126760639024886 +0.00007055216969986599 I,0.0001451075724067155 -0.0001305995412540422 I,-0.000156710098328861-0.0000522393689760471 I,0.0002915092609610277 +0.00005475392750124529 I,0.0003850997284787991 +0.00001992949293920973 I,-0.00003713092406822647+0.0001832526691957935 I,0.0003133329619500492 +0.0000524324008458108 I,-0.00007852845488173397-5.137962580700317*10^-7 I,0.00007666030514433925 -0.00001969211653542386 I,-0.0002405577551865626+0.0001379875055744602 I,-0.00036676936128018+0.00003760305328476127 I,-0.00002533629303670887+0.00005507625280775153 I,0.00009650027953181909 +0.0002372320912257148 I,-0.0001434581165492361-0.0001692692089087382 I,0.00006955692500583004 -0.00004751933775744296 I,0.0002244663068233638 -0.00007238060706562155 I,0.0001354344500351898 -0.0001389776501339426 I,-0.0000442228614076374+0.00005297416086086739 I,0.0001256171208405714 -0.00001119172037407308 I,-0.0002119389411862362+0.0000909077316239593 I,-0.00003701195657829774-0.000122478828231215 I,-0.0003476444687861337+0.0002547151095250539 I,-0.0004120878501094546+0.00016267430747033 I,0.0001384961471439093 -0.00009508759632998183 I,8.970099037124981*10^-6+0.0004276852199191428 I,-0.0001063068812030311+0.00003511454941220154 I,0.0003376394344399879 -0.0001151782110323327 I,-0.00008300352463395298+0.0001797132480443105 I,-0.0002840778651449229+0.00003515702609137471 I,-0.00004338840377918175-0.0002292591592856478 I,-0.0002862810079242966-0.00004105152362126764 I,-0.0000689610345658487-0.0000320929858113364 I,-0.00007508708170870073-0.00005371718776075614 I,0.0001146234148747539 -0.00001644334830211994 I,0.000242812136045248 +0.0001160922012594696 I,9.298732384730205*10^-6-0.0001236163946474065 I,0.0001105776657435696 -0.00008544277144548505 I,0.0001806236667872626 +0.0001064704283708592 I,-0.0001081634490732606-0.0001632409911708751 I,0.0001148742771788732 +0.0002082773538269618 I,0.000157226667686521 +0.0002753592929694294 I,-0.0001445466940418255+0.00002402134106544625 I,0.00002961963101366909 +0.0001969186632752214 I,-0.00008328402108386592-0.0001341562256829676 I,0.0000799142848419811 -0.000031059426210696 I,-0.0001843361572818508-0.0002499572847151639 I,-0.0000626152695771382-0.0002316063291186484 I,0.00009586952329848327 +0.000087691792880715 I,-0.0002103707674290287-0.00009593457232201277 I,-0.0000159028213753823+0.00006131811131014492 I,-0.00003189096022319796+0.0001277588633982856 I,-0.0001713767303837376-0.00006234202723191329 I,-0.00001992633270557205+3.653441997962509*10^-6 I,0.00007131162615642365 +0.00006363513834330923 I,0.0001021539832142284 +5.05633206126515*10^-6 I,0.0001833094901502337 +0.00001178550890944732 I,0.00005128281480053449 +0.0001348010556686308 I,0.0001378666448969203 -2.370341474029492*10^-6 I,-0.00002341208603023957-0.00008746422213121581 I,-0.0001871123266850635+0.00009210609906824021 I,0.00008011812302201186 +0.00001820754684778191 I,-0.00006015028444146959-0.0002535365989730523 I,-0.00004855379915524719+0.00004519365341520049 I,0.0002360354586565349 -0.0001228451519967933 I,0.0001673526705138612 -0.0001756537328897732 I,0.00001935635762921747 +0.00005908285984851136 I,0.0001025334830554153 -0.00006730192939008038 I,-0.00009113967982597065+0.00002749580634046613 I,9.714930054085527*10^-6-0.00004740519675682509 I,-0.00007722584812984217+0.0001090066004475413 I,-0.0001188956575281733+0.00009183388313603106 I,-0.0000184759839478666+0.00002195231118267527 I,0.0001223805587924184 +0.00006969461826637896 I,-0.0000834366011939278-0.00008827722458491671 I,-0.00001536004962725957-0.00006364092413817799 I,0.0001998685628955264 -0.00007973971797259102 I,0.0001261979458529329 -0.0000803153576541854 I,-0.00006519392808651564+0.00005669502129165035 I,0.00008155341533448544 +0.00003547039733916961 I,-0.0001775130031766638+0.0000698232679739411 I,-0.00003045534289569261-0.00008665687047613045 I,-0.0002556402810110583+0.0001138942337751173 I,-0.0002362233938343128+0.00002098134902816899 I,0.00007266918188659967 -0.00003101573701751236 I,-0.00004031295765838275+0.0001917088697339075 I,-0.0000736107414950624-0.00005674605895097812 I,0.000120846241832849 +8.875566970373684*10^-6 I,0.00007226283456311216 -0.00003759951889167486 I,8.937794531843624*10^-7-0.0001625984209507136 I,-1.288392525018679*10^-8-0.00002868838007609524 I,-0.00001084335844232754-0.00008351703308864773 I,-0.000225576724004016+0.00008885323350006584 I,-0.00009582976135069383-0.0001736365453442993 I,-0.0002890084571764545+0.0002417960311043678 I,-0.000222888125853167+0.0002262539396778927 I,0.0002347997580208294 -0.0001972140136400706 I,-0.00007445647397293017+0.0002971430764462397 I,0.0000631164394205981 +0.0002395378018928158 I,0.0002670513301125631 -0.0000937895999533924 I,-0.0002644185209487163+0.0003164523813755691 I,-0.0002677431219142931+0.0001857365124090814 I,0.00001406479870484938 -0.0002036928790800178 I,-0.0002083334929170936-0.00006265632527336287 I,0.00008279686902588754 -0.0001541706943637216 I,0.0000462663777281356 -9.28760216015035*10^-6 I,0.0002027812105462302 -0.0001001257257436493 I,0.0001781268546551771 +0.0001713350310095739 I,0.00003735950615943229 +0.00002623572728769816 I,0.0002310381011697216 -0.0001236866108994604 I,-0.00005075691122777238+0.0002033786774862642 I,-0.00005427608051908545-0.0002546291345529884 I,-0.00001471698019768556+0.00006737574678616448 I,-0.0001085072037121951+0.0003240789423692495 I,-0.000216468351278217-0.00002354516854184588 I,-0.0001382615633337311+0.0003082418918485146 I,9.242035483963658*10^-7+1.25254175039954*10^-6 I,-0.00002716141582438403+0.00007482263576819113 I,-1.559951919237706*10^-7-0.0002770971765419016 I,0.0001874853938087989 -0.0003381444620904655 I,-0.00004893062876782842+0.00003866407120078279 I,-0.0001865364129133546-0.0002391706926860495 I,0.00008548605656886717 -0.00004191354279647953 I,-0.0001987753733920037+0.0001382404872900578 I,8.65743984128218*10^-6-0.0001708950279097963 I,0.0001578991395766893 -0.00006526948057962478 I,0.00001572956788131042 +0.0001921807723808381 I,0.0002164846368421632 +0.00008674918621980518 I,0.00002487774015877712 +0.000200348654313669 I,-0.00008798853967261822+0.0000689548432702945 I,-0.0001699493563890073+0.0002450354950024148 I,-0.0002288689663629786-0.00005299712343913773 I,-0.00006737214812265266-0.0001771108410582169 I,-0.0001220721858270345+0.0003585067679696011 I,0.0001819669102185467 -0.000137616336258792 I,0.0001826086148153276 +0.00001809132679051785 I,0.00002156048815968206 +0.0003261659608667683 I,0.00003201665781204751 +0.0001046668295826997 I,0.00001990046310258704 -0.0001135780804171993 I,-0.00002781290303261136-0.00001747152162253097 I,0.00001947409221902328 -0.000206520425333741 I,0.0001194259794259066 +6.165310020111152*10^-6 I,-0.00001847843995277581-0.0001770102658861432 I,-0.0001019016507476876-0.0001357722717830501 I,-0.00002510913131623978+0.0001117862127472334 I,-0.00006970483249881161+0.0001035085500410215 I,-0.00008781347586402067-0.0001548925234766874 I,0.0001518309967544515 +0.00007196805256198737 I,0.00005100154047248476 +0.00002478153077414662 I,-0.00006544378392749252-0.0001240695651913164 I,0.00003733296596232532 -0.0001527416530659076 I,-0.0001700490334925902-0.0001239187698792379 I,-0.000118143028217489-0.0001444888551801305 I,0.00003531524301904842 -0.0001178042157966123 I,9.885920783275532*10^-7-0.0001781682029868644 I,0.0001295822795952741 -0.00002149715650789127 I,0.0001239716760334103 +0.00005976698589134431 I,-0.0001081419228641855-0.000188430192665184 I,-0.00001361279219688446+0.0001111904190181974 I,-0.00004090942143635256+0.0001073491146741912 I,-0.000127854603640169-0.00007854837335859551 I,-0.00003524366478032987-0.00009502806900367586 I,0.0001011758172034967 -8.990469585604973*10^-6 I,-0.00001704090175253805-0.0001482922443386105 I,0.00003062011571695298 -0.00004486930687783665 I,0.00002382632346332493 -0.00005306209931329344 I,0.0001132784589738157 +2.425043487729793*10^-6 I,0.0001424811790416915 +0.0001869235935998284 I,0.00009333462957748101 +0.00002185386857220858 I,0.0001172498270706316 -0.000221404077830697 I,-0.0000124074128929242+0.0002515556749962698 I,-0.0001587529874550436-0.00009509291297476238 I,-0.0001284317852670404-0.00003639884422978131 I,7.279612477288189*10^-6+0.0002024952385704892 I,-0.00005021001266686069+0.0001184748307354312 I,0.0001230636069904102 +0.0001562279677326764 I,0.0001678626662488987 +0.00001206549947686223 I,0.00007459006113154604 +0.0001327345643021599 I,0.00005053467851146595 -0.0001001601131820493 I,-0.0000820518958669186-0.0001618560026710655 I,-0.00008841880407352715+0.000175318563603244 I,1.71966689062786*10^-6-0.00006326259410804476 I,-0.0002202753192251358-0.0001097967054024348 I,-0.00002529363504141286+0.00008709114432015206 I,-0.00002763690305713492-0.0003358221173840655 I,-0.0001078888004751251-0.0002236591778686397 I,0.00004356751618887354 +0.00008358226882103101 I,0.00003680924026659639 -0.00002289230161904975 I,0.0000793538167180904 +0.0002903278341176519 I,-0.0001547425529993339+0.0001070476476966875 I,0.0001049836223404863 +0.0003122783735412646 I,-3.139748448821133*10^-6+0.00007199707256407573 I,-0.0001959114496578105-0.0001374138876062677 I,0.0001048578068085859 +0.0001983209864269862 I,0.00006145075350323269 -0.0002441507393460733 I,0.00006179143584624663 -0.00001008389087607252 I,0.0002379201214702102 -7.537509134753352*10^-6 I,0.00004724036451205356 -0.000222020618053061 I,0.00002613465741836473 -0.00005192150973499661 I,-0.00007957757333820309-0.0001728830216887471 I,-0.0001883507834817204+0.0001238490768603876 I,-0.0001640000265274905-0.0001470462782183875 I,-0.0000110864049874392+0.0003421842022971574 I,0.000124775054576373 +0.0004320864511871989 I,0.00009352044440971496 -0.0003059020543987088 I,0.0002423005412864553 +0.00007835208378486198 I,0.0002519858211967064 +0.0002742692130341774 I,-0.00005286539797174408-0.0003371002695173507 I,0.00005238672779325602 +0.0003717381910971833 I,0.00006146523864998803 +0.0004580186509882984 I,-0.0002242240140888256-3.08272293085649*10^-6 I,-0.00003638005932330904+0.0002956084767464288 I,-0.00005478471801210501-0.00004210269401653716 I,-0.00002853047044267773+0.00001291881377902466 I,-0.0001427494714647935-0.0001132135924512352 I,0.00003269950335317752 -0.0001500485993845794 I,0.00004406028056279403 -0.0001419180177272885 I,-0.0001601155805503783+0.00004183389335682186 I,0.0002734457612091598 +0.0001157705453202756 I,-0.00001241445795190647-0.00003781411972916726 I,-0.0001200869614097758+0.0003545163545093051 I,0.0000574738525064193 +0.0003686856883935467 I,-0.00004813701210262729-0.00001179961784983358 I,0.00003927130816582772 +0.0001724007968947044 I,0.0001797211356846055 -0.0001987258049913982 I,0.0001355259179958048 +0.0001400706834700773 I,0.00010058614698119 -0.0002800382572752955 I,0.00005972966494082344 -0.0002647640310875845 I,-0.000141973574829588+0.0001034282113513527 I,-0.00004452717511097156-0.00007889587291066896 I,0.00009494809395992228 -0.0001414291917935327 I,-0.0001127832722802935+0.00002605201248206824 I,0.0001260354050938084 +0.0001201182833311428 I,0.0002468783349504577 +0.0001854037392370363 I,-0.00004797551680322051+0.00007667574936629762 I,0.0001800470699887128 +0.000163058425069545 I,0.00007870892648630784 -0.0002370963221342789 I,0.0002436932395912784 +0.00009378477142738854 I,-0.00009985230143285599-0.000385921243027117 I,-0.0002495520904804782-0.0004314601670067739 I,-0.00009068429797212072+0.0003805213009398864 I,-0.000151870484307035-0.00003855398891026599 I,-0.0003692262932423797-0.0003288705800351803 I,0.00005254649605545246 +0.0002493952349413104 I,0.00008243341530589794 -0.0004684735588256355 I,-0.00005663336957047134-0.0004670996171059571 I,0.00009445181548109531 +0.0000378661690710101 I,6.652007216325793*10^-6-0.0001304817301629929 I,-0.00009567805270243717+0.0002485014313973577 I,-0.0001575634271218796-0.00001997857078996131 I,-0.00006582735280935738+0.0002838421730910991 I,-0.00006442911365851019+0.0000405750307572099 I,-0.00007714024161739028-0.0001705005670881918 I,-0.0000650280157574044+0.0002388230205794979 I,0.0001480013600291315 -0.0002119295475325379 I,0.0001160531195461533 +0.00007767890227204467 I,0.0001834695109672447 +0.0001468401884665798 I,0.0001303324346197749 -0.0001794173567668981 I,0.00007992216738121258 -0.00009553348662101351 I,-0.00003261228626467874-0.0002182542455806676 I,-0.0002026922532985411-0.0001231534875071433 I,0.00001901209186811723 -0.000198711392369702 I,-0.0001316708620477275+0.0000655538289102564 I,-0.00004826057388605115+0.0002336808568058193 I,0.0001903077614222255 -0.00007590654524493994 I,0.00002967626980777694 +0.00005571008366785714 I,0.00006260853294328197 +0.0001969166585613 I,0.00006197443924441174 -0.000101761483592713 I,-0.00005666337709959025+0.0002399358429621029 I,0.00002455625628584141 +0.0001967821801607015 I,-0.0000402299277771011-0.00007735409081617457 I,-0.00002500202800301803+0.00002066562099148787 I,-0.00006611058491909875-0.000265610329029627 I,0.0001765999981341349 -0.0001067558058965696 I,-0.0001506670153986214-0.0003041800379816296 I,-0.0001463270088353166-0.00007847971043986135 I,0.0002346918527526193 +0.0002371811622225803 I,-0.0001197236348458282-0.0001122947436651106 I,-0.0003159518432761486+0.0002025840850418071 I,0.0001150512050508696 +0.00007740555850354747 I,-0.0003298785153213222-0.0002443302371102472 I,-0.0003545723539055837-0.000117446728887374 I,0.00006053951102182113 -0.00007789490133844265 I,-0.0001754459406037538-0.00007875616708796822 I,0.00006630050435925145 +0.00009539179757280436 I,-0.00007558453081134151-5.038839114540589*10^-6 I,0.0001480202721565799 +0.00001689784479034728 I,0.0001892629692434882 +0.0001106001040679357 I,0.00009702820250903269 -0.00001504687078570798 I,0.00001426240416654834 -0.0002019237476139484 I,-0.00005287474906309021+0.0002778102925371023 I,-0.00006647774110094402-0.00002560303391760166 I,-0.000311955511005293-0.0001479869509830375 I,-0.0002662473098541056+0.00007267502269230956 I,0.00001253670170639091 +0.00005224901733779847 I,-0.00004607781308890865+0.00007054703761921415 I,0.0002462165296956149 +0.0002502070969444183 I,-0.0001506887443394748+0.0001845082848550187 I,0.000271783564897621 +0.0002172958224991984 I,0.0001835126985283962 +0.00009842683892300257 I,-0.0001643740427186401-0.0001726846579319228 I,0.0001501045146167506 +0.00005665441939541976 I,0.0001437343899781267 +0.00004663770170178929 I,-1.624356994732532*10^-6-0.0001782624869059773 I,-0.0000695694192765556+0.0001696846797539661 I,-0.000201388893010449+0.0003132666457286881 I,-0.0001345206353245516-0.00005612384006241846 I,-0.0001959265514117169+0.0002731691706578942 I,0.0002560706193280667 +0.0001203313869431101 I,-0.0001371724832809137+0.0003100713546235155 I,0.0004209526226459671 -0.0001225126818462572 I,0.0005104897733230232 -0.000397300102349589 I,-0.0005091994004465192-0.0001487958749857936 I,-0.0000382378031600981-0.0001324359969010345 I,0.0004943817231147176 -0.0005617308603764581 I,-0.0003001690004637066+0.0001407494315623808 I,0.0006535664586618151 +0.0001645562454534097 I}",
                                     0.33),

  getRingNoRingSpec: function() {
    return {coeffs: RotorAnimations.RingNoRingCoeffs, freqfn: function(n) {return (-1/6)+bount(n+1);}};
  },

  RingNoRing2Coeffs: parseRotorCoeffs("{-1.08894801374845-1.860487444784657 I,-0.2058587366063385-0.1182019825124719 I,-0.2006974042322546-0.3058662275557784 I,0.2344617259838231 -0.03874713279707742 I,0.02789551886742797 -0.1189663723797991 I,-0.1550048348521057+0.05386962529564474 I,0.01795723005715386 +0.0844467401848845 I,0.02183468201701029 -0.0890628997829579 I,-0.03915821820149327+0.01922530358980329 I,-0.02959740060321513+0.05157358512111791 I,0.008569875955063158 -0.03697653589520697 I,-0.0249431189863796-0.01303260426716138 I,-0.03197297913580943-0.05824352391361026 I,-0.01030285943760946-0.04873079552003316 I,-0.02254473654444367-0.00306166051145506 I,0.01180206718852715 +0.03325539875777584 I,0.01836903179449244 -0.05066131738011864 I,-0.005019094931696763-0.01192959573232555 I,-0.04901802835491569+0.01598136123938382 I,0.03363192574484172 +0.01751514283565131 I,0.002217662651669831 -0.01301448297603531 I,0.002803858244974367 -0.02729408181222922 I,0.02760046661280292 +0.01382343054313542 I,0.04179107494759712 +0.0006921830264210761 I,0.01425880599223981 -0.01010012784910207 I,0.01177221080389566 +0.00943338348114609 I,-0.03350740217710634-0.02560216471828738 I,0.01007455554387568 -0.03099333282837806 I,0.01620342236341015 -0.01694522341447063 I,-0.0207868087181203+0.05042569314744383 I,-0.0007870081992556732+0.003092229237386196 I,-0.02402406333064184-0.03100775709874908 I,0.02183217400032236 +0.007668623158303672 I,0.0003286160965417055 +0.00798510603259886 I,-0.02466335280071947-0.0008880947615910097 I,-0.02140679195417088+0.01282395578693696 I,-0.003769323847410026+0.01564523039917962 I,-0.01121541432771065+0.004217729401310152 I,0.02166823494707272 -0.00212566761638801 I,0.01314511903372661 +0.01007804647465801 I,-0.004204219064705778+0.01030298889101204 I,-0.01687904253506084-0.01930058622102826 I,0.002303422299288836 +0.005031870900860306 I,0.01912482356236366 -0.006497494452433762 I,-0.009224506861042325+0.006661804015869384 I,-0.004452367326290303-0.003052913857489311 I,0.007496483802257047 -0.0104189024929106 I,0.003824193545648894 -0.01235961758900168 I,-0.005253441029968669-0.001909005878659069 I,0.0001982757705077214 +0.0005000910500023228 I,-0.006307401349867785+0.008898203826346294 I,-0.004446835014442697-0.007293700846402967 I,-0.01140285037615735+0.001268026335265045 I,0.01106465443836909 +0.004836178334529457 I,0.01458217682671588 +0.003819132749956468 I,-0.01263812140430923+0.001483492582785464 I,-0.006868416123939194-0.005042592586438949 I,-0.001392089559582347+0.01189110236906974 I,-0.0008598922839153245+0.002971491446287122 I,-0.003288672995640519+0.006803714486544607 I,0.000468953315619372 -0.006983089047018158 I,-0.001145284027829254-0.003391616292756579 I,0.001052862989936568 -0.007765824375697577 I,0.006574285950882814 +0.002385680879117768 I,0.00384034570303159 -0.00007490753647590616 I,-0.009465559633534041-0.002209826822743572 I,0.002832824259477567 +0.001524549723736093 I,0.002832587480961882 -0.005436399038433543 I,-0.006106810581915617+0.00350616740484351 I,-0.002614887244108258+0.00752531689939422 I,0.001408252028281392 -0.005400259507823292 I,-0.004058461627273928-0.002091320643666604 I,-0.005287080119078161-0.009908827901456495 I,-0.001454062232174442-0.008606715807580961 I,-0.004304188374017068-0.001754306860295223 I,0.003528561623066663 +0.005324818644010034 I,0.005488068757422993 -0.009769979701175989 I,-0.006995249897497238+0.0009445792147429136 I,-0.009365652573349424+0.004156656041974676 I,0.007319111063607915 -0.006113792158431071 I,0.002168605261361365 +0.001821968304226707 I,-0.004144171922699736-0.003032743197909723 I,0.0009159760711363058 -0.005682672121540855 I,-0.002240006068590813-0.007839664161380717 I,-0.002526421276732197-0.001846900424482278 I,0.0009925833060932133 -0.003536805566095726 I,-0.002059576299802904+0.007252612116628843 I,-0.005950707439078591+0.001828639874115761 I,-0.00314130938163265-0.002346628524259526 I,0.01018473090427817 -0.001021853953357202 I,0.0005541781652030255 +0.00012232219471484 I,-0.004111239094057263+0.006962217161751251 I,-0.002134839065433335-0.005140874887755599 I,0.0004300039462370436 +0.002777437347930532 I,0.001994773898045164 +0.004565749784986827 I,0.001651559989825637 +0.004259921682054335 I,0.003169788032604501 -0.005984985506647038 I,0.002305008209473479 -0.003183946735894358 I,-0.002126233219009567-0.007761800959113297 I,0.002240865728009053 +0.002457879315406645 I,0.01079443996728528 -0.003306973694755013 I,-0.008525548446519166+0.001598670491490091 I,-0.00805726821916419-0.005187521729146183 I,0.004201508302489655 -0.002462450134624777 I,0.003992265314026505 +0.006678881057091908 I,0.003961410874332688 -0.00165885970116974 I,-0.002759023166248603-0.003794637547402125 I,-0.003111661809319751-0.001503857625312254 I,-0.007773126260188152+0.002466587849304207 I,-0.005281197357011483-0.002471977034296288 I,-0.0000217564892201769+0.007399132868947031 I,-0.0005006128714956874+0.003937564686433233 I,-0.006405536079907837-0.005722064484362731 I,0.007675284126132882 -0.003316413816669499 I,-0.00384736732558025+0.005156944718105524 I,0.00126836427203523 +0.0062597011916423 I,-0.0001591290152360026-0.006133372860722085 I,0.0006236534720673592 -0.003016835226069262 I,0.003040808137806855 +0.00517539841199196 I,0.004834982653620321 +0.003885402247582224 I,0.002525208374103413 -0.0005728561383216715 I,-0.0005428327374812063+0.000950876558497096 I,-0.001484657096580419-0.004550590027482924 I,0.004726617768454603 -0.00002423602679413849 I,0.001071099066289043 -0.004373011201434255 I,-0.004670027582426244+0.001522429483933192 I,-0.005067915044328063+0.005857090355002504 I,0.004362866587808751 -0.004162853516295674 I,0.002665606034795101 -0.002440075738281898 I,-0.00450938519896545-0.005035205558323319 I,0.001086049970003676 +0.00001313079792091346 I,0.0009888292586984057 -0.002853888300597109 I,0.004794524376834247 +0.002020709877033757 I,0.003075554578759075 +0.002139923450660389 I,0.003029879205303142 +0.0003079978503194732 I,-0.002722960031704673+0.0008667135246095219 I,0.0005290026608759475 +0.005436713834122805 I,0.001887232728814642 -0.006188405193195487 I,0.0009883017534969849 -0.003696946306953036 I,0.003416763286266089 +0.00626806386566707 I,-0.003394549588712086-0.0004031244223015314 I,0.002034914926887687 -0.002296296116613042 I,0.005043309286737643 +0.002180077351886082 I,0.007036543974215504 +0.0002133288443001381 I,0.002618142229724205 -0.00142432010928382 I,0.002488333398381184 +0.002394505074163873 I,-0.003401905134923381-0.003767315109566641 I,0.0002180947668960157 -0.004368339310118221 I,0.002135683334516246 +0.00255594581698601 I,-0.002227889235837662+0.001637128503749514 I,0.00358242234348158 -0.003824488830810277 I,-0.0008248998997257293+0.0009227050159472203 I,-0.002867196346498719+0.003255454903437298 I,0.003444774294245645 +0.002929040877036855 I,0.001686148457129375 -0.0008461474017188142 I,0.002412056149270986 +0.001088820168413747 I,-0.004460418744058892-0.00362013551163561 I,-0.0024791096576747-0.002245849801235081 I,-0.006056041331324051-0.001302172706527638 I,0.003983719253551133 -0.001323509462012114 I,-0.0004798727886568847-0.008417972526487275 I,-0.002697128285806176+0.005128326207363565 I,-0.004551812734054961+0.007128762642421306 I,0.00279256390145352 -0.004556963837441694 I,0.002043570675513981 -0.00274410159440177 I,-0.004629495129362301-0.004347632721444042 I,0.003215695827516887 -0.0005777469769762842 I,0.004390548351259953 -0.004939634250900727 I,0.004405025733585536 -0.001841383629816269 I,0.005555909673983093 +0.001129219038137044 I,-0.005996546426146793-0.0007520915764500781 I,-0.004887132201250121-0.006455954952460227 I,0.003152920154183806 -0.000443507209958159 I,0.00364936468616465 +0.01040974816377083 I,0.002196548068987055 -0.004816232515488917 I,-0.01060307141991491+0.0001979824017584462 I,0.004313658725836695 +0.0005886685225460916 I,0.004517300260984266 +0.003798162173733312 I,-0.005661435702388216+0.003817415159636359 I,-0.002563035402212365+0.00801424650161595 I,-0.001085765991513899+0.002650795632577763 I,-0.003722367027977212+0.001181241258637255 I,0.002132005588868905 -0.004603065868356308 I,0.00686411269357793 +0.0002178901380870645 I,0.0007016209599778455 -0.002376972558431357 I,-0.008116054227562449+0.002403748790967353 I,-0.002073553003269528+0.005425683551521073 I,0.002735611236241318 -0.008436078756892364 I,0.00229360887561373 +0.0004544457916664407 I,-0.005101543282096359-0.001648133155937841 I,-0.001292520762016399-0.002898076971819529 I,-0.002578170537058001-0.00393978586937498 I,0.002616120261105071 +0.001041322176339943 I,0.002881485377233296 +0.0005732524274609174 I,0.002198519295920913 +0.000912069474013448 I,-0.003377699183570817-0.0000522250553282263 I,0.002435090217500101 +0.00472153460132935 I,0.00008252079975557686 -0.001279898991943348 I,0.001865297677647274 -0.006040400802100806 I,-0.0009643491416014664+0.001719775107230709 I,-0.0006434934385656649+0.003799563489454336 I,0.003674099930605539 +0.001776008210676971 I,-0.0009516276838119077-0.001237007388655722 I,-0.0003964862808574993+0.0005150629804348222 I,-0.004181469017129808+0.00002648606317657036 I,-0.002337265798590029-0.0008644248369096392 I,-0.002517073167828593+0.001718681807219449 I,0.0006564564290930811 -0.001366230861565312 I,-0.001896421890943781-0.003834262636032502 I,0.001361305320711796 +0.004920165020864062 I,-0.0003240617407128128+0.002419346516221291 I,-0.004006808902106915-0.002443374250145718 I,0.002975805914190785 -0.0009219616272616417 I,-0.00009754934078720547+0.001061633573769153 I,-0.003701943341029323+0.0006955158826992226 I,-0.003461799571577175+0.002260429021046745 I,-0.0008092939913751742+0.002626662808491087 I,-0.00222158438959687+0.0003900319180035253 I,0.003085738407182812 +0.001071135777962702 I,0.001359891880833592 +0.001478925981280181 I,-0.00275162402891955-0.00004270098810545608 I,0.0009065809298436037 -0.001085041927479065 I,0.0007046690382828957 +0.003190656723924583 I,-0.0008661660123528894-0.000822226612366059 I,-0.00003994727885708004-0.002567353181602363 I,-0.002384514377092392+0.0006536848607011476 I,-0.0008174555191768043+0.001153789455400374 I,-0.00182980486631742+0.001096159450843318 I,0.00287395516264281 -0.0001195021272686939 I,0.001316538978866177 -0.0001320355398797737 I,0.003274929426907113 -0.002027492734372618 I,-0.0003200963431605431+0.002488077928507767 I,0.003334863554132575 +0.002858568031828848 I,-0.001714176020175934-0.003650623466307984 I,-0.001474852208455249-0.003993738520067228 I,0.001931944195827762 +0.002630807263086145 I,0.00001370726892762362 +0.002028571691934164 I,0.003340958605667244 -0.0007996821583280845 I,-0.0001847499479066322+0.001061342432945694 I,0.0012751392716514 +0.002756248420661739 I,-0.001254167351315911+0.002086207210802255 I,-0.00236849144670613+0.001778841598890408 I,0.0009435066604500473 -0.001725067201740555 I,0.004095860645354193 -0.0007200513342203365 I,-0.001366787331040446+0.0001127529220282257 I,-0.004657497496403779-0.00002116789890373953 I,0.000984614824977026 +0.003273556282815941 I,0.002026266254658136 -0.004824584484918701 I,-0.0005054990607284357+0.0008842921743366413 I,-0.002935947895947695+0.0004460803950203287 I,-0.0006316533810459646-0.00271020803468609 I,-0.002737287575430548-0.002449198433476662 I,-0.0002757136681246922-0.00008687280027133227 I,0.0001986011404611471 -0.0009994612712709002 I,0.002318569051682332 +0.001729521291442657 I,-0.001138916724370649+0.002646934148598174 I,0.0001893675122942372 +0.001786440362638492 I,0.0003182295973121643 -0.004658749544608826 I,-0.0009327684289315808-0.001679772942612919 I,0.003433695880555857 +0.00272302895282996 I,-0.001441697354105618+0.0009247014307558197 I,0.001442493032972829 -0.001646602490374057 I,0.001978420518419369 -0.0002676372154431703 I,0.002498867798527861 -0.0006846097666081414 I,-0.001083414388647491+0.0005222749177934362 I,-0.0007779771947127533+0.001050460771733149 I,-0.001290884639400573+0.000810537869701524 I,0.0007932737060346328 -0.001745173935521318 I,-0.002782535597442091-0.0002058328196096336 I,0.0005413238553565554 +0.0006892622081211724 I,0.002358707329598804 +0.002547069835535176 I,-0.0006230713535458799-0.001152464111954112 I,-0.001493475858381699-0.001300999509679588 I,-0.001949184750523251+0.001161895195988339 I,0.0007946297792678046 -0.0002125258230029927 I,-0.0001519103870808989-0.0005217485800595289 I,0.001183397922442897 -0.001852936721013149 I,0.001147151533236379 -0.0007766188619205375 I,-0.0001095528066716394-0.00145918542680151 I,0.0001934769738453569 +0.0006126114083787257 I,0.002270579186081805 +0.0003142975276484857 I,-0.002338723637610855-0.0007923034621298629 I,-0.001117370849362714-0.001120581356421872 I,0.002263228331266164 -0.0006908385517971972 I,-0.0002461480486051285+0.001633200530604731 I,-2.486323449419486*10^-6-0.0008126187411017843 I,0.0008483910790437078 -0.001550045242507212 I,0.0006299919448066265 -0.001992185004328041 I,-0.0008307546231121877-0.0004577589736507886 I,0.0004064723141089344 -0.0003203579989747637 I,-0.001457755583528197+0.001084749810546802 I,-0.001036970402629374-0.0007552508274861665 I,0.0001066361192524146 -0.0005384731387945002 I,0.0006157491225392479 +0.00117725165592588 I,-0.0002517866740826386-0.0007790822833907318 I,-0.0004960737879178322+0.0000333751003973419 I,0.0007486657622114853 +0.0007371082086135088 I,0.0009391099820930762 -0.0003206389605385565 I,-0.0005744103672967234-0.0002607431990279971 I,-0.00004838567803255043+0.0000373421580096376 I,-0.001091937268372187+0.001232566912215587 I,-0.000976876798101873+0.0006553187337125154 I,0.0001339691139551594 +0.001067895590269033 I,0.0004158858111552354 -0.0006390386515714363 I,-0.001906000249301258+0.0002467126878981398 I,0.0001461334063446907 -3.218823569289516*10^-6 I,0.001417636793041397 +0.001163380618199682 I,0.00007662815489299169 -0.001000992566703309 I,-0.0008098121427565325+0.0001847707991337944 I,-0.0007881295990221455+0.00057755875601459 I,0.0001567296557334172 -0.001371537391648644 I,-0.0007430275206124727-0.001623330376262863 I,-0.00134731700656395-0.0005392412450891566 I,-0.0003997093102231322-0.0009813020526929042 I,-0.0007182459196665067+0.002491208737518124 I,-0.001308689420205049+0.0004478421612175974 I,-0.002096632402956023-0.001424096950250878 I,0.003581388095075572 -0.0001604775521313934 I,-0.0003923279130788149+0.001511123438455992 I,-0.0006486886145548681+0.00206240396601593 I,0.0001385751693680172 -0.0026346100216081 I,-0.0003916226817673309-0.001372099280588693 I,0.00112128246449198 +0.002130845385842255 I,0.002341510305468241 +0.0008027836469700507 I,0.002692041850548098 +0.0006789025921740274 I,0.00140511509271528 +0.002122205957183427 I,-0.00009873545018488587-0.001803819866495712 I,0.00007896295734212367 -0.002572497422040094 I,0.00005040278546313698 +0.002624495979609575 I,-0.000975168658316141+0.001370144783369139 I,0.003678237014106017 -0.001029871800048344 I,-0.002041154601108675-0.001493745417838818 I,-0.001300350756511742+0.001213868560875162 I,-0.0002398336882337461+0.002720032272901578 I,-0.0007179405925230842-0.001328273371248779 I,-0.001719971564728542-0.0005522019076863139 I,-0.001366388222173563-0.001371399370124669 I,-0.0002127085016723849-0.001455932324623643 I,-0.0005360114476297853+0.001108221043398814 I,-0.001010728801049885+0.0004699623210912619 I,-0.0002746753706215519-0.0007993594711808021 I,0.002012518147614445 +0.0005231227421390386 I,0.0002019009273618016 -0.0004156133317474431 I,-0.001819979990117883+0.00123772702742347 I,0.0002098576916332766 -0.0006131993219178974 I,0.0009028099558327717 +0.0008731887347693203 I,-0.0007377373279624552+0.001082931824364494 I,-0.000394706044848482+0.001716546578662008 I,-0.0002224753110644665+0.000416056265999475 I,-0.000832843186613618+0.00025206254506457 I,0.000728165462904537 -0.000368454693029897 I,0.000839269006127908 +0.00009410655506958692 I,-0.001027620215816462+0.0003451648011167884 I,0.000276749279212841 -0.0004487958444781759 I,0.001449356061282254 +0.0008014115565661587 I,-0.0008810737536287796+0.0001513870216529555 I,-0.001130082512536548-0.0006926193552807444 I,-0.0002845696904521747+0.001570651373170891 I,0.00001925745674324908 +0.0004599020832901858 I,-0.0005027048963171408+0.001106692124889153 I,0.00005345039122519029 -0.001185060467905342 I,-0.0002958482371468139-0.0007466668593236777 I,0.0002149078138333764 -0.001240861128502852 I,0.001176748276511429 +0.0008457259494511259 I,0.000668418960501523 -0.0006156731890349459 I,-0.001334617549189423-0.0005422673169958273 I,-0.0005125039296423862+0.0007981538346998015 I,0.0009096361117603765 -0.0006434619043125585 I,-0.0005004090026513711-0.0001540231434139531 I,-0.0008239419188810562+0.0002650158001984055 I,0.0004763357983275868 -0.00033787239198871 I,-0.0001563170055639837-0.0004263218238375519 I,0.0002624681921756409 -0.001198291169259518 I,0.0004755446230340675 -0.0008669009231715345 I,-0.0003548259702702632-0.000465098532348826 I,-0.0001719702181477768+0.0007193843740526856 I,0.001219087048633519 -0.0007021316698894065 I,-0.000165724715957436-0.0002125718061729764 I,-0.001410152473520296-0.0004623567749918011 I,0.000576274925174646 +0.0007625972217255957 I,0.0004491794307823161 -0.0001389066044280036 I,0.0005689596362755884 -0.0008315144811874872 I,0.0003683966291116413 +0.000806601329685256 I,0.001164381540529646 +0.000759652515553278 I,0.0006390625237390124 +0.0003763890044350166 I,0.0001086478035576791 +0.0008668732032458876 I,-0.0002933614021980052-0.001343303892500448 I,0.001033993781994826 -0.00107975285008437 I,0.0004077524688816448 +0.0002610367246593023 I,-0.001786506285008283+0.001223789732754699 I,0.0007000764899441126 +0.0003564102169450957 I,-0.0004120778798008718-0.001964168317373021 I,0.000207393007305684 +0.0006199112061686545 I,-0.0006728675228566991+0.0007439902768895962 I,-0.0008508704796691775-0.0007290576765867705 I,-0.001364895092382839-0.0001722790290334859 I,-0.00006520463832379673+0.00009380704519781793 I,-0.00005637600916694662-0.0002428149902839562 I,0.001134356045303179 -0.00001825969797580158 I,0.00009748078029787584 +0.0008695933475703506 I,0.0005082667398992298 +0.0009814311664166158 I,-0.0009729031282794147-0.001544706045050807 I,-0.00001768513270215431-0.0008140410924388714 I,0.001366926179027586 +0.0001113090320391634 I,-0.0006511975035629416+0.001066239746790993 I,0.0003621956720974847 -0.0001636917470582693 I,0.0004523084336700242 -0.000902454502426986 I,0.0002781339718122278 -0.0008895182444122907 I,-0.0009897464639336373+0.00006119606030179883 I,-0.0003423535186511794+0.00005699217153019701 I,-0.0005412711062994496+0.001179746146011178 I,-0.0004425301488280515-0.000666957018926536 I,-0.001578444848427626+0.0001481692936604858 I,0.001543114846595089 +0.000279983737264582 I,0.001795679696294352 +0.0004737671345575246 I,-0.001518173212938092+0.0004972242118337964 I,-0.001000678674423624-0.0007250106678165517 I,-0.00006028594045738578+0.001531807048494965 I,-8.431087063966863*10^-7+0.0005218339986144112 I,-0.0002899845432534341+0.001038297016107366 I,0.00009454649803604344 -0.0009799500140098183 I,-0.0001639131306827549-0.0004297686692525597 I,0.0001212371899652191 -0.001212704469434087 I,0.00101482044880313 +0.0003008869592986308 I,0.0006005757710206107 +0.00003064231868721117 I,-0.001499367744568626-0.000332630149882759 I,0.0004765007228592698 +0.0001910598473184252 I,0.0004443657453230231 -0.000858925914316507 I,-0.000987786184857025+0.0005952085493791416 I,-0.0004037127978522765+0.001234224096426765 I,0.0002279820262083385 -0.0008901333687922054 I,-0.0006742492839092536-0.0003467288548498473 I,-0.000880044850478483-0.00165718913895349 I,-0.0002328826940145019-0.001454621898152085 I,-0.0007216221259283024-0.0003105320336786857 I,0.0005941356739947532 +0.0009239764561632795 I,0.000980912178381092 -0.001652878444357762 I,-0.001231726064741791+0.0001238709601228237 I,-0.001632687819246788+0.0006783869665773128 I,0.001320081359887508 -0.001031305103932887 I,0.0003334658177210324 +0.000337145823237756 I,-0.0007137630720576232-0.0005191084945454809 I,0.0002320268608646875 -0.001003719716345938 I,-0.0003229888145615187-0.001392840816998101 I,-0.000459185313443834-0.0004634060898254214 I,0.0002025837133496622 -0.0007136510354134584 I,-0.0005355403283653511+0.001185281030026558 I,-0.000995166368567697+0.0003305228849408894 I,-0.0003972891901355836-0.0006819405778690575 I,0.001699370748924002 -0.00001562858556191488 I,-0.0002224211238652388+0.000135604130404835 I,-0.0005938574477831768+0.001185024662358147 I,-0.0001964469845464763-0.001022755157477822 I,0.00001037296919947483 +0.0001781603640774118 I,0.0005117270169320367 +0.0009648918332586121 I,0.0007092648839756011 +0.0007418596291318418 I,0.0009474830837892054 -0.001024685525079566 I,0.0007251045211869496 -0.0002651289335340681 I,-0.000705872648100373-0.001823033223951655 I,0.0004403516880832233 -0.0002259185004199068 I,0.002302827653017599 -0.0004645432627890699 I,-0.001984021385088291+0.001319967984693916 I,-0.001057459179100273-0.001215812217874626 I,-0.00012140644007586-0.00119982421036966 I,0.001125736961215176 +0.001613548309568445 I,0.0008709615794448135 +0.000362051356787037 I,-0.00141015908854594-0.0007805567907142992 I,-0.00155739830994675+0.0003568091372403925 I,-0.001789767389902926+0.0007547083051869096 I,-0.001568678835206108-0.0005968296105384522 I,0.0008245646565918513 +0.001072523261588383 I,0.0008075119852908592 +0.001482237097088657 I,-0.001007559822819545-0.001140735354114959 I,0.00001020454927122506 -0.001428582387638628 I,-0.00144918497858327+0.001579727478302431 I,0.001851977461162599 +0.0002560490789811565 I,1.495949474783747*10^-6-0.0007893668310686723 I,-0.0005812003013236037-0.001474582939393178 I,0.0009855619784856107 +0.0001942421434430401 I,0.001209067938098052 -0.0005282911508336548 I,0.0007119934945726508 +0.0002659798953344506 I,0.0003740900359988958 +0.0007000186990575512 I,-0.0003376219071430273-0.0002500412985333623 I,0.0002416075140922673 -0.0008648222552520963 I,-0.000404222056818347+0.0003390015056581107 I,-0.0004702008806495314+0.0004757971627693 I,0.0009185199074829096 +0.000623093211354127 I,-0.0001555539883473323-0.0009788455080661947 I,-0.0004928400712782052-0.0000479280743828866 I,-0.0008602844806972433+0.0006234505194507008 I,0.0001373052472076603 -0.0005238086749609845 I,-0.0004234419016579082-0.0005164671993665361 I,0.0001282377931571914 -0.0009383075981917299 I,0.000376214552047107 -0.0006653853478817928 I,-0.0002561037919413124-0.0003813864335356711 I,0.0001030306794315272 +0.0003932357768120991 I,0.001030224444860175 -0.000458562344696313 I,-0.001143633017800813+0.0001025818600951612 I,-0.0009005341093866766-0.00004672100802332788 I,0.000962898186290999 -0.0009613983661262885 I,0.0002691218525140184 +0.0006443712821837076 I,-0.0005249239884475556-0.0003581586691433095 I,0.00006349202385978682 -0.001030829147033986 I,-0.0003944522610028903-0.001309584470773498 I,-0.0003830553491168741-0.0003262050328753843 I,0.0002936519835068667 -0.00059394445298207 I,-0.0004234611584518191+0.0008598882895439338 I,-0.0008078697432043311+0.0003034255826369925 I,0.0003705943464598091 -0.0005638833332406378 I,0.0003939376052864488 +0.0002256393284261923 I,-0.001031086225942276-0.000369085342005575 I,0.000348441508620464 +0.0000644098680078497 I,0.0007809684885451845 +0.000283288755176538 I,0.0002000886335200812 -0.000887077789500282 I,-0.000203065632277809-0.0002991795539219951 I,0.00003563523632085655 -0.0006488860716417263 I,-0.0002860932744783568+0.0009645079051296289 I,-0.0001208946152599156+0.0005260927573601947 I,0.0001353457899386872 +0.001236014338036626 I,-0.0006868786518400753-0.0005799456468802602 I,-0.001395532437532588+0.0006542018690908212 I,0.001314323880532158 +0.0001299832844812626 I,0.001544936548364715 +0.0001899833535597752 I,-0.00115419618641631+0.00003340098242500913 I,-0.0006076018998984502-0.0002704784014442087 I,-0.0003603594334991417+0.001182374815927407 I,-0.0003528254899371709-0.0003397147725792839 I,-0.00116536216266101-0.0001926497726042342 I,-0.0005131618612495505-0.0006901061544471287 I,-0.0001769652300593625-0.001056723044101777 I,0.0003864599090868638 +0.0008323365058290716 I,-0.0006277674894914725+0.001405881559948607 I,-0.00005778035073008224-0.0004682387925431313 I,0.001219074639783294 -0.001486401021454403 I,-0.001107617516370281-0.0001203468590230189 I,0.001010766889589409 +0.001715506644091715 I,-0.0002254698961008713-0.000634552242375928 I,0.00031028121705045 -0.001094119993801977 I,0.001098604304336393 +0.0005673349739113299 I,0.001551222473230069 -0.0002830022919062667 I,0.0004908727595504632 -0.00001873672128947757 I,0.0005127313819816627 +0.0005456949592614011 I,-0.0009556873469658393+0.00006580646461128365 I,-0.0005816127707447163-0.00123955454793989 I,-0.0005047088351345173+0.0002026204096539385 I,0.001117022895212022 +0.001172148322024803 I,0.001341076814997387 -0.0002033776523850928 I,-0.00179692240203349+0.0002316781822167247 I,-0.0002833562458021564-0.0003207745438298782 I,0.0001813596258526075 +0.001285415329916809 I,-0.0004119341819353551+0.0003244768296326086 I,-0.0006101774894574691+0.0007044684349696922 I,-0.0002880201061839672-0.0007384613871937176 I,-0.0002179970915904526-0.0007617460375008005 I,-0.0001852438768985877-0.0002366539403340466 I,0.0001168492716483581 +0.000611400198782841 I,0.0004334831095940438 -0.001019181348281244 I,0.0002649323294513471 +0.000265875224228107 I,-0.001042864105209015+0.0002227667523288611 I,-0.0001569593260201117+0.0002730278641287405 I,0.0006971497545280301 -0.0004542898194957141 I,0.0001423694724422909 -0.0004314829106044153 I,-0.0003395614313284384+0.0005852997670554531 I,-0.00006680987072307364+0.0005712696114587735 I,0.0003410083790249144 +0.0007242724095684592 I,-0.0001970261848885219+0.0004759955048126026 I,0.0007890540629850694 +0.00005090360981372744 I,0.00008814232605263068 +0.0001316687123659123 I,-0.0007514574892251151+0.0007513906729804136 I,0.0006286180333276385 -0.0008890632801597708 I,0.0009570535682476293 +0.000229967254757051 I,-0.0002867218269370342+0.0007731104635079996 I,-0.0009597345899057868-0.00067645737904404 I,-0.0001446490533206974+0.000688729962838604 I,0.0004880903083377759 +0.000582246761944826 I,0.0002760081054850089 +0.0006703997116851266 I,0.0005316045542117875 -0.0009298618303202306 I,0.0002387733915285469 -0.0003214558688644988 I,-0.00002584572262198008-0.001193971646591482 I,0.0005972828460326968 +0.0003837127945898662 I,0.0009414347143036469 -0.00006732837484742574 I,-0.001087407413740509-0.0005064961361906545 I,-0.00036715174828817-0.00008138665690248729 I,0.0009038982230727268 -0.0001095578597423793 I,-0.0005360901573788414+0.0002422189016363462 I,-0.0002313047051441476+0.00006491511723058605 I,0.0007171510561288881 -0.0003522040359434237 I,0.0004339885365485644 -0.000550505854909595 I,-0.00009101289093978344-0.001150065825846301 I,0.0003882576760612362 -0.0007214284092771068 I,-0.001144130203033567-0.0002071783928479761 I,-0.0002366946851873673+1.121409632581414*10^-6 I,0.0007079715205134747 -0.00128387016670725 I,0.0003201164042197838 +0.000944561398102608 I,-0.001102543750231645+0.000250920635733914 I,-0.0002985267478322936+0.0001860089285011232 I,0.0007252009626505785 -0.0007600773924214905 I,-0.0002220101448805712-0.000581430189583114 I,0.0001452248419068986 +0.0009268134225298971 I,0.0007121224748415631 +0.0005541218666446791 I,0.001574650980795643 -0.0000149196487828141 I,0.001035447718924485 +0.0006352472014788005 I,0.00004407877973538555 -0.001760938212535174 I,0.0001480004151842716 -0.0007653093386705212 I,0.001890381969572737 +0.001182078728200648 I,-0.001886126607862693+0.0008085875119774024 I,0.0006206889654451777 -0.001790074228487171 I,-0.0005616212684560033-0.001058179712016937 I,0.00002945754663345561 +0.001824769952654164 I,0.0006774825753619629 +0.001431129516063218 I,-0.001037451044280431-0.001064692732666217 I,-0.001631277674452662+0.00008389779530536398 I,-0.00169885747964102-0.0006409352367053916 I,-0.0008858750047850399-0.00145804875349679 I,-0.00004019340918996533+0.0004925537761175947 I,0.0002665572799931306 +0.001569553450706669 I,0.0007485634671490259 -0.001801249736815171 I,-0.0002920615781152364-0.0003312507829514832 I,-0.002631449342532016+0.0002817050209486961 I,0.001267220413952202 +0.0002142220676178419 I,0.001287290256624828 -0.0001003855529126639 I,0.0002320726823737439 -0.001657627000978228 I,-0.0001150135884858169+0.0003544982827505241 I,0.0004150090759275782 +0.00006344419085194033 I,0.0005011136026925518 +0.001298446790634303 I,0.00002433090565444047 +0.0009212448157065252 I,0.0008399502171532621 -4.148345229888014*10^-6 I,0.00003299903021473515 -0.00008953669732230697 I,-0.0001297175366762036+0.001285420682698248 I,-0.0004227136108106795-0.00105938641097486 I,0.0005790343718216886 -0.0007156542398650478 I,0.0007202798934657969 +0.0004201495338259389 I,-0.0007452519811870444+0.0006684698736047635 I,0.000462835801533622 +0.00007692446360649035 I,0.0005863286019162836 -0.000349009586293495 I,0.0006234072391067554 -0.0002802538881225546 I,-0.0004913031495524359-0.0005231017259232101 I,0.00001458298328133457 -0.0001921527153265207 I,-0.0009812051124804082-0.0002376052262132585 I,0.0001644205581974375 -0.0004969178255414429 I,0.0005981280837809057 -0.000622480114426442 I,-0.0007875893269807771+0.001072471288116516 I,-0.0004051543662056107-0.0003659597318841846 I,-0.0000750880518627236-0.0008179138799630537 I,0.0006056664866081131 +0.0009738600460825489 I,0.0004899882881709947 +0.00006645439829163879 I,-0.000628600409914185-0.0007022565297625448 I,-0.0005835877093839368-0.0002513509784877521 I,-0.001279821103089109+0.0004205426961338156 I,-0.000819447200150186-0.0002497953337897638 I,-0.0002825760024158055+0.0009811292085363777 I,0.0001715984828714651 +0.00001136080106868738 I,-0.0008427488990051348-0.001029274905942933 I,0.0002890243971671809 +0.000742729629913302 I,-0.0003925594339840179+0.001014000616002457 I,-0.0002592405477010705-0.0008637500312657586 I,0.0007514662884333357 -0.0002193515343398988 I,-0.0005037384901540377-0.0002564085695561395 I,-0.0007323244147059631-0.0003354754808084802 I,-0.0009669956831527861-0.0003996459155909831 I,-0.0001268920608422117+0.0008878327498676572 I,-0.000213071659306178+0.0001851572066288138 I,0.001087494603038505 +0.00118564631877304 I,-0.0006867133234431848+0.0004339414136002005 I,-0.001071026469460667+0.001104822287310686 I,0.001271412977476754 -0.001293843807458494 I,0.001200601571253948 -0.0004141476733385588 I,-0.0003990757552147161+0.001069633774311558 I,-0.001090786153345806-0.00004907696927165253 I,0.0003675795134136973 +0.001117107367378942 I,0.0001179290997076292 -0.00009158795818469865 I,-0.0005258582705246547+0.0002996422060404672 I,-0.0009539719146441913-0.00100848777990328 I,-0.0005839524209753453-0.001204183866663921 I,-0.0002286427499736472+0.0002672001815495404 I,0.0002066946004514469 +0.001340136394930716 I,0.0003120061273504432 -0.001507101139029467 I,0.0001870183055599924 -0.0007764239457346016 I,-0.002041318085524646+0.0003989863291900811 I,0.001412164636324673 +0.0007756216412335719 I,0.0005366737399481421 -0.0002623791601302055 I,0.0003285770624074325 -0.001493650694739729 I,0.0005390514903371123 +0.0001920353558272536 I,0.0009556066829536762 -0.0003772030670616529 I,-0.00018095788807255+0.0007382006779091195 I,-0.000238298148522617+0.0005482548513110655 I,-0.000373432465023032+0.0009499868675425871 I,-0.0003642476528861011-0.0006045245400366178 I,-0.001581078391413277-0.000123350216554064 I,0.001743749873003353 +0.0002112327137032707 I,0.0009554043913370725 +0.0007744668629594045 I,-0.001071653730950756+0.0008370642133306142 I,-0.0005083258231043695-0.001149389319167325 I,-0.00006167630861629993+0.0006045635427199145 I,0.0002610367351250922 +0.000846289322500624 I,0.0002574674587547273 +0.0008677364717038327 I,0.0003085169362336386 -0.000759456650480838 I,-0.00004834389019186267-0.0004218023884370651 I,-0.0003761054599145076-0.0009048196344015515 I,0.0006803472938427627 +0.0003760704166885766 I,0.0006873512870200691 -0.001221640536330016 I,-0.0006894227959702412+0.0004530488093627365 I,-0.001495227371448703+0.0007497051251375586 I,0.0006182933175832068 -0.0005204601779910572 I,0.0009324969812538947 -0.0003654733628102125 I,-0.0004143997504094988-0.001069178646745856 I,-0.0001227645617723195+0.0001788553075361017 I,0.00008517979509857954 -0.0002212567782781082 I,0.0005980965444973769 +0.0008348107223281493 I,0.0001962226762957186 +0.0005614485502012408 I,0.0006528930459409111 +0.0003838057108273475 I,-0.0004442734515294699+0.00003178934034496505 I,-0.0005538313146706169+0.0008390567476045141 I,0.0009163971609254562 -0.0008189072493504908 I,0.0005217180087110332 -0.0002412192206809009 I,-0.000127293119508394+0.001138315731689978 I,-0.0004861837669741026-0.0005062686641740506 I,0.000331169031636517 -0.00003752115916178845 I,0.0004609660246562499 +0.0007451032782914232 I,0.0007773704976063688 +0.0006619662644377053 I,0.0004828620747936597 -0.0000522722004490544 I,0.00008592891585254407 +0.0004077486676909221 I,-0.00005827545075183104-0.0007835202666154008 I,0.0004705441775543944 -0.0003806975018768149 I,0.00006392578146992571 +0.0003410276989111008 I,-0.0004847223910558378+2.275865519415837*10^-6 I,0.0005832299449482785 +0.00004146164983762643 I,-0.00007724193853693071-0.0001467222691645235 I,-0.0005116130327798128+0.0000810098050310601 I,-0.00005356374848059647+0.0005606825426708069 I,0.0001772357710145092 -0.00007533310687374761 I,-0.00005673972425101754+0.0001372091305357558 I,-0.0001896525664289893-0.000702497793619763 I,-0.00006597342674785124-0.0005229170651845445 I,-0.0003115793613697631-0.0003303883047104072 I,0.0002614074737529678 +0.000419545411785278 I,0.000454214587569535 -0.0007112879811003218 I,-0.0002165236924092785-0.000144275815141833 I,-0.0008583820306161897+0.0001740700082328577 I,0.0005707975319360253 +0.0003355949580461962 I,0.00007078696573825627 -0.0002246368587817481 I,0.0001141033812363953 -0.0004945303994427124 I,0.0005382743666227595 +0.0003643032024675701 I,0.0008898039990997138 +0.0001615422335865712 I,0.00035722788148195 -0.0003347330183105445 I,0.0002958868718436755 +0.000104700644344835 I,-0.0008326891363614701-0.0008644069582729227 I,0.0003626127057938407 -0.0006078289815859985 I,0.0009004188650222965 -0.0006452859322536522 I,-0.0009618393274228529+0.001419606609327821 I,-0.0005987618596934223-0.0001978344393966311 I,-0.0004219980015903364-0.0009630388407474289 I,0.001022457356415546 +0.0006402564603651885 I,0.000436951331746263 +0.00001640053752942057 I,-0.001057155925772937-0.0001962095901328674 I,-0.0009061778899895237+0.0005012463347600299 I,-0.0008917577549917182+0.0009769034930174098 I,-0.001139840728217973+0.00006909489672969117 I,0.0008339676228523803 +0.000639240207964447 I,0.0008063452859649065 +0.0007199133206257236 I,-0.001254744678068027-0.000452739808124492 I,0.0000547584704342712 -0.00101941577962873 I,-0.0003533841618632439+0.001624464506364165 I,0.00109758551911125 -0.0000162674916564092 I,-0.0003789706987715995-0.000933987508079172 I,-0.0008461085149449681-0.0007321387393198098 I,0.0008955698796333669 +0.0002221081078224397 I,0.0008985117555043588 -0.0003515964770738307 I,0.0008205470137770375 -0.000329832953057168 I,0.0005111863501578818 +0.0002511509331103127 I,-0.000506594211068661-0.0006489442634111781 I,0.000344988512690091 -0.0005432535311317238 I,0.0003195011885521065 -0.0002975916815596865 I,-0.0008669053305775832+0.0008427698155215952 I,-0.0001597827895218188+0.0006674661853616169 I,0.0001363070157052104 -0.001300835427990566 I,0.000368883121768392 -0.00001844643871413423 I,-0.0009171766242358197-0.0001534016615950496 I,-0.0002041321298183157-0.0005228350409902813 I,-0.0005637915410041342-0.0007468518834998234 I,0.0002535139091914204 -0.00004583435178982787 I,0.0003203458916669175 -0.0002488011340643613 I,0.0002296284126468995 +0.0004187167387722283 I,-0.0004538740840632366+0.0004074727083027812 I,0.0001170050071709981 -0.0001092748722445077 I,0.0003190803535729795 -0.0002388075270268843 I,-0.0006298781422196733+0.0001712159525510925 I,0.0003719530079069938 +0.00008077176488420696 I,0.0003666068892852707 -0.0004120898702315323 I,-0.0004761883945021501-0.0007340200836349913 I,0.0001409360736457459 +0.0001277478235458061 I,0.0001654865765548143 -0.0004353578693505311 I,0.0008103823198426458 +0.000327729012646666 I,0.0006199372699191379 +0.0003382334427057854 I,0.0003819780040309189 +0.0001305847538220411 I,-0.0007091340568471478-0.00002196228771526919 I,0.000191997438530636 +0.0008399200905102825 I,0.0005478076519754938 -0.000462310465876183 I,0.0001813483357259111 -0.0007742463493431815 I,-0.0001012997615096657+0.0008116624279666971 I,-0.00008074048604877007-0.00006467339580438808 I,0.0003602284051636406 -0.0001933859448686201 I,0.0001759523901367457 +0.0004382592541454871 I,0.0004892847340845456 +0.0003238720587923748 I,0.0004231202707855306 +0.0001754174643754699 I,0.0002254637621803373 +0.0003973949513024449 I,0.00001839393533305191 -0.0004944951213287528 I,0.00008406141064847754 -0.0003714144646881449 I,0.0003532364880815921 +0.0004115139470841878 I,-0.0004534102312128414+0.0003097514140206524 I,0.0001813547373167778 -0.00040349248192933 I,-0.0002449893533359835-0.0003421769254667165 I,0.0002920205499973652 +0.0003904991951390942 I,0.0001905223153692905 +0.00009530871521007465 I,-0.0005670212673555481-0.00008635026235240218 I,-0.0005643020230474503+0.0002685847841671776 I,-0.0001828999895552714+0.0007135330343893009 I,-0.0004391076418840372+0.0003021100564018563 I,0.001029113004361141 +0.0002686823923290169 I,0.0003259767292458555 +0.0004580062897784346 I,-0.0006814980714948148+0.0008311413410456819 I,-0.0002446708805757211-0.0014477569355389 I,0.0005625783493290401 +0.0003862799352992187 I,0.001046323555436715 +0.00004417775775341249 I,-0.001018884406740537-0.00006279143508357353 I,-0.0007172734467711162-0.00003479647127902755 I,0.000883423829548365 -0.0005848432536331607 I,0.0002982068931441618 -0.001171572043146219 I,0.0002464436430628656 -0.0008825275953277062 I,0.0007968351897267011 -0.0004598335543729116 I,-0.0005532736804486561+0.0004194564034931217 I,-0.001187871939931433+0.0000198499371459746 I,0.0005029059525423166 +0.0001961814281654105 I,0.001117514841168557 -0.00008470167882375986 I,-0.00004635808254831999-0.001274873214641329 I,-0.0005682794964810465+0.001364752567982495 I,-0.00001343316810122552+0.0001326011029103806 I,0.001093534067251936 +0.0002033579901057706 I,-0.00006191230123137005+0.0005576062830194512 I,0.0002726791304100923 +0.0008740255847798403 I,-0.0004928304166516344+0.00008126680903955809 I,-0.0006114005506671902-0.00004457324864197117 I,-0.000112496266399886-0.0001784784625166224 I,0.000651028421326889 +0.0002303155774757889 I,-0.000311345070639343-0.0007452753768008597 I,-0.0001364930252366961-0.0000977076455995037 I,-0.0006516279439356418+0.000928518171458215 I,0.0005528656593696819 -0.0001086991648347446 I,0.0002282880814911823 -0.0006062464020717871 I,-0.000514945234940673-0.0007329290113590285 I,0.0002859530343741601 +0.0002849746258988785 I,0.0003532069760289849 -0.0001724553845554299 I,0.0008529752950706567 +0.000311616518144928 I,0.0004769800358756919 +0.000440714556076989 I,0.000523651381495097 -0.0001475560105881662 I,-0.0003362372914974121+0.00009178361484839411 I,0.0001098254231363998 +0.001022014583958323 I,0.0002149927604949857 -0.001094568794211439 I,0.0002858044784425223 -0.0006785945825207629 I,0.0005775876618530338 +0.001065449141345471 I,-0.000650462235119074-0.00001547432334451915 I,0.0003692448706923054 -0.0003337238862901623 I,0.0008552463461669558 +0.0003541146034911975 I,0.001174776823583949 +0.00003787706658444788 I,0.000443067599818345 -0.0002302350207073026 I,0.0004323541038538111 +0.0004201929472252224 I,-0.0005177539182724577-0.000602292602581605 I,-0.00003261871447883405-0.0007180748415192087 I,0.0003307332972226928 +0.0005335942937908287 I,-0.0002425814187284722+0.0001857527533820464 I,0.0006477143320552843 -0.0007203080807724112 I,-0.0001998516896409056+0.0003226243769435331 I,-0.0004787520758727804+0.0004572006411572998 I,0.0006177481997172367 +0.0004395709669737187 I,0.0002854225669768415 +0.000037671077494332 I,0.0004968606497579152 +0.0003425357479195335 I,-0.0004848201036168068-0.0004846388644758904 I,-0.0002876255699518815-0.0001759159657289085 I,-0.0008208244387875017-0.0004825038360884899 I,0.0006914995530357257 -0.0003061703150222661 I,0.0001789111524457059 -0.001002512991492255 I,-0.0007715695705444701+0.0008175586364417078 I,-0.0006384402948421151+0.0007755095700775652 I,0.0004115132138330247 -0.0008696262977490873 I,0.0004783629850235198 -0.00007217438794441942 I,-0.00048574524082381-0.0006605986595770481 I,0.0001217694710694815 -0.0001818832090536036 I,0.0003269359069825507 -0.0005608342737212135 I,0.0004270196869716665 +0.0003514869620145605 I,0.0003976788977575951 +0.0004940632473765832 I,-0.0001244044408064873+0.0001815453538950842 I,-0.0003866832379570745-0.0007276333794012503 I,-0.0003635318944171253+0.0006170473187418005 I,0.0004003737756857685 +0.0004142809809604113 I,0.00105135611063726 -0.0001124980735328198 I,-0.0007858056060340723-0.00008775640982019268 I,-0.0003620267931363952-0.0001514623319878673 I,-0.0002479686480347081+0.0006366952928423261 I,1.085098787775981*10^-6+0.00006154062175538232 I,-0.0001410930792191011+0.00005925161596401782 I,0.0004093222911034768 -0.0005604861853180709 I,0.0005009310080285958 -0.0002215395587201876 I,-0.0000811920524667214-0.0005755941278578732 I,-0.0002390163622707385+0.00001617814789780484 I,0.001089072117216583 +0.0003128294694486058 I,-0.0005307585969812669+0.0001600749586737713 I,-0.0002078732458045693-0.001134468438944203 I,-0.00005528849788146752+0.00003144395226832228 I,0.0002634221829061733 +0.0008448713249438346 I,0.0006817111067425627 +0.0001088008387931054 I,-0.0004159494475515689-0.0002610759143971066 I,-0.0003203332708847024+0.00008863468800640958 I,-0.0006805567934205299+0.0004347055073039677 I,-0.0004698050554997292+0.00003488005070639806 I,0.0002657200482830189 +0.0005875229534102754 I,-0.0001119376159886273+0.0001706540909272805 I,-0.0006405171025156839+0.0002220733667730419 I,0.0006755291150633171 -0.0006027112345106798 I,0.0003806262072559409 -0.00006791886604003215 I,-0.00006304421496643392+0.0008230544467455376 I,-0.0004693457250332533-0.0002492663056149707 I,0.000342544241113603 +0.0001218881873294596 I,0.000348963854790445 +0.0004356358171000929 I,0.0005279556366253432 +0.0004743994284936911 I,0.0001878982508422074 -0.0001449944925597704 I,0.00003807485525818613 +0.0002239918933187018 I,-0.00004291355740192919-0.00060649817835518 I,0.0003117547467527818 -0.0002233606653523392 I,0.0002561351873202561 +0.0004947417530824716 I,-0.000544105507837129-0.0003198196501976599 I,0.0005300696946401676 -0.0005489752818674967 I,0.0002106689115979772 +0.0002397647982433783 I,-0.0006750904707819196+0.000591937490942751 I,0.000571385500123937 +0.0004908182690425162 I,0.0003632662396361843 -0.0001994588290069406 I,0.0004087197226819032 +0.0001681745306360049 I,-0.0007235245315966817-0.0005835823520266974 I,-0.00037073914420658-0.0002776945210125782 I,-0.0008416280745989105-0.0003130712337245976 I,0.0006325525515730119 -0.000325397734431622 I,0.00006019088920616151 -0.0008051218982202402 I,-0.000735955769662503+0.0007089348101969694 I,-0.00006891309308931109+0.0004570377246166334 I,0.00009999631012152968 -0.0008275840774232835 I,-0.00002287827005034325+0.000394937275362403 I,-0.000107902325454227+0.0003267095005966931 I,-0.0001006437236327059-0.0006187980404691539 I,-0.0003947951479982741-0.0003890402207963664 I,-0.0008552306924565734-0.0003527001268917495 I,-0.0004438465700061537-0.0005775612090899193 I,-0.0003937965478695904+0.000652466556977837 I,0.00004486075051833387 +0.0003433821504079526 I,-0.0005747622512521078-0.000918741859762678 I,0.0005777667472069329 -0.0001561784021629941 I,-0.0003472200023375698+0.000819866071298309 I,0.0003449654411678278 +0.0003495469984716828 I,-0.0003544267967447644-0.0006437471928411849 I,-0.0003432661125658857-0.0001801440685721119 I,0.0008411531242710919 +0.000165786953718159 I,0.0008514657947688439 -0.0003291044401718957 I,0.000319946653958951 -0.00106870933392267 I,0.0006144127007533504 -0.0004532226373055117 I,-0.001447319072778124-0.0005755798060784039 I,-0.0002426925262159666-0.0005312664548500299 I,0.001177051159767526 -0.001274235948514611 I,-0.0001537049446872232+0.00189719075777368 I,-0.001127185001330516-0.0004735682193192892 I,-0.000900734699485661-0.0003601594880790949 I,0.001483316047766403 +0.0003683115573802296 I,0.000890962186728478 -0.0002188599766393358 I,-0.001182343495110724+0.0003652376506812664 I,-0.0005843315693639594+0.001062036572830739 I,-0.0007547816791386072+0.001341083036553784 I,-0.001277751217751578+0.0004406743087508405 I,0.000832670743066993 +0.0003461537162185023 I,0.001135731679791445 +0.0003732898690029416 I,-0.001434944410954467-0.0003823059641133335 I,-0.0001978552266780577-0.0007219605057808826 I,-7.996238042440166*10^-6+0.001865307609979997 I,0.0009185618825498248 -0.0003979709784334245 I,-0.0004546747343698805-0.0009066281830168508 I,-0.001014057747914208-0.000513165737715886 I,0.0008285636959367192 +0.00007710145205347012 I,0.0007396474524379028 -0.0004484192176019005 I,0.0007394238377817322 -0.0004456266848688778 I,0.0005003438619047791 +0.0001457605649403853 I,-0.0005671951045968179-0.0005410979160732259 I,0.000288808964490239 -0.0005439050827442452 I,0.0001921523314227718 -0.0003870893286545149 I,-0.0006633966096147453+0.0008994158236026501 I,-0.00006175331233700099+0.000813385818934198 I,-0.00001809777909922872-0.00124580698069788 I,0.0002978850388729694 -0.0002349691994868149 I,-0.001046230873345382-0.00006730530531516386 I,-0.0001535558658038988-0.0003767067482860804 I,-0.0005415465244027188-0.0006332082662247478 I,0.0005330357106241374 -0.0001556669268930581 I,0.0004796543759955445 -0.0002161850774027885 I,0.0004264843328867808 +0.0001126572237958913 I,-0.0003910110065112326+0.0005375886632381294 I,0.0004046501552955036 +0.0002976497537786838 I,-0.00004417615520310432-0.0008191535884854601 I,-0.0005238471887597373-0.0001707492502695376 I,0.0009686959283363192 +0.0003373156854389767 I,-0.0001696093167085716-0.0001325066308612097 I,-0.0003893913767854765-0.0007692801241897772 I,0.0007795789076723459 -0.00008560973426390857 I,0.0007986991314064333 -0.0008277780718444472 I,0.0007667747962596775 -0.0002815501530857977 I,0.0009557691919363009 +0.0001867876951750368 I,-0.0004296079697906717-0.0001364212891669378 I,-0.0008465674194969822-0.0004345754390060586 I,0.0007869187363219583 +0.0006753597769913761 I,0.000160024253604555 -0.00001524091490322542 I,0.0000886555208223864 -0.00143806502347898 I,0.00006933404652936337 +0.0008245750141101444 I,-0.0001499431796992514+0.0007164366118104022 I,0.001044282382848092 -0.0001550328318093531 I,0.0003926468856116689 -6.743660011142934*10^-6 I,0.0009552440308150791 +0.00004994974808186851 I,-0.0007357340013362344-0.00002528572244277894 I,-0.0003929038229158293+0.0001967482891204334 I,-0.001272004905204354+0.00008202489915822535 I,0.0005176770045371615 -0.00106418730481994 I,-0.0007308913533253878-0.00116098832208015 I,-0.0001736421109663051+0.001692246504479298 I,0.0001485120816030639 +0.001529184502626201 I,-0.0003890493079926704-0.00154601154705123 I,0.0004134429369395447 -0.0005232284543088918 I,-0.001372044187098681-0.0002026147749365526 I,-2.924635002530336*10^-6-0.0005389443821576737 I,-0.0002932367623767071-0.0013309469275328 I,0.0006995220740284604 -0.0002552774785031076 I,0.001070258664880798 -0.0001371071162747621 I,-0.0003232353599457617+0.0008293086780322821 I,-0.001782268902991929-0.0003955514102426228 I,0.00008092639123904923 +0.0005533377091522542 I,0.002103915010273933 +0.0004522185695074221 I,0.0006540191698595168 -0.001305496151360192 I,-0.001672073971610707+0.001694278071145612 I,-0.00003505804053597435-0.000479628633817692 I,0.001004709417079598 +0.0005075717946897357 I,0.0000322913528977806 +0.001226718547197773 I,0.0008084522976873507 +0.001345749050178174 I,0.000427605410603056 -0.0002969173771648009 I,0.0001376453570420626 +0.000287988420893032 I,-0.0008127765634514606-0.001433540661067424 I,0.0007219845899917341 -0.001008434103507642 I,0.0008470919633385665 -0.0005877893633066403 I,-0.00119416867804046+0.002132198670372566 I,0.00004876811206315219 +0.0002166770852514322 I,-0.001152575530137005-0.001803388730069724 I,0.0009799887591385146 +0.0003217953363268716 I,-0.0005617089849686173+0.0005700636403841951 I,-0.001229906168540313-0.0003747335930919337 I,-0.00150546476362154-0.00007128036899143757 I,0.0002371773715561095 +0.0002346791377084151 I,0.0003061497251568012 -0.0003182416553346689 I,0.0009873873384304533 +0.0001782808312857987 I,-0.0007916239391606697+0.0008313403710288954 I,0.001007351361519407 +0.0008408963803596514 I,0.0001979702494963708 -0.0008983704274478559 I,-0.0005013529683043471-0.001884086505221292 I,0.0001762129342870069 +0.001249235178211554 I,0.0002192126424703757 +0.0009168148940437407 I,0.001540546628190474 -0.0002330558707743374 I,-0.0001888650605233248+0.0001677174205305494 I,0.0004146371712125662 +0.000518704666998802 I,-0.0009658425794357241+0.0007067736434462741 I,-0.0007555821251623459+0.000372777247534674 I,-0.0004042397993121043+0.000523371226697628 I,0.0001335303191614067 -0.0006556938166508372 I,-0.001055226986323522-0.0003811655519989803 I,0.0009124214827033869 +0.0009210180558644871 I,0.0006916856608346477 +0.0003867994087675222 I,-0.001435265639889962+0.0001608777705039134 I,0.0002465177414965527 -0.000418405219843829 I,0.0004863613663748711 +0.0006477885295018731 I,-0.0007162066694837467+0.0007041969828124784 I,-0.0003937387977157787+0.001330412117090701 I,-0.0002189251603831349+0.0004935921718535491 I,-0.0007809105793607412+0.0002676705286494568 I,0.000574234532258346 -0.0004021613144987443 I,0.0009995510981856903 +0.00007643422139330485 I,-0.0006973546798723562-0.00007012663943756133 I,-0.0003750693945302185-0.00008757439840119727 I,0.0005747876584922185 +0.001056509709212638 I,-0.0002316666877625584-0.0005716468408524533 I,-0.0004084014983982308-0.0004876732333680951 I,-0.000657483413466168+0.00067554303732965 I,-0.00007401070694545053+0.00006747923296536283 I,-0.0005397320202753719+0.0002831467751544686 I,0.0003048644124640402 -0.0006791026051642989 I,0.0001308756172353884 -0.0005689540353052857 I,0.0003215105687109284 -0.0006173675610235038 I,0.0002299372765890835 +0.0008830548931792056 I,0.0009686555200761405 -0.000262708380795741 I,-0.0005496259330211156-0.0007443548250724607 I,-0.001141863021461586-0.000367131648739311 I,0.0009796701739981256 +0.0006308443394234595 I,0.0001053123647317111 +0.00008633870356545247 I,0.0005034617984702454 -0.0007506369989989929 I,0.0005802512037412398 +0.0004354757001892998 I,0.001187990920052394 +0.0003324633520783212 I,0.000256331448052209 +0.00001468678930754649 I,0.0001168371737224612 +0.0004732570310765644 I,-0.0008815432929717854-0.0009547028858989516 I,0.0006587891664931788 -0.001062179033717908 I,0.0005439186870368973 -0.0005245651194035511 I,-0.001159637858387299+0.001863801530495406 I,-0.00003669131487460755+0.0001599891338716599 I,-0.0008867767011521999-0.00152376771171098 I,0.001000958899922308 +0.0005521375378982486 I,0.0001259951229389402 +0.000418847375223922 I,-0.001352652028299596-0.0001903401884067702 I,-0.001309232569971017+0.0007239046427295699 I,-0.000615636336104275+0.0009449821017037516 I,-0.001044057307881137+0.00009222915845231424 I,0.001264015310426249 +0.0001366535740717753 I,0.001115092615162056 +0.000886726943631976 I,-0.0006896749760169604-0.00004058167606625423 I,-0.0009879201763467898-0.001255966449924887 I,-0.000432618787482961+0.001331871522484578 I,0.001671789857127152 -0.0006544675507132924 I,-0.000360368343281957-0.0002642266345481745 I,-0.0009781424104990857-0.0008520162635144381 I,0.0008016427609475625 -0.0004317840663498728 I,0.0006488067788911642 -0.001016059785435428 I,0.000570958421054246 -0.00009094800236302621 I,0.0005890869745295746 +0.0003560119283548747 I,-0.0003886511503251976-0.00001294393471605121 I,-0.0002003052781014278-0.0007697221154445905 I,-0.00009634559042884293+0.0003792945641607827 I,-0.0002420730749227662+0.0006665168063468409 I,0.0008135302220987984 +0.0001184885052142473 I,-0.0004797622730627075-0.0008529275320753023 I,-0.000188964664384992+0.0001839696886027112 I,-0.0005112963430330697+0.0006456534379293759 I,-0.0002285505724737652-0.0005245826844862766 I,-0.0006626090690512574-0.0003870064720390258 I,-0.0002109450266926902-0.0005205402597035156 I,0.00005564450937200983 -0.0005740310902798419 I,-0.0001367198341375815+0.0001374563362020677 I,-0.0000405780804562468+0.0003576076322352262 I,0.0003105864450767933 -0.0006297954704500624 I,-0.0002299201583987869+0.0002879678588245533 I,-0.0007253520999970265+0.0004255424268194835 I,0.0002718930008079363 -0.0006303902394210455 I,0.0005071999903766046 -0.000147008766276139 I,-0.0006541707518074048-0.0004192377698903506 I,-0.0001686573731020597-0.0002795205582154864 I,-0.0003922578139626225-0.0006244999871813197 I,0.0004093309458082015 +0.0001572597053417653 I,0.0004377325926145586 -0.00005432847620379382 I,0.0003671425057661903 +0.0003937229790245787 I,-0.0007031674667725283+0.0003028300293185672 I,0.0002413602864970108 +0.0004115333958846661 I,0.000509585216704031 -0.0004392516744000442 I,-0.0002567247543151569-0.0005847454095426545 I,0.0001134539158082189 +0.0006509841114670426 I,0.0001269354811849927 -0.00005817585838830243 I,0.0002531539546740799 -0.0004136702365770098 I,0.0001549537299051582 +0.0002973679982441929 I,0.0004609893443126076 +0.0000833946400587125 I,0.0003540896693947791 +0.000265697137313857 I,0.0002308549598213214 +0.0004020055505950039 I,-0.00008757456108224413-0.0002446870749906904 I,-0.00003976771057947201-0.0004843737558863638 I,0.0001459361803125524 +0.0002844117392151926 I,-0.0002353464984681752+0.0005450456272508128 I,0.0002454951501647343 -0.000170630368882326 I,-0.0004049934678562801-0.0004950273958193324 I,0.0003565253170860099 +0.0001890185839267499 I,-0.00009838891062254962+0.00004789117406619195 I,-0.0005180365203700844-0.00009983869235007548 I,-0.0005138899314543357+0.00001674038151032768 I,0.0001441042696297807 +0.0006072581068316478 I,-0.0000336431180847942+0.000349192886612789 I,0.0008432809765415219 +0.0002590667975869274 I,-0.000142685786340554+0.0001119376452199852 I,-0.0004441537597990093+0.00108520564126458 I,0.0001373499256512323 -0.0009188441673206988 I,0.0009727538877327503 -0.0001940523892900086 I,0.0001312054534652055 +0.0002418589994673681 I,-0.0009193256845123315+0.00001216513366794251 I,-0.0003248501528000787+0.0005309546492272331 I,0.0004954977567624032 -0.0002576988567792958 I,0.0000795775515546076 -0.0004395530740097358 I,0.0002232009631556358 -0.0009972690385654382 I,0.0006403802927486903 -0.0005394970587833061 I,-0.0005312560491980584-0.0003301996306616113 I,-0.0005424001343784027+0.00005273611034433762 I,0.001211938958215709 -0.0001268784261051267 I,-0.0001761571964681263+0.0004553059474350685 I,-0.0006022424776420309-0.001146062171373691 I,-0.0002292724905004602+0.0001509020481426294 I,0.0006135780589983008 +0.000660284469376414 I,0.0006852225476498692 -0.0000879078663730521 I,-0.0005634633230430493-0.00005087435376932125 I,-0.0003668839425536818+0.000224649558381032 I,-0.0003960564271920288+0.0006379201420850166 I,-0.0003350057087712483+0.0001697206381652005 I,0.0005384907650680793 +0.000408518315333699 I,-0.0001553852161733721+0.0002512118940574494 I,-0.0002987438492308634+0.0005628147552017372 I,0.0003497681841706952 -0.0008055050120372475 I,0.0002987879273715431 -0.0005141659761779575 I,0.0001833019942577127 +0.0008044969771474823 I,-0.0004550156821500341+0.0001807956535896655 I,0.0005616019655714215 +0.00007625019808533799 I,0.0003306834967426619 +0.00015666248366408 I,0.0005112530550751797 +0.0002444315459935814 I,-0.0002118980281593743-0.0001553229558564351 I,-0.00008991356166134033+0.00007845693395622851 I,-0.0004040403481482347-0.0002985350094426891 I,0.0002236170197843156 -0.0003435746000967611 I,0.0001909052349400815 -0.00006517156507615507 I,-0.0004051681567285643+0.0003579987680612223 I,0.0001397076477399618 -0.0002960441952783817 I,-0.00008944847875270328-0.0001906814743626284 I,-8.411469083260134*10^-6+0.0005292409624551843 I,0.0004279553056420406 +0.0002454323208726424 I,-0.0001507486775376536-0.0002587449887146358 I,-0.00006968606482390645+0.00006511701822466956 I,-0.0007049157780249233+9.896699276378625*10^-6 I,-0.0004495923436280566-0.0001235153451410072 I,-0.0003549428404276704+0.0002203576475452455 I,0.0002934868014007064 -0.0001631673155087524 I,-0.0003302764990551634-0.0005584110013583659 I,-0.0001135908979692903+0.0005078538738553 I,-0.00003440224652043395+0.0004617509355081047 I,-0.0001453644192221082-0.0005148867575263888 I,0.0001964561652975636 +0.0000730688829375407 I,-0.0001074625485661812+0.00009608289916113969 I,-0.0002737439022160606-0.0002643793991328468 I,-0.0003797498511292157-0.0001418702147544373 I,-0.0003875777377748857+0.0001616321780577356 I,-0.00029809164678808-0.0001029113275089916 I,0.0001158359927041089 +0.0004924623284318417 I,-0.00001154199114711621+0.000148529396042334 I,-0.0005599953868874517-0.0001181174143047252 I,0.0003600598575203665 -0.0002514449582963228 I,0.0001971298939842303 +0.0004003802529801658 I,0.0001020613689203485 +0.00009419359730593937 I,-0.0003568399514428616-0.0002283226828315723 I,-0.0002522734556294318+0.00012028765238507 I,0.0003568136247807633 -0.0001576514039862575 I,0.0001569219471431909 -0.0004041377535029674 I,-0.00008892592885403818-0.0006153056942452389 I,0.000246941132769553 -0.0004412715770456847 I,-0.000724714639065112+0.0001975504754643515 I,-0.0004779100271457432-0.00008121085599790479 I,0.0002804822848751853 -0.0007479221678498404 I,0.0007104939764875427 +0.000766814354042124 I,-0.0005815088349419681-0.0002307714087573471 I,-0.0006616089066796792+0.0004736867888793156 I,0.0006822326091023952 -0.0002396147836754782 I,0.0006036195961304476 -0.0002286210096281301 I,-0.0003584703467112619+0.0006788241188412079 I,0.000235156616032556 +0.0008757794124026189 I,0.00001753508850432491 +0.0006272752562196754 I,-0.0004296562302626292+0.0005025548878276233 I,0.0001130840528792423 -0.000390873769415158 I,0.0007636680988713839 -0.0003870018455927454 I,-0.0004371181011787342-0.0001653255738780832 I,-0.0005757976960572499+0.0005190659501581128 I,0.0003037258631596334 +0.0009220764876829021 I,-0.0001277957220256198-0.001020533876395225 I,0.00007695802907210151 -0.0002254520981923957 I,-0.00078124374118085+0.0001401864334730684 I,-0.0001597017744679428-0.0002478742222164514 I,-0.0004204109728294981-0.0003638799952366407 I,0.000411291926596523 -0.0001610218746322903 I,0.0004265917165638778 -0.000121899788712709 I,0.0001970278921933017 -0.00005817127623494885 I,-0.000403866378097015+0.000111129279987255 I,0.000564818944793551 +0.0003770543439620447 I,-0.00007378999436137546-0.00003878389148624498 I,-0.00005489614267493136-0.0008020062402629632 I,-0.0001075764284022065+0.000178653007432668 I,0.0001240097293675697 +0.0005155725031134868 I,0.0005290809432456911 +0.0001144643235246933 I,-0.0002264207386017516-0.0001464270464537633 I,-0.00009691061817744236+0.0000842944702137701 I,-0.0005998770808638279+0.0001575838505406976 I,-0.0003799097797415321-0.0000666391726262023 I,-0.0002970044141370221+0.0003510753620535969 I,0.0000621980694327566 -0.0002035104431748495 I,-0.0004050859762021446-0.0005515270980087381 I,0.0003328574870278755 +0.0007352660332781112 I,-6.571835352554086*10^-6+0.00041423664977817 I,-0.0006848122256099263-0.0003367193188522671 I,0.0004696172241062038 -0.0002043371109001307 I,-0.00001872992542238476+0.0001765954813871653 I,-0.0006066305858866997+0.0001389222072294942 I,-0.0005740864959826781+0.0003802237265158379 I,-0.0001399348181236286+0.0004382258548904035 I,-0.0003801822769890263+0.00005471787667745798 I,0.0005002211998023844 +0.0002119619625332024 I,0.0002137459180194436 +0.0002495337525379094 I,-0.0004821912277056498-0.00006136032472217053 I,0.0001802512497588064 -0.0001280698617803876 I,0.00006600092117037509 +0.0005659443775514169 I,-0.0001576396803906853-0.0001881647438999515 I,0.00008224523161231288 -0.0004171475793144996 I,-0.0004020550561815176+0.00005089443772108204 I,-0.0002156112249122299+0.0001444967988781022 I,-0.0003851759207989937+0.000119616574596134 I,0.0004030216405111828 +0.0001412642117340673 I,0.0001433725864607368 +0.00004168775823676314 I,0.0006448373492815482 -0.0000999395776408006 I,-0.0001346062150134643+0.000388243557853873 I,0.0002343554351150032 +0.0006036301881653113 I,-0.00003521721112630073-0.0007158380289084655 I,0.00007098983917384818 -0.0005669964144885266 I,0.0001992951301809376 +0.0004767451978471025 I,-0.0002357396974256483+0.0002768111126972789 I,0.0004474121629650621 +0.000128799018513193 I,0.0000187358307758861 +0.00002378090157094051 I,0.00002690268773314667 +0.0002682022546906881 I,-0.000434344320980835+4.931150884495772*10^-6 I,-0.0004090682758346407-0.0001102067026686388 I,-0.0000594079548161216+0.00004747568978563985 I,0.0003710263420543468 +0.0001815474704547956 I,-0.0002353717078975662-0.0004964537973225305 I,-0.00009986321181318001-0.00001766839567370716 I,-0.0004198973855001911+0.0005519418237999203 I,0.0003038044669740806 -0.0001416080508829686 I,0.0002314192287520484 -0.0002358160222511185 I,-0.0001474589969886244-0.0004351246638565959 I,0.0000217220599009817 +0.0001143559667574795 I,0.0001628447023969758 +2.846265229840782*10^-6 I,0.0001857744962647388 +0.0004440636687530074 I,-0.00007135992778502904+0.0004071037126719039 I,0.0002527345035184958 +0.0000889403223975908 I,0.0001416573660067011 -0.000236744057045038 I,-0.0005717712859552564+0.0003914298700178463 I,0.0000623488018977672 -0.0001651510814289267 I,0.0006811278531141653 +0.0003916211673482538 I,-0.00008246652926076326-0.0001765638964883101 I,-0.0004608227988984893-0.0001996094039301785 I,-0.0003814282027903728+0.0003274383880749218 I,0.0002146120772335274 -0.0001260219962855784 I,-0.00001347172180338652-0.0002063094924042095 I,0.00009070388203342376 -0.0004822753593007617 I,0.0002015025453348077 -0.0002500597712880906 I,-0.0003145212365665895-0.00009426480983054148 I,-0.00009667268413578027-0.0000760664124313695 I,0.0002150228663737434 -0.0003059655320514193 I,0.00002206034910025724 +0.0004858602094104156 I,-0.0001234287089345909+0.0000596896265827839 I,-0.0003436956386470229-0.000270082967024557 I,0.0002914723302597884 -0.00003616478693861631 I,-0.00007821796187054209+0.0001141573525397973 I,-0.0003393080793970832-0.00001807862915190406 I,-0.0003900056656730222+0.00007892470167736391 I,-0.00009843338309693578+0.0001618119580236643 I,-0.0001703901111741223-0.00006451857753078752 I,0.0002222593836873159 +0.000247328157053924 I,-3.606254547010171*10^-6+0.0001685785786440635 I,-0.0003672901730639996-0.0001467947139055888 I,0.0003817434933832965 +3.368702206281308*10^-6 I,0.00005511531479128104 +0.0004734796823254353 I,-0.0002943808194674139-0.00004111810796708241 I,0.00006072412991014607 -0.0005186030379961038 I,-0.0003919209170726614+0.0001033257982425518 I,-0.0001455760063594731+0.000251584517127134 I,-0.0003021369189948181+0.0001924468430409158 I,0.0004630293543607098 -0.00001957312324246002 I,0.0001674311929941703 -0.00004142488216380802 I,0.0005066162847945444 -0.0002131020879757703 I,-0.00002943144256258956+0.0004102668373157997 I,0.0002171382910936464 +0.0002922861661050895 I,-0.0000163876843188945-0.000568289986621336 I,-0.0001369281132116148-0.00007088103285568786 I,0.000272244965044939 +0.0002831690866862777 I,-0.0001525743777541513-0.0002268027479999595 I,-0.0001804341590163921-0.0001295378746195679 I,0.00025208373240384 +0.0002005773676416724 I,0.0002180279735233219 +3.056080025630903*10^-6 I,0.0005226623442785564 -0.0002277087410658806 I,0.0004240571906043628 +7.620373067731721*10^-6 I,-0.00001020910899674594-0.0004330596810079852 I,-0.0001733333654587835-3.570277793037591*10^-6 I,0.0006658093816027293 +0.0001833153580370406 I,-0.0001701629561722549+0.0001410071398109771 I,-0.0002531897998059813-0.0005581420544658157 I,-0.0001943473244678927+0.0001338256952975432 I,0.0003626281237243767 +0.000139583855697733 I,0.000311897683728616 -0.00009001920767854672 I,-0.000350718804436955+0.0003245024754546575 I,-0.0001113985767446177+0.0005611116796010285 I,0.0002684546639019814 +0.0005414828961239617 I,-0.000114856636734082+0.0004434438840633386 I,0.0007460705685465492 -0.000460202329291006 I,0.0004529374220831524 +0.00007711168970842404 I,0.00008975820848465294 +0.0008813240234185769 I,-0.0009005495938898023-0.0007111859682416685 I,0.0005642943896534537 -0.0001781952990867199 I,0.0005558618569542913 -0.0004180237826999447 I,-0.0006547256424580521+0.0005208862899429819 I,-0.000311102289798139+0.000428234818670454 I,0.0002302792844717771 -0.0007154081207787353 I,-0.0003343637036043827-0.0007042206145866248 I,-0.0003249723488414457-0.0007526307152100164 I,0.0002457263273265721 -0.000713576789148718 I,-0.0003414806620025669+0.0003269444964345563 I,-0.0005963854102786753+0.0004500487137070357 I,0.00061718251529241 -0.0004096981883625778 I,0.0004305244964245015 -0.00003210805094718482 I,-0.0008121522779553803-0.0006893737029060739 I,-0.0000171411043866105+0.0006704169977203764 I,0.0004688904794183269 +0.0001262370540750123 I,0.0006341122372769737 -0.0003232885580094785 I,-0.0002192396619674738+0.0003418840678918697 I,0.00006705015550715514 +0.0004665422988188421 I,-0.0001063311087739007+0.0004827103723694841 I,-0.0002789390049757693+0.0002106819590826219 I,0.0003583939415707623 -0.00004307126020998057 I,0.0002006962539581291 +0.0002347562149384447 I,-0.000030597482534582+0.000174309067582269 I,-0.0001526975398392142-0.0005594198301308859 I,-0.0002460690485296387-0.0001425926213039623 I,0.0005420816722763362 +0.0003501041191757638 I,-0.00006594769034485952+0.0001213074040052185 I,0.0002634317866089998 -0.0004029135128251985 I,0.0002738402215986519 +0.00006157139002496535 I,0.0004733173698863423 -0.00005501456937727145 I,-0.00001201812618429666+0.0001923981801480202 I,0.00001779931708636962 +0.0002590678948858723 I,-0.0001838226856077242+0.00004661056303877759 I,-0.00009745263985035503-0.0003194084230631972 I,-0.0001157275321105661+0.000144785670488672 I,0.0002155748234835042 +0.0001606246517577196 I,0.0002053565648250014 -0.0002866782207656736 I,-0.0002957061946543048+0.0003079629006308718 I,0.00005108162529081633 +0.0001048285004391351 I,0.0005028530426750082 +0.00004302371108361489 I,-0.0001207014853059302+0.0002597521189125827 I,0.000193281889417683 +0.0004879910153147868 I,-0.0002204321023729616+0.0003521628393117484 I,-0.0003760655733871138+0.0003190699075299122 I,0.00002480757991235583 -0.00009480469244109766 I,0.0004262143857382977 -0.0003137471041175171 I,-0.0005201771456971176+0.0000281942195135678 I,-0.00006195641725001237+0.0002187920883361654 I,0.0005616744082816393 +0.0004602714680248208 I,-0.0003955176617705125-0.000294779912528662 I,-0.0002039854146564249-0.0001892392115371558 I,-0.0001926460550886754+0.0004902745572219912 I,-0.0001660171900923116+0.00008359867902543153 I,-0.0003495006575917781+0.0003385563063968329 I,0.00005971984087479922 -0.0002434721666858464 I,-0.00008148578038750793-0.0002352639656383366 I,0.0003039718553189599 -0.0003875837584938303 I,0.0003073538102556292 +0.0005119869893613902 I,0.0004740767961373622 -0.00004816380656459213 I,-0.0005341664633317912-0.0005716755118338628 I,-0.000592016542282727-0.0001972257626809822 I,0.0007594397841244614 +0.0002655683762259917 I,-0.00001127858027749813+0.0002324659354847482 I,0.0003615167369303716 -0.0004596906402421426 I,0.0003329293775338522 +0.00008694341762500405 I,0.0006626095351187032 +0.0001069745863094698 I,-0.0001228339995464547+0.0001476733008460877 I,-0.000162887037777109+0.0003167947047513883 I,-0.0004830828394542309-0.0002892752522737002 I,0.0005016164060837358 -0.0006207822870174212 I,-0.0001514067979835866-0.0003842008907582473 I,-0.0005431755428054035+0.0009604501972083491 I,0.0001183925594409428 +0.0005639406439916319 I,-0.0003946139117336166-0.0009925877575226014 I,0.0004385362832869719 +0.00004448560284142823 I,-0.0002884825844308379+0.0001516997280863606 I,-0.0006125917022229675-0.0001214033323776197 I,-0.0006842035362827353+0.0002077798253139613 I,-0.0000161794771389752+0.0004866347929871861 I,-0.0003138193600254846+0.0001448721799290742 I,0.0008454928408531315 +0.00001031416083215422 I,0.000430494612965473 +0.0004565539148700946 I,-0.0002253445936648374+0.0004425262984328981 I,-0.0006098294016757774-0.0009534305811772037 I,0.00001543015780147141 +0.0003037340955480579 I,0.001000403465727985 -0.0002039411123650208 I,-0.0004183376836847583+0.0001534698381906074 I,-0.0003204547948715508-0.0003467467422367539 I,0.000497954394030568 -0.0004034174603056851 I,0.0003867792452187912 -0.0006492135368422998 I,4.552893458740592*10^-6-0.00006423978128104572 I,0.0001763672931620389 +0.0001701731001349532 I,-0.0003333802503059938+0.000246233265796316 I,-0.0001600895766730101-0.0005206539819178728 I,-0.0004068653420352152+0.0001869317602309232 I,0.0001880808841942038 +0.0003804159946715909 I,0.0007697087746266894 +0.0001895311116493775 I,-0.0005116096124217994-0.0003329796246052326 I,-0.0003126389312908946-0.00004930582514509182 I,-0.0002738397435625631+0.0006138256378034061 I,-0.00008589439455063-0.0001838725259497909 I,-0.0003621047742698583-0.00003011892544284105 I,-0.0001126888398476967-0.0004766979101672214 I,1.917276099912832*10^-6-0.0003891775984973115 I,-0.0001199314710915306-0.000149336430657496 I,0.0001622134898543581 +0.0002000411321181657 I,0.000302360476136107 -0.0003977229861453254 I,-0.0004208044996757951+0.0002262424120872743 I,-0.0003554702323431371+0.0002515953331315164 I,0.0001648963211582722 -0.0005827466317389746 I,0.0002234564471105882 +0.0000900270632966153 I,-0.0003785923927778559-0.000022891109107909 I,-0.000191900516469759-0.0003498142411876942 I,-0.0004600771441939714-0.0003843385122383072 I,-0.00005686154909891419-0.00004433876155325811 I,0.00006326155472450136 -0.0002602239557476399 I,0.000165373652124906 +0.0003353442159824555 I,-0.0002661531848323628+0.0003598462864430565 I,0.00007595020148951103 -0.0001242011017289599 I,0.0002130194224535098 -0.0001876813863188212 I,-0.0004117611615119151+0.00001039484600592758 I,0.0001748224665075998 +0.00009662994850209855 I,0.0002411876179816074 -0.0001440862002850552 I,-0.0001387209676753395-0.0003459989892751336 I,-0.0000353643424577583+0.00006018418019910875 I,-0.00001636883101479116-0.0001431649678522488 I,0.0003209497862512844 +0.0002728138523741308 I,0.0002264971545851331 +0.0002069769910788144 I,0.0003007284335819465 +0.00009773904041729408 I,-0.0003307372504147386-0.00003257713045683441 I,0.00005307671604551421 +0.0005596583762185992 I,0.0002218894824431916 -0.0001821672931805688 I,0.0003252928871558304 -0.0005169944466409198 I,-0.0003266607840414743+0.0002705572396951421 I,-0.00003851364765801153+0.0001657429430975967 I,0.000319313274734423 +0.0002977589877817492 I,-0.0003210312113437717+0.0001171615579408438 I,-0.0003533472088368923+0.000456329166429181 I,-0.0001731552285987202+0.0002460142474398424 I,-0.0003568095824379101+0.00004153094236358817 I,0.0006204519935591817 -0.0001069255301301604 I,0.0004378281609276912 +0.0004468824049787953 I,-0.0001185103783548045+0.000329838477227975 I,-0.0005756781800374352-0.0008922196164819983 I,0.00004236551219962336 +0.0002020067308040078 I,0.0008730602925905849 -0.000199667609360834 I,-0.0004926766254120165+0.0002062237512822651 I,-0.0003917678937081256-0.00005530377430236521 I,0.0003972730148585915 -0.0005630607234229697 I,-0.00005447729099805154-0.0007950097509925839 I,-0.0001344793486002988-0.0004515950772508546 I,0.0002764707610065212 -0.000413908055685636 I,-0.0002151929437650934+0.0004938534329245783 I,-0.0006721364959137974+0.0002945518730797972 I,0.0001903309369633237 -0.00006157646439479718 I,0.0007107914738712994 -0.0003235785898046059 I,-0.0003475328022732111-0.0006085364565822029 I,4.601272767525925*10^-7+0.0009227174121910033 I,0.00005371889378862462 +3.975240183836661*10^-6 I,0.0006332848008108366 -0.0001936388359614593 I,0.00008569181979737391 +0.0003650350728619728 I,0.000352014362762521 +0.0004386064125416868 I,-0.0001928387940041818+0.0002210009577757062 I,-0.0003241740807422618+0.000163258605569791 I,-9.212510416263054*10^-6-0.00008769859927556972 I,0.0003806189924339781 +8.398457804020668*10^-6 I,-0.0002852164249819831-0.0001856615227491 I,-0.0001400880432508577-0.0001718444753051416 I,-0.0001174918564270852+0.0004571255442113867 I,0.0003713303931077036 -0.0000583325825676897 I,-0.00008700445152769724-0.0002200794603901115 I,-0.000255187154930479-0.0002816920565508998 I,0.0002799507294809871 +0.00004162907715240123 I,0.000275072933951456 -0.0001776181905765211 I,0.0003394459442671484 -0.00004550658060638465 I,0.0002735194196271086 +0.0001218037137425511 I,0.00001233371822513248 -0.0001648522687151351 I,-0.0001209633738761296-0.00004700261716731187 I,0.0002778713144331814 +0.0003163400421359933 I,-0.0001639191441865095-0.0002481397508907829 I,-0.00003489923677096472-0.0004355938471370586 I,0.0003126330172399305 +0.000250236352895374 I,-0.000108276417017955+0.0002911705200632839 I,0.000304586206958126 -0.0001964017792271658 I,0.0002241520265278311 -0.00006072421719616699 I,0.0003971179939275177 -0.0001024634062238984 I,-0.0001645727019858088+0.0000726856143201128 I,-0.00003515739728451113+0.0001705993966461635 I,-0.0003200576075563018+0.00006724831018228004 I,-3.002022486835957*10^-6-0.0003902043350776668 I,-0.0001803555120782277-8.08533935092254*10^-6 I,0.00006011017013383366 +0.0003491709524570769 I,0.0002874735727828574 -0.00002868125470599823 I,-0.000295159305212966-0.00009405508923996861 I,5.07748215126697*10^-6+0.0001144469602243671 I,0.0001615619178597443 +0.0001955049625086201 I,-0.0001598766072931911-0.00002379381392280273 I,-0.0001230204676406615+0.0001672249101120726 I,-0.0002525852516631736+0.00009976486091069252 I,-0.0002605511515749484+7.589754712122182*10^-6 I,0.00008460663214708347 +0.00005126244941446251 I,0.0002392483367148664 +0.00006820421267583735 I,-0.0002121179237868047-0.00004922255404071145 I,-0.0001755260029700638-0.0001951744065979209 I,0.00002727470512021253 +0.0002147498793701011 I,0.0003181575361856664 -0.0001001992363565189 I,-0.0001747670507435461+0.00009150239006505803 I,-0.00004058039969914902-0.00006165496482932214 I,0.0001899449575531333 -0.0002288720551180096 I,0.000126976529005673 -0.0002841465666842727 I,-0.0003000151341693126-0.00011117798226097 I,-0.0001203831550732052-0.0001155863541509385 I,-0.000369996870072731+0.0003413940980226221 I,-0.00005994871813027962-0.0001171266181596431 I,-0.000378837061491023-0.0004981514307170029 I,0.0004408379413573175 +0.000351148075658048 I,-0.0001602587765877297+0.000477865656399446 I,-0.0002065640163630758+0.00001989394517810895 I,0.0002611564233018725 -0.0004424285663095862 I,-0.0001346087926725459-0.0002805271464658813 I,-0.00003612849323019054+0.0003325203522644476 I,0.0001815323588267234 +0.0002033957843287562 I,0.0004097070570334479 +0.0003906110290556567 I,0.00006208519791877525 +0.0004596598946428882 I,0.0003114639037932628 -0.0001893827210667598 I,0.0001628959509423565 -0.0002719115232789937 I,-0.0002698319338363512+0.0005966702856210569 I,-0.0002138918832277616-0.0001989493643521738 I,0.0006963469747714245 +0.00006375749893286037 I,0.00001569734845931724 -0.0001961907512438336 I,-0.0004723884937648444+0.0001089603024230227 I,-0.0001790659673251954+0.0003315327165563987 I,0.0001818371024735451 -0.0002690537550622044 I,-6.471336194907931*10^-6-0.0002634028887700873 I,-0.0001733410443020258-0.0003575792944979996 I,0.00003135579682184418 -0.0001949187162182153 I,-0.0003379839429689739+0.0001435378473190594 I,-0.00008802315130367977-0.0001570039831208905 I,-0.0001779836027629432-0.0002715335567794467 I,0.0002738487815332189 +0.0003931015586338717 I,0.000182113135731212 +0.0002699969970260731 I,-0.0004347293708801727-0.0001525495354602676 I,0.00003192234346246397 -0.000249361281417813 I,-0.0001989466642755155+0.0002580702402040256 I,-0.0001665910682336849+0.00009300400695289593 I,-0.0002684246649384906+0.0001552543611509617 I,0.0001221456811457072 -0.00007373166613210483 I,-7.064148999162711*10^-6-0.00008105429001972574 I,0.0002390616457893693 -0.0001021174007497408 I,0.00009359162160226203 +0.0002063141957901116 I,-4.027795089821219*10^-6+0.00008644164125072145 I,-0.0000680063480839304-0.0002749371740553996 I,0.0000675182560931015 +0.0001490246590307903 I,0.0001317321837079836 -8.768622359429923*10^-6 I,-0.0002212430023313721-0.0001425851640543015 I,-0.000244935929615232+0.0001041605992666365 I,0.0001597299632217423 -0.00001464947924207987 I,-0.00001283562356783419-0.00008325869488222859 I,0.0001965162599614625 -0.0003152768280873527 I,0.0001972760608014461 -0.000168301682742511 I,-4.104335262019969*10^-6-0.0001984851614726903 I,-0.00004087606976121766+0.0001742390386520084 I,0.000349334410661145 +0.00001731404313214737 I,-0.0001726591826068087-0.0001913631959834318 I,-0.0002043094837849757-0.0001738293327253935 I,0.0002671669885252836 +0.0001008113154768627 I,-0.00007576955580090053+0.00007669425900118577 I,9.79450903617322*10^-6-0.00009142883758703874 I,0.0001936425744862006 -0.00004311296133670543 I,0.0001866446023315999 -0.0001282225018652604 I,0.00002970336379545259 -0.0001944990216872752 I,0.0001229326236391311 -0.0000829067543356527 I,-0.0002600409879919468-0.00007563982810876533 I,-0.00006135172602495225-0.00007640831496472697 I,0.0001819702115401313 -0.0002098314194679811 I,0.00001602874079901302 +0.0002588770309590789 I,-0.0002011640852447044-7.318232101956388*10^-6 I,-0.00009643404964349012-0.00001222340514112189 I,0.000202891846882489 -0.00009611001578207272 I,-5.23980132212875*10^-6-0.0001283535354682241 I,-0.00004801326305772032+0.0001855792844473976 I,0.0000994857617660108 +0.0001644969168232188 I,0.0002976515989116177 +0.0001348077873644859 I,0.0001323659982677766 +0.000222324650113189 I,0.0001426173291310198 -0.0003474644604450273 I,0.0001388096295359156 -0.0001624200585003276 I,0.0002456529812974332 +0.0003630417272985734 I,-0.0005021313534712361+0.00004555817356297753 I,0.0002749001594129387 -0.0002214528583773478 I,-6.724195442563553*10^-6-0.0003727763709591796 I,-0.000119173312846979+0.0003684012994636977 I,-0.00005541773381170919+0.0003340249693240565 I,-0.0001642816213595396-0.0003505833354297989 I,-0.0004237005672871584-0.0001610807160640744 I,-0.0002919741347012879-0.0002118891189551861 I,-0.00008426211994098866-0.0003599719649630713 I,0.00008324978253465695 +0.0001682995275966688 I,-0.00007851114883247257+0.0004402461895652803 I,0.0002235652536766844 -0.0002027797730751066 I,-0.00001457838382805323-0.0003456955584198892 I,-0.0005449344746847248-0.0001469483782012136 I,0.0004383277149004047 +0.00027073166889442 I,0.0001342379130025604 +0.00007399458672655164 I,0.0001959806054432889 -0.0004026685696402855 I,0.0001220794627293924 +0.00005626674270662544 I,0.0002777972279385572 -0.00001312423189869636 I,-0.00002208578088105603+0.0002655539737299632 I,-0.00008224626033723003+0.0002407179543940483 I,0.00001695396593080561 +0.0000571567496622954 I,0.0001070708669543597 -0.0001681480783912793 I,-0.0002809214081936901+0.000146341814397548 I,-0.0000368479685734864-0.00008179345240473453 I,0.0002800855427272284 +0.0001933696105660953 I,0.00006945292587895078 -0.0000944424359183506 I,-0.0002344306897046319-0.0000653943925779783 I,-0.0001950255868159106+0.0000938081477882748 I,0.000167493420440305 -0.0000729382788462837 I,0.00005603706895953491 -0.000127064828537263 I,0.0001224272603391421 -0.000259107546891316 I,0.0001692919342489835 -0.0001022127498473279 I,-0.00008651962498941912-0.000190134745697867 I,8.752012774150282*10^-6+0.00004689472231187995 I,0.0003518772831561941 +0.00002011152795721122 I,-0.0003535283938108525-0.00006777421077930909 I,-0.0001920244063013627-0.0001972066839888107 I,0.0003438188099074047 -0.0001121952267264767 I,-0.00001181361342453323+0.0002767348290364593 I,0.00001571294158458485 -0.0001414632407291103 I,0.000130269921344566 -0.0002531146481663408 I,0.000104749389268807 -0.0003301534293866001 I,-0.0001371707945703244-0.0000801908237770704 I,0.00007849697646456871 -0.00006478567141717005 I,-0.0002511325171555633+0.0001694958593915448 I,-0.0001812358188376505-0.000107895335276948 I,0.0000709197899947871 -0.00009959320741779494 I,0.0000688601703986686 +0.0001752162021395504 I,-0.000100043218317284-0.0001691118421064953 I,-8.461903972736698*10^-6+0.00002023101069347855 I,0.0001216982706186773 +0.0001607712298803734 I,0.0001681761773737279 -0.0001005695316579571 I,-0.0000370757042771156-0.00008230074719639741 I,0.0000602864951066969 -0.00007757708421186579 I,-0.0002112699049235757+0.0001531130032423736 I,-0.0001292912574439743+0.00007736659692586775 I,-0.000122679946973297+0.0002260373617149441 I,3.678642017128277*10^-6-0.0001778574357250952 I,-0.0003304163190533791-0.00008968681548389068 I,0.0001499793244568326 +0.0001852578322351552 I,0.0002026905980765513 +0.0002644239029468933 I,-0.00013988662957643-0.0001795027350647015 I,-0.00004923476812310955-0.0001112373322131246 I,-0.0002226796434568053+0.00009266506025350322 I,4.544507310916734*10^-6-0.0001088803128776382 I,-0.0001142083158411162-0.0002046087801941083 I,0.00004153312698419563 -0.0001366350051037324 I,0.0001364905371412159 -0.0001234523218415675 I,-0.00005516543504920362+0.0001478557249827339 I,-0.0002683734479416203+0.00004100219645997025 I,0.0000573190592582557 +0.00001089904152510408 I,0.00029515103005942 +2.091846114292146*10^-6 I,-9.570121060766153*10^-6-0.0002065501325421508 I,-0.0001530988046113189+0.0002673793590841852 I,2.930390512705181*10^-6-0.00003103990423946996 I,0.0001622758260734148 +0.00004366678985230201 I,0.00001259015364324633 +0.0001225270493375591 I,0.00008387186669147852 +0.0001349850645563203 I,-0.0000223597677074658-0.00005963350234114943 I,-0.00001589866567903775-0.00003112034682161937 I,-0.0001255908222203335-0.0001056934907824441 I,0.00005899024820267611 -0.00003880777636937257 I,0.0001158691092510518 -0.0001774190466965331 I,-0.00008323187002970005+0.0002102129696305868 I,-0.0001685122204868756+0.00002607057781974924 I,-0.00004858191889175762-0.0001162341157317188 I,0.0001961952830345085 +8.222087940554636*10^-6 I,-5.611953876334458*10^-6-0.00008360419005616398 I,-0.0001223335801991471+0.00002482871835637998 I,-0.00007922718686613733+0.00003277902724413182 I,0.00006756903948566172 +0.0001513564094016991 I,0.00001417188119878153 +0.00007453009390552212 I,0.0001711600162905384 -5.011726718746944*10^-6 I,-0.00002948234033971973+0.0000606657147972895 I,0.0000644818488075142 +0.0001967524009279815 I,-0.00007024898488841266-0.0002025999470882647 I,4.049353854026214*10^-6-0.0001956717313251892 I,0.0001547107845951569 +0.0001167166817112014 I,-0.00007168616317176609+0.0001466074899694593 I,0.0001336513956743167 -0.00005370201327526488 I,0.00009013833828103342 -0.0000531091561687819 I,0.0001362087676051956 -0.00005878130592065696 I,-0.00009931851518825717-0.00001965346026057614 I,-4.958928012616191*10^-6+8.456009289943924*10^-6 I,-0.0001728670303576185+0.00001705723436470647 I,-0.00001945482186118606-0.0001187861959046004 I,0.00005482294990051702 -0.00008973376467476298 I,-0.00004176643755968048+0.0001985304347717417 I,-0.00006973632339218476-0.00009616048046771854 I,-0.00004872771773692581-0.00005864833535952353 I,0.0001485672007718303 +0.0001456842521873602 I,0.0001543159126312263 -0.00006338691716968218 I,-0.0001103099500340748-0.00004830133053505541 I,-0.00001080155585236951+7.706120940207295*10^-6 I,-0.000181353364463826+0.0001997359056700373 I,-0.0001461829906745591+0.00009239355757443811 I,-0.00002901516839539839+0.000175322361271506 I,0.00001811372296349776 -0.0001298368187169073 I,-0.0002591848293131801-0.00004453372744931401 I,0.000134182647487457 +0.0001693140211442456 I,0.0001445070151710064 +0.0001356837231269769 I,-0.000214225802490257-0.00009016750693178186 I,0.0000674324881361452 -0.00005328550803627391 I,-4.597461662102383*10^-6+0.00008127386629699882 I,-0.0001633025658698318+0.00003257627250465153 I,-0.0001613839961589818+0.0001182132506584203 I,-0.00002824712203468289+0.0001494331610401068 I,-0.0001080558704691926+0.00005692048257606509 I,0.0002251849573268123 +0.00002737064793367961 I,0.00008537992298566312 +0.00009516374606246414 I,-0.00009472718359181776+0.0001628751065142479 I,-0.00009545653389325128-0.0002534656811782129 I,0.00008663733084047815 -0.00002001683524783873 I,0.0001795472239525859 +0.00002032061948006514 I,-0.0001419738107504527+0.0001257565657191701 I,0.00006477591566539018 +0.00001622970859327161 I,0.00006832165802799151 -0.0001369501471377179 I,0.00002532348369409547 -0.0001020944483437365 I,-0.0002527862060391614-0.00001984967289853082 I,-0.0001545339231855845-0.00006346482988742901 I,-0.000145054564253975+0.000218272732615125 I,0.0000326870154987731 -0.00002001789106740849 I,-0.0002851466724965645-0.0002713195163018352 I,0.0002090412996583983 +0.0001151691718109826 I,-0.00006691806056579724+0.0003414470687389196 I,-0.00004972881992036584-0.00002930303847682589 I,0.00009619794419461172 -0.0002534877574581544 I,-0.0001384516264471566-0.0001437894626340778 I,2.405343662025298*10^-6+0.0001380364828325944 I,0.0000664537689135638 +0.00005562645823061551 I,0.0002343416298310959 +0.0001689548639467825 I,0.00006359130046793991 +0.0002072784604308444 I,0.0001889137837195612 -0.00009894453546170902 I,0.00005854764689030846 -0.00008308754565870011 I,-0.00006291199753389031+0.0003328465975374931 I,-0.0001616194004169031-0.0001775171756428513 I,0.0002728629671681048 -0.00004725435639577096 I,0.0001198089088560985 -0.0000746787994393325 I,-0.0002129298960950334+0.0001305049115718776 I,-0.00003983160630487067+0.00007381248029290425 I,0.0001045511131804732 -0.0001721145688766499 I,0.00004104387006171966 -0.0001829273446220845 I,-0.0001186506825905141-0.00007509489721008956 I,-5.365075050768403*10^-6-0.00001207990835659742 I,-0.0001333796691963671+0.000150109846865134 I,-0.00006613964843569327-0.0001444977285216691 I,-0.000215741086624803+0.00001063459797063832 I,0.0001775910348978213 +0.0001097240154195072 I,0.0002981737444422509 +0.0001044913892657763 I,-0.0002249162663992902-0.00003322851318662856 I,-0.0001583704099103235-0.00008730576058093214 I,-0.00008895282430882394+0.0002524094179571666 I,8.228158193331688*10^-6-2.561553497795657*10^-6 I,-0.00008157108308671558+0.00005036043118446557 I,9.452853496750343*10^-6-0.0002248327655414143 I,0.00002909812033977116 -0.0001365478826434931 I,-0.00004980132531368482-0.0001704286769497191 I,0.0001107546495372212 +0.0000670137259950351 I,0.0002244823297856288 -0.0000894431211584364 I,-0.0003301263395162027+0.00003500392988393914 I,-0.0001264144003183333-0.00001320196435366762 I,0.0001916449623588469 -0.0002571651633494016 I,0.00001157423823321162 +0.0002071809924904176 I,-0.00007804740759763038+8.563990698912615*10^-6 I,-7.441212152607939*10^-6-0.0002800362057830094 I,-0.0001285719429163344-0.0002800885708257487 I,-0.0002098971300600416-0.0001066524246010187 I,-0.00001570521649599271-0.0001830352991593414 I,-0.0001520304053962184+0.0002336935673313503 I,-0.0001273227213671268+0.00006344503348718793 I,0.00002213284944160333 -0.0002492216779334996 I,0.00007818935507851861 +0.0001441864667768959 I,-0.0002591166296096032+0.00002476242569923943 I,0.00004486992421758143 -0.00009441453526078228 I,0.0002405353431479128 +0.00005958556818119965 I,6.319724387803341*10^-6-0.0001981108261543347 I,-0.0001311108619677024-0.0001130752257154163 I,-0.0001060899280953059-0.0001718035329091428 I,-0.0001187634588429712+0.000310872838674926 I,-0.00008981797510185946+0.0001325705859633297 I,0.0001484605205800839 +0.0004142558772372371 I,-0.0001874792732004665-0.0000762097308429074 I,-0.0004736691778442931+0.0002480131826444354 I,0.0004093188372938224 -0.0001378264201741173 I,0.0004912289664442608 +0.00005135279319269453 I,-0.0002505026489549314+0.00009790742321059786 I,-0.0002977861694590994-0.00005736150856188664 I,-0.00006531871484743686+0.0003892728950524639 I,-0.00002360377745577551-0.0001427680027986169 I,-0.0003062846987154245-0.00008877961259035001 I,-0.0002884827028186416-0.0003214158262691443 I,-0.0001172584051696187-0.0004308987964323075 I,-0.00004067562447196058+0.000308367154255888 I,-0.0001685926878735236+0.0004650464892825716 I,-0.00002918248313403478-0.0004258319208784406 I,0.0004399860254625083 -0.0003323100165101998 I,-0.0005867584884709795+0.00008446750904692594 I,0.00033371687005504 +0.0005413944425705641 I,0.0001199567242450875 -0.0002684268745091008 I,0.0001536181922619046 -0.0005211261568706057 I,0.0002817746186715846 +0.000262202757105633 I,0.0005261170939416758 -6.495815041203769*10^-6 I,0.0001200590086898601 +0.0002318559794911623 I,0.00001590480747235858 +0.0003098467287179171 I,-0.0002057346015775702+0.0001276261768962926 I,-0.00007139484862665404-0.0004118353754412325 I,-0.0004945792231489234+0.00004258479321389807 I,0.0004788559103757505 +0.0002775132716668975 I,0.0005181547887899585 +0.0002733318898376625 I,-0.0005227452305973265+0.00009356591672430079 I,-0.0001954172925244003-0.0003617302944011097 I,-0.000119036898848477+0.0003511263288661161 I,0.00002990324651370248 +0.0002211524734733751 I,-0.00002563302063347779+0.0002491443426693668 I,0.0001236801081562901 -0.0003508763974496507 I,0.00004021914360068449 -0.0002177988226337537 I,-0.0001506359524764674-0.000323560141396736 I,0.0001873240557534933 +0.0001435868781546594 I,0.0003307548589932951 -0.0004595305077628662 I,-0.0002628610345276387+0.0002472316159469372 I,-0.0005767999313849666+0.0002396222212560091 I,0.0001787239881603186 -0.0002697051394739873 I,0.0004134749846524134 -0.0001107360422521387 I,-0.0001993034104861166-0.0003926085577878759 I,-0.000114578158140801+0.00001671540083062921 I,-0.0000690669379754837-0.0001473526672542246 I,0.0002211552461232911 +0.0003381715144008662 I,0.0000898080255271135 +0.0001765670509721561 I,0.0002917151476921646 +0.0002427699656689138 I,-0.0002509360040076631+0.0000827731420719609 I,-0.0002289996065273297+0.0002865703786622432 I,0.000456190220781354 -0.0003529047552782447 I,0.00007120825314613119 -0.00004054532831138487 I,0.00002055283706638969 +0.0004784331383179951 I,-0.00008907292168898467-0.0003143066383844753 I,0.00001816279214940424 -0.0002331416635984434 I,0.0002288274178943432 +0.0003480413825428155 I,0.0004041143085916899 +0.0001348598858451715 I,0.0004557336937937844 +0.0001142777226269428 I,0.0002521245899566343 +0.0003283822593704143 I,0.00007629661311898332 -0.0002663986624286111 I,-0.0000925083660350587-0.0002560829767157564 I,0.00006567254674393535 +0.0005113204097848918 I,0.00005800021034440238 -0.0001162433390702491 I,0.000432488505769539 -0.0003766486657379845 I,-0.0001839550678046679+0.0003308308587806514 I,-0.0002791542471203145+0.00006673837078940097 I,0.0002837898403726424 +0.0002282533895924435 I,0.0001370252592709501 +0.0002004381867439201 I,0.0002730808880901898 +0.0003407881442510324 I,-0.000015848089802409-0.0001957273139656583 I,-0.00006464059677312048+0.00001154916203946824 I,-0.0002429345057144812-0.0004667084754023089 I,0.000415583541579713 -0.0001323276569696651 I,0.0002758614313539402 -0.000296125843885085 I,-0.0006331398428113063+0.0003131639771766767 I,-0.0003298920321362616+0.0002154442026691036 I,0.0003183741780753071 -0.0004954739518535855 I,0.0002706001416909121 +0.0001529755203285883 I,-0.0001448000830207606-0.0003782875328414755 I,-2.070416431044662*10^-6-0.000182403508651532 I,0.0001311322652074928 -0.0003143805934259299 I,0.00005334447013750546 +0.000347842378623461 I,0.00005528581899670235 +0.000322234563937234 I,-0.00002915772785302311+0.0002949909886734518 I,-0.0001732830243189759-0.0004204594318504136 I,-0.000525836463393839+0.0002719736648999891 I,0.0004286630539560251 +0.0002351183333909744 I,0.0006547869969776864 +0.0002225199805803083 I,-0.0005218524341876819-0.00002757821256482211 I,-0.0001782425864154831-0.0003430297293091968 I,-0.0002509690883868+0.0002659600865953168 I,0.00001556560654391778 +0.0002275529949814092 I,0.00003544933436142267 +0.0001816835087156535 I,0.0005049280226835369 -0.0001463679065553178 I,0.000331750327429539 +0.0001449480531791592 I,0.000123845237390009 -0.0005557775818975813 I,0.00007208296234858429 -0.0001873331253418229 I,0.000525235281786833 +0.0004956632928028171 I,-0.0006576021894268093+0.00009029012081086866 I,0.0002766641230703293 -0.0004787242223324364 I,0.00002550436209040213 -0.0004111717146633952 I,-0.00006354422077894695+0.0005694235497872349 I,0.00004840662718259175 +0.0002332738274535944 I,-0.0001796740534009353-0.0004511441195840357 I,-0.0003152326351934545-0.0003209895306615796 I,-0.0003492205252373108+0.00003557212621799931 I,-0.00006058407248988386-0.00008667217622972394 I,0.00002120825855643586 +0.0004602556437757239 I,-0.0003308621285742418-0.00004607262180187288 I,-0.000309303925577007+0.0002248701284706021 I,0.0005373190059014297 -0.0000745224288351829 I,0.0004993150914863311 -0.0002316886603263102 I,-0.0004523977642314921+0.0003593333091785726 I,-0.0002808161464205033-0.00002706013767933007 I,0.0002351298410267236 +0.0004535481274120509 I,-0.00003770056400873408+0.0001333608729476802 I,-0.00005149173932191491+0.0003348733620868743 I,-0.0001652936829472102-0.0002212621309503679 I,-0.000151136435193-0.000125548976062928 I,-0.00008437496850588201-0.0002370873768710127 I,0.0002786756132842679 +5.828155500200104*10^-6 I,0.00008430313340168036 -0.0001527726461635328 I,-0.0003321485648141321+0.0001307710658513234 I,0.00006519875604344831 +0.0001234274662449122 I,-7.629352031673768*10^-6-0.0003284572472229328 I,-0.00007972650609823056+0.0001438526422692709 I,-0.0001094714469779409+0.0002608086318787909 I,-0.00008310544947961643-0.0002373325888675265 I,-0.0002758988727676789-0.0001100057533195228 I,-0.0002375987232235907-0.0002409823913414379 I,-0.00008400235381322396-0.0003004853119570589 I,-0.0000829047777233322+0.00008740126561248039 I,0.00003066113875274624 +0.0002177374261819633 I,0.0001142334874865877 -0.0003304569121742847 I,-0.00009390843250668614+0.00003044574373468317 I,-0.000352426168758566+0.0001195687124061138 I,0.0001973929517145917 -0.0001430752764446024 I,0.0001638669872473062 +0.00003357492433742201 I,-0.00009086246534185472-0.0001893006154745251 I,-0.00002472772111553544-0.0001402224719194474 I,-0.00005439306496362396-0.0002334598064078907 I,-0.0000584571875225758+0.0001208363299478942 I,8.37568607458595*10^-6+0.00002791386040250523 I,7.718127877791455*10^-6+0.0003002537339741438 I,-0.0001992327034172233-0.00006349716702867483 I,-0.0002637441406955087+0.00007924754121314183 I,0.0003432060730515278 +0.00001803931992607678 I,0.0002624698215346665 +0.00004308233416211473 I,-0.0002451669603000642+0.00006169808404211745 I,-0.0001166151401306467-0.000110345138805349 I,-0.00008919509585065811+0.000221727914822036 I,-0.00004028460926975631-0.00002796136273710821 I,-0.0001924395988819614-0.0000395122777332064 I,-0.00003612092501603539-0.0001773563124041251 I,0.00003142427845175498 -0.0002110186748573504 I,0.00003513414433103983 +0.0001239332661814444 I,-0.0001816457041142594+0.0002376990117113303 I,0.00005890306870394516 -0.00007359503812136138 I,0.0002856329404226931 -0.0002219112414042461 I,-0.0001928446814538613-0.0001230367761965103 I,0.00008648762239205407 +0.0003950505655262705 I,-0.00002696278801536157-0.0001155481917960755 I,0.0001368284575802293 -0.0001581817753194181 I,0.000186447646546409 +0.0001746216968896019 I,0.0003234352712751878 +0.00004753123638363515 I,0.0001015495432124407 -0.00002796717259700499 I,0.00009546053388614087 +0.0001078650864556214 I,-0.0002403228480612019-0.0001001367671395801 I,-0.00004391163873739739-0.0002959744764333279 I,-6.619007023168206*10^-7-0.00003034171790550772 I,0.00008408584539942144 +0.0004076988446646982 I,0.0002145386120045748 -0.00002815028787827645 I,-0.00042335847887128-0.0001336440032788743 I,0.00007039792380193456 -0.00001339116777724847 I,-0.00001275857894066275+0.0002662843523191068 I,-0.0002157465195277506+0.00001853967347989057 I,-0.000275666246679741+0.0001184973217842983 I,-0.00002347406420194801-0.00006935832646636059 I,-7.038197351635802*10^-6-0.0001486583865923951 I,0.000113180680635197 -0.00001109727829092398 I,-0.00008074364369383807+0.0002088859651378576 I,0.0001928431786420028 -0.00002379069180493575 I,0.00006061985364065121 -0.0001148694866510596 I,-0.0002187275365201297-0.0002420561219370833 I,0.00002930748635363251 +0.0002351164934650738 I,0.000113479748864366 +0.00008252836681782872 I,0.0002623231233332787 -0.0001034166075723094 I,-0.00005321338369244021+0.00009686044357204147 I,0.00008168561868432709 +0.0001414743695542433 I,-0.0001175788734018079+0.0001936559348248069 I,-0.0001400215736881619+0.0001175339446243379 I,-3.166485193094499*10^-6+0.00008892035363937569 I,0.00002783009559662898 -0.0001130695220784599 I,-0.0002399569366020085+1.671254099699281*10^-6 I,0.0002047586905458403 +0.0001026399648624045 I,0.0001930169666264338 +0.00007056469271342428 I,-0.0002805118558506953+0.00007987044118529596 I,-7.328265528860694*10^-6-0.0001032846532859354 I,0.00009642295686662346 +0.0001552101542633764 I,-0.0001174404669630656+0.0001546678740462564 I,-0.00006811553855235342+0.0002790471821288996 I,-0.00003891783352891871+0.00007032838150229301 I,-0.0001554443191622596+0.00004156441812294749 I,0.0001114153411331442 -0.00009091931692083573 I,0.0002038279720611313 +0.00002552441507317491 I,-0.0001386242417135662-0.00002568562127762029 I,-0.00006672477047111664-0.00001739518255404864 I,0.0001014899465599428 +0.000222577347476422 I,-0.00005257544945661669-0.000116419684240395 I,-0.00005830275933201901-0.0001117917979952154 I,-0.0001390024577896083+0.0001148072093881661 I,-0.00004094976611507695+0.00002585348164505899 I,-0.0001296753108997989+0.00006538842850754893 I,0.00008892976830232976 -0.0000675159958250007 I,0.00002617758963803767 -0.00006142738505014359 I,0.0001507252434272958 -0.0001011314077952815 I,0.00002700694090033637 +0.0001693566351523476 I,0.0001483657866204133 +0.00008482569499893732 I,-0.0001054463684399898-0.000224451382251613 I,-0.00009989766720720864-0.0001422351332908298 I,0.0001677160438194924 +0.0001260664116589514 I,-0.00004176371175144836+0.00009989510385192789 I,0.0001439687786047874 -0.00004917651882351256 I,0.00004516871609555958 +8.869119260266809*10^-8 I,0.00008332302852716598 +0.0000508659080512078 I,-0.0001382819068882305+0.00004742049458056845 I,-0.0001378315987329896+0.00002360492882913135 I,-0.00005430513008659448+0.0000470922037785585 I,0.0001277278628867702 -0.000024552120805361 I,-0.0001918890001654539-0.0001499590848044915 I,0.00002825916948670276 +0.0000371223264992719 I,-0.00002874827260329355+0.0002861914250840115 I,0.00004400339854167002 -0.00007200420291147836 I,0.00002652156006167786 -0.0001790147817418739 I,-0.0001558993034165104-0.000130295140617682 I,0.0000620435966252934 +0.00007707502692866759 I,0.0001092579140197784 -0.00002711722667610557 I,0.0002367934749224329 +0.00009574516340674771 I,0.0001311137384971556 +0.0001892761520432193 I,0.00008730860044962261 -0.00008664819763099473 I,-0.00003731577944418617-0.0001792732887285242 I,-0.00003998535529778309+0.0003481223884700982 I,-0.0000492284042630968-0.00002141197767261183 I,0.0004089814775588609 -0.0001449442918973311 I,-0.0001448218709203981-0.0000557774767021056 I,-0.0002197407493460435+0.0001353044775864591 I,0.00001100340171641018 +0.0002925962967672227 I,-0.00001455554527739379-0.0001440045976369247 I,-0.0001311796556031067-0.00006344509236136541 I,-0.0001909857392194104-0.0001852092850883983 I,-0.00004748532355328074-0.0001737952360482467 I,-0.000117647541354428+0.0001343985790579814 I,-0.00009348939133116223+0.00003093590184176763 I,-0.00007726084380673559-0.000152769477746032 I,0.0002629373202737935 +0.0001051651594514271 I,0.00002895738678006075 +0.00001969037998500881 I,-0.00025015837275886+0.0001310046456245062 I,0.0000333719110931872 -0.0001207641480628263 I,0.0000831324590674984 +0.0001144204034171306 I,-0.0001010205665406219+0.0001564825060483597 I,-0.00006258536266716657+0.0002372605396046079 I,-2.559277353461175*10^-6+0.00005964474116158415 I,-0.0001075726931647112+0.00004075203637801409 I,0.0001257850667774829 -0.00005923338145387704 I,0.0001197925934535724 +0.00002312944688286119 I,-0.0001522890551001336+0.00007032740619363154 I,0.00004576590796842876 -0.0000896444639217165 I,0.0002275423333406584 +0.0001189450115669099 I,-0.0001320612773699195+0.0000346745316477967 I,-0.0001863186341359485-0.0001132908974419917 I,-0.00004594843693306604+0.0002523398684444658 I,6.871785658000499*10^-6+0.00007713122291345254 I,-0.00008252465105713203+0.0001837797195175991 I,8.22114535089678*10^-6-0.0001980794761165761 I,-0.00005294947986620986-0.0001289970322909989 I,0.00003794230595608876 -0.0002077231181209305 I,0.0002052007132148786 +0.0001526241866019141 I,0.0001124980273992045 -0.0001182322825068772 I,-0.0002341625359020009-0.00009365572182022124 I,-0.0001064822699005401+0.0001565273996391296 I,0.0001702979480683629 -0.0001288266087627417 I,-0.00007457438400503073-0.00003377096137048493 I,-0.0001674827545317606+0.0000322258875466891 I,0.00007692329781343077 -0.00007270087587344721 I,-0.00004912882579426159-0.0001038929973395253 I,0.00005222641160384936 -0.0002008764128247823 I,0.00009480001547470949 -0.0001582875634502612 I,-0.00004053405522303524-0.00003838202477089489 I,-0.00007948478268887541+0.0001449672279370125 I,0.0001902332796535244 -0.00009414652382323222 I,0.00005406207503948581 -0.00007604712905494865 I,-0.0002126160979730374-0.0001110058976692815 I,0.00005590970839474741 +0.000210332255386519 I,0.00004067581793184042 -0.00005266190533399792 I,0.0001217236561535743 -0.00008896130731443223 I,0.00007046647979717068 +0.0001767902686944455 I,0.0001893263018920488 +0.0001875281567604787 I,0.0001089606140766969 -6.249220157740536*10^-6 I,0.00001709519199391053 +0.00008547133221567367 I,-0.00006187603969943324-0.0002980847371344715 I,0.0002078936751069035 -0.0001077181536924102 I,0.000195571869099666 -0.00003075589481125218 I,-0.0003920574443745468+0.0001891874584425564 I,-0.00004654972275615392+5.356294315897507*10^-6 I,0.00002901335088754037 -0.0003234964672666974 I,0.0001086527717563285 +0.0001806203647273847 I,-0.00003200358325212543+0.00004689846310729348 I,-0.0001676818929026943-0.000153594565266861 I,-0.0002313950147088107-0.00003772804918074558 I,-0.0001391420240161917+0.00009836206659800053 I,-0.000124033715144924-0.00004148252024430482 I,0.0001779463646966484 +0.0001366303299327459 I,0.00004069822647574217 +0.0001719653408473701 I,-0.0001192854817852838+0.00003969194239911183 I,5.525240160054264*10^-6-0.0002789677004163551 I,-0.00002367505807412653+0.0000770124019050218 I,0.0002155926517033238 +0.00008574468951391204 I,-0.0001281363712711161-0.00002363427654191483 I,-0.00002368230059031465-0.00006841458547459563 I,0.0001531668430804049 -0.00003046413951843309 I,0.0001393355606364688 -0.00008363749975419697 I,-0.00002477777488640181-0.00006086210442537182 I,2.035835164496049*10^-6+8.944038745242652*10^-6 I,-0.000136108452165806+0.00002732493885746197 I,0.00002051839861028873 -0.0001122222530389981 I,-0.0001209544627554291-0.00008584044310373502 I,0.00006420780273894415 +0.0001500013652436169 I,0.0001147962497263865 +0.0001711482277131142 I,-0.0001449707783346407-0.0001049814665472982 I,-0.0000251024227713077-0.0001161385808503226 I,-0.0001298088381105496+0.0001151836079250361 I,-0.0000382673292829049+0.00002266164346390349 I,-0.0001041766512795261+0.00004234975525851061 I,0.00007458057987669189 -0.00007407060744574052 I,0.00002132377424016917 -0.00005050598382227479 I,0.00009194950144138112 -0.00007918408775550194 I,0.00005287860429167255 +0.00009662517403807675 I,0.0000458982227790129 +0.00004086007968596953 I,-0.00009057577266190046-0.0001300378130525502 I,0.00002094009979371734 +0.00004862952278267637 I,0.00009441956572037749 -0.00003364808052948044 I,-0.0001195403300022013-0.0000243174576100546 I,-0.0001164180861051327+0.00006997264215876441 I,0.00008215718913366556 -0.00005114608301823401 I,-0.00002566715518075626-0.0000696011729229936 I,0.00004572739438704604 -0.0001970828776828155 I,0.00008140203571719835 -0.0001334982761888294 I,-0.00004013335418542592-0.00008327147691033498 I,-8.49107433831118*10^-6+0.0001148292146059901 I,0.000198809006685722 -0.00007037847867610359 I,-0.0001204735608977682-0.00007286249148706102 I,-0.0001826203678223027-0.00005516883281053988 I,0.0001865134538454395 -5.246482377322477*10^-7 I,1.402079704545005*10^-6+0.00005680260636674873 I,-0.00002424065546444972-0.0000912305750079734 I,0.00009250464994091639 -0.00007678996228542174 I,0.0000701254661097363 -0.0001496993742286211 I,-0.00001892233796531539-0.00007747549999371386 I,0.00006288026195735812 -0.00005357909026639627 I,-0.0001426962561556099+0.00007919315976623632 I,-0.0001145185391095675-0.0000464487301626001 I,-6.054692726056713*10^-6-0.0001030299943509407 I,0.0001739866957577837 +0.0001318100416641586 I,-0.00001755373569171237+0.00001769437893999972 I,-0.0001557493942081007+0.00005918343475456677 I,0.00004718730083125386 -0.0001252071716168542 I,-0.00003510708161750522+0.00003442359110813787 I,-0.00002035098306707548+0.0001190081650006388 I,-7.87938727309285*10^-6+0.0001007234439939983 I,0.0001667551802468249 -0.00005926435494722196 I,0.0001062990597042227 -7.81557994140327*10^-6 I,0.00006098852400499385 -0.0002007079094181956 I,8.3706785225481*10^-6+0.00004835485706756231 I,0.0002769703719215426 +0.0001101905427344748 I,-0.0002158947846172769-0.00005566428056218013 I,-0.00005927168705636683-0.0002436269314464277 I,0.00008757151390733524 -0.0000166087224760593 I,-0.00001038830683814187+0.0002207416732710214 I,0.0001164239021307684 +0.00005955626668042627 I,-0.00003252412793670575-0.0001319173116674122 I,-0.0000829354360566067-0.00006125478358908802 I,-0.0002289049902106529-0.00006871323021746887 I,-0.00009904815193851761-0.000155376301004575 I,-0.00009970014027477691+0.0001572293598740055 I,-0.00004822565926626478+0.0001132338419279511 I,-0.00003887139260581976-0.0002572080826633497 I,0.0001931116992233718 +0.00003207001731222699 I,-0.0002228612375915585+0.00007858180125300368 I,-0.00002540216793488036+0.0001462225729892666 I,0.0001190791603444141 -0.0001467092109237045 I,0.0000480667647416953 -0.0001049322621184828 I,-7.855021945975824*10^-6+0.0001692860957826493 I,0.00006408361281282445 +0.0001557063301825261 I,0.00009531815009464109 +0.00005450263986721941 I,-0.00001394581583552632+0.00003946445663831059 I,0.00005675245383035234 -0.0001412178608318168 I,0.0001136034618953962 +0.00006788530851427053 I,0.0001099314891610321 -0.00005176653579478551 I,-0.000170655533102084-0.00005582668679639824 I,-0.0002146449479240042+0.00003807993124860164 I,0.0002056765359005779 -0.00004299816314994788 I,0.0001007497178894942 +0.00001342452786130415 I,-0.00002656006686535154-0.0002128092478008831 I,0.00003652704470144189 -7.768910264732552*10^-6 I,0.00008105690297192643 -0.00008836655102043172 I,0.00007211530954755397 +0.0001303171453589201 I,0.00004794962149394424 +0.0001023484270571587 I,0.00004676098058376411 +0.0000822919173204712 I,-0.0001064441872969421-0.00003605097411369369 I,-0.00007574535195775764+0.0001284857621265051 I,0.0001747003371468579 -0.00009645982176849241 I,0.00005982779432078269 -0.00008848556018900826 I,-0.00002825791291564744+0.0002274369552053755 I,-0.00003852488915939527-0.00006915237220230585 I,0.0001115610855849918 -0.000065949558822758 I,0.00007805997803822864 +0.0001605948186283552 I,0.0001977410524449942 +0.0001349211475657684 I,0.0001058090360019114 +0.00007173316541623866 I,0.00002020907005914125 +0.0001581381986231383 I,-3.765040885749947*10^-6-0.0001388087566648803 I,0.00006880823187101946 -0.0001530333840593167 I,-0.00003953605278242558+0.0001578739061082243 I,-0.00004329034809563866+0.00001251890163335114 I,0.0002380581901113503 -0.00004884656163046971 I,-0.0001027301053569103+0.00003595510682023625 I,-0.0001524836626937134+0.00002029667001837129 I,0.00006366064320134141 +0.0001805546133973529 I,0.00004401947036299097 +0.00004693750495232246 I,0.00004389564082420372 +0.0001428308398882022 I,-0.00004317402130662274-0.000173249260642285 I,-0.00004489601519112462-0.0000916943386811003 I,-0.0001130075325644293-0.000212957970296154 I,0.0001995065905495539 +0.00004848465588534309 I,0.0001803518493641644 -0.0002302523733888312 I,-0.0002796644096284656+0.00006749909463390724 I,-0.0002937078902212928+0.0001129024764602545 I,0.000247580616780943 -0.0001215290713126719 I,0.0001193584980317247 +0.00001683381398994905 I,-0.00003573594419027453-0.0002616239486768057 I,0.0001129085692131291 +1.861882891971746*10^-6 I,0.0002494138543016004 -0.0001011507851402481 I,0.0001136991620755128 +0.00008520369990517492 I,0.0001029610516788188 +0.0001793846003009247 I,-0.0001880416951104838-0.0000676306330494118 I,-0.00001036162759978677-0.0003469599911587675 I,-0.00002950135185652832+0.00002554933834427139 I,-0.000022591802866141+0.0004385153228072958 I,0.0002291380111618206 +0.00001740737693448484 I,-0.0003900323335486226-0.0002374303782963143 I,0.0001215219280719072 +7.85601411438757*10^-6 I,-0.00001031455421979894+0.0002021152815251244 I,-0.0002526674822776258+0.00005612227561840728 I,-0.0002272814864063173+0.000226960967160506 I,0.00002959570026270898 +0.00009543970844352672 I,-0.00009076014505500355+0.0000382745002267825 I,0.0002342518313265174 -0.0001840775731749885 I,0.0002416973669606571 +0.0001003029962918767 I,0.0000908009283372565 +0.00009425323074296942 I,-0.0004207873545280558-0.0001344148780988663 I,-0.00004467405686249071+0.00008695374549944673 I,0.0002900228486343669 -0.0002903394646887552 I,-0.00002415676917672796+0.0001445833943057832 I,-0.0001626323455169572-0.00009075889707710628 I,0.00003912159584967421 -0.000221362733851075 I,-0.0000289468974789409-0.0002762128289605717 I,4.087218220317416*10^-6+0.0000312287413575306 I,0.00009185502249089614 +0.00004107154760989542 I,-7.723939741291247*10^-6+0.000150251283282161 I,-0.0001679035699319843-0.0001163804134416275 I,-0.00007799720801054293+0.0001923340801192452 I,0.0001290823067857676 +0.00003465070852355667 I,0.0002854089691856014 -0.0001452658490634923 I,-0.0002055032775494323+0.00003038226635342035 I,-0.0001220712041408319+0.00007672898685655587 I,0.00007018738565963285 +0.0002439190060729835 I,-0.00005368485756919195-0.00004507220033915488 I,-0.0001013227215630877+0.00006369400932295248 I,-0.0001517521403885817-0.0001259089922993773 I,-0.0000856207340227431-0.0001202618384049646 I,-0.00008719320518333983-0.00001832406258060843 I,0.00009216921271337124 +0.00001622877884739494 I,0.00001075854955685579 -0.0001969891777945635 I,-0.00008249624828659006+0.0001758667791716418 I,-0.00006099603015709883+0.0001496204882941507 I,-0.0000779635605839076-0.0002183738789846468 I,0.00009817141199424718 -0.00001539434001296004 I,-0.0001169186304756541+0.00006754729504448247 I,-0.0001383027106736007-0.00006624544860327834 I,-0.0002219367506784929-0.00002113810467693211 I,-0.00002421816331982631+9.023007656481797*10^-6 I,-0.00003296886862486136-0.00008506564569395528 I,0.0001265116894099193 +0.00007153700537667381 I,-0.00001400344493445436+0.0001611664326960819 I,-5.394691641133725*10^-6-0.000038297139114906 I,0.0000366351535817796 -0.0001080652087840834 I,-0.0001117747309109004+0.00007433119330789437 I,0.00006578348984794089 +0.00001196368379312331 I,0.00003679298908836233 -0.00009495771486615423 I,-0.0001004100397039128-0.00007401141092820397 I,8.034415455068647*10^-6+0.000040427341737986 I,-0.00001574778951950195-0.00002169211217400969 I,0.0001634349134119468 +0.00001434769658661042 I,0.0001146827341090689 +0.00002656449332241368 I,0.0001015542185640517 -0.00006460486125687271 I,-0.0000853609601822257+0.00004114278260705247 I,0.0001553340026635376 +0.0001548035905314879 I,-0.00001879673104751312-0.00007147903050333496 I,-0.00002210144580026927-0.0002224445963129142 I,-0.00003264220213531428+0.0001040349790039505 I,0.00005292278467823055 +0.0000799786276622403 I,0.0001514240186230936 -1.401778578219253*10^-6 I,-0.00009485570887291059+0.000085518479287812 I,-0.00002352586493585551+0.0001940513628702232 I,2.429849088430452*10^-6+0.0001696391535754994 I,-0.0001061296823071236+0.0001320699052847758 I,0.0002080799127741151 -0.0001241562719902194 I,0.0002160174493996376 +0.00001611654557951775 I,-0.00004803359639071514+0.0002024204470248472 I,-0.0003229230977623878-0.0001877091331970163 I,0.0001707428314578021 +0.0000819896370634263 I,0.0001964724743065489 -0.000242935415718064 I,-0.0001846600675806514+0.0001267267962049638 I,-0.0001945566504469021+0.0001171279186639222 I,0.0000447284106205262 -0.0002549649520108001 I,-0.0001723843658721245-0.0002585292493131311 I,-0.00005449428483879874-0.0001928278237421658 I,0.0000917186633872656 -0.0002018268566704885 I,0.00002119012240402385 +0.0001176730057272551 I,-0.0001920487056938775+0.0002122161907321798 I,0.0001947580332758521 +0.00001694021515910436 I,0.00007064920274144868 -0.0002239566147768894 I,-0.0002255111713012791-0.0002845908632845154 I,0.0001787222926608638 +0.00026788751262564 I,0.00003655013939435743 +0.0001250333666279544 I,0.0002341810905142053 -0.0001569345091222499 I,0.00004320914376125128 +0.00002814864543385705 I,0.0001371030208428938 +0.0000392998545181893 I,-0.00011613550169703+0.0001576782506242157 I,-0.0001166460798596506+0.0001152488455891605 I,0.00001040631801790473 +0.00008504987004345885 I,0.00006507377535549787 -0.00006919163026466733 I,-0.0002248128245739677+0.00006440044331155857 I,0.00005028243481789482 -0.00009493104033460361 I,0.0001734333334397276 +0.000115450885106192 I,0.00003439164323152636 +0.0000296712231654499 I,-0.0001655134616339906-0.00006165769972946869 I,-0.00006534042950434712+0.00005569413518666778 I,0.0001229095922897324 +3.647835737549383*10^-7 I,0.00008464070946919647 -0.00002662630643554732 I,0.00006573516017339292 -0.0001478295631552902 I,0.000084198070573355 -0.00003570704017019676 I,-0.00007389930064504016-0.0001414548406399395 I,0.00002978231790192913 -0.00001279268303903093 I,0.0002015373508003885 +0.0000185695321165663 I,-0.0002239922401658149-1.849465642288103*10^-6 I,-0.00007637492626388388-0.0001418565328594704 I,0.0001594010533360459 -0.00007192090412285341 I,-7.771480717797007*10^-7+0.0002045718659528386 I,0.0000715525164841539 -0.0000489202250038438 I,0.00003529413723991822 -0.0001523752218531933 I,0.00004139499365509776 -0.0001471544780400859 I,-0.0001631103895544614-1.983958125352872*10^-6 I,-0.00003518012530479405-0.00002300845780746734 I,-0.0001498567981540548+0.0001380856423670304 I,-0.00005491977832140793-0.0000904483330402773 I,-0.00006552178297107141-0.0001022601990806734 I,0.00006756956646555672 +0.0001620163546710631 I,-7.587583559381299*10^-6+3.46099372751233*10^-6 I,-0.00008520150377502129-0.00005338690364673873 I,0.00009646310490763003 +0.00004915698427018265 I,0.00006031598601394331 -0.00001695671069631169 I,-0.0001000949928270741-0.00003219699507312182 I,-0.00006376631837224267+8.664969694293528*10^-6 I,-0.0001065967905174567+0.0001566945127423622 I,-0.0001107117458038087+0.00007123044455295904 I,0.0000811518885035508 +0.0001349638996803665 I,0.00003039007197787672 -0.0000299641355717464 I,-0.0002178933592190212+0.00008384114339765277 I,0.00003329840711000702 -0.00007652014990016789 I,0.0001829097380716902 +0.00008754903746152273 I,0.0000186138304579877 -0.0000635795146709912 I,-0.0001262294949693461+0.00004082030957836737 I,-0.0000525904192562438+0.00009336770306595565 I,0.00001688857139023147 -0.0001579113105608472 I,-0.00009915145574596718-0.0001611032670660852 I,-0.0002009132589324044-0.00006882023629940255 I,-0.00008821781432870324-0.0001349848496157183 I,-0.00007028039816096322+0.0003007872653915813 I,-0.0001209871271652238+0.00009513305547273571 I,-0.0002682852269521446-0.0001971632585972024 I,0.0004115748244602307 -0.00007213899263645702 I,-0.00008322636816883303+0.0002214630206794026 I,-8.761102578539771*10^-6+0.0002520520356949918 I,1.799321714697879*10^-6-0.0003292537754889504 I,-0.00007162482027946724-0.0002045876961430115 I,0.000174465508133626 +0.000243151307970644 I,0.0003139308779138099 +0.00005005001063204616 I,0.0003397230373845368 +0.00007362442787870726 I,0.0001973771863874453 +0.0002610670973969377 I,-0.00003310716606772857-0.0001924208216034894 I,-0.00003719239359885074-0.0003322602499790242 I,-3.529619804576803*10^-6+0.0003390381120915691 I,-0.0000584215241125554+0.0001728766487671045 I,0.0004817886490237916 -0.000164648278120286 I,-0.0002974264299795115-0.0001327872713308648 I,-0.0001727519903642642+0.0001457126214506807 I,6.846166654039061*10^-6+0.0003667037211685636 I,-0.0000955222564093405-0.0001420120636192321 I,-0.0002159032913250796-0.00002989711537482813 I,-0.0001914731935047999-0.00018178337154209 I,-0.00004951020297677792-0.0001970083827694746 I,-0.00007172530884320472+0.0001309952509284151 I,-0.0001078089325389319+0.00007880018551469656 I,-0.00003576915327629953-0.0001359724089249292 I,0.0002573918389074764 +0.0000607915910547073 I,-0.00001051296073181383-0.00003111191766577085 I,-0.000215851985703667+0.0001690053380825294 I,0.00004208231701894536 -0.00009719028239095312 I,0.0001153719449832918 +0.00008346511296829191 I,-0.0000898723398230184+0.0001575475964651734 I,-0.00003484387802716308+0.0002292621858928124 I,-8.735176631697011*10^-6+0.00007581580499151883 I,-0.0001037898930640627+0.00005441079939936578 I,0.0001117851825370349 -0.00004958050646028929 I,0.000104535492573088 +7.510800583559313*10^-6 I,-0.0001475447831668536+0.00007666284970827522 I,0.00005352952507622157 -0.00008463626294062193 I,0.000215690216905141 +0.00009148862472125994 I,-0.00012022134294258+0.00005376797171421682 I,-0.0001718796892763707-0.00009792243513760787 I,-0.00002132439728903801+0.0002160081964165382 I,0.00001735000040564228 +0.00007982513842293929 I,-0.00004381989388353561+0.0001696698816913393 I,0.00001257277789216873 -0.0001651534179052922 I,-0.00004168697074314931-0.00009169336395348958 I,0.00002221964226760729 -0.0001908612225601659 I,0.0001764556607364413 +0.0001024430839640898 I,0.00009197088415163486 -0.00007817253074907713 I,-0.0001974244096683443-0.00007120077107719856 I,-0.00005320018957173531+0.0001111421617052802 I,0.0001217932719228937 -0.00009269149537038555 I,-0.0000833879951502449-0.0000181719613019599 I,-0.0001131682697569319+0.00005266736513735427 I,0.00007117236665608499 -0.00004727197427676526 I,-0.00002032885604567143-0.0000524340785618429 I,0.00003239907148994641 -0.0001853033291261997 I,0.00006483037632610015 -0.0001323279117732737 I,-0.00005955832856613846-0.00007993903024203885 I,-0.00001258413655349101+0.0001075231645436037 I,0.0001878399381698479 -0.0001168919129404256 I,-0.00004039484524314183-0.00002660949005124546 I,-0.0002228937522696672-0.00006127399633759731 I,0.00009748047716990111 +0.0001047331354188348 I,0.00007319284886834435 -0.00001894710676255413 I,0.00008255042293363931 -0.0001359410846498527 I,0.00005821090903643516 +0.0001192840849478692 I,0.0001825853052628275 +0.0001089755944263504 I,0.0001002348589709383 +0.00006331573364998678 I,0.00001924520675953032 +0.0001394825566803003 I,-0.00004634557231888303-0.0002026538944565808 I,0.0001559839474315966 -0.0001747883472010823 I,0.00005602763185559356 +0.00004718116435386633 I,-0.0002716242193632204+0.0001927248914384997 I,0.0001211139662239667 +0.00005570306307114441 I,-0.0000716470994844408-0.0003061626642782224 I,0.00002705017382194803 +0.00009438532605442058 I,-0.0001083227276793836+0.0001231104824901884 I,-0.00013194055231703-0.0001138299810225179 I,-0.0002150520954182539-0.0000273226917570789 I,-6.412966182510536*10^-6+7.417373162718323*10^-6 I,-3.468552203878233*10^-6-0.00004175388990668619 I,0.0001763597002495238 -8.366709876355338*10^-6 I,0.00001248237756400992 +0.0001380319614778203 I,0.00009210944285667006 +0.0001556157587326257 I,-0.0001586396205544854-0.0002416650036897738 I,-7.921145990403612*10^-6-0.0001393652272085157 I,0.0002167375343400972 +0.00001785108438380123 I,-0.00009862267737635185+0.0001764312649243924 I,0.00006348108732685657 -0.00002716336546968102 I,0.00006768021809489363 -0.0001452483149234146 I,0.00004166035491875932 -0.0001410539877938627 I,-0.0001621075149656681+0.00001522312342327646 I,-0.00005776365992941223+0.00001069484849907759 I,-0.00008291991426508585+0.0001931045840720551 I,-0.00007234527821766427-0.0001057650106320971 I,-0.0002572142964793009+0.00002803091124075325 I,0.0002519149675196471 +0.00003954008718045149 I,0.0002927318474076909 +0.00007354145273460202 I,-0.0002448366790201061+0.00008577126370889615 I,-0.0001654091186599446-0.0001167896381894729 I,-6.747495066718721*10^-6+0.0002494984953005209 I,1.913664427561235*10^-6+0.00008593830317284726 I,-0.00004451479563324082+0.0001706672208244241 I,0.00001466370443153718 -0.0001609060306344186 I,-0.00002709071323235356-0.0000699315309637197 I,0.00001824541971700411 -0.0002001819494448064 I,0.0001675306606794415 +0.00004815926891100478 I,0.00009951551927787786 +5.011904772046126*10^-6 I,-0.0002483805194609017-0.00005395023226744525 I,0.00007906944530535631 +0.000030650170835265 I,0.00007336789959756876 -0.0001423836091871169 I,-0.000163669119732158+0.00009950794145033564 I,-0.00006650682912758422+0.0002051600425555426 I,0.00003778420331591219 -0.0001481233186578816 I,-0.0001123174622188158-0.0000577393546360484 I,-0.0001466427411245344-0.00027635766074757 I,-0.00003853939909100284-0.0002430385886103663 I,-0.0001203492372687864-0.00005218805432942352 I,0.00009902974826656683 +0.0001552641981235522 I,0.0001654753351721655 -0.0002763052800334212 I,-0.0002069044770544793+0.00001942120841841075 I,-0.0002750241204548142+0.0001126111881088591 I,0.000223493234244835 -0.0001719997796243665 I,0.00005559907641330145 +0.00005688574065795213 I,-0.0001200445512404879-0.00008855120557763957 I,0.00004086911706253145 -0.0001689194617229013 I,-0.00005226450404438642-0.0002359866583559625 I,-0.00007622368607551334-0.00007971760264224089 I,0.00003625436453127398 -0.0001208115457046457 I,-0.00009311658089600382+0.0001991390900909681 I,-0.0001698135949528637+0.0000544372178685355 I,-0.00006451320983106467-0.0001168195125263557 I,0.0002885961352248894 +1.24080679824202*10^-6 I,-0.00003951703393306597+0.00002155832222916872 I,-0.0001038434373482326+0.0002011292489480944 I,-0.00002998135892515788-0.0001747816177136369 I,2.213727564061882*10^-6+0.00002944103406019868 I,0.00008465966106407257 +0.0001674440481999009 I,0.0001203512095157706 +0.0001304761137957642 I,0.000166323772349855 -0.0001712916528221657 I,0.0001250521437011016 -0.00004137245265168696 I,-0.0001146094119269836-0.000317181957702392 I,0.00007848741618384072 -0.00003886989297376386 I,0.0003976785192208385 -0.0000703646334159691 I,-0.0003499437677200139+0.0002209883606132337 I,-0.0001755423055691058-0.0002124040319265234 I,-0.00001628507906271693-0.0002113315072006856 I,0.0001873015684769083 +0.0002840712179131485 I,0.0001467919302785659 +0.00006775091745134216 I,-0.000241505104439152-0.0001429069966883388 I,-0.0002737975300313883+0.00005349729736796706 I,-0.0003128564950964709+0.0001221799200358674 I,-0.0002683029795221915-0.0001114058796938291 I,0.0001403498392478925 +0.0001900221394983031 I,0.0001319716894744711 +0.0002630396135075984 I,-0.0001676525338693474-0.0001995161256963284 I,7.573775797447581*10^-6-0.0002525894236582581 I,-0.0002589936979240968+0.0002632074240194538 I,0.0003236431173037312 +0.00005627920962774952 I,1.711351363002059*10^-6-0.0001338116695979518 I,-0.00009048907120344236-0.0002591386954862428 I,0.0001720748576869147 +0.00003690238490896219 I,0.0002147894654864893 -0.00008687255029781509 I,0.0001177153481798125 +0.0000497759590722032 I,0.00005858288453514891 +0.0001235365219939877 I,-0.00006045900328487661-0.00004134490692273531 I,0.00004694006903707872 -0.0001514249237725251 I,-0.00007982659358996553+0.00005422213260135515 I,-0.00007783686925931922+0.00008211506273940229 I,0.0001621181008764716 +0.0001197807316148491 I,-0.00002756933015139198-0.000169813105712591 I,-0.00008908651085177762-0.00001782243940666879 I,-0.0001563887501256371+0.0001087497531961151 I,0.00002761624633890381 -0.00008593104432677142 I,-0.00007139728629209074-0.00008673535437248912 I,0.00003306205698880462 -0.0001665451516956096 I,0.00007079164723837721 -0.0001138690340220759 I,-0.00003762417395365773-0.00007662858599378931 I,0.00002078133073651569 +0.00007221229864441671 I,0.0001863810091651776 -0.0000665961941738593 I,-0.0002091162207460914+2.666333198415979*10^-7 I,-0.0001500531335179162-0.00001608697200950919 I,0.0001811016141790008 -0.000160439064714165 I,0.00002876955020884474 +0.0001178960713672431 I,-0.00009043261919349578-0.00005634947067244443 I,0.00002681487340218278 -0.0001840016685312355 I,-0.00005719864959784432-0.0002330300766667546 I,-0.00006917874515035145-0.0000788626003112794 I,0.00005651743909153149 -0.0001137143842158693 I,-0.00009191714239483795+0.0001371608538350582 I,-0.0001356116898826308+0.00005182158215208467 I,0.0000837727871502634 -0.0001105603703245134 I,0.00004848705320915887 +0.00004526211218733257 I,-0.0001919535819279965-0.00006968406278453387 I,0.00007649106750507741 +2.583485842634961*10^-6 I,0.0001337491417721635 +0.00006461116053204543 I,0.00003739389979885472 -0.0001561120470287625 I,-0.00002680860449357501-0.00006878995557106894 I,0.00001108499610204944 -0.0001315487041413156 I,-0.000070899128228913+0.000153763131136914 I,-0.0000257498161414882+0.0000769764508791716 I,-4.469596961843782*10^-6+0.000233687962947654 I,-0.0001257997379006391-0.0001047261239200922 I,-0.0002575708129591069+0.00007917993046958826 I,0.0002550709714992676 +0.00004239394050450151 I,0.0002600790416608793 +0.00005970597685335023 I,-0.0002114412374652647+9.971470336882454*10^-6 I,-0.0001052847621573833-0.00007603949022054104 I,-0.00008121236220018623+0.0002039036485106135 I,-0.0000457112405429318-0.00004557927400710479 I,-0.0001863932760064994-0.00003677154249250554 I,-0.0000593868080337781-0.0001513040470909163 I,1.245133747248498*10^-6-0.000195361289510289 I,0.00003958791008803878 +0.0001125406029096403 I,-0.0001234288512437289+0.0002355663030402716 I,0.00005183098335529606 -0.00009515839858206485 I,0.0001927309640765137 -0.0002168966116019134 I,-0.0002243257565453184-0.00007014506741556605 I,0.0001554870236075996 +0.0002962690695442032 I,-2.786666854920361*10^-6-0.00007490327623727875 I,0.00009820049866404983 -0.0001857105907202144 I,0.0001556245514674086 +0.00009733827843379128 I,0.0002472064640318266 -0.00002151520255877238 I,0.00003521091742649267 +0.00002295456509506319 I,0.0000396835652647882 +0.00008737911872651846 I,-0.0001610675134359337+0.00003633416518263321 I,-0.00007086592680860825-0.0001941562713083196 I,-0.0001326555459901354-0.00001696466853136274 I,0.0002114019275608814 +0.0001985831708673195 I,0.0002008061449240899 +0.00003623529736436338 I,-0.0002946539081814054+0.00004012650044890806 I,-0.00004065334990974645-0.0001068271909397851 I,1.809083698668407*10^-6+0.0001943115574326848 I,-0.00005198583910429679+0.00008417409965074684 I,-0.00007930676127630931+0.0001322638275048377 I,-7.006263426576878*10^-6-0.0001350800095989725 I,-0.00002012938488700608-0.0001236147615965757 I,-0.00003798356171521-0.00008417391014700092 I,0.00004557954625663109 +0.000103285026603415 I,0.0001092256811827242 -0.000197070975102123 I,-8.356809741841503*10^-6+0.0000669016357832128 I,-0.0002266729809791113+0.00006509710876448791 I,9.245342300439685*10^-6-2.500364003061749*10^-6 I,0.0001549051228218502 -0.00008001074189359349 I,-0.00001268597868941073-0.0001188821528296745 I,-0.00006178529591502818+0.00008593393849122354 I,-0.00001830424635259931+0.00005991763352345593 I,0.00008834171438743227 +0.0001508314492002057 I,-0.00001080769988334485+0.00009583913757635633 I,0.0001637554168950568 +0.000035884777103948 I,-0.00002260741392637393+0.00003597424152460845 I,-0.0001388903435636857+0.0001658405836730558 I,0.0001537362329483494 -0.0002045686505707393 I,0.0001602871843943064 +0.00002116297260591995 I,-0.00002458497625642967+0.0001988309506062166 I,-0.0001801019543812943-0.0001471353889309944 I,-0.00002024104588527361+0.00007455079840580926 I,0.0001319465648868379 +0.0001403650197473591 I,0.0001239468614594818 +0.0001184628348406073 I,0.0001578709729538115 -0.0001553933510092274 I,0.00008994543596686043 -8.883149855339973*10^-6 I,-0.00001541912075513741-0.0002517459990509235 I,0.00008368242467169466 +0.00001525096811818606 I,0.0001921836144883068 +0.00006401079772870967 I,-0.000193664656550072-0.00009520403255026025 I,6.360833811017586*10^-7-0.00009433875724092914 I,0.0001380010616089844 +0.00003403037537867663 I,-0.0001444427346055086+0.00007778439460876866 I,0.00002436305100597344 +0.0000511287556984165 I,0.0001610337719264702 -0.00004541809522007648 I,0.0001395436720718438 -0.00005368185911889305 I,-0.0000557783910965921-0.0002556263065525753 I,0.00004267809101037463 -0.0001425828392886621 I,-0.0002990196322092985-0.0001015992322647569 I,0.00004112468884265855 -0.00004137852386279381 I,0.0001514493318379034 -0.0003517211457506428 I,-0.00004053147495877577+0.000279990992543812 I,-0.0002798490528746533+0.0001507868605066945 I,-1.29913589366011*10^-6-0.00008738844204719457 I,0.0001948636703695124 -0.0001550094114620528 I,-0.0001217411827573299-0.0001984697050255623 I,0.00005448756113482267 +0.0001398986309956755 I,0.000187407555106929 -4.388418906228471*10^-7 I,0.0003616015832424563 +0.00002140504250931779 I,0.00027820829571048 +0.0001730967540639105 I,-0.00005871690230656177-0.0002946521219936323 I,-0.0000752074165279162-0.000275989619606912 I,0.0003318668243725867 +0.0002711393181020473 I,-0.0002328235955581471+0.0003087486911527862 I,0.0002442467981895853 -0.0003968373815775851 I,-0.0003321056399137753-0.0001760660442542069 I,0.00003165274227615803 +0.000309092154813841 I,0.0001739361838836923 +0.000371995531451887 I,-0.0002708331131884386-0.0001068202817747648 I,-0.0003320305433930392+0.0001712630216484555 I,-0.00027177168641259-0.0001166710991030472 I,-0.0001777072292858922-0.0002449255692652228 I,0.00002082646194773662 -0.00007177400407465237 I,0.0001766992439557039 +0.0002644004277414823 I,0.0002458282079810986 -0.0003358066849697253 I,-0.0002951591672764487+0.00003214045233018367 I,-0.000510342303175085+0.00008080785471675225 I,0.0002680369315646429 -0.0001903979587339864 I,0.0003044486194747092 +0.00007080051032817201 I,-0.00004810335899796848-0.000306632742711631 I,-0.00009914073724458247-0.00004944522060824352 I,-0.00004526350349354935-0.0001015551947664397 I,0.000104470140313242 +0.0002819738294876748 I,0.00005832692822693952 +0.0001691690843008037 I,0.0002337161468903855 +0.00008218828939747169 I,-0.0001253269288773195-1.001892725173501*10^-6 I,0.00001051087324976756 +0.000394321233194757 I,-1.675849245947789*10^-6-0.0002655600580697843 I,0.0002011482120460795 -0.0003484383920386068 I,0.00006119096941950363 +0.0002035797689839261 I,-0.0001867847484349636+0.0002334615380692188 I,0.0002505565780344248 +0.0001083088415115715 I,0.00007810605450019514 -0.00006942325785241393 I,0.0001149960615338721 +0.00002380410174814042 I,-0.0002455477039212814-0.0001041194912142912 I,-0.0001015874194518886-0.00006091393925862241 I,-0.000269641367066659-0.00001827329349590289 I,0.00009303010050721894 -0.0001493388732333942 I,0.00003474872667447377 -0.0002475315770780535 I,-0.0001240241781157512+0.0003784446836372545 I,-0.0000434089803453477+0.00003084232254823785 I,-0.0001816934014299192-0.0002586139412848283 I,0.0001858768577735273 +0.0001564005018223176 I,0.00008834944256335905 +0.000115101599766985 I,-0.0002617731435200964-0.0001145855835448593 I,-0.0002701824330397387+0.00006405943038972805 I,-0.0003033249401401927+0.000128280424048861 I,-0.0002647608077469259-0.00007719238481047834 I,0.00005621626385703696 +0.0001977367760861974 I,0.0001331551929228177 +0.0000975126387530655 I,-0.0002430439945224617-0.0002533423519210409 I,0.00003272335836650945 +0.00009267533641083238 I,-0.00009278287373222773+0.0003659904387912519 I,-0.00003528614497495347-0.0002439363372940872 I,0.0001385890162305787 -0.0001454756543662043 I,-0.0002364997007321124-0.00002613897079058362 I,-0.0001587538013838517-0.00005050979819050539 I,-0.000277950573750755-0.00007790283699541398 I,0.00009150482821444663 +0.0001216207450716901 I,0.00003319102552160833 -0.00001402303313666897 I,0.000324557456508915 +0.000163335337233901 I,-0.0001809108118426767+0.0002207752778221206 I,-0.00004092632884710042+0.0002940392633729017 I,0.0002224819739024822 -0.0004124955735611677 I,0.0001187887233795859 -0.0002605039728417257 I,1.433201668215776*10^-6+0.0003844506203118823 I,-0.0002026341010100221+0.00003541673845396418 I,0.0002320222151195656 +0.0001624373160461134 I,0.00004619751889098193 +0.00009276813612006662 I,0.00001977500580345367 +0.0002326386911651942 I,-0.000173262134221347-0.0001385735061489797 I,-0.0001743764250604633-0.0001408818584284676 I,-0.00004598556685723695-0.0001564923836817859 I,0.0002691844645148847 +0.0001825103819755195 I,0.000144422775012763 -0.0002994971014652466 I,-0.0003345405583118688-0.00003754951315019546 I,-0.0003942267064363193+0.0001467172397269546 I,0.0003379646941683491 -0.0001876900025038611 I,0.0001626815691165675 +0.0001169827326833697 I,-0.0000115956233091301-0.0002206319692768747 I,-0.00005004822767760791-0.0001549170998724479 I,-0.00008691703923876474-0.0001499670002473766 I,-0.0001991521803974241+0.0002002530520887108 I,-0.0001776757548414213+4.27474760710805*10^-6 I,0.0001151042907058511 +0.0003514087148563417 I,-0.00003807646129925298+0.000146896001571989 I,-0.0003885344388534334-0.00005531667717554759 I,0.0003423582782647978 -0.0003066377754386347 I,-9.430175867601511*10^-6+0.0002446710082826988 I,0.000138507906328297 +0.0002987430214234142 I,-0.0001550257700846311-0.0002750577337819719 I,-0.00002576527075881345-0.0001205266259448247 I,0.0002442460397767584 +0.0001821232974734843 I,0.000294523422369905 +0.000084885978770292 I,0.0001203077558426804 -0.00009747780838830904 I,0.00001589124436437924 +0.000045861212596246 I,-0.0001603974257100576-0.0002018882087567938 I,0.0002064462717771125 -0.00009425108454619378 I,0.00001631998204519911 -0.0002114834808627633 I,-0.0002261235652730763+0.0001810703874192052 I,-0.0001395032621283642+0.0003209739952475338 I,0.0001264961942097932 -0.0003095607179238781 I,0.0001050533543030406 -0.0001051939897555161 I,-0.00028282270760266-0.0001460793352875872 I,0.00001728129245941797 -0.00006901838291092452 I,-0.00004598368508715013-0.0001873760045716977 I,0.0001840233331267306 +0.00001950670940576395 I,0.000145314712374595 +0.00001310860346375538 I,0.0001003856195135157 +0.00004210335964458941 I,-0.0001324333303430594+0.00007821460570148043 I,0.00007047442550383459 +0.0001247299865530032 I,0.00005230270164218587 -0.0001827485617716303 I,-0.00008608879124820241-0.0000994572936959204 I,0.0001631791903115665 +0.0001515016324011711 I,-0.00001686215213286341-0.00002298922633805077 I,9.575090013994036*10^-6-0.0001783109576640235 I,0.0001359323929087828 +0.00004336048525019619 I,0.0001995719437629069 -0.00007623477186617357 I,0.0001301597387391433 +0.00004249987761214096 I,0.0001180755261329038 +0.0001189425081268332 I,-0.0000502436204945146-0.00002571262785246858 I,-0.00009376874302645589-0.0001428352386791141 I,0.00001673662278206019 +0.0001583066970331948 I,0.00005615860303419403 +0.00002711936882876105 I,0.0001450627188291233 -0.0001614373151794702 I,-0.00009322215333723268+0.0001027420484924561 I,-0.00005797756921781353+0.00005562784255719205 I,0.0001270889611671701 +0.00006966254697666242 I,0.00001230732605432588 +0.00004609548644088666 I,0.00006622822982012128 +0.0001062955804321762 I,-0.00005897206619399288-0.00001900729760065495 I,-0.00006475787661098983+0.00001178610551100847 I,-0.00006068734150417414-0.00009136312182463107 I,0.0001377684842083562 -0.00003341626689111837 I,0.00001481704507588002 -0.00009088927661812468 I,-0.0001666927542067215+0.00005301181237710924 I,-0.00007307944721868516+0.0001003407331931958 I,0.000125308261038291 -0.0001283378098566805 I,0.00003174892518033665 +0.00003794131889343319 I,-0.00004323710090456363-0.000102126808462701 I,0.00004613963327584783 -0.00006913042454527382 I,0.00007596266814668162 -0.0001133657146579143 I,-0.00002399378956979413+0.0000410846277207041 I,5.643563402163681*10^-6+0.00005213552719533202 I,-0.000103259011579511+0.00008436077244159196 I,-0.00003065688590914926-0.0001377990699075173 I,-0.000132889336785832-0.0000556207615996344 I,0.0001128512098485922 +0.0001839969164288288 I,0.00008473829623933704 +0.0001207159878428656 I,-0.0001712769328337344-0.00006588757842240022 I,0.00006402061052510988 -0.0001104182732473221 I,-0.00005848543724046263+0.00001536258810047776 I,-0.000072359986104487+0.00008769598713398593 I,-0.00003117902395069716+0.0001022342929610619 I,0.0001241704626768913 +0.00009807306873463277 I,8.700332320404913*10^-6+0.0001124385295065579 I,0.000152280296090133 -0.0001114044346811547 I,0.00009981315025032257 -0.00002504563138924651 I,-0.00001007235243349945+0.0001974474210903737 I,-0.0001950685137171222-0.0001160247803653053 I,0.0001552119964095312 +6.926630136588858*10^-6 I,0.0001183114537571583 -0.0001231683696178276 I,-0.0001376227649490831+0.00009768928855297999 I,-0.00007644403665055027+0.00004817470497432279 I,0.00007142617479095702 -0.000150501188016959 I,6.952399944769154*10^-6-0.0001668119079838891 I,-0.00006727646277499087-0.00007714825000438944 I,0.00002406959220493867 -0.00002999854218670414 I,-0.0001087407556652347+0.0001071053730438209 I,-0.00007320816824565797-0.0001016189483326482 I,-0.000119851657873749-0.00001523383532512005 I,0.0001407879904939523 +0.0001290224673762737 I,0.000185960397973012 +0.0000653014320014368 I,-0.00021185461312213-0.00004053781546158912 I,-0.00005923952785490691-0.00007389113646322558 I,-0.00006296234124248009+0.0001834328206560654 I,-0.00005371000977111184+0.00001845900948180241 I,-0.0001156017278803984+0.00007263698761466166 I,0.00001107467247858039 -0.0001006330761775497 I,-0.00001175477467076763-0.00007802683188189529 I,0.00004836911684922115 -0.0000793780064824434 I,0.00007591322898006808 +0.0000839504942727517 I,0.00005090597117252539 -0.00002243262686928141 I,-0.0001141962279494827-0.00005884291120244599 I,-6.030170979069629*10^-6+0.00007217814572671886 I,0.00007349891035098949 -0.00009266615063671134 I,-0.00007186909857403908-6.477612077838188*10^-6 I,-0.0001209930535483465+0.00006411130437927546 I,0.00004001724187054415 -0.00007466848548567663 I,-0.00006418414768838979-0.00008447059205233927 I,0.00002016029767474062 -0.0001577291353113791 I,0.0000628497621608584 -0.000125978532496117 I,-0.00002466653375177049-0.00003568519797266312 I,-0.00002467611735546249+0.0001089681984588498 I,0.0001558501609176486 -0.00006661413226237984 I,-0.0000787734440714487-0.00005958369819229585 I,-0.0001601882729173365-0.00004434637637763441 I,0.0001463533954037303 +4.271258579294506*10^-6 I,0.00001195077666696657 +0.00004132921145892619 I,-0.00001697371540821036-0.00007862892556010094 I,0.0000683244980887241 -0.0000558538268173989 I,0.00005517260318568375 -0.0001110916702549747 I,-0.00001211552289796672-0.00005570413687940728 I,0.00004674395944331643 -0.00003966157546237122 I,-0.0001094203737099912+0.00004898210807732754 I,-0.0000733834310987718-0.00003189018560068263 I,0.0000117266614781034 -0.00009019376110265491 I,0.000100730219792065 +0.0001015661447626869 I,-0.00004708517698724687+0.00001922949278869101 I,-0.00008068718937737947+0.00003417720897603302 I,0.00005299023646116336 -0.00008789537526476033 I,-0.00002248653813165686-0.00002102321594574727 I,5.949935983249124*10^-6+0.00009582196960199679 I,0.00005382384281921724 +0.0000685714389088301 I,0.0001490645984698966 -0.00002037216700325134 I,0.00009669095693107778 +0.0000465574664493216 I,-1.05506096228662*10^-6-0.0001893807869559176 I,0.00003389575749709854 -0.0000683001305064996 I,0.0002084010020773859 +0.0000887681819222889 I,-0.0002272000687123485+0.0001094855285135678 I,0.00003050935679817349 -0.0001623808710305424 I,-0.00005371666369278642-0.0001586708017276998 I,0.00004388299067658512 +0.0001999174247779933 I,0.00006157294429882994 +0.000136222118773499 I,-0.0001546644161160691-0.0001239325694222883 I,-0.000224348281982986+0.00002112426062077103 I,-0.000197343425437548-1.862684310554663*10^-6 I,-0.0001433717795882439-0.0001315354582897926 I,0.00008809960231183307 +0.00008349775477978081 I,0.00007794156863267653 +0.0002200571663187619 I,1.933818695671827*10^-6-0.0001447667645655026 I,-0.00006799683503000013-0.0001786778538343275 I,-0.0002613446202532331+0.0001079595178157681 I,0.0002789899377559108 +0.00002277768896929571 I,0.00004819970112982945 -0.00002827471219631557 I,-0.0000356746990305456-0.0002306108038588685 I,0.0001020217482943924 -3.265403487741662*10^-6 I,0.000148451353360958 -0.00009040590119814063 I,0.00006851649230370937 +0.00009525397735630136 I,0.000036806369924139 +0.0001239653960184917 I,-0.00001383384742052485+0.00001115227659099469 I,0.0000110411309969134 -0.0001286651988125264 I,-0.00009625476438530377+0.0001094386559719213 I,-0.00003792302829190542+0.00002457432485339711 I,0.000191697679615222 +0.00005582774065007393 I,-0.0000299691189207757-0.0001063425027236919 I,-0.0001165904507757623+8.34021820951276*10^-6 I,-0.0001052821115914093+0.0001225418256792496 I,0.00003685997464317653 -0.00008157202039181966 I,-0.00005101972379856012-0.00007148280293726278 I,1.960738464575095*10^-6-0.0001695035181563501 I,0.0000547374419980504 -0.0001138640878187798 I,-0.00006428281215554071-0.00007387084383261133 I,0.00002691765389130715 +0.00005394138711037956 I,0.0001785776571736416 -0.0000885921606449633 I,-0.0002031166135961783+0.00003528871238456257 I,-0.0001518675277268918-0.00001032390554009478 I,0.0001569988904437625 -0.0001701026873246449 I,0.00004988921675543636 +0.0001156556305557708 I,-0.00008478623376766462-0.00005820645357741038 I,7.383604868841395*10^-6-0.0001745698476791263 I,-0.00006635762214375043-0.0002183845781320526 I,-0.00006276998249479487-0.00005481133955137511 I,0.00005258606860674374 -0.000098251934427372 I,-0.00006990089026551676+0.0001369203673139358 I,-0.0001353050665749632+0.0000507678931132292 I,0.00007331804156620475 -0.00008499690554729954 I,0.00005671324531602456 +0.00002754318779554029 I,-0.0001697864499164395-0.0000763066750332231 I,0.00006990438305727454 +0.00002240417666176997 I,0.0001137475110817989 +0.00005489329482152039 I,0.00004070460676430785 -0.0001429738353814966 I,-0.00001031274250313711-0.00005033717131377163 I,0.00003287912995231855 -0.0001112510399322014 I,-0.00005053814258345197+0.0001204052191192334 I,-8.84149327707223*10^-6+0.00006585881566512982 I,-0.00003290045275418192+0.0001773883922096464 I,-0.0001042916166589304-0.0001064597102675581 I,-0.000188479381155344+0.0000374153870567894 I,0.0002007338089721289 +0.00008719427282952886 I,0.0001825062203163497 +0.00005751186221784305 I,-0.0001899956179666651-0.00001268456443231832 I,-0.00004210914485823579-0.00008198595666316645 I,-0.00008294051413022688+0.0001335307246903352 I,-0.00004477323615015869-3.682020474196174*10^-6 I,-0.0001243878834437069-0.00001144087777329604 I,0.00003963153669291026 -0.00009831112218536695 I,0.00006215350660377847 -0.0001023081588290857 I,0.00004867710182624498 -2.304547884417885*10^-6 I,-0.00009612717785713727+0.000122056443746827 I,0.0001476686450165142 +0.00003057663328934218 I,0.00002487199657426824 -0.0001055859170896032 I,-0.00009421812179070225-0.0001883031366238402 I,0.00004913840157461178 +0.0001353484038297056 I,-3.523381181288587*10^-6+0.00009124442097261265 I,0.0001320079899498168 +4.589440520483664*10^-6 I,6.211973523672374*10^-6-0.00001960989208828822 I,6.729521767203468*10^-6+0.00001188326386509023 I,-0.0001459324948258876-0.0000363230690441652 I,-0.00008359216702892141-0.00008616324399853814 I,-0.00009420454169689797+0.00007141590234242121 I,0.00001661366786917735 +0.0000589061539223647 I,-0.0000275037995888734-0.0002283535174138723 I,0.00007792242045545573 +0.00006585525558136872 I,-0.0002053953474499329+0.0001242572825628175 I,0.00002624584795295583 +0.00002420901814229273 I,0.0001312539582538505 -0.0001163426107650008 I,-0.00002682566017113121-0.0001452612295189259 I,-4.191260251702582*10^-6+0.0001007429538205303 I,0.00005732343107645325 +0.00005192368468132615 I,0.0001313209299021368 +0.00009237307058473706 I,0.00003991271475124771 +0.00008895101779836228 I,0.00007792382403961127 -0.00009541718657811126 I,0.00004796565679694143 -2.735868486221084*10^-6 I,0.00007796907435826851 +0.00009282970009142074 I,-0.0001637883197733792-0.00006675266242297623 I,-0.00003351399782752473-0.00004282238960828173 I,0.0001507454453463022 -0.00006677942703018112 I,-0.0000129919226884363+0.00009320553198296157 I,-0.00002360960825340978-0.00006604140741336327 I,0.00005381279892304684 -0.00009850987609801765 I,0.00003819549430552453 -0.0001369501007495603 I,-0.00002803952913720239-0.00003363591946715815 I,0.000046969216392225 -0.00003031535949674479 I,-0.00009916927184970336+0.00004522058355258096 I,-0.00006359972681818172-0.00002796698457013442 I,0.00008466165574031473 -0.0000784613435476084 I,-0.0000318410604659717+0.0001064035088913734 I,-0.0001475134500816791-0.00005636024786885075 I,0.00004998836844914343 -0.00006063492486509605 I,0.0001494138808634841 +0.00008127595918569269 I,0.00003241648523743726 -0.0001451239997868998 I,-0.00004985455429073083-0.00006301460719018912 I,2.756850585363562*10^-6-0.0001090188037090575 I,-0.00004511218932844283+0.0001561758729355465 I,-6.018966946559384*10^-6+0.00007544587034665215 I,-3.803517136056556*10^-6+0.0001763695464768247 I,-0.0001218911010952072-0.00008288624970790784 I,-0.0001258209972458441+0.00004565626217193757 I,0.0001805663766919259 +0.00007593149573797609 I,0.00006519317011732047 -0.00004230299684734171 I,-0.0001392077030380479+0.000052893034794294 I,0.00007999851266719089 -0.00001294872521023113 I,0.00007634713130635321 -0.00002338391761375332 I,-0.00009578827460767902+0.00005676004312278282 I,-0.00002585976577457758+0.00008293879380877507 I,-8.081026788782299*10^-6+0.0001863746804288685 I,-0.00006930687257174868+0.0001285542093214481 I,0.0001282080722742335 +0.00004174805519910747 I,0.00004509317711855716 -0.00005439954725316594 I,-0.0001442776887427865+0.0001658214217695345 I,-0.00004411976258581135-0.00006580123621566575 I,0.000194540049771464 +0.00001444128015143294 I,4.77831562212971*10^-6-0.00009678347765350735 I,-0.00008891071742243552+0.00009639627577674208 I,-0.00002416658430221745+0.00009329480476291793 I,-0.00004302173512959822-0.0001636051137097975 I,-0.0001457628742803089-0.000132191749898482 I,-0.0002054051741166012+9.977423406644207*10^-6 I,-0.0001152517340369559-0.00009184382078861568 I,0.00001888596704468343 +0.0003077398473206129 I,-0.00009911556547751398+0.0001075299010545183 I,-0.0002990054592021222-0.0001046825473676852 I,0.0003704061366232838 -0.0001458667890284115 I,-0.00001345333927837408+0.0002074336766651985 I,0.00001635122145236269 +0.0002292855111815131 I,-0.00005507937089158235-0.0002909894422430059 I,-0.00008654771900719354-0.0001518370294972469 I,0.0001738142680137917 +0.0001918168834105988 I,0.0002618698600545079 +0.00002281874775361041 I,0.0002924901009783903 +0.00002268969459805014 I,0.0001801820117152113 +0.0001941336475585266 I,-0.00003360154840923197-0.0001737830234976144 I,-0.00002903185093667236-0.0002534808574066618 I,0.00003668119009653958 +0.00026508465630603 I,-0.00008860749348575192+0.0001320946546978087 I,0.0003423702828246762 -0.0001368002789482359 I,-0.0001861151517123074-0.000134026580796891 I,-0.0001157514779852591+0.0001355503757296657 I,-0.00001356161311503545+0.000241600065668778 I,-0.00006765897895305025-0.0001385751797288213 I,-0.000156152953404917-0.00007465753974453818 I,-0.0001308239611791983-0.000107553604676113 I,-0.000012846745147107-0.0001154582841146684 I,-0.00004522459904532681+0.0001280164096851721 I,-0.0001162478518338792+0.00001315560292561288 I,-0.00005596141635426285-0.00002146557287257913 I,0.0002040704139809264 +0.00004639275594756283 I,0.0001088532115972264 -0.00006132423122267108 I,-0.0002095613611957202+0.0001080188312997755 I,-0.0000392766416070545-0.00003358754211158903 I,0.00008148049618547263 +0.0001671149549223056 I,-0.0000646674102187371+0.00006651990579334516 I,-0.00006827061321520024+0.0001533407646331721 I,-0.00006767508570771635-0.00005222235162575644 I,-0.00008430566319402174-0.00005104867860134508 I,3.859784776043837*10^-6-0.00006865309507386329 I,0.0001201485852984902 +0.00002507253141477669 I,-0.0000115409951595703-0.00007914659684099597 I,-0.0001044107100965257+0.00005725991301930583 I,0.00003629565246999285 +0.0001158372003673025 I,-0.00003869573443129691-0.000168851399349353 I,-0.00002522004712578782+0.00001060071929853869 I,-0.0001055316005347189+0.0001417403617580634 I,-0.00007066490242822943-0.0001003887121202132 I,-0.0001999102194537027-0.00004371382955678158 I,-0.00008445196236244137-0.0001206678371209724 I,-0.00003718331417285401-0.0001769949157776436 I,0.00002431928368240474 +0.00005066205112497048 I,8.506015783944873*10^-6+0.0001894968710116646 I,0.00007129390307638965 -0.0001752070901842371 I,-0.00003763410326667119-0.00004888003370931405 I,-0.0002507432020952879+0.00009610555539177294 I,0.0001529957117460153 -0.000071725279434248 I,0.0001134851420228597 -0.0000328704985773866 I,-0.0001164419800392817-0.0001589371268226512 I,-0.00001165937763601026-0.00006533566822490839 I,-0.0000532667853293539-0.0001697093113273629 I,0.00006453741733419014 +0.00009321609860443942 I,0.00008056271944124953 +0.00003435461580840544 I,0.00009085832919447347 +0.0001855781916668372 I,-0.0002221076584341129-3.531643811910878*10^-7 I,-0.0001014969101481452+0.0001896970232422461 I,0.0002903304036194629 -0.0000837856799192988 I,0.0002096707680138945 -0.0001529946806612145 I,-0.0002303342945111698+0.000191626475075234 I,-0.0000955561366090531-0.00003661540334682007 I,0.00007873736420768588 +0.0002258601762974022 I,-0.00009273297964895773+0.00006748332820138438 I,-0.0001824680773765802+0.0001758943201681059 I,-0.00006264346496779531-0.00008097131191757557 I,-0.00009174427999910007-0.0001413403197877278 I,0.0001779642238437279 -0.00005468691062007801 I,0.00009078774506493273 +0.0003090596730994434 I,0.0001173115366403568 -0.00001735604304412577 I,-0.0001467544321378208-0.0003794722377040402 I,-0.0002411080740534887-0.00003482331722411312 I,0.0004119688879627451 +0.0001081497390941711 I,-0.0000886895471235246+0.00008504125154741717 I,4.909054677707772*10^-6-0.0001718405794999737 I,0.0001846933188042277 -0.0001365345128740908 I,0.0001200395690082424 -0.0002538208170511467 I,-0.0001153767415334671-0.00007056606992258628 I,0.00002734503644465561 -0.00008218931490914505 I,-0.0001729359800042083+0.0002627596326610085 I,-0.0002307250542779947-2.286894078399476*10^-6 I,-0.0001417849071081669-0.0001222194077327555 I,0.0004479744929815376 +0.00001149631844344178 I,-0.00002076795698308055-0.00003560275785084454 I,-0.0001791012958856099+0.000352230591434024 I,-3.930227582882144*10^-6-0.0001841162868860847 I,0.0001541251563475609 +0.00001586627701742854 I,0.0000371830781811388 +0.0002410897995090882 I,0.0001217691355387368 +0.0002587010453517447 I,0.00002194469639210287 -0.00003384711058220616 I,-0.00008216430593737241-7.998998857468878*10^-6 I,-0.00003931299667023455-0.0001756625497371408 I,0.000206064296737218 +0.00006000145819923222 I,0.00004772640791733527 -0.000218118236532962 I,-0.0001695318552528595+0.00002481733189386859 I,-0.0002598906359308219+0.000234168524100872 I,0.0002031825506675715 -0.0001271314043983072 I,0.0001211874285762793 -0.0001067767383434274 I,-0.0001640131778973605-0.0002352288862987292 I,0.00005348087819420763 +0.00001362503512044211 I,0.00006052599480325057 -0.0001165038836953961 I,0.0001956898466454083 +0.00009300321046554956 I,0.0001281262667804993 +0.00009213745962947347 I,0.0001104394800694663 +0.00001821289754175563 I,-0.0001191776703407585+0.00002833387029081915 I,0.00003514545118347019 +0.0002035903496619216 I,0.00006608956635241256 -0.0002218817096874104 I,6.037335873585099*10^-7-0.0001545829333566259 I,0.0001517958338089983 +0.0002195303617897969 I,-0.00008590281324177246+5.034469969398934*10^-6 I,0.00007178123447200202 -0.0001387227712668724 I,0.0001758460849496889 +0.00005309571939794949 I,0.0002550268499357827 -0.00005225264051808852 I,0.0000926528925988757 -1.363542037205044*10^-6 I,0.0001079033467769829 +0.0001083580816295572 I,-0.0001199890576965168-0.00005404164438449541 I,-0.00007492073904987677-0.0001776063303735131 I,0.00004238816668716264 +0.0001204153620543261 I,0.00002241469054722805 +0.00007569671685112287 I,0.0001299400188987917 -0.0001865901134518904 I,-0.00008129872202229633+0.00008795895650078457 I,-0.00004411512789720146+0.0001106507767005402 I,0.0001730956362698362 +0.00005274365171880714 I,0.0000144123003319583 +8.448380310906476*10^-6 I,0.00009379847120724542 +0.00007059343125951183 I,-0.0001414583430530953+0.00001231793806941828 I,-0.0001100496123214235+0.00003116872698338545 I,-0.0001290488709131068-3.169281774495931*10^-6 I,0.0001344785291015162 -0.0001177596440384596 I,-0.0001292354098763727-0.0001376607087891104 I,-0.0000879791754856041+0.0001582711163388983 I,0.00003870134665518235 +0.0002264077688887442 I,0.00002674115684792871 -0.0002304784109611897 I,7.504532058636923*10^-6-0.00001837458484244491 I,-0.0001723181375306966-0.00002387159644745649 I,0.00002023269777911261 -0.0001444064831082187 I,-0.00005010521108060151-0.0002386914768234476 I,-0.00002616781970008074-0.00003312024981618524 I,0.00006490743375372563 -0.00005375927970440134 I,-0.00008014486350244206+0.0002498539224107595 I,-0.0002416828215130549-0.00002589819896098791 I,-0.0001651669755304071-0.00003707678558585377 I,0.0004381646109700629 +0.00002901748451030896 I,0.00005604990813338709 +9.869089380584473*10^-6 I,-0.0001918386323932269+0.0002867403427454724 I,-9.360512574540868*10^-6-0.0002393030342217764 I,0.00004052032800748318 -0.00005113340377675262 I,0.000097660752639484 +0.0002552480272106129 I,0.000246783918823205 +0.0001571563296738704 I,0.0002477262770598525 +7.942850629144934*10^-6 I,0.0001355241475137181 +0.0001766710538495542 I,-0.00008561763269988613-0.000255542531499971 I,0.00004454276167581016 -0.0002817078678471391 I,0.00009024454946260135 +0.0001578224850699069 I,-0.0001581036479165184+0.0002833410808539862 I,0.0002869394825021645 -0.00009058158038769174 I,-0.0002637628058474949-0.0002412435956948464 I,-0.00001973269904342713+0.0001284632575078722 I,-0.00004250862529027086+0.0002627850791985201 I,-0.0001586639276741477-0.000128195075002194 I,-0.0002527925213532702-0.00003761368785656883 I,-0.0001017271006517559-0.00008499624189437776 I,-4.291788196799891*10^-6-0.0001393307020994348 I,0.0000328963719434252 +0.0001073170226078004 I,-0.000136591422627389+0.0001000176654117513 I,0.00003846129002984611 -2.966827607766957*10^-6 I,0.0001866836399704586 -0.00002799750668036952 I,-7.118791218106138*10^-6-0.0001618403245612162 I,-0.0001423232696422931+0.0001946179037837493 I,0.00001843170975014967 +6.651892556376778*10^-6 I,0.0001757980967398984 +0.00007158668187368759 I,-0.0000676753892513856+0.0001030150365725786 I,-6.419687487807264*10^-6+0.000182915951850596 I,-0.00008012494180982652+0.0000575529173136303 I,-0.0001146378325856609+0.00003094529203130445 I,0.00002755183620686065 -0.00002141963617657586 I,0.00009623698137183235 -0.00002526848971963404 I,-0.000119215554776449-6.264712843173209*10^-6 I,0.00002404656887210747 +0.00002220198244363286 I,0.0001469474036445578 +0.00009084704533709524 I,-0.0001252273433901273-0.00002578416299885375 I,-0.00008583497788927026-0.0000473293090128503 I,-0.00002002175744693889+0.0001811415139276954 I,-0.00003733184981583778+0.00001982744441110976 I,-0.0000995416096225352+0.0001052953165582236 I,-0.00004824331938463395-0.0001183086027825599 I,-0.00006167537194831603-0.0001101084893979156 I,0.00001249885857807425 -0.00008451207906695504 I,0.0001201155187361021 +0.000115406738377015 I,0.00006230418831225458 -0.000129856347259533 I,-0.0001351350153664696-0.00003284148797789627 I,-0.0001252411529259498+0.0001143898172109516 I,0.0001253655051385316 -0.0001042024300819329 I,5.479077538483141*10^-6-6.925584552815053*10^-6 I,-0.0001037005595886381-0.00003206169342769914 I,0.000028365021126052 -0.00007831264918252105 I,-0.00003994133456762896-0.0001222105430450962 I,-5.236780126998444*10^-6-0.00006077463498744096 I,0.00004108376639807657 -0.00007078730377876603 I,-0.00003114826064524559+0.00009316034893301698 I,-0.0001080954142223056+0.00004034987342928833 I,-0.00002087479648963891-0.00003876798330074758 I,0.0001738746146850525 -0.00001048846714241496 I,-4.111844571517758*10^-7-9.967356217809259*10^-6 I,-0.00008554234082011727+0.0001215016777584998 I,-0.00002823335264771573-0.00009727256792591177 I,-3.378789308208481*10^-6+0.00005442550116028105 I,0.00002821089812225005 +0.000083751214079174 I,0.0000141368260370329 +0.00006711149058818572 I,0.00007363885031309435 -0.0001203144273462331 I,0.00006623018632719431 -0.00007429295656984604 I,-0.00003814765411057155-0.0001282312996257366 I,-0.00001125751276415961+0.0000605220692779839 I,0.0002243933921135939 -0.00005567546034795663 I,-0.00008734738802828359+0.00002448216417127978 I,-0.0001686941138509525-0.0001594657833382748 I,0.0000385021711184027 +0.00004767510926066881 I,0.00008889750033270826 +0.0001117076965642121 I,0.0001474207882933386 -0.00003897358175795798 I,-0.00004685136211161821-5.622148903226332*10^-6 I,2.522632943396616*10^-6+0.00004779131041928731 I,-0.0001483033982634097+0.00006053996460986437 I,-0.0001236368033846346-0.00001491467751255643 I,-0.00005475691346674931+0.0000848987561593049 I,0.00005561889360650491 +1.385919938248125*10^-6 I,-0.000140096942463978-0.0001773448617414439 I,0.0001012445729551438 +0.00008685396893282787 I,-0.00005564290454924107+0.0002012212324966485 I,-0.00006331698536211325-0.00002271552776475009 I,0.00005601620418246552 -0.0001602940469101879 I,-0.00008973583283129841-0.00002229180064911449 I,-9.546273058232336*10^-6+0.00009037327559564695 I,-7.210230693241047*10^-6+0.00005955000284559732 I,0.0001292559143329579 -0.00002475176225453348 I,0.000059546687265127 -9.020616414917934*10^-6 I,0.00003830901691743831 -0.0001217523624277946 I,0.00002576449439040346 +0.0000555708829734428 I,0.0001687095889325646 -0.00001567847537764045 I,-0.0001338040000702398-2.476797722508718*10^-6 I,-0.000164623107899269-0.00006717579815809548 I,0.0001032103760900239 -0.00003778145317888759 I,0.0001043784797650315 +0.00007528373716851871 I,0.00003576080769838362 -0.0001151906482876172 I,-0.00002879343095815503-0.00003578328497623394 I,5.807515528027338*10^-6-0.00005372511661603388 I,-0.00004185723394708695+0.0001036319398706617 I,-0.00002609017565122207+0.00003858497472311076 I,-5.366061569457543*10^-6+0.0001165266480505364 I,-0.0000605639883668953-0.0000327779059844528 I,-0.0001056679780119066-0.0000262446430885812 I,0.0001562911921895292 +0.00006030291625144424 I,4.555027937377183*10^-6+0.00003799283410510694 I,-0.000112857750299295+0.00005112952611830983 I,0.00007668670148456918 -0.00008708866396995243 I,0.00001948623179925146 -0.00003106695020164388 I,-0.00006537135617825347+0.00009674397402510023 I,-0.00001492502996978078+0.00009663955295919279 I,0.00005790171221769857 +0.0001216670920089617 I,-0.00002674137804984148+0.0000857819428998072 I,0.0001089602803287395 -1.023861664943803*10^-6 I,0.00001834151545328556 -9.109784848474958*10^-6 I,-0.00009278316017454941+0.00008994428709304411 I,0.00005906394491895779 -0.0000446558134730263 I,0.00009723041083958337 +0.00004112526856085605 I,-0.00007179368503365261+0.00002881980273129432 I,-0.0000277168080133165-0.00007776602758700556 I,-0.0000312716960094702+0.00004270602990454834 I,-0.00002437120070623251+0.00007318497844841567 I,-0.00003151060930680391+0.00008674495986522935 I,0.00009843632127718666 +0.00001402291836858445 I,0.00002919836910910534 +0.00002878102444838175 I,0.0001169759405025948 -0.0001000791875701966 I,0.0000252360804248683 +0.00005311633246855659 I,0.00009698011817817055 +0.000127174918737006 I,-0.0001040193008547109-0.0001153585080274125 I,-5.532790618964002*10^-7-0.0001289581945043561 I,0.00006396079205540461 +0.00003416112862779112 I,-0.00001427486150500632+0.000110156924233944 I,0.0001033419754823946 +7.22074748228267*10^-6 I,-0.00003995313294780578-0.00002879378425366957 I,-0.00003782948305512056+0.0000466677618337616 I,-0.0001315133559067469+0.00009002005433617497 I,-0.0001431049115313367+0.00001765452581514092 I,0.00007750817820109112 +0.00009135761371208687 I,0.00009581503621666789 +0.00006677036755632061 I,-0.0002019670303702311-0.00003066444670111188 I,0.00003387695291837714 -0.0001563896804780608 I,5.138901757290357*10^-6+0.0002227140474548547 I,0.0001549391024113638 +5.040032527953549*10^-6 I,-0.00008278262363849202-0.0001393173345897559 I,-0.0001565012024049889-0.0001189124130008639 I,0.0001486196211695905 +0.00001028997939763197 I,0.0001363426884738291 -0.0001306897499033721 I,0.000179299427727638 -0.00001331366372370559 I,0.0001589124205660249 +0.0000936146675297321 I,-0.00001605378413270154-6.253553579344483*10^-6 I,-0.000138331215607709-0.0001383394387067211 I,-0.00001254378195233118+0.0002324069765494171 I,0.0001057411679201617 +8.924434192401323*10^-7 I,0.000282756818244871 -0.0001818578568391357 I,-0.0001764698419246798+0.0000927323112074681 I,-0.0001502986496987735+0.00006611626845992648 I,0.00008989290613015753 +0.0002216087198242495 I,-9.53010783181509*10^-6-0.00002391930375928735 I,-0.00005658715066518441+0.00005129980967343963 I,-0.0001467332318127244-0.0001358537499222779 I,-0.00007394638063785221-0.0001269048434230139 I,-0.00009767868556815969+0.00004174467092158118 I,9.745827585953132*10^-6+0.00004378179968233312 I,-0.00004770966735184978-0.0001926485703975047 I,0.0001301028787028135 +0.00008680426817696472 I,-0.00007396898127454241+0.0001109639531700841 I,-0.0001065227040629694+0.00003987870830945573 I,0.00007214424327354861 -0.0001341660800174968 I,-0.00001426875664049481-7.369340341642743*10^-6 I,-0.0000562895763819283+0.000117064965105784 I,-0.00003214117875107901+0.00012178785276958 I,0.00008214155271132827 +0.00008090877773985223 I,-0.00002380273978399417+0.00006321358688746933 I,0.0001477457716227884 -0.00002948774403841263 I,0.00003310858779416638 +0.00004072076935278526 I,-0.000102105138799117+0.0001358250938125112 I,0.00007413115538152461 -0.0001706651377654873 I,0.0001611779913599891 +0.000037552935284683 I,-0.00003013978186026765+0.0001223774749805446 I,-0.0001732915687403252-0.0001072573467839552 I,-0.00002936928145336388+0.0001185380737937418 I,0.00008616659850986541 +0.00009226001266282713 I,0.00004606577098053086 +0.0001106159846670022 I,0.00008870380200750732 -0.0001530879865435248 I}",
                                     0.33),

  getRingNoRing2Spec: function() {
    return {coeffs: RotorAnimations.RingNoRing2Coeffs, freqfn: function(n) {return (-1/6)+bount(n+1);}};
  },

  DragridCoeffs: parseRotorCoeffs("{-0.07647571383805626+0.2325633807968467 I,-0.2406180437149498-0.2912233909368658 I,0.0646385745788352 -0.05095414308618976 I,0.03362696170984475 -0.1208742156628918 I,-0.005578880796895613-0.004965260519571621 I,-0.08547348177480907+0.02728004548576147 I,-0.03366126048130656+0.0429734479192302 I,0.006207761283238911 -0.03342490886460195 I,-0.01821171625702128-0.01804089106084302 I,0.006805955843593385 -0.04240632295493281 I,0.01169696597874588 -0.01867099046964553 I,0.03246592632347805 +0.01163052182280643 I,0.02343151692332089 +0.02410783283222408 I,-0.006996213083200791-0.004460568947972386 I,0.01490196447946536 -0.01029066931638664 I,-0.00482586917610916-0.02343879924091799 I,0.007230730325537556 -0.001978294712670832 I,-0.02531590769064479+0.00202296037346569 I,-0.0137073929177408-0.009748357611782159 I,-0.007604921002368982-0.01485187790428135 I,0.001407165024025566 -0.01386770721059554 I,0.00001338900330945421 +0.01628535863376228 I,-0.002773165966016865+0.01201591948644156 I,-0.012424691788386+0.008753349087905468 I,-0.01464247798237954+0.009679243947639362 I,0.003996558168379957 +0.0004066762838592797 I,0.01066487361300955 +0.008301853676040076 I,0.004650052820608831 -0.00858427481522507 I,-0.006073691695705952+0.0001234691644646808 I,-0.0140589975053654-0.002528572137349262 I,-0.007914034271124314+0.008931149152521925 I,0.004136691965158473 -0.007166349347796475 I,-0.007237664090584465-0.004647459929079013 I,0.0006850306541043962 -0.01056547268564225 I,0.001533922425839393 -0.0007065302391674943 I,0.01269693467158971 -0.001165370155179694 I,0.00658238495437633 -0.0008648351322830062 I,-0.007055881658338275-0.00735941469292403 I,-0.008687775290819568-0.00556541714810466 I,0.000963496082055623 -0.005084295776785388 I,0.001127767016912689 -0.0100219604606565 I,0.004252742535763633 +0.0008667944062409009 I,-0.002534165752682292+0.00171794052896545 I,0.001583679114775783 +0.00988505439447395 I,-0.0006379915161743542+0.004722147246238002 I,0.007097885665879453 -0.005194671896874517 I,0.005923857897340048 -0.003948386730959649 I,0.006036210505446592 +0.002148767188346871 I,0.007696577961984133 +0.004772542466840892 I,0.0004270871143370499 -0.001484006759453655 I,0.005295907866576364 -0.005027084389233999 I,-0.004945472137142439-0.0003653691682106684 I,-0.002561910645588875+0.005307944290901936 I,-0.003014694713562191+0.00532112311683237 I,0.002430780140472516 +0.00594346804404447 I,0.001721564296052094 -0.004774088836435477 I,-0.0007412074909448679-0.0001402459925197055 I,-0.002825763926925055-0.00733598591556734 I,0.001267509952434675 -0.005783741103131609 I,0.004804091981821315 -0.0002762151338654333 I,0.005793705983240036 +0.004184762096677258 I,-0.001636623328592354+0.000837446328810203 I,0.003515793807093291 -0.002319133358466236 I,-0.001841386458908741-0.00547050164574328 I,0.00237374669865578 -0.0005806867846100748 I,-0.00647984641587995+0.00004909559030885809 I,-0.003087134938952303-0.00446981894571507 I,-0.00187723383694381-0.003479697376157281 I,0.002145291429627943 -0.001659648398183838 I,-0.0007095589523780908+0.005890908424532762 I,0.000593924619153583 +0.003617994337510077 I,-0.005020802365699351-0.001244088851281903 I,-0.005163567702053583+0.0004181131460335784 I,-0.003198312244209552-0.0009797392710849864 I,-0.001335209602640098+0.0004414808722974437 I,0.004546862545327999 -0.003657001598277533 I,0.004046756311687614 -0.002458032327297605 I,-0.002874252143262846-0.001870477105595859 I,-0.003214347279491518-0.003388476965084752 I,-0.0023453221053368-0.00227050534946124 I,0.0003058214338445241 -0.004567793181559973 I,-0.0009765582753636609+0.002995658389900573 I,-0.004770415487799828+0.0006479437400172205 I,-0.001132832097229242+0.002226061430173594 I,0.0006300428298373468 -0.001559696793796167 I,0.005007654370703184 +0.000375734112049231 I,0.00295676985842551 -0.00156064126750308 I,-0.001523301060349063+0.003643369103527061 I,0.001005769215179546 +0.003811347021760398 I,-0.0008870426336979242+0.003042719975509692 I,0.0003157544266640462 +0.000508244333495549 I,-0.003808378518602308-0.003099838761669054 I,-0.00223823312546363-0.002351140982627218 I,-0.0006577527880512087+0.003529784939147347 I,-0.0006150934743403744+0.003484528850643917 I,-0.002338464769995903+0.002112995296947225 I,-0.003874145961383624+0.002234211089210656 I,0.0008814259162042335 +0.0008714701794426672 I,0.001439813605380361 +0.00375740908954846 I,0.0004122015463568279 -0.00207715825522214 I,-0.003258892524033228-0.001720147044963993 I,-0.001763778423260484-0.001781782967509261 I,-0.002194783351762744+0.001712576212834069 I,0.002956894431807012 +0.001614859348420223 I,0.0005526923524943651 +0.0009527098555067834 I,0.003297121605660148 -0.002509207899592203 I,0.003736508347435458 -0.001370031371219219 I,0.0002931430743180977 +0.001332349693939122 I,-0.0002224020002506589+0.003665100524679153 I,-0.001500434043865318+0.001149902520851061 I,0.0022252436190894 +0.001739364331377975 I,-0.00001878179172741523-0.002910233482750799 I,-0.0007114766464196361+0.0008085151256889147 I,-0.003513553990837206-0.00153782717221555 I,-0.00112931476879658-0.001179137347082867 I,0.0005592312259082438 -0.003513433930002833 I,0.000615377263955144 -0.002887892390065257 I,-0.001641892520984784+0.001840890232277945 I,-0.002062513379699101+0.001976343857051016 I,-0.002172300787905446+0.00119628753357858 I,-0.003221597255877839+0.0009826481441356193 I,0.001573899899113582 +0.0001389215662302213 I,0.002295363434943353 +0.00226634118309219 I,0.001294953205202684 -0.00114560058943905 I,-0.001551862525436473+0.0001864110128628538 I,-0.003072108475508012-0.001043184862505419 I,-0.001933434113338083+0.00211329057999205 I,0.001172803197837032 -0.001727489511515009 I,-0.001979977775453519-0.001166983663644324 I,0.00008756949465188327 -0.002624397796180002 I,0.0001861500658452929 +0.0001038217453176414 I,0.003362092557119139 -0.000677650697811112 I,0.001518100146243855 -0.001029908636807656 I,-0.002044642879481657-0.001977355912770755 I,-0.003101401826554218-0.001043054255842819 I,0.0001211436465359975 -0.0006305485751491198 I,-0.0005299557232130974-0.002311220700671104 I,0.002271078836048288 +0.000127510936188269 I,0.000734056344462039 +0.001275397077102071 I,0.0009992951349854233 +0.002706312123843814 I,-0.0008117940064226296+0.002176541402663465 I,0.0004432075420309095 -0.002144758822383092 I,0.00107473191668083 -0.0003504188626589709 I,0.002704105639444027 -0.001148746017360768 I,0.001902762958555869 +0.0001020091042320951 I,-0.0007800128465044421-0.002323538198489093 I,0.0005266752456709249 -0.002924958443402053 I,0.0005362128468659902 -0.0005073369517452951 I,0.001935300710735532 +0.0006608165388041396 I,-0.001402686085275134+0.001725696327692152 I,-0.000401293672976434+0.0007381032450841478 I,-0.002100580166106683-0.00194143330562708 I,-0.002331029531948024-0.00123460602995618 I,-0.001299006004637773-0.0001059879045247686 I,-0.001083181775465242+0.001079218110240313 I,0.002336179910107468 -0.00079932147486182 I,0.001134638206676381 -0.001226601496975561 I,-0.001096776404190868-0.0020557834593661 I,-0.002262035487079694-0.001450316296028011 I,0.000310474196496831 -0.0008360521662062893 I,0.0001889336719919606 -0.002701589399328879 I,0.0006928049921713828 +0.00005857460534801177 I,-0.001532053200785728-0.00004971931729272935 I,0.001152565274185736 +0.00198719828204597 I,0.001668849300722201 -0.00001322791532210889 I,0.001682052705261247 -0.001352298615101123 I,0.0002962499482941493 -0.002553348327543929 I,-0.0005948366544296438+0.0005521319325229103 I,0.0009647322664111961 -0.0004040533844613644 I,0.00006043504978763101 +0.002448969717627057 I,-0.001038197260775046+0.001393784257814972 I,-0.001915209600871061+0.000491488135182091 I,-0.001942976030966264-0.001062497427251889 I,0.001364724679992554 -0.00006511251169513178 I,-0.0002730783112632894+0.0008723936179432497 I,0.001005173070592055 +0.002184234038528857 I,-0.0003169136425756767+0.001226317446476563 I,0.002124746362587392 -0.0005910738459359434 I,0.002236447403543618 +0.0007852555116971346 I,0.0005943709272491296 +0.0005940292441215377 I,-0.0006977919769211402+0.001324520388517403 I,-0.001242449463703183-0.001549363437895209 I,-0.0008270262922444613-0.0005299670970128108 I,0.001896324975874448 -0.00122070748449883 I,0.001310579328233523 -0.001677206859908883 I,0.0007769160448557798 -0.0008869316300755456 I,-0.0002564402628261128-0.0006603065492957565 I,0.00008025114100911686 +0.00229928422948448 I,0.00004797326401455716 +0.001458752941521212 I,0.001749940310873187 -0.0006930852075879689 I,0.001622324813398533 -0.0009089282400863003 I,0.001364876579295248 +0.0004834881214227558 I,0.002004933266030364 +0.001085416346776077 I,0.00026541547146466 -0.0004145819310883796 I,0.001411650430987672 -0.001363315613791269 I,-0.001114828576804461+0.0004847253588497966 I,-0.0003074939222285275+0.002021038125049781 I,-0.000330276939305416+0.00096607441401662 I,0.001324291810060101 +0.0005762455568198445 I,0.0006356985904162555 -0.001620858365042795 I,0.0008959386473670831 -0.0003236010585165377 I,-0.001559417889576586-0.001294823605048699 I,-0.0009318557264085738-0.002005089030200607 I,0.0001569586290511714 -0.00006369253769946412 I,0.001585260134499787 +0.0001469079446575117 I,0.0009456539825218621 +0.00118027749644143 I,0.001328849952855482 -0.0006618082274177327 I,-0.001561400620336875+0.000295662359669587 I,0.00006140644725739384 +0.001079290146461917 I,-0.0009644171707536335+0.001593864384615761 I,-0.0001902472885724493+0.0007389361224174611 I,-0.001624258340671016-0.001153766583525795 I,-0.001711990553799305-0.0008848472178612289 I,-0.0000215717073027916+0.0008672983979955588 I,-0.0001483061896415132+0.001635823086023129 I,0.0005706337901633853 +0.001166315312687001 I,-0.0003964576333235939+0.001306329264252687 I,0.001116166152630171 -0.001092652311367461 I,0.002026496991851821 -0.0002362146144165275 I,0.0002092084714432852 -0.0001374530175067127 I,-0.0006043713576682822+0.001415502345670238 I,-0.001292782151248822+0.0003298853570582774 I,0.0004162683636449314 +0.001303530707063485 I,0.0006586588540869011 -0.001329898621808097 I,-0.0004104774952041002+0.0002253790371663726 I,-0.0008115041368318453-0.001767762920959502 I,0.0005415542439513015 -0.0009606809061451358 I,0.001327808765234803 -0.0009763830504312646 I,0.001019716392166599 +0.0000547731029013716 I,-0.001678453525221613-0.0000645707907571515 I,-0.001278621383266896-0.00005133371898758299 I,-0.0002743420152022586-0.001443884711849476 I,-0.0004320284788439227-0.001828658517713916 I,0.0001184205242127858 -0.000433584810041769 I,-0.0007288482532698881+0.00002808577225288747 I,0.001082227131695833 +0.001402202325971131 I,0.000672060166685045 +0.0006958192685649708 I,0.001113732310502139 -0.001196451887008275 I,0.0009495567238728892 -0.001382453489012205 I,0.0008458369367993047 +0.0001390058958295826 I,0.00178655507957952 +0.0005213547983778756 I,-0.00008135840156474133+0.000136939168864911 I,0.0006646671593326071 -0.001089738059341314 I,-0.001330784808500542+0.0001773882929228575 I,-0.000539680521215386+0.001140502047314216 I,-0.0001837351272195378+0.001322954101108519 I,0.0008901683443559483 +0.001289432018009437 I,0.0002706901860625647 -0.0008906223521427985 I,-0.0001077058552672367+0.0001971009051575697 I,-0.0008326151364199397-0.001588922678933815 I,0.0002585940555060106 -0.001428647096543128 I,0.001023876333219502 -0.0002454960269666855 I,0.001422722416819892 +0.0009352287964904225 I,-0.0004147389323902769+0.0003082879063051667 I,0.0008642686802120342 -0.0005628761014274039 I,-0.0004970872140072963-0.001343685498753324 I,0.0006287813939225635 -0.0001513106523895814 I,-0.001627710062549359-0.00001897610929712634 I,-0.0007373902181473222-0.001250631908049615 I,-0.0004469777469485846-0.0008445396672228279 I,0.0006808739557319194 -0.0002648977886817311 I,-0.000250067064066946+0.00154222171051165 I,0.0001901042877306992 +0.0008625228401114095 I,-0.001331650565720767-0.0006134262907615528 I,-0.001281232190817279-0.0001982918833285495 I,-0.00105302423937956-0.0003096451536345757 I,-0.0007187962620094603-0.00004511721689939158 I,0.001321428603001747 -0.0007600462468759659 I,0.001383239743439287 -0.0006371150596652876 I,-0.0003301987889151734-0.0006100305073772844 I,-0.0005233166172614164-0.001381705455612307 I,-0.0006470147858372933-0.000397058611447432 I,0.0006090989695310544 -0.0009538921864476928 I,-0.0003057839284887056+0.00115955065190693 I,-0.0011236722043246-0.00001115049172411577 I,-0.00102834663336154+0.0005994231714215714 I,-0.0003030607963236831-0.0001197379398048667 I,0.001423890953491015 +0.0007122959955147058 I,0.001272905250468021 +0.0003231868624249551 I,-0.0002687323533339409+0.0008979594440828776 I,0.00013235208336186 +0.001501654112265435 I,-0.0005396352999902772+0.0000516893628277063 I,0.0005333977952918869 -0.0005332122545494526 I,-0.0005770599143423569-0.001276083095260532 I,0.0000111308483830134 -0.0001930255260297261 I,-0.0008630788299931711+0.001308640412038157 I,0.00001574723958238874 +0.001046329397669439 I,-0.001075715461203713-0.0004560270221655315 I,-0.001277938420417403+0.0002529639694534461 I,-0.0008538703682578455+5.702874885187958*10^-6 I,-0.0003659300728851388+0.0005141468844917703 I,0.000493948866286502 -0.001323167166916431 I,-0.0000847733023487308-0.001287583777191944 I,-0.0008338916407582123+4.164699546703953*10^-6 I,-0.001318192203270596+0.000601368722570128 I,-0.000317892481454283+0.0003580821594101948 I,-0.001062551941800101-0.0002747133196799343 I,0.001006544898669255 -0.0002822789766422414 I,0.0007640668759120295 +0.0006852003101005645 I,0.001007435903640381 +0.0004454694561511163 I,-0.0000234940135237882+0.0007575661463572344 I,-0.0006343137648174668-0.001122423113858664 I,-0.0001077954350304608-0.0003759179924904376 I,0.0009942901318121182 -0.001032312812042816 I,0.0008531158597445256 -0.0006942525163663351 I,-0.0004476852811523585-0.0008764412509602923 I,-0.0003198055024226258-0.001397612853593788 I,0.00023082825120357 +0.0002029668083079702 I,0.001107242477883846 +0.0004030679824699999 I,-9.61434529996052*10^-6+0.0008566152637850917 I,0.0004306448086734634 +0.0001190418840413664 I,-0.001116226224048256-0.0008059378655323092 I,-0.00080709061182257-0.0002732324356249897 I,-0.001157192056240337+0.00005734879094449234 I,-0.0008047716819565025+0.00002114607436281982 I,0.0008804881981112128 -0.0007866937367654374 I,0.001036872161815825 -0.0007717374675409032 I,-0.0004263207061564321-0.0004051305198108863 I,-0.00078349056395068-0.0008829619409502337 I,-0.000616998527377966-0.000466035899661814 I,0.0000614915269222439 -0.001194379925773474 I,-0.0003023124286280457+0.0006670092771508929 I,-0.001312575330317278-0.0000371245831795161 I,-0.00004854372519200927+0.0004644901053034618 I,0.0006653301234882883 -0.0005531142352161795 I,0.001068857987985676 +0.0001656824986614238 I,0.0003818200550060197 -0.0005632173400185899 I,-0.001006995896546324+0.0006465835920604887 I,-0.0004268130181107353+0.0003813960933706658 I,-0.0003531040325302041+0.001191564424751351 I,-0.0004962910478786836+0.000969301337029197 I,-0.0003457272070536809-0.0007377043283941678 I,-0.0004799994093210888-0.001210101399181199 I,-0.000164868447952337-0.0003248198153473291 I,-0.001154383714934459-0.00002209704626357013 I,0.00009536409253980819 +0.0006808046848940184 I,-0.0002500399858402953-0.0005653274028994906 I,0.001169308119245462 +0.000186852954727559 I,0.0001551283541770446 +0.0005650236625783352 I,0.0003101092941431094 +0.001150580090712089 I,1.930527084373465*10^-6+0.0006632972035374873 I,0.0006278836515944486 -0.0009531477549724489 I,0.0005321343887407695 -0.000980305291052013 I,0.0006832809971515279 -0.0001067394462379883 I,0.001219140368499157 -0.0002054334247764642 I,-0.0002968933079528552+0.0002930317345332636 I,-0.0006417879624611982-0.0006157362300749633 I,-0.0008577834960526887+0.00039513911273776 I,0.0001314539488776062 -0.00006400649238671378 I,0.001099375302557488 +0.0006585694727551681 I,0.0008440104242473557 -0.000252751292147749 I,-0.0004633619832667404+0.0008086222353989842 I,0.0003527266363494262 +0.0009093603230189531 I,-0.00005633738047410898+0.0008195728124965868 I,0.0002917932603533864 +0.0001985133488083937 I,-0.001184131374880829-0.0002850188561684459 I,-0.0009684376223699516+0.0001558435063873857 I,0.0001674416617010038 +0.0007744627547224357 I,0.0006896062798268281 +0.001004977629110115 I,0.0002068129109560564 +0.0002097929074313546 I,-0.000150324580490354+0.0008791925789824714 I,-0.0004578648864328758-0.0007414381907808757 I,0.0003298133920898536 -0.0007091537054830729 I,0.0002737700651789798 -0.0009226238698247623 I,0.0007223932055473169 -0.0001872468781134179 I,-0.0007458971075957877+0.0006385748028147089 I,-0.00029236639450625-0.00001769932504273267 I,-0.00105272330042664-0.000557227799036738 I,-0.0005874705314299042-0.0005665241591762588 I,-0.0005777993870849969+0.0006988076950015304 I,-0.001065769017072711+0.0005575680863948376 I,-0.0001146959297672245-0.0001111396313914892 I,-0.00009138502283713283-0.0008411314631212792 I,0.0008575427264179416 -0.0000726233213479056 I,0.0002760041657720618 -0.0002573946045191883 I,-0.0002480484947878648+0.001112229316217779 I,0.0001652167194683093 +0.0009229268445954939 I,0.0001771119222685714 +0.0007178471499118363 I,0.0002824328798334956 +0.0002555640662220171 I,-0.0009406929334402157-0.0006170666378846523 I,-0.0006750740580488042-0.0004845557075790689 I,-0.0001879352775922166+0.0008186236364844565 I,-0.0001459170140465849+0.0008953786524068098 I,-0.0005363762964548258+0.000520334421961023 I,-0.0009770066669177262+0.0005451313621659752 I,0.0002088125979684857 +0.0002679810662422878 I,0.0002451934692048311 +0.001012348312291272 I,0.00002383041726993092 -0.0005172001103965805 I,-0.0008854072328591399-0.0005781760161667591 I,-0.0002512264481498794-0.0003972498158188731 I,-0.0004426507423942337+0.0004572221213761349 I,0.0006369937209042876 +0.0007174525111357677 I,0.00004094223537509407 +0.0006088653387550682 I,0.0008698050018594732 -0.0004278786893220339 I,0.001127222791796517 -0.0001112812752113142 I,-0.0001198090433949608-0.00003758250117428792 I,-0.0003696097383438441+0.0006280587694595598 I,-0.0007627282149712899+0.0004158183411169513 I,0.00005742995431121144 +0.0008619661379304508 I,-0.00006607895821958898-0.0007226485624107047 I,-0.0007476979390520403-0.00008173321989099751 I,-0.0005968058924282593-0.0005816322629118282 I,-0.0003478379029787168+0.00008065529277230922 I,0.0007977934750771101 -0.00069552697194701 I,0.0004524527325655535 -0.0008798358844965132 I,-0.0004051926409522549-0.0003040232342881661 I,-0.000909750968472928-0.000284551257029626 I,-0.0004199483946892504+0.0003719555379744894 I,-0.0004388273287395931-0.0002390654172731316 I,0.0007363530100308054 +0.0006476508088471128 I,0.0001981555411900907 +0.001016938161748623 I,-0.0001518364797753917+0.0003211617582968128 I,-0.0009237762326697793-0.0001183350959365974 I,9.447898013556955*10^-6-0.0005662000601155088 I,-0.0003869729028902782+0.0003722100414179581 I,0.0008356053753463319 +0.0004359924734489351 I,-0.0000248948926232504+0.0001336149663565841 I,0.0008833924022878506 -0.0006053923280354084 I,0.0008394369527890428 +0.0001012732494076185 I,0.0004539371343329126 +0.0004829974930719991 I,-0.00009314754488085964+0.0007343992047809862 I,-0.0003510382901208471-0.0006870726859096231 I,-0.00004807275667948233-0.0003787636888016424 I,0.0008204510730085297 -0.0005600148164769071 I,0.0008475414878141203 -0.0005180084576684504 I,0.0000513975131129522 -0.0003641351656721569 I,-0.0001877675981458477-0.0006930258734767767 I,-0.0002815409651674675+0.0007187514625671813 I,0.0001969417041442631 +0.0005272183198482213 I,0.0007872505709363207 +0.0004056639162088848 I,0.0009854742530678987 -0.00009355196740379752 I,-0.0001763512471330496+0.000282604979699486 I,0.0002686922495599171 +0.0008849339616098841 I,-0.0003451649583653053+0.0003261597930195773 I,0.0003671367729055471 +0.00005784363347456482 I,-0.0007475361753644073-0.0006115237680004357 I,-0.0008141701574179281-0.00007932017526111104 I,-0.0003280206505162758+0.0005332586096300231 I,-7.7706044485428*10^-6+0.0009369789802389539 I,0.0003817242256761779 +0.0001666646089821075 I,-0.000120402091973828+0.0004800632272225641 I,0.0003015030538929123 -0.0008402995889875561 I,0.0008397667991414039 -0.0005057148420005024 I,0.0002609773908771263 -0.00005926935789672422 I,0.0001337280310039822 +0.000780022259495473 I,-0.0005165310898353796+0.0003608251853221691 I,0.0005127994814441747 +0.0004173333456887099 I,-0.00008035679050119382-0.0007553758900787227 I,-0.0001332482195522114+0.0002549705279696311 I,-0.0008536511085071288-0.0004400911422647914 I,-0.00005471451820926866-0.0004441921077184515 I,0.0001318970054640341 -0.0008823679127940694 I,0.0001690829944622245 -0.0003917546212760615 I,-0.0006857735772527436+0.0005789183451080847 I,-0.0004681418640089872+0.0006933419460118971 I,-0.0004532878927047163-0.000271073136591166 I,-0.0008458957307325535-0.0001687746412447501 I,-4.652719160305047*10^-6-0.0004756423099920544 I,0.0001728367399011428 +0.00004455709593702316 I,0.0009500738526692802 -0.0001653672590038144 I,0.00061657210587194 +0.0001509871591115381 I,-0.0005678884628227355-0.0004775576615380099 I,-0.0002750437303268498-0.0005032804545548463 I,-0.00009712346675966972-0.0007790531033973135 I,0.0002124914023326547 -0.0007795255591353661 I,-0.000516904437778325+0.0001310724396346241 I,-0.000901488581147485-0.0003089127526354801 I,-0.00005116836139780381+0.0001358155014329277 I,0.0004458222800448448 -0.0003901631134368809 I,0.0006753866058090591 +0.0003378359910588848 I,0.0003828837578348911 -0.0001842753828225276 I,-0.0006842117171525211+0.0005145431243640602 I,-0.0002223616276063015+0.0005895191635197322 I,-0.0002593769193851013+0.0006644511240829411 I,-0.00004044423975433351+0.0003767307495672165 I,-0.0004594351703017915-0.0007372214362280447 I,-0.000404434099724132-0.0006785090950586658 I,-0.0003148174240577673+0.0004093538895971116 I,-0.0004798820628364114+0.0005942344610007233 I,-0.0004049590704220289+0.0003685201405878211 I,-0.000845004293609484+0.0001404737162776436 I,0.0003383887658605943 +0.0001661289315918104 I,0.0001609176746825496 +0.0008576674546396006 I,0.0002620469580062217 -0.000196152388544688 I,-0.0005560206748056335-0.0003861286247232034 I,-0.0002053375923757643-0.000600887085185086 I,-0.0004827394372600902+0.00009975966100422465 I,0.000640172148384092 +0.0004490123389409293 I,0.00023996519543691 +0.0001927665916668331 I,0.0008031142530710649 -0.0003281953902524786 I,0.000812435663875714 -0.000178336056963535 I,-0.0001279344820366568+0.0003637780896512461 I,-0.0002510768475536799+0.0007931010572576072 I,-0.0002621672797790432+0.0002701305819392245 I,0.0005009924547623382 +0.000517936565911174 I,0.00009212125601768261 -0.0005470733596271347 I,-0.0002125170155824885+0.000267456117612611 I,-0.000706766208916496-0.0004518515453851351 I,-0.0001444070891249833-0.0002543539227126258 I,0.0002143679454046241 -0.0008260257613696198 I,0.0001531698197453972 -0.0006489502847745428 I,-0.0004819751365101714+0.0003611639436854794 I,-0.0005638276502592601+0.0004459650055629234 I,-0.0004807428226125315+0.0002302811651384845 I,-0.0007719952470233312+0.0001664974944765727 I,0.0004126343925895344 +0.00002515609916521643 I,0.0005497103998139328 +0.0005682774384410097 I,0.0003325681049764551 -0.0002293884914472598 I,-0.0003870898513929848+0.00005663831320159391 I,-0.0007406628823694332-0.0002837787736364609 I,-0.0004803520713601862+0.000520781248848413 I,0.0003015097850323033 -0.0004280051266932193 I,-0.0005056705134690266-0.0002920115832853199 I,0.00001632187296712466 -0.0006548042950538946 I,0.00003222450999330819 +0.000043312288971596 I,0.0008501789249904634 -0.0001957572163692874 I,0.0003695454405328846 -0.0003137050302304354 I,-0.0005271947362385455-0.0004952076337961672 I,-0.0008300653054342033-0.0002256787081099588 I,0.00001269726043316098 -0.0001104863652552972 I,-0.0002047935755616169-0.0005471567962743519 I,0.0006361893960958847 +0.00002202368754805265 I,0.0002789759211217652 +0.0003756364152160662 I,0.0003063109766336554 +0.0006542839467939052 I,-0.0002228950785850738+0.0005826835402157266 I,-0.00002106676530836185-0.0005878471486284161 I,0.0001900959227890706 -0.00005240038138553462 I,0.0007122614462058286 -0.0004283013186307864 I,0.0004326214306662822 -0.0001187629328233427 I,-0.0002659186442749836-0.0006724249274661386 I,-0.00001611660426624316-0.0007937825858375736 I,0.0002698501592380681 -0.0001245780926368891 I,0.0006743120706256731 +0.00005239329973942385 I,-0.0002901821041133544+0.0004143168428403085 I,-0.0001349554285978186+0.00005036828581213179 I,-0.0007194102563255031-0.0004119244074725369 I,-0.0006727487120653932-0.0002927635571254336 I,-0.0003366511772926204+0.0002260168733249574 I,-0.00034985386292586+0.0005017951871184995 I,0.000478550483102116 -0.0003070943898335371 I,0.0001002114705766469 -0.0005489093817006676 I,-0.0002990606776398576-0.0005399859651578486 I,-0.0007898230326836878-0.0002380010033856705 I,0.00008177412286303621 -0.00001109787534334339 I,-0.0001593110341578894-0.0006766200924481216 I,0.0004447878131940478 +0.00003477285510046816 I,-0.0002576484802277462+0.000282264312599251 I,0.0004139560560045549 +0.0005998706271798053 I,0.0002442361350868466 -4.822826883245281*10^-6 I,0.0004018947455881674 -0.0006716701708727728 I,-0.00004966093970787904-0.0007518177819236158 I,0.0001869180980082685 +0.0002546138603351479 I,0.0005127784515675188 -0.00009237573142332843 I,0.0003129993466642495 +0.0005414794198690996 I,-0.00001055997002745896+0.0002338308215264692 I,-0.0006765313338012999+0.0003813048035076468 I,-0.0006682277917013668+0.00008746180170247307 I,0.0004164051682701181 +0.0001599662712272967 I,0.0002854664284599475 +0.0004836555607793945 I,0.0003354837881820085 +0.0004696081860697634 I,-0.0001352574748671006+0.0006398751386337894 I,0.000285440319484525 -0.0003669719909388076 I,0.000713838702209429 +0.00008647166401063832 I,0.0001821723850182661 -0.000298467871349951 I,-0.0002360572885743835+0.0001306034094131923 I,-0.0007236644723761213-0.0001957657497653334 I,-0.0005564701081659342+0.0002348555251066161 I,0.0003878337344901283 -0.0003392540565512605 I,-0.00006610094514165611-0.000450836440266645 I,0.0004517710017732001 -0.0004627971855892203 I,0.0002942584830504866 -7.003404596486794*10^-6 I,0.000394324623122606 +0.0006159506275325853 I,-0.00003666484993951677+0.0006463599211666247 I,0.00004313137371379609 -0.0004429528714923796 I,0.0002973472257893536 -0.00009900739666333458 I,0.0006319077646766338 -0.0003396788345717936 I,0.0004795877536079625 +0.00003709088210591185 I,-0.0001244025182722779-0.0006015058432811643 I,0.0003018886240314732 -0.000692681860341098 I,0.0001830919794598282 -3.98886231963878*10^-7 I,0.0004216403094190377 +0.0004423647029693035 I,-0.0004235179108727803+0.0002178963655638179 I,-0.00006299333729464779+0.00002081818855448385 I,-0.000141891409977868-0.0007560865526822491 I,-0.00009485389250986728-0.0006216091651117785 I,-0.0003577387589363181-0.0002622207972940769 I,-0.0004659709970959323-0.0002365680073249109 I,0.0004740377230880642 +0.0003004974693630038 I,0.0005735507086947826 +0.0001227640838349357 I,0.0003771284453960643 -0.0003128249521577519 I,0.0003141403218759596 -0.000691701808708222 I,-4.57763134359613*10^-6-0.00005489570596569937 I,0.0006342887053977825 +4.866440151338896*10^-6 I,-0.0002458114177266149+0.0003368888130539029 I,-0.0001670720993685726-0.0003584896509705796 I,-0.0006439345754364646+0.00003627655815922106 I,-0.0001505550160022331+0.0001270669236331747 I,0.0002737290058278486 +0.0006744082852005635 I,0.0005061721585355878 +0.0005000910620849125 I,-0.0002344351830254199+0.00005929000457571289 I,-0.0002624673986424223+0.0005177866111428296 I,-0.0003918813304666041-0.0002654444191541909 I,0.0001972675810929014 -0.000254735598076785 I,0.0001901264168416149 -0.0006460307033220004 I,0.0002862324769565219 -0.0001105756387109986 I,-0.0005918597975624594+0.0003304176617830372 I,-0.0002208409590114522+0.0002370866685464264 I,-0.0004875593782962432-0.0004552939678939335 I,-0.000534379267906215-0.0002719120669371519 I,-0.0004314073386529468+0.00002777463040011345 I,-0.0004639208681692358+0.0001991301034865405 I,0.0003460729644671664 -0.000410083065451272 I,0.0001779211588580978 -0.0006907437387205816 I,-0.0001344478755672723-0.0001114212400825283 I,-0.0005861275502477413+1.909588136749559*10^-6 I,-0.0002781417429651796+0.0003389100558669213 I,-0.0004412818172971198-0.00006949631621397308 I,0.0005424606069754454 +0.0001961666271282727 I,0.000371557832663536 +0.0005510585618342285 I,0.0002209304226437729 +0.000195803500518012 I,-0.0003379187406892232+0.0002672722979558681 I,-0.0003012377265080971-0.0004991004057610133 I,-0.0001419192321381164+0.00009971811646061529 I,0.0005992103090322675 -0.0003646406338512414 I,0.0002008169018616698 -0.0002473715135975354 I,-0.00006781983677496014-0.0006423743738473166 I,8.399603270291061*10^-6-0.0005915134496515953 I,0.000406512099002499 +0.00001579352188703327 I,0.0006311086718795665 +0.0001102107900046244 I,-0.0002599089044574757+0.0001849061078431893 I,-0.0001460656978301873-0.00005958475524984395 I,-0.0005355024283769454-0.0004441817359738465 I,-0.0005030853509920539-0.0003775417442464605 I,-0.0003015723185145245+0.000138406241586745 I,-0.0003487181010268269+0.0003335349594048334 I,0.0004960498439879385 -0.0001537489678818983 I,0.0002437406782752298 -0.000348145973154497 I,-0.0002120903524148978-0.0005232645793747045 I,-0.000567971781583011-0.000365969294766797 I,0.00008291445033878808 -0.000181333218695246 I,0.00004265673351162429 -0.0006849852435533194 I,0.0001488204134714695 +4.245369587658599*10^-6 I,-0.0004276935644665534-0.00005302356892518222 I,0.0003233163230129328 +0.0004470569390337864 I,0.0005108206180595257 -0.0000840959527073547 I,0.000367982710354789 -0.000289258871329982 I,-0.00004183962598550335-0.0006137269756395668 I,-0.0003067529963302446+0.0001075932843192963 I,0.00001621373133465321 -0.0002125418150547636 I,0.00003610068135582821 +0.0006590644316884537 I,-0.0003025619331426669+0.0005512764037116371 I,-0.0002469808833559773+0.0001473281200817281 I,-0.0004304635353083699-0.0002931975171909478 I,0.0002608581387841411 -0.0003686725306140938 I,-0.0002512063090578846-0.00003991322042496228 I,0.0004909681089978457 +0.0004081481916898502 I,0.0002356159973285978 -0.00002215855939863162 I,0.0005886059047933122 -0.0002588400661519322 I,0.0004186933299840088 -0.0001767759781735224 I,0.0001552242233893203 +0.0004857647257884237 I,0.0002001437175493321 +0.0006201585241401289 I,-0.00006243725396972647-0.0001877334326108503 I,0.0002037403288718447 -0.0002801790555509443 I,0.0001284546829661143 -0.000569641446171711 I,0.0002534838520476666 -0.0004424234248423046 I,-0.0004409371082470118-0.00006177011458537634 I,-0.0005895151507557832-0.000272558575810276 I,0.000122755016672281 +0.0001363988213246084 I,0.000480898241345077 -0.0001983524889498529 I,0.0004178031499964732 -0.00008720141451956637 I,0.00006068133468138779 -0.000595581101799917 I,-0.0002405778636744926+0.0001838438187959377 I,0.0003338706855837334 -0.0001437587828921463 I,3.224001482395337*10^-6+0.0005614780944476261 I,-0.0003271167661047268+0.00005020947786812436 I,-0.0005058641328454297+0.0002816594027013985 I,-0.0001419001821651138-1.109365064627026*10^-6 I,0.000605672080852386 +0.0002348561679014528 I,0.0005211709101300239 -2.091962652412934*10^-7 I,-0.0001473782898345078+0.0003823515179364123 I,0.0000436023525875148 +0.0005555676777309903 I,-0.0002409228856263606+0.000261445716098306 I,0.00002868778673095524 +0.00006345947906188806 I,-0.0003394123361308273-0.0005606676355812508 I,-0.0001887055576972072-0.0004417850830466573 I,-0.0002859964411721367+0.0003453813560332358 I,-0.000418356622465359+0.0003702017668269152 I,-0.0003150824771134075+0.0001341889317532686 I,-0.0005899307774639547-0.0001617343378067209 I,0.0002166641738079434 +0.000075806037914058 I,-0.000123019566428242+0.0005049306259576348 I,0.0003856550257323552 +0.00007470112757672597 I,0.00003491755072047989 -0.0002745036377790547 I,0.00009939400305419366 -0.0005792253967116417 I,-0.0002752194846300755-0.0004824042270927682 I,0.0001390426361211351 +0.0003028782631475415 I,0.000246609365558168 -0.00009426845243232671 I,0.0003745533179044096 +0.0004551160582466064 I,-0.00002824605255814561+0.0003638576801224429 I,-0.0004544571135901848+0.0002738949754795127 I,-0.0005517123661240896-0.0001227022316899433 I,0.0003019249860524064 -0.00004213930250557881 I,-0.00004829384226719071+0.0002401943582874045 I,0.0002860179929759605 +0.0005178820600760867 I,-0.00008719980163349362+0.0002987326340183404 I,0.0005486000515093684 -0.00009951685368050553 I,0.0005146813268769336 +0.0003072655564769015 I,0.00006949893485658454 +0.0002031454531976671 I,-0.0003350594966506472+0.0002935799303078093 I,-0.0001307829896305646-0.000433087699990516 I,-0.00005413219880710145-0.000126225858695217 I,0.0006167051315017758 -0.00004086531453470629 I,0.0005172467210567623 -0.00005709162662444504 I,0.0002881051495425742 -0.0002116708596395146 I,0.0002400200165780323 -0.0002974443084414429 I,-0.0003041878617896268+0.0003972063588440182 I,-0.0002087555282680963+0.0004642480065344093 I,0.0002208375430117087 +0.0002937742256392809 I,0.0005128534052585912 +0.000342629155692083 I,0.0001075600915049317 -2.280350275989215*10^-6 I,-0.00006294060733737602+0.0004954736578412511 I,-0.0002291896059058834-0.0002963655054082785 I,0.0003159112288457617 -0.0001336788096803666 I,0.00009667619018055297 -0.0005105351738771094 I,-0.00001740422631673537-0.0001282827920558807 I,-0.0005936927514668337+0.0001276559297421813 I,-0.0004872092440473166+0.000300165640742729 I,-0.00006630584487951651-0.0002270231342152088 I,-0.0004083046199798314-0.0003242361240092854 I,0.0002462786592656913 -0.0002209354205045874 I,0.0001181796242663593 +0.0002583694861362745 I,0.0004918944247246622 +0.0002377864207987249 I,0.00002215579609684415 +0.00016893628492662 I,-0.0001582951322050586-0.0005683919553987771 I,-0.0002291651425466438-0.0002314405558750815 I,0.0004530371275398957 -0.0002525535903515989 I,0.0002856629332765995 -0.0004106369538403177 I,0.0001480834196688825 -0.0003176783459518956 I,-0.00006050284259711885-0.0003239207553462109 I,0.0002773627747919803 +0.0004307461869285653 I,0.0004693579099172182 +0.0002919096915431038 I,0.0001702986527513462 -0.0001858697992921381 I,0.0001336571262246881 -0.0005330792761770731 I,-0.0001175189958565193-0.0002263287365207392 I,0.0002637003989361624 -0.0002884767416851168 I,-0.0002045530899500129+0.000412924010958623 I,-0.0004514491566748369+0.0001448414632145679 I,-0.0003329709818973692+0.0001658235257636794 I,-0.0002361617746844946-0.0002739581872352252 I,0.0004798204817851081 -0.00001274873236126818 I,0.00002587861427521518 +0.0000173794443324565 I,0.000134542770811871 +0.0005832845235976606 I,0.00002818516710378059 +0.0003354828962973381 I,0.0004924070293788143 +0.00003975440936845686 I,0.0005259082167630444 +0.000221324621908181 I,-0.0000357294815811328+0.0001692828889663706 I,-0.0002842970485334119+0.0003521730996825412 I,-0.0002244414100486667-0.0003136824994868994 I,-0.0001143945577936004-0.0000700396793025608 I,0.0005174167124428145 -0.0002569989752937737 I,0.0003634553749978973 -0.0003409315293961347 I,0.0001743769539757123 -0.0002667812824663878 I,-0.00003552326779071645-0.000218664929709038 I,-9.491079891706457*10^-7+0.0005468130681924404 I,0.00002163420643770713 +0.0003775255509252454 I,0.0004299642831619031 -0.0001356508948581334 I,0.0004136541766912399 -0.0002218794605416656 I,0.0003321608848980681 +0.0001174284836519785 I,0.0005056110984311819 +0.0002641961375657283 I,0.0000765514522246766 -0.0001066248834705023 I,0.0003572409238754104 -0.0003439972657381956 I,-0.000264374383714+0.000158261384951337 I,-0.00004880947439957618+0.0005413985001791729 I,-0.00005826563491731124+0.0002006099709386942 I,0.0003491960573548894 +0.0000586871470656024 I,0.0001761544537004127 -0.0004254096566464364 I,0.0003030160275939042 -0.0001309815643641873 I,-0.0004010935106343703-0.0002565157287652066 I,-0.0003103693529572058-0.0004817320124136229 I,-0.00005188349564153578+0.00005332126215308863 I,0.000320484496959789 +0.0000110529794953936 I,0.0003779708806778557 +0.0002941553581548693 I,0.0003881192877409503 -0.0001590273474208116 I,-0.00031615242642589+0.0002337695455010873 I,0.00007804839056994487 +0.0004114579706457546 I,-0.0001059865796792802+0.0003771238526670376 I,0.0001715712371879613 +0.0001814415346350799 I,-0.0004513760574613518-0.0002471385096319167 I,-0.0005204836455374693-0.00002777316485337144 I,-0.0001274348111705987+0.000193919366921487 I,-0.00002818005591835647+0.0004683437325540425 I,0.000308914717311016 +0.00009114068041158262 I,-0.00001166713205150036+0.0001728076761933561 I,0.0002253384286731675 -0.0004913931849883371 I,0.0004421761131643086 -0.0002642381508090691 I,0.0002026372837953377 +0.0001192501153273667 I,0.0000911862788236193 +0.0005177158035579218 I,-0.0002056271672736105+0.00001974123552916723 I,0.0001662159955436523 -0.0004524829465513916 I}",
                                  1.5),

  getDragridSpec: function() {
    return {coeffs: RotorAnimations.DragridCoeffs, freqfn: function(n) {return (1+(n>>1)*2)*Math.pow(-1, n+1);}};
  },

  // Fails at 1280 on Chrome due to STCircle tangent near PI/2, fixed now in STCircle
  SextuplePointTriCoeffs: parseRotorCoeffs("{0.5+0.288675134594813I, 0.133481351607106-0.153908110153918I, -0.0243488649545296-0.0378842674509074I, -0.0613218538774728-0.0669754400591662I, 0.015135681709512-0.00697522391342414I, -0.0608439573972173-0.033017500365535I, -0.0508967617448603+0.0477828435997477I, 0.0393998147726789-0.0271733757486783I, 0.00768114273933714+0.0616970252368099I, -0.0382538423861649-0.0190387604244493I, 0.0101073581497329-0.00863268557355125I, 0.01273333949353-0.024892143042259I, 0.0112255750517602+0.0279478227761424I, 0.00388445431508399-0.0157316844784372I, -0.0147287823812226+0.00581349056027736I, 0.00368893587010576-0.00580518404430077I, 0.0141714404172319+0.0238446361734154I, 0.0132157594301761-0.022778895655711I, -0.00759974849244624+0.0246508059631431I, 0.0064643930465512-0.0246245939894099I, 0.0012051392314197+0.0220417586166957I, -0.0117152746664615-0.0107814702551875I, 0.00256339459762763+0.016792959982832I, -0.00239221285319805-0.0158916332203316I, 0.0000561482579553714+0.0202675992105821I, 0.00300596815715956-0.00883855625768394I, -0.00326622558069413+0.0198171169249186I, -0.00684896946427584-0.0185607307221294I, 0.0031038098606677+0.00845991248266339I, -0.0051908826884302-0.0100980713114539I, -0.00693874411924399+0.0123511374977336I, 0.00458078566470191-0.0142099472188495I, 0.00288682100961299+0.012536929941485I, -0.00633193488024179-0.00950853434414814I, -0.00528693244129987+0.0000116236895918137I, 0.00936548772406403-0.00946915819444722I, 0.00576809045271743+0.00722924620363836I, 0.0057119723402405-0.0162276522466889I, -0.00137053983704712+0.00935897885018852I, 0.00455859382849747+0.00267402212499413I, 0.00146377924899352+0.00219726753794697I, 0.00173231125100959-0.00614709864906035I, -0.00333946655761029+0.000608894483534599I, -0.00391397181081316-0.0089482881962799I, 0.00710184031508497-0.00132889777829379I, -0.00171939082382586-0.00890945113165239I, 0.00883256421891717+0.00476616637991805I, 0.00574827531692807-0.000433362902456989I, -0.00288457282536945+0.00706199177469026I, 0.0027200743846224-0.00155330734364092I, -0.00359313427784301+0.00308764636447952I, 0.00398687229576273-0.0105981379437764I, -0.00339577016985369+0.00313965334798669I, 0.00277133530327088-0.00227015199519076I, -0.00686523947505871-0.00470687226219393I, -0.000523795890689586-0.00318972410580396I, -0.0101692715307038+0.00987132683678745I, 0.00509968105562818-0.0141538804648215I, 0.00591194693013376+0.00850964533561723I, -0.00388768511285128+0.000732953201592368I, -0.00220116285843072+0.00226627339043536I, 0.00211802328033399-0.00716877268769103I, 0.00433421200686827-0.00213933745321476I, 0.00491924361633427-0.00638460479103962I, -0.00341083770002441+0.00439698293255811I, 0.00177608986830815-0.00602523734138973I, 0.00139614863779039+0.00189691672397844I, 0.00109736233513666-0.0100891448303462I, 0.00187724711434068+0.00649112109450338I, -0.00365680309983102-0.00633911119775388I, 0.00188769873119629+0.00573930926007875I, 0.000705557466317781+0.000052853911895244I, -0.000826728519818047+0.00431344511838255I, 0.000406084798761525-0.00251557629176293I, 0.00398217194629063+0.00357045561916378I, 0.0029679364172252-0.00634092750232619I, 0.00350375050905954+0.00602820840660418I, -0.00272125916669683-0.00068513288695575I, 0.000294420837740301+0.00482435933242612I, 0.00201081696262809-0.00250464005321634I, -0.000791293607681628+0.00722061126388607I, -0.000439395426445044-0.00520476737528149I, -0.00209749337032828+0.000350035951721984I, -0.000536146221317277-0.00357521109709816I, -0.00202193172735127+0.00519687649815841I, -0.00220038247171026-0.00359694170946189I, -0.00117730586838661-0.000454269564367366I, -0.00569665614321497-0.00693650679988039I, -0.000718627173741644+0.00228326331722559I, -0.00101485618434891-0.00362275427581956I, 0.00253144397579673+0.00628548913997225I, 0.00183650168866485-0.00671564727059051I, -0.00491300117052039+0.000866385731143415I, 0.00184158213323157+0.00190045143873118I, 0.00115168086022856+0.00330413812834673I, -0.00521197085888585-0.000345436333007098I, 0.00234006863313042+0.00193845693731907I, -0.000951011532387106-0.000932690158451613I, -0.00222773723284688+0.00135256166551264I, -0.00134440602306785-0.00090420166702448I, 0.00171389540041154+0.0014958069788052I, 0.00121171265764572-0.00251108083602215I, 0.00226268974393091+0.000351310026217925I, 0.0000914782114200133-0.0051311159735274I, -0.00112865340317909+0.000179338753344935I, -0.00260058906580245-0.00554462090151432I, -0.00434128981180794+0.00362698502328I, 0.000752559557027822-0.00444900992364947I, 0.00181812985905803+0.00467282599619292I, -0.00392479439986721-0.0031323615885547I, 0.00346063589962853+0.00186315468482406I, 0.00122967012867922-0.00245290922722279I, 0.00274182609548401-0.00108356796179484I, 0.00298359951391887-0.00308131377126542I, 0.000896749436539833-0.000274027361672735I, -0.00145418827497946-0.0039570195298475I, -0.0000172613075349495+0.00310036518101038I, 0.00316512505574898-0.00343896023552289I, 0.000638766729132632+0.00251685285554853I, -0.00185716891189173-0.00381048335781034I, -0.00117544687727385+0.00387391468188199I, -0.000556573787275761-0.00200956021538543I, -0.00150588507894904+0.000586331977542411I, -0.00146248975699419-0.000809938747789288I, 0.00136489105705558+0.00432889305787253I, 0.000474241745137806-0.00119289398571729I, 0.000130803005320012+0.00295043418468178I, -0.000847459125628963-0.00149855544698879I, -0.00197188561660643+0.00219287310624011I, 0.000109072730438173-0.00115903960781741I, 0.00116565708823177+0.00284426464915332I, -0.00150525422985784-0.00275319908783192I, 0.000846680216500245+0.00185742191398985I, 0.000747958906348171-0.000905475707077309I, -0.00131567565586475+0.00466666665485716I, -0.00123807926500101-0.00396584856024498I, -0.00154321605917454+0.00128559945229722I, -0.00358235974861743-0.00192515018985293I, 1.50065043260652*10^-7+0.00241609389584504I, 0.0011317481084468-0.00419145269630176I, -5.55582436799968*10^-6+0.00272982558243583I, -0.000363756423631562-0.00425987004235019I, -0.00297573272846942+0.00438652913850968I, 0.00280993511739533+0.0000909785115683784I, -0.00207671203255445+0.00202856024848113I, -0.00119686838882394+0.00175795659331559I, -0.00140797594304849-0.000231856724916597I, 0.000115987214802057-0.00263922834051829I, -0.00128950756516911+0.00243298962537795I, 0.000248018963521167-0.0016750682522425I, 0.000645906789615451+0.0015569455063688I, -0.000487070668703797-0.00151953623399169I, 0.000525809064481408+0.000325368686201936I, -0.00184184519646347-0.00483308023377306I, -0.00110845398121429+0.00481401401467472I, -0.00319916596573351-0.00280118324705988I, 0.00139636916062955+0.00441746780642823I, -0.00276167024770239-0.000168394018331046I, 0.000204245601550567+0.000506081454524091I, -0.00126576602799889-0.00157455229506643I, 0.00242250337200404-0.000384536666189307I, 0.000292642930308557-0.0017325208847949I, -0.00334731973999224+0.000858960184869608I, -0.000884823064351586-0.00361107547615874I, -0.000796179902326666-0.000771235148945999I, -0.00248799367498863-0.0000263239531875428I, 0.000262421805790286+0.00380987657067199I, 0.00116078492930236-0.00244475712980641I, -0.00134260234120608+0.000787281598631127I, 0.000679955122266763-0.000561322337019581I, -0.000315393969359567+0.0033186087428591I, -0.00130656258142456-0.00302682951424937I, -0.00168346740133514+0.00274536840880095I, 0.000625430637950295-0.00110078007718496I, 0.000138830538143988+0.00244975675189234I, 0.000559036563227778-0.000287212988296264I, -0.0017261263867205+0.00217515018691693I, 0.00061924310909648-0.0034998597789482I, -0.000436455909177517+0.00292571634759371I, -0.000486080370430318-0.00188577068629032I, -0.000608033819416205-0.000374330610907279I, 0.000572324304871344-0.00064304593481257I, 0.00134843513181923+0.00212672535821632I, 0.000562737140233955-0.00246851695471836I, 0.00148295625376857+0.00086489768682336I, 0.000513096590914434-0.00141751240456163I, 0.00115019185900507+0.00361215206629434I, 0.00326365029507179-0.000523990401765366I, 0.000417289326669533+0.00201812767698483I, -0.000396100474061135-0.00134312348266708I, -0.000835870477643746+0.000839115764607196I, -0.000730290720437598-9.37830869706598*10^-6I, 0.00273544555028094+0.00214483160466001I, -0.000845618072389127-0.00434072321295638I, 0.000497761800281693+0.0020192870169855I, 0.00248555217459285-0.00312248097397599I, -0.000432305024973971+0.00293014965239439I, 0.0002591622592592-0.00165150765455746I, 0.000718293303425031+0.00229711082580552I, 0.000446049610618771-0.00191550261766229I, -0.000822697270714898+0.00163931342175724I, 0.000350286317703923-0.0019029659961272I, -0.00125632269379719+0.00209121264336218I, -0.00034926420352386+0.000225496161826841I, 0.000932003428172468+0.00107299079225583I, 0.000788539336649218+0.000233395866366797I, 0.000465863584060727+0.000726060488148189I, 0.001448634955744-0.00135843185657786I, -0.000415122237157461+0.00173468218724012I, 0.000715133627564961-0.000694036842443399I, -0.0019857992763443+0.00111903031729199I, -0.0000169913116274435-0.00187142891889075I, 0.000187057286169973+0.0029057453986292I, -0.000486201007303514-0.000941057031902222I, -0.00118565903296254+0.00279366221513869I, 0.00156262261507896-0.000936640125853269I, 0.000681023132096541+0.000656281446302669I, -0.00102099958842753-0.00211069034971988I, 1.2570064831403*10^-6+0.00233140567508202I, 0.000257936403610735-0.00184327179689963I, 0.00101786112789043+0.00104419496828043I, 0.00027068599280562-0.000367040891834606I, 0.000962301905918866+0.0022396803906767I, 0.00055304100226865-0.00205363543019172I, 0.0002886682169221+0.000524142899292168I, 0.000494300115236739-0.00284807635415708I, 0.00116822736445431+0.00181095432016457I, -0.000676336029917262-0.000618119589650172I, 0.00117481835116884+0.00135962722465465I, 0.000595928045438529-0.001518127594504I, -0.0006767044972612+0.0028243438976306I, 0.000393783442090115-0.00203659624903182I, -0.00150624274535615+0.000496743232109116I, -0.000759210002575679-0.00167125427737176I, -0.000586441662658426+0.00109040529582563I, -0.000380387233210755-0.000564306028118438I, -0.000382999552644268-0.000284892837613193I, -0.00213945136295745-0.00271149293316876I, -0.000538258623507228+0.000707640391571539I, -0.000857750931554578-0.000698471065078096I, 0.00166022508447339+0.0024144933541089I, 0.000786837361009068-0.00283796250382993I, -0.00223440696503214-0.000133586188952557I, 0.0010647666541795+0.00111289611267429I, 0.000968255018665781+0.00141866413599428I, -0.002508989157063+0.000309810163545576I, 0.00118549150969656+0.000828981821644481I, -0.000720876493677117-5.65459857847395*10^-6I, -0.0011851013566029+0.0002702845838247I, -0.000854638073772229-0.0000974357130259582I, 0.00107980994766368+0.000318807281009962I, 0.000582123172374633-0.000632502235563944I, 0.00136950481610358-0.000338477920685957I, -0.000455749127701491-0.00270122790353712I, -0.0000771603772729061+0.000411949024467608I, -0.00187089536403864-0.00272792800878172I, -0.00150614604745995+0.0013935677720655I, -0.000168148507182469-0.000881459433060503I, 0.000450438583740397+0.00178729751028235I, -0.00208215237590697-0.00176575023962222I, 0.00240163842656954+0.000289269131944938I, 0.000559208174664972-0.000751402257504196I, 0.000315420840610571-0.0000985521505088561I, 0.000676688237899677-0.00148276624442066I, 0.00040676249008075-0.000995334647572086I, -0.00145017351023466-0.000863139813711536I, 0.00012882756511993+0.00225247504394788I, 0.00178860723696306-0.00115359396448876I, -0.00049517152878085+0.000699406773468154I, 0.0000448893554584532-0.00117072558665447I, -0.000745850808904836+0.00214966659358929I, -0.00105229222678512-0.0017708330505894I, -0.00103586796622123+0.000842200028147575I, -0.000334421118124536-0.000509782589663688I, 0.000141327524531965+0.00226964360127401I, 0.000248371868221864+0.000226141780865333I, -0.00112661321244155+0.00152896239644652I, 0.0000139718401721431-0.00216990874602965I, -0.000690054456593029+0.00148037112739999I, -0.000507633930810653-0.000906064690505I, -0.000245639357894884+4.78658051693127*10^-6I, 0.000163098145374592-0.000698831883411226I, 0.00116300815191471+0.00136083388376243I, 0.000278454030837794-0.00127473399028466I, 0.000409062722465117+0.00071205882495798I, 0.0003514549396471-0.00134214316751441I, 0.000752046539016641+0.00194019379776826I, 0.00164383224714454-0.000938270290858721I, 0.000273600929452925+0.00128368842509377I, 0.0000223526313173286-0.00094852455559813I, -0.000578282382376177+0.000753345974650767I, -0.0000991288444208889-0.00030242163138105I, 0.00118528300807834+0.000964241734825034I, -0.000682534399358363-0.00201791930424113I, 0.000308198010278985+0.00116571050875447I, 0.000635594237861703-0.00195523365205582I, 0.000119263246893754+0.00168658078824206I, -0.0000332324040342963-0.000933404811707432I, 0.000273154850121863+0.00109807153372061I, -0.0000656803693457651-0.00106902203530356I, -0.000125990947016556+0.00104457659743249I, 0.000254650780409374-0.00106901951716501I, -0.000101533189394533+0.00117308588415641I, 0.0000822159139771701-0.000508862948157605I, 0.000491910686545406+0.000330262448465674I, 0.000231481624156863-0.000924183500253318I, -0.000478603356153285+0.0003573930559108I, 0.000814865538046431-0.0010846480186046I, -0.0000910294005775839+0.00155672984667724I, -0.000328612389833473-0.00126369610647145I, -9.67862777875922*10^-6+0.00112354932070666I, 0.000137069448991416-0.000898174823427505I, 0.000926791856042118+0.00123445427118677I, 0.000321189119957909-0.000510252947557016I, 0.000235944943942943+0.000937402323581947I, 0.000230885370892562-0.00119795412299689I, 0.000119779215574446+0.000962427644934186I, 0.000466073105461114-0.000702182554028799I, 0.000041508806719303+0.00119165349876269I, 0.0000483384663937931-0.00147562282583036I, -0.000268719906173593+0.000975207886738189I, -0.000245753803077749-0.00101393493134493I, 0.000191960112857153+0.000562468355954368I, -0.000472846634040032-0.00107816206673581I, -0.00021602132098633+0.00146060218037641I, -0.000271936696856074-0.000654637664748363I, -0.000193570524300671+0.00101053822978848I, -0.000136130330784467-0.0011878400401578I, -0.000169848457984702+0.000514873826263535I, -0.000368611154911008-0.000941430852244451I, 0.000154114964031632+0.00105513338170337I, -0.0000809940887422993-0.0012174597070992I, 0.000239928936243832+0.00147317798975988I, -0.000250219413535963-0.000635019056477839I, -0.000400670231189622+0.000799660700469704I, 0.000604399541998497-0.00115865791333739I, -0.0000827454802169094+0.00100416375511099I, -0.0000555153433734333-0.000852702467402216I, -0.000139839404680975+0.00116988205031905I, 0.000584574215476562-0.000390948419992153I, 0.00012960312151855+0.000251660162061372I, -0.000230454348824327-0.001079096895057I, -0.000620047978449681+0.00118544805797493I, 0.000368116185186703-0.00106982573760246I, 0.000190819114703304+0.000164707277800193I, 0.0000594251828256304-0.000283255617467153I, 0.000333058509312654+0.000305285097010335I, 0.000646887568558588-0.000472059686967272I, -0.000484352392962277+0.00117270403880343I, 0.000415642046190686-0.000391230725101651I, -0.000328887421661578+0.00047057627644693I, 0.000299887914635983-0.00127064015356561I, -0.000377943024829381+0.000260927134076801I, -0.000158990685943437-0.0010404880250455I, -0.000945588362321089+0.000452191445236798I, -0.000752457377155482-0.00071193290062198I, -0.000710061630315043+0.00191921364538234I, -0.000119657973684695-0.00144728824044027I, 0.000720778230164667+0.000831845024269781I, -0.000727485084611947-0.0000605632251098797I, 0.000339966589228504-0.0000274803443608082I, 0.000315692650491995-0.00107184246587145I, -0.000389828557676628-0.000393166540049363I, 0.000418640960755696-0.00159248817028104I, -0.000567019429104387+0.000023919660674258I, -0.00062807287979873-0.000583997084459203I, 0.000152202167488824+0.00106485316291796I, 0.000531030624276839-0.001736583958416I, -0.0000981614303177892+0.000652797008216296I, -0.000413566612343787-0.000573885233620037I, 0.0000640071281498606+0.00151395714925197I, -0.000248686684972153-0.000733751357459666I, -0.000867732712529344+0.00103856668565671I, 0.000272026203744306-0.00017568869588119I, 0.000660933053240266+0.000968478572279535I, 0.000598846359447345-0.000492351896343084I, -0.000185882306670838+0.00111039174789112I, 0.0000353147664187714-0.000922028419976467I, -0.000325608489308957+0.00151133578680295I, 0.000200319475121165-0.00065276521538058I, -0.000153075068365269+0.000248860116924016I, 0.0000282236866570727-0.000356858218643284I, 0.000289080397688813+0.000425152881968269I, 0.000436844400825229-0.00114137649906272I, 0.000622939345556311+0.000963985758227311I, -0.000423951035323708-0.0006091541555429I, 0.000166767768302407+0.00125131319363715I, 0.00060680785230892-0.000334443140824504I, 0.000144572041314885+0.00074192989589309I, -0.000668299989155218-0.00103936339170716I, -0.0000154386477356913+0.00104304889552903I, -0.000216169126202294-0.000517668838735969I, 0.000877614338045953+0.000943284856496777I, 0.00011172804504803-0.00139309728031829I, 0.000318202575113292+0.00150056662574887I, 0.000598258165351929-0.00100295399033063I, -0.000184117335413757+0.00156249550638651I, -0.000452431982808315-0.000512527985899061I, 0.0000918634690506801+0.000750858607008339I, -0.000281091596091873-0.000665154212870782I, 0.000138208585793171+0.000520838595837865I, 0.00041752925866556-0.000488166818504181I, 0.0000202271166402317+0.000520419313693864I, -0.000482272830223577-0.000311789980105141I, 0.00105991670224572+0.000159236157916047I, -0.000265624523459997-0.00053409938849758I, -0.000376991876700862-0.000208453816433248I, 0.00111475032129975-0.000279904947719569I, -0.000275929662178192+0.0014341433248707I, -0.000755655042585552-0.000984659303915204I, 0.000165433978994008+0.000328450614856483I, 0.00032371454396224-0.000584990122371171I, 0.000973610853198133+0.00116967694243011I, 0.000345148611042805-0.000164138227049286I, -0.0000713746656323154+0.000272202374798644I, 0.000187353730393983-0.000576164254385041I, 0.000456858309678478+0.000881844329648514I, 0.000977035571214896-0.000756965449641519I, -0.000177283137097986+0.00106854801895858I, 0.0000603585565752594-0.00143829010235181I, -0.000264027855022836+0.00114070608933166I, -0.000739843979443457-0.000587168909473584I, -0.000131931400212783+0.000362220746549699I, -0.000538554095778224-0.000482644563532613I, 0.000211342143144597+0.00166688068524968I, 0.000271537885711497+0.0000582474935329815I, -0.000414523997030281+0.0013028643141124I, -0.000577241580218861-0.00122334605317974I, -0.000538427598474263+0.000617248772999738I, -0.000459694665259111-0.000251183536607207I, -0.000206548210226131+0.000452810017027875I, -0.0000736778390967732-0.000820897297213983I, 0.001067516749555+0.000754240452408662I, 0.000107340180105904-0.000424194795273926I, -0.000421900293133337+0.000651434434194963I, 0.000290342417563636-0.00158417373389522I, 0.000550367621392768+0.000920071712192017I, 0.000777841052441449-0.0010898479171637I, 0.000292138904130146+0.000958354114690776I, 0.000786514537764525-0.000942219889000089I, -0.00090367946805257+0.000351811557126487I, 1.88482517308526*10^-6-0.00018155276223522I, 0.000512028422634094+0.00090488281218957I, -0.000648274555966438-0.0012918542672512I, -0.000420673710891555+0.000310916422271182I, 0.00057439278333658-0.00128005934677637I, -0.000192353461257831+0.00059884453185611I, 0.000455477646100323-0.00107854455355746I, 0.000290364340363821+0.00105784279942386I, 0.000393846114701527-0.00095608167360681I, -0.000438774202203365+0.000941106391147353I, -0.000100256087674701-0.0010349868356477I, -0.00052570463235245+0.00103245065590596I, 0.0000986562006221929-0.00028128007304725I, -0.0002808582177762+0.000755196029602876I, 0.000278449491075928-0.00014788106212934I, 0.000154811459081893+0.000943312742330172I, 0.000171343085571342-0.000943497650839386I, 0.000102748484676577+0.000591285959202208I, 0.000208577418741561-0.000199668730255639I, -0.00060053081186699+0.000485064007077367I, 0.0000386106937165324-0.000881057996519151I, -0.0000850829646975367+0.000698465168084169I, 0.0000127748280258995-0.000746580671639663I, -0.000474800186722991+0.00088386709339526I, 0.000281754515418231-0.000509660302892172I, 0.000226713741355851+0.000470615146677253I, -0.0000304914888622274-0.00114542142746301I, 2.68956096242764*10^-6+0.000887481725794831I, -0.000119417179494675-0.000811163027971387I, 0.000249828230509069+0.000829150371734264I, -0.000131232831298497-0.00028409021322858I, -0.0000829451806830745+0.000822927355091385I, 0.0000744227964278626-0.000487276299310156I, 0.000405566332240934+0.000729556421837282I, 0.000452323756710655-0.00064387292616107I, 0.000130751930971162+0.00105122207981961I, -0.000407562607693026-0.000631600711053798I, -0.000082775816216804+0.000869400920178787I, 0.0000434994286406772-0.00035401271657186I, -0.00030377480832419+0.000678086021275442I, 0.0000322656618029995-0.000676834426806851I, 0.000336387234245331+0.000214996407103473I, 0.000102471996670172-0.000654350296513916I, -0.000175554499537498+0.000671250456810697I, -0.000176303678575746-0.00104280406623684I, 0.000374119489832343+0.000497214707961434I, 0.000280266432638636-0.00119842128847462I, 0.000205157650437516+0.000611280573484372I, 0.000201842876181424-0.000965586606518668I, -0.000512015052972773+0.000857841075781371I, 0.00025650764513261-0.000357090891347433I, 0.000368386954312353+0.000397467322663855I, -0.000629842971686255-0.000687156885143426I, -0.000258920825973072+0.000626018842854206I, 0.0000407579391377276-0.00127293014330594I, -0.000056410614748243+0.00074622292616655I, 0.000060932095717999-0.000964346103963255I, 0.000314608763933846+0.000662938239655531I, 0.0000745069471907463-0.000922995710631256I, -0.000135863205414513+0.00094365839361297I, -0.000161632457203938-0.000781141638975842I, -0.000136822387433203+0.00107433090248967I, 0.000310852990044894-0.000353741686588683I, -0.0000891055865924571+0.000645752472740668I, 0.000512997127765151-0.000443626154947867I, 0.0000997431288812033+0.000598653635723299I, 0.000102871594945803-0.000940675342174261I, 0.0000164253118098226+0.000600225896843722I, 0.000383408234516389-0.000470204649447131I, -0.000516964586832059+0.000904237536055258I, -0.0000812340778023916-0.000758453809411606I, 0.000137414025789224+0.000677177784786445I, 0.000051363446956502-0.000588345847231578I, -0.000105342211442651+0.0010923528747135I, 0.00035939410314652-0.000804418194366662I, -0.0000200250167425416+0.000274223014633072I, -0.000326045762090894-0.000826112278110176I, 0.000215953155800737+0.000773194197930691I, -0.000227092973417813-0.000684959215286683I, 0.00022176706487141+0.000464610105434304I, 0.000283018024523099-0.000259050514025271I, 0.00011952705682476+0.000640649554572883I, 0.0000956005797089454-0.000570920426150994I, 0.000246875215337897+0.000338955884826112I, 0.000169846094255907-0.000986743526305016I, 0.00044715249513171+0.000629898745081754I, -0.000153266658338777-0.00012194277264565I, 0.0000701592554609281+0.000669470569341404I, 0.000308234062361119-0.000490746672091173I, 0.0000412638028339385+0.000966696152593979I, -0.000133942838419931-0.000672818453580121I, -0.000325601101552485+0.000176697458821519I, 0.0000818702308327387-0.000566911469337659I, -0.000105783576955576+0.000969115098098169I, -0.000501360596878644-0.000574857274995748I, -0.000285763741107916+0.000176775792976473I, -0.000795703599236966-0.000709903846652128I, -0.0000287182443303993+0.000417412855154045I, -0.000262860257725825-0.000991495223932549I, 0.000202619965668545+0.00108620999765649I, 0.000108187140006702-0.000977106840237106I, -0.000365995767062279+0.000617417499198418I, 0.000386937033285248-0.0000135948621146668I, -0.000185348651460941+0.000955660287807481I, -0.000363842795848511-0.000158743089253405I, -0.00020233126928144+0.000562399058363605I, -0.000370612919502-0.000624923487920882I, -0.0000913884834880557+0.000410647609702397I, -0.000248965865006448-0.000575534563670968I, 0.000322180905557119+0.000527442645849812I, -0.0000307593933754134-0.000268978140702547I, 0.000314031971958934+0.000401781228103899I, -0.000291732789502737-0.000855986714378946I, 0.000403483082838375+0.00084229917425967I, -0.000297978298271613-0.000762194004166373I, 0.000194393578741285+0.000298344427235946I, -8.55346316593676*10^-6-0.0000691102386535775I, -0.000233693635589931+0.000612753816823718I, -0.000210882906841658-0.000839060936151759I, 0.000312105165548994+0.000431181859787514I, 0.0000501762712747961-0.000426608299771814I, 0.0000380997076341729+0.000839715302606325I, -0.0000694565137303364-0.000395671141312287I, 0.0000565910586582571+0.000320377312224321I, -0.0000585501987767663-0.000401736231476321I, 0.000117388114040113+0.00073660201228841I, 0.000284217625565291-0.000444479595161168I, -0.000143779968790554+0.000593640574123619I, 0.000130272100588861-0.000703262388914596I, -0.000103733092385327+0.000672893122086496I, -0.000260650481738424-0.000583295938189799I, -0.0000300204967619837+0.000486808794032394I, -0.000136666906992823-0.000504015445313903I, 0.0000220378488019424+0.000825236267310613I, 0.0000901822283149092-0.000307426333149241I, -0.000148496900198709+0.000728428558421928I, -0.000181304159977342-0.00070467660777927I, -0.000145662888294477+0.000542313075755122I, -0.000110702837190396-0.0003835597405995I, -0.0000937348626021775+0.00046488850258925I, -0.0000250864404253633-0.000561472944332118I, 0.000313871812228105+0.0004571422996292I, 0.000074886367659214-0.000465220968497357I, -0.000128047035144698+0.000585363809009325I, -0.0000683311036985418-0.00083243952783836I, 0.000155328136628359+0.000514435831350528I, 0.0000864613891829106-0.000729971157689209I, 0.000121132504288052+0.000554094317093927I, 0.000153149748313652-0.000828705615312561I, -0.00030836188430084+0.000646471672902382I, 0.0000819897459730068-0.000452407570858548I, 0.00014029662980903+0.000581404118451421I, -0.000164879413344634-0.000480587417060443I, -0.000294442302156836+0.00062592239876743I, 0.0000194453508599885-0.000667397398815887I, -0.000222635578634027+0.000544281676779266I, -0.0000953613014581733-0.000807752066406393I, 0.00015476636317957+0.000539914903868331I, -0.0000396229638420977-0.000740688512579278I, 0.0000583114806040391+0.000705680784657273I, -0.000162387100627039-0.000429602091938055I, 0.0000625157343754394+0.000709951855155325I, -0.000054767334798501-0.000583736085948025I, 0.000218616562392659+0.000885558008341482I, 0.000089802011199445-0.000542092140869437I, 0.000381980218037017+0.000346851791053221I, -0.000098896736590954-0.000207005217196029I, -0.000241741475109842+0.000434314019208923I, 0.000175742970400219-0.000725726361559565I, -0.0000291421984307496+0.000628263222341998I, -0.0000728324528860724-0.000449899211062515I, -0.0000801872876319872+0.00092261127696394I, -0.000199097864620213-0.000361979604242397I, 0.0000815689629394636+0.000615742877857134I, 0.000072679162631008-0.000445373350778722I, 0.0000166156980650547+0.00057640719025093I, -0.0000322315273522226-0.000297053507122195I, -0.0000767795683189665+0.000541289460069824I, 0.000177379999697669-0.000567739455820075I, -0.0000496243452037528+0.000441839030328103I, -0.0000731060457481112-0.000605758236562127I, 0.000131736942625103+0.000487084044816374I, -0.0000847188293559473-0.000623278691201618I, -0.000155480583485105+0.000597262703988828I, -0.000108937946855783-0.000426816776421061I, -0.0000993719469704294+0.000498437059717617I, -1.25917789328604*10^-6-0.000625627740486228I, -0.0000190522280200306+0.000395012122887409I, -0.00011523460071347-0.000512350980110631I, 0.0000279317213939929+0.000468154697604067I, -0.000019085709904901-0.000526675600270548I, 0.000122243319384822+0.000645356415351662I, 2.13757944437983*10^-6-0.00042727152587986I, -0.0000424173248165056+0.000491815706305477I, 0.0000950352793813759-0.000604329141341028I, 3.84071925360497*10^-6+0.000613443272745526I, 0.000111393805286791-0.000349268850145845I, 0.0000297662933291418+0.000578243742392353I, 0.000132634117429527-0.000564083491768084I, -0.000133556860895847+0.000344830856409939I, -0.000151336434177887-0.000405437056916516I, 0.000139388164557719+0.000786254218458913I, 0.0000764665363961983-0.000771400771556547I, -0.000162336734631754+0.000503523594690208I, 0.000399264301761143-0.000450513508341869I, -0.000269576699560164+0.000497760096231628I, 0.0000636537387325351-0.000646518353541796I, 0.0000877154941826523+0.000684623355170156I, 0.000126484904451212-0.000617027706320989I, -0.000124402845062935+0.000529199220983113I, -0.000042077833074914-0.000482128625315392I, -0.000246869846649457+0.000560709456594343I, -0.000265856838695587-0.00027984713279678I, 0.00025716677551658+0.00079410923805668I, 0.0000263450153258518-0.0000683246909859597I, 0.000475718466180729+0.000348117036666362I, 0.000136157377000525-0.0000919867254148605I, -0.000261262531040835+0.000368094572466085I, 0.000240496066388333-0.0003574013296487I, -0.000429273355336938+0.000273001603409657I, -0.0000343399707304292-0.000521785797495701I, -0.00021452301022224+0.00112936151499549I, -0.000350228477573245-0.000284393373996807I, -0.000357333417558294+0.00076569078377911I, 0.000401060587287015-0.00012236515452871I, 0.000265264885846794+0.000391646485888417I, -0.000226259787322822-0.000550328648858587I, -0.000157223453512673+0.000730319713538433I, 0.000278501237151293-0.000624515999919927I, 0.000261892752709239+0.000406354572915334I, -0.000130992228802366-0.000234853963105857I, 0.000316779544218869+0.000752187138055733I, 0.000101076796975401-0.00070306957715896I, 0.0000225904508998217+0.000387178938876263I, 0.00019573037336625-0.000698383739378994I, 0.000168026455237081+0.000743838053403863I, -0.000355602499453485-0.000485724912573842I, 0.000292297149304773+0.000436379106495603I, 4.24307779000459*10^-6-0.000391020278923359I, -0.000373111997250167+0.000780728665953229I, 0.00017895501471709-0.000695040010915843I, -0.0000763006522789524+0.000184531308283213I, -0.00022008009574238-0.000507078038012414I, -0.000327808639848027+0.000247402843406307I, -0.0000115665276275052-0.000583136105793721I, 0.00023679412909696+1.93710513687668*10^-6I, -0.000216979750716538-0.00125524076077006I, 8.16313545201428*10^-6+0.000326288208143309I, 0.0000765950033201543-0.000464016495331257I, -0.0000277189726512687+0.000784777802743877I, 0.000423178227987482-0.000592894953607848I, -0.000313527581527494-0.000119701824332092I, -0.000311753225847295+0.000161258478781131I, -0.0000325238676577465+0.000391265947819188I, -0.000678217016024329-0.000602707298115596I, 0.000235603340499656+0.000355158504058393I, -0.000142299306700145-0.0005206669113131I, 3.6907705701904*10^-6+0.000199330960615773I, -0.000184074948084438-0.000500009591137992I, 0.000214036224913304+0.000580816137053719I, -0.000128591132672583-0.000358458285764992I, 0.000347114639117136+0.000544059860089202I, 0.000197981551234127-0.000740762046325027I, -0.0000352976721046643+0.000539629813705696I, 0.0000595152259967311-0.000847779455206405I, -0.0000109784892787308+0.00037835266301331I, -0.000204961525315949-0.000402106827042828I, -0.0000650873739916824+0.000452529189064544I, 0.0000209455652537452-0.000778358725382204I, 0.000269094586973691+0.000723197079488278I, -0.00006029061362363-0.000337301692843768I, 0.0000874108257182082+0.000520998336990816I, -0.0000233008674704419-0.000338236933173676I, 0.000307001312275555+0.000427801954920804I, -0.0000627632008193624-0.000564018261048749I, -0.000108824763721931+0.000526974233913004I, 0.0000396977048394775-0.000132988660599315I, 1.70771586574514*10^-6+0.000395907839153723I, 0.00010043618749076-0.00049620652135354I, -0.000148708096485226+0.000339676701730231I, -8.92407579446745*10^-6-0.000591182558238843I, 0.000127487966367859+0.000295381001318063I, -0.000155581493988042-0.000559508623016227I, -0.000211855320860858+0.000542723336771869I, -0.000209269741480915-0.000369404091709067I, -0.0000767308185423668+0.000350594841189331I, 0.0000232887357964585-0.00049540961416538I, -0.000027999213888153+0.000230887411121383I, -0.000119899405328543-0.000461592136557912I, 0.000108493699194567+0.000502038326880424I, -0.0000625478054313421-0.000541511069571073I, 0.0000377783330880668+0.000625585490611284I, -0.0000698532967504267-0.000283048814035806I, -0.00017620265631934+0.000462959384403153I, 0.000117446972286457-0.000565448789038831I, -0.000100137663047361+0.000381929455245571I, -0.00022014493325767-0.000422600193733055I, -0.0000292689746443513+0.000493737930760671I, 0.00023837305848796-0.00046809018086043I, -0.000022605271249479+0.000321284371758896I, -0.000043651085447149-0.000572961068627042I, -0.000300139101994084+0.000628536573357505I, 0.000198048737311068-0.000271309628447319I, -0.000223126238461839+0.000288037000980862I, -0.0000644738199772577-0.0000935420669277884I, -0.000146965091962176+0.000132523555593899I, 0.0000857522338256213-0.000557036984444665I, -0.0000860486222801204+0.000506979407399117I, 0.000083475735846972-0.000457996440928869I, 0.0000135612786144212+0.000439971911438632I, -0.000132970266430904-0.000385508156964372I, -1.83617105941273*10^-6+0.000345952826702279I, -0.000200034873298833-0.000702991280285027I, -0.000100238745141137+0.000919767771206159I, -0.000251125150779062-0.00043074922079265I, 0.000421873671816607+0.000666059336774459I, -0.000409818790102563-0.0000556439561523558I, -0.000114560171939387+0.000131339760852736I, 0.0000917838973246921-0.000423282861523894I, 0.00012112554028599+0.000215558655564238I, -0.0000534755268821725-0.0003832536816998I, -0.000577359891823619+0.000645179939502723I, -0.000362080031562705-0.000526193241185989I, -0.00012670626690135+0.000301241774534538I, -0.000154544121166114-0.0000609301508209502I, 0.0000630316017949117+0.000645078791444727I, -0.0000755035483312582-0.000337009826087155I, -0.000252172342814711+0.000334846968505058I, 0.000323081910184987-0.000244684130774371I, 0.000040082627968899+0.000465943946203354I, -0.00018510680083482-0.000597408037490123I, 0.0000258382327056754+0.000670993732089952I, 0.000131058350605424-0.000555247320809378I, -0.000160166530049882+0.000408818994101532I, 7.73464077838291*10^-6-0.000330097614437422I, -0.000224603513642536+0.000483041479352942I, 0.0000252112071395026-0.000747312189999968I, 0.000157835279677086+0.00046523282502446I, -0.000140085059590305-0.000495989582289052I, -0.000244868567693397+0.000232505247324691I, 0.000221473137824562-0.000374545221155003I, 0.000127180187572568+0.000496973855766347I, -0.000102908782896462-0.000548361382831776I, 0.000072968645262193+0.0000419989766598932I, 0.000236003688909546-0.000372862029367918I, 0.000329035121943861+0.000536766056327666I, 0.000538061301129104-0.000612650714595158I, 0.0000485944238392844+0.00042233498598026I, 0.000041135183441528-0.000218164654284741I, -0.000156206704933403+0.000318676533761119I, 0.000102474995089333-0.000136437118471365I, 0.00031634201160763+0.000132501709823729I, -0.000415847935574327-0.000592564686052283I, 0.0000990636728808757+0.000332352764661893I, 0.0000708885187352841-0.000848844489083364I, 0.00010780410027309+0.000556469350763805I, 0.0000259503577868478-0.000476243224599187I, 0.00016815324287724+0.000388331602032675I, 4.00886640451867*10^-6-0.000528245474097231I, -0.0000637497197990377+0.000541629612316147I, -0.0000653014184236829-0.000442423411680054I, -0.0000361495008659182+0.000648109722771366I, 0.000157488572295148-0.000235098995187253I, 0.0000901156924659755+0.000410146221950581I, 0.000314127763749542-0.000341615023205244I, 0.000153779729488527+0.000213586686210387I, 0.0000443616770326672-0.00036185733002872I, -0.000148251269907725+0.000377754772222206I, 0.000305312601214189-0.000521962013375595I, -0.000247473642852364+0.000653166890835454I, -0.000114185206889539-0.000380997459370825I, 0.0000322835522732957+0.000799960517953402I, -0.000200785045034368-0.000196188528856859I, 0.0000800048479187479+0.000752759424209274I, 0.000269365059497971-0.00041264967602949I, -0.0000207979123570833+0.000294153214382572I, -0.000303601432922518-0.000137048450072099I, 1.45765410993774*10^-6+0.000487788707139874I, 0.000220326795766529-0.000446930033045464I, 0.0000655888893391747+0.000128632150716155I, 0.0000999488210551576-0.000413847271644389I, 0.000390263256412548+0.000451740797335754I, -7.5492251175926*10^-6-0.000737657472346591I, -0.000289383191025493+0.000237438910307831I, -0.000227474646927397-0.000639906235853675I, 0.000103969400797609+0.000290898182513566I, -2.7649500262471*10^-6-0.000380345110952709I, 0.000296476975913875+0.000121933772362568I, -0.0000666421993612744-0.000551715768938927I, -0.0000302720124603363+0.000691033085004434I, 0.0000954333181297402-0.000619710050811209I, -0.000291775184817668+0.000575850730150973I, -0.00031149972727086-0.000338806988591449I, -0.000240442122665052+0.000169856315881443I, 0.000270968232274906-0.000271169901474717I, -0.0000894393183764508+0.000132834799640276I, -0.000352936646341563-0.00062690269145314I, -0.000162219182560942+0.000369733428299481I, 0.000141779228829384-0.0000256618017193356I, 0.000302244120080131+0.000276656317581438I, 0.000100309965819106-0.000669521973682517I, -0.000629323727813243+0.000174484430944176I, 0.000190870953340066+0.0000392097760656613I, 0.000136650111775349+0.0000542731631446877I, -0.000474754249857222+0.000025102713153653I, 0.000285800573453024+5.83665248508641*10^-6I, 0.000134504767790051-0.000184599588964201I, -0.000286806514571939+0.000299205400761649I, 0.0000255754205073283-0.000152415518056574I, 0.0000673755180650051+0.000228390803293024I, 8.51484637929557*10^-6-0.000399615057228647I, 0.000135994888952345+0.0000885048478786785I, -0.000033509850177258-0.000788893199209062I, -0.000406632940277025+0.000459708244993528I, -0.000398626516930578-0.000635615846140662I, -0.0000975168108366986+0.000821473194213438I, -0.000441676790061435-0.000353506493287488I, 0.000169147670252974+0.000198377057278205I, -0.000202695247314803-0.000334384109377776I, 0.000439632277757035+0.000145187570813598I, 0.000023951268230855-0.000332999869950519I, -0.000466230206814366+0.0000516487549443797I, -0.000108940982062613-0.000664483068703983I, -0.0000145167674182634-0.0000313605828135805I, -0.000440119429287288-0.000141720290050662I, -0.0000405262516433049+0.000675354422646506I, 0.00010683167291078-0.000357268378142626I, -0.000205265489414977+0.000138277528313348I, 0.000174711268090476-0.0000834672282195027I, -0.0000698523084108997+0.000533262052115034I, -0.000174218098780189-0.000663405755363979I, -0.000207037165574633+0.000554402549019013I, 0.000144870621341326-0.000364287249034398I, -0.000126296996182361+0.000397187730683129I, -0.0000170772335610323-0.000159007706829064I, -0.000344619248571609+0.000351750893973482I, 0.000196639548615904-0.00076331426624123I, 0.0000673564743687467+0.00051161072625505I, -0.000141635155374305-0.000513383050479781I, -0.000170274131891799-0.0000206469950482325I, 0.00023340978037926-0.000188835684017694I, 0.000149624996998978+0.000560743940740496I, -0.0000428967370732578-0.000587898794560588I, 0.000303051161219458-0.0000502345312487135I, 0.000306701216427109-0.000156830827819536I, 0.000333564302407746+0.00073516172944365I, 0.000827714867363968-0.000286661792036334I, 0.0000285215607156961+0.000410453133761799I, -0.000108548582269495+0.0000165431439843083I, -0.0000605431207920937+0.000126707712041867I, -0.0000279819963551336+0.0000293770557791053I, 0.000534847288739705+0.000109445759300073I, -0.000395581224789517-0.000855680790381362I, 0.000370420290047393+0.000302138582898005I, 0.00027811482504158-0.000882257382611843I, 0.00024252564110255+0.000682338442020702I, 0.0000630395347102531-0.000194636358839715I, 0.00012117672492368+0.00036050198238363I, 0.0000114136327198063-0.000325008439998476I, -0.000132482258134058+0.000342927735924947I, 0.000139730814919678-0.000456386470760739I, -0.000116780083713548+0.00046573890753955I, 0.000166846430570281-7.07390410465296*10^-6I, 0.0000920495685900419-0.000094013672998858I, 0.000242716949193039-0.000241732454285873I, -0.000279769297373196+0.000120255242779186I, 0.000361253649176461-0.000548856610674837I, 0.0000263416833286289+0.000560034274203732I, -0.000018356309534221-0.000359263498732137I, -0.000159335515449103+0.00046380485942663I, 0.0000245221364171617-0.000353364321855748I, 0.000349671022510806+0.000405717900767127I, 0.000109096762166098-0.000187574343190796I, 0.0000458729884218938+0.000452164377008501I, 0.000154502569046203-0.000435469775512839I, 0.0000316734534956307+0.000260066924263659I, 0.0000494054752205663-0.000286796464816016I, 0.000037235456435267+0.00043421419464772I, 0.0000204754733081177-0.00048453064284987I, -0.0000278327215258883+0.000309268220329028I, -0.0000567913618998823-0.000318770993485861I, 0.0000940441119495734+0.00027219340913413I, -0.0000987348167568677-0.00040075596405435I, -0.0000558200867165479+0.000448106605547527I, -0.0000362485308819164-0.000273566317449138I, -0.0000521912740370263+0.000397256490172824I, -0.0000855926057435695-0.00045746260438649I, -6.77099855302051*10^-6+0.00024369654369286I, -0.00012963965047148-0.000315515418976305I, -0.0000655741074591956+0.000339203403536216I, 0.0000496465065658389-0.000411232880556268I, 0.00016235634266958+0.00043253189486989I, -0.0000829012752921689-0.000297475551834076I, -0.000128148571715408+0.000176454053484171I, 0.000214806148381852-0.000494958016951187I, 0.000203506155430363+0.00036931302274997I, 0.000305843107016297-0.000570715018867274I, 0.0000489296496101227+0.000400663511720005I, 0.000261898194079651-0.000227659842659304I, -0.00023338102105167+0.000186788979242468I, 0.000101582515953529-0.000128090368845912I, 0.000136200161649435+0.000170090494390762I, -0.000425519216049547-0.000463336470790616I, -0.0000408888163431541+0.0000870703003429774I, 0.0000188120162221035-0.00074431137938935I, 0.000116709459524569+0.000292677005566001I, 0.000198799808855442-0.000485545959366929I, 0.000153899485334563+0.000381451758300759I, 0.000124284513532979-0.000484257928351547I, -0.000159015973558526+0.00053920731552355I, -0.000140628829683052-0.000521811808017036I, -0.000106422774557336+0.000654714285564203I, 0.000346963819589116-0.000233890164037639I, -0.00029407270235442+0.000374113548147231I, 0.000411483377606377-0.000328494244955316I, 0.0000826292323690888+0.000494736732471752I, -0.000204162258186006-0.000646135489293725I, 0.0000578800906189417+0.000190863375707047I, 0.000430822623981511-0.000313668999657771I, -0.000273679444872299+0.00071828202452401I, -0.000166067735721078-0.00040579560864326I, -0.00013544078115089+0.000336632081872076I, -0.000117132092627552-0.000409626578145048I, 0.000132894958387458+0.000742119763721629I, 0.0000912060709844789-0.000535076212462214I, -0.000219600410634307+0.000185285580280802I, -0.00048074178384744-0.000167527027713126I, 0.000134066677499094+0.000255848572065705I, 0.000023205003955828-0.000119396277843854I, 0.0000678751002789413+0.0000342229214373417I, 0.000369310455198757-0.000461886886560673I, 0.00021694985913927+0.000408560819730323I, 0.000143616265879365-0.000581744052467169I, -0.000281897460891892-0.0000597506261831341I, -0.000334594038624822-0.000683899627197001I, 0.000131589692101451-0.0000403453083201716I, 0.000373995948603447-0.000192210114560296I, 0.000281744134594134+0.00020634856817497I, 0.0000861114876526606-0.00068156399380771I, 0.000204935206096896+0.000441874576592564I, 0.000049385316986803-0.000313837404065635I, -0.000523336399031625+0.000656311643246404I, -0.000152711504822862-0.000436465448679761I, 0.000243542603137229+0.000217371365193946I, 0.000197306754227676+0.000211919958725645I, -0.000318383643608032+0.000485127314657513I, -0.0000966320411358635+0.000221367133752319I, -0.00022388238652833+0.000342931838557126I, -0.000309326457893615+0.000172503602150874I, 0.000562360664646651+0.0000476929560075045I, -0.000300442568025566-0.000444734324977272I, -0.000122972955148382+0.000477241616165792I, 0.000475249637718181-0.000634288711461041I, 0.00048998428600924+0.000314636969333905I, 0.000290064192044423+0.0000172627145529438I, 0.000164836529519625+0.000435142283495506I, 0.0000529950951547306+0.00013096714994597I, -0.000224791626055403+0.000394367938003496I, 0.0000378067992801149-9.950066910821*10^-6I, -0.000085985783900043-0.0000290202503927077I, 0.000325789169405895-0.000340478907021107I, -0.000186514277543948-0.0000207569200419836I, -0.000239821081453233-0.0000378837603927063I, 0.000170939396117186-0.000139469582068454I, -0.000196520316295954-0.0000285584912687602I, -0.000307890831904851+0.000166566931902808I, 0.00048752316863044-0.000302737791658371I, 0.0000416248275799749+0.000541654048022913I, -0.000251939376955302-0.000132099475734737I, -0.000137834431837808+8.79653432068956*10^-6I, 0.000123617872823596-0.000343316120932814I, 0.000255734988059514+0.000399950086289335I, 0.0000811228273909803-0.0001650808734551I, -0.000218256085547271+0.000298241771972229I, 0.000222880750298481-0.000202808447296727I, 0.000196933457725611+0.000226083793946443I, 0.0000791280471746126-0.000430024097345842I, -0.0000281705899820509+0.000493136929264458I, 0.0000405754232232381-0.000505033886648643I, 0.0000872373241471813+0.000328900382582988I, -0.0000931334755427795-0.000139058885442382I, 0.000134704861961764+0.000341413612022214I, -0.0000355934585569266-0.000393505810301979I, 0.0000296342005959537+0.000346439613466129I, 0.0000516345687660757-0.000389029516126094I, 0.0000843777883244425+0.000431671661139859I, -0.00016679876545489-0.000309616011765825I, 0.000115531126173706+0.000208975947868873I, -0.0000431454015656301-0.000304189782881601I, -0.000107609100638449+0.00049726341215662I, 0.0000657239040869862-0.000464228708158194I, -0.0000761285895988912+0.000324723150379642I, -0.000180209103473731-0.000278592155980384I, -0.00019001499282619+0.000162598170377743I, 0.000175518909596982-0.000295311889654628I, 0.0000217068580610426+0.000147804331836002I, -0.000114304863116186-0.000513447126732045I, -0.0000803935664782824+0.000308265207496587I, 0.000114896645413801-0.0000360730827883129I, 0.000146817102506732+0.000192625253440297I, 0.0000421766982754442-0.000406417653197951I, -0.00030095658448194+0.000158311072056302I, 0.0000523777785378707-0.000213394531523904I, 0.000171126180304844+0.000042795851565689I, -0.000135718275282549-0.000152654761902315I, 0.000239479177423729+0.000126674639731427I, 0.00018967735548402-0.000101050110242793I, -0.000169704259400981+0.000326439425563325I, 0.0001078641773988-0.000112967045989796I, -0.0000950069349459949+0.000157272679314002I, 0.000114040776586739-0.000437371006809481I, -0.000113965603164252+0.000116540818685715I, -3.07063697339109*10^-6-0.000288134487616128I, -0.000303987887622276+0.0000887280173916599I, -0.000159714715124457-0.000191701571260648I, -0.000238852734470329+0.000595657766918264I, 0.0000288819313469233-0.000512370169562939I, 0.000214256531335554+0.000290000511227894I, -0.000124116155911141+0.0000105489185987687I, -0.0000645991488851994+0.0000401709221319565I, 0.0000716363776659575-0.000399323122288532I, -0.000130449394844548+0.0000432013075307815I, 0.0000429421228512177-0.000470659730332609I, -0.000282598796917189+0.000234543014565803I, 0.0000122014684571265-0.000154845577555412I, 0.000103580764595315+0.000248592075373851I, -0.0000565450682767353-0.000616498754002854I, -0.0000298320371911956+0.000344361983496585I, -0.000045635771261235-0.000239543667695959I, 0.000189396700692826+0.000391980928067399I, -8.82956393853477*10^-6-0.000112015862178323I, -0.0000536552960227092+0.000522004399123005I, 0.000203252273114533-0.00026580044725042I, 0.000193501864333049+0.000136087636984304I, 0.000244950268203581-0.000486448090250544I, 0.000150202842356064+0.000443353211362795I, -0.000125121129852358-0.000219850093038933I, 0.000192674689094609+0.000469659921201497I, 0.000142498216476289-0.000288545351777984I, -0.000249256074315124+0.000415654660396061I, 0.000156656349413292-0.000321011870951939I, -0.000198829658620287+6.59669529912852*10^-6I, -0.0000978496338540546-0.000476276114196039I, 9.83742002054874*10^-6+0.000142690978559526I, -0.000102976461202159-0.000112294533920923I, 0.000125143529858594+0.0000382550245747805I, -0.0000983560382173436-0.000706301637382341I, -0.0000432898235338762+0.00013271323040517I, -0.000260279518029505-0.000141796382614941I, 0.000219337986495542+0.000587693692713091I, 0.000278317594990798-0.000405579555354458I, -0.000156259468574358-0.000264485534811569I, -0.000140812117645415+0.000111147580005038I, 0.00030802172173817+0.000356766195904358I, -0.000540959296024228-0.000389640734134722I, 0.000382928480215447+0.000398483644359661I, -0.000188098286858022-0.0000333786036123379I, -0.000104460455833024-6.77027410432395*10^-7I, -0.00023306771653505-0.000111279064461428I, 0.000207564768500875+0.000193496552258017I, 0.000120903652469065-0.000210781933869319I, 0.000318576565630221+0.000152416451196227I, 0.000167890549300125-0.000431251776316413I, -0.0000163585685251329-0.00013084855592455I, -0.0000844029777242281-0.000654887089795115I, -0.000500850808249825+0.000182649073706485I, 0.00010166808502311-0.000482027458019247I, 0.000141445914656763+0.00051902659042516I, -0.000331937354487277-0.000455223490718521I, 0.000379589627737052+0.000370997672233356I, 0.0000817738425191418-0.000213026225491022I, 0.000334125513014321-0.0000595014458085029I, 0.000268900771327377-0.000288453481810491I, 0.000213963299463334+0.0000176660959057457I, -0.00018965901617526-0.000439531611393321I, -0.0000648016451399285+0.000357632921939963I, 0.000262640690290819-0.000257787854646613I, 0.0000414688905822042+0.000209673096229622I, -0.00010906051981065-0.00030850343165906I, -0.000137169611463125+0.000380452870049703I, -0.0000547475065826844-0.000338022367310656I, -0.0001535190485762+0.000158534207417713I, -0.0000583639742872592-0.000160610120259207I, 0.0000520920519381721+0.000397537239359513I, 0.0000311522131075018-0.000153066118927663I, -0.0000793141333832332+0.000294240451525101I, 0.0000175800759360294-0.000320921939178794I, -0.0000907471874649526+0.000348862451571338I, 5.83529962357345*10^-6-0.000252355781424237I, -0.0000144632233756167+0.00017699962460548I, 0.0000185253759800712-0.00021254741614714I, 0.0000575224993545287+0.000223345561451959I, 0.0000478387035628144-0.00035279256217748I, 0.0000923537428811242+0.000257669975760872I, -0.0000654051696613583-0.000225185359175905I, 0.0000871107173964784+0.000306685388369408I, 0.000132400883733668-0.000354894303401394I, 0.0000260431163275243+0.000225972761627657I, -0.000179965960565833-0.000276761754026216I, 0.0000176403079346653+0.000427360166932651I, 0.000117633710097804-0.000223459916785648I, 0.00015416438796483+0.000021904216948267I, -0.000175269686182293-0.000189545901709746I, 0.000161001852440574+0.000442401946365809I, -0.000204402657320972-0.000523413634759302I, 0.00015704884639854+0.000513586825853164I, -0.000231513632288709-0.000180442515535876I, 0.0000528957420906262+0.0000675369640134399I, -0.000239491794435226-0.000256939211224471I, 0.000182586407274496+0.000290085489238346I, 0.0000587906072040626-0.000144430854459636I, 0.000292960157154833+0.00032101927878201I, 0.000174071602347953-0.000294579624252057I, 0.00025894615778424-0.0000393408579315313I, 0.000130553877197849-0.000628687732410522I, -0.000317443818312429-0.000138888448921112I, 0.000205943896704919-0.000319822380638845I, -0.0000609988177173938+0.000564658338357849I, -0.000182019238709283-0.000714472608085643I, 0.000321183769135226+0.000632299782900161I, 0.0000102944801748317-0.000114066271843586I, 0.000581382170934439+0.00026076398532881I, 0.000238456862911159-0.000013485603765309I, 0.0005049473024241+0.000177422737902531I, -0.0000766069189944103-0.000622329457497568I, -0.000160588546641973+0.000328547776847055I, 0.000267078182898884+0.000158590034492462I, 0.000106401151236179+0.000209684072962443I, -0.0000253215650218261-0.00043094992659404I, -0.000336272826974141+0.000173867425773014I, 1.85074292380727*10^-6-0.000521232453018438I, 0.0000546096357390988-0.000135638413184142I, -0.000332314050844711-0.000311640847071256I, -0.000283380845541389+0.000531504444844308I, -0.000385536710842946-0.0000182208979038914I, -0.000205324098225909+0.000042496897865633I, 0.000198483164800385-0.000380835060184908I, -0.000206879441469351-0.000020058574993121I, -0.000205926503545243-0.000357930109272082I, 0.000333824443432556+0.000158170407296645I, -0.000143137289592385-0.00027319656968083I, 0.000120508511646251+0.000747857807517617I, 5.53660862059186*10^-7-0.0000932092216991295I, 0.0000145078735077505+0.000317785861421938I, 0.000283927489188217-0.000278769904387335I, -0.000187297712821291+0.000587356159388997I, 0.000132585957342193+0.000269734385740638I, -0.0000571641620982579+0.000447594928092389I, 0.000163378809844506-0.0000837346647232597I, 0.0000400282897101198-0.000182744971374854I, -0.000343752096460812-0.000199398141537403I, 0.000056175684353497+0.000721939598111416I, 0.000266421606947116-0.000744241424016119I, 0.0000623499241766684+0.000126941738514197I, 0.000500001283071113-0.0000812742359678981I, -0.0000990019031693818+0.000237989073381775I, 0.000187579129428693-0.000152381992634844I, -0.0000598326990824724+0.000482994748869121I, 0.000161202152879716-0.000176421609031878I, -0.000184111772812185+0.000123392578054012I, 0.000140318684266508-0.000339704178795055I, -0.000267323194039816+0.000148654739981129I, -0.000172544831458737-0.0000171133629480913I, 0.0000925036969594481+0.000115508116637577I, -0.000053520815213951+0.0000267847593941196I, -0.0000473322608051816+0.000184345406056221I, 0.000302624514883085-0.000218177051151807I, -0.0000209209872759071+0.000363664128891289I, -0.0000780065464729987-0.000156064702425I, -0.000169666633881205+0.000103978617115382I, 0.0000581456846068731-0.000274450535826869I, 0.000184424613891549+0.000415528162489162I, 0.000035686483981877-0.00010363076737636I, -0.0000917307859655525+0.000304563358165431I, 0.000178848509057453-0.000235141524692695I, 0.000111070144911844+0.000199312640019442I, 0.0000775675909643869-0.000230288711520966I, -1.20816733914552*10^-6+0.000385779688606285I, 0.0000414011017330598-0.000443089550011131I, -0.000015944387820044+0.000230631953163768I, -0.0000658267384356178-0.000207435979929504I, 0.000131290777560054+0.000166688816718714I, -0.000138584758028568-0.000338692858927955I, -0.00007302522932415+0.000373754864526862I, -0.000094316430504358-0.000222166706071743I, -9.51641641293409*10^-6+0.00027486657668567I, -0.000061401060220805-0.000297052098475779I, -0.0000100928967681996+0.0000747876178961292I, -0.000114560172340815-0.000268998262764951I, 0.0000593851606048676+0.000350510721144675I, -0.0000311993296221303-0.000372683920888414I, 0.0000249909834807394+0.00045103513944832I, -0.000105739717156077-0.000140164767138991I, -0.000147095192380954+0.000212816164393875I, 0.000216440657193551-0.000298691726351988I, -0.0000757772366945894+0.000277399414604813I, -0.0000536468600366903-0.00015345474305514I, -0.0000667037439685366+0.000341936920821329I, 0.000172405255639531-0.0000556421693368993I, 0.0000852206459724338-7.48906698836425*10^-6I, -0.000140741364861533-0.000307155127005703I, -0.000182437625083837+0.000422433341676405I, 0.000198299641351343-0.000402578353727827I, 0.0000839307298796774+0.0000468440331394115I, 0.000162965830939659-0.0000199147161211761I, 0.0000484499314913454+0.000115966084067575I, 0.00018232552450137-0.0000958217055537432I, -0.000139145683280744+0.000379609411742272I, 0.000129756412822191-0.0000989564302457631I, -0.000122771933676426+0.0000866817180790543I, 0.000140650265979723-0.000330891234181055I, -0.000179859889331628+0.0000514822382710796I, -0.000160631608048626-0.000147244414362916I, -0.0000504367941221158+0.0000889764879882495I, -0.000191362066952458-0.0000503918927947558I, -0.000137715268521986+0.000317181439497845I, 0.000218284911188182-0.00023564753651852I, 0.0000613810976185928+0.000331597844624665I, -0.000201499856446735-0.0000821217356811949I, -0.0000471935435854997-0.0000375229863811919I, 0.000104053706864528-0.000269524156398346I, 0.0000968085510740261+0.00026504649450821I, 0.0000579163464836134-0.000176452662987739I, -0.000182224851155295+0.000135156752445548I, 0.0000929567717058447-0.000115222653946373I, 0.000174538264316339+0.000251623597151971I, 0.000191289550073902-0.000343008206647487I, -0.0000691609946405164+0.000373170476671513I, 0.000050146816782236-0.000437649893183509I, 9.41541280239683*10^-6+0.000345023730132737I, -0.000204209274981742-0.000144804053610998I, 0.0000365492156836771+0.000220937254449815I, -0.000115429425045705-0.000252519043018991I, 0.0000432125620575872+0.000452686906597413I" /* 1275 */+", 0.0000784043089466259-0.000113292259672614I, -0.0000632996727679185+0.00043751364995261I, -0.000223283975343651-0.000388955660395694I, -0.000040179048017907+0.000152170146287887I, -0.000156984634554512-0.000140028281619537I, -0.000127581122512224+0.000289213805028033I, 0.000038159491401266-0.000379088823046192I, 0.000229830607242131+0.000285491942855373I, -0.000111088092464022-0.000145306232013621I, -0.000255547659552173+0.0000988604264546611I, 0.00022654330176646-0.000491730589030159I, 0.000194395802679731+0.000177019529336863I, 0.000154921041161426-0.000568172359531651I, 0.0000327352940547321+0.000312029250456607I, 0.000347725335218178-0.000120403396891929I, -0.000193813460090448+0.0000783980949388163I, 0.000100361452039349-0.000150827597746368I, -0.0000815355002211853+0.000107032581351254I, -0.000271707503327634-0.000257619865318934I, -0.0000341744461636161-0.0000916323915475755I, -0.0000657569483957986-0.000412670704218514I, 0.000161794318686367+0.0000661627728534573I, 0.00024841093441846-0.000273815640977901I, 0.0000102233100426997+0.000300432796465239I, 0.000145288440808692-0.000244921887452102I, -0.000150017257815965+0.00030458976421842I, -0.0000311628072766649-0.00042860354754129I, -0.000115227061594089+0.000331960939043488I, 0.000204488063934732-0.000195972355165994I, -0.000338581301567326+0.000139022906411645I, 0.000133423857185193-0.000216409765881579I, -0.000133158571604247+0.000468016377948707I, -0.0000440590055958598-0.000541401462704214I, 0.000153219981273045+0.000215608703294444I, 0.0000819391189625099-0.000125636111876766I, -0.000113873874669843+0.000333403984433403I, -0.0000191818316129624-0.00030053931508983I, 0.0000415141307681239+0.0000864614364179064I, 0.0000701653365590024-0.000279387050158888I, 0.0000194891780560007+0.000358192510884958I, 0.000063826606717129-0.000364516777328851I, -0.0000564228440141666+0.000145189693193847I, -0.0000908375070864324-0.000222962779952737I, 0.0000879186864912924+0.000259972660038591I, -0.0000449468245576426-0.000271729545302019I, 2.60965001420302*10^-6+0.000145976814931056I, 0.00010336863568233-0.00022782099457433I, 0.000110136991870938+0.000174310350336121I, -0.0000594500177832578-0.000333987511637897I, -0.0000974096795167888+0.000198921190978063I, -0.000155770167457161-0.000327384837009415I, 0.000062735415750591+0.000141346224828463I, 0.0000394083688233614-0.000188155749906858I, 0.000058117844095255+0.000056726245302784I, -0.0000572933310607279-0.00032668526854664I, 0.000113479841343645+0.00039133911671324I, -0.000027220380416209-0.000370766515120752I, -0.000133061317489795+0.000468508758511993I, -0.000164852364679731-0.000132112684320429I, -0.000146744122126279+0.000179153737081438I, 0.000265371190177537-0.000188403865913022I, -0.000152303682184752+0.000201827883236128I, -0.000186164641782103-0.000155670606327552I, -0.000133063540638823+0.000316285825455938I, 0.00020230220470028+0.000119184419867104I, 0.000224160320638865-0.0000817498889516831I, -0.000115968004557631-0.000390941555837119I, -0.000423746352089158+0.00032041511758746I, 0.000251910066248335-0.000255793733346519I, 0.000187769834266451-0.000161493019065505I, -0.0000298883136029733+0.00012944563033185I, 0.000249437723554741-0.0000816888047553399I, 0.000341082675880925+0.0000655768624126504I, -0.000285537002192595+0.000345058154641619I, 0.000209107112253743+0.0000766425023696021I, -0.000176355878879441-0.0000208263083105247I, 0.000193608469081786-0.000453824233260494I, -0.000199627428264146-0.000104844971531169I, -0.0000140315280894698-0.000246738672710269I, -0.000520995165607524-0.000139619270242431I, -0.000312649330941108-0.000135189174969883I, -0.000504042857672975+0.000732646510350701I, 0.0000623057391314571-0.000664275372869655I, 0.000420568774319295+0.00028952281236931I, -0.000383376940642505+0.00019760801406245I, 0.000103517888146611-0.000149351760726834I, 0.000174917120878032-0.000346979924284226I, 0.0000395132527178685-0.000464351491013953I, 0.000349920959468338-0.000508172918391114I, -0.000229694591604698-0.000107188478249883I, -0.000183607628894034-0.000251610304276716I, 0.0000652786201342632+0.000209380760067782I, 0.000306159793319869-0.000648698266209773I, 0.0000692073457286695+0.000210700599286691I, -0.000321406100500146-0.000267180239281978I, -4.7472557444203*10^-6+0.000493828275341709I, -0.0000728695881753062-0.0000559501865550991I, -0.000339818145947956+0.000124491621449443I, -0.0000541636989767528+0.000113177563868776I, 0.000349725121745238+0.000395332038339086I, 0.00023065710254922-0.0000418215220753654I, 0.0000161031385897516+0.000367486037437658I, -0.000127678796279823-0.000128023899358028I, -0.000296093869718091+0.000352747431316145I, 0.0000380917767000226-0.0000105864736192311I, 0.0000826734758254629+0.000145853015204637I, -0.000178608958716185-0.000178445438666579I, 0.000251008043552947+0.000085585976680138I, 0.000191764076128308-0.0000757053277964195I, -0.0000755810722270545+0.000520937676625945I, -0.000163970656148692-0.000515579116162828I, -0.0000792276919107539+0.000248432680226178I, -0.000143690189650365-0.0000966968664127238I, 0.0000808020760352433+0.000294475454583092I, 0.000231395707841028-0.00054030474467773I, -0.000222645909548356+0.000174440907851322I, -0.000139133276400934-0.000262029352847704I, -0.0000593037461280127+0.000655644963511833I, 0.000203336577049956-0.000306140203042392I, -0.000365234268831447+0.000157270872670621I, 0.000253782652240247+0.0000708989837099657I, -0.000325922919652342-0.0000342006065425215I, 0.000194788357934278-0.000429570389035569I, -0.0000303614341959304+0.000461967574784557I, 0.000235681782485858-0.000287947429775762I, -0.000150137243114711+0.000234494607927894I, -0.0000955404504169199-0.00031130766083658I, -0.00026582215439548+0.000149288041218993I, -0.00026668055572701-0.000309908510957003I, -0.000235690175794344+0.00067635973816553I, -0.000269528039072831+0.0000871159362046905I, 0.000390746217780158+0.000683114967503762I, -0.000297149320023867-0.0000550658159668231I, 0.0000446526646777229-0.000103998833581399I, 0.0000706462123860612+0.00019167360294902I, -0.00015651815543612-0.000255817472608468I, 0.000023647128120346-0.000347449143108086I, -0.000691635244749024+0.000193551139075198I, -0.000257265552106571-0.000538460239045327I, -0.000514053402334988+0.000102486781224524I, -0.000100930269600475+0.000239003772947861I, 0.000191506868642411+0.000320867675267923I, -0.0000816086923015379-0.000648003926812617I, -0.000214439624160301+0.000214381811639783I, 0.0000923749942939177-0.0000228415094719493I, 0.00025179491512751+0.000459306816629754I, -0.000183580269837468-0.000102006278333205I, -0.000202125807551546+0.000638104488819944I, 0.000307865070684003-0.000109356566970304I, 0.000266584588677708+0.000133108804753538I, 0.000414137537935472-0.000215704429699596I, -0.0000448658669853896+0.000515805247092772I, -0.000167253191910615-0.000385640553526468I, 0.000103011226263184+0.000580534143217143I, 0.0000580777660739871-0.00011948838250842I, -0.000414156587589721+0.000043798122206254I, 0.000221396881481921-0.000106368141181371I, 0.000181684351693696-0.000136697976601037I, 0.000047028677299578-0.000454834373502292I, 0.000114895585581889-0.0000140834657761624I, -0.0000888146129442612-0.000273919955737952I, 0.000459137044013379+0.00021053419373776I, 0.000540121865884429-0.000693370477888399I, 0.000151416161052076+0.000132043925021398I, -0.000078743971278059-0.000291001723525592I, -0.000308397843519631+0.000449280230971498I, 0.000284851088019157+0.000109488420939111I, 0.000485847876748634-0.000264021936975175I, -0.000674874366646288-0.000275362701793733I, 0.0000304372144348998+0.000293982477979203I, -0.000079406236946765-0.00094746714289422I, 0.000131619360168885+0.000494992030894825I, -0.0000628672172559746-0.000384931589452001I, 0.000282914809779936+0.000111588319345729I, -0.0000602918294331593-0.000430693575491958I, -0.000027801194452936+0.000475847456525592I, -0.000111280297085799-0.000293661496320097I, 0.000034556938371655+0.000617866754485705I, 0.000375259958103535-0.0000281998718520733I, -0.0000100461388060376+0.000111500641334893I, 0.000500210640963685-0.000259478755363762I, -0.0000134844065047564+0.0000726017578031497I, 0.0000391050198052049-0.00049619931596835I, -0.0000136841501004932+0.000245338549767923I, 0.000335890963341079-0.000294535269673445I, -0.000277254025204836+0.000649867055871798I, -0.000117446595261092-0.000265236857122909I, 0.000164620262426769+0.000283829746761821I, 0.0000116710106874061-0.000157241102832688I, 0.000145029292252462+0.00061556488948108I, 0.000202252503075526-0.000470649040676549I, -0.00014408877251208+7.76884515234384*10^-6I, -0.000324918492805026-0.000131028713962238I, 0.000175983374402334+0.000248643210593734I, -0.0000777430862996331-0.000190829936264697I, 0.0000693277497335862-0.000015986195862195I, 0.000294289601034972-0.000165564938719732I, 0.000187967974938135+0.000198941429776318I, 0.0000480902068044259-0.000346189243720651I, -0.0000832694726838679-0.0000388219536627297I, -0.000148808512877464-0.000521068621819384I, 0.000232800082960844+0.0000266568872830244I, 0.000115728029980297+0.0000396637566450424I, 0.00014667835948216+0.000121299953419385I, 0.000133229505717485-0.000342651495703738I, 0.000151239602218897+0.000449062567360525I, -0.0000532674981394559-0.000279409831923159I, -0.000372943591784191+0.00025732298045406I, -0.0000688488608468773-0.000204439655870566I, 5.13091442754763*10^-6+0.00032273933394863I, -0.0000593795674024291+1.9342822177706*10^-6I, -0.000298708901601506+0.000093800575468281I, -0.000448901651547467-0.0000233226082303883I, -0.000133885570214005+0.000175658939066622I, -0.000147801738269412-0.000122385601840482I, 0.000349401682770056+0.000229501172291833I, -0.000102915729448072-0.000486818717166229I, -0.00035184133211918+0.00034312142948432I, 0.000437268036096338-0.0000599758008940289I, 0.0000980746698317933+0.000247532559314901I, -0.0000690089381283593+0.000246745317945004I, -1.8949446882689*10^-6+0.000102745023292049I, -0.0000284314024498709-0.0000379310388505715I, -0.000204782895585742+0.000217099366161117I, -9.37505320328318*10^-6-0.0000426633101449706I, 0.0000580858369735393+0.0000186694974087489I, 0.0000982612821018872-0.000146043220097296I, 5.05592669359685*10^-6-0.000070843283802136I, -0.00025547296565218-0.000322901418305037I, 0.000031993306918186+0.000249938453553877I, -0.000332907050170425-0.000155213604089765I, 0.000014966830910361+0.000297545783796692I, -0.0000133960610508432-7.9044520850238*10^-6I, -2.01486147726642*10^-6+0.00017803171821981I, -0.000176038549721153-0.000108673405568136I, 0.000106363506345661-0.000115552430143571I, 0.000064737102911204-0.000205332123069118I, -0.000207848131484051+0.000244973512369545I, -0.00010490808296299-0.000267151772494199I, -0.000206056920233988+0.0000260194289607241I, -0.0000584510576744965+0.0000617773658855488I, 0.000141343916242268+0.000322735580790942I, 0.0000621886301995872-0.000325933532661234I, -0.00016117302741726+0.000227265098424998I, 0.000141667110390315-0.000174784135514992I, 0.0000799140164468516+0.000325350962287497I, -0.000169868829376599-0.0001933098485509I, -0.0000253271007124117+0.000403259782670786I, 0.000110974883897926-0.000244317904572429I, 0.0000306416303399546+0.000214085983205302I, 0.000137879210673069-0.000199053701181754I, -0.000070451978175011+0.000359320688831657I, -0.0000902616465825548-0.000376731157859482I, 0.000140302729274855+0.000298759176590654I, -0.0000390037837411128-0.000226883689224774I, -0.000255301664990994+0.000172190884700695I, 0.00019028895701784-0.000223798795459784I, 0.0000147777434702284+0.00014035935946549I, -0.000123412768026679-0.000349598381445696I, 6.63335446451265*10^-6-0.0000827477962782888I, 0.000136665349443221-0.00011923047534832I, 0.000248574735694998+0.000158208153089016I, 0.000262524869824842-0.000530439888707829I, -9.79687533524771*10^-6+0.000157367120809148I, -0.0000602474321883547+0.0000593885992899474I, 0.0000515771949107686+0.000231383695698966I, 0.000176807056499933-0.000124092037854283I, 0.000035783854221289-0.000232483626683365I, -0.000276411664233035-0.000141532625547796I, 0.000295533553867748+0.0000985662107090086I, -0.000241374846932431-0.000454914436568883I, 0.000361943990220888+0.000296675333101081I, 0.0000117278996034994+9.48697847677466*10^-6I, -0.00004195828096769+0.0000662783289812835I, -0.0000701327156794946-0.0000633562641259023I, 0.0000185148333818766+0.000137821145253148I, 0.000128854447450528-0.000265491002570249I, 0.0000934925553533844+0.00014681975611555I, 0.000199180736869602-0.00016482398050434I, -0.000128007542862401-0.000257487193207969I, 0.0000157859031689409-0.000362428921894859I, -0.000488235471807034+0.000188840265533126I, 0.000182709901646969-0.000516802594545508I, 0.000191471579171809+0.000415090931020385I, -0.000256934365224028-0.00021138851011556I, 0.000169621968840888+0.000244599176649197I, 0.0000882121232262572-0.000191413739241947I, 0.000344078675844777-0.00014500791634294I, 0.000289779910607894-0.000205337428327358I, 0.000109798439496368+0.0000239172980365253I, -0.0000920316358345903-0.000359725040450235I, -0.0000245181718223843+0.000197888374372293I, 0.00025302017388264-0.000244364675682833I, 0.0000705460787454904+0.000186067652035869I, -0.000146180307595671-0.000294975258046059I, -0.0001134495181378+0.000298850337276913I, -0.000068488673356898-0.000191230921762967I, -0.000131925939622244+0.0000322567363690249I, -0.000128278635881302-0.000064412366235924I, 0.0000766790352851378+0.000383882089040216I, 0.0000222980382263717-0.0000153864380329767I, -0.0000814446938906905+0.000245671366990335I, -0.0000474909151367937-0.000253618789839151I, -0.000173202021424945+0.000184075143438563I, -0.0000791498337883673-0.000113431328453887I, 0.0000415487915069+0.00010921161064976I, -0.0000532673498376008-0.000191183630383177I, 0.000211591298041887+0.000251671654535797I, 0.0000415379262319728-0.000121685318160429I, -0.000054698590069748+0.000216955423385037I, 0.0000887619848274904-0.000360200857166288I, 0.000073863230003335+0.000291098165375889I, 0.000181737042808768-0.000194808104582755I, 0.0000469598023449873+0.000273864233066807I, 0.000198481067392433-0.000185878027901538I, -0.000172553859671933+0.0000399494048497866I, -0.0000392420042520642-0.0000735036429096863I, 0.0000907855282977107+0.000273063259224209I, -0.000132555553509584-0.000388212563062989I, -0.0000509459557253583+0.0000230779927938088I, 0.000161008288122817-0.000312368857490032I, 0.0000145984101292287+0.000124158533680526I, 0.000205117514463575-0.000233464768347661I, 0.000030864350318856+0.000314170232773321I, 0.000160568002661934-0.000207681267038028I, -0.000172635249992413+0.00022358224190588I, 3.04834925679249*10^-6-0.000364481810598549I, -0.000197629996732594+0.000249973647395959I, 0.0000849920553410948-0.0000758671379052311I, -0.00025691057190173+0.000120334081579584I, 0.0000670462658298911-0.0000304199771933751I, -0.0000768808442558281+0.000432318741104765I, 4.4601220883881*10^-6-0.00044846172132731I, 0.000158826747274007+0.000146804990771079I, 0.0000490120749913319+0.0000714502188906353I, -0.000190948311105646+0.000110861365012679I, 0.000019342804211918-0.000306229592492027I, -0.000086750746992477-0.0000121543149065464I, 0.0000581761461180469-0.000344395689442971I, -0.000191601183288784+0.000248739411661823I, 0.0000324830490298391-0.00017566279386105I, 0.0000309479770396829+0.000100836660372661I, -0.0000741144458113075-0.000449444639994322I, 0.0000407643712259607+0.000220976877299292I, -0.000114751498894595-0.000148397721913211I, 0.000125578327442746+0.000253176078102399I, 0.0000437282550920588-0.0000526772549266335I, -0.0000818459239870705+0.000298928484030686I, 0.000114302538992514-0.000110896324695902I, 0.000175417286224709+0.0000994336412945951I, 0.000166823661380795-0.000283391594112331I, 0.000103949043395698+0.000274136835890687I, -0.0000669306795542273-0.000114149443025052I, 0.0000335161940066646+0.000323375710515083I, 0.0000915214261778034-0.000150466108254376I, -0.000094681593736036+0.000220439422023246I, 0.0000375210639487372-0.00017508567367346I, -0.0000351478444469149+0.0000227712499725208I, 0.0000122712038956382-0.000262179774292936I, 0.0000207370364795659+0.000180306773460041I, -0.0000751640487293752-0.000170657048851276I, 0.0000724359312443737+0.000123481605047062I, 0.0000320299839229077-0.000341776110539147I, 0.0000143231666477728+0.000145106750817585I, -0.0000607686353179323-0.000174456442514158I, 1.8772752105523*10^-6+0.000283227358957959I, 0.0000889665574750795-0.000166128497155762I, 0.0000327912069143022+0.0000171936199224601I, -0.00012731198324849-0.000151057620729969I, 0.0000886704853019697+0.000180189039554258I, -0.0000927640016204699-0.00031922558908261I, 0.000131324159901082+0.000237458183466783I, 0.0000175064768981869-0.000121773532897749I, -2.18348174255352*10^-6+0.000145161537788896I, -5.09012941403915*10^-6-0.000141929798837718I, -0.0000202877786347493+0.000174956363777009I, 0.0000565660721847954-0.000256541199568743I, -7.28556889212645*10^-6+0.000183806689284606I, 0.000110576964428167-0.000130503041549394I, -0.000125482126646141-0.0000404841267094941I, 0.000029642311853491-0.000194878394856218I, -0.000252413455794307+0.000262460019971604I, 0.000102047435565941-0.00043945522300446I, 0.000155077198475691+0.000274687025624603I, -0.000108571410700751-0.0000586006037581459I, -0.0000161127847892346+0.000155624914514223I, 0.0000599381703709123-0.000245389618945276I, 0.000129967796824182-0.0000891141372098403I, 0.000185511193268469-0.000275563999515561I, -0.0000734453820605962+0.000125369758310618I, -0.0000190298845234196-0.000249052539964304I, 0.0000205341189671546+0.000122617749512436I, 0.0000956966993613515-0.000410946387965931I, 0.0000647357080945425+0.000208483748198804I, -0.000177845905435635-0.000216200359547434I, 0.0000404085558130202+0.000299570798953783I, -0.0000120988283548468-0.0000427183638698553I, -0.000154353908162392+0.000177921699500525I, 0.0000254331597861108-0.0000109014467457622I, 0.000234630544012502+0.000219906406441815I, 0.000207973029415714-0.000155894390916707I, 0.0000657673533553844+0.000314703398588588I, -0.00012789862854-0.000110962566057405I, -0.0000994806048551627+0.000332610191120433I, 0.0000756067126252771-0.0000428740093135101I, -0.0000784030546140468+0.000163278010758489I, -0.0000221057576180966-0.000148146203317489I, 0.000121317989453707-0.0000574142114049789I, 0.000101463053575083-0.000205112381052505I, -0.00001911281909001+0.000271162341473734I, -0.0001737479993704-0.000329266859931697I, 0.000126278588137404+0.000110638554594375I, 0.0000637275467996462-0.000412591385514967I, 0.0000950167861082493+0.000141118857466465I, 0.0000137240563310243-0.000386014596485127I, -0.000202010876168817+0.000369363957004915I, 0.000151858118436688-0.000091035366227112I, 0.000155112889576883+0.0000302516142204546I, -0.000290244914790524-0.000108308974294168I, -0.000113468671705236+0.000240007784240141I, -0.00010129285110413-0.000486125722837545I, 1.79530775913765*10^-6+0.000253798673488984I, -0.0000352432127105339-0.00033328539353591I, 0.000140920953789889+0.000129270682451407I, -0.0000274961333819896-0.000317213223553621I, 3.69925778617099*10^-6+0.000345997413440363I, -0.0000957654900020683-0.000227720736606014I, 0.0000429818617279439+0.000399944595078202I, 0.000222846134587691-0.000150700720063529I, -0.0000833132029745827+0.000176598042340871I, 0.000250601816685842-0.000253207425014397I, -0.0000230453092606125+0.000193262689639155I, -0.0000474165473225693-0.000403811524596419I, 0.0000480510153682319+0.000177386439315428I, 0.000158920448798017-0.000191767242859273I, -0.00010929995305382+0.000419036582007772I, -0.0000528941691237831-0.00022946445151934I, 0.000074744321452466+0.0000708096403272983I, 0.0000764573140906174-0.000230952857520714I, 0.0000742329446929902+0.000351772986466055I, 0.0000470368851291999-0.000364073420671678I, -0.000102129320267893+0.0000621988039961325I, -0.000136332916697666-0.000234670614206917I, 0.000132729334486281+0.00017415487902586I, -0.000150261147098999-0.000145692559501759I, 0.0000415425150991501+0.000115917886950585I, 0.000177572308829642-0.000107246033915108I, -0.0000121291594499315+0.000159426682314872I, 0.0000479475113243956-0.000134174678667544I, 0.0000862128789330588+0.0000499596916455888I, 0.000028893660861781-0.000326203418535657I, 0.000161286044541599+0.000123715258947325I, 0.000022271053431198+0.0000234735156454925I, -2.14506538572834*10^-6+0.00023868440036958I, 0.000143142037241701-0.000164301172834498I, 0.000066658222654574+0.000274073825597234I, -0.000065767537704825-0.000164403369796237I, -0.000136745691379142+0.0000326553331545905I, 0.0000684806719184292-0.000193336621826715I, 0.0000267019072724124+0.000349414980426368I, -0.000216999988469874-0.00015115076787903I, -0.00010937978292791+0.0000696799355548912I, -0.000245027020033625-0.000198187230638681I, 2.39613393128029*10^-6+0.000114180218561944I, -0.00012358176630771-0.000354516650557349I, 0.0000453343388392154+0.000389083391307997I, 0.0000655101894509923-0.000282473185466932I, -0.0000607334369753515+0.000152797989982309I, 0.0000387692969853794+0.0000234338487932071I, -0.0000766775332178299+0.000332087916320701I, -0.000168164588596398-0.000159373917461389I, -0.0000513174019536789+0.000208800786308087I, -0.000140934069872535-0.000249693331828237I, 0.000025923161269585+0.0000896626798192448I, -0.000103813619146007-0.000239169201405952I, 0.000117966454095085+0.000243967616907353I, -0.0000715836233553926-0.000100072910341716I, 0.000153362107080043+0.000240793894446944I, 0.0000611295792311355-0.000297745563189118I, 0.0000543371861724066+0.000281365435193187I, 0.0000610728506628792-0.000350795045737618I, 0.0000561956586990447+0.000121014647499513I, -0.000122795552719025-0.000162429249106444I, -0.0000442949825673503+0.000160389630466608I, 0.0000710419801053327-0.00033642447342722I, 0.0000970374524514841+0.000364868267283786I, -0.0000547422669452144-0.000144107727190925I, 6.5094539319444*10^-6+0.000173630705978304I, -6.24026424748694*10^-6-0.000185794877447735I, 0.000176608364351419+0.000221490047264027I, -0.0000662998476293154-0.000305655824427932I, -0.00012297220101553+0.000194740541519957I, -0.0000705101177113273-0.0000440255723615451I, 0.0000537690670565844+0.000101374042990041I, -0.0000212487773796082-0.000103993415692491I, -0.0000648037260731581+0.0000968346827723333I, 0.000123371698605159-0.000291527233707339I, 0.0000127029978271846+0.000120074907310481I, -4.24519648102123*10^-6-0.000211618297873193I, -0.000105539500131776+0.000106239006299091I, -0.000152286679442464-0.000223341129207651I, 7.32568011995803*10^-8+9.86885373446776*10^-6I, 0.000178115097675728-0.000122869046145802I, 1.54015623660486*10^-6+0.000139495808590719I, 0.0000308271968043534-0.000283372406725671I, 0.000158176970124436+0.000159154781064433I, -0.0000498741934534173-0.000131121507951389I, -0.000133798335315792+0.000314016640523663I, 0.0000239111894019742-0.000181599184002433I, 0.000123151593972527+0.000248081113699151I, 3.20296601270631*10^-6-0.0000257326866714507I, -0.000176727978318078+0.00029121431277724I, -0.0000738314943811437+0.000147537817579389I, -0.0000598071063472972+0.000202601892812715I, -0.000109117218882049-0.000117791731254625I, 0.000160496638657494+0.000084846530668879I, -0.000168398179729327-0.000231855765979531I, -0.0000137074599696238+0.000384352293082161I, 0.00025319969518465-0.000320186428468789I, 0.0000749748360830443+0.000233423661176571I, 0.00018282281083793+0.0000167332970876189I, -0.0000633978536868512+0.000201817732536007I, 7.18664274446114*10^-6-0.0000816869223995829I, -0.0000687352864261756+0.000242527783108017I, 0.0000311304039367175-0.000102677020250857I, -0.0000301610355087154+0.0000677466280086349I, 0.0000814464248896721-0.000167605370802158I, -0.0000863729238426821+0.0000557545555090185I, -0.000142749557617828-0.000114335165724075I, 0.0000619714248264035+0.000140436611018136I, -0.000111734756216208-0.0000488239552594908I, -4.03582415308949*10^-6+0.000181258375570383I, 0.0000843432976768852-0.000121484949082255I, 6.60226524722015*10^-6+0.000186053940459766I, -0.0000622710563972844-0.0000681495833719152I, -0.0000467230399508288+0.0000157906472703I, 0.0000361654590209753-0.000202097786822214I, -0.0000525780965752526+0.000172761276154771I, -0.0000116235379582063-0.000210411471647334I, -0.00013305289296311+0.000139340288032918I, 0.0000105266008270537-0.0000855613014218249I, 0.0000632468260417003+0.000171606846660766I, -2.28132533826295*10^-6-0.000295793494752184I, -0.000029189964004741+0.000193988973605018I, -0.0000110745921237077-0.00014737309497372I, 0.0000654666243255419+0.000226637250018487I, -0.0000146934150630166-0.00011955973771306I, -0.0000481029062485005+0.000260628255171931I, 0.0000827536088037973-0.000137882828745095I, 0.0000819959761830965+0.000121713625434623I, 0.0000961176596387232-0.0002181203829724I, 0.0000318712534860143+0.000217272310198003I, -6.52997529380851*10^-6-0.00015311671676455I, 0.0000405546862238246+0.000274194921823285I, 0.0000656500415151167-0.000177560188533019I, -0.0000647247929776681+0.000153715783842253I, 0.0000432516141459278-0.000124278744893989I, -0.000057496573580136+0.0000747800559452842I, 0.0000238107431662866-0.000264508444772303I, 0.000098946785616609+0.000154510318757889I, -0.0000786193510080718-0.0000648052497985479I, 0.0000240526581366833+0.000178053243507182I, 0.0000475781265724491-0.000161701619263567I, -2.44824009954714*10^-6+0.000124060612853187I, -0.00018111253122878-0.000156153815318675I, 0.0000768077542452943+0.000266505373513866I, 0.0000138608094291637-0.000156261037550633I, 0.000103884894752171+0.00008116363546157I, -1.84117541266015*10^-6-0.000196743724972415I, 0.000130194867045215+0.00030507059160538I, 2.66480595261353*10^-6-0.000235486311178402I, 0.0000491370316010514+0.000332323235173056I, -0.000112185212848289-0.0000743603657424097I, 4.200120074279*10^-6+0.000108763564144246I, -0.0000890747699307907-0.000125940155528138I, 0.0000539687159703735+0.000117501120908931I, 0.0000881843794561815-0.000116775687981833I, 0.0000510699368179178+0.000128724977780399I, -0.0000219740270093283-0.0000883102407007743I, 0.000170598468392965-0.0000201235465085892I, -4.91569156672053*10^-6-0.000179364430814214I, -0.000148916200873375-0.0000184335586638743I, 0.000210177391460681-0.000163861491545949I, -0.0000167353715317225+0.000317392217006093I, -0.000129301322381329-0.000219474858647208I, 0.000034191082899579+0.000156244795153477I, 0.0000508739369432336-0.000145846735007395I, 0.000221657430521821+0.000174166520662054I, 0.000102485190775828-0.0000831830208965508I, 0.000020122570642682+0.000112310114344404I, 0.0000312055353796841-0.000197610367168086I, 0.000043793479228291+0.000158773510848539I, 0.000131195953252393-0.000190797453705201I, 0.0000101834127502031+0.000211058587572246I, -0.00003968895115861-0.000258159294067705I, -0.0000305921621757633+0.00022058326463058I, -0.0000658383422871747-0.000124302453342964I, -0.000040899769698887+0.000112119499682581I, -0.0000499687458966147-0.000104161805825175I, 0.0000681441502001984+0.000261654759100251I, 0.0000675875771696104-0.0000906428696549924I, -0.0000220549974183829+0.000240191370789442I, -0.0000678698622380999-0.000191467324536482I, -0.0000699742483245555+0.000202167297366808I, -0.000011966888167404-0.0000998005794274077I, -0.0000383591522776749+0.000122096193622072I, -1.18720779166701*10^-6-0.000143950052661858I, 0.000111191975642431+0.000101010687600715I, 0.0000479039074418255-0.000179615176832248I, 8.38097927277979*10^-6+0.000170987451615959I, -0.0000381072531502168-0.000232499570964523I, 0.0000979530936441505+0.000196008794108172I, 0.0001570018068799-0.00023429132246515I, 0.0000559004149092388+0.000162116059363073I, -0.000014666744500465-0.000229448590632846I, -0.000120700439737732+0.000214942490933579I, 0.0000468491412801088-0.0000427589855625902I, 0.000207107618160926+0.0000882042270518453I, -0.000200009165969761-0.000260820128434587I, 4.00303960792727*10^-6+0.00022872282500477I, 0.0000610072902257168-0.000433864899192523I, 8.54511260590615*10^-6+0.000314356312715648I, -0.0000452615260524121-0.000215086615137526I, 0.000121136655183428+0.00014181856316503I, -0.0000366030295418258-0.000246131841795719I, -0.0000151704658684519+0.000248633594561496I, -3.60833305078835*10^-6-0.000180312199506851I, -1.16282096504451*10^-6+0.000323262984070724I, 0.000136139441929831-3.24346263416469*10^-6I, 0.000105394106422237+0.0000272429059510675I, 0.000243908845763778-0.000151158431958822I, -0.0000740962252160465-0.000027349641658435I, 0.000182483077687462-0.000292591253632537I, -0.000027907620646246+0.000279334463006271I, 0.0000916969551428441-0.000236984723980158I, -0.000146845329345748+0.000382609946176655I, -0.0000247707295692671-0.000173200695597258I, 0.000288426416581854+0.000223907055971638I, 0.000102725458067316-0.0000375587243476184I, 0.00011177538284214+0.00036033445610675I, 0.000165607176031706-0.000367129086097468I, -0.0000516568724856917+0.0000188243876850236I, -0.0000651850860570235-0.00010661135749553I, 0.000136652658800146+0.00023122560292192I, -0.000105039816003112-0.000265357797456163I, -9.94261450802567*10^-6+0.0000500361831175399I, 0.000140143779651986-0.0000881424157308694I, 0.0000971063508661832+0.0000380940110044232I, -0.0000725246973010963-0.000181894653337316I, -9.29177717290012*10^-7+0.000119881451360932I, -0.000074733426891401-0.000273593191112883I, 0.000145200184305456+0.0000830438550890773I, 0.0000101839031209962+0.0000262677773097626I, -0.0000221215664372668+0.0000793650724686124I, 0.0000719810349267378-0.000163026816576157I, 0.000145214835459655+0.000310750688181106I, -0.000108653404073556-0.000207243352263705I, -0.000137801176701129+0.000156943292442708I, 0.000023519924609688-0.0000927489453101516I, -0.0000401636393783405+0.000344095579057864I, -0.000115091362103879-0.000134797698199575I, -0.000210523396720508+0.0000855817046941459I, -0.000323583560877849-0.0000179568303216552I, -0.0000451028585875949+0.000148009447860069I, -0.0000697945807934255-0.000252711070556832I, 0.000130494273301643+0.000219469728384671I, -0.0000738484369398214-0.000319273482963646I, -0.000155461391764904+0.000325548566934101I, 0.000263907442714454-0.0000788394988157333I, -0.0000378489508036349+0.000268906942338463I, 5.95495694173922*10^-6+0.0000787715281506004I, -0.000101209974352342+0.000146596293447965I, -0.0000890380656689304-0.000130885423331484I, -0.0000623796095204554+0.000159829126723947I, -0.0000506062467729166-0.000126746750169519I, 0.0000700209534045688+0.0000936329386057559I, 0.0000365499569342528-0.0000700438771028016I, 0.0000424094833277309+0.0000602799206036924I, -0.000147365614327795-0.000187988088143459I, 0.000177103977647927+0.000197481181314777I, -0.000108728259007066-0.000155083447728412I, 0.000037059731373198+0.0000298812510475952I, 0.0000956862967934382+7.26511650207531*10^-6I, -0.0000842184924512882+0.000218842480202018I, -0.0000912633349660387-0.000237232629516033I, 0.0000579090584073824+0.0000979568598198331I, 0.0000300103443729735-0.000117204302462955I, 0.0000992831810092166+0.000286934036108212I, 6.60681695187767*10^-6-0.0000532863607747808I, 0.0000209798850645102+0.000109081993802649I, 0.0000359489761786698-0.000140751301392127I, 0.0000500851670351792+0.000187166478991651I, 0.000107033809804131-0.0000992491905285819I, -0.0000227170530238528+0.000202955952127664I, 0.0000300234376463382-0.000260413119579523I, -0.0000506813964908802+0.000172489380024278I, -0.0000704559541078253-0.000161633298990781I, 0.0000281246051171364+0.0000843595027514713I, -0.0000941337571215428-0.000161695078782588I, -0.000021221552397819+0.000270571989021526I, -0.0000255950423023666-0.0000786917623172831I, -0.0000401211575929111+0.000186676138661823I, -0.0000417091435407658-0.00019370724484098I, -0.0000672423297226879+0.0000978295799139134I, -0.0000568292067079338-0.00012205682064184I, 0.0000350739469108097+0.000152778066589605I, -0.0000429079333705584-0.000179812867025052I, 0.0000926346738232552+0.000210930803167876I, 0.0000192851656657767-0.0000944608405966647I, -0.000048213255420919+0.000199053594423768I, 0.0000262490620691274-0.00024525488613221I, -0.0000135080589354729+0.000203220650568407I, 0.000025529264036217-0.0000843200391134647I, 0.0000191720562153466+0.000201619114016818I, 0.0000757336688089838-0.000211174189403757I, -0.000073876802020799+0.000088260790458821I, -0.0000749547911370644-0.000118207077311625I, 0.0000645550496745292+0.000308623290681832I, 0.000033404993262802-0.000296898366329261I, -0.0000752734225838157+0.00016430005411791I, 0.000193389217645582-0.00015580368238187I, -0.000117986913610088+0.000173659547270827I, 0.0000276277028408309-0.000219531667077308I, 0.000046607463063233+0.000244570099529494I, 0.0000511641179864083-0.000213368911184727I, -0.0000609203048259801+0.000175641141499369I, -2.18624870566099*10^-6-0.000161805779830109I, -0.000111760186342343+0.00020452497626055I, -0.000086686357104231-0.0000216944508065499I, 0.000137556790776634+0.000222361983288811I, 0.0000635805483140462+0.0000239519356840483I, 0.000147249057762885+0.0000390567245117355I, 0.000141728941628744-0.0000373681813069598I, -0.000110972076741613+0.000165835650173102I, 0.0000991144212821392-0.000134633976067889I, -0.000211482383238385+0.00013169760514294I, -0.0000124725766224428-0.000169872211592535I, 0.000051587604238801+0.000444392216171841I, -0.0000887334930862352-7.46395916721945*10^-6I, -0.0000901914793244013+0.000329842943822235I, 0.00023134676663099-0.000098514722435661I, 0.0000934975232085456+0.0000722031612456679I, -0.000084861080438055-0.000130194572221923I, -0.0000117173188270346+0.000294012859765712I, 0.0000861783561806145-0.000290634795302553I, 0.0000740338695521277+0.0000755641813917268I, -0.0000156876323769112-0.000049970429333747I, 0.000194798704444362+0.000179318405006993I, -0.0000389157201949528-0.000284442662394588I, -0.0000397865827500333+0.000148897966113809I, -9.59650506147761*10^-6-0.000259113012928411I, 0.0000947452418558267+0.000219044650472942I, -0.00013537329668363-0.000125459811772638I, 0.0000988545092349867+0.0000349009074589357I, -0.000032197276223326-0.000142402797674305I, -0.0000601012007230921+0.000350723127103777I, 0.0000218882531654977-0.000305998196059923I, -0.0000570318388272772+0.000153872731025713I, -0.00013108871429188-0.0000932833105258782I, -0.000206112862372664+0.0000953992617466508I, 0.0000891844111710204-0.000221826307546984I, 3.46881445739973*10^-6-0.0000245981406520041I, -0.000174199716993943-0.00038349989354648I, -0.0000400139466445255+0.000150425584142079I, 0.00013997136374266-0.0000616394330714205I, 0.0000417780498093246+0.00013484427729103I, 0.0000842478273199318-0.000252548430817952I, -0.000258470174389071+0.0000423753900183749I, -0.0000107915145220958+0.0000194655792575734I, 9.76255207581826*10^-6-0.00001956490635879I, -0.000186538846836175-0.0000777317064317731I, 0.000125682068862853-5.98387750493569*10^-6I, 0.0000975908913429303-0.000109387644645923I, -0.0000775597605658748+0.000140631033905882I, 0.0000433095140647579-0.0000916771408458315I, -0.0000230144036905713+0.000130514992234307I, 6.5320270726133*10^-6-0.000221320590216733I, -4.04381442467625*10^-6+0.000106545736060919I, 0.0000652415654712289-0.000207535629435747I, -0.000198830969864203+0.0000885403997878937I, -0.000032297876682978-0.000183651211035481I, -0.0001219531834184+0.000312094212894283I, -0.000035795674653592-0.000303856612503366I, 0.000113083337013855+0.000149414778303866I, -0.0000425828199045177-0.0000678244129755495I, 0.0000108533632543129+0.000129590001990076I, 0.0000207551442232095-0.000186659998693534I, 0.0000164493736821823-0.0000119320095703005I, 0.0000804548852274059-0.000212025350020957I, -0.0000279967019150403+0.000134901138391469I, -4.40883600414743*10^-6-0.000198577884833923I, -6.28396889681735*10^-6+0.000113111821864584I, 0.0000151860091321799-0.000234343495378362I, 0.0000460452726496175+0.0001634549501628I, -0.0000846567109221331-0.000174756728602899I, 0.0000160393157856146+0.000168983839573969I, 0.0000308896046785493-0.0000847743830218534I, -0.0000265974405283515+0.000122164804804422I, -0.000014283670379004-0.000105107562783438I, 0.0000722110429321632+0.000151758328639906I, 0.0000371568421748873-0.000175517872891753I, 0.0000656467041130242+0.000169750900036596I, -0.0000460717992872902-0.0000689035652981931I, -0.0000373851569149228+0.00015018138418242I, 0.0000335136320018457-0.000104791333151848I, 0.0000317950646997057+0.000212589505534057I, -0.0000543467338604417-0.000178865340162369I, -4.57290182786061*10^-6+0.0000976686467751653I, 0.0000222737706553606-0.0000998947496712613I, -0.0000637356587633141+0.000246819829969994I, -0.0000638052430790033-0.000218374345055307I, -0.0000587667980675079+0.0000853691668336968I, -0.000147978126436697-0.000163261580915938I, 3.96706104522557*10^-6+0.000151216983168715I, 0.0000525173260063887-0.000240910913566087I, -0.0000206214914535908+0.000178475216281724I, -8.8584582584785*10^-6-0.000208111775762465I, -0.0000850982385278052+0.000234622975612175I, 0.0000668194411323503-0.0000997474827997425I, -0.000101377234038764+0.000142884234636348I, 0.0000103842339362407-0.000062506164735609I, -0.0000781819527318705+0.0000776768309511181I, 0.0000249537293576882-0.000212326522749788I, -3.92476995865145*10^-6+0.000191253278063713I, 0.000039948021915354-0.000181928413509372I, -0.0000194833983260056+0.00016895509030089I, -0.0000365837553861753-0.000159440116048815I, -0.0000481439276076236+0.00016176694933135I, -0.0000484898014791157-0.000151258739665805I, -0.0000143726691393086+0.000271229820798705I, -3.97473253025067*10^-7-0.0000665375764644979I, 0.000126725342299272+0.000197029874776893I, -0.0000305198115546995-0.0000979811787352103I, -0.0000323904085362581+0.0000856767491112181I, 0.0000904059021362342-0.0000869879058624745I, -0.00011965348646818+0.000111319839625337I, -0.0000203806695377239-0.000180464155108776I, -0.0000947723533963239+0.000262271476551607I, -0.0000854761114860343-0.000139668102351471I, -0.0000983299337692451+0.000259307456698656I, 0.000112292110477788-0.0000889368839180751I, 0.0000437874001368294+0.0000967247862010949I, -0.000121748015220494-0.000190897490483099I, -7.83249240950497*10^-6+0.000213615635299492I, 0.0000403831151302943-0.00017012661315883I, 0.0000945172794734319+0.0000981868905905193I, 0.000035308032972933-0.0000626565363435906I, 0.000109315861573651+0.000228806500365706I, 0.0000464418217226563-0.000231219865720071I, 4.49111787941231*10^-6+0.0000592190648028155I, 0.0000252403303206835-0.000296303804972858I, 0.00011648577813355+0.000179670457880779I, -0.0000734117600214005-0.000072864731930701I, 0.000119118052970164+0.000113828187195561I, 0.0000409181572409037-0.000161172880586445I, -0.0000566028010704922+0.000317691357412162I, 0.0000292862600037261-0.000245698675137026I, -0.000139105984437276+0.0000878163522101915I, -0.0000976463077125407-0.000144139059206692I, -0.000114685137215974+0.000120454324618339I, 2.02965975057129*10^-6-0.000120833076928639I, -0.0000391437053376501-0.0000357012039350677I, -0.000234755015708641-0.000324262327379999I, -0.0000542759496359263+0.000105735536612047I, 0.000012307061085817-0.0000745383021656366I, 0.000127190357736498+0.000207781086944402I, 0.0000822681323067156-0.000297130695273229I, -0.000273544446279475+0.0000222574850573339I, 0.0000733506326774973+0.00008521272187278I, 0.0000513556502439244+0.0000630007635159983I, -0.000235004400851359+3.56923304418332*10^-6I, 0.000124511454799628+0.0000226547370128361I, 0.0000255137606355433-0.0000554545380736152I, -0.000113551284187599+0.0000946017792825899I, -6.9507787060856*10^-6-0.0000466485978065735I, 0.0000339722854268605+0.0000790582879972074I, 0.0000346228969548457-0.00016466777812998I, 0.0000480463180529811+0.0000290196064552477I, 5.55309524920192*10^-6-0.000245218955828748I, -0.000146300958980537+0.0000725560608059054I, -0.00012214957398955-0.000204873387624906I, -0.000143305619623864+0.000285370773997098I, -0.0000277963403032362-0.000238412237358223I, 0.000108613475233319+0.000157465078527367I, -0.000119769563696352-0.0000611402405560823I, 0.0000782749896692634+0.0000471208683651518I, 0.0000466224566351868-0.000169658857757585I, -4.08582195658014*10^-6-0.0000358816387946989I, 0.0000815474180018675-0.000219954964363725I, -0.0000696616772320066+0.0000509715435888216I, -0.0000329919184397433-0.000130570045433209I, 0.0000326347080957787+0.000137040940112474I, 0.0000627498196832073-0.000283791186366973I, 0.00001104334737377+0.000162567131291871I, -0.0000781692251382091-0.000170471785068215I, 0.0000393641346404141+0.000214808136719673I, -0.000020436276323881-0.0000376338219679095I, -0.0000614909380250497+0.000160075601973785I, 0.0000177422580508856-0.0000678631692152755I, 0.000139800278589677+0.000153600659468648I, 0.000134669257626411-0.000172301833842217I, 0.0000829152344526206+0.000242598414308218I, -0.000127601655215586-0.0000587388917139865I, -0.0000179810551179175+0.000201099022322306I, 0.0000639829749047555-0.0000476308764371299I, -0.0000649282626558689+0.000234691659355973I, -0.0000198288306768432-0.000190614615983794I, 9.31337048058165*10^-7-0.0000482286450234883I, 0.0000107387930722393-0.00012244407204829I, -0.000117162956140405+0.000223715327569178I, -0.000145169896835373-0.000249382371321374I, 0.0000124629471278038-0.0000649360481633286I, -0.000243596336939219-0.000412779995896971I, 0.0000214789103825454+0.0000793534722994757I, 0.0000314686364405446-0.00029712255025206I, -0.0000304747548482679+0.000352322726580945I, 0.000167851177900126-0.000276769649299748I, -0.000209249534928775+8.39732673976029*10^-7I, -0.0000327545236070447+0.000224941697346845I, -0.000127900007070811+0.000153081162274077I, -0.000353240137374342-0.0000740056026412477I, 0.0000239463285535211+0.0000167088909021977I, -0.0000695542783578075-0.000221261884401332I, -0.0000324125116260695+0.0000339789110204676I, -0.0000738994461216401-0.0001657824284115I, 0.000125221712657293+0.000203008376272739I, -0.000100530631704657-0.0000921765416979459I, 0.000187664207138265+0.000140993412427616I, 0.0000662439638092439-0.000412148932307495I, -0.000133383119451692+0.000304953115707972I, -0.0000775627783325712-0.000389294393531938I, 0.0000298959945679289+0.000285067541489307I, -0.000281719837699804-0.000142488218814699I, 0.0000374056014632141+0.0000521522222506443I, -0.0000197103122411656-0.000235911769289066I, 0.000233331796614995+0.000204022995146311I, -0.0000249548908390284-0.000111611237211914I, -0.000155363089616283-0.0000181592023977485I, -9.27879378707274*10^-6-0.000269946010976631I, 0.000107751181452496+0.0000137677137616778I, -0.000206342919205964-0.000163872449547819I, -0.0000927388749229757+0.000248386675528143I, 0.0000267408918616416-0.000106513414811819I, -0.0000168960288955984+0.0000192147564563421I, -0.0000140150999923073-0.0000183231738775026I, -0.0000592501318801836+0.000181704192590271I, 0.0000131075149494059-0.000255642290911081I, -0.000113048551523481+0.000148073730736933I, 0.0000374438898169832-0.000088246938867408I, -0.0000108216848290099+0.000133151597274884I, -0.0000125422500030741-0.0000875093792520496I, -0.0000709853881982066+0.0000899189605485424I, 0.0000970044296396669-0.000173273303978932I, -0.0000405904098189006+0.000216914150303946I, 0.0000211616640036215-0.000163213722851511I, 0.0000273564761037468+0.0000278535567312987I, -2.86476054215207*10^-6-0.0000550153533042663I, 0.0000190678876146552+0.000136773189407019I, 0.0000723738837070403-0.000189857919466807I, 0.000121589877901713+0.00017225051262776I, -0.0000705295964077906-0.0000991618229109964I, 4.41480787842784*10^-6+0.000224803686023823I, 0.0000782647758972043-0.0000499576483251493I, 0.000026108161383011+0.000129879463056165I, -0.000101800681557435-0.000223036538021824I, -0.0000244635824662652+0.000204673492397163I, -0.0000114806039827781-0.0000997185525339332I, 0.000142444032463857+0.000163904624427129I, -0.0000183289177432049-0.000176522414405734I, -0.0000214741939155388+0.000281427540344948I, 0.0000406154702500231-0.000213004629533381I, -0.0000761292370492207+0.000255732800299244I, -0.000116106331921786-0.000195433589441618I, 0.0000660737625027478+0.000108186741262856I, -0.000076992354368704-0.000208815807772447I, 0.0000648596520371214+0.000187513928859225I, -0.000030942118418939-0.0000572564057389203I, 0.0000744353130579769+0.000207470471617577I, -0.0000188414147821453-0.000138360374084288I, 0.0001997773977591+0.000213849190449591I, 0.0000652433284411731-0.000192700293354502I, 0.000111158053202979-0.0000590048197505438I, 0.0000293684271156344+0.0000147309326892547I, -0.000139380248328269+0.000168901559537596I, 0.0000571462425240333-0.000323183490151511I, 0.0000228594991957827+0.000259460650546657I, -0.0000371222744919561-0.0000772818596801585I, 0.0000813394062077905+0.000372465341763704I, -0.000070845226564262+6.26543837777726*10^-6I, 0.000129262038359691+0.000204610950392776I, 0.0000593185311843562-0.000165265645543771I, -0.0000157050042892489+0.00016588426370331I, -6.07156935237278*10^-6+0.0000687817817699029I, -0.0000204911921356971+0.000160832270871606I, 0.000117370123381504-0.000209422664169149I, -0.0000687435773061548+0.0000522187843567426I, -0.000027201474442733-0.000225514193756062I, 0.000142647389768212+0.0000722755816258551I, -0.000101729754853287-0.000269551232484053I, -0.0001735980701309+0.000207021461409595I, -0.000153216218240175-0.000103826875310037I, -0.0000639923731502229+0.0000959358708014867I, 4.25164039196333*10^-6-0.000230571018664962I, 0.0000252397415214316-0.000026484764770928I, -0.000123778339405792-0.000182625059215477I, 0.0000311479613037242+0.000184368782122007I, 8.50119557598954*10^-6-0.000234397689989017I, 0.0000406142738449936+0.000326700301991712I, -0.000110679952683165-0.0000541593116086637I, -0.000136786054877492+0.0000401707075260752I, 0.000217464893793378-0.000219845820683098I, 0.0000204575223686387+0.000138855249614217I, 0.0000297829486099481-0.000211504098404436I, -0.0000267485438618485+0.000207816153268862I, 0.000243220452586357+8.13435590933245*10^-6I, -0.0000446794047938477-0.0000535952526859361I, 5.80814192240344*10^-6-0.000130527809734095I, -0.000151333672631189+0.000135444100659989I, -0.0000748352233051337-0.000141920008281079I, -0.0000311871835572747-0.000138380131856691I, -0.0000444208745676117-0.000131077816097562I, 0.0000950782800464766-0.0000708609026738886I, 0.000225059791096069-0.000164275482314707I, -0.0000373729986734696+0.000207385699436824I, 0.000147296055111127-0.00012824361387689I, -0.000122684059256283+0.000175037706192467I, -0.0000620968113031921-0.00029750120963419I, -0.0000997273158201098+0.00017924909244166I, 0.000151707130818033-0.000202045750637688I, -0.000399335419517563+0.000185899885851288I, 0.0000497401093291848-0.000144930056422997I, -0.0000234672853036762+0.000506973437335016I, -0.000260420474398303-0.000406936081031357I, 0.000178806692181642-0.0000188958545145034I, 0.000134287012315629+0.0000341650847383084I, -0.0000530636147555818+0.000185806944799035I, -0.0000506256247721547-0.000222821089647015I, -0.0002514134791858-0.000149154919152443I, -5.77297978510708*10^-6-0.000381546888981686I, -0.000029870818393145+0.000199680796258405I, -0.000114879795726309-0.000205265738294278I, -0.000125820184690279+0.000102629377741632I, -0.00020159454410963-0.000235068423795275I, 0.0000683477026127438+0.0000240956079866464I, -0.0000992120764039938+0.0000790381566523976I, 0.0000589637308961299+0.0000986164691961061I, 0.000198769132433512-0.000184544134525391I, -0.0000774774652682762+0.000223265842582993I, 0.000153199349995394-0.000117578389950724I, 3.74073660558811*10^-6-0.0000884041912326194I, -0.0000512231170483518-0.000288788035256476I, 0.0000719835705561252-0.00002881464365134I, 0.000215336487752377-0.0000182886837188978I, 0.0000680207439932027+0.000226150391832454I, 0.00012727742316111-0.000269357277452536I, 0.000110785468119841+0.000103235916307087I, -9.38476894919767*10^-6-0.0000321840537871329I, -0.000233718432621851+0.000178433921061348I, 0.0000449642448502165-0.000230927997619891I, 0.000231999716815779+0.000195092436108356I, -0.0000655632179129187+0.000132143399550784I, -0.000174693504731745+0.000256799041411488I, -0.0000440600122002472+0.000218597070272793I, -0.0000687545100655025+0.0001220597070393I, -0.00024706962560721-0.0000523136131778998I, 0.000228383015585699+0.000101968706069689I, -0.000168365305770045-0.000175059843436046I, 0.0000615051157222517+0.000288546132501493I, 0.000250209178505509-0.00031738069063571I, 0.000162688740656207+0.000268858858908312I, 0.000212910310251752+6.54559798434093*10^-6I, -0.0000363841699783088+0.000259764076910872I, -0.0000390852208424497+8.7515567010304*10^-6I, -0.0000550003618705462+0.000180537905891044I, 1.82497348621969*10^-6-0.0000488920920606227I, -0.0000257710774339007+7.2553755239996*10^-6I, 0.000122716078888392-0.000118282420997389I, -0.0000957936393608236+0.0000222137897718253I, -0.000136309380416372+0.0000172975665840116I, 0.000142576993173787+1.91916348045519*10^-6I, -0.0000521523307372167+0.0000398793917381122I, -0.0000519305487830239+0.0000417640903288769I, 0.000211225728482248-0.000107305924301929I, -9.15688240322701*10^-6+0.000207304671342084I, -0.0000438732063180894-0.0000219495861038409I, -0.000139227104916314+9.08410672747972*10^-6I, 0.0000376471778587546-0.00017356959845494I, 0.0000673658161934508+0.000190171036679589I, 0.0000171761031934622-0.000100456633206824I, -0.000142395604613401+0.000199184401420687I, 0.000133659053148649-0.0000951958402941978I, 0.0000780135473620538+0.0000452752747609132I, -0.0000581150247465055-0.000263847297842046I, 0.0000235638515889518+0.000219868343886635I, -0.0000453385352923687-0.000175348964485513I, 0.000105159534514582+0.000124271107648742I, 0.0000530421336153287+0.0000114792084582678I, 0.0000409265130757475+0.000195426352989314I, 0.0000660976768498893-0.000134275200765017I, 0.0000995453285268334+0.0000275884852242801I, 0.000087365123627933-0.000297492070552979I, 0.000164512012969826+0.000166276980330926I, -0.000056312900942246+0.0000284698615020234I, 0.0000814434240384513+0.0001756427648946I, 0.000127617634330209-0.000126787622518738I, -0.0000226732564317621+0.000284688569699826I, -8.30438924417752*10^-6-0.000164207681345954I, -0.000195449459727956-0.0000233932899300353I, -8.64321968326009*10^-6-0.00016941806573467I, -6.71499274858403*10^-6+0.000220237946644035I, -0.000170287681187561-0.0000236665471679607I, -0.000123289666553293-0.0000378752341091594I, -0.000328201941654083-0.000165876656989741I, -0.0000510473954176164+0.0000446748671417081I, -0.000192764197911435-0.000187013432073863I, 0.000203572039556831+0.000330901429371779I, 0.0000387592651554775-0.000330463314208623I, -0.000196204748740505+0.0000824078590134619I, 0.000220398907936235+0.000119309901799748I, 0.000053285969139742+0.000293183228598469I, -0.000197553404241999+0.0000946184400131912I, 3.91343545933431*10^-6+0.00015417055410439I, -0.000179030245742988-0.000036744368477545I, -0.0000972518825003121+0.0000247052023549668I, -0.000139553180937304-0.0000520499478237712I, 0.000162528398743347+0.0000393057829892754I, 0.0000509711948240113+0.0000128007806394915I, 0.000166510716138095-0.0000200190709648003I, -0.000134069661963503-0.000254353217143457I, 0.000191486588910721+0.000130924120396215I, -0.000172538914608815-0.000250137603660067I, -0.0000252871444225911-0.0000178493618205611I, 0.0000485051060415594+0.0000655282483648466I, -0.0000803127417839328+0.00020502400420896I, -0.000155362029571754-0.000275796948481896I, 0.000189993485023986+0.0000725336962117655I, 0.0000336843819379578-0.0000565776835672248I, 0.0000803967885312247+0.000224813026408075I, -3.38704147598147*10^-6-0.0000385127456912181I, 0.0000677660340836956+7.85069861576896*10^-6I, -0.0000312961751506516-0.0000938294059370216I, 0.0000332170007541486+0.00021195150374414I, 0.000119764455734915-0.0000276367659545241I, -0.000051557023103996+0.000138575258858931I, 0.0000719387655117008-0.000195172397075713I, -0.0000677938781227119+0.000143157742042545I, -0.0000787377855945058-0.000182285340192893I, 0.0000356663019139381+0.0000800913658465583I, -0.0000725675565767897-0.000170998238364317I, -0.000071829595669552+0.000218984282130892I, -0.0000597348558750862-0.00007216847399886I, -0.0000576607431266157+0.000132538893511818I, -8.11219411054115*10^-6-0.00018835934491074I, -5.45867890276277*10^-6+0.0000536485087353932I, -0.0000603915266802734-0.000146341441341497I, 0.0000188224820812932+0.00014761693308426I, -3.65764923157856*10^-6-0.000173930429537151I, 7.50310749054961*10^-6+0.000220368188993339I, -0.0000486830640387496-0.0000900052045179974I, -0.0000539743183981196+0.000104965985017559I, 0.0000806790152188281-0.000131356081677294I, -0.0000418574465510153+0.00012518724707986I, -0.0000700382663308658-0.0000946244655605542I, -0.0000367343775984036+0.00015464053867771I, 0.0000500109058008854-0.0000546960869637093I, 0.0000703315265196804+0.0000511228292840059I, -0.0000471322288033286-0.000201952587425194I, -0.000140837974951587+0.000193279537005825I, 0.000139499403873772-0.000109312500574918I, 0.0000253901600662562+0.0000585431302556368I, -5.14950427785703*10^-6+0.000053221128470091I, 0.0000155221920930835+0.0000343236531879939I, 0.0000532487885889828-0.0000589935211053356I, -0.0000955036351787316+0.000161978134987781I, 0.0000382495090701314-0.000047384522507529I, -0.0000128195189482595+0.0000440255504510409I, 0.0000428600903239067-0.000136792201916356I, -0.0000336012303351168-2.24842624995288*10^-6I, -0.000103815159775754-0.000182222619835011I, -0.0000587242663073761+0.000141294071847649I, -0.000165200993260869-0.0000900542897519038I, -0.0000134598222451016+0.000237128678506079I, -0.0000285192104987507-0.0000801779303301602I, 0.0000375623394281851+0.0000988639731799876I, -0.0000910008589601141-0.0000391437142687632I, 0.0000485455460968941-0.000044598534677063I, 0.0000364070035263157-0.00014109677057287I, -0.000119349937068298+0.0000808685343647175I, -0.0000281285131946415-0.000193792245926464I, -0.000113339637783352+0.0000270532563440175I, -0.0000490369821249571-0.0000115477576422406I, 0.0000581279519349349+0.00017629473733698I, 0.0000300099526560783-0.000219487632762618I, -0.0000608746619605085+0.000125147256248192I, 0.0000277307213912766-0.0000947741976672858I, 0.0000419877593906023+0.000194192793645733I, -0.0000534276840760061-0.000110164254411972I, -0.0000439401292664107+0.000214459798828836I, 0.0000638920532382512-0.000113808792484267I, 0.0000449925017105326+0.000110324734509047I, 0.0000728443934480517-0.000141152545304856I, -6.16993726763877*10^-6+0.000179090177262892I, -0.0000178786663406882-0.000152893261475686I, 0.0000469760108001778+0.000192254910573182I, 0.0000249030628668295-0.000136972249727415I, -0.000074106935600308+0.000117682054670252I, 0.0000511031665752654-0.000113242305819305I, -0.0000399382101824873+0.0000732311191472462I, -0.0000161056128797198-0.000190452650949252I, 0.0000415149845451668+0.0000788166085958871I, -0.0000210393441815052-0.0000492504324949078I, 0.0000213529959794865+0.000104696091727626I, -5.2772648817974*10^-6-0.000155991133828209I, -0.0000184723303364509+0.0000925288791157121I, -0.000108403124605857-0.0000778229308511801I, 0.0000922954574882424+0.000183099510382656I, 0.0000166451822585143-0.000166440492289197I, -0.0000226307866274297+0.0000516736155473335I, 0.0000603465078906387-0.000078872106314014I, 0.000102887624092597+0.000186890429661359I, -0.0000591676611454262-0.0000665806350221423I, 0.000054559900950461+0.000179035390363501I, -0.0000711617675950434-0.0000245180451222311I, -0.0000558595837475358+0.0000775755795800974I, -0.0000626439859789626-0.0000503665941637165I, 0.0000596880709602131+0.0000456337516694708I, 0.0000730929375247319-0.0000691256028547647I, 0.0000560920182068171+0.0000149031195391536I, -0.0000875037127109443-0.000145263341519636I, 0.000104390237124756+0.0000394710799578255I, -0.00012600461013116-0.000148696534686652I, -0.0000788822103966227+0.000042331296517061I, 0.0000948851766852852-0.0000295421176182089I, -0.0000210295712924228+0.000198956858726272I, -0.000148198685037988-0.000156861666203482I, 0.0000996936849030846+0.0000139127138511001I, 0.0000500560270480794-0.0000922981693174147I, 0.0000548441411569605+0.000161029386074425I, 9.42759974985103*10^-6-0.0000787712647368321I, -0.0000304391602861512+5.77809205173214*10^-6I, -0.0000179074618222393-0.0000397312738085935I, 0.000072453939298282+0.000185995036475157I, 0.000117149706715622-0.000130483769001723I, -0.000067559891039841+0.000150084948500048I, 0.000061392332373667-0.000174656576715277I, -0.0000140318343471639+0.000188418479709168I, -0.000116965052844867-0.000126843227373246I, -8.97529125380276*10^-6+0.0001467412683553I, -0.0000143422851451978-0.000131033734346527I, 2.94164582554124*10^-6+0.000203781810995848I, 0.0000450478932728026-0.000062269121381555I, -0.0000626787160932988+0.000205679433929929I, -0.0000600287968389407-0.000224169913350256I, 0.0000185621136956842+0.000133642914879596I, -0.0000495271772147309-0.000113930600690905I, -0.0000958789140445102+0.0000987004185691795I, 0.0000662052127064772-0.000140372300497422I, 0.0000542648052678201+0.000122215117530862I, -0.0000481891541728537-0.000162130955372375I, -0.0000160623074279327+8.64156511074012*10^-6I, 0.0000746207096670702-0.000114051400805918I, 0.000105128189253633+0.000122160306035415I, 0.000124036087218827-0.000236491145309767I, -1.29288122687787*10^-6+0.000117210520518353I, -0.0000149969879125215-0.000013506746340676I, 0.0000157322958678622+0.000112144155353195I, 0.0000482785888593618-0.0000854424719700612I, 0.000028718486349565-0.0000227822504620749I, -0.0000838532061756675-0.000137138058851416I, 0.000132466554743957+0.0000922381630225056I, -0.0000501063392858623-0.000205990131736771I, 0.000133494597562027+0.000183284578583858I, -6.88294296758596*10^-6-0.0000116436346739092I, -0.0000245599995735369+0.0000792447908166732I, -0.0000349171513320105-0.0000486087340055846I, 7.12193211939746*10^-6+0.0000658869004840183I, 0.0000904070620768558-0.000130061777451814I, 0.0000217864194031522+0.0000659676471667609I, 0.0000203437739057284-0.000059690116835113I, 0.0000387851700960322-0.00010781441146509I, -0.0000251256314151846-0.000137250216432255I, -0.000214170462434196+0.0000307321915661767I, 0.00019415675280785-0.000185463770463947I, 0.0000404573432979132+0.000266653585902346I, -0.000161382737379881-0.000132215314269564I, 0.0000457653367264251+0.0000805176839046156I, 0.0000657564338326873-0.000106626529880998I, 0.0002324919181273+0.0000768728184022771I, 0.000126465259136124-0.0000368125254513304I, 4.66087753271297*10^-6+0.0000437876810388741I, 0.0000448803817733714-0.000145722911271886I, 0.0000633447187250295+0.000108032117393803I, 0.000163220388063853-0.000145964814093105I, 6.76143625204302*10^-6+0.000189696196654034I, -0.0000276419627667182-0.000264909147399535I, -0.0000343037633147488+0.000174696265995448I, -0.0000985861859762579-0.0000514669594503137I, 1.86238855667968*10^-6+0.0000414929254214162I, -0.0000990590752075829-0.0000888471053611662I, 0.0000524403481418231+0.000261189448097623I, 0.0000532702982897445-0.0000395736627552913I, -0.000011176056566674+0.000227952753930213I, -0.00014203380415771-0.00015348234697667I, -0.0000581314136048771+0.0000734676166890208I, -0.0000600983849529305-0.000033034886039233I, -0.000051244034307455+0.000168426806879344I, -7.11384872137454*10^-6-0.00019813332257574I, 0.000118261379379593+0.0000904176644599214I, -0.000039337276199652-0.0000564948742374781I, -0.000157144060856305+0.0000979798613551059I, 0.0000529571675813286-0.000269597816080358I, 0.0000794353350454384+0.0000289567449391096I, -0.0000170627925711585-0.000335442067811105I, 0.0000188799275514069+0.000132395434137705I, 0.000154198995011654-0.000108355487000562I, -0.0000813931092907291+0.000111466320039158I, 0.0000907064395480305-0.000129914424743874I, -0.000108658842127031+0.0000151381080358122I, -0.000109141502577331-0.0000132567660037901I, -0.0000253508189426075-0.0000178782687937397I, -0.00014439739265536-0.000155467383779583I, 0.000102169007628729+0.0000159899415968601I, 0.0000796817547723864-0.000110502369410089I, -0.000029223780404076+0.00010179543679018I, 0.0000285513330590183-0.0000864019856567168I, -0.0000202539315249457+0.000126016203418176I, 3.44940043013507*10^-6-0.000183884047506355I, 0.0000121045737651707+0.000105262236906243I, 0.0000923420778857118-0.000163981999617312I, -0.000177891943425032+0.0000202949555382234I, -0.000018147620779627-0.000181800308766694I, -0.000166632415232186+0.000244883619204752I, -6.90330497544086*10^-6-0.000290882849498188I, 0.00012246589581194+0.000152939174969512I, -0.000096479390735513-0.0000618031533336457I, 0.0000618335507857183+0.000104974176855227I, 0.0000410658955696009-0.000139702434158464I, 0.0000859446157237472-0.0000947438591640396I, 0.000138387323601836-0.000170523108577031I, 2.58750622036578*10^-6+0.0000346099266134941I, -0.0000432262642750531-0.000184397117740542I, -1.55636681844324*10^-6+0.000102386011646449I, 0.000117288785732249-0.000207701459104519I, 0.0000451042506852956+0.000124798992997753I, -0.000112260469084636-0.000177911169009958I, -0.0000264984006764018+0.000186539018088311I, -0.0000276075750255641-0.0000509247823389703I, -0.0000820864947178137+0.0000382580179198525I, -0.0000646030476515418-0.0000140542641290739I, 0.000111872285114495+0.000211866698020198I, 0.0000748152677506813-0.0000483666145034437I, 0.0000211839876578169+0.000188208405983468I, -0.000102557364854882-0.0000723483652170031I, -0.000115849079821138+0.000123735335137145I, -1.6816387762817*10^-6-4.77500371225374*10^-6I, 0.0000156804816454664+0.000145488706251389I, -0.0000749108878288688-0.000153934967157548I, 0.000118021255262694+0.000034846510014136I, 0.0000453705182762612-0.0000247574442164272I, -0.000130457876814967+0.000234347732075144I, -0.0000697848128455967-0.000314665216425643I, 1.16141422897147*10^-7+0.000024703216386223I, -0.000140733231767182-0.000249303961370397I, 0.0000431212201316747+0.000129536347293568I, 0.000170725309956047-0.000284643320325746I, -0.000136626330819722+0.00016613319225245I, 0.0000574179687470194-0.00017497315579953I, -0.000134740039042393+0.000172885427926332I, -0.0000158907941631704+0.0000434962837375803I, -0.000214042014122958+0.0000381812713267498I, -0.000116225367382145-0.0000361035497617386I, -0.0000769099819393923-0.0000684965350692175I, 0.0000705641780279646-0.000260479493465382I, -7.68317324786669*10^-6+0.000146817419329272I, 0.0000625815304305514-0.000175761574982949I, -7.04649940966842*10^-6+0.000191102662008824I, -0.000116321163882035-0.000153295420690694I, 4.0818328339454*10^-6+0.000142646345443883I, 0.0000136794573854453-0.00029718233714648I, -0.00022230670569301+0.000368034168238752I, -0.0000722582302630651-0.000169242012464282I, 0.000147194313604403+0.000393740653282595I, -0.000302402418028718-0.000128437651744927I, 0.000058340278591755-0.0000639171479861636I, 0.0000671901561286318-0.000031608729452869I, 0.0000566562385985569+0.000058863365652992I, -0.0000281577136832811-0.000152561797455178I, -0.000327501373054578-0.000016279550560474I, -0.0000861576668976097-0.000316162641187904I, -0.0000767971283330327+0.0000507128366046263I, -0.000152802850984039-0.0000257718519999818I, -0.0000302425991440221+0.00018979134286336I, -0.000061968523421174-0.000221265542520132I, -0.0000500647515297342+0.0000281811261168216I, 1.95060001383877*10^-6+0.0000441296051304587I, 0.0000481534725843406+0.000185262191556747I, 1.9980390448063*10^-6-0.000156845025067965I, -0.000119491983946242+0.000248624453049072I, 0.000129899609513531-0.000062673373995878I, 0.0000601404215057968+0.0000431596211231593I, 0.0000895750476315308-0.000121251955279409I, -0.0000321873497680229+0.000135719726862226I, 0.000043659974096284-0.000153395027630734I, 0.0000216171443275348+0.000263486736552718I, 0.0000487231966905926-0.000125546967448526I, -0.0000749339931080337+0.0000135763975556989I, 0.0000537561331360429-0.0000307499517551632I, 0.0000142506010139348+0.0000279889494181629I, 0.0000530250965013213-0.000214826700864088I, 0.000121236942925618+0.0000839957858648708I, -0.0000690164058808196-0.0000725979586359897I, 0.0000894008699344209+0.000166444882624462I, 0.000153117403923861-0.000148025111410828I, 0.000039519043169403+0.0000869891771709221I, -0.0000954851607413938-0.000165487121208284I, -0.0000553135483788181+0.000202184970058575I, 0.0000446137834425722-0.0000322563249982518I, 0.000175119641917989+0.0000324468007443102I, -0.000131468267809443-0.000155736285775562I, 8.04040946520046*10^-6+0.00020638074769931I, 0.0000150821470914724-0.000286041000368011I, -9.71550237751*10^-6+0.000227506543394257I, -0.0000557170957332216-0.000169013592255185I, 0.0000798345882028988+0.0000899692037870209I, -0.0000315941637049475-0.000183331982687022I, 0.0000110286483762609+0.000181541454106674I, -0.0000337289083471446-0.000109678722491471I, 0.0000153557756121313+0.000214740451611314I, 0.0000545753525255466-0.0000783190364602408I, 0.0000447543417868464+0.000142070855825126I, 0.00010578746522174-0.00011369438101041I, 0.0000580929638448316+0.0000675017080579778I, -0.0000138721725135098-0.000126702984159829I, -0.0000330614073939611+0.0000977590472585869I, 0.000104472074780793-0.000129779574997507I, -0.0000792794759761601+0.000195449384184018I, -0.0000363607982646136-0.000132674792554353I, -0.0000371765464297629+0.000176200010101775I, -0.0000510332276219019-0.0001238961231729I, -0.0000104037550817482+0.000220783275451391I, 0.000047943606615468-0.000119252104977644I, -0.000017017188690899+0.0000778931951394973I, -0.000123706834876862-0.000134726047194029I, 0.0000127850911911011+0.000123939498403798I, 8.44115078662793*10^-6-0.0000695548715418984I, 0.0000656239368543254+0.0000724437629573096I, 0.0000684368274753517-0.0000850839802893679I, 0.0000388745887585755+0.000197670657753625I, 0.0000880111164462327-0.000153532560502848I, 0.0000137389870353747-5.28912860776207*10^-6I, 0.0000330815141087709-0.000247007287316179I, 0.0000853074053653564+0.000105996616811816I, 8.34046257515274*10^-6-0.0000446781422935519I, 0.000100218289044983+0.000168896382728938I, 0.0000865368939989806-0.000148964874511561I, -0.0000503729517754711+0.000184220130100077I, 0.0000423543693878342-0.000114891408194887I, -0.000146937722004795+9.7754823373288*10^-6I, -0.0000189040274371368-0.000194282569246004I, 0.0000398612024449654+0.000107481595002358I, -0.000102920329618377-4.95665687738461*10^-6I, -0.0000156904482340632+0.0000118457089358518I, -0.000124278719442633-0.000212269007126286I, -0.0000232928598775263+0.0000368727389970242I, -0.000169414109246523-0.000129386159503137I, 0.000118562366858341+0.000294210322143521I, 0.000095395549136451-0.000203667663213548I, -0.0000650020225292129-0.0000496202706670741I, 0.0000185002899894208+0.0000913031117666228I, 0.000069507780579958+0.00024454936790136I, -0.000221494830020762-0.0000991138116212245I, 0.0000614054340464551+0.000185520136840389I, -0.000166292584430376-0.0000697727528678433I, -0.0000227094754724013-0.0000185446759057367I, -0.00014585197198896-0.0000948334978526525I, 0.00014956523999087+0.000104987153082498I, 5.91902741944831*10^-7-1.73381515217943*10^-6I, 0.000192300194472924+0.000082189693936371I, -5.41329352133994*10^-7-0.000236103374715155I, 0.000132934217073018+0.000113918475209273I, -0.0000450262710784699-0.000302510471606191I, -0.0000207996672137563-0.0000237096758053122I, -0.0000330601013049472+3.07575069978616*10^-6I, -0.000071813977651454+0.000160678690341748I, -0.0000607450892616146-0.0003102832678506I, 0.000188349890093089+0.000192265060892326I, -0.0000111293404765069-0.0000383803795903177I, 0.0000447006319061607+0.000187617662892925I, -0.0000314282898302545-0.000053762138192047I, 0.000143331518304213+0.0000424063481567002I, -0.000072556907149276-0.000118619273486438I, -0.0000347162271035364+0.00020602840074883I, 0.0000447617199719308+0.0000459891276420508I, -0.0000462863825076709+0.000070237454675039I, 0.000100242167442322-0.0000988976864292743I, -0.0000758696946109868+0.0000925819392510631I, -0.000041516031233298-0.000240788890577702I, 0.0000376751299694238+0.00010329804213778I, -0.0000210911841379817-0.000191640702337342I, -0.000133962171862442+0.000144951842244871I, -0.000097707727605279-0.000074084760462674I, -0.0000984805572210855+0.0000685471889918141I, 0.0000698608095857408-0.000231058095236451I, 0.0000537333436075075+0.0000692806574211646I, -0.0000660323455965251-0.000197272417311347I, -0.0000212673526532995+0.0000670592201054143I, 0.0000728052528000991-0.000114424592979425I, -0.0000225883588328409+0.000237164039471421I, -0.0000754595574666668-0.000167139470516896I, 0.0000458346738935286-0.0000367360091356553I, 0.000141001370544604-8.08694546875218*10^-6I, 0.0000560275666821602+0.000176840586892242I, 0.000149938069723804-0.000119264103761166I, -0.0000366394528675681+0.000120360209915637I, -0.000050592415111376+0.000089195697403404I, 0.0000909134232661513+0.0000403190034591024I, 0.0000145684528224062-0.0000777313613483978I, 0.0000216125231513285-0.0000311754627625162I, -0.0000576121808839138-0.000156756203453716I, 0.000201808337810525+0.0000656649689958942I, -0.0000564667352086924-0.000188564265073224I, 0.000180472159973026+0.000182225564418305I, -2.68280605067023*10^-6+0.0000427791952782485I, -0.0000432524311154761+0.0000533877473596349I, -0.0000461291042327487-0.0000132409165524886I, 0.0000143571344394429+0.0000415327340646295I, 0.0000978559763442387-0.000123521642788824I, 0.000050778805562242+0.0000488021682336181I, 0.0000671333146977253-0.0000784373344578652I, 3.66735853782145*10^-7-0.000155330422924291I, -0.0000124298425534686-0.000199214909049728I, -0.000264534984052656+0.0000404581602953314I, 0.000138075431179866-0.000226683753431086I, 0.0000685922822400088+0.00025476830465779I, -0.000159803166745328-0.000167153431608033I, 0.000117985267652892+0.000139735181020721I, 0.0000481072160715456-0.0000827425442791109I, 0.000222688208896058-0.0000136648719446506I, 0.000141928703505118-0.000068028112078648I, 0.0000962968297617531+0.0000100714922771206I, -0.0000444497322953122-0.000196381003883392I, -0.0000134609008143275+0.00012977238754073I, 0.000146334783366379-0.0000696012458734673I, 0.0000235450472483859+0.000104033774470066I, -0.0000351956260010402-0.000173262986921474I, -0.0000853585224215806+0.000143665681772318I, -0.0000476277180295105-0.00014702988037811I, -0.0000341062729607569+0.0000126391696695843I, -0.0000826702977526283-0.0000835324012527662I, -0.000020811089978818+0.00021681330147226I, -0.000037992536982183-0.0000123595675189086I, -0.0000707192769119511+0.000111789138575192I, 3.73107673515791*10^-6-0.000177045053880107I, -0.000058401343330563+0.0000717036970800226I, -0.0000653928718019865-0.000108658231115404I, 0.0000182273053407538+0.0000626212208365447I, 3.75248855479064*10^-6-0.000114834959039996I, 0.0000767829566503511+0.000195085978121708I, -0.000020637496858-0.0000970994281807918I, -8.96949234763817*10^-6+0.0000536270019239044I, 0.000107031060530119-0.000126111261204512I, 0.000049880163764524+0.000176256089945547I, 0.000140316936225118-0.000109641418089168I, -2.64209238228478*10^-6+0.000145333945207307I, 0.0000504067729411856+6.66264091496037*10^-6I, -0.0000136719979008143+7.53127834399375*10^-6I, -7.6037589866889*10^-6-0.0000441501369723224I, 0.0000447060610388727+0.0000564068213077737I, -0.0000857127966023668-0.000211445662271057I, 0.000107927547472955+0.0000154924319057084I, 0.000026091499586458-0.000208192082790908I, 0.00012567464632798+0.000132385864636349I, 0.000074655600436956-0.0000144719259712359I, -0.0000153813256768098+0.000112703514639179I, 0.000026215605745746-0.0000427243636601011I, -0.0000572583973494156+0.0000695670927236063I, 0.0000797725124764802-0.000185522223104173I, -0.0000398971027633972+0.0000885439562943109I, 0.0000974918191746076-0.0000177817048664087I, -0.0000936073235917934-0.000155281134262328I, 0.0000376602504006923-0.000110446736935576I, -0.0002700312098389+0.000126886078117848I, 0.000152780510833969-0.000336479060585971I, 0.000128793532384904+0.000232179053696678I, -0.000122412413530195-0.0000462136685297149I, 5.48145756777*10^-6+0.000119677008557625I, 0.000056247801495262-0.000134267810208018I, 0.000228509918747683-0.000108799080244472I, 0.000200741849609742-0.000114878727757322I, 0.000020123644329523+0.0000586451223406659I, 1.9075900719752*10^-6-0.000231778551577895I, -5.19450459002894*10^-6+0.0000475603572775634I, 0.00012318521325623-0.000198502657270634I, 0.0000838296518782288+0.000131724309059105I, -0.00015135475777514-0.00018887645249838I, -0.0000328107817134078+0.000160204683960969I, 5.19772272422678*10^-6-0.0000278665679603862I, -0.0000807932153828711-1.40481100102043*10^-6I, -0.0000684691005778142+5.89102026818804*10^-6I, 0.000116968315843415+0.000184930490093223I, 0.0000603550824335889-0.0000569190342985145I, 0.000039489818054188+0.000148188159642985I, -0.0000656328504650168-0.0000332828037331837I, -0.000115689987749619+0.000123661545951152I, 0.000024162798120853-0.0000178693492594378I, 0.0000422511468311159+0.000123308925992177I, -0.0000741808632312534-0.000111810612491253I, 0.0000717052389018076+0.000029533840606684I, 0.0000626306379078184-0.000053209002584937I, -0.0000536559933774934+0.00022752309780949I, -0.0000930858447383297-0.000215497374444712I, -0.0000199277032672655+0.0000614709568378566I, -0.0000897164946486799-0.000151385688564948I, 0.000031816582794625+0.000104112656193151I, 0.0000414937446618437-0.000236991024725277I, -0.0000704207671371734+0.000175653795216783I, 0.0000331253983169534-0.000133302801132793I, -0.0000266838596953526+0.000130260615339099I, -0.0000156943778153219-0.0000248935925169553I, -0.0000936943191796646+0.000130663717810154I, -0.0000668138810923643-0.000107120081970978I, -0.0000372298286870817+0.000078820334890276I, -0.0000258039078714586-0.000164363167660295I, 0.000014364411122051+0.0000869243783910002I, -0.0000228527410188327-0.000137721076821809I, 0.0000346243376789479+0.000144574311612506I, -0.0000357008255211903-0.0000907275825002802I, 0.0000561112443242675+0.000130892081707625I, 0.0000319109382283147-0.000173068513695615I, -0.0000166131079967549+0.000145529589745051I, 1.57313429259351*10^-6-0.00019142334179605I, -7.26380997524184*10^-6+0.000115597111032884I, -0.0000521181508847855-0.000120639988650158I, 6.70364824616557*10^-6+0.000109715291715636I, -0.0000224649855547442-0.000166336696071001I, 0.0000829404283584786+0.000157422587304261I, -1.48518901006898*10^-6-0.0000880189068409481I, 0.000027573286861389+0.0000529377999061256I, 0.0000370303469014697-0.00011700713625316I, 0.0000764960400427693+0.0000572538536660768I, -0.0000655018523326561-0.000155649425760228I, -0.0000373294638812968+0.000140949164192896I, 0.0000604561150642486-0.0000643956039562329I, 9.27102113841329*10^-6+0.0000712965849214299I, -0.0000182075451895135-0.000104550030656835I, -0.0000604754986568004+0.000124283338092193I, -0.0000148159238860429-0.000162130762797061I, -0.0000423795560706555+0.0000515237449031979I, -0.0000417548816763685-0.0000790604430312229I, -0.0000203584161194496+0.000165568330540441I, -0.0000333982493379256-0.0000294938323739569I, -0.0000629332624557388+0.0000876699680411833I, 0.0000322815734763355-0.000155256140391639I, -0.0000599174641634093+0.0000993550400436737I, -0.0000402992725534247-0.000105828086697264I, 0.00003234634229499+0.0000370492545836868I, -0.0000113839849102836-0.0000811382899556471I, 0.000077021606195164+0.000168262720401096I, 0.0000216377140030758-0.000097508307397303I, 0.0000265361195323248+0.0000949148296627447I, 0.0000521693379306636-0.000133739098032217I, 0.0000355785853676255+0.000195836256884488I, 0.000144007455357203-0.000056724554495493I, 0.0000187064946345425+0.000141716901488072I, 0.0000394994271428772-0.0000769457096535649I, -0.000062820919502147+0.000030220945119703I, -0.0000278930314023388-0.0000162804551608472I, 0.000115261529188205+0.000122430012933726I, -0.0000909459066631064-0.000246433392423128I, 0.0000278663712184018+0.0000683124099541118I, 0.000101771644750053-0.000234222681066209I, 0.0000273426861681906+0.000153083555668272I, 0.0000542761808286757-0.0000994319013609718I, 0.0000420309942017351+0.000136829803801188I, 0.0000353115993928005-0.000115427832301531I, -0.0000657427336718533+0.000122804691341326I, 0.0000271686926971111-0.000169338597288822I, -0.0000648324769162394+0.000163061584474257I, 0.0000867725405321728+0.0000128861359917164I, -0.0000357463864046414-0.0000380283998146209I, 0.000113046197246458-0.0000586884523583729I, -0.000122116793397017+0.0000849858180501072I, 0.000122989953993889-0.000271729244395651I, 0.0000595588484769133+0.000179068642977935I, 2.89401870263425*10^-6-0.0000609792754551478I, -0.0000926573251833454+0.000173666232457751I, 0.0000141337844876735-0.000137432377898468I, 0.000177793898417415+0.0000179864206479695I, 0.000124521822224914-0.0000810696364346707I, 0.0000184727107367249+0.000172661382816645I, 0.0000731120358219414-0.000230465909240885I, -0.0000162493005948701+8.7672444518556*10^-6I, 0.0000168412307174077-0.000156107418293432I, 0.0000949391897624269+0.000147304430557866I, -0.000123351966122954-0.000179441243749565I, -2.22301382088363*10^-6+0.0000905407999219974I, 0.0000601662432914917-0.0000169581783303598I, -6.3843978999716*10^-6+0.0000135028967431328I, -0.0000555538793844218-0.0000469813954968757I, 0.0000779060032691028+0.000121303665517494I, 0.0000206530914795227-0.000126501995690757I, 0.0000888367876049263+0.000108401040383745I, -0.0000508475966250143+0.0000164579025574763I, -0.0000735349711184936+0.0000850638676998692I, 0.0000422931424447822-0.0000390582439956983I, 0.0000698075250647555+0.000179986619702411I, -0.0000878761327458119-0.000133456839240558I, 4.54935273591181*10^-6+0.000039771254332816I, 0.0000458997909089157-0.0000322574856783547I, -0.0000751337155007332+0.000250339922563082I, -0.00010220509472211-0.000199214570794153I, -0.0000777705943024326+0.0000261671473632925I, -0.000186157670209459-0.000122763666700739I, 0.0000140545632030491+0.000100020543585678I, 0.0000520341775123287-0.000247260479433421I, -0.0000410775802796509+0.000170778508170536I, 0.0000154515280893709-0.000172760830351816I, -0.0000897532602679903+0.000177419062888522I, 0.0000437085639682051+4.10926883369063*10^-6I, -0.000132901414726561+0.000121922315209515I, -0.0000591891937650098-0.0000398138423500181I, -0.0000780482666641354+0.0000337559186355972I, -0.0000132953698938029-0.000191613136595012I, 9.04138650822716*10^-6+0.0001051634732012I, -1.50397753569527*10^-6-0.000153941686090133I, 0.0000257649038490199+0.00015343423549756I, -0.000064995638237344-0.0000903114100311382I, 0.0000336218249010553+0.000142071463256598I, 0.0000152138113869016-0.000178685895165971I, -0.0000303101060073065+0.00022949237132787I, 0.0000196378596466612-0.00015374170694767I, 0.000084614320771486+0.000144209866018434I, -0.000103070240799465-0.0000926297247949038I, -0.0000161136452982359+0.0000521425346468531I, 0.000064834841647946-0.000142997663095255I, 0.0000135418569096886+0.000171292702046754I, -0.0000318359409369434-0.000101261938447076I, -0.0000473608196127919+0.000111353859765974I, -0.0000235627613232452-0.000119684384630613I, 0.0000531067299383737+0.000144811554433848I, -0.0000124402426369272-0.000149561371472468I, -0.0000486790784699947+0.000101707011045282I, -0.0000535384523269689-0.0000543430260086359I, 0.0000245634584927095+0.0000815968761839276I, -5.7186588329046*10^-6-0.0000805769192521549I, -0.0000141181020037905+0.0000584753131101348I, 0.0000597919374646331-0.000135969607639281I, 0.0000286804068854272+0.0000824787993604307I, -7.7610195541656*10^-6-0.000134170579962273I, -0.0000469967420524138+0.0000685164046275928I, -0.0000681416736584447-0.000143752665480579I, 0.000022374922733066+0.0000392161891536472I, 0.0000504761925908467-0.0000646602245221675I, 0.0000120480467573528+0.0000622357745546321I, 8.8885768349575*10^-6-0.000141935699700257I, 0.0000653743056948273+0.000142500360436547I, -0.0000266262449761264-0.000122105327160617I, -0.0000683552821103014+0.000160755318176578I, -0.0000189212411271931-0.0000795605274959869I, -9.64363158449457*10^-6+0.000134408354228068I, 0.0000191180091521073-0.00007153487868101I, -0.0000832189216321438+0.0000944963449620922I, -0.00010624337125304-0.0000404657779907283I, -0.000036111058136667+0.000114299032243628I, 0.0000134064277467412-0.0000628291713537636I, 0.0000747317178139582+0.0000638852573992283I, -0.00003318184483867-0.000172489963186768I, -0.000118159855751379+0.000152727158783946I, 0.000100390205188881-0.0000681978944731192I, 0.0000172040716263304+0.0000535967277425827I, -0.0000338281099588405+0.0000159810126076254I, 0.0000274275651501356+0.0000281175361567083I, 0.0000457055032910695-0.000055038966076188I, -0.0000663824610555635+0.000114593550161903I, 0.0000272360612983648-0.0000462249309071847I, -0.0000107363903685664+0.0000576107568099289I, 0.0000208115324324277-0.000132540373635837I, -9.09636726472394*10^-6+0.0000334895609257913I, 4.21728317853366*10^-6-0.000153778369892426I, -0.000120715859586253+0.0000684406379613525I, -0.0000693799594082855-0.000128743425386716I, -0.0000863778780798281+0.000231387165685107I, -0.0000522672717038222-0.000200629220329381I, 0.0000938777397889122+0.0000908218662517026I, -0.0000621049570477762-0.0000320809869129143I, 0.0000574490255008789+0.0000578050970474907I, 0.0000216029722282127-0.000127246230055242I, -0.0000327633557321941-0.0000906547517749096I, 0.000074022620643736-0.000206394581057381I, -0.0000147230542391602+0.0000291449061225821I, -0.00008842288626749-0.00013978433594482I, -0.000031643049658295+0.000106368050204824I, 0.0000301306419345468-0.000191689296170655I, 0.0000343051268915356+0.0000562472768195483I, -0.00010101210693953-0.0000487540130043466I, -4.41817078055598*10^-6+0.000150576651727038I, 0.0000432651242803108-0.0000929335921189781I, -0.000110065933270746+0.0000895525087646613I, 0.0000302052378522701-7.49211698339965*10^-6I, 0.0000822726990940746+0.0000774634794374762I, 0.0000404125583182161-0.000100458023725865I, 0.0000240302969655686+0.0000774294917171548I, 0.0000466875133837656-0.0000299994495326514I, -0.0000587413061232649+0.000179483737796467I, 0.0000712278208562114-0.0000877104792511357I, 0.0000571685665627312+0.0000614803983314653I, -0.0000495157285198894-0.0000356872382810978I, -0.000025850901426842+0.0000417199569653739I, 0.000089209008937061-0.000119758568899707I, 0.0000907994709689468+0.000219222698498435I, -0.000132797454047974-0.0000628528967555757I, -0.0000771616633567996+0.000134147429030801I, -0.0000688336344660782+0.0000336912596389572I, 8.47871123371498*10^-6+0.0000828051829314562I, -0.000116219895956285-0.000214163937155754I, 0.0000309745784470137+0.000179989161040577I, -0.0000483985571883005-0.000132355849279384I, 0.0000531078308556918+0.000199522371726646I, 0.000108303243892117-0.000110609408604191I, -0.000027760488547942+0.000249052373945512I, 0.0000523630404541287-0.0000396401371012403I, -0.0000979547568292146+0.000174084392098077I, -0.000100844740289754-0.000113361890793804I, 9.38264279925751*10^-6+0.0000905114889325626I, -0.0000561572214812411-0.000116852453353569I, 0.0000569380110807499+0.0000887008188729034I, 8.48323568279202*10^-6-0.0000258902962413408I, 0.0000425265919023371+0.0000812027791796173I, -0.0000786914107364778-0.000106902776282746I, 0.000158208670643757+0.000141776979436975I, -0.0000307229603761273-0.000113055112429727I, 0.0000518756363520641-0.000021670292289625I, 0.0000477072158970141+0.0000171354927864459I, -0.0000752393495618492+0.00013438251004272I, -0.0000286071263098909-0.000181755917266305I, 0.0000396757602340707+0.000103593768805881I, 4.13782629452638*10^-6-0.0000677935738880668I, 0.0000534738126543837+0.000193992168417602I, -8.58982897507795*10^-6-0.0000384100644216328I, 0.0000428425473073139+0.0000824907225580145I, 5.36153951581558*10^-6-0.00010426949167065I, 6.91179800986226*10^-6+0.000124801991544013I, 0.0000436026656555278-0.0000397754055266487I, -0.0000100046039915162+0.000106837512180754I, 0.0000211679083093335-0.000133427518257821I, -0.0000387903125286889+0.0000954060440413214I, -0.000019520575858679-0.000130647704791897I, 0.0000136357390183458+0.000059261461091051I, -0.0000424478539926476-0.000109591777152003I, -0.0000338894392285259+0.000143017736278944I, -0.0000395060210287475-0.0000658461007834191I, -0.0000281894557230861+0.0000839342153429852I, 0.0000174235483511366-0.000116838775115239I, -0.0000275595644322773+0.0000700172564639099I, -0.0000218759679116424-0.000107319140553168I, 0.0000388318628517154+0.000086669188148049I, -0.0000213470126723719-0.0000978365854056741I, 0.0000154176808866355+0.000149779900955718I, 0.0000102345513478787-0.0000813515321530004I, 8.87674054015401*10^-6+0.000121605978210427I, 0.0000137353242514537-0.0000991150747077906I, -0.0000332351738771109+0.000141039157278116I, 1.03467186113572*10^-6-0.0000162967560916844I, -5.61711647471398*10^-6+0.000119056249547596I, -2.85501241879227*10^-6-0.000100043480108423I, 0.0000122328310590186+0.0000626146650298217I, -0.0000518924832973615-0.000100524351910107I, 0.0000163130926721174+0.000175870468357777I, 0.0000586573501056267-0.000162043684278246I, 4.92363999726851*10^-6+0.000117762385131673I, 0.0000755005723757359-0.0000634605430434883I, -0.0000329024774576936+0.00011774367537211I, -2.77558162143926*10^-6-0.0000841286349642628I, -6.03111136398549*10^-6+0.000124413545296347I, 4.44859579452464*10^-6-0.0000898967242038392I, -9.28895379818712*10^-6+0.0000758364026121235I, 0.0000268784449267277-0.0000909510207805451I, -0.0000247196312798052+0.0000783743114667261I, -0.0000416569668179537-0.0000584065555684203I, 0.000061025818984693+0.0000727868195362909I, -9.81120285510549*10^-6-0.0000606636650756558I, -9.78349550467383*10^-6+0.0000445099307727127I, 0.0000787669937056466-0.0000741511073009463I, -0.0000196408571223292+0.00013845362170841I, -0.0000272581205821652-0.000105618369187255I, -0.000020534627682392+0.0000830233782745738I, 0.000014594979305816-0.000094638402214028I, 0.0000767715904122772+0.00014003243302287I, 0.0000253522913406763-0.0000487759293782658I, 1.27921894021462*10^-6+0.00010768574989546I, 0.0000444206941424908-0.000117019047701537I, 0.0000200427109685386+0.0000805395362813342I, 0.0000333110888240782-0.0000877769214124814I, 0.0000141405602509637+0.000132357825122104I, -0.0000128203416644949-0.000158827360941512I, -0.000013779032189066+0.0000917290891841473I, -5.49336342801348*10^-6-0.0000751063465396861I, 0.0000252230727647848+0.0000510505195281371I, -0.0000497326308333721-0.000100483614327871I, 5.46988978203646*10^-7+0.000135805789487852I, -0.000017899136408965-0.0000912605059798825I, 0.0000179067917212774+0.000101967764591423I, -0.0000256414555565517-0.0000716619730023978I, -0.0000287820946605305+0.0000530666870363784I, -0.0000103182929791377-0.0000805165389548115I, 0.0000405443295258573+0.000141505031168596I, -0.0000410106003595191-0.000131956407203943I, 4.60586431337323*10^-6+0.000119530082096575I, -2.55523373191063*10^-6-0.0000456733145586552I, -0.00005628632024648+0.00015125144135145I, 2.34894142761577*10^-7-0.000140605666821536I, -0.000058563017523112+0.0000726761959493979I, -0.00010717133907646-0.0000692167574327096I, -0.0000120613352061535+0.000114863683482805I, 0.0000534792449273514-0.000121507692060134I, 0.0000134593814884559+0.0000719580523798018I, -0.0000331801260094927-0.000157019213315553I, -0.0000988334836209627+0.000186125174671577I, 0.0000983779780382116-0.0000701931473629613I, -0.0000478165094369449+0.0000654091673385837I, 5.66586332766633*10^-6+0.0000271656178376127I, -0.0000374149661751079+0.0000185535263702672I, 0.0000354563593021806-0.000100533035921184I, -0.000044210738335412+0.000131852177432578I, 0.0000337659087884908-0.0000766058417015831I, -0.0000112111508506077+0.0000710513004275726I, -2.62412451642581*10^-7-0.000105741446437591I, -0.0000262990915336043+0.0000429699492787012I, -0.000056723088827539-0.000141563582614519I, -0.0000502339157788827+0.000153546288341091I, -0.0000830122277708407-0.000074641638494983I, 0.0000260021342734865+0.000184284764791802I, -0.0000477101910219754-0.0000778079119768594I, 0.0000231770510598736+0.0000600546907066458I, -0.000029495881270983-0.0000413598682102947I, 0.0000194261724128558+0.0000132027623874671I, 0.0000139263670928437-0.000111394626453914I, -0.0000847954272367852+0.0000530878874129943I, -0.0000102945112647773-0.000152837011715014I, -0.0000581264489702066+0.0000475217241284832I, -0.0000396918592024561-0.0000508474903761266I, 0.0000178239002339892+0.000120458689787087I, 0.0000142743736105673-0.000157919316632318I, -0.0000200431392911095+0.0000854261886040406I, -0.000010909742221311-0.0000691317322494181I, 0.0000172503120517468+0.000138228226243222I, -0.0000138608239656866-0.0000898908406232063I, -0.0000484297630564387+0.00013087883697174I, 0.0000319224077918407-0.00006414803353725I, 0.0000411412213905309+0.0000898877600942343I, 0.0000421518326419177-0.0000928284181707809I, -2.86684126991846*10^-6+0.00011492723006991I, 4.32338048059343*10^-6-0.0000954758351744782I, -7.07486885383461*10^-6+0.000146820958562861I, 0.0000236168440122108-0.0000918928798868019I, -0.0000127037419427792+0.0000701301208128632I, 3.41356780444899*10^-6-0.0000677203890869505I, -1.9268717488156*10^-6+0.000062703088111684I, 0.0000280751266751496-0.000124339660739231I, 0.0000458256863655192+0.000113353697086022I, -0.0000421229532308926-0.0000699669216556639I, -2.38454561719451*10^-6+0.000116833969299406I, 0.0000142928226452811-0.0000586443820394169I, 4.64319970824094*10^-6+0.0000864944879508827I, -0.000064029732654974-0.000118597070690475I, 0.000017330333242725+0.00012734223382974I, -0.0000175297270896104-0.0000926352996465239I, 0.0000491239391022869+0.00011183892747452I, 0.0000296080658332118-0.000128057190838437I, 0.0000247422359723861+0.000162145065970678I, 0.0000398916938707731-0.0000948525990823646I, -0.0000194395272576977+0.000156346076963948I, -0.0000440043882779442-0.0000733811113389319I, 2.9381734856975*10^-6+0.0000917513607849843I, -0.0000262306064305518-0.0000868069853047583I, 0.0000160327678811922+0.000073325184974071I, 0.0000337768983682467-0.0000690607263442628I, 2.67588442093895*10^-6+0.0000743784088888356I, -0.0000434671004172236-0.0000552690470044442I, 0.0000947695409877601+0.000050627248544541I, -0.0000150072031691988-0.000069228230543718I, -0.0000237590626394766+9.72438977291903*10^-6I, 0.00009993925984631-0.0000571573949821447I, -0.0000270698393310055+0.000151747185049417I, -0.0000474750446856221-0.000108954181483167I, -0.000010061858179281+0.000067447041388857I, 0.0000229689186290815-0.0000883467913541577I, 0.0000873706813337372+0.000143250052683398I, 0.0000270268358433107-0.0000411419398162786I, -0.0000152199885681033+0.0000880750954071729I, 0.0000463191536776395-0.0000909816728259951I, 0.000039143594061019+0.0000824808114450874I, 0.0000489881007709173-0.000110364557008654I, 4.43786272172435*10^-7+0.000138270984251599I, -8.0999073771864*10^-6-0.000160206202988315I, 1.14204714087199*10^-6+0.000110537023823136I, -0.0000307089497558847-0.0000501586126240891I, 0.0000114111166862648+0.0000737392369432288I, -0.0000298095502661163-0.0000836556604963813I, 0.0000343481465817718+0.000135065167436215I, 0.0000348726677773311-0.0000883309083364176I, 0.0000217675249044685+0.000141267558284766I, -0.0000622868966174952-0.0000774718315691037I, -0.000016827058509038+0.0000892378555098972I, 3.92095445061196*10^-6-0.0000557024893861349I, -0.000017854429681886+0.000133825620773122I, -0.0000121046473353537-0.000122053965517701I, 0.0000171732703322781+0.0000459402114663225I, 7.51013961116732*10^-7-0.0000781701475056618I, -0.0000509234505056238+0.000122397043422053I, -0.0000380746044757937-0.000137837186384145I, 3.54293542990553*10^-6+0.0000341927113335976I, -0.0000707158457911425-0.000171266896934037I, 6.59424255011285*10^-6+0.000079422495429963I, 3.44984384696916*10^-6-0.000139695699390126I, -5.69940491011362*10^-6+0.000155664902356303I, 0.0000431967867942634-0.000132926157438437I, -0.000050274985940519+0.0000563403801795105I, -1.77240494705909*10^-6-1.73615138097801*10^-6I, -0.0000188754457373213+0.000116310135175932I, -0.0000912085387968008-0.0000809689391670344I, 0.0000119615523412066+0.000088214901882183I, -0.0000437861921265616-0.0000943220048227874I, -0.0000101189226717119+0.0000539242729837228I, -0.0000444043928737313-0.000088629311341067I, 0.0000496729849805921+0.0000957955287318035I, 2.03986407413021*10^-7-0.0000622822000848259I, 0.0000699799485125534+0.0000799618440640287I, 0.0000100254312647012-0.000143834511986354I, 0.0000289795774233184+0.0000733898845570188I, -0.0000216540922541708-0.000176661967146567I, -0.0000516277358550664+0.0000507532089281012I, 0.0000140886616095802-0.0000885975695302874I, -2.61171528472175*10^-6+0.000140020270944722I, -0.0000674693418271624-0.000168649449122109I, 0.0000944460304667673+0.000126701770894769I, 0.0000152718445125531-0.0000643290430507342I, 0.0000973540295917472+0.0000615927789469172I, 0.0000623810142112794-0.0000618313636297244I, 0.0000815938278853203+0.0000325295775715245I, -0.0000380158111873195-0.000151557657342441I, -0.000018548782192531+0.000118910333898956I, 0.0000981243590178975-0.0000436477797306213I, 0.0000168813101984256+0.0000826563576311292I, -0.0000300423870217062-0.000139485157561079I, -0.0000725635811557527+0.000110990863941078I, -0.00002346790675881-0.00012490009965414I, -0.000029589865359367-1.29462167307559*10^-6I, -0.0000808982706021585-0.0000565177551378273I, -8.44295217161994*10^-6+0.00018347448858669I, -0.0000366129271846784-7.82035338464948*10^-6I, -0.0000430724021080349+0.0000808455913556275I, 6.91251671075083*10^-6-0.000105099885028333I, -0.0000939067205949959+0.0000583830500246868I, -0.0000347811088654503-0.0000650427940307735I, 0.0000728796661415293+0.0000583658182861082I, -0.00006162987593063-0.0000857489535819013I, 0.0000804913700178404+0.000146416165602233I, 0.000046118392426667-0.0000336327724723401I, -0.0000124958863478219+0.000168558569692028I, -2.05118825851582*10^-6-0.000163967285182232I, -0.0000425865360257416+0.000153701087512182I, -1.69664916883863*10^-6+0.0000205613229131339I, 0.0000166130497289247+0.0001342390128685I, 0.0000577969300200287-0.000163657377538586I, -0.0000570089273444318+0.0000357655780085698I, -0.0000801065223150225-0.0000789365011737578I, 0.0000372036667109875+0.000260201295484108I, 0.0000697909109740965-0.000194597773717348I, -0.0000825563424977427+0.000106071805745664I, 0.000150329358032066-0.0000404705599165725I, -0.000115401889672121+0.0000816649462567273I, 0.0000235317012863524-0.000144838058306917I, 0.0000180305075739786+0.000166859291755446I, 0.0000435435582423246-0.000126487611029809I, -0.0000390018242461919+0.00009639808344757I, -5.02393722823989*10^-6-0.000091553094521106I, -0.000076970313941761+0.0000988528304888278I, -0.0000875379807278038-0.0000367286032320947I, 0.0000738476516505755+0.000170774619810979I, -8.84862136049064*10^-6+0.0000209202759334026I, 0.000108533844851886+0.0000554935839332426I, 0.0000592186403875144+1.81654189083219*10^-6I, -0.000060578922438707+0.0000816701481926068I, 0.0000312214853825344-0.0000620112125686382I, -0.0000978220605506696+0.0000312687706036443I, 4.19066861653137*10^-6-0.00010158554688997I, -6.60460641333126*10^-6+0.000236953864730281I, -0.0000497770856512679-0.0000310886059949788I, -0.0000761939821776177+0.000146281737654992I, 0.00010097565719736-0.0000326256386072205I, 0.000067585474833011+0.0000671961549774069I, -9.79012355543747*10^-6-0.000108682601739518I, -0.0000223060708601706+0.000164617005920534I, 0.0000401425365045238-0.000167605308819984I, 0.0000384417634933901+0.0000845950487687428I, -0.0000394289958259212-0.0000271640674829399I, 0.0000694923219147033+0.000108044519784405I, -0.0000240110570671749-0.000128989217670223I, 0.0000134145555485243+0.000119275107183163I, 0.0000331065949432582-0.000114174218565484I, 0.0000374921190952312+0.000160129926646211I, -0.000101712964871939-0.0000859994369196716I, 0.0000209350844378242+0.0000519994780892709I, -0.0000184017243494754-0.0000504266621462253I, -0.0000478168360762086+0.000178924853006735I, 1.81443860039542*10^-6-0.000166973018597946I, 0.0000202193693986928+0.0000531757159395822I, -0.0000445841292042844-0.0000494643522527473I, -0.000125978507300692+0.000085706195331201I, 0.0000125988284029349-0.000182411988070323I, 0.0000240762590226292-0.00001735563046651I, -0.0000987320840635807-0.000256782136635801I, 2.51940354261145*10^-6+0.0000873137546353536I, 0.000100598139136126-0.000111997267824707I, -0.0000337452750456588+0.000119358913230375I, 0.0000704576699955728-0.000143374019937021I, -0.000132650110430453+0.0000353465789405237I, -0.000033282602151577+0.0000433138880723167I, -0.0000609603068394789+0.0000125862465332032I, -0.000131624228159454-0.000061659738464725I, 0.0000314683750054431-0.0000126551474611606I, 0.0000391266529199792-0.00012446806392725I, -0.0000226789953105371+0.0000794122444620771I, 0.0000172279171801727-0.0000943079638190529I, 5.02244541323164*10^-6+0.000115749956682635I, -0.0000377841473519288-0.000116701369443243I, 0.00002344201059708+0.0000943459414388947I, 0.0000414454878742674-0.000171839772230678I, -0.000120431814159764+0.00013891629826432I, -0.0000177769099288916-0.000142882245338544I, -6.4764958733056*10^-6+0.000203597910268413I, -0.000095016803976605-0.000148083130307821I, 0.0000481302125907449+0.0000506083047110539I, 4.88252705670346*10^-6-0.0000667273204136944I, 0.0000319829692878415+0.0000970378496771679I, -3.8075191887502*10^-6-0.000105908998562871I, -0.0000529415910018916+3.57536148427554*10^-6I, 0.0000161646331813079-0.000150134899255005I, -2.15131206417752*10^-6+0.000075848753892919I, -0.0000351055396641037-0.000109274569269658I, -0.0000206094414770111+0.0000908152292935808I, -8.83417096168057*10^-6-0.00012956863040154I, 0.0000170213016618024+0.0000753735359116275I, -0.0000404173452980659-0.0000719990989989501I, 9.00990772856607*10^-6+0.000102263190485594I, 0.0000267567495070394-0.0000721416736627753I, -0.0000283349605719729+0.0000925077123392548I, 0.0000133240757483249-0.0000625853891734149I, 0.0000405620964503782+0.0000715950519057668I, 0.0000263718654279982-0.000117111860569099I, 0.0000394946054844169+0.0000931528861547519I, -0.0000116512453054086-0.0000339613081382577I, -0.0000121678448980612+0.000113357678020826I, 0.0000422278641976644-0.000070401013763959I, 0.0000147806298843031+0.000125756374036852I, -0.0000290033347429612-0.000096269237467849I, -0.0000319308860933147+0.0000299193935083876I, 0.0000229911417922928-0.0000802334204925262I, -0.0000179913937401455+0.000164736735064519I, -0.000088662065073642-0.000104914126317473I, -0.0000514786218228132+0.0000233666555182255I, -0.000154380101762875-0.000115945392475877I, -2.25814049233179*10^-6+0.0000647204746162057I, -0.0000302190802235837-0.000180404617197601I, 0.0000287127126445339+0.000185642437783331I, 0.0000271795787779972-0.000181996201629863I, -0.0000996504838516156+0.000103990706684738I, 0.0000854769210685401+0.0000580987108165335I, -0.0000594062767891853+0.000152669049037801I, -0.000115109939608372+0.000023468974433I, -0.0000420826792188485+0.000053395426549746I, -0.0000752238142273168-0.000111761736735229I, -0.0000331780071738674+0.000048194808566375I, -0.0000573213575389172-0.0000887886565379279I, 0.0000859790350075105+0.000084861106889991I, -0.0000248455656634548-0.0000171182287277558I, 0.000100282938545149+0.0000406576195051229I, -0.0000617262154931848-0.000222277636449957I, 0.0000528496026768803+0.000193448307074552I, -0.000104916359516837-0.000193669933488104I, 0.0000536499892533616+0.0000725958404107465I, -0.0000797153140545956+0.0000387614022270954I, -0.0000463509082908319+0.0000757748780682988I, -0.0000671726061823295-0.000187575670042964I, 0.000154064859607553+0.0000636310166894035I, 7.17267533869054*10^-6-0.000041241937929206I, -0.0000403612683016207+0.000122748925544378I, -0.0000287548830383721-0.0000788217539967002I, 0.0000549622056354375-0.000023991872106319I, -0.0000882451958683644-0.000046727495403024I, 3.80912969116394*10^-6+0.000186875190659867I, 0.0000987422601000275-0.000017960221312193I, -0.0000580054394463804+0.0000570910427552449I, 0.0000606975408450296-0.0000958981907951929I, -0.0000661232481699847+0.000131999447282457I, -0.0000899966892353768-0.000168704608617297I, -0.0000288772471925268+0.0000656750328931759I, -0.0000466825284371164-0.0000870323494097168I, -0.0000442354186239877+0.000190793384115954I, -0.0000186071377870315+0.0000265391845548839I, -0.000109096748960017+0.00011530395591147I, -7.4599439565055*10^-6-0.000204377840085378I, -0.0000472078758157412+0.0000751812546474338I, -0.0000776242433572636-0.0000739414554372007I, -0.0000222012530191339+0.0000112377045278171I, 0.0000148536567587515-0.0000842433009495459I, 0.000126780163583337+0.000144600218470266I, 8.61522836921543*10^-7-0.0000748751012692152I, -0.0000181791069823655+0.0000373242390225493I, 0.0000714853054756444-0.000167494512815683I, 0.0000877551300703985+0.000146766892196347I, 0.000149495524187801-0.000137485714473377I, 0.0000332902532867279+0.000121885906136166I, 0.0000890706317148401-0.0000870434706345257I, -0.000109292619266651+0.0000387151674347208I, 0.0000208293884657477-9.08543822452209*10^-6I, 0.0000789656805064178+0.0000666808057104667I, -0.000127617591943985-0.000143637094318563I, -0.0000326970814209246+0.0000365614792117772I, 0.0000185953082432045-0.000210539911778879I, 0.0000112517872573169+0.0000894618620438433I, 0.0000373949206689002-0.000140281503956553I, 0.0000481054989834101+0.0000998641801523576I, 0.0000182413095199808-0.000132579057119288I, -0.0000272508477926112+0.000144272074030216I, -0.0000312403545291506-0.000120737446493547I, -5.65525809080485*10^-6+0.000163884280594263I, 0.000076787834650114-0.0000766842006997129I, -0.0000401465613001431+0.0000848091458256751I, 0.0000793701213745926-0.000119175162624469I, -0.0000115543277726376+0.0000909341202707637I, -0.0000132615652481935-0.00015050297388039I, 9.19730328609085*10^-6+0.0000959838114455668I, 0.0000331800158492342-0.00012666519976139I, -2.85322213858587*10^-6+0.000177075454726356I, -0.000014008116141942-0.0000843408772531332I, 0.0000610514619555652+0.0000704228239156817I, 0.0000358075159471267-0.0000695810092968423I, 0.0000731924285086455+0.000114987063892453I, -7.20761775749273*10^-9-0.000165768577870933I, -0.0000381773798674218+0.0000768568452238405I, 0.0000209781615927337-0.0000345854070852393I, 0.0000357752678069011+0.0000828799477418584I, -0.0000269999160513234-0.000121012369112993I, -0.0000487240758612881+0.0000677314183328174I, 0.0000132576110462313-0.00011145682035596I, 7.64665370484839*10^-6+7.78281683298569*10^-6I, -0.000070901377649963-0.0000810411552367083I, -0.0000260548522051175+0.00014370772016504I, -0.0000522656363982796-0.0000390168883567225I, -0.0000183714737302707+0.0000646450393990908I, -1.10412960959826*10^-6-0.0000867264406354238I, -0.000070125874933054+0.0000312307687550984I, -0.0000390161653214117-0.0000619700776319293I, 0.0000677076069046186+0.0000882230816189732I, -0.0000591331741561264-0.000109691190378737I, 0.0000746519445090041+0.000142429352006661I, 0.0000210258454013219-8.54022228639405*10^-6I, -0.0000654549711929531+0.000151696598523292I, 0.0000185103029089611-0.000204468083807486I, -0.0000249050720332431+0.000101861754699581I, -0.0000318276355741961-0.0000713963960624015I, 0.0000222673221116674+0.000130823221852643I, 0.000142698119924722-0.000170743534713094I, -0.000105759743780553+0.0000361577413166241I, -0.0000190929723244676-0.0000826039044190578I, -0.0000300233980032901+0.000202900235732935I, -0.0000145535059535319-0.0000912499476893655I, -0.000155439614651291+0.0000224628972860643I, 0.0000333984483907946-0.0000724662273376398I, -0.0000938816007574169-0.0000167690380344358I, 0.000069556105135365-0.000221884394934336I, 0.0000354963584009073+0.000147149838135824I, 0.0000655974827051611-0.000173434620001835I, -0.0000388815694883533+0.000172421347232559I, -0.000103905928918714-0.000121715309120398I, -0.0000348863188357452+0.00017632180532533I, 0.0000317637193948838-0.000154953606635881I, -0.000104812462862631+0.000301659076367207I, 0.0000547524656741784-0.0000946261658248982I, 0.000183563251598866+0.000212010422436377I, -0.000201870697537447-0.0000738727535926153I, -0.0000210769927203763-0.0000476897969088633I, 0.000164019253687201-0.0000969707803808762I, -0.000037494768886867+0.000168424806739453I, -0.000068730283623952-0.0000998021777484909I, -0.000191444359642844+0.000129296712824481I, -0.000109720041458231-0.00015026710669522I, 0.0000385808309704339+0.000175216191400988I, -0.0000362798524793442-0.0000999992604517517I, -0.0000755045000715968+0.00010552735943793I, -0.000141802822211234-5.20522462996661*10^-6I, 6.33506568384805*10^-7+0.0000288675974967956I, 0.0000486430658189367+0.0000131690811106249I, -3.84277509858858*10^-6+0.0000129501565378936I, 0.0000850560036454885-0.00019189274898427I, 0.0000451288096812654+0.000118019378655382I, 0.0000336447815869941-0.000167669542104493I, -0.000123213358948851+2.2160500292507*10^-6I, -0.000132616019346811-0.000137778645810388I, -0.0000244627820930907-0.0000373618234168684I, 0.000143305238521235-0.000105465018429848I, 0.0000515032090336106+0.000041360324447276I, -0.0000171610876487713-0.000199901079751299I, 0.0000858107100799895+0.0000649853876527704I, 1.5523751841772*10^-6-0.000070304456772641I, -0.0000867535124972284+0.000253414290549149I, -0.0000287356087209076-0.0000933417273603084I, 0.0000757885011015478+0.0000575365261538696I, 0.000107059186679083+0.0000182958870729796I, -0.0000851082258840913+0.000206759395322375I, 0.0000460857720759797+0.000134185937933945I, -0.0000487838437381324+0.000137216481285591I, 6.93057757435069*10^-7+0.0000406703107977747I, 0.0000867020222379501-0.0000786616892876644I, -0.000136645353793854-0.0000754361292552454I, 7.25396126384085*10^-6+0.000236644094352891I, 0.00012051567439747-0.000278343360884741I, 0.0000666338168601924+0.000025414221365449I, 0.00019595500773753-2.92914959605433*10^-6I, -0.0000157991888865798+0.0000689643635341598I, 0.000103481397612086-0.0000314007286063184I, -0.0000357916181593044+0.000186251267434603I, 0.000094800050659632-0.000047229789778453I, -0.000100665835338359+0.000030273479698057I, 0.0000459678278825545-0.00015445249652963I, -0.000147197561241314+0.0000481832861653777I, -0.0000511497695953684+0.0000130667646155976I, -0.0000531144890760582+0.0000586866184189127I, -5.40391848162125*10^-6+0.0000703240108743849I, 0.0000125794521307726+0.000165253981242683I, 0.0000453863223129419-0.000133413276673364I, 0.0000343652716898863+0.0000598902374237363I, 0.0000491107678075216+0.0000510756120224942I, -0.000147030906573835+0.0000127043242185287I, -3.30804141822815*10^-7-0.000141352835050723I, -0.0000630057810909717+0.0000918024127995956I, -0.0000261919198979875-0.000121667280634023I, -0.000112038060114143+0.000174250834327252I, 0.0000753653419793969-0.0000634363113861839I, 0.0000225798730912072+0.0000228176424682033I, -0.000108667714433419-0.000165546286402126I, 0.0000208955902675956+0.000120708893155163I, -0.0000183988500109933-0.000067940392618121I, 0.0000775451921626029+0.0000518387709444526I, 0.0000742522310977691-0.0000288312249212753I, 0.0000452707649534886+0.000143512915725457I, 0.0000561365533972472-0.000123305966500483I, 0.000012240926190711-0.0000192289742899264I, -7.32148374997226*10^-6-0.000214360667400376I, 0.0000957303944009023+0.0000527406992346494I, 0.0000255846508042573+4.84352190549247*10^-6I, 0.0000750882220481825+0.000078454714618134I, 0.000060060418333866-0.000136735977148616I, 0.0000314413156804972+0.000180204912970395I, -7.47207079793267*10^-6-0.000112445958062257I, -0.000159118097453204+0.0000928803949262037I, -0.0000360595875298468-0.000094432808843042I, 9.06009700792414*10^-6+0.000117387358229096I, -0.000013568570919341+0.0000224104091284992I, -0.000127424752431109+0.0000409358103762139I, -0.000189471463846055+0.0000127330844178701I, -0.0000675159425072255+0.0000733947161952356I, -0.0000717918521815338-0.000013172913902942I, 0.000181264730944485+0.0000694989120289323I, -0.0000693034772516915-0.00021615255199801I, -0.000171136371726543+0.000161281002068372I, 0.000236812842355964-0.0000481112405793141I, 0.000078845398363615+0.0000848384657485762I, -1.65793103304103*10^-6+0.000163191346204741I, 0.0000161199616175208+0.0000292750171790397I, 0.0000126292506998042+0.0000398959980517059I, -0.000125011754757932+0.000106129846622417I, 0.0000146415284119807+0.0000336918668910478I, 6.19722624299875*10^-6-0.0000459006658382396I, 0.0000870721281121384-0.0000737278668069611I, -0.000030384454451856-0.0000967717569547319I, -0.000143069372510003-0.000113145609739723I, -7.37479907545638*10^-6+0.0000337164727761617I, -0.000200407056591896-0.000027407006612578I, -0.0000704171768131566+0.000151590426165115I, 0.0000538332212614918-0.0000360314417061199I, 0.0000446795468227026+0.000106096617546573I, -0.000157199197523008+0.0000165197649908436I, 0.0000557892729937353-0.000128841469168894I, 0.0000663031030443217-0.0000968778285962922I, -0.0000386683603956567+0.0000175712567596248I, 0.0000268033025203302-0.000125727840139838I, -0.000127533034569215-0.0000519735630676874I, -0.0000348805790422732+0.0000216687108523373I, 0.0000862869213663049+0.000123723184208335I, 0.000107782715298934-0.000206683505228181I, -0.000051235354712036+0.000107620982408199I, -8.77470774954529*10^-6-0.000111287136021585I, 0.0000239517334701688+0.000186212472194664I, -0.0000938009799002458-0.0000307164165769483I, -0.0000717906452004605+0.000119863920849633I, 8.83925714620509*10^-6-0.0000216606714300309I, 0.0000933631865097916+0.000146431947631513I, 0.000106930458751683-0.0000297607793903343I, -0.0000186402938069438+0.000184702758312685I, -0.0000823741867675583-0.00011674611838441I, -0.0000335487622448805+0.000137897297603561I, -6.77503554753677*10^-6-0.0000228741898913434I, -0.0000664912595280714+0.0000628457124857296I, 8.80290956956685*10^-6-0.0000854710008596964I, 0.0000825973928501621+0.0000140304931032056I, 0.0000169898741085721-0.0000818252294571838I, -0.0000335453187696814+0.0000840856361817597I, -0.0000196564247659884-0.000157231536823255I, 0.0000590578939605146+0.0000616844928041109I, 0.0000265201118714959-0.000167680271662836I, 0.0000281099709563783+0.0000846207868254131I, 0.0000482489620403249-0.000129594439065124I, -0.0000678226125051959+0.000104484648837753I, 0.0000266245655721992-0.0000697003726826045I, 5.81490983513009*10^-6+0.000075279943582071I, -0.0000408505674879239-0.0000710513925962741I, -0.0000495549723563194+0.0000598598932991788I, -4.54114795287314*10^-6-0.0000969605368704557I, -0.0000160054676298315+0.0000500460286713849I, 0.0000332159006829076-0.000121791876653486I, 6.7491177901558*10^-6+0.000105454341429119I, 0.0000328039770008915-0.000100619404457363I, -0.0000234158141979551+0.000102269314178486I, -0.0000209240742661097-0.000110802291883097I, -0.0000328207958455138+0.0000929020380261561I, -6.73842938329016*10^-6-0.0000982631521330221I, -0.0000649913935814336+0.000137311790258708I, -0.0000186743893075963-0.0000488699908351147I, 0.0000398310950591377+0.000171182800123128I, -0.0000531470647567925-0.0000971332928794475I, 0.0000238496323387449+0.0000342450884980511I, 0.0000232346390024943-0.0000114404012939075I, -0.0000351027272816667+0.0000300574868606414I, -4.09176194298183*10^-7-0.000113116235099436I, -0.000105151939601847+0.0000555930980253388I, -0.0000311121201811378-0.000147097271332516I, -0.0000797868603429763+0.0000847642555607447I, -9.74110116578473*10^-6-0.0000301708324796328I, 0.000019281391976605+0.0000842662304885839I, -0.0000382111721977408-0.000164376263711154I, -0.0000179807860407206+0.0000832868898223886I, -2.28157125249636*10^-6-0.0000440731666830972I, 0.0000518431877258466+0.000110411218188949I, 1.38651824690681*10^-6-0.0000548349653892985I, -0.0000263083946279509+0.000156242543371799I, 0.0000629913925197407-0.0000712325921570288I, 0.0000469570923641866+0.0000382732938581407I, 0.0000600848394030427-0.000122497725545667I, 0.0000239979596930995+0.000113391225312093I, -0.000011611115042887-0.0000740134626527331I, 0.0000343212616708547+0.000142239318402229I, 0.000034926818577269-0.0000849799530464235I, -0.000047843197166953+0.0000852470611139841I, 0.0000281785974793743-0.0000703640107699601I, -0.0000295654734927872+0.0000256645767738905I, 1.2327369233298*10^-6-0.000126646135099983I, 0.0000287040855893648+0.0000673091091222675I, -0.0000306387667869253-0.0000485129280066979I, 0.0000211357369230652+0.0000668204324039301I, 9.16230018537124*10^-6-0.000116302997821706I, -8.51972121363505*10^-7+0.0000622380528472146I, -0.0000534013313927358-0.0000745404873096525I, 0.0000253674107048556+0.000124721846301605I, 0.0000184535294274476-0.0000854914999013933I, 0.0000148055104249229+0.0000354446165844415I, -8.15255620170414*10^-6-0.000075013994232179I, 0.0000409323772185218+0.000109250456229455I, -0.0000153805186240049-0.0000986802986476173I, 0.0000272304634181041+0.000115752637929919I, -0.0000159435988311811-0.00005267522953131I, -7.0452344875768*10^-6+0.0000699294934688482I, -0.0000115334595344321-0.0000641677841887313I, 4.38205350933923*10^-6+0.0000651127661557209I, 0.0000264055187340192-0.0000851240581667106I, -2.4507185566715*10^-6+0.0000645063392426703I, -6.65320228837862*10^-6-0.0000598092767466821I, 9.83068902367967*10^-6+0.0000300705747788897I, -0.0000117112630273639-0.0000633523249543001I, -0.0000454038819937012+0.0000752293301310201I, 0.0000502490851988625-0.000105336271614086I, 0.0000173561122008555+0.000109840977037218I, -0.0000302313739517985-0.0000479428324701113I, -0.0000201250993886595+0.0000482825122891303I, 0.0000179085700612246-0.0000972437425748156I, 0.0000258991087841764+0.0000626882423399402I, 0.0000233581008187878-0.0000872959273648171I, -0.0000453366695341507+0.0000800417735898586I, 0.0000249836275721267-0.0000702198521468829I, 0.0000276404087582003+0.0000628356768089397I, 6.84429937774406*10^-6-0.000144150678941377I, 3.99912863300228*10^-6+0.000108936684935056I, -0.000022448149060654-0.000101252054328648I, 0.0000315788442982217+0.000102538154442267I, -4.30277904179062*10^-6-0.0000296270319492111I, -7.49591614554364*10^-6+0.000104733468194036I, 0.000019001060878598-0.0000665159899530101I, 0.0000527265043717886+0.0000707362106577009I, 0.000053908331624104-0.000114381092015443I, 0.0000413631368334761+0.000121312610309545I, -0.0000427920974551513-0.0000501350351314396I, 0.0000153003265858413+0.000109217746844755I, 0.0000302304545593548-0.0000633463169783425I, -0.0000330184243053395+0.000121169692206246I, 7.74534627025059*10^-6-0.0000967783448087135I, -0.0000271272435021022+0.0000190638611754536I, -7.45898411562582*10^-6-0.0000964843810547854I, -0.0000183413246876213+0.0000896104460831233I, -0.0000360240418184653-0.0000743218480386735I, 1.34358134645886*10^-6+0.0000248884435535349I, -0.0000601592483359073-0.000145891944069825I, -7.36594713119156*10^-6+0.0000586632023194924I, -0.0000313663865378296-0.000078736538736439I, 0.0000382739791636487+0.000133709598020293I, 0.00003286182921972-0.000122485305985587I, -0.0000542499377152542+0.0000253758074307642I, 0.0000166239396045066-0.0000149416660834482I, 0.0000380298983331847+0.0000890974295411657I, -0.0000731842719197623-0.0000527472798436131I, 0.0000500809574151125+0.0000803851186048364I, -0.0000146287341496948-0.0000274719526439863I, -0.0000374207798282515+0.0000528227144285016I, -0.0000204125482468145-0.0000321214282127135I, 0.0000201026410805937+0.0000416747685950314I, 0.0000434501103840398-0.000081790989480072I, 0.0000221119968448083+0.0000220257474277784I, -4.9333395507362*10^-6-0.0000885045690954558I, -0.0000190464739784849-0.0000184134072112677I, -0.0000582163159815062-0.0000981227655100214I, -0.000125080711526844+0.0000938475106401168I, 0.0000703501182515939-0.000140368162174741I, 0.0000581308035244198+0.000142404730234019I, -0.000117336160872637-0.0000391359138948169I, 0.0000452186674868222+0.0000150982291394109I, 0.0000510953430631927-0.0000911415573416676I, 0.0000912638583008385-0.0000361813995052525I, 0.000104704422988525-0.0000965759175163773I, -0.0000377992175931237-5.0776507365636*10^-6I, -0.0000143311542938407-0.000092445443755964I, 0.0000381945915341804+0.0000707596682315816I, 0.000118244585211743-0.000188225669056286I, 0.0000177197057796342+0.000107276491113676I, -0.000088231309316889-0.000142899862230508I, -1.52854723256525*10^-6+0.000163135413472175I, -0.0000498758404649855+6.40484865557186*10^-6I, -0.0000853673516641333+0.0000404340452718402I, -0.0000326266736074754+0.0000298175793216044I, 0.00013888423350918+0.00016051842573792I, 0.000122315240551315-0.0000213528275916015I, 0.0000265605124498076+0.000180115685700195I, -0.000115239279150085-0.0000376011705372963I, -0.000099661480245567+0.000136159076447584I, 0.0000252799542707747+0.0000413374495851616I, -0.0000275808870956336+0.0000860431841439483I, -0.0000517477206982309-0.0000838196357337587I, 0.000115837764076465-0.0000672779952754949I, 0.0000743443386119647-0.0000360994125615993I, -0.000075037494211407+0.000191269173931626I, -0.000132093360977832-0.000245604215794348I, 0.0000384643380646971-1.05965287854787*10^-6I, -0.0000743800610956672-0.000219196631156697I, 0.0000655117908111369+0.0000639758259727177I, 0.0000730322493742411-0.000294681533744738I, -0.000154799001627265+0.00020674951208927I, 0.0000782783958490026-0.0000851221669399037I, 4.40426012331101*10^-6+0.0000870755582401627I, -0.000084421697526173+0.0000407369708043823I, -0.00019262797576522+0.000118952451717483I, -0.0000756901758401277-0.000129138115128621I, -0.000103232651430601+0.0000330538311745662I, -0.0000270256915084752-0.000246247427356946I, 0.0000676839143518689+0.000074268828940023I, -2.86662759427262*10^-6-0.000192865620092562I, 0.0000264690210156946+0.000193018149845349I, -0.000114576018608912-0.0000685918495303007I, 0.0000364586529999994+0.000188379163803031I, 0.0000375397364503148-0.000162692529629771I, -0.0000378214177629815+0.000286865214740768I, 0.0000595408041638574-0.000119976624792705I, 0.000168398203679161+0.000141166898058832I, -0.000161767056366754-0.0000519430830048246I, -0.0000343552944409613-0.000022355854821586I, 0.000131101604762123-0.0000972747466475998I, -0.0000310108588855011+0.000150951048539863I, -0.0000500144731785493-0.0000947738125560644I, -0.000142556783105628+0.000113908009090794I, -0.0000751951646839801-0.000135177657262922I, -7.43021949520451*10^-7+0.000152071847046966I, -8.69016037096194*10^-6-0.000077322102754071I, -0.000036751465606014+0.0000742835267346867I, -0.000108193016756806-0.0000950515465810223I, 7.05945619529501*10^-6+0.0000581306999252978I, -4.40854173808068*10^-6-0.0000142644414646685I, 0.0000430420183190289+0.0000544755056195544I, 0.0000634390515394876-0.0000728480706302303I, -8.10881453631092*10^-9+0.000138624402998488I, 0.0000672453322076303-0.000083067311273963I, 0.0000241865262017444-8.52233610966603*10^-6I, 0.0000298930606353439-0.000161882534833257I, 0.0000607356420723263+0.0000648858888951384I, 0.0000144264957152697-9.87908013686666*10^-6I, 0.0000339846825641457+0.000136057685512796I, 0.0000756641810986421-0.0000821510933333633I, -6.74909140820887*10^-6+0.000112246476512642I, -6.81182388124352*10^-6-0.0000617245795309836I, -0.0000744677475554209-0.000017721158847884I, 0.0000362557963787907-0.000112983117545647I, 0.0000320504297988111+0.000147531424559286I, -0.000134929104030257-0.0000546559399980021I, -0.0000359422485701417+0.0000158978813180311I, -0.000122984288686278-0.000110945606784277I, 9.65076353693087*10^-6+0.0000314687365578852I, -0.0000957536980353199-0.000208838976020821I, 0.0000160209417207782+0.000235568756395215I, 0.000048540143530563-0.000138437360640633I, -0.0000114130004292343+0.0000603241008777342I, 0.0000257374510391568+0.00005589653746173I, -0.0000721533237220981+0.000216948899710099I, -0.0000958788161503645-0.000064430246571805I, -0.0000771418896035624+0.000118234180583313I, -0.000129322933510489-0.00016032275592397I, 0.0000349545095180541+0.0000219726604789726I, -0.0000846995153068454-0.000150838159307236I, 0.0000991218897476771+0.000137907882832289I, -0.0000697800372721837+0.0000121422010863317I, 0.000111444480002737+0.000133440372035172I, -0.0000216788223497561-0.000167885199027083I, 0.00013038029562575+0.000249167460525609I, 0.000028160044004203-0.000167185224402501I, 0.000166868963008433-0.0000222200994396915I, -0.0000922134804667843+0.0000756302209117159I, -0.00012034256397186+0.0000369835823110394I, 0.0000971293541428013-0.000226910860243895I, 0.000034161384246021+0.000183938395878014I, -0.0000555325231212803-0.0000389778234634428I, -0.0000608361599806855+0.000261353500829166I, -0.000113869338909689-0.0000254670431110979I, 0.0000862209445669351+0.000153944508186584I, 0.0000185015362430383-0.0000784556201602365I, -0.0000377950067623165+0.000107845597067644I, -0.0000980986734366572+0.0000568560521408051I, -0.0000193054379874434+0.0000671944490843487I, 0.0000913484978916401-0.000052053424317298I, -7.4698313376653*10^-6-5.85162777552677*10^-6I, 0.0000467536145813846-0.000148580126510392I, 0.000103061139838469+0.000104153996356876I, 0.0000144003241628996-0.000190465784301488I, -0.000119117460180078+0.0000133149713543301I, -0.0000995110875637932-0.000162185937785255I, 0.0000155381917868738+5.59771941654352*10^-6I, 0.0000636995349374742-0.0000700056481312675I, 0.0000906939520645867+0.0000223012940865159I, 7.57950338268384*10^-6-0.000162666138275584I, 0.0000269013348508098+0.000144554620990432I, 0.0000185283425192839-0.000111751363007022I, -0.000135916390377401+0.000144374241173305I, -0.0000594837618715541-0.000095617163245917I, 5.88342254472013*10^-6+0.0000561990560142373I, 0.0000220187001618816+0.0000236558156140557I, -0.0000830692399471623+0.0000415852878435363I, -0.000138989662157453-0.0000408994781057159I, -0.0000579858401606137+0.0000623487197856461I, -0.0000696856035331761-0.0000112287719864869I, 0.000151069871605122+0.000093886286529958I, -0.0000124154994723444-0.000191221717479616I, -0.000146335713041073+0.0000743994007090742I, 0.000149654854114788+0.0000245390489594559I, 0.000061681999616137+0.000103972535739652I, -0.0000949414244503333+0.0000851800845498646I, 0.0000281795750199606+0.0000506685352350871I, -0.0000528709039390858-5.35614432061353*10^-6I, -0.000082033666550071+0.0000410539019913624I, -0.0000490886184004519-0.000014840259201875I, 0.0000695480155696665+0.0000121387597123306I, 0.0000281346985510657-0.0000184379986146581I, 0.0000738606845786809-0.0000270701764544992I, -0.0000834186273423981-0.000175134836657595I, 0.0000524184142058938+0.000105069417882284I, -0.000128971828436251-0.000145260353962206I, 4.46074972420576*10^-6+0.0000606900432641021I, -0.000032380116046122+0.0000371804631665665I, -0.0000319568397041115+0.0000823460987631508I, -0.0000771611101975181-0.000139476886809049I, 0.000121387588461997+0.0000184655922938187I, 0.0000119269934995924-0.0000425728770507727I, -0.0000454674985443897+0.000122301305429822I, -0.0000441283977218614-0.0000686643380879245I, 0.0000141233298388506-5.55774760045891*10^-6I, -0.0000515540911883161-0.0000142342601094053I, 0.0000180245419643482+0.000153138656529581I, 0.0000377008925138093-0.0000307002687380628I, -0.0000590619447953881+0.0000627860205189951I, 0.0000758114247690221-0.0000671105721606445I, -0.0000190388737117165+0.0000949991485588195I, -0.0000521979895648562-0.000132115670823325I, 0.0000129247614749919+0.00011131267956273I, 9.82303292417234*10^-6-0.000120684475351796I, -0.0000519197475567101+0.0000938084384680425I, -0.0000173803261942276-0.0000669570085502663I, -0.0000473413880701156+0.0000862148004184641I, 0.0000113446111271093-0.000144814466327013I, 0.0000484685777127021+0.0000743320413358524I, -0.0000236542415853767-0.000115927515937311I, -0.0000457208336415631+0.0000698669183269267I, 0.0000529501030841359-0.0000875770562511856I, -0.0000272164835933162+0.000105474138704593I, -0.0000478029775686604-0.000120944832610853I, 0.0000139811080936699-4.80049008797409*10^-6I, 0.0000454104370719227-0.0000152386232216831I, 0.0000331647739086046+0.0000686711833235999I, 0.0000258111679217563-0.000132052360033564I, -0.0000231395259491379+0.0000603616524898267I, -0.000060377880618484+0.0000101458603620551I, 0.0000758367508663423+0.0000993502314809256I, 0.0000408572096450212-0.000102198288323178I, -0.00003400684066333-0.0000396371868404649I, -6.80855067765592*10^-6-0.0000130292060705996I, 0.000109985634979908+0.0000925856509375001I, -0.000114739923910544-0.0000825756334275591I, 0.000103177941141624+0.000120212733313695I, -0.0000620622334104761+0.0000145354343630727I, -0.0000400544059242452+2.85774643983786*10^-6I, -0.0000783923373760792-0.0000159639567325195I, 0.0000691384386240381+0.0000316366684225654I, 0.000052299290670675-0.0000317365576981177I, 0.000103942584149033+0.0000172318705969102I, 0.000012772084548038-0.000118232852240464I, 0.0000681063322361674-0.0000511415667432331I, -0.00004610440406151-0.000199510286780921I, -0.000143607496107753-0.0000288039982323108I, 0.0000779268963344666-0.0000654345341236278I, -4.24487539548017*10^-6+0.000184844279986282I, -0.000134882450274792-0.000204264067757913I, 0.000155209378294718+0.000114237519724679I, 0.0000309046465554448-0.0000199443649913053I, 0.000181759858764788+0.0000606797477386715I, 0.0000833311047622631+6.84294312416808*10^-6I, 0.000128316033209338-0.0000207905715512374I, -0.0000375398608414304-0.000142168479797612I, -0.000011287909091646+0.000124475899485848I, 0.000148472146138755+0.0000381847032897316I, 2.54404473595982*10^-6+0.0000748535058641484I, 0.0000115592910258419-0.000164846945373237I, -0.000109876135872086+0.0000848010735766451I, -0.0000621145972199585-0.000138573419748377I, 0.0000130606557509612-0.0000501619076791593I, -0.000129224747335349-0.0000854765127441736I, -0.0000709068341692966+0.00021647132610616I, -0.0000875224191453992+0.0000378006144534226I, -0.0000802469107511802+0.0000654971304133486I, -8.70062660528966*10^-6-0.000143457393910019I, -0.0000766917854872951-0.0000287188991705342I, -0.0000960420544475059-0.0000650903676399214I, 0.0000626559266143759+0.0000674843738091206I, -0.0000405847280121698-0.000123682996962842I, 0.0000945076131450827+0.000206632496503942I, -0.000027353065739292+0.0000125541457316893I, -0.0000900175462503013+0.000077627340462943I, 0.000113311354180431-0.000171448467198972I, -0.0000190434345609032+0.000107531361307286I, -8.52164724490186*10^-6-0.0000495373018213884I, -8.42546785093636*10^-6+0.00013783242932783I, 0.000160886520767218-0.0000361414223096355I, -0.0000375277035481924-0.0000560746134614251I, -0.0000461568593768318-0.0000935947957370649I, -0.000100216568912849+0.000176054850877592I, 0.0000481571015051763-0.000104721126620679I, -0.0000497676974108073-0.0000608687186562327I, 0.0000303992430269203+0.0000227249484262945I, -3.85082162908631*10^-6-0.000056898718762227I, 0.0001182154373419-0.0000810595598217464I, -0.0000498166364800068+0.00014319524105823I, 0.0000850179606447506-0.0000483353785935888I, -0.0000605805461672812+0.0000553238641929912I, -5.43239893763229*10^-6-0.000135078156699798I, -0.0000679686289793066+0.0000287711902039769I, -0.0000243585395078154-0.000119939162461198I, -0.000146075476385168+0.000122574939191152I, -0.0000793130605435029-0.0000421661593209404I, 5.80247558604222*10^-6+0.000241938968993846I, -0.0000794177925531887-0.000109479128105477I, 0.0000617889693718917+0.0000176319270901142I, -0.0000225398891676005+0.0000171771986513116I, 0.0000100392506531661-0.0000184610819095002I, 0.0000138323792515454-0.000102494711049595I, -0.000115189302663878-4.95777298610975*10^-6I, -0.0000116322297883181-0.000154314910434595I, -0.0000739307183144383+0.0000241945170397529I, -0.0000439684263308515-0.0000205175507400878I, 0.0000201729522441112+0.000100525681422402I, 9.9565269250529*10^-6-0.000144366576403672I, -0.0000264010048042826+0.0000666274512949615I, 2.20810109699585*10^-6-0.0000520859796835957I, 0.0000233520468169243+0.000118480667494458I, -0.0000284166480898641-0.0000655503113524307I, -0.0000325077152820056+0.000119265382209964I, 0.0000258734654478932-0.0000633204693103281I, 0.0000297508364447897+0.0000795088472937527I, 0.0000415389269544944-0.0000714981092786264I, -7.06391381878981*10^-6+0.000114653721679473I, -0.0000307002800444685-0.0000987658665007251I, 0.0000128745866025565+0.0000984626729083271I, -5.87115770078697*10^-6-0.0000649608230964046I, -0.0000460930987063886+0.0000819233050815522I, 0.0000239330243538401-0.0000944328787005715I, 0.0000145536801374245+0.0000525001307736636I, -0.0000267524017097469-0.0000822876063348076I, -0.0000373574806441669+0.0000395661634052926I, 0.000023534704135208-0.000100814111607517I, 0.0000419530235214106+0.0000311400242190926I, -9.05274342869073*10^-6-0.000190835245406068I, -1.62734821391072*10^-6+0.0000694988282720104I, 0.0000451214583615527-0.0000424168406157471I, -4.20755901132328*10^-6+0.0000881298599860334I, 0.0000633419351705039-0.00010006730012245I, -0.0000836427574050875-0.0000183261360196479I, -0.0000526156758584611+0.0000246869503736926I, 0.000013162985192845-9.88498779801931*10^-7I, -0.000141716543301815-0.0000771567734659348I, 0.0000906114727622382+4.47332609136768*10^-6I, 0.00004047247920302-0.0000529861587957316I, -0.0000387109438676518+0.0000421206479156881I, 4.95114862621457*10^-6-0.0000392675640438359I, 8.49687510721145*10^-6+0.0000743260490446737I, -7.58252715330077*10^-6-0.000113788003497306I, 0.0000445441588873817+0.0000504435758903012I, 0.000083069287601256-0.000171178855282638I, -0.000174495966273112+0.000029912149997588I, -0.0000378131015389829-0.000185160105402577I, -0.000122856407636408+0.000225904640803475I, -0.000106707906297834-0.000221266093343698I, 0.000116015962014337+0.0000620441040202857I, -0.0000644472912511122-0.0000464532570367639I, 0.000117570787569303+0.0000883451403375076I, 0.0000137001052104012-0.0000887094667883623I, -0.0000261471607459289-0.000171919621040162I, 0.000092830317011662-0.000197001892950308I, 0.0000430038410356316-0.0000251756739579815I, -0.000138138140646054-0.000144527335228221I, -0.0000648214380341261+0.000102712986085896I, 0.0000524945913301542-0.000132262587617283I, 0.0000382419336392573-7.13042984791663*10^-7I, -0.000104248027477157-7.95787343021612*10^-6I, -0.0000379738467039214+0.000127790974035915I, 0.0000392005436321008-0.000104428241637678I, -0.000128983674309767+0.0000386804820071183I, 0.0000125073298354584+0.0000260289827278649I, 0.0000618639600967267+0.0000741750508285175I, 0.000016675233381489-0.0000393125744689654I, -8.13389105679871*10^-6+0.0000357817862258994I, 0.0000641612691277476-0.0000262559502602943I, -0.0000814497582035378+0.000142519109138404I, 0.0000513575651214456-0.0000626989137585616I, 0.0000746174856463514+0.0000114007207638337I, -0.0000582990120601855-6.21090970906538*10^-6I, 1.37717535650817*10^-6+0.0000449606678200513I, 0.0000930914526757874-0.000077894078145925I, 0.000085050727712901+0.000196493626306992I, -0.000116477233430259-0.0000579888862155696I, -0.0000804308070458153+0.000122777017686895I, -0.000070326397465449+0.000072722814091563I, 0.0000108058926992864+0.0000689613563874882I, -0.0000827658759326694-0.000197764810793909I, 0.0000158095449262972+0.000134298327912272I, -0.0000574523271521731-0.000117030865543484I, 0.0000265455897065917+0.000205855188859087I, 0.000132850262387123-0.0000687513587736454I, -0.0000619860030769496+0.000204582033138943I, 0.0000436540457111027+0.0000351818597477465I, -0.000121102634955291+0.000105027169403515I, -0.0000893792240201054-0.000106187506463777I, -6.09739392892003*10^-6+0.0000752382861604803I, -0.0000461060221387849-0.0000918582209471283I, 0.0000639105970923715+0.0000615898000072529I, -0.0000109045954968951+0.000010607174829783I, 0.0000511302375409812+0.0000319421100628761I, -0.000117662266102354-0.000143568390720237I, 0.000139174572556021+0.000197073365589904I, -0.0000962471306694018-0.000105454436622424I, 0.000112152098755226-9.1833846576665*10^-7I, -0.0000308553795563581+0.000114600682128348I, -0.0000905388524173636+0.0000621811606595896I, -0.0000423742417027431-0.000166074710065421I, 0.0000989372098443167+0.0000267393947925261I, 3.23963613775076*10^-6-0.0000267855704157358I, -0.0000487954563898797+0.000190998901748741I, -0.0000652878561305749-0.0000353215475207374I, 0.0000289826391243923-6.16626743145983*10^-6I, -0.000051595809006598-4.52803875415972*10^-6I, 0.0000221812715671359+0.000165264920481698I, 0.0000717758255932983+9.39319660333828*10^-6I, -0.0000703109155983546+0.0000575736282722896I, 0.0000902020551862877-0.0000835438735728284I, -0.0000528248530725703+0.0000995130861415194I, -0.0000874668107845476-0.000157236358957439I, 3.45951584491833*10^-6+0.0000678182236634312I, -0.0000395131027109779-0.000104184370864128I, -0.0000720199600842475+0.000158219430363459I, -0.0000419603177330884+0.0000196952495703283I, -0.000106741810004891+0.0000861734486708263I, 9.2017763240986*10^-6-0.000195091517578637I, -0.0000170250250246872+0.0000403661609324655I, -0.00008382589086226-0.0000880235176986404I, -0.0000160421311262283+0.0000106447412210983I, 0.0000246620971333507-0.0000796584935462547I, 0.0000886752854909268+0.000169483471548084I, -0.0000265674388115012-0.0000618410840359259I, -0.000014350279022207-5.89428720588625*10^-7I, 0.000112060346970645-0.000111680992856063I, 0.0000590186894829286+0.000140447735448645I, 0.000127846746067146-0.0000788342750889277I, 5.84995631738372*10^-6+0.000115342171974076I, 0.0000877536923533276-5.17249943518992*10^-6I, -0.0000476282059411538-0.000025941792896103I, -0.000014553941137542-0.0000249194569975908I, 0.0000186003511747663+0.0000792748291734559I, -0.0000506040048307758-0.000142985665227492I, 0.0000115399110752994-0.0000199359008189059I, 0.0000314028278997329-0.000102835076487207I, 0.000037584331298103+0.0000319169266543258I, 0.0000812226514989901-0.0000660120460156533I, -0.0000101871319903157+0.000107034643442681I, 0.0000507950150577836-0.0000580108724440496I, -0.0000493404757625538+0.0000712706766934518I, 3.65351534903717*10^-6-0.000128585882313877I, -0.000041232564387639+0.0000675045652477287I, 0.000033176306627402-0.000083730418872363I, -0.000116258471369611+0.0000474688867716474I, -0.0000178121984933358-0.0000764638334729277I, -0.0000530602366366992+0.000182620056926976I, -0.0000464797573697714-0.000159685187795473I, 0.0000703832570926025+0.0000552455013732846I, -0.0000273709462632827-0.0000190278195504422I, 0.000028707495313923+0.0000485495326048227I, 0.000011300037675224-0.0000922585672042985I, -0.0000403636812819942-0.0000562249025230468I, 0.0000400675762375035-0.00015068841814302I, -8.90448273540022*10^-6+0.0000168027985867815I, -0.0000763107082578007-0.0000877392838427039I, -0.0000216932563062715+0.0000937931698537744I, 0.0000348874905646319-0.000120368854262861I, 7.86377715098312*10^-6+0.0000333385811979761I, -0.0000476126433292841-0.0000327563614948371I, -0.0000184056159988595+0.000117376391203871I, 1.42146824624279*10^-6-0.0000979312033560215I, -0.0000830678331344641+0.0000634629106505821I, 7.36229258658543*10^-6-0.000013280226922958I, 0.000032686149123764+0.0000897453064395508I, 0.0000119932051740363-0.0000255914106052369I, -0.0000337404451827016+0.0000607663520804164I, 0.000040369625824878-0.0000830921180327894I, -0.0000565988737676206+0.000116193685695257I, 5.4420909370637*10^-6-0.0000679192596701811I, 0.0000350372904826996+1.32063448664316*10^-6I, -0.0000245688766929566-0.0000225929558494057I, 0.0000425074382365614+0.0000857775726078738I, 0.0000566254039731656-0.0000769331930511828I, 0.0000631876572081178+0.000117011779522862I, -0.0000251046814167134-0.0000753112285420491I, -0.0000158015970848334+0.000154151127655404I, 0.0000565306776873167+0.0000421804029480323I, 0.0000159547084187034+0.0000928924133365412I, -0.0000181610213266692-0.000117084432230345I, -0.0000279722033525644+0.0000476384057615149I, -0.0000638293432578334-0.0000405746013566046I, 0.0000855221774099287+0.000178225096215436I, 0.0000412493636701222-0.000177103033415922I, -0.0000249270217228016+0.000117203351417442I, 0.000123962758723744-0.0000608781943506967I, -0.0000741468949632494+0.000106932391963327I, -1.00515826695763*10^-6-0.0000921188062237602I, 0.0000163791140314254+0.000117026687047645I, 0.0000184487372208978-0.0000886277672209317I, -0.0000213652264441553+0.0000665694759742156I, 9.67679866330881*10^-6-0.0000671186542041082I, -0.0000465887322344132+0.0000682931321850301I, -0.0000616020239959154-0.000033249875008009I, 0.0000588562638653167+0.000102020761650537I, -0.0000198108733355856-6.80029848402448*10^-6I, 0.0000482642411520651+0.00004559895555917I, 0.000042155137398707-0.000017117414677427I, -0.0000290268700198109+0.0000750042445969291I, -8.10136909543721*10^-6-0.0000516991858396281I, -0.0000346082369095551+0.0000193209002738062I, 0.0000118964562009165-0.00007656000403837I, -5.306722130659*10^-6+0.000127287774048955I, -0.0000139959270492481-0.0000540823539115094I, -0.0000456938025561543+0.0000664508220416465I, 0.0000243589243706831-0.0000337699719744151I, 0.0000390140075041181+0.0000762799735234519I, 0.0000292602831777999-0.000097523250206763I, -0.0000186967392484425+0.0000992152426587643I, 9.02470274250812*10^-6-0.000105891517449714I, 7.7905204676426*10^-6+0.0000972059707931465I, -0.0000400049950760344-0.0000476862537391374I, -3.75322016529123*10^-6+0.0000739759030038313I, -0.0000171725437986247-0.0000589269444036311I, 0.0000275189202901623+0.00011331984653064I, 0.0000344732572838813-0.0000439333831677431I, -5.75322577837798*10^-6+0.000118786218563968I, -0.0000517627223387754-0.0000847781265748729I, -0.0000270627963557097+0.0000738074433800462I, -0.0000145195425314062-0.0000316412223079263I, -0.0000189584535659881+0.00007786224953914I, -0.0000132500991439449-0.0000895642152533993I, 0.0000593133233515385+0.0000444182838592708I, 0.0000110079206946967-0.0000429492517275213I, -0.000052115856198786+0.0000960949614763165I, -0.0000157652456661112-0.000155321282169641I, 0.0000206607197863828+0.0000431540001895711I, -0.0000277606539091617-0.000135258801567686I, 0.0000224464175073297+0.0000794355718047432I, 0.0000716569005198835-0.000142789431265532I, -0.0000694451094178044+0.0000914433072128907I, 0.0000249533322402473-0.0000835075268995178I, -0.0000338443751356157+0.000094596937178239I, -0.000018752376977981-0.0000158496129394836I, -0.000097310742422678+0.0000494671704979145I, -0.0000380922865246853-0.0000525741735354331I, -0.0000468379112768886+5.47881897379381*10^-6I, 0.000018111246460035-0.000150159542968383I, 9.46545892838979*10^-6+0.0000854137424370751I, 0.0000194982163292503-0.000114318114039213I, 4.42202353282713*10^-6+0.0001165698028828I, -0.0000628234779513046-0.0000743367279087289I, 0.0000116615567133778+0.0000985269655604068I, -1.91369188790718*10^-6-0.000159325337088686I, -0.0000648509765298964+0.000217340378338968I, -0.0000239206095869635-0.000107028588310203I, 0.000106230613146305+0.000166310623326892I, -0.000150884038079227-0.0000342791533483506I, -6.68280696757674*10^-6-0.0000162547672693273I, 0.0000501236007662996-0.0000768072980026495I, 0.000038713923837085+0.0000711969284274655I, -0.0000259174297000037-0.0000725122309139781I, -0.000157216507553173+0.0000644241899888641I, -0.0000679189523650491-0.000141349258676618I, 1.25565891863955*10^-6+0.0000500822608854054I, -0.0000771984308689629-0.0000336651645139884I, -0.0000279124975137377+0.000125545797776011I, -0.0000346770188841467-0.0000580666409531003I, -0.0000327089785916282+0.0000194372264691702I, 0.0000344842653580842+7.25190345338111*10^-6I, -4.41135627119442*10^-6+0.0000818721962777101I, 1.83997679778558*10^-6-0.000140546349256203I, -0.0000316915661815622+0.000117125117130981I, 0.0000403209938448596-0.0000792612457414742I, -0.0000332925663507494+0.0000482487979267768I, -0.0000206093651722694-0.0000576634454519225I, -0.0000493987587436386+0.0000417075056738128I, 0.0000652377769447056-0.000119651353560115I, 0.0000105848251842804+0.000102875539685285I, -7.15392910483735*10^-6-0.000111311516105096I, 3.01624564884493*10^-6+5.99558104130296*10^-6I, 0.0000197963722614685-0.0000307669925346918I, 2.10212555399034*10^-6+0.00011439412975581I, 0.0000134884047923412-0.000105730944849212I, 0.0000730988375815877+0.0000451559949336325I, 0.0000216729081001154-0.0000276550317493973I, 9.04843940563021*10^-6+0.000149622245238873I, 0.0000896551120252851+8.39520577115834*10^-6I, 8.00854614478746*10^-7+0.0000847712327363354I, -0.0000329702461186305-0.000049672436915037I, 4.13194197400231*10^-6+0.0000381160130804657I, -0.000040466028669989-0.000035811503069265I, 0.000074161921080212+0.000109003828231267I, 9.64494036942354*10^-6-0.000157506486776166I, 0.0000230304084793087+0.0000949622109518653I, 0.0000781400318837947-0.0000863411111810633I, -0.0000209402824369232+0.000110527050349149I, -1.06781805727943*10^-6-0.0000678572004044087I, 0.0000101403931127114+0.0000929368507419286I, 6.67495729174901*10^-6-0.0000764066432803755I, -0.0000127890467005032+0.0000656148298827164I, 9.19290802058889*10^-6-0.0000691198719429689I, -0.0000226425666661326+0.0000717685856267078I, -0.0000272294442208769-0.0000496855265433974I, 0.0000353193632734858+0.0000814401137482077I, -5.62377608697246*10^-6-0.0000461009835010315I, 0.0000264860329342651+0.0000521362563384535I, 0.0000112774659818628-0.0000380528024535422I, -0.0000200807718378518+0.0000702511612084438I, 3.54891494943462*10^-6-0.0000769312427003555I, -6.47155428380363*10^-6+0.0000580127213870977I, -1.18785392723016*10^-6-0.0000679523876684713I, -0.0000157165400378614+0.000107552132842338I, -0.0000207239279245277-0.0000646259023746607I, -2.90859988261232*10^-6+0.0000681668796096927I, -5.23912149687832*10^-6-0.0000553038252474898I, 4.70185873305238*10^-6+0.0000881009647317532I, 4.91619888946831*10^-6-0.0000526984334432593I, -0.0000147531596496827+0.0000678131705817472I, 0.0000219465458907899-0.0000682328318738814I, -0.0000112791296637666+0.0000727876242238124I, -0.0000123851324664895-0.0000995037766652229I, 2.68610649228769*10^-6+0.0000704312633786581I, -9.34147192718295*10^-6-0.0000831174893881667I, -0.0000275664748881019+0.0000868143223015996I, -0.0000271848208430733-0.0000493447109164319I, -0.0000303675475953409+0.0000582854105546505I, 0.0000238430167651181-0.000101446811094862I, -8.46877134365314*10^-6+0.0000564813451794988I, -0.0000240296609373648-0.0000880486060832047I, 0.0000207933941758824+0.0000475379158322647I, -5.52303945499426*10^-6-0.0000663680355885669I, 0.0000164551654467041+0.000127579372826302I, 1.79515124966409*10^-7-0.0000595009725692858I, 0.0000118202901824194+0.00006788128030048I, 0.0000396736173933431-0.0000705233955002236I, -0.0000186454681485222+0.000121918597846957I, 0.0000254521352566794+8.402420426686*10^-6I, -6.78802419466742*10^-6+0.0000990728572143329I, 0.0000281436921038407-0.0000505057118761494I, 2.83854976130542*10^-6+2.05307561782021*10^-6I, -0.0000619586091493868-0.0000669473081969934I, 1.47498743573045*10^-6+0.000164163891182622I, 0.0000661781547030135-0.000149326402203927I, -9.8675038608817*10^-6+0.0000454086412560824I, 0.0000960881080752117-1.84569144479193*10^-6I, -0.0000412937383078873+0.0000413339447810154I, 0.0000476736775889831-0.0000668749788620052I, -0.0000221036201547863+0.000129203126008448I, 0.0000503716182929365-0.0000567990122987465I, -0.0000410348997049947+0.0000416955147097813I, 0.0000132420421380758-0.0000888077195379466I, -0.0000702490836195977+0.0000281702407861663I, -0.0000786510137902321-0.0000602405502292761I, -0.0000159829664938641+0.000119200167638713I, -0.0000728754221335168+0.000012276414190573I, 0.0000546192934296861+0.000132721866744576I, -0.0000111101784172096-0.0000182634945056995I, 8.11734478507486*10^-7+0.0000316132727589606I, -0.0000154520064384933+7.76546467227427*10^-6I, -0.0000285054413686309-0.0000489426008755106I, 0.0000163016228730082-0.0000888941786260869I, -0.000107217130584322+0.0000909705838808485I, -0.0000476220481556659-0.000107836287150175I, -0.000105913096386826+0.0000323166830464866I, -9.88812163503781*10^-6+0.0000324653691035932I, 0.0000559123093385885+0.0000986589528304548I, 0.0000137243492931901-0.000142439752229703I, -0.0000537090220742161+0.0000818863240104752I, 0.0000337851602441925-0.0000575968266340964I, 0.0000377122600897718+0.000124553668137678I, -0.0000634294146810029-0.0000514829424767452I, -0.0000309211868700323+0.000138964426093776I, 0.0000357547556115839-0.0000547049957260397I, 0.0000412839837770481+0.0000879290228753483I, 0.0000733258207671418-0.0000424748522806762I, -0.0000284093798859102+0.000140943455927357I, -0.0000480201610948752-0.000126432478280683I, 5.71703532048947*10^-6+0.000120140490775453I, -0.0000146280128202201-0.0000416778932178923I, -0.0000789771170789961+0.0000339781241620649I, 0.000036327150309886-0.0000620803017909639I, 0.0000698943530830547+0.000024031426541847I, 4.32183457275186*10^-6-0.0000925021402146186I, -7.01153274245532*10^-6+0.000024980819070678I, 6.6420310201907*10^-6-0.000120462629687825I, 0.0000925480098540137+0.0000729432839517042I, 0.000102589453262785-0.000165737032602445I, 0.0000339103261979614+0.0000703426039362884I, 0.0000393734583331904-0.0000962591570774808I, -0.0000889593841679743+0.0000864389584359087I, 0.0000437260712549546-0.0000108080994295361I, 0.0000696831817092263+0.0000192909340976764I, -0.000116004294429193-0.0000838152806094251I, -0.0000411765244009639+0.0000524361199085674I, 1.20132495209512*10^-6-0.000174086154796747I, -7.51362089024038*10^-6+0.0000739354700398435I, 0.0000206556097096513-0.000137447459327382I, 0.0000493657843640968+0.0000796586254949416I, 0.0000203828315658041-0.00012621494880542I, -0.0000218451291651558+0.000130808096256256I, -0.0000491568467274153-0.0000944543373377137I, -0.0000175889591867749+0.000145920650626241I, 0.0000407923580382674-0.000070949821534546I, -0.000040604849866072+0.000146467713948587I, 0.0000605833868711353-0.0000579978975237211I, 0.0000834665594047132+0.000111974629779801I, -0.0000755932496929144-0.0000752004309677099I, -0.0000133665747855175+9.73865969930288*10^-6I, 0.0000986923147396717-0.0000580203428362928I, -0.0000574372814882281+0.000108549621855779I, -0.0000350900897123494-0.0000866204641644278I, -0.000102139708124509+0.000114016778724124I, -0.0000716120508536669-0.0000999915461438805I, -0.0000190269917696018+0.000139010150514063I, 0.0000127840472206616-0.0000461072067168347I, -0.0000150540693520809+0.0000652250340576313I, -0.0000990285941007478-0.000064392479303498I, -0.0000117635550434173+0.0000618180267056644I, 0.0000401868268994047-0.0000195974005726883I, 0.0000387880846702578+0.0000367481746151179I, 0.0000332992949463784-0.0000860995825103844I, 0.0000381356417567259+0.000136803424184934I, 0.0000542870147757309-0.000123907921310369I, -0.0000361876302241568-1.51541301823272*10^-6I, -0.0000169068857250556-0.000136635797392157I, 0.0000170209981967661+0.0000489268930878892I, 0.0000277582667491506-0.0000750671707229044I, 0.0000754410741514004+0.0000733531805065972I, 0.0000108796356364821-0.000118623754383125I, -0.0000266355863231614+0.0000954277210457698I, 0.0000403511156733557-0.0000823051323025051I, -0.0000720754287024353+0.0000860903112672028I, -0.0000461930162785144-0.000103904900665596I, 0.0000104886106694245+0.0000125589767697964I, 0.0000378285256645123-8.57128316792257*10^-6I, 2.5457151890506*10^-6+0.0000503266816281619I, -0.0000115735196035364-0.000103235601253693I, -0.0000276588627184714+0.0000598790354523735I, -0.000017852972613378+0.0000113539243379457I, 0.0000654682875552274+0.0000614448878012638I, 0.0000197391413628765-0.000100285820109098I, -0.0000676494807624176+5.98210542474736*10^-6I, 0.0000104433481556591-0.0000234260239266675I, 0.0000609098280801478+0.0000250007674317092I, -0.0000750926738183358-0.0000349231423577125I, 0.0000746068295574368+0.0000340003389291176I, 0.0000269026563057004-0.0000173897620381623I, -0.000047626746926551+0.0000526123483828046I, 6.81552573910379*10^-6-0.0000211883963188705I, 1.93738777966838*10^-6+0.0000392692066964512I, 0.0000142642790779219-0.0000922421304569109I, 0.0000138614832549249+0.0000203497797053321I, 0.0000181176672756787-0.000128335912885531I, -0.000106583979869604+0.0000418522704245353I, -0.0000619990474393177-0.000113347803762274I, -0.0000636973206580779+0.000182403275597903I, -0.0000778492725171866-0.000131185675675687I, 0.0000711979640739277+0.0000430160757352393I, -0.0000445733916913595-0.0000221710035015237I, 0.0000706165765555611+0.0000250097581655974I, 0.00001051853778371-0.0000834590496267099I, -0.0000906075871741968-0.0000713127551356557I, 0.0000143905902243352-0.000173382600647563I, -0.0000219216894101456-0.0000104652393928279I, -0.00010450293692166-0.0000483191707744675I, -0.0000210805178247642+0.00011082524641947I, 0.0000103109907662757-0.000130203426335796I, -0.0000148347732505328+0.0000120495859485087I, -0.00002400286065344+0.0000155067366867737I, 4.04926417777838*10^-6+0.000120252354274127I, 2.15462108666526*10^-6-0.000103951336632751I, -0.0000912615545279905+0.000116845013012656I, 0.0000622272965748513-0.0000216424648628903I, 0.0000323645298429798+0.0000393997934307372I, 0.000033618086411533-0.000054900938644634I, -0.0000327422150679366+0.00005893466528959I, 0.0000578257804985335-0.0000919685745217589I, -2.35511380484742*10^-6+0.000153083315067197I, 0.0000279119750425751-0.0000920018586154601I, -0.0000135568643373178-0.0000133034068595473I, 0.0000241085711822616-4.07464043786042*10^-7I, -6.21796244680437*10^-6+0.0000526646395126228I, 0.0000386713141492482-0.000133552708095781I, 0.000111544970429832+0.0000552721983947969I, -0.0000289112688798296+3.12727781565462*10^-6I, 0.0000190322381972993+0.000143158881444751I, 0.000100498079487645+7.09533925947325*10^-6I, 6.75793990601859*10^-6+0.0000599187121708152I, -0.0000986283586047334-0.0000607127361288369I, 0.0000202303366287212+0.0000845405771867999I, -0.0000265409718302702-0.0000274221545549223I, 0.00011262416208581+0.0000681223823460688I, 5.46687090252955*10^-7-0.000148338122105052I, 0.0000574148026131909+0.000142441282281457I, 0.0000655509231486223-0.000108814854025177I, -6.05528904606383*10^-6+0.000158697958811576I, -0.0000408051698395517-0.0000393242504348979I, 0.0000134016479412145+0.000064543459610827I, -0.0000232052542857201-0.0000613301435541929I, 7.83917879219553*10^-6+0.0000494607081669629I, 0.00003149244187746-0.0000492404495329934I, -2.7662494421373*10^-6+0.0000562166232523965I, -0.0000324178770855781-0.0000311348931582849I, 0.0000748104677500698+0.0000297643766683154I, -0.0000141256141773799-0.0000467946495212581I, -0.0000119553737583203+0.0000105758317601142I, 0.0000542353368987406-0.0000333995907869284I, -0.0000189815887564269+0.0000968530306241101I, -0.0000304201783705575-0.0000713402652558146I, 2.57501330552167*10^-6+0.0000348223714175883I, 0.0000129801542705706-0.0000642173120745168I, 8.37084217176115*10^-6+0.0000955059015326414I, -7.81821052677411*10^-6-0.0000605244387579226I, -0.0000251124381301889+0.0000428257785172144I, -5.34309781284023*10^-6-0.0000299525191214028I, 0.0000257681045865358+0.0000873528020565381I, 0.0000231622096388731-0.0000875407377391407I, -0.0000272406466394877+0.0000729376123819678I, 0.0000214774244136732-0.0000643703075916135I, 8.02041721932066*10^-6+0.0000972304023906084I, -0.0000373871826793773-0.0000717241938886066I, -0.0000192692591991643+0.000101713667554524I, 0.0000220501111195501-0.0000623974131300725I, 0.0000131285334899009+0.0000783868585777103I, 0.0000350076259451226-0.0000508414566768474I, -0.0000269083530692427+0.000100694299790493I, -9.78579390606961*10^-6-0.000112386765206814I, 0.0000135350928344166+0.000106190459604976I, -5.71778201148695*10^-6-0.0000688748990232251I, -0.0000531722431302354+0.0000322388387313446I, 0.0000396951579933452-0.0000500966577201111I, 0.0000229693684567271+0.0000467846306652904I, -3.36264179364388*10^-6-0.000111509326353753I, 0.0000311610101019262+0.0000135664811205817I, 0.0000114115352893378-0.0000494903061854004I, 0.0000675890480419837+0.000086371522943683I, 0.00010135249907703-0.00011992189932329I, 0.0000115943179528919+0.00005767918486334I, -0.0000340173019567011-0.0000367653546371469I, -0.0000128773547692952+0.0000876595304821127I, 0.0000363819734309251-0.0000231173057304935I, 0.0000686318015308219-0.0000225306468170279I, -0.0000863407649069974-0.0000820170034943239I, 0.0000562426513095246+0.0000787507853861394I, -0.0000294778625881266-0.000165486338646479I, 0.0000596181314173364+0.000123776649489408I, -0.0000177781600800791-0.0000424129621900422I, 0.000016411695766465+0.0000350508077871736I, -0.0000265871558467024-0.0000602538157592545I, 0.0000102216086946823+0.0000730827699024144I, 0.0000168402146329918-0.0000734109215963706I, 0.0000296311635510469+0.0000826634456896993I, 0.0000541842479265133-0.0000580331960739348I, -7.09079449846316*10^-6-0.0000171037521942427I, 0.0000253125752461737-0.000116434093002691I, -0.0000886740722051079+0.0000485938579040595I, 0.0000233544896118449-0.000133484628921525I, 0.0000320398704703715+0.000109203687390231I, -0.000039577356925809-0.000088779074045516I, 0.0000455561082855827+0.000103474548670126I, 0.0000103597342177919-0.0000632771906015966I, 0.0000609101865500499-0.0000144701200541096I, 0.0000592456259389254-0.0000854022842271306I, 0.0000433485717275493+0.000027193144757952I, -0.0000494894180017882-0.000113463901458196I, -0.0000260714751832622+0.0000784602666065394I, 0.0000479418872407647-0.0000744337205716025I, 0.0000179221510152328+0.0000397174457258522I, -0.0000427109036387217-0.0000533926702814348I, -0.0000327447983019936+0.0000954096615083932I, 1.26102585351566*10^-7-0.0000913772188616197I, -0.0000630225995016085+0.0000377800887647091I, -6.52237880651537*10^-6-0.0000162325138234695I, 0.0000228625295588072+0.0000918974844099744I, 7.87254779215579*10^-6-0.0000200258753716147I, -0.000031759914195933+0.0000566647230358483I, 0.000032756995105266-0.0000779839501580361I, -0.0000522414374039247+0.000106439018691786I, 6.1828435779859*10^-6-0.0000591719941007718I, 0.0000235076061261972-1.81699237650672*10^-6I, -0.0000152786513792734-0.0000156780768461671I, 0.0000428289995841802+0.0000642467624695843I, 0.000056112175618986-0.0000839795725958476I, 0.000067835589299406+0.0000986470607918016I, -0.0000369151168696516-0.0000641102019815738I, 8.52236844383427*10^-6+0.00013976361408542I, 0.0000828352157259203+0.0000117558270762411I, 0.0000242838387612108+0.0000737571276396455I, -0.0000445365972460064-0.000119655175834297I, -0.0000399376806023985+0.0000774808468903186I, -0.0000307461837545251-0.0000138592658081497I, 0.000126823348490296+0.000114629870202543I, -0.0000188578802098083-0.000153767634818504I, -0.0000142247284920738+0.000143019821778556I, 0.0000906121165373711-0.000129768389134238I, -0.0000590148028047036+0.000146952740311906I, -0.0000429011353286111-0.000100966957296591I, 0.0000489755073976103+0.0000764254057372783I, -0.0000193950591614907-0.00010937184806405I, 6.08875156213474*10^-6+0.0000918147307701886I, -7.62358801686824*10^-6-0.0000405010749735111I, -5.34500896609013*10^-7+0.000110296679742908I, -0.0000171230140526406-0.0000293529702054583I, 0.000102290008196707+0.0000943912581506481I, 0.000040844404230696-0.0000535980064402823I, 0.0000588907853747033-0.0000250507019399936I, 0.0000298541758376742-2.66832615327631*10^-6I, -0.0000615330214577625+0.0000786840446220965I, 0.000032974094798696-0.000124644401016576I, -0.000016110813594165+0.000105331691830599I, -0.000014879541481986-0.0000491949694843828I, 0.0000176620080069994+0.000164733145759882I, -0.0000354975129726576-0.0000222073046101595I, 0.0000346164736523247+0.0000941636166696913I, 0.0000120812658886943-0.0000655949834033235I, -5.44915605447178*10^-6+0.000082968649712698I, -7.8642980403493*10^-6-1.42921939982406*10^-6I, -0.000014938606046603+0.0000639809604746937I, 0.0000450286868453214-0.0000659862327670602I, -0.0000232807736221073+0.0000466596255027419I, -0.0000101334744070735-0.000119914462618847I, 0.0000250931924240008+0.0000620520096158674I, -0.0000126351385952524-0.000102515227271393I, -0.0000632155554816278+0.0000806167458252547I, -0.0000539720112503162-0.0000409469373615398I, -0.0000511098134053259+0.0000373872058826943I, 0.0000483395230901972-0.000126558585723738I, 6.37595255856411*10^-6+0.0000453812275982889I, -0.0000395647241662324-0.000108154675570385I, 0.0000141863657039929+0.0000175032251134646I, 0.0000189002461189113-0.0000475224094167179I, 0.0000181951055049605+0.00015619128011763I, -9.15902257301258*10^-6-0.0000831163585993503I, 0.0000510581480249122+0.000013123112455905I, 0.0000807105451848498-0.000027562150872086I, 0.0000191892043378155+0.000169459368043803I, 0.000148112266522402+0.0000329914520456553I, -4.82226859931528*10^-6+0.000100077706333136I, -5.05480385310075*10^-6+0.0000156752194291915I, -7.35773222356917*10^-7-0.0000347503395278143I, -0.0000674959106769628+0.0000125685058545444I, 0.000115110787764206+0.000110567574780758I, -0.0000201312070668734-0.000252448735583317I, 0.0000707184214812284+0.0000485628603046172I, 0.000164564071757363-0.000135794924935589I, 9.01643606914215*10^-6+0.000132877348523245I, 0.0000464943583899602-0.0000280871760078845I, 0.0000218191850527775+0.000119423329799772I, 0.0000422034608810799-0.0000568189106758817I, -0.0000716509778550105+0.000044865812474126I, 0.0000508499929144364-0.000109913493503834I, -0.000098905627536598+0.0000802090264631418I, -5.72122950981932*10^-6+0.0000752355362217837I, 0.00003526264845751-0.0000382310054634863I, 0.0000562363730283234+0.0000542954495021866I, -0.0000436814413335294+0.0000178789488235668I, 0.000142092747433592-0.000114669186033326I, 6.22212059843284*10^-6+0.000116335784773342I, 0.0000122690780150674+6.42983342260776*10^-7I, -0.000129160399109939+0.0000453386243586352I, 0.000014185320968872-0.000095691398088028I, 0.0000844419797607586+0.000104845502432698I, 0.000026204794156526-0.0000249172228763534I, -0.000062246210945778+0.00013917455618502I, 0.000102133338048004-0.0000738841537486345I, 0.0000394858239491723-1.7836164017098*10^-6I, -0.0000229284034050482-0.000118441515185027I, 0.0000254586268607827+0.00012782800717802I, -0.0000246223355216357-0.000120701690227345I, 0.0000422664346495013+0.0000529388496740465I, 0.0000142893062600246+0.000010182890251026I, 0.0000380430708771274+0.0000690897368426049I, -2.74268321162521*10^-6-0.0000729334857781728I, 0.0000376816805929758+0.0000514147632475594I, 0.0000293583779647007-0.000115178901115735I, 0.0000625757269088127+0.0000945001518137607I, -0.0000496555120852961-0.0000163406317825378I, 0.0000166331763383151+0.0000550235958835303I, 0.0000206183906822584-0.0000480339433062674I, -9.77248062429946*10^-6+0.000133085982841558I, -8.65181356486806*10^-6-0.000100139892795628I, -0.0000308177694193092+0.0000293295836029456I, -0.0000160245999148259-0.0000518728841344456I, -0.0000461523828925627+0.0000862647584992389I, -0.0000171837124919485-0.0000782556270975957I, -0.0000215230948671049+7.95767627898421*10^-6I, -0.0000836596987965933-0.000107014872763998I, -0.0000108114829480967+0.0000572575479098398I, 0.0000118238792308141-0.0000678693105939389I, 0.000024928759726873+0.0000832278319534666I, 0.000010895817947336-0.000109274728384978I, -0.0000736533421992728+0.0000612096668282557I, 0.0000353766605099204-0.000016897596222694I, -3.13293123519809*10^-6+0.0000452844336117079I, -0.0000370463324271101-0.0000106519945804029I, 0.0000128760541723609+0.0000257017133168739I, 0.0000179201414281142-0.000049684747948562I, -0.0000310544702007945+0.0000691218564946865I, 0.0000137232838784717-0.000041628754073246I, -3.86751626483272*10^-6+0.0000456608158177251I, 8.76953906653515*10^-6-0.0000791702428131557I, -0.0000104932748630834+0.0000296811935214316I, -0.0000152406506688392-0.0000876715196462989I, -0.0000531035658506397+0.0000644780481632704I, -0.0000470119432202388-0.0000544224504302797I, -0.0000203834700205814+0.000134853346795433I, -0.0000230271993853017-0.0000914313847431409I, 0.0000388096104759908+0.0000487872610134933I, -0.0000262861782851222-4.69853231515405*10^-6I, 4.98483077955727*10^-6+5.91457531536425*10^-6I, 0.0000140524129213638-0.0000866868783944143I, -0.0000566185327393667-4.00954793839549*10^-6I, 8.01080096999738*10^-6-0.000126429323483622I, -0.000059495870610711+0.0000315867380340799I, -0.0000304339208248676-0.0000334511076455426I, 0.0000133423588337428+0.0000700069148750493I, -1.7685672674246*10^-6-0.000150312312065871I, -7.43137001287117*10^-6+0.0000559467090150525I, -0.0000291990950835773-0.0000280404401681401I, 0.0000344735267058961+0.000106533569266517I, 2.35924923189354*10^-6-0.0000380802677602621I, -0.0000571035960824348+0.00011593408021943I, 0.0000577347698543525-0.0000219448958674396I, 0.0000668061603549272+0.0000298858858705699I, 0.0000697514277252085-0.0000882153173789681I, 0.0000210549535928514+0.0000894733836565607I, 1.73558601059978*10^-6-0.0000412421801198772I, 8.8837870421762*10^-6+0.000146951339603839I, 0.000054268541488714-0.0000590552105961893I, -0.0000325006438572786+0.0000430310729374369I, 0.0000153669169729105-0.0000223815754838465I, -0.0000304057938323921-0.0000167742085944585I, 0.0000375118659194132-0.000123310328381909I, 0.0000730375669015478+0.0000836644936044756I, -0.0000867836061279208-5.7901127657744*10^-6I, 4.61346202923516*10^-6+0.0000672431423275118I, 6.28348028923175*10^-6-0.000046486317402495I, 6.4275495631701*10^-6+0.0000291934307481742I, -0.000131777124414157-0.000103116167047396I, 0.0000406143563539219+0.000156668197341652I, 6.08069252278919*10^-6-0.0000679104555967431I, 0.0000719964651078641+0.0000332068476038973I, 0.0000255528020668529-0.0000627885177441624I, 0.0000512637191526618+0.000193076292051441I, 3.751238976999*10^-6-0.0000801802791723199I, -0.0000119837842428141+0.00017714778398982I, -0.000108558905098281-0.0000256564533326468I, 7.88263285923424*10^-6+0.0000215154139511337I, -0.0000773353447337441-0.0000516260114509932I, 0.0000612941060949974+0.0000345423850832897I, 0.0000468759762791454+6.25159466397576*10^-6I, 0.0000542750873095611+0.0000313175196065041I, -0.0000637923135947905-0.0000325457543609285I, 0.00018120752675838-2.98690280507071*10^-7I, -0.0000336327423924911-0.0000723087925691968I, -0.0000340857111529037-0.000107118300351335I, 0.000140685262776215+0.0000374889638889819I, -0.0000654326790392357+0.000158212285960451I, -0.0000929946684406456-0.000138792721919342I, 0.0000414543520637955+0.0000323254269666025I, 0.0000348488395020739-0.0000274937534966418I, 0.000137101576654722+0.000164092918155448I, 0.0000262297451850206+0.0000385439830843071I, 9.56154397158054*10^-6+0.0000129787421903967I, 0.0000399822899187481-0.0000360339870927869I, 0.0000622007024917594+0.0000845645241526382I, 0.000115729058556318-0.0000427954896661394I, -0.0000306710297397235+0.00011576248740822I, 0.0000303919309585921-0.000164934991289064I, -0.0000275526274208363+0.000102961675853985I, -0.000100658546215705-0.0000357391816395522I, 0.0000105476958074353+0.0000278014657709557I, -0.0000640848919986475-0.0000520852005373423I, 0.0000233206765822519+0.000175953628026502I, 0.0000491106407847182+0.0000143402020120028I, -0.0000354536310320883+0.000156515431462263I, -0.000106301843539205-0.00012068402485066I, -0.0000412689222628394+0.000050169296026273I, -0.0000486743394014233+7.11297177253293*10^-6I, -0.0000648471672638107+0.0000649946323238936I, 8.68131790347544*10^-6-0.0000987075131101412I, 0.000119652854133857+0.0000170378679209042I, -4.75861323798013*10^-6-0.0000376825672921255I, -0.0000887960484615811+0.0000414233219421039I, 4.60020188946109*10^-6-0.000187360843910563I, 0.0000996069600575636+0.0000130846593850656I, 0.0000457098448032473-0.000242503321358868I, 0.0000427477867918726+0.0000619880630669875I, 0.0000850433954165145-0.000124304153814815I, -0.000117833164926371+0.000112856914101777I, 0.0000929938238346917-0.0000306273515992187I, 0.0000143085369336179-0.0000314547124071498I, -0.000146519763855607+0.0000179947080189315I, -0.0000686000532307331+0.0000390150239833172I, -0.000104252942816431-0.000175890062885785I, 0.0000176027534721935+0.0000481898601578779I, -0.0000135507533978934-0.000137082086840899I, 0.0000455546533543809+0.0000262766877545717I, -0.0000254761723649867-0.000117914459814954I, 0.0000282590225418461+0.000136485949382721I, -0.0000565814996949541-0.0000585003227141707I, 0.0000662937404907038+0.000137672282315733I, 0.0000772347131452678-0.000123770671352403I, -0.0000248503883314833+0.000108732971725492I, 0.0000512801833291827-0.000163557818539327I, 0.0000145725794043551+0.0000577815206381754I, -0.0000718040013566674-0.0000746179036038863I, -0.0000118324061696534+0.0000548896711735267I, 0.000021390590035946-0.000151349755131985I, 0.0000684785474804962+0.000157289111251023I, -0.000019885472496097-0.0000381449992076184I, 0.0000139342304109005+0.0000581031028395911I, 3.01922465175816*10^-6-0.0000563773777111199I, 0.0000949572712049732+0.0000455915820262553I, -0.0000524720985502293-0.00010579239288054I, -0.0000433691454563036+0.000101825568024687I, 0.0000336487347547139+0.0000165341647287171I, -1.93618085372769*10^-6+0.0000284578096844507I, 0.0000177495095007453-0.0000580636512419356I, -0.0000607119198777368+0.000061106639525266I, -0.0000203839914942715-0.000137026750479629I, -0.0000128469441141754+0.0000190817622172216I, -0.000042837262162412-0.0000656775115702495I, -0.0000526519568159883+0.000121551198027367I, -0.0000499638365749604+0.0000149930486157525I, -0.0000770505740338464+0.0000434504553805995I, 0.000030178957336867-0.000136863859185297I, -0.0000458287373793072+0.0000408066834515752I, -0.0000589639449875332-0.0000673748888074584I, 0.0000200882537829263-7.39367255895845*10^-6I, -2.0785312434999*10^-6-0.0000468445680328714I, 0.0000933258772975016+0.000139174442920512I, 0.0000123162042571011-0.0000479402390017421I, 0.0000100046337247254+0.000034253193279025I, 0.0000687200866167987-0.000115788095846695I, 0.0000545332652861182+0.000151780476140544I, 0.000158651651520513-0.0000380904246773888I, 0.0000279089913970135+0.000105072471383003I, 0.0000765990368704211-0.0000554720519729656I, -0.000101223651516498-0.0000168969869534976I, -0.0000195421628629885+0.0000359626517870311I, 0.000118005427768235+0.0000926392484061424I, -0.000111319566558254-0.000192922002938859I, -0.0000292614184060669+0.0000157338347320957I, 0.00010496741698973-0.000194548003829887I, -0.0000160946327497485+0.0000815389693240898I, 0.0000611527531165151-0.000122117593883901I, 0.000064057304209512+0.000109644479541158I, 0.000050144155392861-0.000120283287241081I, -0.0000665701325898119+0.000121821937813734I, -0.0000297590655995004-0.000118758642950444I, -0.0000669929279688804+0.000160198328965536I, 0.0000682302252145954+7.93391271908314*10^-6I, -0.0000427548983554544+0.0000707165682805045I, 0.000115355211490802-0.0000140426431428202I, 0.0000317449238932632+0.0000813215149702645I, -2.67865470367887*10^-6-0.000137172005311445I, 6.67731215362698*10^-6+0.0000429236187314293I, 0.0000994238780429196-0.0000371953431735677I, -0.000104358627371942+0.000136818481993198I, -0.0000282665443285796-0.0000900825728486255I, -2.18675666970461*10^-6+0.0000765591677879277I, -8.702984792968*10^-6-0.0000698377329908694I, 2.61513674312285*10^-6+0.000162658154467554I, 0.0000448063407735377-0.000107215946493693I, -0.0000269550175818961+0.000015210211880193I, -0.0000787305572976689-0.000065362887546186I, 0.0000353004817350642+0.0000698768471987692I, -0.0000200568526344789-0.0000490934811666255I, 0.0000220532045260018+0.0000230943066184437I, 0.0000586690817569653-0.0000531513662685041I, 0.0000253289521387174+0.0000696612985163301I, 0.0000170546973868512-0.0000794963874131263I, -6.40996729329931*10^-6+0.0000138325137344049I, -0.0000178476589411469-0.00011324385759506I, 0.0000369850745583148+0.0000326268726015751I, 0.0000241719393803038-0.0000251327658582788I, 0.0000169582182085863+0.0000628544319102074I, 0.0000237592308093424-0.0000811801458356696I, 0.0000222708002898674+0.000083056629612418I, -7.66986040043179*10^-6-0.000059327590601203I, -0.0000410244936123631+0.0000635045595865723I, 6.00412612022495*10^-6-0.0000710465041643092I, 0.0000196400854435026+0.0000823442523855453I, -0.0000188252089068116-0.0000391710581173756I, -0.0000241252358656173+0.0000686152229305665I, -0.0000127971862508683-0.0000328960914947487I, -3.86513013222326*10^-6+0.0000598140948335338I, -0.0000286689394684716-0.0000732062086195658I, 0.0000127898243597857+0.0000748721615211637I, -8.91180841612757*10^-6-0.0000605510811179701I, 0.0000241057327144033+0.0000742686657947261I, 1.78958286954218*10^-6-0.0000816297262190204I, 3.69109838397354*10^-6+0.0000938664424144548I, 0.0000216788629256526-0.0000857651298946303I, -0.0000140438978423473+0.0000957598203930676I, -0.0000213484860438095-0.0000727150108888175I, 0.000019081190181501+0.000060979899006812I, -0.0000130924431340535-0.0000815547822922517I, 5.27279863257972*10^-6+0.0000763899569222673I, -3.3068593155617*10^-6-0.0000532476804980881I, 4.86158572946248*10^-6+0.0000914121950449834I, 8.98251702457865*10^-6-0.0000376411755562446I, 0.0000486565640116078+0.0000612602922630421I, 0.0000441519018682941-0.0000620545361117685I, 0.0000158177750420305+5.36037029977153*10^-6I, 0.0000303927266998603-0.0000514554584520287I, -0.0000302924474264415+0.0000810885167920742I, 0.0000295859681979498-0.000102749035372267I, -0.0000246842643662108+0.000114064841645794I, -0.0000121934293355656-0.0000550744837086789I, 0.0000492815371096872+0.000121099110544601I, -4.04533503643489*10^-6-0.000018666586025114I, 0.0000370969884704505+0.000109643604875299I, 0.0000373844510619621-0.0000956436929501094I, -0.0000106212129752702+0.0000469926056602597I, -0.0000160280294720445-0.0000133740502511483I, 0.0000148477461539628+0.0000770632809816582I, 0.0000107717402747356-0.0000893596063507433I, -0.0000142667856257783+0.0000280663289210064I, 0.0000154765095791503-0.0000733223029227452I, 0.0000434853143855166+0.0000337288504693337I, -0.0000303275626006944-0.000094153986964692I, -0.0000389000673233259+0.0000705727670413868I, -0.0000450907578893189-0.0000690594218879603I, 3.06984382811272*10^-6+0.0000408768948855473I, 6.73313687460418*10^-6-0.0000609134909141207I, 2.35447030580246*10^-7+0.0000177081202607656I, -0.0000191755724648451-0.0000741144987720184I, 0.0000284415600272538+0.0000855643808082622I, -0.0000130796676281831-0.0000851286875943525I, -3.00458973007558*10^-6+0.000108213974893124I, -0.000016082905309762-0.0000366963771000744I, -0.000027837922101207+0.0000656471103138238I, 0.0000343396297739752-0.0000792834846583533I, -0.0000189187223594914+0.0000669325969284845I, -0.0000141908541756873-0.0000485093093597352I, -7.3497250491763*10^-6+0.0000804566633066585I, 0.0000445894823771101-0.0000476658633986675I, -5.27113436102362*10^-6+0.0000223986687502278I, -0.0000150041710133919-0.000065635257889951I, -0.0000253221797316855+0.0000933782229239012I, 5.40170356695405*10^-6-0.0000795139510782993I, -0.0000158892683627256+0.0000235079626022212I, 0.0000179668180014822-0.0000532275481836012I, -3.82815387418837*10^-6+0.0000310787548112095I, 0.0000359103793608858-0.0000758425922656375I, -9.0184416993974*10^-7+0.0000846638938306906I, 0.0000268815282670579-0.0000699915640047703I, -0.0000238683469862889+0.0000722010775720108I, -0.000011668383320367-0.0000857547442444072I, -0.0000246767717577591+0.0000776886915078519I, 0.0000158103348947944-0.0000592855768373556I, -0.0000401533948207081+0.0000803487639256484I, 0.0000205909128180703-0.0000509923331352948I, 0.0000143119042314893+0.0000980637242964681I, -0.0000235856885831924-0.0000896299898337735I, 0.0000104095233222677+0.0000438206984715361I, 0.0000331809689882274-0.0000491228129794486I, -0.0000254833126978647+0.000082407019562038I, -0.0000112831950952924-0.000072641625346583I, -0.0000175141655920062+0.0000557786674007447I, -5.26870621819468*10^-6-0.000074124946987815I, 5.25624956078201*10^-6+0.0000927741054477246I, 6.20975297751699*10^-6-0.0000819615595269612I, -0.0000164378639615094+0.0000503443361607403I, -0.0000297874156472868-0.0000548326297013343I, 0.0000141345085985378+0.0000586969023975284I, -6.96606848812398*10^-6-0.0000541725799287019I, 6.8653013886508*10^-7+0.0000436305493696336I, 0.0000266469384539391-0.0000699794759845963I, 0.0000119774855352068+0.0000558111692798724I, -2.95771494534767*10^-6-0.0000734373135390788I, -0.0000158652871529987+0.0000481758194360618I, -0.0000263106407453073-0.0000761388469344666I, 8.24625045141186*10^-6+0.0000380634434426448I, 0.0000192105574264351-0.0000499123462390058I, -7.97507381811798*10^-7+0.0000454685078238981I, -1.52987186583697*10^-6-0.0000743388624048901I, 0.000028559375388843+0.0000713975272710365I, -0.0000144186449378121-0.0000675695257241359I, -0.0000104934450863356+0.0000910643299720031I, -7.94032079483939*10^-7-0.0000462615643353353I, -5.25946849976807*10^-6+0.0000780312671418413I, 0.0000130845105724251-0.0000691373685896387I, -0.0000255148937828385+0.0000741010000784744I, -0.000016215208106686-0.0000285154501834802I, -4.51377490685616*10^-6+0.0000768160487504719I, 0.0000313025396948025-0.0000644502574742136I, -6.35787850563498*10^-6+0.0000308235919053171I, -0.0000227384980782379-0.0000652833661887987I, -0.0000149864329981165+0.000110727419608254I, 0.0000164141682378575-0.0000844170214318526I, -0.0000301115077184991+0.0000353273774930711I, 0.0000323527555244923-0.0000434232290195713I, -0.0000249710670978717+0.0000282146418177227I, 0.000037397933708117-0.0000911222956296717I, 3.28477123952921*10^-6+0.0000932400444481354I, 0.0000357563257343952-0.0000806365593787275I, -0.0000285080347256491+0.0000791305862008476I, -0.0000256189419570666-0.0000875767536635109I, -0.0000357985426093044+0.0000847455718574949I, 8.16942724475624*10^-6-0.0000673416265332372I, -0.0000545455726365234+0.000123763849189616I, 0.0000208348173086183-0.0000365606760359867I, 0.0000588938138459489+0.000129218849205744I, -0.0000680583041328266-0.0000803841690063677I, 8.09421280989411*10^-6+6.06337958689619*10^-6I, 0.0000724566974761999-0.0000266392483164858I, -0.0000472705503581054+0.0000763021166567556I, -0.0000259732906377193-0.000084077733148229I, -0.0000947758601469796+0.0000595777183645516I, -0.0000446958636159904-0.000109955624005217I, -0.0000199816934516154+0.000115697863579583I, -2.61544436629108*10^-6-0.0000597277695432557I, -0.0000262478560406601+0.000049930455023292I, -0.0000932851473111544-0.0000727042001565555I, 7.82684774724815*10^-6+0.000042333085693746I, 2.78107517081236*10^-6+3.75305404744342*10^-7I, 0.000032342468079355+0.0000309621043008729I, 0.0000627472796093251-0.0000776116280323962I, 0.0000122770401911144+0.000109995494223762I, 0.000056942192993213-0.0000895191880880356I, -0.0000208683151888109-0.0000251908293790475I, -0.0000255563406526595-0.000136076601102644I, 0.0000314834317630333+9.33996142043753*10^-6I, 0.0000661339893899228-0.0000274007362056099I, 0.0000535514854962316+0.0000801145189648546I, 0.0000412590402084688-0.000119260719954987I, 0.0000218492830373451+0.0000735140148282632I, 0.000011448761354356-0.0000418192970676983I, -0.000103298035597651+0.0000778703274477485I, -5.80824775956337*10^-6-0.000102193367227058I, 0.0000685945702785978+0.0000623983409997299I, -0.000015458288163438+0.0000447481302749474I, -0.0000581027331785934+0.000081218487973218I, -0.0000368847410241872+0.0000307322905520613I, -0.0000325274293284406+0.0000483072070330799I, -0.0000989340514571173-0.0000181059660954958I, 0.00010341136659951+0.0000680424908306082I, -0.0000394005663331504-0.0000964975441455575I, -9.09457674980589*10^-6+0.0000808470178582661I, 0.000092494509742022-0.0000801700989414928I, 0.0000722525152588229+0.000115498299193102I, 0.0000213939151939032+7.37451593544741*10^-7I, 7.4580698521745*10^-6+0.000105240982103353I, -0.0000398062878141209+1.25797800886329*10^-6I, -0.0000316151627546305+0.000051225386617907I, -0.000027956309532521-0.0000221438404249105I, 0.0000232943957353321+0.0000117559447922492I, 0.0000425836843841084-0.0000313200267021558I, 0.0000132938517101133+7.36869726023137*10^-6I, -0.0000500064327467468-0.0000511118875520922I, 0.0000718098476341088+0.0000198340801096978I, -0.0000428422408075554-0.0000590279357277168I, -0.0000256887482103242+4.53946751592212*10^-6I, 0.0000524349640698506-0.000016560684347623I, -0.0000195100728852686+0.0000969815208962596I, -0.0000447017599071042-0.0000853863323986005I, 0.0000276447583357341+0.0000359372986720688I, 0.0000121227339125724-0.0000469201761969287I, 0.0000333497026593601+0.0000916330151690595I, 9.38132839364276*10^-7-0.0000346242514175177I, 9.18048773492143*10^-6+0.0000381388371960465I, -1.41633476006664*10^-6-0.0000525070341738279I, 9.9434796340408*10^-6+0.0000778104209205407I, 0.0000224038460460071-0.0000360366199990911I, -0.0000120089617610504+0.0000634956413789998I, 0.0000202401402482258-0.0000701834527990837I, -0.0000156669856150564+0.0000607900848393839I, -0.000011808060977023-0.0000849389760034787I, 9.77111991069124*10^-6+0.0000546232100676646I, -9.62833021274816*10^-6-0.0000786098763586491I, -0.0000304735875624626+0.0000716703528911548I, -0.000030110960773805-0.000052796599106374I, -0.0000216277868044047+0.000044309897231449I, 0.0000279412182370452-0.0000838331880344225I, 8.70500641648615*10^-6+0.0000456304040474876I, -0.0000144621942451572-0.0000905376885662899I, 0.000014692082041505+0.0000486745330920725I, 7.91668247747197*10^-6-0.0000562661331153436I, -0.0000181737441439015+0.000112945546158906I, -0.0000130639494210789-0.0000713803284648485I, 0.0000289940015445821+0.0000384782659304645I, 0.0000391213241479968-0.0000126078330020126I, -0.0000193459226770514+0.000103936890434968I, 0.0000287458769057634+0.0000131221534103843I, -0.0000203055131011542+0.0000728145449749507I, -0.0000314689756279399+2.63666410004461*10^-6I, 0.0000539905076461293+0.0000121657725326513I, -0.000047132985371447-0.0000600449115277203I, 0.0000108235874263602+0.0000902565726669091I, 0.0000562243998405888-0.000129701482418395I, 0.0000692619406846271+0.0000635362132911575I, 0.0000600689878073233-0.0000261037778444303I, 0.0000196641739716455+0.0000901991821589268I, 4.34827343000013*10^-6+6.61501575290656*10^-6I, -0.000028833972589571+0.0000744231320045948I, 1.43590336893005*10^-6-0.0000136911178536878I, -0.0000114711267346096+6.0154670175679*10^-6I, 0.0000577140172221306-0.0000594524573720536I, -0.0000247314303972629+4.27076100682722*10^-6I, -0.0000437355280187381-0.000015337417554573I, 0.000044817790825449-0.0000259060651130676I, -0.0000423711849915914-0.0000250405522590448I, -0.0000676554989750618+0.0000182156173101176I, 0.0000976263639284826-0.0000525184507850369I, 5.83780520014123*10^-6+0.000117215070263135I, -0.0000756877612425997-0.0000482388737699111I, 8.98175831138722*10^-6+5.00126147268894*10^-6I, 0.0000321104350088355-0.0000553128099565544I, 0.0000762450945983298+0.0000604072949881798I, 0.0000393320101241206-0.0000274340086963328I, -0.0000167816326435862+0.0000183916846637331I, 0.0000122734136286873-0.0000515310286117766I, 0.0000341591843983775+0.0000648467028171723I, 0.0000717318481897919-0.0000715085393411801I, -7.48040333152843*10^-6+0.0000836392406674619I, -4.10981637232121*10^-6-0.000109123240296906I, -0.0000182484393142588+0.0000889908097015575I, -0.0000454298905022863-0.0000493293992340226I, -0.000010667385892273+0.0000303301227707306I, -0.0000385355813785646-0.0000433312929230836I, 0.0000117341505712334+0.000121541849277611I, 8.54027942607943*10^-6-9.39716452186564*10^-6I, -0.0000273081820800397+0.0000895286372876808I, -0.000027413853376808-0.0000913224785673131I, -0.0000357196984448448+0.000046293035224145I, -0.000032503204456915-0.0000386040594768932I, -1.04619889749282*10^-6+0.0000419130697500093I, -7.37567291713229*10^-6-0.0000660570918440525I, 0.0000607848043821355+0.0000830186021080081I, 3.23430470723558*10^-6-0.0000397735205698592I, -0.0000194768065748407+0.0000538630531877859I, 0.0000347771408943703-0.000101361951259294I, 0.0000223738912486877+0.0000902695519308556I, 0.0000574419240425462-0.0000502711902996604I, 0.0000105890500563484+0.0000835037081221776I, 0.0000469772972114243-0.0000503984746767114I, -0.0000405691025585897+0.0000105585105949697I, -0.0000220002920022774-0.0000237078788165328I, 0.0000364601853747973+0.0000918934120585094I, -0.0000228259264657629-0.000132648056972436I, -4.30692318130532*10^-6+0.0000275883748775739I, 0.0000687510803132765-0.0000893143528721156I, -5.75951131987325*10^-6+0.000060850998537809I, 0.0000405391139161727-0.0000630743029817554I, 0.0000116255897854806+0.0000941134944143946I, 0.0000331410929836831-0.0000633672821990087I, -0.0000414091947323753+0.0000603572795537106I, 0.0000124075532674151-0.000090124694097164I, -0.000051923015962269+0.0000720503137339858I, 3.34134723560157*10^-6-5.6129449789082*10^-6I, -9.96015676960352*10^-6+0.0000272656775446673I, 0.0000241704218879755-7.22491108761561*10^-6I, -0.0000170922352986757+0.0000610612748212798I, 0.0000513780832428616-0.0000927220152369655I, 0.0000112697510545527+0.000076319199992882I, 4.08902813785428*10^-6-0.0000260796611364367I, -0.0000539766155783667+0.0000514534804731489I, 7.48048654055209*10^-6-0.0000740616602619853I, 0.0000406591012434958+0.0000636164792444704I, 0.0000215085426864852-0.0000433637963453025I, -0.0000203556940829154+0.0000866971096927765I, 0.0000401546313221333-0.0000747832415366095I, 0.0000133415292766166+0.0000308939499315617I, 4.44673507230362*10^-6-0.0000754358883518484I, 0.0000168516868166428+0.0000858358028426088I, -0.0000180635165853192-0.0000945829106727073I, 4.10109563292341*10^-6+0.0000545992816979729I, 3.83589546201805*10^-6-0.0000336801789758865I, 0.0000153365396605988+0.000036877392583952I, -0.0000254804033030746-0.0000595864637062648I, 7.45513234680918*10^-6+0.0000758333805179942I, -7.47742581545554*10^-6-0.0000599522025306747I, 0.0000156472949447949+0.0000638350569029223I, -0.0000172528206395988-0.0000439663492061044I, -0.0000165087869869919+0.0000311715443109615I, -0.0000107434140191551-0.0000496995171931975I, 0.0000243703713189412+0.0000849386692609853I, -0.0000236720757493083-0.0000820123520016829I, 9.6460336296499*10^-6+0.0000838653646073307I, -6.11081289896048*10^-6-0.0000237755912276178I, -0.0000371075948639427+0.0000826394381334934I, 0.0000237517876537574-0.0000906681694686039I, -0.0000297145298726253+0.0000590237081847404I, -0.0000364922799094912-0.0000304612443284481I, -7.94981445279024*10^-6+0.0000818280789377499I, 0.0000577143706942859-0.0000502022387441892I, -1.6484486566653*10^-6+5.28397059599342*10^-6I, -0.0000379201234730991-0.0000764222973589213I, -0.0000489272789672718+0.000131251900474214I, 0.0000528674549459388-0.0000925113764812894I, -0.0000215979168171944+3.90621108417615*10^-6I, 0.0000512727157852942+9.54802045759005*10^-6I, -0.0000169289553079118-1.03107400395916*10^-6I, 0.0000677655640480078-0.0000532295118476996I, -0.0000291933520660299+0.000110727857977205I, 0.0000591013206733025-0.0000379613170941021I, -0.0000489359954830679+0.0000345613692953134I, 0.000012594022484022-0.000100627768282902I, -0.0000713633656456222+0.0000226383693082744I, -0.0000344130329654449-0.0000466053484064629I, -0.0000759391502194058+0.0000673012538251747I, -0.0000486929176444155+0.000012431771435922I, -2.92363471757351*10^-6+0.000158748300027886I, -6.94704974983038*10^-6-0.0000928593255982274I, 0.0000472871317546491+0.0000311132398125903I, -0.0000182980362163203+0.0000480486936610495I, -0.0000411290444050457-0.0000349831742328202I, 0.0000202691058708271-0.0000944964074379031I, -0.0000655367017317482-5.47086666352629*10^-6I, 9.12565266776782*10^-6-0.000119710002055927I, -0.0000974144862207029+0.0000352178991956045I, -6.51069199454345*10^-6-0.0000141609883686986I, 0.0000335342789455373+0.0000492994160510136I, 7.06643631239926*10^-6-0.000167995057207243I, -6.84652555698259*10^-6+0.0000688370092548351I, -0.000036143007028062-0.0000469419691930429I, 0.0000420504440859424+0.00010838324789298I, -0.0000137264986206238-6.44702124062801*10^-6I, -0.0000538106833412089+0.0000985829842278718I, 0.0000356388414328377-6.3060452324682*10^-6I, 0.0000838569501021867+0.0000520769295354303I, 0.0000827824909018817-0.0000602668865263424I, 0.0000219027324124842+0.000108776917643239I, -0.0000379042371425929-0.0000387486945851789I, -0.0000184732641871345+0.000121730093897599I, 0.0000314841946420052-0.0000191411262830143I, -0.00003262333758836+0.0000465289139229586I, -3.85462055262111*10^-6-0.0000390029185091914I, 0.0000284913265031502-0.000021826577098044I, 0.000038620715275524-0.0000718675699083691I, 0.0000106669147807218+0.0000924724102947935I, -0.0000657570096070926-0.0000898902721497938I, 0.0000235887518390147+0.0000468697703197585I, 7.51466057146193*10^-6-0.0000876036535400732I, 0.0000265412671211802+0.0000463498317857762I, -0.0000143296260059372-0.000134637918083076I, -0.0000482376050171417+0.000113547131530621I, 0.0000157528888725361-0.0000413191317445731I, 0.0000516065194902474+0.0000634475303450351I, -0.0000305332539591126-0.0000588182352785724I, -0.0000464250850700498+0.000103928094017655I, 0.0000178834131409046-0.0000977247293336409I, -0.0000454674017839506+0.0000842392011259998I, -0.0000229328034060718-0.000107729013568177I, 0.0000351796931554521+0.0000627376470818322I, -4.7612358541184*10^-6-0.0000996807522693526I, 3.40070805777822*10^-6+0.0000914062204314451I, -0.0000237043557393059-0.0000501807619675236I, -4.01974993433419*10^-6+0.000098617462580418I, -4.17185480074451*10^-6-0.0000463957547624695I, 0.0000332578690249819+0.000107654076849718I, 0.0000311095549249688-0.0000392079053274989I, 0.0000581773832993382+0.0000366637275232309I, -3.08676905308714*10^-6-0.0000336415299088926I, -0.0000284832412817472+0.0000447244983571926I, 0.0000423201749333884-0.0000628935495397862I, -0.000040255844357467+0.0000746162061975941I, -0.0000117195099225764-0.0000638879218031142I, -0.0000146162605566899+0.000107683594805616I, -0.0000233648568548344-0.0000488705080134837I, -0.000010716517440656+0.0000979491220258703I, 0.0000278277565144854-0.000055332823516843I, 3.64272224637132*10^-6+0.0000455181911450343I, -0.0000281377315116344-0.0000609220617729647I, 3.13791344756706*10^-6+0.0000709941324765531I, 2.18516093673874*10^-6-0.000062898464395443I, 0.0000140290402719711+0.0000467948392772144I, 0.0000123546704717074-0.0000460882218226059I, 0.0000149318191893168+0.0000656601408417089I, 3.49257035492881*10^-6-0.00006484526029023I, 4.5716135776212*10^-6+0.0000467082059352465I, 2.46537258787311*10^-6-0.0000766739799766687I, 0.0000189682958514899+0.0000586588154981132I, -5.36440773538011*10^-6-0.0000386104035766202I, -9.5340781265432*10^-8+0.000057184965491312I, 0.0000103110005854929-0.0000536490629016072I, 0.0000106546944051378+0.0000750905714645335I, -0.0000152656873772414-0.0000630115709572534I, -6.95470087499147*10^-6+0.000049584277735312I, 0.0000110381715471538-0.0000460602491369327I, -8.22133546358129*10^-6+0.00009124790632692I, -0.0000236811349285557-0.0000739053986879312I, -0.0000261783942027356+0.0000514263158283069I, -0.000044733063139505-0.0000343679345394868I, 2.06880477183337*10^-6+0.0000621316953448604I, 0.0000105807470693696-0.000101443320472452I, -8.1262514897364*10^-6+0.0000640352905828498I, -0.000021748095495849-0.0000772215735272209I, -0.0000150985211860734+0.000122120216296066I, 0.0000478289606577539-0.0000608473295625281I, -0.0000522018317638938+0.0000730567690946666I, 0.0000373170235388432-5.79842035622001*10^-6I, -0.0000633125464236101+0.0000342529525995338I, 4.21940041432195*10^-6-0.0000983164583240477I, 1.58227362336915*10^-6+0.0000889325081403241I, 0.0000206766006742448-0.0000821048056737641I, -7.30172035211635*10^-6+0.0000672839642078724I, -0.0000246896242383528-0.0000507727042879741I, -0.0000275476450057342+0.000059143412840339I, -0.0000572859633405933-0.0000744727542233225I, 0.0000137285274209439+0.000167928222889708I, -0.000028439461666551-0.0000113298832779836I, 0.000115700808109523+0.0000877078058716712I, -0.0000535654002752075+0.0000232985005784202I, -0.0000378701280033221-2.1496678622644*10^-6I, 0.0000500748896286166-0.0000329546664218899I, -0.0000377243803699572+9.98746568836622*10^-6I, -0.0000152265159527378-0.0000669289337501511I, -0.000126062334145443+0.000139766505615811I, -0.000089653670501967-0.00007811198466526I, -0.0000625236971168629+0.0000803204718622344I, 8.24755545430182*10^-6+0.0000232663159475201I, 0.0000247440159818565+0.0000785671893362076I, -0.0000582847149699577-0.0000714446581422711I, -0.0000449453795291712+0.0000606625505653025I, 0.0000614704465545129-0.0000201936849566627I, 0.000041432465744596+0.0000571542036750398I, -0.000013479840766852-0.0000693286732586108I, 0.0000212760279419347+0.000139221164046633I, 0.0000542939050387417-0.0000984573237345305I, -0.0000197889284315697+0.0000168082558703122I, 0.0000123969101661949-0.0000905586487502662I, -9.72609933227466*10^-6+0.000069911238514715I, 0.0000101663309444717-0.0000973336884793413I, 0.0000627280589701154+0.0000814885846867352I, -1.84893146720228*10^-6-0.0000924228261915093I, -0.0000528084291809054+0.0000482242257304089I, 0.0000522409341000396-0.0000525046695117214I, -0.0000279083825445405+0.0000649742466314742I, -0.0000300269711181934-0.000106172778381988I, 0.0000336816024031319-0.0000162092582441414I, 0.000040626445193198-2.69504142954269*10^-6I, 0.0000412952876316152+0.0000785714420190854I, 0.0000823865336608588-0.0000722118256480798I, -9.68505134840497*10^-6+0.0000538088662894988I, -0.0000359032653473454+0.0000185449480900444I, 0.0000293618199749657+0.0000379160679481402I, 5.18880975443052*10^-7-0.0000255426267665162I, 0.0000440035437986881+4.06078457713662*10^-6I, -0.0000329157888125694-0.000115742498256768I, 0.000080167377399788+0.0000476417162836739I, 0.0000312875337354312-0.000107730528000724I, 0.0000532567493810508+0.000102095054994849I, 0.0000155451902738322-6.91009221979044*10^-6I, -2.40774396136227*10^-6+0.0000612488264538993I, 8.6574066709463*10^-6-0.0000307369967249163I, -0.0000266065726483503+0.0000329798661613212I, 0.0000424472422344217-0.0000830907516644664I, -0.0000365650731876377+0.0000465783466968892I, 6.76673640652664*10^-6+0.000010592136806304I, 1.051746934791*10^-6-0.000039038046886537I, 0.0000128421530543818-6.16608614381922*10^-6I, -0.0000699869723323417+0.0000449707401722455I, 0.0000868733329918936-0.000109638719301941I, 0.0000275668077393487+0.0000981174663770256I, -0.0000254332368727189-2.44693052596678*10^-6I, -0.0000544607678051214+0.0000242894324833594I, 0.0000210953703950649-0.0000796852424980415I, 0.0000525513850597602+0.0000349780900362552I, 0.0000365908050390387-0.0000535587798858153I, -0.0000551192100175345+0.0000708859338435402I, 0.0000447860267987592-0.0000531426798320267I, 0.0000331337987636084+0.0000203114897343125I, 8.74152577549532*10^-6-0.00013009227627013I, 0.0000129100993935007+0.0000942136901817317I, -0.0000345539092439677-0.0000887121127903491I, 0.0000349882816469577+0.0000781217663537852I, -5.29516573259649*10^-6+0.0000112335180874377I, -9.84201475248364*10^-6+0.000070537581271941I, 0.0000111059381049661-0.0000290143913757597I, 0.0000682770411688809+0.0000549478227810606I, 0.0000666927216788828-0.000079093383930125I, 0.0000452599949661768+0.000106911801602123I, -0.0000586706397419493-0.0000189314181270809I, -4.6203606977875*10^-6+0.0000879602488532541I, 0.0000301766818793025-0.0000159409048520004I, -0.000032085659237572+0.000089980249112435I, -5.8856305236717*10^-6-0.0000672285123526142I, 6.00177618379689*10^-6-0.0000265148808993938I, 0.0000155606557359235-0.000057512324128662I, -0.000028393726444649+0.0000902412949817265I, -0.0000637522158761066-0.0000944949860979728I, 0.0000128480391855914-4.26180430971948*10^-7I, -0.0000527809264550686-0.000132130960165468I, 0.000016584025779756+0.0000361592187628657I, -4.01771302005505*10^-6-0.000127748507833097I, -0.0000283638682931962+0.000130775553475613I, 0.0000409875511927457-0.0000751377584586555I, -0.000012504067892332+0.000032091001665553I, -0.0000201562501708401+0.0000108087679727603I, -0.0000498727249623676+0.0000894082488670884I, -0.0000489162664848205-0.0000654875063252741I, -0.0000270984438553753+0.0000529467095425971I, -0.0000301860375391628-0.000102650818407388I, 0.0000170368975529969+0.0000420801095920184I, -0.0000163661615704359-0.0000897247817974865I, 0.0000245241022259911+0.0000879590591945895I, -0.0000333904275792456-0.000038614359065737I, 0.0000259024058042839+0.000082827182250201I, -5.17361211118023*10^-6-0.0000909353500100321I, 0.0000149565761918747+0.000132054277281358I, 4.48661903890847*10^-6-0.0000717533570175577I, 0.0000691863373117354+0.0000579805728374063I, -0.0000496889869089189-0.0000121884443660357I, -0.0000295377254471647+0.0000237122396154083I, 0.0000395421848198175-0.0000737044677381527I, -4.17224359077319*10^-6+0.0000656359484594711I, -0.0000166121975399667-0.0000558364874502571I, -0.0000647390651367132+0.000106372903362476I, -0.0000519545606991961-0.000065728243784929I, -0.0000115858373834632+0.0000762207727417378I, 1.19136028221487*10^-7-0.0000235975972712684I, 2.34449422325402*10^-6+0.0000706420702579956I, -0.0000401440206949967-0.0000527611841939168I, -0.0000220166557180713+0.0000538054147954035I, 0.0000355222436806241-0.0000311252847947212I, 0.000021490246858969+0.0000508895946013209I, -1.27831191931496*10^-6-0.0000669960796126556I, 0.0000142855309255073+0.000105044553410726I, 0.0000357758629422386-0.0000824577912942705I, -0.0000119858698012539+0.0000265078520000133I, 0.0000101750131233164-0.0000837997250637016I, -3.56422128965763*10^-7+0.0000636726362287921I, 4.34854107987034*10^-6-0.0000746706211325444I, 0.0000420073468327582+0.000077594265090875I, 7.13790190005726*10^-6-0.0000758590142225852I, -0.0000385778575062188+0.0000560015043265347I, 0.0000348217770044693-0.0000532623537083257I, -0.0000232824988121519+0.0000425559020726297I, -0.000016174745063693-0.0000939296585207787I, 0.0000205897689816995+0.0000141645009564424I, 5.51416501404914*10^-6-0.0000234558869989304I, 0.0000328666571813427+0.0000524302487424537I, 0.0000396233729500561-0.0000972311730422128I, -1.64246171695224*10^-6+0.0000430077697693979I, -0.0000362299577438675-0.0000282022765980581I, 0.0000141698036892298+0.0000801833197861715I, 0.000027624384188063-0.0000421812155270981I, 0.0000266297691966379-0.0000107518091894238I, -0.0000446422179568459-0.0000510289156128048I, 0.0000425157431530374+0.0000699192127725243I, -0.0000281846133737359-0.000110633108120231I, 0.0000393024762575992+0.000094030577878271I, -0.0000184179771955622-0.0000398066669416065I, 9.13062707084196*10^-6+0.0000329901141836019I, -0.0000203016846115986-0.0000545530279891242I, 0.0000100953043595893+0.0000609456426180779I, 8.33446885364389*10^-6-0.0000563366131452257I, 0.0000185130461709937+0.0000702068581688993I, 0.000028834213219391-0.0000449248848567089I, 9.40320977379291*10^-6+0.0000149992284617882I, 0.0000257942159332481-0.0000757735774028524I, -0.0000343690673900341+0.0000330601261988214I, 0.000018856770262806-0.000083539394991295I, 5.27207311983087*10^-6+0.0000775924551054574I, -2.68122387912947*10^-7-0.0000702657455662544I, 9.28675758345814*10^-7+0.000085329195926408I, -1.51995782222056*10^-6-0.0000571825968899068I, 0.0000293760931158403+0.0000450325365492221I, 0.0000144188781446154-0.000055337564541218I, 0.0000160803122827754+0.0000667679197072721I, 6.56466583751602*10^-7-0.0000776577378059949I, -0.0000107953497157451+0.0000495015842178162I, -7.2982023122123*10^-6-0.0000570307828834361I, 0.0000111661436695179+0.0000536666071255956I, -0.0000117861526051348-0.0000500549136651716I, -7.0742443870203*10^-8+0.0000525874234292578I, 0.0000158951940263579-0.0000589185826249212I, -5.58006204425289*10^-6+0.0000573630852938689I, 9.73705779641448*10^-6-0.0000526723742511054I, 3.54255671009242*10^-6+0.0000419497872111929I, -6.04035662753808*10^-7-0.0000689230156898868I, 7.06683370584199*10^-6+0.0000451944181686066I, 0.0000184397600206427-0.0000451906486321087I, 2.86391779313401*10^-6+0.0000718348812820861I, 0.0000164061284431827-0.0000669728034396911I, 7.86571748117162*10^-6+0.0000486672959967787I, -1.27436963675204*10^-7-0.0000392777573376582I, -0.0000201848647623426+0.0000508374962633375I, 0.0000144136642038357-0.0000753672023512283I, 0.0000371650576806177+0.0000673170671574164I, -0.0000229130711472228-0.0000234293313876244I, -0.0000130920026231361+0.0000774523667188836I, 0.0000108703116839173-0.0000102309032638212I, -1.58206108580741*10^-6+0.0000509172833800033I, -0.0000536116038046782-0.0000642196664819797I, 0.0000210828946845529+0.0000708529033499151I, -0.0000217629827073803-0.0000495223287779978I, 0.0000457564752312345+0.0000759014706474739I, 0.0000252642361310528-0.0000965315916540088I, 0.0000217896413870839+0.000111280901743032I, 0.0000498654405434138-0.0000610137511949192I, -0.000023632355229367+0.000111631824407715I, -0.0000365165058486595-0.0000452884893260224I, 0.0000106864264338353+0.00005835005580845I, -0.0000187642698571629-0.0000596249836455945I, 8.900366667367*10^-6+0.0000446438676548578I, 0.0000202014450127528-0.0000314032174742367I, -5.92222601488031*10^-6+0.0000525111649996027I, -0.0000428849191037394-0.0000127436309018294I, 0.0000971618148122561+0.0000451378061224653I, 2.24749856784366*10^-6-0.0000177057730604236I, 0.0000233671317131666-0.0000295637444792215I, 0.000074363424078872+8.43577947819849*10^-6I, -0.0000486746027826736+0.0000861883315324996I, -3.03478617409309*10^-6-0.0000755403784574818I, -0.0000341931922441285+0.0000390948784842126I, 5.28828351855405*10^-6-0.0000477442968877727I, 0.0000382947272131839+0.000157972526988072I, -0.0000246817836581322+2.62847921643235*10^-6I, -0.0000210779146472194+0.0000803051775231143I, 0.0000576676032824202-0.0000235889936491469I, 0.0000387453762518103+0.0000501295367835979I, 2.79142360326303*10^-6-0.0000429701271451365I, -0.0000203146053881245+0.0000961120128631033I, 0.0000412003653428276-0.0000986695764640746I, 0.0000141508180982013+0.0000493027450433628I, -0.0000325942190186068-0.0000361211673184118I, 0.0000457397031607318+0.0000712723601377585I, -8.54427840369016*10^-6-0.0000901937654444231I, -0.0000114494593441126+0.0000718770609095206I, 0.0000107398612352065-0.0000635622457787596I, 1.80645779583981*10^-6+0.0000917429743504844I, -0.0000433124076979204-0.0000813484943112871I, 0.0000291369518474828+0.000038177407171613I, -0.00001998473820084-0.0000512990285899746I, -0.0000450231523459497+0.0000835689587343789I, 0.0000276317137915346-0.0000866131109529346I, 8.77951452064289*10^-6+0.0000507126390789161I, -0.0000360795807327162-0.0000606523720093637I, -0.0000392761895193319+6.63750413389232*10^-6I, 0.0000359801542065357-0.0000700623705295726I, 0.000044349041095348+0.0000261233913523025I, 0.0000259840741478462-0.000141565835544305I, -6.31258367355081*10^-7+0.0000528028498787918I, 0.0000236064751439821-0.0000165555747564302I, -7.45315090180561*10^-6+0.0000545907982033128I, 0.0000371883981899838-0.0000449602521557359I, -0.0000118821128676535-0.0000155018933620065I, -0.0000578837299094775-0.0000385534432710587I, 0.00002865252144426+0.0000188496080739774I, -0.0000402067394321754-0.000102317576126487I, 0.0000541025768028832+0.0000538390636885486I, 0.000017019408653117-0.0000389392972009286I, -5.49856394715027*10^-7+0.0000433181731526723I, 1.92061883016801*10^-6-0.0000454375964660796I, -8.65429258369398*10^-6+0.000057673536713276I, 9.83207242433713*10^-6-0.0000767160835778954I, -3.19029836095659*10^-7+0.0000625400960941268I, 0.0000353313078639465-0.0000419197102474448I, -0.0000329317097734417+2.46751861911744*10^-6I, 0.0000171001186930625-0.0000604120462003589I, -0.0000534069558928982+0.000067339685947635I, 0.0000229942652435863-0.000107833800102259I, 0.0000274506877888131+0.0000750376884791582I, -0.0000133386469611761-0.000037443772118338I, -9.4509716438918*10^-6+0.0000593329417000318I, 8.41391016862609*10^-6-0.0000654322874153096I, 0.0000302107924341585+0.0000201129556019888I, 0.0000273830616559638-0.0000617355097013204I, -0.0000115970804470388+0.0000556863630165973I, 8.61786647839695*10^-6-0.0000641057311306532I, 6.95754505724633*10^-6+0.0000417602958873491I, 9.95759935138239*10^-6-0.0000881879906328246I, 8.10034639436175*10^-6+0.0000664093551897291I, -0.0000190702598912662-0.000066545586229119I, 0.0000113877854040725+0.0000698656348624258I, -6.64102600000185*10^-6-0.0000288660295333725I, -0.0000129342420384511+0.0000617646141147992I, 6.73758005705946*10^-6-0.0000378639622851503I, 0.0000340824870659184+0.000059802052886997I, 0.0000379501170049167-0.0000564695995044103I, 0.0000115313253180756+0.0000853596142462834I, -0.0000305047332109855-0.0000519773292423966I, -3.34805738990186*10^-6+0.0000796950951861862I, 9.23451525738954*10^-6-0.0000332100541360212I, -0.0000314581183561227+0.0000564252549704549I, 0.0000102427132731008-0.0000541143059672351I, 0.0000205457670909089+8.64732780144667*10^-6I, 7.35250474984291*10^-6-0.0000693387249278599I, -6.01775506667915*10^-6+0.0000486085367501711I, -0.0000224984928733216-0.000077178461309693I, 0.0000444808454267192+0.0000368283386109011I, 0.0000364272708859557-0.000127017919410951I, 0.000019193193255315+0.0000427486629289775I, -5.25190442653332*10^-6-0.0000797039137588483I, -0.0000407812408933699+0.0000983877773986122I, 0.000046193363632955-0.000023583962314226I, 0.0000488388783966544-0.0000117158633885921I, -0.0000916349036587233-0.0000416235848595695I, -3.53549023981891*10^-6+0.0000710318668371409I, -0.0000308418062006024-0.000157517987723636I, 0.0000195573529338207+0.0000938555201745763I, -0.000020060127377118-0.0000820416277867318I, 0.0000416675972377961+0.000030752303598611I, -0.0000201286549031155-0.0000890321253504629I, 6.8284350655872*10^-6+0.000097329101622403I, -0.0000176784274667147-0.0000600767299126067I, 0.0000212802807774882+0.000119696486489136I, 0.0000662112604346226-0.0000306355008566149I, 0.0000102770471468379+0.000030073319752764I, 0.000086387898643079-0.0000799170442481517I, -0.0000150439807547649+0.0000131172974669966I, 0.0000177433968200015-0.000103796383600741I, -5.57886911325769*10^-6+0.0000750761073523919I, 0.0000467733120726088-0.0000887510383662355I, -0.0000334358082647648+0.000145004620857688I, -0.000018156206674374-0.0000576785968876426I, 0.0000610496898666794+0.0000659263428521699I, 0.0000178311104957744-0.000032119079010975I, 0.0000456303374904145+0.000121387018554384I, 0.0000362246981762429-0.000118244647550544I, -0.0000297799984275241+0.0000176995083729032I, -0.0000418665746472112-0.0000305539545854616I, 0.0000389132102532252+0.0000629033758928722I, -0.0000231533185928199-0.0000631609097919298I, 9.75235871084507*10^-7+0.0000144675195482573I, 0.0000523702931968901-0.0000462420461848594I, 0.0000301723792852012+0.0000351521839172969I, -2.9615915870606*10^-6-0.0000706885955218756I, -0.0000104629280029384+0.0000214425309435262I, -0.0000255119096566795-0.000102340717676097I, 0.0000419622120218901+0.0000226194506330629I, 0.0000179236289110736-4.45707944521408*10^-6I, 0.0000131890127191202+0.0000415469478206437I, 0.0000278729769026172-0.0000704150777247061I, 0.0000313614833199361+0.0000936066514029612I, -0.0000153201658425385-0.0000609163822878304I, -0.0000591037914522668+0.0000504454198740065I, 9.82624482608172*10^-7-0.0000572431132453445I, 8.87865076331335*10^-6+0.0000875761556195199I, -0.0000331128487632105-0.0000194128505959759I, -0.0000483560497263352+0.0000357727035263135I, -0.0000703579631138863-0.000024733257158222I, -0.0000153832780814935+0.000041783271736055I, -0.0000520560183999939-0.0000634796106865441I, 0.0000520375060748254+0.0000874546238881415I, -6.4988160762953*10^-6-0.0000910431536856305I, -0.0000226625477242435+0.0000603984350477424I, 0.0000503166372343758-0.0000287290154750665I, 0.0000219586631627814+0.000103768130382137I, -0.0000168744794112519-0.0000239078624880111I, -2.92404842272423*10^-6+0.0000855146207931934I, -0.0000493879046542091-0.0000301759627155714I, -9.97442654189265*10^-6+0.0000320239580720106I, -0.0000390170169067284-0.0000425209668261144I, 0.0000350291583367158+0.0000365889057460786I, 0.0000207167272572528-0.0000190616927116652I, 0.0000362048133608267+0.0000352771579762785I, -0.0000212319282138365-0.0000497518889646725I, 0.0000820602894010664+0.0000216820171095573I, -7.18196554766597*10^-6-0.0000773161850115692I, -0.0000250709490123815-0.0000325761386572696I, 0.0000649076820570422-0.0000167716618897875I, -0.0000315251964480125+0.000106467160845667I, -0.0000387476329729077-0.000113666359203126I, 0.0000287713924693276+0.0000729068611928044I, 0.0000104920368869657-0.0000310448504623354I, 0.0000969580170712705+0.000100904166103417I, 0.000026318435892642+9.4788848686417*10^-6I, 0.0000442347357724992+0.0000452356059674057I, 0.0000233301695016764-0.0000815380054253365I, 7.76179334933945*10^-6+0.0000561984597428903I, 0.0000522611290905611-6.9017561951727*10^-6I, 6.30888863712541*10^-6+0.0000750724770433456I, 3.80644203798755*10^-6-0.000112739151075604I, -0.0000361251551994454+0.0000484454492447199I, -0.0000230576758763432-0.0000583934136930035I, 0.000021591571646849-2.42407467390374*10^-6I, -0.0000597501759328941-0.0000586112055994536I, -0.0000184027315047588+0.000113308907181921I, -0.0000255731354511289-0.0000147718648418616I, -0.0000163224617419558+0.0000600923998102039I, -0.0000214129537859933-0.0000661005749668146I, -0.0000382954856606283+0.0000111867730911437I, -0.0000322354029318541-0.000034141044818681I, 0.000025560363097554+0.000060399166782958I, -0.0000267712838774839-0.0000736682137545097I, 0.0000490617923319706+0.00008510066298499I, 5.09032445814279*10^-6-0.0000124752967063624I, -0.0000409549135013886+0.0000775218946344473I, 0.000017453992884811-0.000113661843990236I, -5.20363785619814*10^-6+0.0000644117390023544I, 3.73578470261614*10^-7-0.0000424576514278524I, 0.0000111652292745339+0.0000770465277509513I, 0.0000603982879882797-0.0000862859257807387I, -0.0000492557252758815+0.0000228424486690212I, -0.0000198696861965197-0.00003954409058215I, 0.0000114480992999647+0.000114415338379979I, -4.30461944070976*10^-6-0.0000925892988305526I, -0.0000551868815805999+0.0000390805006120624I, 0.0000587285575128619-0.0000618029162100482I, -0.0000499995584043234+0.0000393022106513857I, 0.0000210808714204257-0.000102632936655651I, 0.0000254882808272217+0.0000893990772528416I, 0.0000257757566612852-0.0000920742950551985I, -0.0000251295450467039+0.0000816245646301856I, -0.000022773716603046-0.0000620873412181819I, -0.0000374335313353639+0.0000940649107492223I, -0.0000131640454065257-0.0000216020012141801I, 0.0000278853734610558+0.000111653746325765I, 0.0000415646790453231-2.91511153352276*10^-6I, 0.0000794077466740088+0.000029467260360679I, 0.0000136669463729538-0.0000149373097121685I, -0.0000437497482895415+0.0000372938587538885I, 0.0000661412371257394-0.0000591558861532047I, -0.0000771228887048245+0.0000739231955091727I, -0.0000184058200990992-0.0000581597440516023I, -3.50187280286302*10^-6+0.000165376274999131I, -0.0000472437102267719-5.67786021550408*10^-6I, -0.0000129580102841405+0.000134121314272613I, 0.0000775931569069114-0.0000443653641327133I, 0.0000162878549701668+0.0000270074897046703I, -0.0000514636330436214-0.0000201877079198661I, -1.53254105355476*10^-6+0.0000938265287564823I, 0.0000402787087292252-0.0000911419195443929I, 0.0000197070970840403+9.22125850214288*10^-6I, 7.07226002270852*10^-6-0.0000369934999216881I, 0.0000819309104050537+0.0000593503693896082I, -0.0000172854895682335-0.000116216522952462I, -0.0000403395993132841+0.0000457084616175151I, -0.0000307004864191579-0.0000929662486383357I, 0.0000259951863880315+0.000056896238217128I, -0.0000295887890145078-0.0000510292388348221I, 0.0000385588312577602-4.65103566932625*10^-6I, -0.0000222677597428046-0.0000634915441208473I, -6.35222629949818*10^-6+0.000121204034984873I, 5.36419050599181*10^-6-0.000110633472530647I, -0.0000210005183372854+0.0000853282273223515I, -0.0000526709200831979-0.0000242104914125596I, -0.0000745294617528816+0.0000277769440822569I, 0.0000558941759039146-0.0000807988347984803I, -4.84798081901439*10^-6+7.37320386504494*10^-6I, -0.000049019295021644-0.000114824011518884I, -0.0000166406477845109+0.0000654336899120031I, 0.0000751219790156668-9.81003486380915*10^-6I, 5.98972042758471*10^-6+0.0000151472550013336I, 0.0000142667868044657-0.0000794711203804947I, -0.00008982229136163+0.0000404273105360791I, -5.20721475808827*10^-6-0.0000216634165880812I, -9.20820094618053*10^-6-0.0000321116197005685I, -0.0000344314728147276-0.0000281631047233853I, 0.0000331658000211402-0.000019905101953288I, 0.0000643992637720799-0.0000549992521246239I, -0.0000214795832840844+0.0000729283621910592I, 0.0000424503609619894-0.0000445073775551279I, -0.0000335506865358611+0.0000568262016905094I, -0.0000108415583051209-0.0000982556706094018I, -0.0000335839156665447+0.0000554319125441257I, 0.0000272369233857912-0.0000616912608304627I, -0.0000994560170900532+0.0000627688987867208I, 8.03078018246511*10^-6-0.0000335576280125179I, -4.61087570862157*10^-6+0.000144141677176812I, -0.0000385335277528773-0.000115318439487385I, 0.0000404516849261838+0.0000214325325820408I, 0.000033559933419537+4.8008123347756*10^-6I, -0.0000440292953961685+0.0000480386060798769I, -7.07416019693426*10^-6-0.0000814344056930977I, -0.0000454999787042201+0.0000147813637027898I, -4.49144704140203*10^-6-0.0000909893849014735I, -0.0000366650526787603+0.0000917684430413062I, 0.0000197585583365056-0.0000571198464718042I, -3.62518456876143*10^-6+0.0000245866102947466I, -0.000053306445893862-0.0000948495656357207I, 0.0000163100696931425+0.0000625467716229249I, -0.0000182780336311835-0.000041725973085385I, 0.0000363090899275844+0.0000397218977518839I, 0.0000362132031976451-0.0000203204926364719I, 0.0000152481018328606+0.0000792083019569086I, 0.0000258089205244146-0.0000632046874082449I, 0.0000187386591388474+7.66499788467225*10^-6I, 0.0000130490184434998-0.000111763400020344I, 0.0000511348973816948+0.0000528384215096173I, -0.0000118410045752783-7.40180585980918*10^-6I, 0.0000290839650059613+0.0000560922533066265I, 0.0000291141872408962-0.0000561325618507877I, -2.90539414669482*10^-6+0.000105582582239195I, -3.06875351131751*10^-6-0.0000747991833353863I, -0.0000506227543248202+0.0000240652541397897I, -0.0000140707553748186-0.0000512329123086599I, -0.0000250300526150028+0.0000709216852911149I, -0.0000229088284450321-0.0000483966189752386I, -0.0000263210064014627+1.84331953849251*10^-7I, -0.0000861283078201922-0.0000901072618806026I, -0.0000119431926718709+0.0000404387214003927I, 1.10070608765867*10^-6-0.0000619894842856448I, 0.0000274607253962999+0.0000830891848708933I, 0.0000194699379387942-0.0000986480172947995I, -0.000071531992267387+0.0000400612473519849I, 0.0000242798913114397+0.0000132805756869984I, -0.0000144807074133758+0.0000438912693831166I, -0.0000550570105925399-0.0000127630825191532I, 5.45963604968013*10^-6+0.0000166350240123165I, 2.7477100128909*10^-6-0.0000604679101704393I, -0.0000163047083029892+0.000046521798377607I, 1.92723986259135*10^-6-0.0000529097955377631I, 8.83797345928149*10^-6+0.0000575565873901321I, -0.000015955384616841-0.0000558329267402488I, 9.92546446344683*10^-6+0.000050672929276853I, 3.8927296831453*10^-6-0.0000860568665891596I, -0.0000340511080363732+0.0000916669568917626I, -4.98136582912516*10^-6-0.0000638120092934942I, 0.0000256598791079773+0.0000930144135174016I, -0.0000463932098306526-0.0000558288316644324I, 3.71131488567409*10^-6+0.0000242659952893152I, 0.0000312693407608183-0.0000443470859789098I, -0.0000102889413505427+0.0000607587282546952I, -0.0000134921199627832-0.000061997737188987I, -0.0000491021794639895+0.0000591268339429152I, -0.0000277710045232088-0.000071688197086637I, -9.46092282181547*10^-6+0.0000837986941559887I, 9.69267738827151*10^-6-0.0000489287201795647I, -9.0220411659423*10^-6+0.0000453726450785792I, -0.0000549122103908998-0.0000538840749883257I, 2.41494993345166*10^-6+0.0000533501152237112I, 0.0000112435236849282-0.0000337230380763515I, 0.0000235221684294473+0.000028268844469717I, 0.0000326175026356013-0.0000503167844703376I, 0.0000309780596619549+0.0000819840599006423I, 0.0000258413518042838-0.0000875432749132457I, -0.0000179551502471745+3.90161929481618*10^-6I, -0.000016447854595397-0.000112836106921135I, 0.0000337485695435781+0.0000334235970534556I, 0.0000136225396632497-0.0000269400952260078I, 0.0000488415084308798+0.0000420789090155691I, 0.0000186243021905341-0.0000841663411820152I, 1.14519007908824*10^-6+0.000100214724798865I, 0.0000110402878917849-0.0000745291039227946I, -0.0000748444283861778+0.0000612477776633591I, -0.0000341073551129024-0.000062751388974557I, -7.8454895454231*10^-6+0.0000410056141361881I, 0.0000115331029596551-5.56175593415874*10^-6I, -0.0000357412573694275+0.0000179960910704286I, -0.0000733514532919757-0.0000582958835567247I, -0.0000310608447473072+0.0000433895816608323I, -0.0000207551024920159-4.98973376697791*10^-7I, 0.0000769969182962597+0.0000518979852137539I, 8.55137173438279*10^-9-0.000106868628934312I, -0.000087652557533396+0.0000335937403051521I, 0.0000593745894464221-8.23540106521322*10^-6I, 0.0000469706845161634+0.0000298766136510908I, -0.0000472710905341912+0.0000193881983808514I, 0.0000449379867354129+0.0000224387177985848I, 0.00001298721553423+3.45380082952378*10^-6I, -0.0000507013038602893+0.0000452460416622653I, 3.10579353209033*10^-6-2.19146834717837*10^-6I, 3.00130075474986*10^-6+8.98505927303136*10^-6I, 0.0000313288885753055-0.0000622387951512011I, -7.63915867387541*10^-7-6.67985694629983*10^-6I, -0.000015804893137798-0.0000675804528822701I, -0.0000405517148954737+5.20830502796287*10^-6I, -0.0000538782776423215-0.0000498666581337001I, -0.0000620627574078156+0.00009698778858664I, 0.0000162607336684791-0.0000849900078789174I, 0.0000409293932254665+0.0000661938022152476I, -0.0000519466785513996-6.13181920331945*10^-6I, 0.0000131506045170194-2.42975077187155*10^-6I, 0.0000226078671057455-0.0000664439011753719I, 4.05087775104983*10^-6-7.69148325786789*10^-6I, 0.0000307845510497035-0.0000771952533443229I, -0.0000434477453135331+0.0000218626670465342I, 3.69825027108476*10^-7-0.0000394999190118288I, 0.0000224195224679115+0.0000435201689210671I, 0.0000204227389308539-0.0001165819244659I, 3.09033262135404*10^-6+0.0000682811308027091I, -0.0000290084836101651-0.0000672739804243738I, 0.0000237602713195687+0.0000798809308055157I, -7.99685625403661*10^-6-2.33977583885043*10^-6I, -0.000018385883569408+0.0000655289364863919I, 0.0000109447487180544-0.0000267091548129879I, 0.0000575615435797045+0.0000512460161542769I, 0.0000563492995634961-0.0000724302531215164I, 0.0000391220722334414+0.0000946528518722945I, -0.0000537341337906195-0.0000161841764253647I, 3.69917839063348*10^-7+0.0000748509086621909I, 0.0000272803221398193-0.0000167772957828077I, -0.000029398069588513+0.0000957053473963432I, -5.01803792269132*10^-6-0.0000748465955497465I, -7.06600221187292*10^-6-0.0000241385373534788I, -8.19837732219036*10^-7-0.000045677087187264I, -0.0000488488626259271+0.0000805967655794146I, -0.0000539281587307905-0.000090805360676195I, 3.72600864479345*10^-6-0.0000338424077632208I, -0.000100485749839033-0.000162364236809181I, 5.64753008197466*10^-6+0.0000271246387048085I, 0.0000144751416394854-0.000106209267687886I, -7.0152213006416*10^-6+0.000131993273947857I, 0.0000662343101187158-0.000108088488485443I, -0.0000904414977918201-5.21158532986541*10^-6I, -0.0000114364967528471+0.0000932153551239867I, -0.0000481641171407393+0.0000483611087418884I, -0.00014024105351152-0.0000217302290355766I, 0.0000135077245295959-4.4688712513002*10^-6I, -0.0000180407818860897-0.0000827372746194061I, -0.0000157747524597314+0.0000135567254716456I, -0.0000213130148431458-0.000060367256328417I, 0.0000427346947348631+0.0000763980766611948I, -0.0000418271858345565-0.0000411442549235898I, 0.0000656055735207472+0.0000527071873304359I, 0.000030392421626417-0.000159712084688865I, -0.0000730793761761604+0.000121880377871984I, -0.0000285534074755678-0.000143117206133675I, 0.0000159634572179184+0.000132081226641463I, -0.000126854602948414-0.0000663633867619362I, 0.0000244008361715772+3.97315586539148*10^-6I, 6.31746396015778*10^-6-0.0000712174252188978I, 0.0000791480688632481+0.0000746312143268261I, -0.0000142560636753993-0.0000497462648184991I, -0.0000885906910642317-0.0000242747904810327I, -0.0000113035231480292-0.0001236164790631I, 0.0000289628630377745+0.0000134069830780235I, -0.0000838576136559596-0.0000584456248621094I, -0.0000420163816379933+0.0000885795923054612I, -0.0000169025271815147-0.0000558762884321081I, -3.26419287043593*10^-6-1.36270080452877*10^-6I, -0.000011699762286047+0.0000180047657875243I, -7.39433499323524*10^-6+0.000064417112487744I, 0.0000256270917472358-0.0000935454600443914I, -0.0000484534243237374+0.0000759897496586398I, 0.0000405561639599056-0.0000322338333851345I, 2.77147386654727*10^-6+0.0000162529601217824I, 8.13694576808296*10^-7-0.0000601763637713524I, -0.0000104997044127449+0.0000216578971763134I, 0.0000544202242747341-0.0000457872155569033I, 2.35529663118909*10^-6+0.000100308843708698I, 0.0000321471030738141-0.0000747154903248358I, 0.0000110740057235091+0.000013151549000031I, 3.66640745214238*10^-6-8.06405230535789*10^-6I, -0.00003116032614071+0.0000359682838366766I, 0.0000323701411438901-0.0000915851707082682I, 0.0000753242228024153+0.0000711284138203064I, -0.0000521115256514481+2.91654226667019*10^-6I, -0.0000165155319713203+0.0000824963031973879I, 8.56327683240461*10^-6+5.37218490084857*10^-6I, 1.75909800774411*10^-6+0.0000353576223690113I, -0.0000901654171434014-0.0000842359014514688I, 0.0000276981952231205+0.000102290974344855I, -9.53796065422589*10^-6-0.0000529640544868373I, 0.0000561173899147512+0.0000582482094461567I, 0.0000248936509603119-0.0000569488560700668I, 0.000014411020849039+0.000145930105137264I, 0.0000109350573764213-0.0000607768765771249I, -0.0000319574053702457+0.00012351474664161I, -0.0000785981563316694-0.000054831696765263I, 0.0000180699938194206+0.0000260997711973303I, -0.0000535694169322345-0.0000691233775398632I, 0.000047149623708884+0.0000553951992276168I, 1.33563436179556*10^-6+4.04472156449444*10^-6I, 0.0000481898319287138+0.00006020575962296I, -0.0000317879419407013-0.0000527808057750229I, 0.000127729708501703+0.0000729252761969518I, 8.15637535822046*10^-6-0.0000794091795562436I, 0.0000430600430453549-0.000067789045785243I, 0.0000360477770358792+0.0000456370334214428I, -0.0000758708288190487+0.0000811784875288239I, 3.63553830623425*10^-6-0.000148629828598617I, 0.0000251869713702106+0.0000917489460084344I, -0.0000101338377742198-0.00001611346692272I, 0.0000513375101927428+0.000177432775519818I, -0.0000339398891450078+0.0000291321640175777I, 0.0000576701976893756+0.000068441249888815I, 0.000029911220037102-0.0000522142569970115I, 4.06510291048646*10^-6+0.000072300273543463I, 0.0000118826704353445+0.0000481149088971555I, -0.0000188859391841326+0.0000679994873834209I, 0.0000650810752508109-0.0000941976740770487I, -0.0000338702725828237+0.0000172670970206427I, -0.0000269412082285217-0.0000922165410822039I, 0.0000696611562158155+0.0000231458509175776I, -0.0000502394455862584-0.000116639917008162I, -0.0000825935124808797+0.0000929855300011574I, -0.0000676642272692728-0.0000308272210432221I, -0.0000353734525322155+0.000041229031828656I, -5.67067846100303*10^-6-0.000104688103701144I, 0.0000155226231888153-0.0000259361881342877I, -0.0000626514446528184-0.0000718584904201387I, 4.39178887081635*10^-6+0.000074102389375904I, 0.0000105248058995294-0.000100455172690165I, 0.0000220900122359182+0.000139170153266015I, -0.0000576868502754415-0.0000157993138807682I, -0.000067825826477496-4.31732565620255*10^-6I, 0.00010750276624364-0.0000895589454216714I, 0.0000178319863928932+0.0000484879755526276I, 0.0000187075136451036-0.0000995656223945227I, -0.0000133368347218323+0.0000831570921202377I, 0.000112956873966346+0.000024460740869959I, -0.000017909058434146-0.0000359233375477228I, 8.71871708311455*10^-6-0.0000495540937398634I, -0.0000748991769003756+0.0000376993705347706I, -0.0000391953285354169-0.000047447646593911I, -6.50064560050605*10^-6-0.0000802588104259929I, -0.0000315988719270196-0.0000499828705248884I, 0.0000545774186043846-0.0000455981432869517I, 0.000104530540244375-0.0000577355049648713I, -0.0000215611937766953+0.000080175075847754I, 0.0000662151649369165-0.0000422013315264175I, -0.0000546346081047319+0.0000659737210686773I, -0.0000258223231995028-0.000126738499168867I, -0.0000420605672304316+0.0000657702261034829I, 0.0000716235985845234-0.0000874915142436254I, -0.00019119759124108+0.0000680712247477874I, 0.0000157238811327429-0.0000616470408856432I, -0.0000200615878701953+0.000229475025460468I, -0.00012430188264063-0.000180632890352887I, 0.0000884732821936102-0.0000196170824651545I, 0.0000540131144543631+0.00002944753436397I, -0.0000149574409505163+0.0000689732426841528I, -0.000021166587804414-0.0000918994290224446I, -0.000122001732417679-0.0000923113283016993I, -8.22184390869497*10^-7-0.000173420518988852I, -0.0000160894382143123+0.000070364602298424I, -0.0000637547502085507-0.0000794536364253174I, -0.0000583235111585682+0.0000419327758101413I, -0.0000889994704666668-0.000103869382670716I, 0.0000281738501462484-5.08232082146341*10^-6I, -0.0000465887336223163+0.0000552568853496503I, 0.000027589937073973+0.0000417032826203907I, 0.0000886021789187492-0.0000775990423715527I, -0.0000461914498210298+0.0000985032457786787I, 0.0000780606494061011-0.0000377341120416394I, 6.32920023561249*10^-6-0.0000532786973696132I, -0.0000169134401511418-0.000119097814781904I, 0.0000279053987375749-0.0000221610514357991I, 0.000103699887972091-3.13244537829329*10^-6I, 0.0000317369556200772+0.000105423839851508I, 0.0000615054642552345-0.000114946988440581I, 0.0000449034299184703+0.0000251176053675318I, 1.3633087952907*10^-6+5.17473830886577*10^-6I, -0.000105717922421298+0.0000662220054050246I, 0.000025750859942531-0.000106644119946305I, 0.000122019640734706+0.0000738892993186619I, -0.0000346015304920162+0.0000802846608481821I, -0.0000715002539159083+0.000117312591926515I, 1.61434758532476*10^-6+0.000114541939340954I, -0.0000286092620878285+0.0000430675047883909I, -0.000129514951700314-0.000013216110442638I, 0.000103457438991131+0.0000424623729366956I, -0.0000763526622950354-0.0000582437474510928I, 0.0000537401476765971+0.000113308635911949I, 0.000102455833402342-0.000148764386001313I, 0.000084902643894531+0.000127023342757267I, 0.000106611848814561-8.62666943055276*10^-6I, -0.0000149092820167752+0.00013056988150535I, -0.0000278352230497969+0.0000159553399107818I, -0.0000162703377701614+0.000068689029696447I, -6.20185398110673*10^-6-0.0000162416014610741I, -0.0000100623194016555-4.01088052827452*10^-6I, 0.0000596595412634072-0.0000412570309215464I, -0.0000420862827177722+8.86552959060614*10^-6I, -0.0000573829598578265+0.0000319999900763205I, 0.0000834204787255336-0.0000219954826184528I, -8.89211333791469*10^-6+0.0000300376844108066I, -0.000028411342115859-0.0000127310623393199I, 0.000114040743916559-0.000041058528092859I, -0.000010332382163147+0.0000954762431232359I, -0.0000154634249574796-8.13343035637067*10^-6I, -0.000072815267239147+7.17333023233872*10^-6I, 0.000015768331811982-0.0000688797011487134I, 0.0000498889179716799+0.0000880859940716122I, 0.0000111537643647189-0.0000255942482192392I, -0.0000586459808994478+0.0000938652805586862I, 0.000072403039129559-0.0000432325779908052I, 0.000033973443234114+3.54698994805921*10^-6I, -0.0000318836450855107-0.000105957500285462I, 0.0000167504323670262+0.0000954243246005085I, -0.0000233403704258693-0.0000755276507687241I, 0.0000484994748921525+0.0000399828028760191I, 0.0000318590304307104+0.0000177707027958019I, 0.0000258940973917266+0.0000758546198474873I, 0.0000287262172960794-0.0000553357343114971I, 0.0000436320754270416-1.77155053298613*10^-6I, 0.0000355713730584289-0.000134967767931825I, 0.0000829928806854706+0.0000615047442129164I, -0.0000230636158148762+0.0000317654151799438I, 0.0000395564603401412+0.0000679968566824417I, 0.0000635454239602376-0.0000512807542203421I, -4.15895969501986*10^-6+0.000127990031289989I, -7.29149682202205*10^-6-0.0000664611972186552I, -0.000100517966316922-0.0000206804083457677I, -2.7793210160968*10^-6-0.0000682028610553876I, -4.65121194884968*10^-7+0.000099060516726298I, -0.0000853206646968905+5.62598629608658*10^-6I, -0.0000675790281321663-0.0000288254187495841I, -0.000164720896399126-0.0000556561360403965I, -0.0000262809272100296+9.16109955629271*10^-6I, -0.0000989030581759384-0.0000803806074923034I, 0.000103203066437261+0.000147389510731158I, 0.0000125983317678292-0.00014924694646584I, -0.0000927039832709964+0.0000367790082550262I, 0.000116018961676254+0.0000647935694246338I, 0.0000252225109286546+0.000136774260307711I, -0.0000859204660543937+0.0000631672363565304I, -5.80650279181888*10^-6+0.0000668748602902449I, -0.0000909619443512542-6.25898728142004*10^-6I, -0.0000456795418357057+1.83827065874507*10^-6I, -0.0000685956227457256-0.0000149386011027383I, 0.0000791836039246571+5.82826522121786*10^-6I, 0.0000252276885308521+0.0000223332472132374I, 0.0000780164203003631-0.0000212668680250135I, -0.0000722208710168478-0.000104842749218927I, 0.000106433897702853+0.000056198496031264I, -0.0000810317633077036-0.000100902011430494I, -1.52598190946542*10^-6-0.0000318432481988157I, 0.0000300444867438157+0.0000552327847079588I, -0.0000479798000555737+0.0000870825858925541I, -0.0000682257974461189-0.000123450358133491I, 0.000082000152265903+0.0000232710098985226I, 0.0000142439358386649-0.0000149046910210287I, 0.0000391886465375585+0.00011554495688686I, -9.52506579204108*10^-6+1.48389740790308*10^-6I, 0.0000297398280307329+2.65390193306944*10^-6I, -3.46385204550574*10^-6-0.0000301152231293734I, 0.0000192063423435526+0.0000864695116074576I, 0.0000485192720303115+2.32225654607425*10^-6I, -0.000025584355029396+0.0000605722286144658I, 0.0000404995663498373-0.0000854530342554446I, -0.0000278598041329467+0.0000503045165182358I, -0.0000345613787205503-0.0000725764706089694I, 0.0000287284524602058+0.0000305593848925169I, -0.0000332741267670802-0.0000795584065812587I, -0.000038958790571972+0.0000871219349184204I, -0.0000317126731704573-0.0000341755689777445I, -0.0000202389254920193+0.0000507557506102615I, -5.34975130084284*10^-6-0.0000735052267827202I, 6.46816608043241*10^-6+9.3622886680975*10^-6I, -0.0000261887995816346-0.0000625237102692282I, 8.70765736483046*10^-6+0.0000711481740919545I, -8.88801468141635*10^-7-0.000077941602307638I, -9.41153902199053*10^-6+0.0000940447434251799I, -0.0000292867745422312-0.0000315552422683554I, -0.0000305600912682357+0.0000393156010589427I, 0.0000376168785061988-0.000045816880888677I, -0.0000281967826918562+0.0000390936197909985I, -0.0000533743321221339-0.0000359759362625066I, -0.0000226599645605635+0.0000603728088223824I, 0.0000202634316276573-0.000011985503833823I, 0.0000465609810449219+0.0000170220355089809I, -0.0000224507552142133-0.0000985277896729905I, -0.0000858326158016394+0.0000819150624868875I, 0.0000822491348313738-0.0000280230218431883I, 0.0000137808742113023+0.00001749376738084I, -0.0000120440931648907+0.0000554877259915199I, 7.8195706511457*10^-6-1.64983765823361*10^-6I, 0.0000232101029888099-0.0000128478196984009I, -0.0000540171209131456+0.0000650745996400496I, 0.0000169084086596356-6.5773035890484*10^-6I, -1.66058585739553*10^-6+4.46026370707856*10^-6I, 0.0000213160946126007-0.0000508605577788942I, -0.0000129695447609023-0.0000219935619786455I, -0.0000594603143413806-0.0000859193130801459I, -0.0000285617424107205+0.0000654019082743646I, -0.0000926434943021593-0.0000330430826230526I, -7.55760306800345*10^-8+0.000109112738983693I, -0.0000219671704490945-0.0000169840242754749I, 0.0000165481777573559+0.0000307468578880967I, -0.0000459550411336319-5.05995068464414*10^-6I, 0.0000281385440864274-0.0000432389736084892I, 0.0000177007466278646-0.0000569226112510687I, -0.0000739254259732853+0.0000309231128222499I, -0.0000225142372063959-0.0000877395774861556I, -0.0000625956457475339-2.47822261509877*10^-6I, -0.0000273269918375319+0.0000171596753769106I, 0.0000312927460004707+0.0000788550451477373I, 9.59818168612127*10^-6-0.0000992821176698465I, -0.0000358410348865076+0.0000479122273578279I, 0.0000197295861198553-0.0000277194954386197I, 0.0000261105095229061+0.0000848493927079057I, -0.0000272691212876976-0.0000424908717332778I, -0.0000215512634564457+0.000104813216577185I, 0.0000409748485708352-0.0000469614414941984I, 0.0000225407357196291+0.0000342421491979524I, 0.0000404195544925812-0.000064359798752325I, -1.42095735290538*10^-6+0.0000776883050679653I, -7.62099883178394*10^-6-0.0000638962575244528I, 0.0000328437813151028+0.0000890028818501744I, 0.0000169934259844732-0.0000593677355653461I, -0.0000444997989548967+0.0000469274766591486I, 0.0000320593183852619-0.0000421183335955001I, -0.0000295160883324311+0.0000174049403105282I, -0.0000106884337258632-0.0000906314734645597I, 0.0000264432279345381+0.000020549918883749I, -0.0000144810430210505-7.44415784894156*10^-7I, 0.0000126432115307715+0.0000349410392679088I, -4.78085187276577*10^-6-0.0000709796952232278I, -0.0000112685606533028+0.0000280045016568889I, -0.0000677464344099793-0.0000222288197252802I, 0.0000558556319326506+0.000088615684063835I, 0.000013804036274592-0.000074333067119998I, -0.0000130138139554713+8.67258692782017*10^-8I, 0.0000310444499090782-0.0000167136317913982I, 0.0000604409153994552+0.0000885229537788563I, -0.000041536671476207-0.0000202497211970053I, 0.0000335320075231759+0.0000830387515364071I, -0.0000451056166577068+6.46687440555582*10^-6I, -0.000030239779088709+0.0000180450608207042I, 0.0000377882729331524+7.43937360878239*10^-6I, 0.0000376667662720921-9.02961555985216*10^-6I}"),

  getSextuplePointTri: function() {
    return {coeffs: RotorAnimations.SextuplePointTriCoeffs, size: 2.0, freqfn: function(n) {return -2*bount(n+1)/8.0;},
            xofffactor: 0.5, yofffactor: 1.75
           };
  },

  end: ""
}
